# -*- coding: utf-8 -*-

from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import
from builtins import super

from .multiplyForm import multiplyForm
from .coForm import coForm
from ..transform.graphTransform import graphTransform3D

class areal(multiplyForm):
    """
    An areal form specifies a form with an areal operational behavior,
    applicable to 2D planar segments, and using a boundary
    classification approach to determine overlapping segment parts.
    An areal form is a collection of co-forms, where each co-form is a co-equal
    collection of areal elements. An areal form is maximal if no each
    of its co-forms is maximal.

    The Areal class extends on the :class:`~sortal.form.multiplyForm.multiplyForm` class and
    specifies the areal operational behavior of forms.
    """

    type = 'form-areal'
    
    def __init__(self, sort, *args, **kwargs):
        super(areal, self).__init__(sort)
        
    def toString(self, string = '', printKey = False):
        """
        get the form in string format
        """
        if len(self) == 0:
            return '{}'
        string += '  '
        s = '{\n' + string
        printComma = False
        for i in self.elements:
            if printComma:
                s += ',\n' + string
            else:
                printComma = True
            s += i.toString(string, printKey)
        s += '\n'+ string[:-2]+ '}'
        return s

    def oneLineString(self, printKey = False):
        """
        get this form in one-line string format
        """
        s = '{'
        printComma = False
        for i in self.elements:
            if printComma:
                s += ', '
            else:
                printComma = True
            s+= i.oneLineString(printKey)
        s += '}'
        return s
        
    def printE(self, string = ''):
        """
        print out the form for easier assessment

        :param string: prefix string to be added to the beginning of all printed lines
        :type string: str
        """
        print(string + '|' + self.sort.name+' ('+str(self.sort.baseSort().characteristic.type)+'):',len(self.elements),'elements')
        for ind in self.elements:
            ind.printE(string+'   ')
        return
        
    def rhinoString(self, dX = 0, dY = 0):
        """
        get the string format of the form that can be pasted into Rhinoceros for visualization

        :param dX: x-coordinate offset from (0,0)
        :type dX: float
        :param dY: y-coordinate offset from (0,0)
        :type dY: float
        """
        s = ''
        for cf in self.elements:
            s += cf.rhinoString(dX, dY)
        return s
        
    def totalLength(self):
        """
        The total number of individuals in this form
        """
        length = 0
        for i in self:
            length += i.totalLength()
        return length
        
    def accept(self, visitor, direction=-1):
        """
        Go through all the elements in this form and perform visitEnter and visitLeave 
        functions of the visitor on each of the element.

        :param visitor: visitor to perform visitEnter and visitLeave functions with
        :param direction: any extra information for the visitor's function
        """
        if visitor.visitEnter(self, direction) and not self.isEmpty():
            for element in self.elements:
                element.accept(visitor, direction)
        visitor.visitLeave(self)
        return
                
    def edit(self, editor):
        """
        transforming the form with a transformation set(editor).

        :param editor: the transformation set to be applied
        :type editor: ~sortal.transform.transformSet.transformSet
        """
        result = areal(self.sort)
        for element in self.elements:
            a = element.edit(editor)
            result.insert(a)
        if editor.current is None:
            transf = editor.retrieveTransform(graphTransform3D)
            if transf is not None:
                for i in transf.extra:
                    result.insert(i)
        return result
        
    def insert(self, plane):
        """
        Inserting a plane form (coForm or areal) or a plane segment into the form 
        at appropriate location in the element list. This function will reset the form's 
        maximal status to False.

        :param element: element to be added
        :type element: sortal.element.element
        """
        if plane.type == 'form-areal':
            for i in plane:
                self.insert(i)
            return
        elif plane.type == 'form-coForm':
            if not self.sort == plane.sort:
                raise ValueError('Inserting an object of sort '+plane.sort.name+' into a form of sort '+self.sort.name)
            self.maximal = False
            thisCoForm = self.first()
            index = 0
            while thisCoForm is not None:
                if thisCoForm.compareValue(plane) == 0:
                    for i in plane.elements:
                        thisCoForm.insert(i)
                    return
                elif thisCoForm.compareValue(plane) == 1:
                    self.add(index, plane)
                    return
                else: 
                    index = index+1
                    thisCoForm = self.get(index)  
            self.add(index, plane)
            return
        else:
            if not self.sort == plane.sort:
                raise ValueError('not of the same sort')
            self.maximal = False
            thisCoForm = self.first()
            index = 0
            while thisCoForm is not None:
                if thisCoForm.compareValuePlane(plane) == 0:
                    thisCoForm.insert(plane)
                    return
                elif thisCoForm.compareValuePlane(plane) == 1:
                    newcoForm = coForm(self.sort)
                    newcoForm.insert(plane)
                    self.add(index, newcoForm)
                    return
                else: 
                    index = index+1
                    thisCoForm = self.get(index)      
            newcoForm = coForm(self.sort)
            newcoForm.insert(plane)
            self.add(index, newcoForm)

    def maximalize(self):
        """
        maximalize the form:

        * remove Nil elements
        * merge/split elements based on their attributes
        * reorder the elements in ascending order
        """
        if self.maximal:
            return
    
        self.maximal = True
        
        if len(self.elements) == 0:
            return
        
        index = 0
        disorganized = []
        while index < len(self):
            coF = self.elements[index]
            if len(coF) == 0:
                self.elements.pop(index)
                continue
            j = index + 1
            while j < len(self):
                if self.elements[j].compareValue(coF) == 0:
                    for ele in self.elements[j]:
                        coF.insert(ele)
                    self.elements.pop(j)
                else:
                    j += 1
            coF.maximalize()
            if index < len(self) - 1 and coF.compareValue(self.elements[index+1]) == 1:
                disorganized.append(coF)
                self.elements.pop(index)
                continue
            index += 1
        for i in disorganized:
            self.insert(i)
        return
    
    def sum(self, other):
        """
        Modify this form to be the sum of this form and another. 
        The other form will be unchanged.

        | e.g.  
        |       formA: a, b, c
        |       formB: b, c, d
        |   formA.sum(formB) --> 
        |       formA: a, b, c, d

        :param other: other form to find the sum with
        :type other: ~sorta.form.form.form
        """
        if self is other:
            raise ValueError('Operation on two instances of the same form')
        if not self.sort.__eq__(other.sort):
            raise ValueError('Operation on two forms of different sorts '+str(self.sort)+'  and '+str(other.sort))

        for i in other:
            self.insert(i)
        self.maximalize()
        return

        self.maximalize()
        other.maximalize()
        
        thisCoForm = self.first()
        otherCoForm = other.first()
        index = 0
        indexOther = 0
        while thisCoForm is not None and otherCoForm is not None:
            if thisCoForm.compareValue(otherCoForm) == 0:
                #if has attribute, sum
                
                thisCoForm.sum(otherCoForm)
                index += 1
                indexOther += 1
                thisCoForm = self.get(index)                            
                otherCoForm = other.get(indexOther)
            elif thisCoForm.compareValue(otherCoForm) == 1:
                self.add(index, otherCoForm)                            
                indexOther += 1
                otherCoForm = other.get(indexOther)
            else:                          
                index += 1
                thisCoForm = self.get(index)                            
        while otherCoForm is not None:
            self.add(len(self.elements), otherCoForm)                            
            indexOther += 1
            otherCoForm = other.get(indexOther)
        return

    def difference(self, other):
        """
        Modify this form to be the difference of this form and another. 
        The other form will be unchanged.
        
        | e.g.  
        |       formA: a, b, c
        |       formB: b, c, d
        |   formA.difference(formB) --> 
        |       formA: a

        :param other: other form to find the difference with
        :type other: ~sorta.form.form.form
        """
        if self is other:
            raise ValueError('Operation on two instances of the same form')
        if not self.sort.__eq__(other.sort):
            raise ValueError('Operation on two forms of different sorts '+str(self.sort)+'  and '+str(other.sort))
        
        self.maximalize()
        other.maximalize()
            
        thisIndex = 0
        otherIndex = 0
        thisCoForm = self.first()
        otherCoForm = other.first()

        while thisCoForm is not None and otherCoForm is not None:
            if thisCoForm.compareValue(otherCoForm) == 0:
                #if has attribute, calculate difference
                thisCoForm.difference(otherCoForm)
                if len(thisCoForm.elements) == 0:
                    self.pop(thisIndex)
                    thisCoForm = self.get(thisIndex)
                otherIndex += 1
                otherCoForm = other.get(otherIndex)
            elif thisCoForm.compareValue(otherCoForm) == 1:                            
                otherIndex += 1
                otherCoForm = other.get(otherIndex)
            else:                             
                thisIndex += 1
                thisCoForm = self.get(thisIndex)
        return

    def product(self, other):
        """
        Modify this form to be the product of this form and another. 
        The other form will be unchanged.

        | e.g.  
        |       formA: a, b, c
        |       formB: b, c, d
        |   formA.sum(formB) --> 
        |       formA: b, c

        :param other: other form to find the product with
        :type other: ~sorta.form.form.form

        """
        if self is other:
            raise ValueError('Operation on two instances of the same form')
        if not self.sort.__eq__(other.sort):
            raise ValueError('Operation on two forms of different sorts '+str(self.sort)+'  and '+str(other.sort))
            
        self.maximalize()
        other.maximalize()
                       
        thisIndex = 0
        otherIndex = 0
        thisCoForm = self.first()
        otherCoForm = other.first()
        while thisCoForm is not None and otherCoForm is not None:
            if thisCoForm.compareValue(otherCoForm) == 0:
                #if has attribute, product
                thisCoForm.product(otherCoForm)
                if len(thisCoForm.elements) == 0:
                    self.pop(thisIndex)
                    thisCoForm = self.get(thisIndex)
                otherIndex += 1
                otherCoForm = other.get(otherIndex)
            elif thisCoForm.compareValue(otherCoForm) == 1:                
                otherIndex += 1
                otherCoForm = other.get(otherIndex)
            else:                             
                self.pop(thisIndex)
                thisCoForm = self.get(thisIndex)
        while thisIndex < len(self.elements):   
            self.pop(thisIndex)
        return
        
    def symdifference(self, other):
        raise NotImplementedError

    def partition(self, other):
        raise NotImplementedError
        

    def containsInd(self, ind):
        """
        check if this form contains a certain individual
        """
        if self.sort == ind.sort:
            for i in self.elements:
                if i.compareValuePlane(ind) != 0:
                    continue
                if i.containsInd(ind):
                    return True
        return False
    
        
    def partOf(self, other):
        """
        check if this form is a part of other form

        :param other: other form
        :type other: ~sortal.form.form.form
        :returns: boolean
        """
        if self is other:
            raise ValueError('Operation on two instances of the same form')
        if not self.sort == other.sort:
            if other.sort.contains(self.sort):
                for i in other:
                    if i.sort == self.sort or i.sort.contains(self.sort) and self.partOf(i):
                        return True
                return False
            raise ValueError('Operation on two forms of different sorts '+str(self.sort)+'  and '+str(other.sort))
            
        self.maximalize()
        other.maximalize()
                       
        thisIndex = 0
        otherIndex = 0
        thisCoForm = self.first()
        otherCoForm = other.first()
        while thisCoForm is not None and otherCoForm is not None:
            if thisCoForm.compareValue(otherCoForm) == 0:
                if not thisCoForm.partOf(otherCoForm):
                    return False
                thisIndex += 1
                otherIndex += 1
                thisCoForm = self.get(thisIndex)
                otherCoForm = other.get(otherIndex)
                
            elif thisCoForm.compareValue(otherCoForm) == 1:
                otherIndex += 1
                otherCoForm = other.get(otherIndex)
            else:                             
                return False
        return thisCoForm is None
        
