# -*- coding: utf-8 -*-

from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import
from builtins import super

from .multiplyForm import multiplyForm
from copy import copy
from ..transform.graphTransform import graphTransform

class coForm(multiplyForm):
    """
    A co-form specifies a form of co-equal areal elements (2D or 3D plane
    segments) and is a subform of an areal form. Co-equal elements
    share the same co-descriptor or carrier (2D or 3D plane).
    A co-form uses a boundary classification approach to determine overlapping
    areal element parts. A co-form is maximal if all areal elements are
    disjoint or if they touch (but do not overlap) and their attribute forms
    are different.

    The CoForm class extends on the :class:`~sortal.form.multiplyForm.multiplyForm` class and
    specifies the areal operational behavior of subforms.

    """
    type = 'form-coForm'
    
    def __init__(self, sort, *args, **kwargs):
        super(coForm, self).__init__(sort)

    def toString(self, string = '', printKey = False):
        """
        get the form in string format
        """
        if '3D' not in self.sort.baseSort().characteristic.type:
            string += '  '
        printComma = False
        s = ''
        for i in self.elements:
            if printComma:
                s += ',\n' + string
            else:
                printComma = True
            s += i.toString(printKey)
            if i.sort.type == 'sort-attribute' and i.hasAttribute():
                s += i.attribute.oneLineString(printKey)
        if '3D' not in self.sort.baseSort().characteristic.type:
            s = '{\n' + string + s
            s += '\n'+ string[:-2]+ '}'
        return s

    def oneLineString(self, printKey = False):
        """
        get this form in one-line string format
        """
        s = '{'
        printComma = False
        for i in self.elements:
            if printComma:
                s += ', '
            else:
                printComma = True
            s += i.toString(printKey)
            if i.sort.type == 'sort-attribute' and i.hasAttribute():
                s += i.attribute.oneLineString(printKey)
        s += '}'
        if '3D' in self.sort.baseSort().characteristic.type:
            s = s[1:-1]
        return s
        
    def clearElements(self):
        self.elements = []
        
    def insert(self, ind):
        """
        Inserting a plane segment into the form at appropriate location in the element list. 
        This function will reset the form's maximal status to False.

        :param element: element to be added
        :type element: sortal.element.element
        """
        if not self.sort == ind.sort:
            raise ValueError('Inserting an object of sort '+ind.sort.name+' into a form of sort '+self.sort.name)
        if ind.type == 'form-areal':
            for i in ind:
                self.insert(i)
            return
        if ind.type == 'form-coForm':
            for i in ind:
                self.insert(i)
            return
        self.maximal = False
        self.add(0, ind)
        
        
    def insertClassification(self, ind):
        """
        insert a plane segment into the form for classification purpose. Thus there
        is no need to check for duplication or overlapping.
        """
        if not self.sort == ind.sort:
            raise ValueError('not of the same sort')
        self.maximal = False
        self.elements.insert(-1, ind)
    
    def printE(self, string = ''):
        """
        print out the form for easier assessment

        :param string: prefix string to be added to the beginning of all printed lines
        :type string: str
        """
        print(string + '|' + self.sort.name+' ('+str(self.sort.baseSort().characteristic.type)+'):',len(self.elements),'elements')
        for ind in self.elements:
            ind.printPlaneSegment(string)
            if ind.hasAttribute():
                ind.attribute.printE(string+'  ')
        return        
    
    def move(self, other):
        """
        empty the content of this form and move them into another form
        """
        for element in self.elements:
            element.assoc = other
            
        other.elements = self.elements + other.elements
        self.elements = []
        other.maximal = False
        
    def copy(self, other):
        """
        insert all content of this form into another form
        """
        for element in other.elements:
            self.insert(element)
        self.maximal = False
        
    def coEquals(self, other):
        """
        check if the plane segments in this coForm are on the same plane
        as the content of the other form
        """
        return self.first().coEquals(other.first())
        
    def compareValue(self, other):
        """
        comparing the first plane segment in the two forms
        """
        return self.first().compareValue(other.first())
        
    def compareValuePlane(self, other):
        """
        comparing the planes of the first plane segment in the two forms
        """
        if len(self) == 0:
            return -1
        return self.first().compareValue(other)
        
    def edit(self, editor):
        """
        transforming the form with a transformation set(editor).

        :param editor: the transformation set to be applied
        :type editor: ~sortal.transform.transformSet.transformSet
        """
        result = coForm(self.sort)
        for element in self.elements:
            result.insert(element.edit(editor))
        if editor.current is None:
            transf = editor.retrieveTransform(graphTransform)
            if transf is not None:
                for i in transf.extra:
                    result.insert(i)
        return result
        
    def maximalize(self, ignoreMaximal=False):
        """
        maximalize the form:

        * remove Nil elements
        * merge/split elements based on their attributes
        * reorder the elements in ascending order
        """
        if self.maximal and not ignoreMaximal:
            return
                
        keyList = []
        for plane in self:
            if plane.key is not None:
                keyList.append(plane.key)
            for bound in plane.boundaries:
                bound.maximalize()
        if len(self) <=1:
            return
        if self.sort.type == 'sort-attribute':
            result = coForm(self.sort)
            temp = coForm(self.sort)
            one = coForm(self.sort)
            two = coForm(self.sort)
            temp.insertClassification(self.elements.pop(0))
            while len(self.elements) != 0:
                elements = self.elements.pop(0)
                two.insertClassification(elements)
                #two.printE('++++++')
                #print('########################################################')
                element = two.elements[0]
                while len(temp)!=0 and not two.isEmpty():
                    #temp.printE('temp______')
                    current = temp.elements.pop(0)
                    one.insertClassification(current)
                    #one.printE('one_____')
                    #two.printE('two_____')
                    classes = element.classification(one,two)
                    if classes.disjoint():
                        #print('*************************disjoint')
                        result.insertClassification(current)
                        one.clearElements()

                    elif classes.touches():
                        #print('*************************touches')
                        #current.attribute.printE('1')
                        #element.attribute.printE('2')
                        if current.attribute == element.attribute:
                            #print('*-------> ==')
                            two.clearElements()
                            element.construct(classes.sumSet(), two)
                            element = two.elements[0]
                            for ind in two.elements:
                                ind.addAttribute(current.attribute)
                            one.clearElements()
                        else:
                            #print('*-------> not part of')
                            result.insertClassification(current)
                            one.clearElements()
                    elif current.attribute == element.attribute:
                        #print('*************************equal attribute\'s')
                        two.clearElements()
                        element.construct(classes.sumSet(), two)
                        element = two.elements[0]
                        for i in two.elements:
                            i.addAttribute(current.attribute)
                        one.clearElements()
                    elif current.attribute.partOf(element.attribute):
                        #print('*************************current attribute part of element\'s')
                        if not classes.isContained():
                            one.clearElements()
                            element.construct(classes.diffSet(), one)
                            for i in one.elements:
                                i.addAttribute(current.attribute)
                                result.insertClassification(i)
                        one.clearElements()
                    elif element.attribute.partOf(current.attribute):
                        #print('*************************element attribute part of current\'s')
                        self.insertClassification(current)
                        one.clearElements()
                        two.clearElements()
                        if not classes.contains():
                            #print('*-------> not part of')
                            element.construct(classes.otherDiffSet(), two)

                            for ind in two.elements:
                                ind.addAttribute(element.attribute)
                            element = two.first()

                        
                    else:
                        #print('*************************else')
                        if not classes.isContained():
                            #print('*------> is not contained')
                            # update result by subtracting element from current
                            one.clearElements()
                            element.construct(classes.diffSet(), one)
                            for ind in one:
                                ind.addAttribute(current.attribute)
                                result.insertClassification(ind)
                            one.clearElements()
                            # update result with the product of element and current
                            element.construct(classes.productSet(), one)
                        
                        attr = current.attribute.duplicate()
                        attr.sum(element.attribute)
                        for ind in one:
                            ind.addAttribute(attr)
                            self.insertClassification(ind)
                        one.clearElements()
                        # update element by subtracting current
                        two.clearElements()
                        if not classes.contains():
                            #print('*------> does not contain')
                            element.construct(classes.otherDiffSet(), two)
                            for ind in two.elements:
                                ind.addAttribute(element.attribute)
                            element = two.first()
                
                #result.printE('result___')
                    

                result.move(temp)
                two.move(temp)
                #print('~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~')
                
            temp.move(self)
        else:
            element = self.pop(0)
            result = coForm(self.sort)
            one = coForm(self.sort)
            result.insertClassification(element)
            while not self.isEmpty():
                element = self.pop(0)
                one.insertClassification(element)
                classes = element.classification(result, one)
                if classes.disjoint():
                    one.remove(element)
                    result.insertClassification(element)
                elif classes.isContained():
                    result.clear()
                    one.remove(element)
                    result.insertClassification(element)
                elif classes.contains():
                    one.clearElements()
                else:
                    one.clearElements()
                    result.clear()
                    element.construct(classes.sumSet(), result)
            result.move(self)
    

        for i in self.elements:
            i.assoc = self
            for j in i.boundaries:
                j.maximalize()
        
        for k in keyList:
            for i in self.elements:
                if k.ind == i:
                    i.key = k
        self.maximal = True
                
        
    def sum(self,other):
        """
        Modify this form to be the sum of this form and another. 
        The other form will be unchanged.

        | e.g.  
        |       formA: a, b, c
        |       formB: b, c, d
        |   formA.sum(formB) --> 
        |       formA: a, b, c, d

        :param other: other form to find the sum with
        :type other: ~sorta.form.form.form
        """
        if self is other:
            raise ValueError('Operation on two instances of the same form')
        if not self.sort.__eq__(other.sort):
            raise ValueError('Operation on two forms of different sorts '+str(self.sort)+'  and '+str(other.sort))
        if other.isEmpty():
            return
        if self.isEmpty():
            self.copy(other)
        for i in other:
            self.insert(i)
        self.maximalize()
        return
            
        self.maximalize()
        other.maximalize()
        otherForm = other.duplicate()
        #self.printE('+++')
        #other.printE('###')
        if self.sort.type == 'sort-attribute':
            result = coForm(self.sort)
            temp = coForm(self.sort)
            one = coForm(self.sort)
            two = coForm(self.sort)
            otherind = other.first()
            for otherind in otherForm:
                two.insertClassification(otherind)
                #print('\n-----------------------------------------------------------')
                #two.printE('>>>')
                while not self.isEmpty():
                    thisind = self.first()
                    self.remove(thisind)
                    one.insertClassification(thisind)
                    
                    #print('@===============================')
                    #one.printE('one___')
                    #two.printE('two___')
                    classes = thisind.classification(one, two)
                    if classes.disjoint():
                        #print('@@@ disjoint')
                        one.move(temp)
                    elif classes.touches():
                        #print('@@@ touches')
                        if thisind.attribute == otherind.attribute:
                            #print('@@@@@ same attribute')
                            two.clearElements()
                            thisind.construct(classes.sumSet(), two)
                            otherind = two.first()
                            for i in two.elements:
                                i.addAttribute(thisind.attribute)
                            one.clearElements()
                        else:
                            #print('@@@@@ diff attribute')
                            one.move(temp)
                    elif thisind.attribute == otherind.attribute:
                        #print('@@@ same attribute')
                        if classes.contains():
                            one.move(temp)
                            break
                        elif classes.isContained():
                            one.clearElements()
                        else:
                            two.clearElements()
                            thisind.construct(classes.sumSet(), two)
                            for i in two.elements:
                                i.addAttribute(thisind.attribute)
                            otherind = two.first()
                            one.clearElements()
                    elif thisind.attribute.partOf(otherind.attribute):
                        #print('@@@ part of attribute')
                        #update this by subtracting otherind from thisind
                        one.clearElements()
                        if not classes.isContained():
                            thisind.construct(classes.diffSet(), one)
                            while not one.isEmpty():
                                ind = one.first()
                                ind.addAttribute(thisind.attribute)
                                one.remove(ind)
                                temp.insertClassification(ind)
                        thisind.clear()
                    elif otherind.attribute.partOf(thisind.attribute):
                        #print('@@@ other part of attribute')
                        one.move(temp)
                        # update otherind by subtracting thisind
                        two.clearElements()
                        if not classes.contains():
                            otherind.construct(classes.otherDiffSet(), two)
                            for ind in two.elements:
                                ind.addAttribute(otherind.attribute)
                            #otherind.clear()
                            otherind = two.first()
                        else: 
                            #otherind.clear()
                            break
                    else: 
                        #print('@@@ else overlapping attribute')
                        # update this by subtracting otherind from thisind
                        
                        one.clearElements()
                        if not classes.isContained(): 
                            #print('@@@@@ is not contained')
                            thisind.construct(classes.diffSet(), one)
                            while not one.isEmpty():
                                ind = one.first()
                                ind.addAttribute(thisind.attribute)
                                one.remove(ind)
                                temp.insertClassification(ind)
                        # update result with the product of otherind and thisind
                        thisind.construct(classes.productSet(), one)
                        attr = thisind.attribute.duplicate()
                        attr.sum(otherind.attribute)
                        thisind.clear()
                        while not one.isEmpty():  
                            ind = one.first()
                            ind.addAttribute(attr)
                            one.remove(ind)
                            #otherForm.insertClassification(ind)
                            result.insert(ind)
                        # update otherind by subtracting thisind
                        two.clearElements()
                        
                        if not classes.contains(): 
                            #print('@@@@@ not contains')
                            otherind.construct(classes.otherDiffSet(), two)
                            for ind in two.elements:
                                ind.addAttribute(otherind.attribute)
                            
                            #otherind.clear()
                            
                            otherind = two.first()
                        else: 
                            #otherind.clear()
                            break
                #result.printE('result__')
                #result.maximalize(True)
                #result.sum(temp)
                #result.printE('result__')
                #temp.printE('temp____')
                two.move(result)
                temp.move(self)
            result.move(self)
        else:
            element = self.first()
            classes = element.classification(self, other)
            if classes.disjoint():
                self.copy(other)
            elif classes.isContained():
                self.clear()
                self.copy(other)
            elif not classes.contains():
                self.clear()
                element.construct(classes.sumSet(), self)
        self.maximal = False
        self.maximalize()
    

    def difference(self, other):
        """
        Modify this form to be the difference of this form and another. 
        The other form will be unchanged.
        
        | e.g.  
        |       formA: a, b, c
        |       formB: b, c, d
        |   formA.difference(formB) --> 
        |       formA: a

        :param other: other form to find the difference with
        :type other: ~sorta.form.form.form
        """
        if self is other:
            raise ValueError('Operation on two instances of the same form')
        if not self.sort.__eq__(other.sort):
            raise ValueError('Operation on two forms of different sorts '+str(self.sort)+'  and '+str(other.sort))
        if other.isEmpty():
            return
        if self.isEmpty():
            return

        self.maximalize()
        other.maximalize()

        if self.sort.type == 'sort-attribute':
            result = coForm(self.sort)
            one = coForm(self.sort)
            two = coForm(self.sort)
            while not self.isEmpty():
                thisind =  self.first()
                self.remove(thisind)
                one.insertClassification(thisind)
                otherind =  other.first()
                otherIndex = 0
                while otherind is not None:
                    two.insertClassification(otherind)
                    classes = thisind.classification(one, two)
                    if classes.isContained():
                        if (thisind.attribute == otherind.attribute 
                                or thisind.attribute.partOf(otherind.attribute)):
                            one.clearElements()
                        else:
                            while not one.isEmpty():
                                ind =  one.first()
                                ind.attribute.difference(otherind.attribute)
                                one.remove(ind)
                                result.insert(ind)

                        two.remove(otherind)
                        break
                    elif not classes.disjoint() and not classes.touches():
                        one.remove(thisind)
                        if (not thisind.attribute == otherind.attribute
                                and not thisind.attribute.partOf(otherind.attribute)):
                            # update result with the product of otherind and thisind
                            attr = thisind.attribute.duplicate()
                            attr.difference(otherind.attribute)
                            
                            if attr == thisind.attribute:
                                one.insertClassification(thisind)
                            else:
                                thisind.construct(classes.productSet(), one)
                                while not one.isEmpty():
                                    ind =  one.first()
                                    ind.addAttribute(attr)
                                    one.remove(ind)
                                    result.insert(ind)
                                # update one by subtracting otherind from thisind
                                thisind.construct(classes.diffSet(), one)
                                for ind in one.elements:
                                    ind.addAttribute(thisind.attribute)
                                thisind.clear()
                        else:
                            # update one by subtracting otherind from thisind
                            thisind.construct(classes.diffSet(), one)
                            for ind in one.elements:
                                ind.addAttribute(thisind.attribute)
                            thisind.clear()
                    two.remove(otherind)
                    otherIndex += 1
                    otherind = other.get(otherIndex)
                one.move(result)
            result.move(self)
        else:
            element = self.first()
            classes = element.classification(self, other)
            if classes.isContained():
                self.clear()
            elif not (classes.disjoint() or classes.touches()):
                self.clear()
                element.construct(classes.diffSet(), self)
        self.maximal = False
        
        
    def product(self, other):
        """
        Modify this form to be the product of this form and another. 
        The other form will be unchanged.

        | e.g.  
        |       formA: a, b, c
        |       formB: b, c, d
        |   formA.sum(formB) --> 
        |       formA: b, c

        :param other: other form to find the product with
        :type other: ~sorta.form.form.form

        """
        if self is other:
            raise ValueError('Operation on two instances of the same form')
        if not self.sort.__eq__(other.sort):
            raise ValueError('Operation on two forms of different sorts '+str(self.sort)+'  and '+str(other.sort))
        if other.isEmpty():
            self.clear()
            return
        if self.isEmpty():
            return
            
        differenceForm = self.duplicate()
        differenceForm.difference(other)
        self.difference(differenceForm)
        return
                
        
    def partOf(self, other):
        """
        check if this form is a part of other form

        :param other: other form
        :type other: ~sortal.form.form.form
        :returns: boolean
        """
        if self is other:
            raise ValueError('Operation on two instances of the same form')
        if not self.sort == other.sort:
            if other.sort.contains(self.sort):
                for i in other:
                    if i.sort == self.sort or i.sort.contains(self.sort) and self.partOf(i):
                        return True
                return False
            raise ValueError('Operation on two forms of different sorts '+str(self.sort)+'  and '+str(other.sort))
            
        if self.isEmpty():
            return True
        if other.isEmpty():
            return False
        self.maximalize()
        other.maximalize()
        for i in self.elements:
            if not other.containsInd(i):
                return False
        return True

    def containsInd(self, ind):
        """
        check if this form contains a certain individual
        """
        if not self.sort == ind.sort:
            return False
        if not ind.hasAttribute():
            posPlanes = copy(self.elements)
        else:
            posPlanes = []
            for i in self.elements:
                if ind.attribute.checkPartOf(i.attribute, flagDesc = False):
                    posPlanes.append(i)
        if len(posPlanes) == 0:
            return False

        for i in posPlanes:
            if i.equals(ind):
                return True

        mergedPlanes = []
        while len(posPlanes)>0:
            plane1 = posPlanes.pop(0)
            index = 0
            while index < len(mergedPlanes):
                plane = mergedPlanes[index]
                check = plane1.merge(plane)
                if check is not None:
                    plane1 = check
                    mergedPlanes.remove(plane)
                    index -= 1
                index += 1
            index = 0
            while index < len(posPlanes):
                plane = posPlanes[index]
                check = plane1.merge(plane)
                if check is not None:
                    plane1 = check
                    posPlanes.remove(plane)
                    index -= 1
                index += 1
            mergedPlanes.append(plane1)
        
        for i in mergedPlanes:
            if i.contains(ind):
                return True
        return False
        
    def duplicateForm(self):
        """
        duplicate the form and its content
        """
        duplicate = coForm(self.sort)
        for i in self.elements:
            duplicate.insert(i.duplicate())
        return duplicate
