# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import range
from builtins import super
from builtins import str

from .multiplyForm import multiplyForm


class curve(multiplyForm):
    """
    A curve form specifies a form with an interval operational behavior,
    applicable to Bezier Curves.

    The curve class extends on the :class:`~sortal.form.multiplyForm.multiplyForm` class and
    specifies the interval operational behavior of forms.
    """

    type = 'form-curve'
    
    def __init__(self, sort, *args, **kwargs):
        super(curve, self).__init__(sort)
    
    def printE(self, string = '', *args, **kwargs):
        """
        print out the form for easier assessment

        :param string: prefix string to be added to the beginning of all printed lines
        :type string: str
        """
        print(string + '|' + self.sort.name+' ('+str(self.sort.baseSort().characteristic.type)+'):',len(self.elements),'elements')
        for ind in self.elements:
            print(string, end = ' ')
            print(ind.printVal())
            if ind.hasAttribute():
                ind.attribute.printE(string+'  ')
        return
        
    def insert(self, ind):
        """
        Inserting an element into the form at appropriate location in the element list. 
        This function will reset the form's maximal status to False.

        :param element: element to be added
        :type element: ~sortal.element.element
        """
        if not self.sort == ind.sort:
            raise ValueError('Inserting an object of sort '+ind.sort.name+' into a form of sort '+self.sort.name)
        if ind.type == self.type:
            for i in ind:
                self.insert(i)
            return
        #thisInd = self.first()
        index = 0
        for index, thisInd in enumerate(self.elements):
            #print(ind.printRhino(), thisInd.printRhino())
            if thisInd.compareValue(ind) != 0:
                continue
            comp = thisInd.compareInterval(ind)
            if comp == 0:
                if thisInd.hasAttribute() or ind.hasAttribute():
                    thisInd.attribute.sum(ind.attribute)
                return
        self.elements.append(ind)
        self.maximal = False
        return
                
    
    def oneStepForward(self, index):
        """
        Used in maximalization, checking the element at index position with the element right after
        to see whether they are duplicate or can be merged and act accordingly.
        """
        current = self.get(index)
        nextInd = self.get(index+1)
        if nextInd is not None and current.compareValue(nextInd) == 0 and current.touches(nextInd) and nextInd.attribute == current.attribute:
            self.elements[index+1] = nextInd.combine(current)
            if current.hasAttribute():
                self.elements[index+1].addAttribute(current.attribute)
                self.elements[index+1].attribute.sum(nextInd.attribute)
            nextInd = self.elements[index+1]
            self.remove(current)
        else:
            index += 1
        return index
        
    def maximalize(self, ignoreMaximal = False):
        """
        maximalize the form:

        * remove Nil elements
        * merge/split elements based on their attributes
        * reorder the elements in ascending order
        """
        if self.maximal and ignoreMaximal == False:
            return
        if len(self.elements) == 0:
            return
        oneI = 0
        while oneI < len(self.elements):
            one = self.get(oneI)
            twoI = oneI+1
            changedOne = False
            while twoI < len(self.elements):
                two = self.get(twoI)
                cmp = one.compareValue(two)
                if cmp == 0:
                    cmpI = one.compareInterval(two)
                    if cmpI == 0:
                        if one.hasAttribute() or two.hasAttribute():
                            one.attribute.sum(two.attribute)                
                        self.pop(twoI)
                    elif one.touches(two) :
                        if (not one.hasAttribute() and not two.hasAttribute()) or one.attribute == two.attribute:
                            self.pop(twoI)
                            self.elements[oneI] = one.combine(two)
                            if one.hasAttribute():
                                self.elements[oneI].addAttribute(one.attribute)
                            changedOne = True
                            break
                        else:
                            twoI += 1
                    elif one.disjoint(two):
                        twoI += 1
                    elif one.contains(two):
                        if (not one.hasAttribute() and not two.hasAttribute()) or one.attribute == two.attribute:
                            self.pop(twoI)
                        else:
                            temp = one.complement(two)
                            if one.hasAttribute() or two.hasAttribute():
                                two.attribute.sum(one.attribute)
                            
                            self.pop(oneI)
                            
                            for i in temp:
                                self.insert(i)
                            changedOne = True
                            break
                    elif two.contains(one):
                        if (not one.hasAttribute() and not two.hasAttribute()) or one.attribute == two.attribute:
                            self.pop(oneI)
                            changedOne = True
                            break
                        else:
                            temp = two.complement(one)
                            if one.hasAttribute() or two.hasAttribute():
                                one.attribute.sum(two.attribute)
                            self.pop(twoI)
                            for i in temp:
                                self.insert(i)
                            changedOne = True
                            break
                    else:
                        if (not one.hasAttribute() and not two.hasAttribute()) or one.attribute == two.attribute:
                            self.pop(twoI)
                            self.elements[oneI] = one.combine(two)
                            if one.hasAttribute():
                                self.elements[oneI].addAttribute(one.attribute)
                            one = self.elements[oneI]
                            changedOne = True
                            break
                        else:
                            temp1 = one.complement(two)[0]
                            temp2 = two.complement(one)[0]
                            one = self.elements[oneI] = one.common(two)
                            if one.hasAttribute() or two.hasAttribute():
                                one.attribute.sum(two.attribute)
                            self.pop(twoI)
                            self.insert(temp1)
                            self.insert(temp2)
                            changedOne = True
                            break
                else:
                    twoI += 1
                                
            if not changedOne:
                oneI += 1
        for ind in self.elements:
            if ind.hasAttribute():
                ind.attribute.maximalize()    
        self.maximal = True
    
    
    def sum(self, other):
        """
        Modify this form to be the sum of this form and another. 
        The other form will be unchanged.

        | e.g.  
        |       formA: a, b, c
        |       formB: b, c, d
        |   formA.sum(formB) --> 
        |       formA: a, b, c, d

        :param other: other form to find the sum with
        :type other: ~sorta.form.form.form
        """
        if self is other:
            raise ValueError('Operation on two instances of the same form')
        if not self.sort.__eq__(other.sort):
            raise ValueError('Operation on two forms of different sorts '+str(self.sort)+'  and '+str(other.sort))
        
        for i in other.elements:
            self.insert(i)
                
        self.maximalize()
        other.maximalize()
        return

    
    def difference(self, other):
        """
        Modify this form to be the difference of this form and another. 
        The other form will be unchanged.
        
        | e.g.  
        |       formA: a, b, c
        |       formB: b, c, d
        |   formA.difference(formB) --> 
        |       formA: a

        :param other: other form to find the difference with
        :type other: ~sorta.form.form.form
        """
        if self is other:
            raise ValueError('Operation on two instances of the same form')
        if not self.sort.__eq__(other.sort):
            raise ValueError('Operation on two forms of different sorts '+str(self.sort)+'  and '+str(other.sort))
        
        self.maximalize()
        other.maximalize()
        thisInd = self.first()
        thisIndex = 0
        while thisInd is not None:
            changed = False
            for otherInd in other:
                if thisInd is None:
                    break
                if thisInd.compareValue(otherInd) == 0:
                    cmpI = thisInd.compareInterval(otherInd)
                    if thisInd.disjoint(otherInd) or thisInd.touches(otherInd):
                        if cmpI == 1:
                            continue
                        else:
                            thisIndex += 1
                            thisInd = self.get(thisIndex)
                            changed = True
                    elif cmpI == 0:
                        if thisInd.hasAttribute() or otherInd.hasAttribute():
                            thisInd.attribute.difference(otherInd.attribute)
                        if not thisInd.hasAttribute():
                            self.remove(thisInd)
                            thisInd = self.get(thisIndex)
                            changed = True
                        continue
                    elif thisInd.contains(otherInd):
                        temp = thisInd.complement(otherInd)
                        if thisInd.hasAttribute() or otherInd.hasAttribute():
                            for i in temp:
                                i.addAttribute(thisInd.attribute)
                            thisInd.attribute.difference(otherInd.attribute)
                            if thisInd.hasAttribute():
                                self.elements[thisIndex] = thisInd.common(otherInd)
                                self.elements[thisIndex].addAttribute(thisInd.attribute)
                                thisInd = self.elements[thisIndex]
                                for i in temp:
                                    self.insert(i)
                                thisIndex += 1
                                if thisInd == self.get(thisIndex):
                                    thisIndex += 1
                                thisInd = self.get(thisIndex)
                                changed = True
                            else:
                                self.remove(thisInd)
                                for i in temp:
                                    self.insert(i)
                                thisInd = self.get(thisIndex)
                                changed = True
                        elif not thisInd.hasAttribute():
                            self.remove(thisInd)
                            for i in temp:
                                self.insert(i)
                            thisInd = self.get(thisIndex)
                            changed = True
                            
                    elif otherInd.contains(thisInd):
                        if thisInd.hasAttribute() or otherInd.hasAttribute():
                            thisInd.attribute.difference(otherInd.attribute)
                        if not thisInd.hasAttribute():
                            self.remove(thisInd)
                            thisInd = self.get(thisIndex)
                            changed = True
                        else:
                            thisIndex += 1
                            thisInd = self.get(thisIndex)
                            changed = True
                    elif cmpI == -1:   
                        temp = thisInd.complement(otherInd)
                        if thisInd.hasAttribute() or otherInd.hasAttribute():
                            for i in temp:
                                i.addAttribute(thisInd.attribute)
                            thisInd.attribute.difference(otherInd.attribute)
                        if thisInd.hasAttribute():
                            self.elements[thisIndex] = thisInd.common(otherInd)
                            self.elements[thisIndex].addAttribute(thisInd.attribute)
                            thisInd = self.elements[thisIndex]
                            self.insert(temp[0])
                            thisIndex += 2
                            thisInd = self.get(thisIndex)
                            changed = True
                        else:
                            self.remove(thisInd)
                            self.insert(temp[0])
                            thisIndex += 1
                            thisInd = self.get(thisIndex)
                            changed = True
                    else:
                        temp = thisInd.complement(otherInd)
                        if thisInd.hasAttribute() or otherInd.hasAttribute():
                            for i in temp:
                                i.addAttribute(thisInd.attribute)
                            thisInd.attribute.difference(otherInd.attribute)
                        if thisInd.hasAttribute():
                            self.elements[thisIndex] = thisInd.common(otherInd)
                            self.elements[thisIndex].addAttribute(thisInd.attribute)
                            thisInd = self.elements[thisIndex]
                            thisIndex += 1
                        else:
                            self.remove(thisInd)
                        self.insert(temp[0])
                        thisInd = self.get(thisIndex)
                        changed = True
                else:                      
                    continue
            if not changed:
                thisIndex += 1
                thisInd = self.get(thisIndex)
        
        
        index = 0
        while index < len(self.elements):
            index = self.oneStepForward(index)
        return
        
    def product(self, other):
        """
        Modify this form to be the product of this form and another. 
        The other form will be unchanged.

        | e.g.  
        |       formA: a, b, c
        |       formB: b, c, d
        |   formA.sum(formB) --> 
        |       formA: b, c

        :param other: other form to find the product with
        :type other: ~sorta.form.form.form

        """
        if self is other:
            raise ValueError('Operation on two instances of the same form')
        if not self.sort.__eq__(other.sort):
            raise ValueError('Operation on two forms of different sorts '+str(self.sort)+'  and '+str(other.sort))
            
        self.maximalize()
        other.maximalize()
        differenceForm = self.duplicate()
        differenceForm.difference(other)
        self.difference(differenceForm)
        return
                
    def containsInd(self, ind):
        """
        check if this form contains a certain individual
        """
        com = None
        if self.sort == ind.sort:
            for i in self.elements:
                if not i.compareValue(ind) == 0:
                    continue
                if i.containsInd(ind):
                    return True
                elif i.containsPoint(ind.start) and (not i.end == ind.start) and self.sort.type == 'sort-attribute':
                    if ind.hasAttribute():
                        if not ind.attribute.checkPartOf(i.attribute):                    
                            continue
                    com = i
                elif com is not None:
                    if i.start == com.end and i.containsPoint(ind.end) and ind.attribute.checkPartOf(i.attribute):
                        com = com.combine(i)
                        if com.contains(ind):
                            return True
                        return False
                    elif i.start == com.end and ind.contains(i) and ind.attribute.checkPartOf(i.attribute):
                        com = com.combine(i)
        return False
        
        
    def partOf(self, other):
        """
        check if this form is a part of other form

        :param other: other form
        :type other: ~sortal.form.form.form
        :returns: boolean
        """
        if self is other:
            raise ValueError('Operation on two instances of the same form')
        if not self.sort == other.sort:
            if other.sort.contains(self.sort):
                for i in other:
                    if i.sort == self.sort or i.sort.contains(self.sort) and self.partOf(i):
                        return True
                return False
            raise ValueError('Operation on two forms of different sorts '+str(self.sort)+'  and '+str(other.sort))
        self.maximalize()
        other.maximalize()
        thisInd = self.first()
        otherInd = other.first()
        for thisInd in self.elements:
            check = False
            com = None
            for otherInd in other.elements:
                if not otherInd.compareValue(thisInd) == 0:
                    continue
                if thisInd == otherInd or otherInd.contains(thisInd):
                    if thisInd.hasAttribute() and not thisInd.attribute.partOf(otherInd.attribute):
                        return False
                    check = True
                    break

                elif otherInd.containsPoint(thisInd.start):
                    if thisInd.hasAttribute():
                        if not thisInd.attribute.partOf(otherInd.attribute):
                            continue
                    com = otherInd.duplicateInd()
                elif com is not None:
                    if (otherInd.start == com.end and otherInd.containsPoint(thisInd.end) 
                    and self.sort.type == 'sort-attribute' and thisInd.attribute.partOf(otherInd.attribute)):
                        com = com.combine(otherInd)
                        if com.contains(thisInd):
                            check = True
                            break
                    elif (otherInd.start == com.end and thisInd.contains(otherInd) 
                    and self.sort.type == 'sort-attribute' and thisInd.attribute.partOf(otherInd.attribute)):
                        com = com.combine(otherInd)

            if check is False:
                return False
        return True
