# -*- coding: utf-8 -*-

from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import
from builtins import super

from .multiplyForm import multiplyForm


class discrete(multiplyForm):
    """
    A discrete-form specifies a form with a discrete operational
    behavior, corresponding a mathematical set: any two individuals in the
    form are either identical or disjoint. A discrete form is maximal
    if no two individuals are identical, and the individuals' attribute forms,
    if any, are also maximal.
    
    The discrete class extends on the :class:`~sortal.form.multiplyForm.multiplyForm` class and
    specifies the discrete operational behavior of forms.
    """

    type = 'form-discrete'
    
    def __init__(self, sort, *args, **kwargs):
        """
        Initializing the discrete form
        """
        super(discrete, self).__init__(sort)
    
    def printE(self, string = ''):
        """
        print all the form's elements and their subsidiaries in tree format

        :param string: prefix string to be added to the beginning of all printed lines
        :type string: str
        """
        print(string + '|' + self.sort.name+' ('+str(self.sort.baseSort().characteristic.type)+'):',len(self.elements),'elements')
        for ind in self.elements:
            print(string, end = ' ')
            print(ind.printVal())
            if ind.hasAttribute():
                ind.attribute.printE(string+'  ')
        return

    def rhinoString(self, dX=0, dY = 0):
        """
        get the string format of the form that can be pasted into Rhinoceros for visualization

        :param dX: x-coordinate offset from (0,0)
        :type dX: float
        :param dY: y-coordinate offset from (0,0)
        :type dY: float
        """
        if self.sort.baseSort().characteristic.type == 'ind-label':
            s = [0,0,0]            
            for lbl in self.elements:
                if lbl.value == "0":
                    s[0] = 255
                if lbl.value == "1":
                    s[1] = 255
                if lbl.value == "2":
                    s[2] = 255
            s = ' SelLast -Properties o c o '+str(s[0])+','+str(s[1])+','+str(s[2])+' EnterEnd SelNone\n'
            return s
        # elif self.sort.baseSort().characteristic.type == 'ind-description':
        #     for desc in self.elements:
        #         if str(desc) == '`x1?=ptHeight.index1`':
        #             s = ' SelLast -Properties o c o 255,0,0 EnterEnd SelNone\n'
        #             return s
        s = ''
        for ind in self.elements:
            st = ind.printRhino(dX, dY)
            if st is not None:
                s += st + '\n'
                if ind.hasAttribute():
                    s += ind.attribute.rhinoString(dX, dY)
        return s
        
    def insert(self, ind):
        """
        Inserting an individual into the form at the end of the element list. 
        This function will reset the form's maximal status to False.

        :param ind: individual to be added
        :type ind: ~sortal.ind.individual.individual
        """
        if not self.sort == ind.sort:
            raise ValueError('Inserting an object of sort '+ind.sort.name+' into a form of sort '+self.sort.name)
            
        if ind.type == self.type:
            for i in ind:
                self.insert(i)
            return
        self.maximal = False
        thisInd = self.first()
        index = 0
        while thisInd is not None:
            if thisInd.compareValue(ind) == 0:
                if thisInd.hasAttribute() or ind.hasAttribute():
                    thisInd.attribute.sum(ind.attribute)
                return
            elif thisInd.compareValue(ind) == 1:
                self.add(index, ind)
                return
            else:
                index = index+1
                try:
                    thisInd = self.elements[index]
                except:
                    thisInd = None
        self.add(index, ind)
        return
        
    def maximalize(self):
        """
        Maximalize the form:

        * remove Nil elements
        * merge/split elements based on their attributes
        * reorder the elements in ascending order
        """
        if self.maximal:
            return
        
        if len(self.elements) == 0:
            return
            
        
        ind = self.first()
        index = 1
        nextInd = self.get(index)
        disorganized = []
        while nextInd is not None:
            #if nextInd is Nil, remove it and assign nextInd as the next one
            if nextInd.isNil():
                self.remove(nextInd)
                nextInd = self.get(index)
                continue
            #if ind is Nil(should only happen for the first element), move 
            #   ind and nextInd one step forward
            elif ind.isNil():
                self.remove(ind)
                ind = nextInd
                nextInd = self.get(index)
                continue
            #if ind is equals to nextInd, sum up their attributes and remove 
            #   nextInd, assign nextInd one step forward 
            cmp = ind.compareValue(nextInd)
            if cmp == 0:
                if ind.hasAttribute() or nextInd.hasAttribute():
                    ind.attribute.sum(nextInd.attribute)                
                self.remove(nextInd)
                nextInd = self.get(index)
            #if ind is greater than nextInd, remove nextInd, add it to 
            #   disorganized to be added back in later in the correct place, 
            #   assign nextInd one step forward
            elif cmp == 1:
                self.remove(nextInd)
                disorganized.append(nextInd)
                nextInd = self.get(index)
            #if ind is lesser than nextInd, nothing happens, more ind and 
            #   nextInd one step forward
            else:
                ind = nextInd
                index += 1
                nextInd = self.get(index)
        
        #add all the disorganized element back in the form again in the correct place
        for ins in disorganized:
            self.insert(ins)
            
        #maximalize the attributes of the individuals in the form
        for ind in self.elements:
            if ind.hasAttribute():
                ind.attribute.maximalize()       
                
        self.maximal = True
        return
    
    def sum(self, other):
        """
        Modify this form to be the sum of this form and another. 
        The other form will be unchanged.

        | e.g.  
        |       formA: a, b, c
        |       formB: b, c, d
        |   formA.sum(formB) --> 
        |       formA: a, b, c, d

        :param other: other form to find the sum with
        :type other: ~sorta.form.form.form
        """
        if self is other:
            raise ValueError('Operation on two instances of the same form')
        if not self.sort.__eq__(other.sort):
            raise ValueError('Operation on two forms of different sorts '+str(self.sort)+'  and '+str(other.sort))
        
        self.maximalize()
        other.maximalize()
        
        thisIndex = 0
        otherIndex = 0
        thisInd = self.first()
        otherInd = other.first()
        while thisInd is not None and otherInd is not None:
            if thisInd.compareValue(otherInd) == 0:
                #if has attribute, sum
                if thisInd.hasAttribute() or otherInd.hasAttribute():
                    thisInd.attribute.sum(otherInd.attribute)
                
                thisIndex += 1                 
                otherIndex += 1
                thisInd = self.get(thisIndex)           
                otherInd = other.get(otherIndex)
            elif thisInd.compareValue(otherInd) == 1:
                self.add(thisIndex, otherInd)  
                thisIndex += 1                          
                otherIndex += 1
                otherInd = other.get(otherIndex)
            else:                          
                thisIndex += 1                 
                thisInd = self.get(thisIndex)           
        while otherInd is not None:
            self.elements.append(otherInd)
            otherIndex += 1
            otherInd = other.get(otherIndex)
        return

    def difference(self, other):
        """
        Modify this form to be the difference of this form and another. 
        The other form will be unchanged.
        
        | e.g.  
        |       formA: a, b, c
        |       formB: b, c, d
        |   formA.difference(formB) --> 
        |       formA: a

        :param other: other form to find the difference with
        :type other: ~sorta.form.form.form
        """
        if self is other:
            raise ValueError('Operation on two instances of the same form')
        if not self.sort.__eq__(other.sort):
            raise ValueError('Operation on two forms of different sorts '+str(self.sort)+'  and '+str(other.sort))
        
        self.maximalize()
        other.maximalize()
            
        thisIndex = 0
        otherIndex = 0
        thisInd = self.first()
        otherInd = other.first()
        while thisInd is not None and otherInd is not None:
            if thisInd.compareValue(otherInd) == 0:
                #if has attribute, calculate difference
                if thisInd.hasAttribute() or otherInd.hasAttribute():
                    thisInd.attribute.difference(otherInd.attribute)
                if not thisInd.hasAttribute():
                    self.pop(thisIndex)
                    thisInd = self.get(thisIndex)
                otherIndex += 1
                otherInd = other.get(otherIndex)
            elif thisInd.compareValue(otherInd) == 1:                            
                otherIndex += 1
                otherInd = other.get(otherIndex)
            else:      
                thisIndex += 1                       
                thisInd = self.get(thisIndex)
        return

    def product(self, other):
        """
        Modify this form to be the product of this form and another. 
        The other form will be unchanged.

        | e.g.  
        |       formA: a, b, c
        |       formB: b, c, d
        |   formA.sum(formB) --> 
        |       formA: b, c

        :param other: other form to find the product with
        :type other: ~sorta.form.form.form

        """
        if self is other:
            raise ValueError('Operation on two instances of the same form')
        if not self.sort.__eq__(other.sort):
            raise ValueError('Operation on two forms of different sorts '+str(self.sort)+'  and '+str(other.sort))
            
        self.maximalize()
        other.maximalize()
        
        thisIndex = 0
        otherIndex = 0
        thisInd = self.first()
        otherInd = other.first()
        while thisInd is not None and otherInd is not None:
            if thisInd.compareValue(otherInd) == 0:
                #if has attribute, product
                if thisInd.hasAttribute():
                    thisInd.attribute.product(otherInd.attribute)
                thisIndex += 1
                otherIndex += 1
                thisInd = self.get(thisIndex)                            
                otherInd = other.get(otherIndex)
            elif thisInd.compareValue(otherInd) == 1:                
                otherIndex += 1
                otherInd = other.get(otherIndex)
            else:                             
                self.pop(thisIndex)
                thisInd = self.get(thisIndex)                            
        while thisIndex < len(self):   
            self.pop(thisIndex)
        return

        
    def partOf(self, other):
        """
        check if this form is a part of other form

        :param other: other form
        :type other: ~sortal.form.form.form
        :returns: boolean
        """
        if self is other:
            raise ValueError('Operation on two instances of the same form')
        if not self.sort == other.sort:
            if other.sort.contains(self.sort):
                for i in other:
                    if i.sort == self.sort or i.sort.contains(self.sort) and self.partOf(i):
                        return True
                return False
            raise ValueError('Operation on two forms of different sorts '+str(self.sort)+'  and '+str(other.sort))
        self.maximalize()
        other.maximalize()

        if self.sort.baseSort().characteristic.type == 'ind-description' and len(self) == 1:
            desc = self.elements[0]
            if len(desc.tree) == 1 and desc.tree[0].type == 'description-literal' and (desc.tree[0].name == 'nil'
            or desc.tree[0].name == 'e'):
                return True
        thisIndex = 1
        otherIndex = 1
        thisInd = self.first()
        otherInd = other.first()
        while thisInd is not None and otherInd is not None:
            if thisInd.compareValue(otherInd) == 0:
                if thisInd.hasAttribute() and not thisInd.attribute.partOf(otherInd.attribute):
                    return False
                thisInd = self.get(thisIndex)
                otherInd = other.get(otherIndex)
                thisIndex += 1
                otherIndex += 1
                
            elif thisInd.compareValue(otherInd) == 1:
                otherInd = other.get(otherIndex)
                otherIndex += 1
            else:
                return False
        return thisInd is None
        
    def checkPartOf(self, other, flagDesc = True):
        """
        check if this form is a part of other form.
        This function should only be used in the geometry check of the detect

        :param other: other form form
        :type other: ~sortal.form.form.form
        :param flagDesc: check for whether the descriptions in 
        :type flagDesc: boolean
        :returns: boolean
        """
        if self.sort.baseSort().characteristic.type == 'ind-description':
            if (len(self) == 1 and len(self.elements[0].tree) == 1 
            and self.elements[0].tree[0].type == 'description-literal' 
            and (self.elements[0].tree[0].name == 'nil' or self.elements[0].tree[0].name == 'e')):
                return True
            if len(self) > len(other):
                return False
            if other.type == 'form-metaForm':
                check = False
                for i in other:
                    if i.sort == self.sort:
                        other = i
                        check = True
                        break
                if not check:
                    return False
            for i in self:
                if not i.checkPartOf(other):
                    return False
            if not flagDesc:
                return True
            for i in self.elements:
                i.registerDelay(other)
            return True
        else:
            return self.partOf(other)