# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import super

from ..element import element

class form(element):
    """
    A form is a collection of :class:`~sortal.ind.individual.individual` 
    (or forms) of the same :class:`~sortal.sort.sort.sort`. Forms belong 
    to a few types, where each type specifies a particular behavior for 
    forms under common arithmetic operations such as sum, difference, and 
    product. Each characteristic :class:`~sortal.ind.individual.individual` 
    has a behavior (Form class) assigned. 

    Form types distinguish between singly- and multiply-associated sorts.
    A :class:`~sortal.form.singlyForm.singlyForm` specifies a form to 
    contain only a single individual, resulting in singly-associated sorts. 
    An example is the :class:`~sortal.form.ordinal.ordinal` specifying an 
    ordinal behavior on numbers: any set of numbers is reduced to the single 
    maximum number, all other numbers in the original set are a part
    of this maximum number.

    A :class:`~sortal.form.multiplyForm.multiplyForm` allows a form to 
    contain any number of individuals, resulting in multiply-associated sorts. 

    A :class:`~sortal.form.discrete.discrete` form specifies a discrete 
    behavior corresponding to a set: any two individuals are either identical 
    or disjoint. 

    An :class:`~sortal.form.inteval.inteval` form specifies a behavior of 
    intervals on a one-dimensional axis: if two intervals overlap, these 
    are combined into a single interval. Similarly, 
    :class:`~sortal.form.cyclical.cyclical` form 
    :class:`~sortal.form.curve.curve` form both specifies a behavior of 
    intervals similar to interval form, but for arcs on circles 
    and ellipses (cyclical form) and bezierCurves (curve form).

    Finally, a :class:`~sortal.form.metaForm.metaForm` specifies a form 
    consisting itself of forms from different sorts. This corresponds to 
    a :class:`~sortal.sort.disjunctiveSort.disjunctiveSort` as
    a disjunctive combination of various sorts. Any individual belongs to
    exactly one of the operand sorts.

    The Form class extends on the :class:`~sortal.element.element` class. It 
    specifies the interface as well as some common functionality for all 
    subclasses. Each subclass specifies a particular behavior for forms. A form 
    is typically represented as a list of elements, either individuals or forms. 
    If a form is assigned an attribute form to an individual, this individual 
    is denoted the associate to this form and is represented as such.

    A form is denoted maximal if it is represented in its canonical form.
    That is, the elements are ordered and no two elements in this form can be
    combined into a single element of the same class. Forms are always
    maximalized before these are compared or output.

    **Object Parameters:** (not including inherited attributes)

    * **name** (str) -- a value in the bezier function.
    * **prev** (:class:`~sortal.form.form.form`) -- previously discarded form from the form register that shares the same name.
    """

    formRegister = {}
    
    def __init__(self, sort, *args, **kwargs):
        """
        Initializing a form. This function is called for every form initialization

        :param sort: sort of the form
        :type sort: ~sortal.sort.sort.sort
        """
        super(form, self).__init__(sort)
        self.name = None
        self.prev = None
        
    def __repr__(self):
        return self.oneLineString()
        
    def __str__(self):
        return self.sort.name + ': ' + self.toString()
        
    def printRhino(self, dX=0, dY=0):
        print(self.rhinoString(dX, dY))
        
    def totalLength(self):
        raise NotImplementedError 
        
    def rhinoString(self, dX=0, dY=0):
        return ''
        
    def geometricFrame(self):
        return None
        
    def compareValue(self, other):
        if 'form' not in other.type:
            raise ValueError('comparing a form with a non-form object')
        s = self.totalLength()
        o = other.totalLength()
        if s < o:
            return -1
        if s > o:
            return 1
        return 0
        
    def register(self, name):
        """
        register this form in the form register with a form name

        :param name: name of the form
        :type name: str
        """
        if self.name is not None and self.name in form.formRegister:
            form.formRegister.pop(self.name)
        if name in form.formRegister:
            self.prev = form.formRegister.pop(name)
        form.formRegister[name] = self
        self.name = name
        
    def isNil(self):
        """
        Check if the form is Nil: whether the form is empty
        """
        return self.isEmpty()
    
    def isMaximal(self):
        raise NotImplementedError 
   
    def size(self):
        raise NotImplementedError 
    
    def first(self):
        raise NotImplementedError 
    
    def next(self, currentElement):
        raise NotImplementedError 
    
    def parse(self, reader, assoc=None):
        return
        
    def updateAssoc(self):
        for i in self:
            if i is not None:
                i.assoc = self
                i.updateAssoc()
        
    def checkPartOf(self, other, flagDesc = True):
        return self.partOf(other)
        
    def maximalize(self):
        raise NotImplementedError 

    def sum(self, other):
        raise NotImplementedError 
        
    def difference(self, other):
        raise NotImplementedError 
        
    def product(self, other):
        raise NotImplementedError 
        
    def partition(self, other):
        raise NotImplementedError 
        
    def partOf(self, other):
        raise NotImplementedError 
        
    def merge(self, other):
        raise NotImplementedError 
        
    def add(self, element):
        raise NotImplementedError 
        
    def remove(self, element):
        raise NotImplementedError 
    
    def edit(self, editor):
        raise NotImplementedError 
        
