# -*- coding: utf-8 -*-

from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import

from builtins import super

from .multiplyForm import multiplyForm
from ..transform.graphTransform import graphTransform, graphTransform3D
class metaForm(multiplyForm):
    """
    A meta-form denotes a form of a :class:`~sortal.sort.disjunctiveSort.disjunctiveSort`.
    It is composed of zero, one or more forms corresponding to the component
    sorts of the disjunctive sort. The operational behavior of a meta-form is
    defined by the individual behaviors of the element forms.
    A meta-form is maximal if it contains at most one form
    corresponding to each component sort and each element form is maximal.

    The meta-form class extends on the :class:`~sortal.form.multiplyForm.multiplyForm` class and
    specifies the operational behavior of meta-forms in terms of the behaviors
    of its constituent element forms.
    """

    type = 'form-metaForm'
    
    def __init__(self, sort, *args, **kwargs):
        super(metaForm, self).__init__(sort)
        if sort.type == 'sort-compound':
            self.compound = True
        else:
            self.compound = False
        
    def toString(self, string = '', printKey = False):
        """
        get the form in string format
        """
        string += '  '
        s = '{\n' + string
        printComma = False
        for i in self.elements:
            if printComma:
                s += ',\n' + string
            else:
                printComma = True
            s += i.sort.name+': '+i.toString(string, printKey)
        s += '\n'+ string[:-2]+ '}'
        return s
        
    def totalLength(self):
        """
        The total number of individuals in this form
        """
        length = 0
        for i in self:
            length += i.totalLength()
        return length
        
    def oneLineString(self, printKey = False):
        """
        get this form in one-line string format
        """
        s = '{'
        printComma = False
        for i in self.elements:
            if printComma:
                s += ', '
            else:
                printComma = True
            s += i.sort.name+': '+ i.oneLineString(printKey)
        s += '}'
        return s
        

    def __eq__(self, other):
        if other is None or self.type != other.type or self.sort != other.sort or len(self.elements) != len(other.elements):
            return False
        
        if not self.maximal:
            self.maximalize()
        if not other.maximal:
            other.maximalize()
            
        lenSelf = 0
        for i in self:
            lenSelf += len(i)
        lenOther = 0
        for i in other:
            lenOther += len(i)
        if lenSelf != lenOther:
            return False
            
        for i in self.elements:
            check = False
            for j in other.elements:
                if i.type != j.type or i.sort != j.sort:
                    continue
                if i == j:
                    check = True
                    break
            if check == False:
                return False
        return True
        
    
    def printE(self, string = ''):
        """
        print out the form for easier assessment

        :param string: prefix string to be added to the beginning of all printed lines
        :type string: str
        """
        print(string + '|' + self.sort.name+':',len(self.elements),'elements')
        for ele in self.elements:
            ele.printE(string+'  ')
        return
        
    def rhinoString(self, dX = 0, dY = 0):
        """
        get the string format of the form that can be pasted into Rhinoceros for visualization

        :param dX: x-coordinate offset from (0,0)
        :type dX: float
        :param dY: y-coordinate offset from (0,0)
        :type dY: float
        """
        s = ''
        for ele in self.elements:
            if ele.type in ('form-metaForm','form-interval','form-discrete','form-coForm','form-areal','form-cyclical','form-curve'):
                s += ele.rhinoString(dX, dY)
        return s
        

    def accept(self, visitor, direction=-1):
        """
        Go through all the elements in this form and perform visitEnter and visitLeave 
        functions of the visitor on each of the element.

        :param visitor: visitor to perform visitEnter and visitLeave functions with
        :param direction: any extra information for the visitor's function
        """
        if not self.compound and isinstance(direction, int) and -1 == direction:
            direction = self.sort.name
    
        if visitor.visitEnter(self, direction) and not self.isEmpty():
            for element in self.elements:
                element.accept(visitor, direction)
        visitor.visitLeave(self)
        return

    def splitElements(self):
        splits = []
        for i in self.elements:
            splits += i.splitElements()
        return splits
        
    def switchSort(self, newSort):
        """
        Create a new form with the content of this form under a new sort

        :param newSort: the new sort
        :type newSort: ~sortal.sort.sort.sort
        :returns: :class:`~sortal.form.metaForm.metaForm` -- the new metaForm
        """
        if not newSort.contains(self.sort):
            raise ValueError(self.sort.name+' is not a subset of '+newSort.name)
        form = newSort.newForm()
        for i in self.elements:
            form.elements.append(i.duplicate())
        return form
        
    def splitSort(self, newSort, complementSort = None):
        """
        Split the content of this form into two metaForms under two diffent sorts (newSort and complementSort).
        The newSort must be a part of this form's sort.

        :param newSort: the sort of one of the splitted form
        :type newSort: ~sortal.sort.sort.sort
        :param complementSort: (optional) the sort with the leftover content
        :type complementSort: ~sortal.sort.sort.sort
        :returns: (:class:`~sortal.form.metaForm.metaForm`, :class:`~sortal.form.metaForm.metaForm`) -- a 
            tuple of the two new forms
        """
        if not self.sort.contains(newSort):
            raise ValueError(newSort.name+' is not a subset of '+self.sort.name)
        form = newSort.newForm()
        
        if complementSort is None:
            complementSort = self.sort.complementSort(newSort)
        else:
            for i in self.sort.component:
                if i not in newSort.component and i not in complementSort.component:
                    raise ValueError('Sorts '+newSort.name+' and '+complementSort.name+' do not sum up to become '+self.sort.name)
        complementForm = complementSort.newForm()
        for i in self.elements:
            if i.sort in newSort.component:
                form.elements.append(i.duplicate())
            else:
                complementForm.elements.append(i.duplicate())
        return form, complementForm
                
    def mergeSort(self, otherForm, newSort):
        """
        Merge the content of this and other form into a new metaForm under a new sort.
        The new sort must contains the sorts of this and other form.

        :param otherForm: the other form to be merged with
        :type otherForm: ~sortal.form.metaForm.metaForm
        :param newSort: the new sort
        :type newSort: ~sortal.sort.sort.sort
        :returns: :class:`~sortal.form.metaForm.metaForm` -- the new merged form
        """
        if not newSort.contains(self.sort):
            raise ValueError(self.sort.name+' is not a subset of '+newSort.name)
        if not newSort.contains(otherForm.sort):
            raise ValueError(self.sort.name+' is not a subset of '+newSort.name)
        form = newSort.newForm()
        for i in self:
            form.insert(i)
        for i in otherForm:
            form.insert(i)
        return form
                
    def insert(self, element):
        """
        Inserting an element into the form at appropriate location in the element list. 
        This function will reset the form's maximal status to False.

        :param element: element to be added
        :type element: ~sortal.element.element
        """
        if self.compound:
            if element.type == 'form-metaForm':
                if not self.sort.contains(element.sort):
                    print('~~~~~~~')
                    print("ERROR: Argument sorts do not adhere to containment relationship:")
                    print(self.sort.name,"does not contain",element.sort.name)
                    return
                for i in self.elements:
                    if i.sort == element.sort:
                        for j in element:
                            i.insert(j)
                        return
                self.elements.append(element)
                return
            for i in self.elements:
                if i.sort.contains(element.sort):
                    i.insert(element)
                    return
            for i in self.sort.component:
                if i.contains(element.sort):
                    newForm = metaForm(i)
                    newForm.insert(element)
                    self.elements.append(newForm)
                    return
            return
        if 'form' in element.type:
            if not self.sort.contains(element.sort):
                print('~~~~~~~')
                print("ERROR: Argument sorts do not adhere to containment relationship:")
                print(self.sort.name,"does not contain",element.sort.name)
                return
            self.maximal = False
            if element.type == 'form-metaForm':
                for i in element:
                    self.insert(i)
                return
            index = 0
            for i in self.elements:
                if i.type == 'form-metaForm':
                    if i.sort.contains(element.sort):
                        i.insert(element)
                        return
                elif i.sort.baseSort().dimension() == element.sort.baseSort().dimension():
                    if i.sort.name == element.sort.name:
                        i.merge(element)
                        return
                    elif element.sort.name < i.sort.name:
                        self.add(index, element)
                        return
                elif i.sort.baseSort().dimension() > element.sort.baseSort().dimension():
                    self.add(index, element)
                    return
                index = index + 1
            self.add(index, element)
        else:
            if not self.sort.contains(element.sort):
                print('!!!!!!!!!!')
                print(self.sort.name, element.sort.name)
                print("Argument sorts do not adhere to containment relationship")
                return
            self.maximal = False
            index = 0
            for i in self.elements:
                if i.sort.baseSort().dimension() == element.sort.baseSort().dimension():
                    if i.sort.name == element.sort.name:
                        i.insert(element)
                        return
                    elif element.sort.name < i.sort.name:
                        form = element.sort.newForm()
                        form.insert(element)
                        self.add(index, form)
                        return
                elif i.sort.baseSort().dimension() > element.sort.baseSort().dimension():
                    form = element.sort.newForm()
                    form.insert(element)
                    self.add(index, form)
                    return
                index = index + 1
            form = element.sort.newForm()
            form.insert(element)
            self.add(index, form)
        return
               
    
    def maximalize(self):
        """
        maximalize the form:

        * remove Nil elements
        * merge/split elements based on their attributes
        * reorder the elements in ascending order
        """
        if self.maximal:
            return
        self.maximal = True
        for i in self.elements:
            i.maximalize()
        i = self.first()
        disorganized = []
        index = 1
        j = self.get(index)
        while j is not None:
            if i.sort.baseSort().dimension() == j.sort.baseSort().dimension():
                if j.sort.name == i.sort.name:
                    i.sum(j)
                    self.remove(j)
                    j = self.get(index)
                elif i.sort.name > j.sort.name:
                    self.remove(j)
                    disorganized.append(j)
                    j = self.get(index)
                else:
                    index += 1
                    i = j
                    j = self.get(index)
            elif i.sort.baseSort().dimension() > j.sort.baseSort().dimension():
                self.remove(j)
                disorganized.append(j)
                j = self.get(index)
            else:
                index += 1
                i = j
                j = self.get(index)
        for i in disorganized:
            self.insert(i)
        return
        
    def sum(self, other):
        """
        Modify this form to be the sum of this form and another. 
        The other form will be unchanged.

        | e.g.  
        |       formA: a, b, c
        |       formB: b, c, d
        |   formA.sum(formB) --> 
        |       formA: a, b, c, d

        :param other: other form to find the sum with
        :type other: ~sorta.form.form.form
        """
        if self is other:
            raise ValueError('operation on two instances of the same form')
            return
        if not self.sort.__eq__(other.sort):
            if self.sort.contains(other.sort):
                for i in self.elements:
                    if i.sort == other.sort or i.sort.contains(other.sort):
                        i.sum(other)
                return
            raise ValueError('not of the same sort')
        if self.compound:
            if other.compound:
                for i in other.elements: self.insert(i)
            else: self.insert(other)
            self.maximalize()
            return
            
        self.maximalize()
        other.maximalize()
        thisIndex = 0
        otherIndex = 0
        thisForm = self.first()
        otherForm = other.first()
        while thisForm is not None and otherForm is not None:   
            if thisForm.sort.baseSort().dimension() == otherForm.sort.baseSort().dimension():
                if thisForm.sort.name == otherForm.sort.name:
                    #if has attribute, sum
                    thisForm.sum(otherForm)
                    thisIndex += 1
                    otherIndex += 1
                    thisForm = self.get(thisIndex)                            
                    otherForm = other.get(otherIndex)
                elif thisForm.sort.name > otherForm.sort.name:
                    self.add(thisIndex, otherForm)      
                    otherIndex += 1                      
                    otherForm = other.get(otherIndex)
                else:
                    thisIndex += 1
                    thisForm = self.get(thisIndex)                            
            elif thisForm.sort.baseSort().dimension() > otherForm.sort.baseSort().dimension():
                self.add(thisIndex, otherForm)
                thisIndex += 1                 
                otherIndex += 1
                otherForm = other.get(otherIndex)
            else:
                thisIndex += 1
                thisForm = self.get(thisIndex)                            
                 
        while otherForm is not None:
            self.elements.append(otherForm)                            
            otherIndex += 1
            otherForm = other.get(otherIndex)
        return
        
    def difference(self, other):
        """
        Modify this form to be the difference of this form and another. 
        The other form will be unchanged.
        
        | e.g.  
        |       formA: a, b, c
        |       formB: b, c, d
        |   formA.difference(formB) --> 
        |       formA: a

        :param other: other form to find the difference with
        :type other: ~sorta.form.form.form
        """
        if self is other:
            raise ValueError('operation on two instances of the same form')
            return
        if self.sort != other.sort:
            if self.sort.contains(other.sort):
                for i in self.elements:
                    if i.sort == other.sort or i.sort.contains(other.sort):
                        i.difference(other)
                return
            raise ValueError('not of the same sort')

        self.maximalize()
        other.maximalize()
        
        if self.compound:
            if other.compound:
                for i in other.elements: 
                    for j in self.elements:
                        if j.sort.contains(i.sort):
                            j.difference(i)
                            break
            else: 
                for j in self.elements:
                    if j.sort.contains(other.sort):
                        j.difference(i)
                        break
            return
        
        thisIndex = 0
        otherIndex = 0
        thisForm = self.first()
        otherForm = other.first()
        while thisForm is not None and otherForm is not None:
            if thisForm.sort.baseSort().dimension() == otherForm.sort.baseSort().dimension():
                if thisForm.sort.name == otherForm.sort.name:
                    #if has attribute, calculate difference
                    thisForm.difference(otherForm)
                    if thisForm.isEmpty():
                        self.pop(thisIndex)
                        thisForm = self.get(thisIndex)
                    otherIndex += 1
                    otherForm = other.get(otherIndex)
                elif thisForm.sort.name > otherForm.sort.name:                            
                    otherIndex += 1
                    otherForm = other.get(otherIndex)
                else:                             
                    thisIndex += 1
                    thisForm = self.get(thisIndex)
            elif thisForm.sort.baseSort().dimension() > otherForm.sort.baseSort().dimension():
                otherIndex += 1
                otherForm = other.get(otherIndex)
            else:
                thisIndex += 1
                thisForm = self.get(thisIndex)
        return        
        
    def product(self, other):
        """
        Modify this form to be the product of this form and another. 
        The other form will be unchanged.

        | e.g.  
        |       formA: a, b, c
        |       formB: b, c, d
        |   formA.sum(formB) --> 
        |       formA: b, c

        :param other: other form to find the product with
        :type other: ~sorta.form.form.form

        """
        if self is other:
            return
        if not self.sort.__eq__(other.sort):
            if self.sort.contains(other.sort):
                for i in self.elements:
                    if i.sort == other.sort or i.sort.contains(other.sort):
                        i.product(other)
                return
            raise ValueError('not of the same sort')
        differenceForm = self.duplicate()
        differenceForm.difference(other)
        self.difference(differenceForm)
        return

        self.maximalize()
        other.maximalize()
                       
        thisIndex = 0
        otherIndex = 0
        thisForm = self.first()
        otherForm = other.first()
        while thisForm is not None and otherForm is not None:
            if thisForm.sort.baseSort().dimension() == otherForm.sort.baseSort().dimension():
                if thisForm.sort.name == otherForm.sort.name:
                    #if has attribute, product
                    thisForm.product(otherForm)
                    if thisForm.isEmpty():
                        self.pop(thisIndex)
                        thisForm = self.get(thisIndex)
                        otherIndex += 1
                        otherForm = other.get(otherIndex)
                    else:
                        thisIndex += 1
                        otherIndex += 1
                        thisForm = self.get(thisIndex)
                        otherForm = other.get(otherIndex)
                elif thisForm.sort.name > otherForm.sort.name:                
                    otherIndex += 1
                    otherForm = other.get(otherIndex)
                else:                             
                    self.pop(thisIndex)
                    thisForm = self.get(thisIndex)
            elif thisForm.sort.baseSort().dimension() > otherForm.sort.baseSort().dimension():
                otherIndex += 1
                otherForm = other.get(otherIndex)
            else:
                self.pop(thisIndex)
                thisForm = self.get(thisIndex)
        while thisForm is not None:   
            self.pop(thisIndex)
            thisForm = self.get(thisIndex)
        return

        
    def partOf(self, other):
        """
        check if this form is a part of other form

        :param other: other form
        :type other: ~sortal.form.form.form
        :returns: boolean
        """
        if self is other:
            raise ValueError('Operation on two instances of the same form')
        if not self.sort == other.sort:
            if other.sort.contains(self.sort):
                for i in other:
                    if i.sort == self.sort and self.partOf(i):
                        return True
                return False
            raise ValueError('Operation on two forms of different sorts '+str(self.sort)+'  and '+str(other.sort))
            
        self.maximalize()
        other.maximalize()
        
        for thisForm in self.elements:
            if thisForm.isEmpty():
                continue
            check = False
            for otherForm in other.elements:
                if thisForm.sort.baseSort().dimension() == otherForm.sort.baseSort().dimension():
                    if thisForm.sort.name == otherForm.sort.name:
                        if thisForm.partOf(otherForm):
                            check = True
            if check == False:
                return False
        return True
        
    def edit(self, editor):
        """
        transforming the form with a transformation set(editor).

        :param editor: the transformation set to be applied
        :type editor: ~sortal.transform.transformSet.transformSet
        """
        prev = None
        if not self.compound and self.assoc is None:
            if editor.current is not None:
                prev = editor.current
            editor.current = self.sort.name

        self.maximal = False
        result = self.sort.newForm()
        for i in self.elements:
            result.insert(i.edit(editor))
        if prev is not None:
            editor.current = prev
        return result

    def checkTransform(self, transformSet, compareForm):
        """
        check if this form is contained by the compareForm after transforming
        with a transformSet
        
        :param transformSet: the transformation set to be applied
        :type transformSet: ~sortal.transform.transformSet.transformSet
        :param compareForm: the form to check for the containment of this transformed form
        :type compareForm: ~sortal.form.form.form
        :returns: boolean
        """
        prev = None
        if not self.compound and self.assoc is None:
            if transformSet.current is not None:
                prev = transformSet.current
            transformSet.current = self.sort.name
            
        if compareForm.compound:
            if self.compound:
                for i in self.elements:
                    if i.isEmpty():
                        continue
                    check = False
                    for j in compareForm.elements:
                        if i.sort.name == j.sort.name:
                            if not i.checkTransform(transformSet, j):
                                return False
                            check = True
                            break
                    if not check:
                        return False
                if prev is not None:
                    transformSet.current = prev
                return True
            else:
                for j in compareForm.elements:
                    if self.sort == j.sort:
                        if not self.checkTransform(transformSet, j):
                            return False
                        return True
                return False
        for i in self.elements:
            if i.isEmpty():
                continue
            if not i.checkTransform(transformSet, compareForm):
                return False
        if prev is not None:
            transformSet.current = prev
        return True      
                
        
    def checkPartOf(self, other, flagDesc = True):
        """
        check if this form is a part of other form.
        This function should only be used in the geometry check of the detect

        :param other: other form form
        :type other: ~sortal.form.form.form
        :param flagDesc: check for whether the descriptions in 
        :type flagDesc: boolean
        :returns: boolean
        """
        if self is other:
            raise ValueError('Operation on two instances of the same form')
        if not self.sort == other.sort:
            if other.sort.contains(self.sort):
                for i in other:
                    if i.sort == self.sort and self.partOf(i):
                        return True
                return False
            raise ValueError('Operation on two forms of different sorts '+str(self.sort)+'  and '+str(other.sort))
        self.maximalize()
        other.maximalize()
        for thisForm in self.elements:
            if thisForm.isEmpty():
                continue
            check = False
            for otherForm in other.elements:
                if thisForm.sort.baseSort().dimension() == otherForm.sort.baseSort().dimension():
                    if thisForm.sort.name == otherForm.sort.name:
                        if thisForm.checkPartOf(otherForm, flagDesc):
                            check = True
                        
            if check == False:
                return False
        return True
        

