# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import super

from .form import form
from ..transform.graphTransform import graphTransform, graphTransform3D
class multiplyForm(form):
    """
    A multiply-form denotes a form of a multiply-associated sort.
    Multiply-forms may contain any number of individuals.

    Different types of multiply-forms are distinguished based on the operational
    behavior of their forms. 

    A :class:`~sortal.form.discrete.discrete` form specifies a discrete behavior corresponding to a set: any 
    two individuals are either identical or disjoint.

    An :class:`~sortal.form.interval.interval` form specifies a behavior of intervals on a
    one-dimensional axis: if two intervals overlap, these are combined into a
    single interval. 

    An :class:`~sortal.form.areal.areal` form applies to two-dimensional planar segments and
    three-dimensional volumes, and uses a boundary classification approach to
    determine overlapping segment parts.

    A multiply-form is maximal if no two individuals combine into a single
    individual under the respective operational behavior, and if no individuals
    equal to nil.

    The MultiplyForm class extends on the :class:`~sortal.form.form.form` class.
    It implements the form container as an ordered tree set and maintains
    for each form a reference to the attribute sort to which it belongs.

    It provides common implementations for some interface methods.

    **Object Parameters:** (not including inherited attributes)

    * **elements** (list of :class:`~sortal.element.element`) -- a list of elements contained within the form. The elements must be other forms for metaForm and areal form, and must be individuals for all others.
    * **maximal** (boolean) -- whether the form is maximal.
    """
    def __init__(self, sort, *args, **kwargs):
        """
        initialize the multiply form. This function is called for every multiply form initialization
        
        :param sort: sort of the form
        :type sort: ~sortal.sort.sort.sort
        """
        super(multiplyForm, self).__init__(sort)
        self.elements = []
        self.maximal = True
        
    def __iter__(self):
        return iter(self.elements)
        
    def __len__(self):
        return len(self.elements)
        
    def toString(self, string = '', printKey = False):
        if len(self) == 0:
            return '{}'
        if len(self) == 1:
            s = '{'
            s += self.elements[0].toString(printKey)
            if self.elements[0].sort.type == 'sort-attribute' and self.elements[0].hasAttribute():
                return s+self.elements[0].attribute.oneLineString(printKey)+'}'
            return s + '}'
        string += '  '
        s = '{\n' + string
        printComma = False
        for i in self.elements:
            if printComma:
                s += ',\n' + string
            else:
                printComma = True
            if 'ind' in i.type:
                s += i.toString(printKey)
                if i.sort.type == 'sort-attribute' and i.hasAttribute():
                    s += i.attribute.oneLineString(printKey)
            else:
                s += i.toString(string, printKey)
        s += '\n'+ string[:-2]+ '}'
        return s
        
    def oneLineString(self, printKey = False):
        s = '{'
        printComma = False
        for i in self.elements:
            if printComma:
                s += ', '
            else:
                printComma = True
            if 'ind' in i.type:
                s += i.toString(printKey)
                if i.sort.type == 'sort-attribute' and i.hasAttribute():
                    s += i.attribute.oneLineString()
            else:
                s += str(i)
        s += '}'
        return s
        
        
    def rhinoString(self, dX = 0, dY = 0):
        """
        get the string format of the form that can be pasted into Rhinoceros for visualization

        :param dX: x-coordinate offset from (0,0)
        :type dX: float
        :param dY: y-coordinate offset from (0,0)
        :type dY: float
        """
        #print(string + '|' + self.sort.name+':',len(self.elements),'elements')
        s = ''
        for ind in self.elements:
            s += ind.printRhino(dX, dY) +'\n'
            if ind.hasAttribute():
                s += ind.attribute.rhinoString(dX, dY)
        return s
        
    def map(self, assignment, other):
        if self.sort != other.sort:
            return False
        count = 0
        for i in self.elements:
            if i.map(assignment, other):
                count+=1
        if count > 0:
            return True
        return False 
        
    def isEmpty(self):
        """
        check if the form is empty
        """
        return len(self.elements) == 0

    def totalLength(self):
        return len(self)
        
    def getElements(self):
        return self.elements
        
        
    def get(self, index):
        try:
            return self.elements[index]
        except IndexError:
            return None
            
    def first(self):
        """
        return the first element of the form
        """
        if len(self.elements)==0:
            return None
        else: return self.elements[0]
        
    def last(self):
        """
        return the last element of the form
        """
        if len(self.elements)==0:
            return None
        else: return self.elements[-1]
    
    def geometricFrame(self):
        minCoord = [None, None, None]
        maxCoord = [None, None, None]
        if len(self) == 0:
            return None
        for i in self:
            frame = i.geometricFrame()
            if frame is not None:
                for i in range(len(minCoord)):
                    if frame[0][i] is None:
                        continue
                    if minCoord[i] is None or minCoord[i] > frame[0][i]:
                        minCoord[i] = frame[0][i]
                for i in range(len(maxCoord)):
                    if frame[0][i] is None:
                        continue
                    if maxCoord[i] is None or maxCoord[i] < frame[1][i]:
                        maxCoord[i] = frame[1][i]
        if minCoord[0] is None: return None
        return [minCoord, maxCoord]
                    
    def next(self, current):
        """
        return the next element in the form after current element
        """
        try:
            return self.elements[self.elements.index(current)+1]
        except:
            return None
        
    def current(self, obj):
        """
        return the element that is equal/same as obj
        """
        for i, e in enumerate(self.elements):
            if e == obj and i+1 < len(self.elements):
                return self.elements[i]
        return None
        
    def previous(self, current):
        """
        return the previous element in the form before current element
        """
        try:
            return self.elements[self.elements.index(current)-1]
        except:
            return None
    
    def contains(self, current):
        """
        check if the form contains a certain element
        """
        for i, e in enumerate(self.elements):
            if e.__eq__(current):
                return True
        return False

    def __eq__(self, other):
        """
        check equality between two forms
        if all elements are the same between two forms, they are equal
        """
        if other is None or other.__class__ != self.__class__ or not other.sort == self.sort:
            return False
        if not self.maximal:
            self.maximalize()
        if not other.maximal:
            other.maximalize()
        if len(self) != len(other):
            return False
        for i in self.elements:
            check = False
            for j in other.elements:
                if i.type != j.type or i.sort != j.sort:
                    continue
                if i == j:
                    if 'ind' in i.type and i.sort.type == 'sort-attribute':
                        if i.attribute == j.attribute:
                            check = True
                    else:
                        check = True
            if check == False:
                return False
        return True
        
    def merge(self, other):
        """
        merging another form into this form
        i.e. inserting all elements belonging to other form into this form
        """
        for i in other.elements:
            self.insert(i)
            
    def move(self, other, append = False):
        if append:
            for i in self.elements:
                other.elements.append(i)
                i.assoc = other
            self.clear()
        else:
            for i in self.elements:
                other.insert(i)
            self.clear()
            
    def add(self, index, element):
        """
        add an element into this form at a certain index
        """
        if element.assoc is None:
            self.elements.insert(index, element)
            element.assoc = self
        else:
            newCopy = element.duplicate() 
            self.elements.insert(index, newCopy)
            newCopy.assoc = self
        self.maximal = False
        
    def duplicateForm(self):
        duplicate = self.sort.newForm()
        for i in self.elements:
            duplicate.insert(i.duplicate())
        return duplicate
        
    def remove(self, element):
        """
        remove an element from this form
        """
        try:
            self.elements.remove(element)
            element.assoc = None
        except ValueError:
            for i in self.elements:
                if i is element or i == element:
                    self.elements.remove(i)
            element.assoc = None
            
    def pop(self, index):
        """
        pop an element with index out of the list and return it
        index = 0 to pop the 1st element ... -1 for last element
        """
        return self.elements.pop(index)
        
    def accept(self, visitor, direction=-1):
        """
        going through this form for rule analysis or rule detection
        """
        if not self.isEmpty() and visitor.visitEnter(self, direction):
            if self.sort.type == 'sort-attribute':
                for element in self.elements:
                    if element.hasAttribute():
                        element.accept(visitor, direction)
                for element in self.elements:
                    if not element.hasAttribute():
                        element.accept(visitor, direction)
            else:
                for element in self.elements:
                    element.accept(visitor, direction)
        visitor.visitLeave(self)
        return
        
    def edit(self, editor):
        """
        transforming the form with a transformation set(editor).
        """        
        self.maximal = False
        result = self.sort.newForm()
        for i in self.elements:
            transformedElement = i.edit(editor)
            if transformedElement is not None:
                result.insert(transformedElement)
        return result
        
    def splitElements(self):
        splits = []
        for i in self.elements:
            newForm = i.sort.newForm()
            newForm.insert(i)
            splits.append(newForm)
        return splits
                
    def containsInd(self, ind):
        """
        check if this form contains a certain individual
        """
        if self.type == 'form-metaForm':
            for i in self.elements:
                if i.sort == ind.sort or i.type == 'form-metaForm':
                    if i.containsInd(ind):
                        return True
        elif self.sort == ind.sort:
            for i in self.elements:
                if i.containsInd(ind):
                    return True
        return False
        
    def checkTransform(self, transformSet, compareForm):
        """
        check if this form is contained by the compareForm after transforming
        with transformSet
        """
        for i in self.elements:
            if not i.checkTransform(transformSet, compareForm):
                return False
        return True      
    
    def clear(self):
        """
        clear the content of the form
        """
        for i in self.elements:
            i.assoc = None
        self.elements = []
