# -*- coding: utf-8 -*-

from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import

from builtins import super

from .singlyForm import singlyForm

class ordinal(singlyForm):
    """
    An ordinal-form specifies a form with an ordinal operational behavior
    on numbers: any set of numbers is reduced to the single maximum number, all
    other numbers in the original set are a part of this maximum number.
    An ordinal-form is maximal if it is not equal to nil and its
    individual's attribute form is maximal. If it is equal to nil, it
    should not have an individual, i.e., its individual should be null.

    The OrdinalForm class extends on the :class:`~sortal.form.singlyForm.singlyForm` class and
    specifies the ordinal operational behavior of forms.s
    """

    type = 'form-ordinal'
    def __init__(self, sort, *args, **kwargs):
        super(ordinal, self).__init__(sort)

    def printE(self, string = ''):
        print(string, end = '|')
        print(self.sort.name+':')
        print(string, end = ' _ ')
        print(self.individual.printVal())
        if self.individual.hasAttribute():
            self.individual.attribute.printE(string+'  ')
        return
        
    def insert(self, ind):
        """
        Insert the individual into the form: replacing the current individual
        """
        if self.sort != ind.sort:
            raise ValueError('Inserting an object of sort '+ind.sort.name+' into a form of sort '+self.sort.name)
        if ind.type == self.type:
            if ind.individual is not None:
                self.add(ind.individual)
            return
        self.add(ind)
        return
        
    def sum(self, other):
        """
        Modify this form to be the sum of this form and another. 
        The other form will be unchanged.

        If other.individual > self.individual: self.individual = other.individual,
        then sum attributes.
        """
        if self is other:
            raise ValueError('Operation on two instances of the same form')
        if not self.sort.__eq__(other.sort):
            raise ValueError('Operation on two forms of different sorts '+str(self.sort)+'  and '+str(other.sort))
            
        self.maximalize()
        other.maximalize()
        
        if self.individual is None:
            self.add(other.individual)
            return
        if other.individual is None:
            return
        if self.individual.compareValue(other.individual) == -1:
            a = self.individual.attribute
            self.add(other.individual)
            self.individual.attribute = a
        if self.individual.hasAttribute() or other.individual.hasAttribute():
            self.individual.attribute.sum(other.individual.attribute)
        return
        
    def difference(self, other):
        """
        Modify this form to be the difference of this form and another. 
        The other form will be unchanged.

        If other.individual > self.individual: self.individual = 0,
        then find the difference of two attributes
        """
        if self is other:
            raise ValueError('Operation on two instances of the same form')
        if not self.sort.__eq__(other.sort):
            raise ValueError('Operation on two forms of different sorts '+str(self.sort)+'  and '+str(other.sort))
        
        self.maximalize()
        other.maximalize()
            
        if other.individual is not None and self.individual is not None:
            if self.individual.hasAttribute():
                self.individual.attribute.difference(other.individual.attribute)
            if not self.individual.complement(other.individual):
                self.remove()
        self.maximalize(True)
        return 
        
        
    def product(self, other):
        """
        Modify this form to be the product of this form and another. 
        The other form will be unchanged.

        if other.individual < self.individual: self.individual = other.individual,
        then find the product of two attributes
        """
        if self is other:
            raise ValueError('Operation on two instances of the same form')
        if not self.sort.__eq__(other.sort):
            raise ValueError('Operation on two forms of different sorts '+str(self.sort)+'  and '+str(other.sort))
            
        self.maximalize()
        other.maximalize()
                       
        if self.individual is None or other.individual is None:
            self.remove()
            return
        if self.individual.compareValue(other.individual) == 1:
            self.add(other.individual)    
        if self.individual.hasAttribute() or other.individual.hasAttribute():
            self.individual.attribute.product(other.individual.attribute)
        return
        
    def symdifference(self,other):
        """
        Modify this form to be the symdifference of this form and another. 
        The other form will be unchanged.
        
        self.individual = min(other.individual, self.individual),
        then find the symdifference of two attributes
        """
        if self is other:
            raise ValueError('Operation on two instances of the same form')
        if not self.sort.__eq__(other.sort):
            raise ValueError('Operation on two forms of different sorts '+str(self.sort)+'  and '+str(other.sort))
            
        self.maximalize()
        other.maximalize()

        if other.individual is None:
            return        
        if self.individual is None:
            self.add(other.individual)
            return
        if self.individual.compareValue(other.individual) == -1:
            self.add(other.individual)
        if self.individual.hasAttribute() or other.individual.hasAttribute():
            self.individual.attribute.symdifference(other.individual.attribute)
        return 

    def partition(self, other):
        if self is other:
            raise ValueError('Operation on two instances of the same form')
        if not self.sort.__eq__(other.sort):
            raise ValueError('Operation on two forms of different sorts '+str(self.sort)+'  and '+str(other.sort))
            
        self.maximalize()
        other.maximalize()
        if other.individual is None or self.individual is None:
            return None
        common = self.sort.newForm()
        if self.individual.compareValue(other.individual) == 0:
            common.add(self.individual)
            if self.individual.hasAttribute() and other.individual.hasAttribute():
                common.individual.attribute.product(other.individual.attribute)
                self.individual.attribute.difference(common.individual.attribute)
                other.individual.attribute.difference(common.individual.attribute)
            if not self.individual.hasAttribute():
                self.remove()
            if not other.individual.hasAttribute():
                other.remove()
        elif self.individual.compareValue(other.individual) == 1:
            common.add(other.individual)
            other.remove()
        else:
            common.add(self.individual)
            self.remove()
        return common
        
    def partOf(self, other):
        """
        Check if this form is a part of other form

        * If other.individual < self.individual: false
        * If other.individual > self.individual: true
        * If other.individual = self.individual: depending on attributes
        """
        if self is other:
            raise ValueError('Operation on two instances of the same form')
        if not self.sort == other.sort:
            if other.sort.contains(self.sort):
                for i in other:
                    if i.sort == self.sort or i.sort.contains(self.sort) and self.partOf(i):
                        return True
                return False
            raise ValueError('Operation on two forms of different sorts '+str(self.sort)+'  and '+str(other.sort))
            
        self.maximalize()
        other.maximalize()
        
        if self.individual is not None and other.individual is not None:
            if self.individual.hasAttribute() and not self.individual.attribute.partOf(other.individual.attribute):
                return False   
            if self.individual.compareValue(other.individual) < 1:
                return True
            elif self.individual.compareValue(other.individual) == 1:
                return False
        return self.individual is None