# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import super

from .form import form
from ..transform.graphTransform import graphTransform, graphTransform3D

class singlyForm(form):
    """
    A singly-form denotes a form of a singly-associated sort.
    Singly-forms may contain only one individual.
    Different types of singly-forms are distinguished based on the operational
    behavior of their forms. 

    An :class:`~sortal.form.ordinal.ordinal` form specifies an ordinal behavior on numbers: any set of 
    numbers is reduced to the single maximum number, all other numbers in the 
    original set are a part of this maximum number.

    A :class:`~sortal.form.specialForm.specialForm` specifies a special behavior: 
    the individual comparison and operations are defined within the individual
    class (equals, compareValue functions for comparison; contains, combine,
    common and complement functions for other operations)

    An singly-form is maximal if it is not equal to nil and its
    individual's attribute form is maximal. If it is equal to nil, it
    should not have an individual, i.e., its individual should be None.

    The SinglyForm class extends on the :class:`~sortal.form.form.form` class.
    It implements the form container in a trivial way, containing a single
    individual.

    It provides common implementations for some interface methods.
    """

    def __init__(self, sort, *args, **kwargs):
        """
        initialize the form
        self.individual is the single individual that belongs to this form
        """
        super(singlyForm, self).__init__(sort)
        self.individual = None
        self.maximal = True
        
    def __iter__(self):
        return iter([self.individual])
        
    def splitElements(self):
        return [self]
                
    def totalLength(self):
        if self.isEmpty():
            return 0
        return 1
        
    def isEmpty(self):
        """
        check if the form is empty
        """
        if self.individual is None:
            return True
        return self.individual.isNil()
        
    def __len__(self):
        if self.individual is None:
            return 0
        return 1
        
    def toString(self, string = '', printKey = False):
        s = '{'
        if self.individual.sort.type == 'sort-attribute' and self.individual.hasAttribute():
            s += self.individual.toString(False) + str(self.individual.attribute)
        else:
            s += self.individual.toString(False)
        s += '}'
        return s
        
    def oneLineString(self, printKey = False):
        return self.toString(printKey)
        
    def getElements(self):
        return [self.individual]
        
    def first(self):
        """
        return the first element of the form which is the only individual
        """
        return self.individual

    def last(self):
        """
        return the last element of the form which is also the only individual
        """
        return self.individual

    def next(self, current):
        return None

    def previous(self, current):
        return None

    def maximalize(self, ignoreMaximal = False):
        """
        maximalize the form
        
        * if the individual is Nil(empty, or has value of 0/min): remove the individual
        * if the individual has attribute, maximalize the attribute form
        """
        if self.maximal and ignoreMaximal == False:
            return
        self.maximal = True
        if self.individual is None:
            return
        if self.individual.isNil():
            self.individual = None
        elif self.individual.hasAttribute():
            self.individual.attribute.maximalize()

    def __eq__(self, other):
        """
        check if two forms are equal
        if the individual in the two forms are equal, the two forms are equal 
        """
        if other is None or other.__class__ != self.__class__ or not other.sort == self.sort:
            return False
        if self.individual is None:
            if other.individual is None:
                return True
            return False
        if other.individual is None:
            return False
        if not self.maximal:
            self.maximalize()
        if not other.maximal:
            other.maximalize()
        if self.individual == other.individual:
            if self.sort.type == 'sort-attribute':
                return self.individual.attribute == other.individual.attribute
            return True
        return False
        
    def merge(self, other):
        """
        merge the other form into this form

        if other.individual > this.individual -> this.individual = other.individual
        """
        if other.individual.compareValue(self.individual) == 1:
            self.add(other.individual)
        return
        
    def add(self, ind):
        """
        assign ind as this form's individual
        """
        self.maximal = False
        if ind.assoc is None:
            self.individual = ind
            ind.assoc = self
        else:
            newCopy = ind.duplicateInd() 
            self.individual = newCopy
            newCopy.assoc = self
        
    def remove(self):
        """
        remove the individual from this form (make it minimum)
        """
        try:
            self.individual = self.individual.getMinInd()        
        except AttributeError:
            self.individual = None
    
    def accept(self, visitor, direction=-1):
        """
        check this form's individual for rule analysis or rule detection
        """
        if not self.isEmpty() and visitor.visitEnter(self, direction):
            self.maximalize()
            self.individual.accept(visitor, direction)
        visitor.visitLeave(self)
        return
        
    def edit(self, editor):
        """
        transforming the form with a transformation set(editor).
        """        
        result = self.sort.newForm()
        result.individual = self.individual.edit(editor)
        if editor.current is None:
            transf = editor.retrieveTransform(graphTransform)
            if transf is not None:
                for i in transf.extra:
                    result.insert(i)
            transf = editor.retrieveTransform(graphTransform3D)
            if transf is not None:
                for i in transf.extra:
                    result.insert(i)
        return result
        
        
    def containsInd(self, ind):
        """
        check if this form contains a certain individual
        """
        if self.individual == ind:
            return [self.individual]
        return False
        
    def checkTransform(self, transformSet, compareForm):
        """
        check if this form is contained by the compareForm after transforming
        with transformSet
        """
        if self.individual.checkTransform(transformSet, compareForm):
            return True
        return False           
        
    def clear(self):
        """
        clear the content of the form
        """
        if self.individual is not None:
            self.individual.assoc = None
            self.individual = None
        
    def duplicateForm(self):
        newForm = self.sort.newForm()
        newForm.insert(self.individual.duplicate())
        return newForm
    
