from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import


from .ruleAnalysis import ruleAnalysis
from .ruleApplication import ruleApplication
from ..struct.planeStruct import PlaneTransformationError
from ..map.similarityMap import similarityMap2D
from ..map.similarityMap3D import similarityMap3D
from ..map.parametricMap import parametricMap
from ..map.exactMap import exactMap
from ..transform.paramSet import paramSet
from ..transform.transformSet import transformSet
from ..ind.description import description
from ..ind.descriptions.tree import DescriptionEvaluatingError
from ..visit.elementVisitor import elementVisitor

class detection(elementVisitor):
    """
    Detection: Object for detecting matches when applying a rule onto a shape
    """
    applicationDict = {}
    
    def __init__(self, rule):
        """
        a detection object is initialized when a rule is created
        """
        self.rule = rule
        self.applicationDict[rule.name] = []

        # maximalize the lhs
        self.rule.lhs.maximalize()

        # perform rule analysis on the lhs and rhs
        self.analysis= ruleAnalysis(self.rule.lhs, self.rule.rhs)

        self.form = None
        self.target = None
        self.uptrail = []
        self.downTrail = []
        self.success = False
        self.applications = []
        
    def isValid(self):
        """
        check if a rule is valid
        """
        if len(self.analysis.maps) == 0:
            return False
        for i in self.analysis.maps:
            if not self.analysis.maps[i].isValid():
                return False
        return True
        
    def count(self):
        return len(self.applications)
        
    def getApplication(self, index = 0):
        """
        get one application from the latest detection
        """
        if index >= len(self.applications):
            raise IndexError
        return self.applications[index]
        
    def matches(self):
        """
        get all the matched forms from the rule applications

        :returns: list of :class:`~sortal.form.form.form` -- list of transformed lhs matches
        """
        matches = []
        for i in self.applications:
            match = i.matchForm
            if len(matches) == 0 or match not in matches:
                matches.append(match)
        return matches
    
    def results(self, match):
        """
        get all the result forms of a single matched form from the rule applications
        
        :returns: list of :class:`~sortal.form.form.form` -- list of transformed rhs results of a 
            single transformed lhs match 
        """
        r = []
        for i in self.applications:
            if match is i.matchForm or match == i.matchForm:
                r.append(i.resultForm)
        return r
        
    def detect(self, form, printing = None):
        """
        Main function for detecting rule matches

        :param form: the shape for the rule to be applied on
        :param printing: (optional) list of options for printing, should only be used in ruleObj.debugDetect(), 
            should be None otherwise. Structure of printing: [printGeoMatch, printBase,  printPredicate, printDirective] 
        :type form: ~sortal.form.form.form
        :type printing: list of 4 booleans
        :returns: list of :class:`~sortal.grammar.ruleApplication.ruleApplication` - list of rule application for the shape
        """
        # clear all previous detect processes
        self.clear()

        # set the shape
        self.form = form

        # if rule is invalid, raise error
        if not self.isValid():
            raise ValueError('DETECTION ERROR: Invalid Rule')

        # if the rule is detected on this exact same shape before,
        # get the saved rule applications and return them
        for i in self.applicationDict[self.rule.name]:
            if i[0] == form:
                self.applications = i[1]
                if printing:
                    print('---------------------------------------------\nExisting matches found:', len(i[1]),'results')
                return i[1]

        # go through fragments and analyze the shape
        #   similarity mapping (non-parametric): analyze counterparts for the rule's distinguishables
        #   graph mapping (parametric): add all parametric objects into a graph
        #   description: add all description into a param map
        self.analysis.unsetVisitedAll()
        fragments = self.analysis.fragments
        for i in fragments:
            self.target = i
            self.uptrail= []
            if not self.target.content.visited:
                visit = True
                self.uptrail.append(self.target)
                while(self.uptrail[-1].parent is not None):
                    self.uptrail.append(self.uptrail[-1].parent)
                    if self.uptrail[-1].content.visited:
                        visit = False
                if visit:
                    self.form.accept(self)
                    self.target.content.visited = True
                    if self.target.parent is not None:
                        self.target.parent.content.visited = True
        
        ## print ##
        if printing:
            print('==================================================================')
            print('                    GENERATING MATCHINGS')
            print('==================================================================')
            print('_ Geometric matchings:')
            printGeoMatch, printBase, printPredicate, printDirective = printing
            frame = self.form.geometricFrame()
            try:
                dispX = frame[1][0] - frame[0][0] + 1
                dispY = frame[1][1] - frame[0][1] + 1
            except TypeError:
                dispX, dispY = 30, 30

        ########## detect geometric transformations ##########
        resultForms = []   
        transformlists = [None] * len(self.analysis.maps)
        lists = 0
        # for each of the mapping
        for map in self.analysis.maps.values():
            # potential transforms (only for parametric/non-parametric geometry matching, empty transform for descriptions)
            transformlists[lists] = map.potentialTransforms()
            i = 0

            # remove duplicate transforms
            while i < len(transformlists[lists])-1:
                j = i+1
                while j < len(transformlists[lists]):
                    if transformlists[lists][i] == transformlists[lists][j]:
                        transformlists[lists].pop(j)
                    else:
                        j += 1
                i += 1

            if printing is not None and not map.type == 'map-parametric':
                print('   +',map.type+': '+str(len(transformlists[lists]))+' results')

            # set transform id (the id being the name of the shape's disjunctive sort)
            transfId = map.id.split('=|=')
            if len(transfId) > 1:
                for i in transformlists[lists]:
                    i.id = transfId[0]
            if transformlists[lists] is not None:
                lists += 1
        
        ########## detect description transformations ##########
        paramsets = None
        if self.analysis.hasMap(parametricMap):

            # potential description matches
            paramsets = self.analysis.maps[parametricMap.type].potentialParamSets()

            # if no match, create one empty paramSet to not raise error
            if printing:
                print('_ Description matchings:')
                if len(paramsets) == 0:
                    paramsets = [paramSet()]
                    print('   + No description matches --> creating an empty description transform set')
                else:
                    print('   + Description matches (before simple resolving):', len(paramsets),'results')
            elif len(paramsets) == 0:
                paramsets = [paramSet()]

            # resolve simple description delays: all the conditions/references that only require checking
            # within the description match without having to check the geometry.
            i = 0
            while i < len(paramsets):
                if not paramsets[i].resolveSimpleDelays():
                    paramsets.pop(i)
                else:
                    i += 1
            if printing:
                print('   + Description matches (after simple resolving):', len(paramsets),'results')
        indices = []
        totalcount = 1
        for i in range(lists):
            indices.append(0)
            totalcount *= len(transformlists[i])

        sBase = ''
        sPredicate = ''
        sDirective = ''
        
        countGeo = 0
        countBase = 0
        mergedCount = 0
        predicateCount = 0
        directiveCount = 0
        firstDescResolvedCount = 0
        secondDescResolvedCount = 0

        ########## Check each combination of geometry+description matches ##########
        for i in range(totalcount):
            # add the potential transformation for each mapping into the tfset
            tfset = transformSet()
            for j in range(lists):
                tfset.add(transformlists[j][indices[j]])
            j = 0
            while indices[j] == len(transformlists[j])-1 and j < lists-1:
                indices[j] = 0
                j += 1
            if indices[j] == len(transformlists[j]):
                break
            indices[j] += 1

            # check geomtry transform to see if the transformed lhs is part of the shape
            try:
                check = self.rule.lhs.checkTransform(tfset, self.form)
            except Exception:
                continue
            ## print ##
            if printing and printGeoMatch:
                #tfset.printTransf()
                try:
                    print()
                    tlhs = self.rule.lhs.transform(tfset) 
                    tlhs.printRhino(countGeo*dispX, -dispY)
                    #print('check:', check)
                except Exception:
                    print('Unable to transform the lhs')
                    pass
                    #print(ex)
            countGeo += 1

            # if transformed lhs is not part of the shape, continue to next set of transformation combination
            if not check:
                continue
            
            ## print ##
            if printing and printBase:
                #tfset.printTransf()
                tlhs = self.rule.lhs.transform(tfset) 
                sBase += tlhs.rhinoString(countBase*dispX, -dispY) + '\n'
            countBase += 1
            # no description
            if paramsets is None:
                mergedCount += 1

                # transform the lhs
                try:
                    tlhs = self.rule.lhs.transform(tfset) 
                except (PlaneTransformationError, ZeroDivisionError):
                    continue

                # check predicate
                if not self.rule.evaluatePredicate(tfset, tlhs):
                    continue
                
                ## print ##
                if printing and printPredicate:
                    sPredicate += tlhs.rhinoString(predicateCount*dispX, -2*dispY) + '\n'

                predicateCount += 1

                # check directives
                for tfS in self.rule.solveDirectives(tfset, tlhs):
                    directiveCount += 1

                    # transform the rhs
                    try:
                        trhs = self.rule.rhs.transform(tfS)
                    except (PlaneTransformationError, ZeroDivisionError, ValueError) as ex:
                        continue

                    ## print ##
                    if printing and printDirective:
                        sDirective += tlhs.rhinoString(directiveCount*dispX, -3*dispY)
                        sDirective += trhs.rhinoString(directiveCount*dispX, -4*dispY) + '\n'

                    # for map in self.analysis.maps.values():
                    #     # analyze rhs for parametric geometry (both 2D and 3D)
                    #     if map.type == 'map-graphMap':
                    #         transformlists[lists] = map.solveRHS(self.rule, self.form, transformlists[lists])


                    # if the rule is additive and the rhs is a part of the shape -> reject
                    if self.rule.ruleType == 'ADDITIVE' and trhs.partOf(self.form):
                        continue
                    
                    # if the same transformation already exist -> reject
                    if [tlhs,trhs] in resultForms:
                        continue

                    # if tlhs is not fully part of the shape -> reject
                    if not tlhs.partOf(self.form):
                        continue
                    
                    # create a ruleApplication
                    resultForms.append([tlhs,trhs])
                    app = ruleApplication(self.rule, tfS, None, tlhs, trhs)

                    if app not in self.applications:
                        self.applications.append(app)
            else:
                description.clearDelays()
                # transform lhs
                try:
                    tlhs = self.rule.lhs.transform(tfset) 
                except (PlaneTransformationError, ZeroDivisionError):
                    continue
                # check each of the description matches
                for paramset in paramsets:
                    mergedCount += 1
                    params = paramset.duplicate()
                    params.lhs = tlhs
                    params.transform = tfset
                    params.keys = self.rule.keys
                    # check predicates
                    if not self.rule.evaluatePredicate(tfset, tlhs, params):
                        continue
                    ## print ##
                    if printing and printPredicate:
                        sPredicate += tlhs.rhinoString(predicateCount*dispX, -2*dispY) + '\n'

                    predicateCount += 1

                    # resolve description delays: all the conditions/references that requires referencing to
                    # the lhs geometry (e.g. line.length, plane.area, ...). Description requiring information on
                    # rhs geometry will be resolved later
                    if not (params.resolveDelays()):
                        continue

                    firstDescResolvedCount += 1
                    # resolve directives
                    for tfS in self.rule.solveDirectives(tfset, tlhs, params):
                        params.transform = tfS

                        # transform rhs
                        try:
                            trhs = self.rule.rhs.transform(tfS)
                        except (PlaneTransformationError, ZeroDivisionError, ValueError):
                            continue
                        
                        ## print ##
                        if printing and printDirective:
                            sDirective += tlhs.rhinoString(directiveCount*dispX, -3*dispY)
                            sDirective += trhs.rhinoString(directiveCount*dispX, -4*dispY) + '\n'
                        directiveCount += 1

                        # resolve description delays: all the conditions/references that requires referencing to
                        # the rhs geometry (e.g. line.length, plane.area, ...).
                        params.rhs = trhs
                        if not description.resolveDelays(params):
                            continue
                        secondDescResolvedCount += 1
                        glhs = tlhs.edit(params)
                        grhs = trhs.edit(params)

                        # transform lhs & rhs with the description match
                        try:
                            glhs = tlhs.edit(params)
                            grhs = trhs.edit(params)
                        except (PlaneTransformationError, ZeroDivisionError, DescriptionEvaluatingError, AttributeError) as ex:
                            print('Error:', str(ex))
                            continue
                        # if the rule is additive and the rhs is a part of the shape -> reject
                        if (self.rule.ruleType == 'ADDITIVE' or glhs.partOf(grhs)) and grhs.partOf(self.form):
                            continue
                        
                        # if the same transformation already exist -> reject
                        if [glhs, grhs] in resultForms:
                            continue

                        # if tlhs is not fully part of the shape -> reject
                        if not glhs.partOf(self.form):
                            continue

                        # create a ruleApplication
                        resultForms.append([glhs, grhs])
                        app = ruleApplication(self.rule, tfS, params, glhs, grhs)
                        if not app in self.applications:
                            self.applications.append(app)
        
        ## print ##
        if printing:
            if paramsets is None:
                print('\n\n==================================================================')
                print('                    EVALUATE MATCHINGS')
                print('==================================================================')
                print('  _ Total geometric transformations:', countBase)
                if printBase:
                    print('*** base geometric results: ***\n\n')
                    print(sBase, end = '')
                    print('*******************************\n')
                print('  _ Total transformations after geometric check:',mergedCount)
                print('\n<===================== Evaluate Predicates ======================>\n')
                if printPredicate:
                    print('*** predicate results: ***\n')
                    print(sPredicate, end = '')
                    print('**************************\n')
                print('  _ Total geometric transformations after predicate check:',predicateCount)
                print('\n<===================== Evaluate Directives ======================>\n')
                if printDirective:
                    print('*** directive results: ***\n')
                    print(sDirective, end = '')
                    print('**************************\n')
                print('  _ Number of transformations after directive check:',directiveCount)
                print('\n<=================== Final Containment Check ====================>\n')
                print('  _ Number of final results:', len(self.applications),'\n\n')
            else:
                print('\n\n==================================================================')
                print('                    EVALUATE MATCHINGS')
                print('==================================================================')
                if printBase:
                    print('*** base geometric results: ***\n')
                    print(sBase, end = '')
                    print('*******************************\n')
                print('  _ Total geometric transformations:', countBase)
                print('  _ Total description transformations:', len(paramsets))
                print('\n<=============== Merging Description and geometric ==============>\n')
                print('  _ Total transformations after merging:',mergedCount)
                print('\n<===================== Evaluate Predicates ======================>\n')
                if printPredicate:
                    print('*** predicate results: ***\n')
                    print(sPredicate, end = '')
                    print('**************************\n')
                print('  _ Total geometric transformations after predicate check:',predicateCount)
                print('\n<==================== Evaluate Descriptions =====================>\n')
                print('  _ Number of transformations after first description check:',firstDescResolvedCount)
                print('\n<===================== Evaluate Directives ======================>\n')
                if printDirective:
                    print('*** directive results: ***\n')
                    print(sDirective, end = '')
                    print('**************************\n')
                print('  _ Number of transformations after directive check:',directiveCount)
                print('\n<==================== Evaluate Descriptions =====================>\n')
                print('  _ Number of transformations after second description check:',secondDescResolvedCount)
                print('\n<=================== Final Containment Check ====================>\n')
                print('  _ Number of final results:', len(self.applications),'\n\n')

        # register the application detection and return
        self.applicationDict[self.rule.name].append([self.form.duplicate(), self.applications])
        return self.applications
        # if len(self.applications) == 0:
        #     if mergedCount == 0:
        #         raise RuntimeError('DETECTION ERROR: No Geometry match detected')
        #     if predicateCount == 0:
        #         raise RuntimeError('DETECTION ERROR: No Predicate solution detected')
        #     if firstDescResolvedCount == 0:
        #         raise RuntimeError('DETECTION ERROR: No Description solution detected')
        #     if directiveCount == 0:
        #         raise RuntimeError('DETECTION ERROR: No Directive solution detected')
        #     if secondDescResolvedCount == 0:
        #         raise RuntimeError('DETECTION ERROR: No Description solution detected')

    def clear(self):
        self.form = None
        for map in self.analysis.maps.values():
            map.clear()
        self.target = None
        self.uptrail = []
        self.downTrail = []
        self.success = False
        self.applications = []

    def visitEnter(self, element, assignment):
        """
        process to be performed on each of the element in a form in form.accept(this_detection)
        """
        if 'ind' in element.type:
            if element.sort not in self.analysis.sorts:
                return False
            if len(self.uptrail) == 0:
                try:
                    mapping = self.analysis.getMap(element.sort, assignment)
                except KeyError:
                    return False
                result = False
                fragment = self.target.content
                if len(self.downTrail) != 0:
                    fragment = self.downTrail[0].content
                # try to map the individual to all elm in the fragment by performing mapping.map()
                for elm in fragment:
                    if assignment == 'refl' and mapping.type == 'map-parametric':
                        continue
                    try:
                        if mapping.map(elm, element):
                            result = True
                    except AttributeError:
                        pass
                if len(self.downTrail) == 0:
                    self.success = (result or self.success)
                return result
        elif 'form' in element.type:
            if element.type == 'form-metaForm':
                return True
            if element.sort not in self.analysis.sorts:
                return False
            elif len(self.uptrail) != 0 and element.sort != self.uptrail[-1].content.sort:
                return False
            elif len(self.uptrail) != 0:
                self.uptrail.pop(0)
            else: 
                try:
                    fragment = self.downTrail[0]
                except IndexError: 
                    fragment = self.target
                for frag in fragment.children:
                    if element.sort == frag.content.sort:
                        self.downTrail.append(frag)
                        break
                if len(self.downTrail) >0 and self.downTrail[-1].content.visited:
                    return False

        return True
    
    def visitLeave(self, element, asignment = None):
        if 'ind' in element.type:
            if (self.success and len(self.uptrail) != 0):
                mapping = self.analysis.getMap(element.sort, asignment)
                for elm in self.uptrail[-1].parent.content:
                     mapping.map(elm, element)
                self.success = False
        elif 'form' in element.type:
            if element.type == 'form-metaForm':
                return
            elif len(self.downTrail) != 0:
                self.downTrail.pop(0)

            elif (len(self.uptrail) != 0 and self.uptrail[-1].parent is not None
                    and element.sort == self.uptrail[-1].parent.content.sort):
                self.uptrail.append(self.uptrail[-1].parent)
            elif element.sort == self.target.content.sort:
                self.uptrail.append(self.target)
        return
