# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import str

from ..map.parametricMap import parametricMap
from ..visit.elementVisitor import elementVisitor
from ..struct.fragmentTree import treeLink
from .fragment import fragment

class ruleAnalysis(elementVisitor):
    """
    RuleAnalysis: object under :class:`sortal.grammar.detection.detection`,
    used for analyzing the lhs of the rule for matching
    """

    def __init__(self, lhs, rhs):
        self.fragments = []
        self.sorts = []
        self.roots = []
        self.maps = {}
        self.current = None

        self.analyzeLHS(lhs, rhs)



    def getMap(self, sort, disjSort):
        """
        get the mapping for a specific sort

        :param sort: the sort to get mapping for
        :param disjSort: the overarching disjunctive sort 
        """
        if 'sort' in sort.type:
            mapping = sort.baseSort().mapping()
        else:
            mapping = sort
            
        if mapping.type == 'map-parametric' or not isinstance(disjSort, str):
            mapString = mapping.type
        else:
            mapString = disjSort + '=|=' + mapping.type      
            if mapString not in self.maps:
                mapString = mapping.type
        return self.maps[mapString]
        # return self.maps[mapping]

    def hasMap(self, mapping, mappingSort = None):
        """
        check if the rule contain a specific mapping or a type of mapping

        :param mapping: the sort to get mapping for
        :type mapping: ~sortal.map.mapping.mapping
        :param mappingSort: (optional) the overarching disjunctive sort name, if left as None, it will check whether the specific type of mapping in present
        :type mappingSort: string
        """
        mapString = mappingSort + '=|=' + mapping.type if mappingSort is not None else mapping.type
        for i in self.maps:
            if mapString in i:
                return True
        return False
        # return map in self.maps


    def analyzeLHS(self, lhs, rhs):
        lhs.accept(self)
        descMap = parametricMap.getInstance()
        if not self.hasMap(descMap) and (self.checkDesc(lhs) or self.checkDesc(rhs)):
            descMap.id = descMap.type
            self.maps[descMap.type] = descMap
        for i in self.maps.values():
            i.postAnalysis()


    
    def makeFragments(self, sort, parent, mappingSort):
        """
        create a new fragment for containing the individuals
        """
        frag = treeLink(fragment(sort), parent)
        self.fragments.append(frag)
        if parent is None:
            self.roots.append(frag)
        else: parent.children.append(frag)
        mapping = sort.baseSort().mapping()
        
        if mapping.type == 'map-parametric' or not isinstance(mappingSort, str):
            mapString = mapping.type
        else:
            mapString = mappingSort + '=|=' + mapping.type      

        if mapString not in self.maps:
            newMapping = mapping.getInstance()
            newMapping.id = mapString
            self.maps[mapString] = newMapping
            
        return frag
        
    def visitEnter(self, visit, assignment):
        """
        process to be performed on each of the element in a form in form.accept(this_ruleAnalysis)
        """
        # if individual
        if 'ind' in visit.type:
            # add sort to sort list
            if visit.sort not in self.sorts:
                self.sorts.append(visit.sort)
            # create a new Fragment if there is none
            if len(self.fragments) == 0:
                self.current = self.makeFragments(visit.sort, None, assignment)
            # add the individual into the fragment (fragment.addIndividual(...))
            self.current.content.addIndividual(self.getMap(visit.sort, assignment), visit) 
        # if form
        elif 'form' in visit.type:
            if visit.sort.type == 'sort-disjunctive' or visit.sort.type == 'sort-compound':
                return True
            # add sort to sort list
            if visit.sort not in self.sorts:
                self.sorts.append(visit.sort)
            # create a new fragment if there is none
            if visit.assoc is None or self.current is None:
                self.current = self.makeFragments(visit.sort, None, assignment)
            # if current fragment's sort is different from the object sort, create a new fragment
            elif self.current.content.sort != visit.sort:
                found = False
                for i in self.current.children:
                    if i.content.sort == visit.sort:
                        self.current = i
                        found = True
                if found == False:
                    self.current = self.makeFragments(visit.sort, self.current, assignment)
        return True
        
    def visitLeave(self, visit):
        if 'ind' in visit.type:
            return
        elif 'form' in visit.type:
            if visit.sort.type == 'sort-disjunctive':
                return
            if self.current is not None and self.current.content.sort == visit.sort:
                self.current = self.current.parent
        return 
        
    def visitAssoc(self):
        return 
        
    def unsetVisitedAll(self):
        for i in self.fragments:
            i.content.visited = False
    
    def checkDesc(self, element):
        """
        check whether the element is a description or is a form that contains descriptions
        """
        if 'form' in element.type:
            for i in element:
                if self.checkDesc(i):
                    return True
        else:
            if element.type == 'ind-description':
                return True
            if element.type == 'ind-color':
                if element.description is not None:
                    return True
            if element.hasAttribute():
                return self.checkDesc(element.attribute)
        return False
            
    def analyzeRHS(self, rhs):
        for i in self.maps.values():
            i.analyzeRHS(rhs)
