# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import range
from builtins import super
from builtins import str

from .rule import rule
from .ruling import FlowTermination
from .flowApplication import flowApplication
from .ruleDisjunction import ruleDisjunction
from .ruleDisjunctionP import ruleDisjunctionP
from .ruleNegation import ruleNegation
from .ruleSequence import ruleSequence

from ...io.parseReader import parseReader, ParsingError

class flow(rule):
    """
    flow: a set of rules that are to be performed in a certain order

    **Object Parameters:**

    * **name** (str) -- the name of the flow.
    * **description** (str) -- the description of the flow.
    * **rules** (:class:`~sortal.grammar.rules.ruling.ruling`) -- the ruling objects containing all the rules and ruling in the flow.
    """

    type = 'rule-flow'
    register = {}
    
    def __init__(self, name, description, rules, *args, **kwargs):
        if not rule.isValidName(name):
            raise ValueError('Invalid Flow name')
        self.name = name
        self.description = description
        if self.description is None or not isinstance(self.description, str):
            self.description = ''
        # if len(rules.rules) == 1 and rules.quantifier == [1,1]:
        #     self.rules = rules.rules[0]
        # else:
        #     self.rules = rules
        self.rules = rules
        self.process = None
        if self.name in rule.register:
            raise ValueError('a rule/flow called '+self.name+' already exists in the rule register')
        if self.name in flow.register:
            raise ValueError('a flow called '+self.name+' already exists in the flow register')
        flow.register[name] = self
        rule.register[name] = self
        
    @staticmethod
    def existingFlows():
        """
        get the list of existing flow names
        """
        return flow.register.keys()
        
    @staticmethod
    def cleanup():
        """
        clear the flow register
        """
        for i in list(flow.register.keys()):
            del flow.register[i]
            
    def setName(self, newName):
        """
        change flow name
        """
        if newName in flow.register:
            raise ValueError('There already exists a flow with flow name '+str(newName))
        flow.register[newName] = flow.register[self.name]
        del flow.register[self.name]
        
        from ...io.sdlParser import sdlParser
        if self.name in sdlParser.flows:
            sdlParser.flows[newName] = sdlParser.flows[self.name]
            del sdlParser.flows[self.name]
        self.name = newName
    
    @staticmethod
    def getFlow(flowName):
        """
        get a specific flow from the flow register
        """
        return flow.register[flowName]

    def __str__(self):
        return str(self.rules)
    
    def detect(self, *args, **kwargs):
        raise NotImplementedError

    def apply(self, form, printProcess = True, random = True, *args, **kwargs):
        """
        apply the flow on a shape, terminate the process when coming up with one valid result

        :param form: the form to be performed on; the form will not be modified to become the result of the flow application
        :type form: ~sortal.form.form.form
        :param printProcess: whether the process of resolving flow will be printed out in string format for debugging
        :type printProcess: boolean
        :param random: whether the selection of applications are chosen at random
        :type random: boolean
        :returns: :class:`~sortal.grammar.rules.flowApplication.flowApplication` -- the resulting flow application
        """
        flowApp = flowApplication(printProcess, random)
        try:
            self.rules.apply(form.duplicate(), flowApp)
        except FlowTermination:
            print('Backtrack into possessive ruling. Flow Terminated')
            return flowApp
        print()
        flowApp.finalizeResult()
        return flowApp
        #raise NotImplementedError

    def allApplications(self, form, printProcess = False, random = True, *args, **kwargs):
        """
        find all applications of the flow on a shape

        :param form: the form to be performed on; the form will not be modified to become the result of the flow application
        :type form: ~sortal.form.form.form
        :param printProcess: whether the process of resolving flow will be printed out in string format for debugging
        :type printProcess: boolean
        :param random: whether the selection of applications are chosen at random
        :type random: boolean
        :returns: list of :class:`~sortal.grammar.rules.flowApplication.flowApplication` -- the list of flow applications 
            of this flow
        """
        applications = []

        flowApp = flowApplication(printProcess, random)
        try:
            self.rules.apply(form.duplicate(), flowApp)
        except FlowTermination:
            print('Backtrack into possessive ruling. Flow Terminated')
            return []
        if printProcess: print()
        flowApp.finalizeResult()

        while len(flowApp.path) > 0:
            applications.append(flowApp)
            flowApp = flowApp.duplicate()
            try:
                self.rules.apply(form.duplicate(), flowApp, backtrace = True)
            except FlowTermination:
                print('Backtrack into possessive ruling. Flow Terminated')
                flowApp.path = []
            flowApp.finalizeResult()
            if printProcess: print()
        return applications

    def manualApply(self, form):
        """
        Manually apply a flow

        :param form: the form to be performed on; the form will not be modified to become the result of the flow application
        :type form: ~sortal.form.form.form
        :returns: :class:`~sortal.grammar.rules.flowApplication.flowApplication` -- the resulting flow application
        """
        # create new flow application
        flowApp = flowApplication(False, False)

        # mark as a new invocation of the rulings contained
        newInvoke = True
        count = 1
        currentForm = form
        currentPath = ''

        # loop
        while True:
            # find all possible application choices
            #   returned value: result - result[0]: list of flowApplication objects to be chosen
            #                            result[1]: boolean value indicating whether the 
            result = self.rules.manualApplyPeek(form.duplicate(), flowApp, newInvoke)
            newInvoke = False
            apps = result[0]
            if result[1] == False and len(apps) == 1:
                flowApp = apps[0]
                form = flowApp.resultForms[-1]
                continue
            print('\n\n_____________________________________________________________________')
            print('FLOW APPLICATION ITERATION', count)
            print('CURRENT FORM:')
            print(currentForm)
            print('CURRENT APPLICATION PATH:')
            print(currentPath)
            if len(apps) == 0:
                if result[1] == False:
                    print('NO POSSIBLE FURTHER APPLICATION, EXITING')
                    return
                print('NO FURTHER APPLICATION, FLOW COMPLETES')
                flowApp.finalizeResult()
                return flowApp
            print('\n\nAPPLICATIONS:')
            count += 1
            for index, application in enumerate(apps):
                print('\nApplication', index + 1)
                print('->'.join(application.path))
                print(application.resultForms[-1].toString('   '))
            if result[1] == True:
                print('\nApplication', len(apps) + 1)
                print('->'.join(flowApp.path) + '-> END')
            option = input('\nSelect one application: ')
            try:
                option = int(option) - 1
            except Exception:
                break
            if option >=0 and option < len(apps):
                flowApp = apps[option]
                form = flowApp.resultForms[-1]
                currentForm = form
                currentPath = '->'.join(flowApp.path)
            elif result[1] == True and option == len(apps):
                flowApp.finalizeResult()
                return flowApp
            else:
                break
        return
        # flowApp.finalizeResult()
        # return flowApp

    def manualApplyPeek(self, form, flowApp, newInvoke = False):
        """
        Get all the possible first step applications of the ruling on the form based on the flowApp

        :param form: the form to apply the ruling on; this form will be modified to 
            become the resulting form at the end of this method
        :type form: ~sortal.form.form.form
        :param flowApp: the flow application indicating the characteristic of the application 
        :type flowApp: ~sortal.grammar.rules.flowApplication.flowApplication
        :param newInvoke: indicate if this is a new invocation of the ruling (i.e. not continuing from previous manualApplyPeek)
        :type newInvoke: boolean
        :returns: a list of two values: 
            first value is a list of flowApplication objects that are the different possible step results of the flow, 
            second value is a boolean indicating whether the ruling is already satisfied -> can continue on to the next ruling
            or end the flow if it's the topmost ruling in the flow
        """

        return self.rules.manualApplyPeek(form, flowApp, newInvoke)

    @staticmethod
    def readQuantifier(reader):
        """
        parse a quantifier string and return a tuple of two objects: the quantifier and the matching. 
        Used in sdl parsing (:class:`~sortal.io.sdlParser.sdlParser`)
        """
        quantifier = []
        matching = 'G'
        if reader.newToken() == '{':
            quantifier = []
            if reader.newToken() != parseReader.NUMBER:
                raise ParsingError(reader,'Number bigger than 0 expected in the set bracket as quantifier for flow sequence')
            quantifier.append(int(reader.value()))
            if reader.newToken() == '}':
                quantifier.append(quantifier[0])
            else:
                if reader.token() != ',':
                    raise ParsingError(reader,'Number bigger than 0 expected in the set bracket as quantifier for flow sequence')
                if reader.newToken() != '}':
                    if reader.token() != parseReader.NUMBER:
                        raise ParsingError(reader,'Number bigger than 0 expected in the set bracket as quantifier for flow sequence')
                    quantifier.append(int(reader.value()))
                    if reader.newToken() != '}':
                        raise ParsingError(reader,'"}" expected')
                else:
                    quantifier.append(None)
        elif reader.token() == '?':
            quantifier = [0,1]
        elif reader.token() == '*':
            quantifier = [0,None]
        elif reader.token() == '+':
            quantifier = [1,None]

        if reader.previewToken() is not None and reader.previewToken() in ['?', '+', '*']:
            if reader.newToken() == '?':
                matching = 'L'
            elif reader.token() == '+':
                matching = 'PO'
            elif reader.token() == '*':
                matching = 'PR'
        
        return quantifier, matching
        
        
    @staticmethod
    def readFlowRules(reader, thisNegation = False, starting = False):
        """
        recursively parsing the ruling in the flow/ruling. 
        Used in sdl parsing (:class:`~sortal.io.sdlParser.sdlParser`)
        """
        ruleList = []
        paren = '('
        if not starting:
            paren = reader.token()
            if reader.newToken() == '*':
                paren += '*'
                reader.newToken()
        while reader.token() is not None and reader.token() not in (')', ']'):
            negation = False
            if reader.token() == '!':
                negation = True
                reader.newToken()
            if reader.token() in ('[','('):
                ruleList.append(flow.readFlowRules(reader, negation))
            else:
                ruleName = reader.value()
                try:
                    singleRule = rule.register[ruleName]
                except KeyError:
                    raise ParsingError(reader, 'rule '+ruleName+' does not exist in rule register')
                if reader.previewToken() is not None and reader.previewToken() in ['?','*','+','{']:
                    quantifier, matching = flow.readQuantifier(reader)
                    singleRule = ruleSequence([singleRule], quantifier, matching)
                if negation:
                    singleRule = ruleNegation(singleRule)
                ruleList.append(singleRule)
            reader.newToken()
        quantifier = [1,1]
        matching = 'G'
        if reader.token() is not None and (reader.previewToken() is not None and
                                           reader.previewToken() in ['?','*','+','{']):
            quantifier, matching = flow.readQuantifier(reader)
            
            
        if paren == '(' or len(ruleList) == 1:
            r = ruleSequence(ruleList, quantifier, matching)
        elif paren == '[*':
            r = ruleDisjunctionP(ruleList, quantifier, matching)
        elif paren == '[':
            r = ruleDisjunction(ruleList, quantifier, matching)
        if thisNegation:
            r= ruleNegation(r)
        if matching == 'PO':
            r.setPo()
        return r



    @staticmethod
    def createFlow(name, description, string):
        """
        create a new flow
        """
        reader = parseReader(string)
        r = flow.readFlowRules(reader, starting = True)
        flowObj = flow(name, description, r)
        return flowObj
