# -*- coding: utf-8 -*-

from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import
from builtins import str

import sys
import re

from random import randint

from .ruling import ruling
from .flowApplication import flowApplication
from ..detection import detection
from ...struct.key import key
from ...io.parseReader import parseReader, ParsingError
from ...sort.sort import sort as __sortClass__

class rule(ruling):
    """
    A rule consist of a left-hand-side form and a right-hand-side form. It can be
    applied on a shape where parts of the shape that match the lhs would be detected
    and transformed into the rhs in the following manner:
    
    applied_shape = shape - matched_lhs + matched_rhs

    **Object Parameters:** 

    * **name** (str) -- the rule name; must be a unique identifier.
    * **description** (str) -- the rule description.
    * **lhs** (:class:`~sortal.form.form.form`) -- the left hand side form of the rule.
    * **rhs** (:class:`~sortal.form.form.form`) -- the right hand side form of the rule.
    * **keys** (list of key) -- list of all the object keys/tags (both lhs and rhs) in the rule.

    """
    type = 'rule'
    
    register  = {}   
    
    def __init__(self, name, description, lhs, rhs, keys = None, predicates = None, directives = None, *args, **kwargs):
        """
        :param name: the rule name; must be a unique identifier.
        :type name: str
        :param description: the rule description.
        :type description: str
        :param lhs: the left hand side form of the rule.
        :type lhs: ~sortal.form.form.form
        :param rhs: the right hand side form of the rule.
        :type rhs: ~sortal.form.form.form
        :param keys: (optional) list of all the object keys/tags (both lhs and rhs) in the rule.
            if keys is not provided (is None), the rule will iterate through the lhs and
            rhs to compile a list of keys automatically
        :type keys: list of :class:`~sortal.struct.key.key`
        :param predicates: (optional) the predicate string of the rule if there is.
        :type predicates: str
        :param directives: (optional) the directive string of the rule if there is.
        :type directives: str
        """
        super(rule, self).__init__()
        if not rule.isValidName(name):
            raise ValueError('Invalid rule name')
        if name in rule.register:
            raise ValueError('A rule called '+ name +' already exists in the rule register')

        self.name = name
        self.description = description if description is not None else ''

        # parse lhs and rhs if they are in string format
        from ...io.sdlParser import sdlParser
        if isinstance(lhs, str):
            if lhs[0:2] == '<<':
                lhs = lhs[2:]
            # lhsKeys = {}
            lhs = sdlParser.readFormStruct(lhs)
        if isinstance(rhs, str):
            if rhs[0:2] == '>>':
                rhs = rhs[2:]
            # rhsKeys = {}
            rhs = sdlParser.readFormStruct(rhs)

        # self.keys = dictionary of all the object keys/tags in the rule
        # go through the lhs and rhs identifying all keys if not already provided
        self.keys = {}
        if keys is not None:
            self.keys = keys
        else:
            keys = {}
            lhs.accept(self, keys)
            rhs.accept(self, keys)
            self.keys = keys

        # maximalize lhs and rhs
        lhs.maximalize()
        rhs.maximalize()
        
        # identify the rule type
        self.ruleType = 'MIXED'
        if (rhs.partOf(lhs)):
            self.ruleType = 'SUBTRACTIVE'
        elif (lhs.partOf(rhs)):
            self.ruleType = 'ADDITIVE'

        self.lhs = lhs
        self.rhs = rhs
        
        # applicational lhs and rhs
        self.applhs = lhs.duplicate()
        self.applhs.difference(rhs)
        self.apprhs = rhs.duplicate()
        self.apprhs.difference(lhs)

        # rule analysis
        self.detection = detection(self)

        if not self.isValid():
            raise InvalidRuleError(self)
            # print('a rule called '+self.name+' already exists in the rule register')

        rule.register[self.name] = self

        if predicates is not None:
            self.parsePredicate(predicates)
        if directives is not None:
            self.parseDirective(directives, analyzeRHS = False)
        self.detection.analysis.analyzeRHS(self.rhs)

    
    def __eq__(self, other):
        return self.name == other.name

    def __lt__(self, other):
        return self.name < other.name
        
    def __gt__(self, other):
        return self.name > other.name
        
    def __le__(self, other):
        return self.name <= other.name
        
    def __ge__(self, other):
        return self.name >= str(other)

    def __str__(self):
        s = 'rule ' + self.name + ' "'+ self.description +'"\n'
        s += '<< ' + self.lhs.sort.name + ': ' + self.lhs.toString(printKey = True) + '\n'
        predList = self.predicateList()
        if predList is not None and predList != []:
            s += str(self.getPredicates()) + '\n'
        s += '>> ' + self.rhs.sort.name + ': ' + self.rhs.toString(printKey = True)
        dirList = self.directiveList()
        if dirList is not None and dirList != []:
            s += '\n' + str(self.getDirectives())
        s += ';'
        return s


    def isValid(self):
        return self.detection.isValid()

    @staticmethod
    def isValidName(text):
        if not isinstance(text, str):
            return False
        if not (text[0].isalpha() or text[0] == '_'):
            return False
        if re.match(r'^[A-Za-z0-9_\-]+$', text):
            return True
        return False

    @staticmethod
    def existingRules():
        return rule.register.keys()
        
    def visitEnter(self, obj, direction):
        if 'ind' in obj.type:
            if obj.key is not None:
                if obj.key.name in direction and direction[obj.key.name].ind != obj:
                    raise ValueError('There are multiple objects with the same tag name '+obj.key.name)
                direction[obj.key.name] = obj.key
            return True
        elif 'form' in obj.type:
            return True
            
    def visitLeave(self, obj):
        return
        
    @staticmethod
    def cleanup():
        """
        clear all existing rules from the rule register
        """
        for i in list(rule.register.keys()):
            del rule.register[i]
        
    def retrieveKey(self, keyValue):
        if keyValue not in self.keys:
            raise ValueError('tag name '+keyValue+' does not exist in rule '+self.name)
        return self.keys[keyValue]
        
        
    def setName(self, newName):
        """
        change rule name
        """
        if newName in rule.register:
            raise ValueError('There already exists a rule with rule name '+str(newName))
        rule.register[newName] = rule.register[self.name]
        del rule.register[self.name]
        
        detection.applicationDict[newName] = detection.applicationDict[self.name]
        del detection.applicationDict[self.name]
        
        from ...io.sdlParser import sdlParser
        if self.name in sdlParser.rules:
            sdlParser.rules[newName] = sdlParser.rules[self.name]
            del sdlParser.rules[self.name]
        self.name = newName

        
    def setDescription(self, newDesc):
        """
        change rule description
        """
        self.description = newDesc if newDesc is not None else ''
        
    @staticmethod
    def getRule(ruleName):
        """
        get a specific rule from the rule register with rule name
        """
        return rule.register[ruleName]
        
    def detect(self, form):
        """
        detect and return matches for a specific shape(form)
        
        :param form:  the form to detect rule matches on
        :returns: list of :class:`~sortal.grammar.ruleApplication.ruleApplication` -- a list of rule applications, 
                if there is no matches for this rule, an empty list will be returned
        """
        form.maximalize()
        det = self.detection.detect(form)
        return det
        
    def debugDetect(self, form, printGeoMatch = False,
                                printBase = False, 
                                printPredicate = False, 
                                printDirective = False):
        """
        detect and return matches for a specific shape(form), also can be flagged for 
        printing out the rhino code for debugging.

        :param form: the form to detect rule matches on
        :type form: ~sortal.form.form.form
        :param printGeoMatch: printing the geometric matches (before geometry checking, no description, no predicate, directive)
        :type printGeoMatch: boolean
        :param printBase: :param printing the base matches (after geometry checking, no description, no predicate, directive)
        :type printBase: boolean
        :param printPredicate: :param printing the matches remaining after predicate (after predicate checking, no directive)
        :type printPredicate: boolean
        :param printDirective: :param printing the matches remaining after directive (after directive checking)
        :type printDirective: boolean
        :returns: list of :class:`~sortal.grammar.ruleApplication.ruleApplication` -- a list of rule applications, 
                if there is no matches for this rule, an empty list will be returned
        """
        form.maximalize()
        det = self.detection.detect(form,[
                                         printGeoMatch,
                                         printBase, 
                                         printPredicate, 
                                         printDirective
                                         ])
        return det
        
    def detectApplyAll(self, form):
        """
        detect and apply all the application on a specific shape(form). 
        
        :param form:  the form to detect and apply rule matches on
        :returns: the transformed form
        """
        form.maximalize()
        det = self.detection.detect(form)

        tForm = form.duplicate()
        for app in det:
            tForm = app.perform(tForm)
        return tForm

    def apply(self, form, flowApp = None, backtrace = False):
        """
        Detect matches and apply one of the results onto a specific shape(form),
        the rule application to be applied is selected at random so the result may
        differ when the code is rerun.

        Used in flow.

        :param form: the form to apply the rule on; this form will be modified to 
            become the resulting form at the end of this method
        :type form: ~sortal.form.form.form
        :param random: indicate whether the rule applications will be selected at random 
        :type random: boolean
        :param backtrace: indicate if this is a backtrace from a flow -- *only used in flow*
        :type backtrace: boolean
        :param s: a list of string indicating the process of a flow -- *only used in flow*
        :type s: list of str
        :param printProcess: whether the process of resolving the rule will be printed out 
            in string format for debugging -- *only used in flow*
        :type printProcess: boolean
        :returns: the resulting form after applying the rule; 
            the input form is also modified to become the resulting form
        """
        if flowApp is None:
            flowApp = flowApplication()
            backtrace = False


        ruleTrace = flowApp.getRulingTrace(self.name)
        if ruleTrace is None:
            ruleTrace = []
            flowApp.setRulingTrace(self.name, ruleTrace)

        if backtrace:
            if len(ruleTrace) == 0:
                if flowApp.printProcess: print(' **'+self.name+'** ', end='')
                sys.stdout.flush()
                return False
            tracing = ruleTrace.pop(-1)
            app = tracing[0]
            form = tracing[2].duplicate()
        else:
            form.maximalize()
            app = self.detect(form) #detect the possible matches
            tracing = [app, list(range(len(app))), form.duplicate()]

        if len(tracing[1])>0:
            if flowApp.random:
                randInt = randint(0, len(tracing[1])-1)
            else:
                randInt = 0
            appIndex = tracing[1].pop(randInt)
            app[appIndex].perform(form)
        else:
            if flowApp.printProcess:
                print(' ##'+self.name+'## ', end='')
            sys.stdout.flush()
            return False
            
        ruleTrace.append(tracing)
        if flowApp.printProcess: 
            print(' '+self.name, end='')
            print('('+str(appIndex+1)+'/'+str(len(app))+')', end = '')
        sys.stdout.flush()

        if flowApp is not None:
            flowApp.addPath(' '+self.name+'('+str(appIndex+1)+'/'+str(len(app))+')')
            flowApp.addResult(form.duplicate())
        return form
        
    def applySpecific(self, form = None, choice = 0):
        """
        detect matches and apply one of the results onto a specific shape(form),
        the rule application to be applied is specified by its index

        :param form: the form to apply the rule on
        :param choice: the index of the application in the result list
        :type form: ~sortal.form.form.form
        :type choice: int
        """
        form.maximalize()
        app = self.detect(form)

        if choice >= len(app):
            print('no result')
            return form
        app[choice].perform(form)

        print(' '+self.name, end='')
        print('('+str(choice+1)+'/'+str(len(app))+')', end = '')
        sys.stdout.flush()

        return form

    def applyAll(self, form, applications):
        tForm = form.duplicate()
        for app in applications:
            tForm = app.perform(tForm)
        return tForm

    def manualApply(self, form, flowApp = None):
        if flowApp is None:
            flowApp = flowApplication()
        print('applying:', self.name)
        app = self.detect(form)
        print(len(app),'results')
        if len(app) == 0:
            return False
        elif len(app) == 1:
            print('  _ application:')
            tForm = form.duplicate()
            tForm = app[0].perform(tForm)
            print('     ',tForm.toString('      '))
            return tForm
        applications = []
        for i in range(len(app)):
            print('  _ application ' + str(i) + ':')
            tForm = form.duplicate()
            tForm = app[i].perform(tForm)
            applications.append(tForm)
            print('     ',tForm.toString('      '))
        choice = input('  application choice, please select a number between 0 and ' + str(len(app) - 1) + ': ')
        while not choice.isnumeric() or int(choice) < 0 or int(choice) >= len(app):
            choice = input('  invalid application choice! Please select a number between 0 and ' + str(len(app) - 1) + ': ')
        choice = int(choice)
        flowApp.addPath(' '+self.name+'('+str(choice+1)+'/'+str(len(app))+')')
        flowApp.addResult(applications[choice].duplicate())
        return applications[choice]

    def manualApplyPeek(self, form, flowApp, newInvoke = False):
        apps = self.detect(form)
        if len(apps) == 0:
            return [[], False]
        flowApps = []
        count = 1
        for app in apps:
            formDup = form.duplicate()
            newFlowApp = flowApp.duplicate()
            newFlowApp.addPath(' '+self.name+'('+str(count)+'/'+str(len(apps))+')')
            newFlowApp.resultForms.append(app.perform(formDup))
            flowApps.append(newFlowApp)
            count += 1
        # if len(apps) == 1:
        #     return [flowApps, False]
        return [flowApps, False]


    def addPredicate(self, predicate, keyList):
        """
        add a predicate to the rule's predicate list

        :param predicate: the predicate to be added
        :param keyList: the list of all object tags in the rule
        :type predicate: predicate in list form
        :type keyList: list of key
        """
        for i in self.detection.analysis.maps:
            if 'graphMap' in i or 'similarityMap' in i:
                self.detection.analysis.maps[i].addPredicate(predicate, keyList)
                return
        if 'map-parametric' in self.detection.analysis.maps:
            self.detection.analysis.maps['map-parametric'].addPredicate(predicate, keyList)

    def parsePredicate(self, predicateString, clearCurrent = False):
        """
        parse predicates from a string and add them to the rule's predicate list

        :param predicateString: the predicate in string format ("::" at the start of the predicate string is optional)
        :param clearCurrent: True: replace the current predicates with the predicate in the string;
                            False: add the parsed predicates to the end of the current predicate list
        :type predicateString: str
        :type clearCurrent: boolean
        """
        if isinstance(predicateString, str):
            from ...io.sdlParser import sdlParser
            predicateString = predicateString.strip()
            if predicateString[0:2] != '::':
                predicateString = '::' + predicateString
            reader = parseReader(predicateString)
            predicates = sdlParser.readPredicates(reader, self.keys)
        else:
            predicates = predicateString
        if clearCurrent:
            self.clearPredicates()
        for pred in predicates:
            self.addPredicate(pred[0], pred[1])


    def clearPredicates(self):
        """
        clear all current predicates in the rule
        """
        for i in self.detection.analysis.maps:
            if 'graphMap' in i or 'similarityMap' in i:
                self.detection.analysis.maps[i].clearPredicates()
                return
        if 'map-parametric' in self.detection.analysis.maps:
            self.detection.analysis.maps['map-parametric'].clearPredicates()

    def addDirective(self, directive, keyList, analyzeRHS = True):
        """
        add a directive to the rule's directive list

        :param directive: the directive to be added
        :param keyList: the list of all object tags in the rule
        :type directive: directive in list form
        :type keyList: list of key
        """
        for i in self.detection.analysis.maps:
            if 'graphMap' in i:
                self.detection.analysis.maps[i].addDirective(directive, keyList)
                if analyzeRHS:
                    self.detection.analysis.analyzeRHS(self.rhs)
                return

    def parseDirective(self, directiveString, clearCurrent = False, analyzeRHS = True):
        """
        parse directives from a string and add them to the rule's directive list

        :param directiveString: the directive in string format ("::" at the start of the directive string is optional)
        :param clearCurrent: True: replace the current directives with the directive in the string;
                            False: add the parsed directives to the end of the current directive list
        :type directiveString: str
        :type clearCurrent: boolean
        """
        if isinstance(directiveString, str):
            from ...io.sdlParser import sdlParser
            directiveString = directiveString.strip()
            if directiveString[0:2] != '::':
                directiveString = '::' + directiveString
            reader = parseReader(directiveString)
            directives = sdlParser.readDirectives(reader, self.keys)
        else:
            directives = directiveString
        if clearCurrent:
            self.clearDirectives()
        for pred in directives:
            self.addDirective(pred[0], pred[1], False)
        if analyzeRHS:
            self.detection.analysis.analyzeRHS(self.rhs)


    def clearDirectives(self):
        """
        clear all current directives in the rule
        """
        for i in self.detection.analysis.maps:
            if 'graphMap' in i:
                self.detection.analysis.maps[i].clearDirectives()
                return

    def getPredicates(self):
        """
        return the rule's predicates (in string format)
        """
        return self.predicateString()
        
    def printPredicateString(self):
        """
        print the rule's predicates in string format 
        """
        print(self.predicateString())
    
    def predicateList(self):
        """
        get the predicates in the rule
        """
        for i in self.detection.analysis.maps:
            if 'graphMap' in i or 'similarityMap' in i:
                return self.detection.analysis.maps[i].predicates
        if 'map-parametric' in self.detection.analysis.maps:
            return self.detection.analysis.maps['map-parametric'].predicates
        return None

    def predicateString(self):
        """
        return the rule's predicates in string format
        """
        predicates = self.predicateList()
        if predicates is None:
            return '::'
        s = ':: '
        writecomma = False
        for pred in predicates:
            if writecomma:
                s += ', '
            else:
                writecomma = True
            if pred[0] == 'description':
                s += '{'
                commaCheck = False
                for descS in pred[1]:
                    if commaCheck:
                        s += ', '
                    else:
                        commaCheck = True
                    s += rule.writeDescPredDir(descS[0])
                    s += "?"
                    if descS[1] == '[':
                        s += '['
                        c = False
                        for subObj in descS[-1]:
                            if c:
                                s += ', '
                            else:
                                c = True
                            s += rule.writeDescPredDir(subObj[1:-1])
                        s += ']'
                    elif descS[1] == '{':
                        s += '{'
                        c = False
                        for subObj in descS[-1]:
                            if c:
                                s += ', '
                            else:
                                c = True
                            s += rule.writeDescPredDir(subObj[1:-1])
                        s += '}'
                    else:
                        s += rule.writeDescPredDir(descS[1])
                        s += rule.writeDescPredDir(descS[-1][1:-1])
                s += '}'
                        
            else:
                s += pred[0]+': {'
                checkComma = False
                for obj in pred[1]:
                    if not checkComma:
                        checkComma = True
                    else:
                        s += ', '
                    if isinstance(obj, list):
                        writeComma = False
                        s += '('
                        for i in obj:
                            if not writeComma:
                                writeComma = True
                            else:
                                s += ', '
                            if isinstance(i,list):
                                if pred[0] == 'void' or pred[0] == 'inside' or pred[0] == 'outside':
                                    s += '(' + str(i[0])
                                    for j in i[1:]:
                                        s += ', '+str(j)
                                    s += ')'
                                else:
                                    s += str(i[0])+'('+str(i[1])
                                    for j in i[2:]:
                                        s += ', '+str(j)
                                    s += ')'
                            else:
                                s += str(i)
                        s += ')'
                    else:
                        if pred[0] == 'bound':
                            s += obj.getBoundString()
                        else:
                            s += str(obj)
                s += '}'
        return s

    def getDirectives(self):
        """
        get all the rule's directives
        """
        return self.directiveString()

    def printDirectiveString(self):
        """
        print the rule's directives in string format 
        """
        print(self.directiveString())

    def directiveList(self):
        """
        get all the rule's directives
        """
        for i in self.detection.analysis.maps:
            if 'graphMap' in i:
                return self.detection.analysis.maps[i].directives
        return None

    def directiveString(self):
        """
        return the rule's directives in string format
        """
        directives = self.directiveList()
        s = ':: '
        writecomma = False
        for pred in directives:
            if writecomma:
                s += ', '
            else:
                writecomma = True
            if pred[0] == 'description':
                s += '{'
                c = False
                for cond in pred[1]:
                    if not c:
                        c = True
                    else:
                        s += ', '
                    func = cond[0]
                    if func == 'angle':
                        s += 'angle('+cond[1][0].name+'.direction, '+cond[1][1].name+'.direction):='
                        s += str(cond[1][2])
                    elif func == 'length':
                        s += cond[1][0].name+'.length:='
                        s += str(cond[1][1])
                s += '}'
            else:
                s += pred[0]+': {'
                checkComma = False
                index = 0
                while index < len(pred[1]):
                    obj = pred[1][index]
                    if isinstance(obj, list) and any(x is None for x in obj):
                        for ind0, i in enumerate(obj):
                            if i is None:
                                ind = ind0
                                break
                        pred[1].pop(index)
                        pred[1].insert(index, obj[ind+1:])
                        pred[1].insert(index, obj[:ind])
                    index += 1
                for obj in pred[1]:
                    if not checkComma:
                        checkComma = True
                    else:
                        s += ', '
                    if isinstance(obj, list):
                        writeComma = False
                        s += '('
                        for i in obj:
                            if not writeComma:
                                writeComma = True
                            else:
                                s += ', '
                            if isinstance(i,list):
                                s += str(i[0])+'('+str(i[1])
                                for j in i[2:]:
                                    s += ', '+str(j)
                                s += ')'
                            else:
                                s += str(i)
                        s += ')'
                    else:
                        s += str(obj)
                s += '}'
        return s

    @staticmethod
    def writeDescPredDir(string):
        s = ''
        reader = parseReader(string)
        while reader.token() is not None:
            if reader.token() == '`':
                reader.newToken()
                continue
            if reader.token() == '-':
                s += ' - '
                # file.write(' - ')
                reader.newToken()
                continue
            if __sortClass__.checkSort(reader.value()):
                if reader.trailerToken(3) == ':' and reader.trailerValue(4) == 'tag':
                                            # sortname (current)
                    reader.newToken()       # .
                    prop = reader.newValue()# property
                    reader.newToken()       # :
                    reader.newToken()       # key
                    reader.newToken()       # .
                    reader.newToken()       # value
                    reader.newToken()       # =
                    keyName = reader.newValue()[1:-1]

                    s += keyName+'.'+prop
                    # file.write(keyName+'.'+prop)
                    reader.newToken()
                    continue
            s += reader.value()
            # file.write(reader.value())
            reader.newToken()
        return s

    def evaluatePredicate(self, transfSet, tlhs, params = None):
        """
        check if the transformation in transfSet and params can satisfy all the predicates in the rule

        :param transfSet: transformation set used for solving predicates
        :param tlhs: the rule lhs transformed with the transfSet
        :param params: (optional) description transformation set
        :type transfSet: ~sortal.transform.transformSet.transformSet
        :type tlhs: ~sortal.form.form.form
        :type params: ~sortal.transform.paramSet.paramSet
        """
        if params is None:
            params = self.keys
        for i in self.detection.analysis.maps:
            if 'graphMap' in i or 'similarityMap' in i:
                if not self.detection.analysis.maps[i].evaluatePredicate(transfSet, self.lhs, tlhs, self.detection.form, params):
                    return False
                return True
        if 'map-parametric' in self.detection.analysis.maps:
            if not self.detection.analysis.maps['map-parametric'].evaluatePredicate(tlhs, params):
                return False
            return True
        return True
                    
    def solveDirectives(self, transfSet, tlhs, params = None):
        """
        solve the directives in the rule with transfSet and params

        :param transfSet: transformation set used for solving directives
        :param tlhs: the rule lhs transformed with the transfSet
        :param params: (optional) description transformation set
        :type transfSet: ~sortal.transform.transformSet.transformSet
        :type tlhs: ~sortal.form.form.form
        :type params: ~sortal.transform.paramSet.paramSet
        """
        if params is None:
            params = self.keys
        for i in self.detection.analysis.maps:
            if 'graphMap' in i:
                a = self.detection.analysis.maps[i].solveDirectives(transfSet, self.lhs, tlhs, self.rhs, self.detection.form, params)
                return a
        return [transfSet]

    def matches(self):
        """
        get all the matched forms from the rule applications
        """
        return self.detection.matches()
    
    def results(self, match):
        """
        get all the result forms of a single matched form from the rule applications
        """
        return self.detection.results(match)


class InvalidRuleError(Exception):
    """
    Custom Error that is raised when the rule created is invalid
    """
    def __init__(self, ruleObj):
        super(InvalidRuleError, self).__init__('Rule '+ ruleObj.name +' Creation Error: Invalid Rule')
        self.rule = ruleObj
        for mapID in self.rule.detection.analysis.maps:
            if 'map-distinguishable-similarityMap' in mapID :
                mapping = self.rule.detection.analysis.maps[mapID]
                if mapping.isValid():
                    continue
                self.casus = mapping.casus
                self.distinguishables = mapping.distinguishables
                self.discardedDist = mapping.discardedDist

class InvalidObjectError(Exception):
    """
    Custom Error that is raised when the rule created is invalid
    """
    def __init__(self, message):
        super(InvalidObjectError, self).__init__(message)
