# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import range
from builtins import super
from builtins import str

from .ruling import ruling, FlowTermination
from .flowApplication import flowApplication

from random import randint
from copy import copy

class ruleDisjunctionP(ruling):
    """
    ruleDisjunctionP: [* r1 r2 r3 ...] -- performing one of the ruling object in the bracket.
    The trial and error process is done randomly

    **Object Parameters:** 

    * **rules** (list of :class:`~sortal.grammar.rules.ruling.ruling`) -- the ruling objects contained in the ruleDisjunctionP.
    * **quantifier** -- the quantifier of the ruleDisjunction.
    * **matching** -- the matching property of the ruleDisjunction.
    """

    type = 'rule-disjunctionP'
    
    def __init__(self, rules, quantifier = '', matching = 'G', *args, **kwargs):
        super(ruleDisjunctionP, self).__init__()        
        self.rules = rules
        self.quantifier = quantifier
        self.matching = matching
        # self.lenS = []
        
    
    def apply(self, form, flowApp = None, backtrace = False, *args, **kwargs):
        """
        applying the rule disjunction onto a shape form. This function calls on greedyApply and lazyApply 
        functions below based on their matching and backtracking status:
        
        greedy matching: only greedyApply

        posessive matching: only greedyApply

        lazy matching: only lazyApply

        probabilistic matching: choose a matching number randomly, do greedyApply up to that number of times.
        if backtracking, backtrace with greedyApply. If all backtracking is exhausted, do lazy apply from the 
        application that reaches the randomly chosen matching number.

        :param form: the form to apply the ruling on; this form will be modified to 
            become the resulting form at the end of this method
        :type form: ~sortal.form.form.form
        :param flowApp: the flow application indicating the characteristic of the application 
        :type flowApp: ~sortal.grammar.rules.flowApplication.flowApplication
        :param backtrace: indicate if this is a backtrace from a flow
        :type backtrace: boolean
        :returns: the resulting form after applying the ruling; 
            the input form is also modified to become the resulting form
        """
        if flowApp is None:
            flowApp = flowApplication()
            backtrace = False

        if self.matching == self.PROBABILISTIC:
            traces = flowApp.getRulingTrace(self.toString(True))
            if traces is None:
                traces = []
                flowApp.setRulingTrace(self.toString(True), traces)
            if not backtrace:
                if self.quantifier[1] is None:
                    targetTrace = self.quantifier[0]
                    if targetTrace == 0: targetTrace = 1
                    while randint(0,1) == 1:
                        targetTrace += 1
                else:
                    targetTrace = randint(self.quantifier[0], self.quantifier[1])
                trace = [targetTrace, True, []]
                traces.append(trace)
            else:
                trace = traces[-1]
            if trace[1] is True:
                result = self.greedyApply(form, flowApp, backtrace)
                if result is False:
                    trace[1] = False
                else:
                    return result
            else:
                result = self.lazyApply(form, flowApp, True)
                if result is not False:
                    return result
            if trace[1] is False:
                while len(trace[2]) > 0:
                    appSet = trace[2].pop(0)
                    app = appSet[0]
                    # form.elements = []
                    # for i in appSet[1]:
                    #     form.elements.append(i)
                    flowApp.updateApp(app, self.toString(True))
                    result = self.lazyApply(appSet[1], flowApp, True)
                    if result is not False:
                        return result
                if result is False:
                    traces.pop(-1)
                    flowApp.setRulingTrace(self.toString(True), traces)
                    return False
            return result
        elif self.matching == self.LAZY:
            return self.lazyApply(form, flowApp, backtrace)
        return self.greedyApply(form, flowApp, backtrace)




    def greedyApply(self, form, flowApp, backtrace):
        """
        applying the rule disjunction onto a shape form in a greedy characteristic. Used for ruling with
        lazy or probabilistic matching. The Application would go for highest possible number of iterations (or
        up to a specific number for probabilistic) and backtrack down

        :param form: the form to apply the ruling on; this form will be modified to 
            become the resulting form at the end of this method
        :type form: ~sortal.form.form.form
        :param flowApp: the flow application indicating the characteristic of the application 
        :type flowApp: ~sortal.grammar.rules.flowApplication.flowApplication
        :param backtrace: indicate if this is a backtrace from a flow *only used in flow*
        :type backtrace: boolean
        :returns: the resulting form after applying the ruling; 
            the input form is also modified to become the resulting form
        """
        rulingCycles = flowApp.getRulingCycle(self.toString(True))

        if self.matching == self.PROBABILISTIC:
            traces = flowApp.getRulingTrace(self.toString(True))
            trace = traces[-1]
            tempMax = trace[0]
        else:
            tempMax = self.quantifier[1]

        # not backtrace -> fresh invocation of the disjunction:
        #  _ start new cycle
        #  _ determine the maximum number of cycles
        if not backtrace:
            # self.lenS.append([len(s)])

            # start new cycle
            cycle = [[form.duplicate(),0]]
            # trace is a list of number from 0 to len(self.rules), different from normal ruleDisjunction which is only a number
            rulingTrace = list(range(len(self.rules)))
            # # determine tempMax, if probabilistic + max quantifier is not None => random number between min + max quantifier
            # if self.matching == self.PROBABILISTIC and self.quantifier[1] is not None:
            #     self.tempMax = randint(self.quantifier[0], self.quantifier[1])
            # # otherwise use the max quantifier
            # else:
            #     self.tempMax = self.quantifier[1]
        # ############# OLD POSSESSIVE
        # elif self.matching == 'PO':
        #     l = self.lenS.pop()[0]
        #     while len(s) > l:
        #         flowApp.removePath()
        #     if flowApp.printProcess: self.printS(flowApp.path)
        #     return False

        # if "greedy" backtrace -> the last cycle fails
        #  _ pop out the last cycle, and continue from there.
        #  _ If there's only one member in cycle -> is back at the very beginning -> disjunction fails
        #  _ trace and form are set as the continuation from the last cycle
        else:
            if self.matching == self.POSSESSIVE:
                raise FlowTermination()
            # pop out the last cycle and assign it as "cycle" variable
            cycle = rulingCycles.pop(-1)
            # one cycle
            if len(cycle) == 1:
                return False
            # trace is set as the trace from the last cycle
            rulingTrace = cycle[-1][1]
            # if all ruling/rule choices are not exhausted for the last cycle -> form would be the second last cycle last form
            if len(rulingTrace) > 0:
                form = cycle[-2][0]
            cycle.pop(-1)
                                        
        # run this loop up until reaching tempMax
        while self.quantifier[1] is None or len(cycle)-1 < tempMax:

            # if all ruling/rule options run out -> pop the last cycle iteration
            if len(rulingTrace) == 0:
                cycle.pop(-1)
                rulingTrace = cycle[-1][1]
                rulingTrace.pop(0)
                
            # no backtrace -> fresh iteration -> start from the sequence first ruling/rule -> trace = full list
            if not backtrace:
                rulingTrace = list(range(len(self.rules)))

                # rIndex is a random number between 0 and len(trace)
                rIndex = randint(0, len(rulingTrace)-1)

                # swap trace[0] with trace[rIndex], making the first trace value to be "random"
                rulingTrace[0], rulingTrace[rIndex] = rulingTrace[rIndex], rulingTrace[0]

            # backtrace -> form being the last iteration result
            else:
                form = cycle[-1][0].duplicate()
            
            # go through the rulings/rules in the disjunction (rulingTrace being the list of indices left of the current ruling/rule)
            while len(rulingTrace) > 0:

                # if it's a rule, pop out one process string if backtrace
                if self.rules[rulingTrace[0]].type == 'rule':
                    if backtrace:
                        flowApp.removePath()
                    if flowApp.printProcess: self.printS(flowApp.path)

                # apply the ruling/rule represented by trace[0]
                try:
                    app = self.rules[rulingTrace[0]].apply(form, flowApp, backtrace = backtrace)
                except FlowTermination as ft:
                    raise ft
                except:
                    # if there's any error -> app = false
                    if flowApp.printProcess and self.rules[rulingTrace[0]].type == 'rule':
                        print(' ##'+self.rules[rulingTrace[0]].name+'## ', end='')
                    app = False
        
                backtrace = False

                # successfully applying the ruling/rule -> assign form as the result iteration is successful, continue to new cycle or end
                if app is not False:
                    form = app
                    cycle.append([form.duplicate(), rulingTrace])
                    if self.possessiveCheck is True:
                        self.setPoCheck(None)
                    break
                # if not successful
                else:
                    if self.matching == ruling.POSSESSIVE:
                        if self.rules[rulingTrace[0]].possessiveCheck is True:
                            self.setPoCheck(True)
                            return False
                        elif self.possessiveCheck is None:
                            self.setPoCheck(True)

                    # pop out the first trace
                    rulingTrace.pop(0)
                    # if no more ruling left 
                    if len(rulingTrace) == 0:
                        if self.possessiveCheck:
                            if flowApp.printProcess: print('>> POSSESSIVE RULING FAILS <<')
                            self.resetS(flowApp, cycle)
                            return False

                        # if the number of iterations is more than the minimum -> disjunction is successful
                        if len(cycle) > self.quantifier[0]:
                            rulingCycles.append(cycle)
                            return cycle[-1][0].duplicate()
                        # if only 1 iteration left in the cycle -> disjunction fails
                        if len(cycle) == 1:
                            return False
                            
                        # otherwise, turn on backtrace and go back to the last iteration
                        backtrace = True
                        rulingTrace = cycle[-1][1]
                        cycle.pop(-1)
                        
                        ####### ????????????
                        # flowApp.removePath()
                    else:
                        # rIndex is a random number between 0 and len(trace)
                        rIndex = randint(0, len(rulingTrace)-1)
                        # swap trace[0] with trace[rIndex], making the first trace value to be "random"
                        rulingTrace[0], rulingTrace[rIndex] = rulingTrace[rIndex], rulingTrace[0]

                    form = cycle[-1][0].duplicate()    
                    
            # # if PROBABILISTIC *, flip a coin to see if the disjunction is to be stopped or not
            # if self.matching == self.PROBABILISTIC and self.quantifier[1] is None:
            #     if randint(0,1) == 0:
            #         break
        # add the cycle into the list
        rulingCycles.append(cycle)
        # if PROBABILISTIC *,
        if self.matching == self.PROBABILISTIC:
            savedApp = flowApp.duplicate()
            savedForm = form.duplicate()
            trace[-1].append([savedApp, savedForm])

        return form


    def lazyApply(self, form, flowApp, backtrace):
        """
        applying the rule disjunction onto a shape form in a greedy characteristic. Used for ruling with
        lazy or probabilistic matching. The Application would go for lowest possible number of iterations (or
        from a specific number for probabilistic) and backtrack up

        :param form: the form to apply the ruling on; this form will be modified to 
            become the resulting form at the end of this method
        :type form: ~sortal.form.form.form
        :param flowApp: the flow application indicating the characteristic of the application 
        :type flowApp: ~sortal.grammar.rules.flowApplication.flowApplication
        :param backtrace: indicate if this is a backtrace from a flow *only used in flow*
        :type backtrace: boolean
        :returns: the resulting form after applying the ruling; 
            the input form is also modified to become the resulting form
        """

        # not backtrace -> fresh invocation of the disjunction:
        #  _ start new cycle
        #  _ determine the maximum number of cycles
        maxed = False
        rulingCycles = flowApp.getRulingCycle(self.toString(True))

        # not backtrace -> fresh invocation of the disjunction:
        #  _ start new cycle
        #  _ determine the maximum number of cycles
        if not backtrace:
            # start new cycle
            cycle = [[form.duplicate(),0]]
            rulingTrace = list(range(len(self.rules)))
            if self.quantifier[0] == 0:
                rulingCycles.append(cycle)
                return cycle[0][0].duplicate()

        # if "greedy" backtrace -> the last cycle fails
        #  _ pop out the last cycle, and continue from there.
        #  _ If there's only one member in cycle -> is back at the very beginning -> disjunction fails
        #  _ trace and form are set as the continuation from the last cycle
        else:
            # pop out the last cycle and assign it as "cycle" variable
            cycle = rulingCycles.pop(-1)

            # one cycle
            if self.quantifier[1] is not None and len(cycle) > self.quantifier[1]:
                maxed = True
                # trace is set as the trace from the last cycle
                rulingTrace = cycle[-1][1]
                cycle.pop()
            form = cycle[-1][0].duplicate()

        # run this loop up until reaching max (the loop should be broken halfway for most situations)
        while maxed or self.quantifier[1] is None or len(cycle)-1 < self.quantifier[1]:
            # if all ruling/rule options run out -> pop the last cycle iteration
            if maxed and len(rulingTrace) == 0:
                cycle.pop(-1)
                rulingTrace = cycle[-1][1]
            
            # not maxed -> fresh iteration -> start from the sequence first ruling/rule -> trace = full list
            if not maxed:
                rulingTrace = list(range(len(self.rules)))
                # rIndex is a random number between 0 and len(trace)
                rIndex = randint(0, len(rulingTrace)-1)
                # swap trace[0] with trace[rIndex], making the first trace value to be "random"
                rulingTrace[0], rulingTrace[rIndex] = rulingTrace[rIndex], rulingTrace[0]

            # go through the rulings/rules in the disjunction (rulingTrace being the index of the current ruling/rule)
            while len(rulingTrace) > 0:
                
                # if it's a rule, pop out one process string if backtrace
                if self.rules[rulingTrace[0]].type == 'rule':
                    if maxed:
                        flowApp.removePath()
                    if flowApp.printProcess: self.printS(flowApp.path)

                # apply the ruling/rule represented by trace
                try:
                    app = self.rules[rulingTrace[0]].apply(form, flowApp, backtrace = maxed)
                except:
                    # if there's any error -> app = false
                    if flowApp.printProcess and self.rules[rulingTrace[0]].type == 'rule':
                        print(' ##'+self.rules[rulingTrace[0]].name+'## ', end='')
                    app = False

                maxed = False
                # successfully applying the ruling/rule -> assign form as the result iteration is successful, continue to new cycle or end
                if app is not False:
                    form = app
                    cycle.append([form.duplicate(), rulingTrace])
                    # if trace is at first ruling/rule and it's above the min number of iterations (quantifier[0]) -> finish the sequence
                    if len(cycle) > self.quantifier[0]:
                        rulingCycles.append(cycle)
                        return form

                    break

                # if not successful
                else:
                    # increase trace
                    rulingTrace.pop(0)
                    # if no more ruling left 
                    if len(rulingTrace) == 0:
                        # if only 1 iteration left in the cycle -> disjunction fails
                        if len(cycle) == 1:
                            return False
                            
                        # otherwise, turn on backtrace and go back to the last iteration
                        maxed = True
                        rulingTrace = cycle[-1][1]
                        cycle.pop(-1)
                    else:
                        # rIndex is a random number between 0 and len(trace)
                        rIndex = randint(0, len(rulingTrace)-1)
                        # swap trace[0] with trace[rIndex], making the first trace value to be "random"
                        rulingTrace[0], rulingTrace[rIndex] = rulingTrace[rIndex], rulingTrace[0]

                    form = cycle[-1][0].duplicate()

        # add the cycle into the list
        rulingCycles.append(cycle)
        return form


    def resetS(self, flowApp, cycle = None):
        rulingCycles = flowApp.getRulingCycle(self.toString(True))
        if cycle is None:
            cycle = rulingCycles.pop()
        while len(cycle) > 1:
            trace = cycle.pop()[1]
            if self.rules[trace[0]].type == 'rule' or self.rules[trace[0]].type == 'rule-negation':
                flowApp.removePath()
            else:
                self.rules[trace[0]].resetS(flowApp)

    def manualApplyPeek(self, form, flowApp, newInvoke = False):
        """
        Get all the possible first step applications of the ruling on the form based on the flowApp

        :param form: the form to apply the ruling on; this form will be modified to 
            become the resulting form at the end of this method
        :type form: ~sortal.form.form.form
        :param flowApp: the flow application indicating the characteristic of the application 
        :type flowApp: ~sortal.grammar.rules.flowApplication.flowApplication
        :param newInvoke: indicate if this is a new invocation of the ruling (i.e. not continuing from previous manualApplyPeek)
        :type newInvoke: boolean
        :returns: a list of two values: 
            first value is a list of flowApplication objects that are the different possible step results of the flow, 
            second value is a boolean indicating whether the ruling is already satisfied -> can continue on to the next ruling
            or end the flow if it's the topmost ruling in the flow
        """
        if newInvoke:
            cycle = 0
            trace = None
        else:
            cycle = flowApp.getRulingCycle(self.toString(True))
            trace = flowApp.getRulingTrace(self.toString(True))
        if trace is not None and self.rules[trace].type not in ('rule-sequence', 'rule-disjunction', 'rule-disjunctionP'):
            cycle += 1
            trace = None
            flowApp.setRulingTrace(self.toString(True), None)
            if self.quantifier[1] is not None and cycle >= self.quantifier[1]:
                return [[], True]
            

        opFlowApp = flowApp.duplicate()
        resultApps = []
        opFlowApp.rulingCycles[self.toString(True)] = cycle

        contCheck = False
        if cycle >= self.quantifier[0]:
            contCheck = True
        while True:
            if trace is not None:
                app = self.rules[trace].manualApplyPeek(form, opFlowApp, False)
                if app[1] == False:
                    if len(app[0]) == 1:
                        opFlowApp = app[0][0]
                        continue
                    resultApps += app[0]
                    break
                resultApps += app[0]
                cycle += 1
                trace = None
                if self.quantifier[1] is not None and cycle >= self.quantifier[1]:
                    return [resultApps, True]
                continue
            for ruleIndex in range(len(self.rules)):
                newFlowApp = opFlowApp.duplicate()
                newFlowApp.setRulingTrace(self.toString(True), ruleIndex)
                app = self.rules[ruleIndex].manualApplyPeek(form, newFlowApp, True)
                resultApps += app[0]
            break
        if cycle >= self.quantifier[0]:
            contCheck = True
        return [resultApps, contCheck]