# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import range
from builtins import super
from builtins import str

from copy import copy
from .ruling import ruling, FlowTermination
from .flowApplication import flowApplication
from random import randint

count = 0

class ruleSequence(ruling):
    """
    ruleSequence: (r1 r2 r3 ...) -- performing the ruling objects in sequence

    **Object Parameters:** 

    * **rules** (list of :class:`~sortal.grammar.rules.ruling.ruling`) -- the ruling objects contained in the ruleSequence.
    * **quantifier** -- the quantifier of the ruleSequence.
    * **matching** -- the matching property of the ruleSequence.
    """
    type = 'rule-sequence'
    
    def __init__(self, rules, quantifier = [1,1], matching = 'G', *args, **kwargs):
        super(ruleSequence, self).__init__()
        self.rules = rules
        self.quantifier = quantifier
        self.matching = matching

        self.rulingTrace = 0

        # ############# OLD POSSESSIVE
        # self.lenS = []
        # #############

        
    def apply(self, form, flowApp = None, backtrace = False, *args, **kwargs):
        """
        applying the rule Sequence onto a shape form. This function calls on greedyApply and lazyApply 
        functions below based on their matching and backtracking status:
        
        greedy matching: only greedyApply

        posessive matching: only greedyApply

        lazy matching: only lazyApply

        probabilistic matching: choose a matching number randomly, do greedyApply up to that number of times.
        if backtracking, backtrace with greedyApply. If all backtracking is exhausted, do lazy apply from the 
        application that reaches the randomly chosen matching number.

        :param form: the form to apply the ruling on; this form will be modified to 
            become the resulting form at the end of this method
        :type form: ~sortal.form.form.form
        :param flowApp: the flow application indicating the characteristic of the application 
        :type flowApp: ~sortal.grammar.rules.flowApplication.flowApplication
        :param backtrace: indicate if this is a backtrace from a flow
        :type backtrace: boolean
        :returns: the resulting form after applying the ruling; 
            the input form is also modified to become the resulting form
        """
        if flowApp is None:
            flowApp = flowApplication()

        if self.matching == self.PROBABILISTIC:
            traces = flowApp.getRulingTrace(self.toString(True))
            if traces is None:
                traces = []
                flowApp.setRulingTrace(self.toString(True), traces)
            if not backtrace:
                if self.quantifier[1] is None:
                    targetTrace = self.quantifier[0]
                    if targetTrace == 0: targetTrace = 1
                    while randint(0,1) == 1:
                        targetTrace += 1
                else:
                    targetTrace = randint(self.quantifier[0], self.quantifier[1])
                trace = [targetTrace, True, []]
                traces.append(trace)
            else:
                trace = traces[-1]
            if trace[1] is True:
                result = self.greedyApply(form, flowApp, backtrace)
                if result is False:
                    trace[1] = False
                else:
                    return result
            else:
                result = self.lazyApply(form, flowApp, True)
                if result is not False:
                    return result
            if trace[1] is False:
                while len(trace[2]) > 0:
                    appSet = trace[2].pop(0)
                    app = appSet[0]
                    # form.elements = []
                    # for i in appSet[1]:
                    #     form.elements.append(i)
                    flowApp.updateApp(app, self.toString(True))
                    result = self.lazyApply(appSet[1], flowApp, True)
                    if result is not False:
                        return result
                if result is False:
                    traces.pop(-1)
                    flowApp.setRulingTrace(self.toString(True), traces)
                    return False
            return result
        elif self.matching == self.LAZY:
            return self.lazyApply(form, flowApp, backtrace)
        return self.greedyApply(form, flowApp, backtrace)



    def greedyApply(self, form, flowApp, backtrace):
        """
        applying the rule Sequence onto a shape form in a greedy characteristic. Used for ruling with
        lazy or probabilistic matching. The Application would go for highest possible number of iterations (or
        up to a specific number for probabilistic) and backtrack down

        :param form: the form to apply the ruling on; this form will be modified to 
            become the resulting form at the end of this method
        :type form: ~sortal.form.form.form
        :param flowApp: the flow application indicating the characteristic of the application 
        :type flowApp: ~sortal.grammar.rules.flowApplication.flowApplication
        :param backtrace: indicate if this is a backtrace from a flow *only used in flow*
        :type backtrace: boolean
        :returns: the resulting form after applying the ruling; 
            the input form is also modified to become the resulting form
        """
        rulingCycles = flowApp.getRulingCycle(self.toString(True))

        if self.matching == self.PROBABILISTIC:
            traces = flowApp.getRulingTrace(self.toString(True))
            trace = traces[-1]
            tempMax = trace[0]
        else:
            tempMax = self.quantifier[1]

        # not backtrace -> fresh invocation of the sequence:
        #  _ start new cycle
        #  _ determine the maximum number of cycles
        if not backtrace:

            # ############# OLD POSSESSIVE
            # self.lenS.append([len(s)])
            # #############
            if self.possessiveCheck is True:
                self.setPoCheck(None)

            # start new cycle
            cycle = [form.duplicate()]

            # # determine tempMax, if probabilistic + max quantifier is not None => random number between min + max quantifier
            # if self.matching == self.PROBABILISTIC and self.quantifier[1] is not None:
            #     self.tempMax = trace[0]
            # # otherwise use the max quantifier
            # else:
            #     self.tempMax = self.quantifier[1]

        # if "greedy" backtrace -> the last cycle fails
        #  _ pop out the last cycle, and continue from there.
        #  _ If there's only one member in cycle -> is back at the very beginning with 0 loop -> sequence fails
        else:
            if self.matching == self.POSSESSIVE:
                raise FlowTermination()
            cycle = rulingCycles.pop()
            if len(cycle) == 1:
                return False
            cycle.pop()

        # run this loop up until reaching tempMax
        while tempMax is None or len(cycle)-1 < tempMax:
            # no backtrace -> fresh invocation -> start from the sequence first ruling/rule -> trace = 0
            if not backtrace:
                self.rulingTrace = 0
            # backtrace -> back up by one ruling/rule (in the sequence rule list)
            else:
                self.rulingTrace = len(self.rules) - 1
            
            # go through the rulings/rules in the sequence (rulingTrace being the index of the current ruling/rule)
            while self.rulingTrace < len(self.rules):
                # if it's a rule, pop out one process string if backtrace
                if self.rules[self.rulingTrace].type == 'rule':
                    if backtrace:
                        flowApp.removePath()
                    if flowApp.printProcess: self.printS(flowApp.path)
                # apply the ruling/rule represented by trace
                try:
                    app = self.rules[self.rulingTrace].apply(form, flowApp, backtrace = backtrace)
                except FlowTermination as ft:
                    raise ft
                except:
                    
                    # if there's any error -> app = false
                    if flowApp.printProcess and self.rules[self.rulingTrace].type == 'rule':
                        print(' ##'+self.rules[self.rulingTrace].name+'## ', end='')
                    app = False
                # successfully applying the ruling/rule -> assign form as the result, no backtrace
                if app is not False:
                    form = app
                    backtrace = False
                    if self.possessiveCheck is True:
                        self.setPoCheck(None)

                # if not successful
                else:
                    if self.matching == ruling.POSSESSIVE:
                        if self.rules[self.rulingTrace].possessiveCheck is True:
                            self.setPoCheck(True)
                            self.resetS(flowApp)
                            return False
                        elif self.possessiveCheck is None:
                            self.setPoCheck(True)

                    # if trace is at first ruling/rule -> check if it's above the min number of iterations (quantifier[0])
                    #  _ if above -> finish the sequence
                    #  _ if below -> remove the latest iteration, backtrack to the previous one 
                    if self.rulingTrace == 0:
                        if self.possessiveCheck:
                            if flowApp.printProcess: print('>> POSSESSIVE RULING FAILS <<')
                            self.resetS(flowApp)
                            return False
                        if len(cycle) > self.quantifier[0]:
                            rulingCycles.append(cycle)
                            return cycle[-1].duplicate()
                        # only one cycle left -> sequence fails
                        if len(cycle) == 1:
                            return False
                        # remove the latest cycle, backtrack to the previous one, trace = the last rule 
                        # (combination of trance = len(self.rules) and the next line trace -= 1)
                        cycle.pop()
                        self.rulingTrace = len(self.rules)
                    # reduce trace by one
                    self.rulingTrace -= 1
                    
                    # set backtrace as true
                    backtrace = True
                    if flowApp.printProcess: print('>>backtrace<<', end = '')

                    # skip to the next round of the loop
                    continue
                
                # increasing trace -> continue to the next ruling/rule
                self.rulingTrace += 1
            
            # ############# OLD POSSESSIVE
            # self.lenS[-1].append(len(s))
            # #############

            # after finishing all the rulings/rules in the sequence, add a cycle
            cycle.append(form.duplicate())
            
            # # if PROBABILISTIC *, flip a coin to see if the sequence is to be stopped or not
            # if self.matching == self.PROBABILISTIC and self.quantifier[1] is None:
            #     if randint(0,1) == 0:
            #         break

        # add the cycle into the list
        rulingCycles.append(cycle)
        # if PROBABILISTIC *,
        if self.matching == self.PROBABILISTIC:
            savedApp = flowApp.duplicate()
            savedForm = form.duplicate()
            trace[-1].append([savedApp, savedForm])
        return form


    def lazyApply(self, form, flowApp, backtrace):
        """
        applying the rule Sequence onto a shape form in a greedy characteristic. Used for ruling with
        lazy or probabilistic matching. The Application would go for lowest possible number of iterations (or
        from a specific number for probabilistic) and backtrack up

        :param form: the form to apply the ruling on; this form will be modified to 
            become the resulting form at the end of this method
        :type form: ~sortal.form.form.form
        :param flowApp: the flow application indicating the characteristic of the application 
        :type flowApp: ~sortal.grammar.rules.flowApplication.flowApplication
        :param backtrace: indicate if this is a backtrace from a flow *only used in flow*
        :type backtrace: boolean
        :returns: the resulting form after applying the ruling; 
            the input form is also modified to become the resulting form
        """
        # not backtrace -> fresh invocation of the sequence:
        #  _ start new cycle
        #  _ determine the temporary maximum number of loops

        # maxed: check for whether the sequence has reached the maximum, used in place of backtrace for most of the operation
        maxed = False
        rulingCycles = flowApp.getRulingCycle(self.toString(True))
        if not backtrace:
            # start new cycle
            cycle = [form.duplicate()]
            if self.quantifier[0] == 0:
                rulingCycles.append(cycle)
                return cycle[0]

        # if "lazy" backtrace -> the last cycle fails
        #  _ pop out the last cycle, and continue from there.
        else:
            cycle = rulingCycles.pop()
            if self.quantifier[1] is not None and len(cycle) > self.quantifier[1]:
                maxed = True
                cycle.pop()
            form = cycle[-1].duplicate()

        # run this loop up until reaching max (the loop should be broken halfway for most situations)
        while maxed or self.quantifier[1] is None or len(cycle)-1 < self.quantifier[1]:
            # not maxed -> fresh invocation -> start from the sequence first ruling/rule -> trace = 0
            if not maxed:
                self.rulingTrace = 0
            # maxed -> back up by one ruling/rule (in the sequence rule list)
            else:
                self.rulingTrace = len(self.rules) - 1
            
            # go through the rulings/rules in the sequence (rulingTrace being the index of the current ruling/rule)
            while self.rulingTrace < len(self.rules):
                # if it's a rule, pop out one process string if maxed
                if self.rules[self.rulingTrace].type == 'rule':
                    if maxed:
                        flowApp.removePath()
                    if flowApp.printProcess: self.printS(flowApp.path)
                # apply the ruling/rule represented by trace
                
                try:
                    app = self.rules[self.rulingTrace].apply(form, flowApp, backtrace = maxed)
                except:
                    # if there's any error -> app = false
                    if flowApp.printProcess and self.rules[self.rulingTrace].type == 'rule':
                        print(' ##'+self.rules[self.rulingTrace].name+'## ', end='')
                    app = False
                # successfully applying the ruling/rule -> assign form as the result, no more backtrace, reset maxed
                if app is not False:
                    form = app
                    maxed = False
                    backtrace = False

                # if not successful
                else:
                    # if trace is at first ruling/rule -> check if it's above the min number of iterations (quantifier[0])
                    #  _ if above -> finish the sequence
                    #  _ if below -> remove the latest iteration, backtrack to the previous one 
                    if self.rulingTrace == 0:
                        # if len(cycle) > self.quantifier[0]:
                        #     rulingCycles.append(cycle)
                        #     return cycle[-1]
                        # only one cycle left -> sequence fails
                        if len(cycle) == 1:
                            return False
                        # remove the latest cycle, backtrack to the previous one, trace = the last rule 
                        # (combination of trance = len(self.rules) and the next line trace -= 1)
                        cycle.pop()
                        self.rulingTrace = len(self.rules)
                    # reduce trace by one
                    self.rulingTrace -= 1
                    
                    # set backtrace as true
                    backtrace = True
                    maxed = True
                    if flowApp.printProcess: print('>>backtrace<<', end = '')

                    # skip to the next round of the loop
                    continue
                
                # increasing trace -> continue to the next ruling/rule
                self.rulingTrace += 1
            # after finishing all the rulings/rules in the sequence, add a cycle
            cycle.append(form.duplicate())
            # if trace is at first ruling/rule and it's above the min number of iterations (quantifier[0]) -> finish the sequence
            if len(cycle) > self.quantifier[0]:
                rulingCycles.append(cycle)
                return cycle[-1].duplicate()
        # add the cycle into the list
        rulingCycles.append(cycle)
        return form

    def totalRules(self):
        """
        total number of rules inside the sequence (including rules in subordinate rulings)
        """
        ctrace = 0
        for i in range(len(self.rules)):
            if self.rules[i].type == 'rule':
                ctrace+=1
            else:
                ctrace+= self.rules[i].totalRules('max')
        return ctrace
    
    def resetS(self, flowApp, cycle = None):
        rulingCycles = flowApp.getRulingCycle(self.toString(True))
        if cycle is None:
            cycle = rulingCycles.pop()

        if self.rulingTrace == 0:
            for j in range(0, len(cycle) - 1):
                for i in range(0, len(self.rules)):
                    if self.rules[i].type == 'rule':
                        flowApp.removePath()
                    else:
                        self.rules[i].resetS(flowApp)
        else:
            for i in range(0, self.rulingTrace):
                if self.rules[i].type == 'rule' or self.rules[i].type == 'rule-negation':
                    flowApp.removePath()
                else:
                    self.rules[i].resetS(flowApp)

    def manualApplyPeek(self, form, flowApp, newInvoke = False):
        """
        Get all the possible first step applications of the ruling on the form based on the flowApp

        :param form: the form to apply the ruling on; this form will be modified to 
            become the resulting form at the end of this method
        :type form: ~sortal.form.form.form
        :param flowApp: the flow application indicating the characteristic of the application 
        :type flowApp: ~sortal.grammar.rules.flowApplication.flowApplication
        :param newInvoke: indicate if this is a new invocation of the ruling (i.e. not continuing from previous manualApplyPeek)
        :type newInvoke: boolean
        :returns: a list of two values: 
            first value is a list of flowApplication objects that are the different possible step results of the flow, 
            second value is a boolean indicating whether the ruling is already satisfied -> can continue on to the next ruling
            or end the flow if it's the topmost ruling in the flow
        """
        if newInvoke:
            cycle = 0
            trace = 0
            flowApp.rulingCycles[self.toString(True)] = cycle
            flowApp.setRulingTrace(self.toString(True), trace)
        else:
            cycle = flowApp.getRulingCycle(self.toString(True))
            trace = flowApp.getRulingTrace(self.toString(True))


        opFlowApp = flowApp.duplicate()
        resultApps = []

        continueCheck = False
        skipCheck = True

        if not newInvoke and self.rules[trace].type not in ('rule-sequence', 'rule-disjunction', 'rule-disjunctionP'):
            trace += 1
            newInvoke = True
        
        while True:
            if trace == len(self.rules):
                cycle += 1
                trace = 0

            opFlowApp.rulingCycles[self.toString(True)] = cycle
            opFlowApp.setRulingTrace(self.toString(True), trace)

            if trace == 0 and cycle >= self.quantifier[0]:
                if continueCheck or (self.quantifier[1] is not None and cycle >= self.quantifier[1]):
                    if not skipCheck:
                        return [[opFlowApp], False]
                    return [resultApps, skipCheck]
                continueCheck = True

            app = self.rules[trace].manualApplyPeek(form, opFlowApp, newInvoke)
            if app[1] == False:
                if len(app[0]) == 1:
                    opFlowApp = app[0][0]
                    trace += 1
                    newInvoke = True
                    skipCheck = False
                    continue
                resultApps += app[0]
                break
            resultApps += app[0]
            trace += 1
            newInvoke = True
        return [resultApps, continueCheck]
