# -*- coding: utf-8 -*-
"""
Quadratic bezier curve is defined by the quadratic equation: B(t) = a * t^2 + b * t + c.
"""
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import range
from builtins import super
from builtins import str

from .individual import individual
from .point import point2D, point3D
from .plane import plane3D
from ..struct.coordinate import c0, c1, c0p5, coordinate
from ..struct.vector import vector2D, vector3D
from ..transform.affineTransform import affineTransform2D, affineTransform3D
from ..transform.graphTransform import graphTransform
from ..io.parseReader import ParsingError

v0 = vector2D(0,0)

class bezierCurve2D(individual):
    """
    Quadratic Bezier curve in 2D. The bezier curve is defined by the quadratic equation:

    | B(t) = a * t^2 + b * t + c
    | * a, b, c: constant 2D vectors defining the equation
    | * t: the numeric variable of the function; B(0) would be the start point of the curve while B(1) would be the end point.

    **Object Parameters:** (not including inherited attributes)

    * **a** (:class:`~sortal.struct.vector.vector2D`) -- a vector in the bezier function.
    * **b** (:class:`~sortal.struct.vector.vector2D`) -- b vector in the bezier function.
    * **c** (:class:`~sortal.struct.vector.vector2D`) -- c vector in the bezier function.
    * **start** (:class:`~sortal.struct.vector.vector2D`) -- start point of the bezier curve (t = 0).
    * **controlPoint** (:class:`~sortal.struct.vector.vector2D`) -- control point of the bezier curve.
    * **end** (:class:`~sortal.struct.vector.vector2D`) -- end point of the bezier curve (t = 1).
    * **A** (:class:`~sortal.struct.coord.coordinate_float.coordinate`) -- 4(a.x^2 + a.y^2)
    * **B** (:class:`~sortal.struct.coord.coordinate_float.coordinate`) -- 4(a.x * b.x + a.y * b.y)
    * **C** (:class:`~sortal.struct.coord.coordinate_float.coordinate`) -- b.x^2 + b.y^2
    * **D** (:class:`~sortal.struct.coord.coordinate_float.coordinate`) -- 2(a.y * b.x - a.x * b.y)

    **Class Properties:**

    * **Behavior**: :class:`~sortal.form.curve.curve`
    * **Sort Argument**: :class:`~sortal.sort.parameter.argument.argument.argumentType`.NONE
    * **Mapping**: :class:`~sortal.map.similarityMap.similarityMap2D`
    """

    type = 'ind-bezierCurve2D'

    def __init__(self, sort, p1, p2, p3, isFunction = False, *args, **kwargs):
        """
        Initializing a 2D bezier curve.

        :param sort: sort of the bezier curve
        :type sort: ~sortal.sort.sort.sort
        :param p1: start point of the curve (if isFunction is false) | value of a in the function (if isFunction is true)
        :type p1: ~sortal.struct.vector.vector2D
        :param p2: control point of the curve (if isFunction is false) | value of b in the function (if isFunction is true)
        :type p2: ~sortal.struct.vector.vector2D
        :param p3: end point of the curve (if isFunction is false) | value of c in the function (if isFunction is true)
        :type p3: ~sortal.struct.vector.vector2D
        :param isFunction: if False, the curve initialization takes 3 points on the curve (start, end and control point) if True, the curve initialization takes the value of a, b and c of the function
        :type isFunction: boolean
        """
        super(bezierCurve2D, self).__init__(sort)
        if isFunction:
            self.a = p1
            self.b = p2
            self.c = p3
            self.start = p3
            self.controlPoint = self.b.scale(c0p5).add(p3)
            self.end = self.a.subtract(p3).add(self.controlPoint).add(self.controlPoint)
        else:
            self.a = p3.add(p1).subtract(p2.scale(2))
            self.b = (p2.subtract(p1)).scale(2)
            self.c = p1
            self.start = p1
            self.controlPoint = p2
            self.end = p3
        self.A = 4 * (self.a.x ** 2 + self.a.y ** 2)
        self.B = 4 * (self.a.x * self.b.x + self.a.y * self.b.y)
        self.C = self.b.x ** 2 + self.b.y ** 2
        self.D = 2 * (self.a.y * self.b.x - self.a.x * self.b.y)
        self.evaluateCurve()

    def vertex(self):
        """
        vertex of the curve (positon on the curve at t = -B/2A)
        
        :returns: :class:`~sortal.struct.vector.vector2D` -- vertex vector
        """
        t = -self.B/(2*self.A)
        return self.pos(t)
    
    def evaluateCurve(self):
        """
        Re-evaluate the orientation of the curve
        """
        # find gradient of the curve at vertex
        t = -self.B/(2*self.A)
        vg = self.vectGradient(t)

        # reverse the curve's order if vertex gradient is negative (or vertex gradient is 0 and gradient at t-1 is negative)
        if vg < v0 or (vg == v0 and self.vectGradient(t-1) < v0):
            self.start, self.end = self.end, self.start
            self.b = self.a.scale(-2).subtract(self.b)
            self.c = self.start
            self.B = 4 * (self.a.x * self.b.x + self.a.y * self.b.y)
            self.C = self.b.x ** 2 + self.b.y ** 2
            self.D = 2 * (self.a.y * self.b.x - self.a.x * self.b.y)
    
    def geometricFrame(self):
        """
        the geometric frame of the curve
        """
        return [[min(self.start.x, self.end.x, self.controlPoint.x), min(self.start.y, self.end.y, self.controlPoint.y), None],[max(self.start.x, self.end.x, self.controlPoint.x), max(self.start.y, self.end.y, self.controlPoint.y), None]]
        
    def toString(self, printKey):
        """
        converting the curve to string format
        """
        s = ''
        if printKey and self.key is not None:
            s += str(self.key)+' '
        return s + '<'+str(self.start)+','+str(self.controlPoint)+','+str(self.end)+'>'
        # return str(self.a)+'*t^2 + '+str(self.b)+'*t + '+str(self.c)
    
    def getValue(self):
        return str(self)

    def printVal(self):
        """
        print out a string representing the value of this individual, used in form.printE()
        """
        return '<'+str(self.start)+','+str(self.controlPoint)+','+str(self.end)+'>'

    def printEquation(self):
        return str(self.a)+'*t^2 + '+str(self.b)+'*t + '+str(self.c)

    def printRhino(self, dX = 0, dY = 0):
        """
        print the individual in a format that can be pasted into Rhinoceros for visualization
        """
        s = ''
        s += '_Curve '+str(self.start.x + dX)+','+str(self.start.y + dY)+'  '+str(self.controlPoint.x + dX)+','+str(self.controlPoint.y + dY)+'  '+str(self.end.x + dX)+','+str(self.end.y + dY)+' EnterEnd'
        return s
        
    def equals(self, other):
        """
        check equality between two bezier curves

        :param other: <individual> - other object to be compared with
        :returns: boolean -- whether the curve and the other object are the same geometrically
        """
        return self.a == other.a and self.b == other.b and self.c == other.c
    
    def isNil(self):
        """
        check if the curve is Nil (start == controlPoint == end)
        """
        return self.start.x == self.controlPoint.x == self.end.x and self.start.y == self.controlPoint.y == self.end.y
        
    def compareValue(self, other, params = None):
        """
        Comparing the geometric value of this curve with another
        (consider the two curves as equal if they are on the exact same curve)
        
        :param other: other individual to be compared
        :type other: ~sortal.ind.bezierCurve.bezierCurve2D
        :returns: int -- -1 if this curve is smaller, 1 if bigger, 0 if equal
        """

        if params is None:
            l = 1/(((other.A**(3/2)) * (self.D**2))/((self.A**(3/2)) * (other.D**2)))
            phi = (self.A * other.D)/(other.A * self.D)
            v = (self.B * other.D - other.B * self.D)/(other.A * self.D * 2)
        else:
            l, phi, v = params
        c = self.pos(0).compareValue(other.pos(v).scale(l))
        if c != 0:
            return c
        return self.pos(1).compareValue(other.pos(phi+v).scale(l))
    
    def compareInterval(self, other):
        """
        Comparing the geometric value of this curve with another
        (compare the position of the start and end points of the two curves if they are on the exact same curve)

        :param other: other individual to be compared
        :type other: ~sortal.ind.bezierCurve.bezierCurve2D
        :returns: int -- -1 if this curve is smaller, 1 if bigger, 0 if equal
        """
        l = 1/(((other.A**(3/2)) * (self.D**2))/((self.A**(3/2)) * (other.D**2)))
        phi = (self.A * other.D)/(other.A * self.D)
        v = (self.B * other.D - other.B * self.D)/(other.A * self.D * 2)
        c = self.compareValue(other, params = (l,phi,v))
        if c != 0:
            return c
        c = v.compareValue(c0)
        if c != 0:
            return c
        c = (phi+v).compareValue(c1)
        if c != 0:
            return c
        return 0
        
    def pos(self, t):
        """
        find the position of a point on the curve given the t value
        
        :param t: other individual to be compared
        :type t: ~sortal.struct.coord.coordinate_float.coordinate
        :returns: :class:`~sortal.struct.vector.vector2D` -- position of the point on the curve with t value
        """
        tc = coordinate(t)
        return self.a.scale(tc**2).add(self.b.scale(tc)).add(self.c)
        
    def curvature(self, t):
        """
        find the curvature at a point on the curve given the t value

        :param t: t value
        :type t: ~sortal.struct.coord.coordinate_float.coordinate
        :returns: :class:`~sortal.struct.coord.coordinate_float.coordinate` -- curvature value
        """
        tc = coordinate(t)
        return self.D / ((self.A * (tc ** 2)+ self.B * tc + self.C)**(3/2))
                
    def gradient(self, t):
        """
        find the gradient at a point on the curve given the t value
        
        :param t: t value
        :type t: ~sortal.struct.coord.coordinate_float.coordinate
        :returns: :class:`~sortal.struct.coord.coordinate_float.coordinate` -- gradient value
        """
        dx = 2 * self.a.x * t + self.b.x
        dy = 2 * self.a.y * t + self.b.y
        if dx == 0:
            return None
        return dy/dx
        
    def vectGradient(self, t):
        """
        find the gradient vector at a point on the curve given the t value

        :param t: t value
        :type t: ~sortal.struct.coord.coordinate_float.coordinate
        :returns: :class:`~sortal.struct.vector.vector2D` -- gradient vector
        """
        dx = 2 * self.a.x * t + self.b.x
        dy = 2 * self.a.y * t + self.b.y
        return vector2D(dx, dy)
                
    def codescriptor(self):
        """
        codescriptor of the curve: the curve's vertex

        :returns: sortal.point2D - codescriptor
        """
        point = point2D(self.sort, self.vertex())
        point.assoc = self
        return point
        
    def solveT(self, point):
        """
        solve for t value given its position vector

        :param point: coordinates of a point on the curve
        :type point: ~sortal.struct.vector.vector2D
        :returns: :class:`~sortal.struct.coord.coordinate_float.coordinate` -- if point is not on the curve, returns None, else corresponding t value
        """
        a = self.a.x
        b = self.b.x
        c = self.c.x - point.x
        
        if a == 0:
            x1 = -c/b
            x2 = None
        else:
            d = b*b-4*a*c # discriminant
            if d < 0:
                return None
            elif d == 0:
                x1 = (-b)/(2*a)
                x2 = None
            else:
                x1 = (-b+d.sqrt())/(2*a)
                x2 = (-b-d.sqrt())/(2*a)
            
        a = self.a.y
        b = self.b.y
        c = self.c.y - point.y
        if a == 0:
            y1 = -c/b
            y2 = None
        else:
            d = b**2-4*a*c
            
            if d < 0:
                return None
            elif d == 0:
                y1 = (-b)/(2*a)
                y2 = None
            else:
                y1 = (-b+d.sqrt())/(2*a)
                y2 = (-b-d.sqrt())/(2*a)
        x1r = round(x1.value,3)
        y1r = round(y1.value,3)
        if x1r == y1r:
            return x1
        if y2 is not None and x1r == round(y2.value,3):
            return x1
        if x2 is not None:
            x2r = round(x2.value,3)
            if x2r == y1r:
                return x2
            if y2 is not None and x2r == round(y2.value,3):
                return x2
        return None
        
    def containsPoint(self, point):
        """
        check if the curve contains a point

        :param point: the vector coordinates of the point to be checked
        :type point: ~sortal.struct.vector.vector2D
        :returns: boolean -- result of the check
        """
        if self.solveT(point) is not None:
            return True
        return False
            
    def compareEnd(self, other):
        """
        compare the endpoints of two curves

        :param other: the other curve
        :type other: ~sortal.ind.bezierCurve.bezierCurve2D
        :returns: int -- -1 if this curve's end point is smaller, 0 if equals and 1 if bigger
        """
        return self.end.compareValue(other.end) 
        
    def length(self):
        """
        # NOT IMPLEMENTED
        """
        return 0
        
    def touches(self,other):
        """
        check if two curves touches one another

        :param other: the other curve
        :type other: ~sortal.ind.bezierCurve.bezierCurve2D
        :returns: boolean -- result of the check
        """
        phi = (self.A * other.D)/(other.A * self.D)
        v = (self.B * other.D - other.B * self.D)/(other.A * self.D * 2)
        return self.compareValue(other) == 0 and ((v == 1) or (phi + v == 0))

    def contains(self,other):
        """
        check if this curve contains another

        :param other: the other curve
        :type other: ~sortal.ind.bezierCurve.bezierCurve2D
        :returns: boolean -- result of the check
        """
        phi = (self.A * other.D)/(other.A * self.D)
        v = (self.B * other.D - other.B * self.D)/(other.A * self.D * 2)
        return self.compareValue(other) == 0 and (v <= 0 and (phi + v >= 1))

    def disjoint(self,other):
        """
        check if the two curves are disjoint (but are on the same curve)

        :param other: the other curve
        :type other: ~sortal.ind.bezierCurve.bezierCurve2D
        :returns: boolean -- result of the check
        """
        phi = (self.A * other.D)/(other.A * self.D)
        v = (self.B * other.D - other.B * self.D)/(other.A * self.D * 2)
        return self.compareValue(other) == 0 and (v > 1 or (phi + v < 0))

    def align(self, other):
        """
        check if the two curves align with one another

        :param other: the other curve
        :type other: ~sortal.ind.bezierCurve.bezierCurve2D
        :returns: boolean -- result of the check
        """
        phi = (self.A * other.D)/(other.A * self.D)
        v = (self.B * other.D - other.B * self.D)/(other.A * self.D * 2)
        return self.compareValue(other) == 0 and (v == 0 or (phi + v == 1))

    def reconstruct(self, phi, v):
        """
        constructing a new curve segment (on the same curve but different start, end and control point) with phi and v
        For finding sum/difference/product of two curves

        :param phi: the difference between the t value of the new end point and t value of the new start point
        :type phi: ~sortal.struct.coord.coordinate_float.coordinate
        :param v: t value of the new start point
        :type v: ~sortal.struct.coord.coordinate_float.coordinate
        :returns: :class:`~sortal.ind.bezierCurve.bezierCurve2D` -- new bezier curve
        """
        phi = coordinate(phi)
        v = coordinate(v)
        newA = self.a.scale(phi **2)
        newB = (self.a.scale(2*v).add(self.b)).scale(phi)
        newC = self.a.scale(v*v).add(self.b.scale(v)).add(self.c)
        newBezierCurve = bezierCurve2D(self.sort, newA, newB, newC, True)
        return newBezierCurve
        
    def createCurve(self, startPoint, endPoint):
        """
        constructing a new curve segment (on the same curve but different start, end and control point) with new start and end point
        For finding sum/difference/product of two curves

        :param startPoint: new start point
        :type startPoint: ~sortal.struct.vector.vector2D
        :param endPoint: new end point
        :type endPoint: ~sortal.struct.vector.vector2D
        :returns: :class:`~sortal.ind.bezierCurve.bezierCurve2D` -- new bezier curve
        """
        if self.start == startPoint:
            tS = coordinate(0)
        else:
            tS = self.solveT(startPoint)
        if self.end == endPoint:
            tE = coordinate(1)
        else:
            tE = self.solveT(endPoint)
        phi = tE - tS
        v = tS
        return self.reconstruct(phi, v)

    def common(self,other):
        """
        Find the common curve segment of two curves

        :param other: the other curve
        :type other: ~sortal.ind.bezierCurve.bezierCurve2D
        :returns: :class:`~sortal.ind.bezierCurve.bezierCurve2D` -- common bezier curve
        """
        if self.disjoint(other) or self.touches(other):
            return False
        phi = (self.A * other.D)/(other.A * self.D)
        v = (self.B * other.D - other.B * self.D)/(other.A * self.D * 2)

        nStart = self.start
        nEnd = self.end
        if phi+v > 1:
            nEnd = other.end
        if v < 0:
            nStart = other.start
        return self.createCurve(nStart, nEnd)

    def combine(self,other):
        """
        Find the combination of two curves
        
        :param other: the other curve
        :type other: ~sortal.ind.bezierCurve.bezierCurve2D
        :returns: :class:`~sortal.ind.bezierCurve.bezierCurve2D` -- common bezier curve
        """
        if self.disjoint(other):
            return self
            
        phi = (self.A * other.D)/(other.A * self.D)
        v = (self.B * other.D - other.B * self.D)/(other.A * self.D * 2)

        nStart = self.start
        nEnd = self.end
        if phi+v < 1:
            nEnd = other.end
        if v > 0:
            nStart = other.start
        return self.createCurve(nStart, nEnd)
        
    def complement(self,other):
        """
        Find the complement of two curves
        
        :param other: the other curve
        :type other: ~sortal.ind.bezierCurve.bezierCurve2D
        :returns: list of :class:`~sortal.ind.bezierCurve.bezierCurve2D` -- list of complement bezier curve
        """
        result = []
            
        phi = (self.A * other.D)/(other.A * self.D)
        v = (self.B * other.D - other.B * self.D)/(other.A * self.D * 2)
        if v < 0 < phi+v:
            nStart = self.start
            nEnd = other.start
            new = self.createCurve(nStart, nEnd)
            if self.hasAttribute():
                new.addAttribute(self.attribute)
            new.key = self.key
            result.append(new)
        if v < 1 < phi+v:
            nStart = other.end
            nEnd = self.end
            new = self.createCurve(nStart, nEnd)
            if self.hasAttribute():
                new.addAttribute(self.attribute)
            new.key = self.key
            result.append(new)
        return result
        
    def intersectSegment(self, other):
        """
        # NOT IMPLEMENTED
        """
        pass      

    def replicateInd(self, sort = 0):
        """
        Create a duplicate of the individual without attributes
        
        :returns: the duplicate object
        """
        if sort == 0:
            sort = self.sort
        return self.__class__(sort, self.start, self.controlPoint, self.end)
    
    def transformInd(self, transformSet):
        """
        Transforming the individual with a transformation set (editor)

        :param transformSet: the transformation set to be applied
        :type transformSet: ~sortal.transform.transformSet.transformSet
        :returns: the transformed object
        """
        affine = transformSet.retrieveTransform(affineTransform2D)
        if affine is None:
            p0 = self.start
            p1 = self.controlPoint    
            p2 = self.end    
            newCurve = self.duplicateInd()
        else:
            p0 = self.start.transform(affine)
            p1 = self.controlPoint.transform(affine)
            p2 = self.end.transform(affine)
            newCurve = self.__class__(self.sort, p0, p1, p2)
        
        transf = transformSet.retrieveTransform(graphTransform)
        if transf is not None:
            p0 = transf.apply(p0)
            p1 = transf.apply(p1)
            p2 = transf.apply(p2)
            newCurve = self.__class__(self.sort, p0, p1, p2)
        return newCurve
    
    def containsInd(self, ind):
        """
        Check if this individual contains another individual
        """
        if not self.contains(ind):
            return False
        if self.sort.type == 'sort-attribute':
            # if not self.hasAttribute():
            #     if ind.hasAttribute():
            #         return False
            #     else:
            #         return True
            if ind.hasAttribute():
                if ind.attribute.checkPartOf(self.attribute):
                    return True
                else:
                    return False
            else:
                return True
        return True
        
    @staticmethod
    def parse(reader, sort):
        """
        Parse the 2D bezier curve from a sortal parseReader (from string)

        :param reader: the reader of the string to be parsed
        :type reader: ~sortal.io.parseReader.parseReader
        :param sort: sort of the new object
        :type sort: ~sortal.sort.sort.sort
        :returns: :class:`~sortal.ind.lineSegment.lineSegment2D` -- the individual parsed
        """
        if reader.token() != '<':
            raise ParsingError(reader,'expected a "<" at the beginning of creating a line-segment')
        if reader.newToken() != '(': #bracket
            raise ParsingError(reader,'expected a "(" ')
        x,y = vector2D.readVector(reader)
        p0 = vector2D(x,y)
            
        reader.newToken() # bracket (current token would be ',')
        
        x,y = vector2D.readVector(reader)
        p1 = vector2D(x,y)
        
        reader.newToken() # bracket (current token would be ',')
        
        x,y = vector2D.readVector(reader)
        p2 = vector2D(x,y)
            
        ind = bezierCurve2D(sort, p0, p1, p2)

        if reader.token() != '>':
            raise ParsingError(reader, '">" expected to close the line-segment')
            
        reader.newToken() #next token
        
        return ind
        
        

class bezierCurve3D(plane3D):
    
    """
    Quadratic Bezier curve in 3D. The bezier curve is defined by the quadratic equation:
    
    | B(t) = a * t^2 + b * t + c
    | * a, b, c: constant 2D vectors defining the equation
    | * t: the numeric variable of the function; B(0) would be the start point of the curve while B(1) would be the end point.

    The equation is in 2D based on the plane the curve belongs to,

    **Object Parameters:** (not including inherited attributes)

    * **a** (:class:`~sortal.struct.vector.vector2D`) -- a value in the bezier function.
    * **b** (:class:`~sortal.struct.vector.vector2D`) -- b value in the bezier function.
    * **c** (:class:`~sortal.struct.vector.vector2D`) -- c value in the bezier function.
    * **start** (:class:`~sortal.struct.vector.vector3D`) -- start point of the bezier curve (t = 0).
    * **controlPoint** (:class:`~sortal.struct.vector.vector3D`) -- control point of the bezier curve.
    * **end** (:class:`~sortal.struct.vector.vector3D`) -- end point of the bezier curve (t = 1).
    * **localStart** (:class:`~sortal.struct.vector.vector2D`) -- localized start point of the bezier curve.
    * **localControlPoint** (:class:`~sortal.struct.vector.vector2D`) -- control point of the bezier curve.
    * **localEnd** (:class:`~sortal.struct.vector.vector2D`) -- localized end point of the bezier curve.
    * **A** (:class:`~sortal.struct.coord.coordinate_float.coordinate`) -- 4(a.x^2 + a.y^2)
    * **B** (:class:`~sortal.struct.coord.coordinate_float.coordinate`) -- 4(a.x * b.x + a.y * b.y)
    * **C** (:class:`~sortal.struct.coord.coordinate_float.coordinate`) -- b.x^2 + b.y^2
    * **D** (:class:`~sortal.struct.coord.coordinate_float.coordinate`) -- 2(a.y * b.x - a.x * b.y)

    **Class Properties:**

    * **Behavior**: :class:`~sortal.form.curve.curve`
    * **Sort Argument**: :class:`~sortal.sort.parameter.argument.argument.argumentType`.NONE
    * **Mapping**: :class:`~sortal.map.similarityMap3D.similarityMap3D`

    """

    type = 'ind-bezierCurve3D'

    def __init__(self, sort, p1o, p2o, p3o, *args, **kwargs):
        """
        Initializing a 3D bezier curve.

        :param sort: sort of the bezier curve
        :type sort: ~sortal.sort.sort.sort
        :param p1o: start point of the curve (if isFunction is false)
        :type p1o: ~sortal.struct.vector.vector3D
        :param p2o: control point of the curve (if isFunction is false)
        :type p2o: ~sortal.struct.vector.vector3D
        :param p3o: end point of the curve (if isFunction is false)
        :type p3o: ~sortal.struct.vector.vector3D
        """
        super(bezierCurve3D, self).__init__(sort, [p1o,p2o,p3o])
        p1 = self.localize(p1o)
        p2 = self.localize(p2o)
        p3 = self.localize(p3o)
        
        self.a = p3.add(p1).subtract(p2.scale(2))
        self.b = (p2.subtract(p1)).scale(2)
        self.c = p1
        
        self.start = p1o
        self.controlPoint = p2o
        self.end = p3o
        self.localStart = p1
        self.localControlPoint = p2
        self.localEnd = p3
        
        self.A = 4 * (self.a.x ** 2 + self.a.y ** 2)
        self.B = 4 * (self.a.x * self.b.x + self.a.y * self.b.y)
        self.C = self.b.x ** 2 + self.b.y ** 2
        self.D = 2 * (self.a.y * self.b.x - self.a.x * self.b.y)
        self.evaluateCurve()
        

    def vertex(self):
        """
        vertex of the curve (positon on the curve at t = -B/2A)
        
        :returns: :class:`~sortal.struct.vector.vector3D` -- vertex vector
        """
        t = -self.B/(2*self.A)
        return self.pos(t)
        

    def evaluateCurve(self):
        """
        Re-evaluate the orientation of the curve
        """
        t = -self.B/(2*self.A)
        vg = self.vectGradient(t)
        if vg < v0 or (vg == v0 and self.vectGradient(t-1) < v0):
            self.localStart, self.localEnd = self.localEnd, self.localStart
            self.b = self.a.scale(-2).subtract(self.b)
            self.c = self.localStart
            self.B = 4 * (self.a.x * self.b.x + self.a.y * self.b.y)
            self.C = self.b.x ** 2 + self.b.y ** 2
            self.D = 2 * (self.a.y * self.b.x - self.a.x * self.b.y)
        

    def geometricFrame(self):
        """
        the geometric frame of the curve
        """
        return [[min(self.start.x, self.end.x, self.controlPoint.x), 
                 min(self.start.y, self.end.y, self.controlPoint.y), 
                 min(self.start.z, self.end.z, self.controlPoint.z)],
                [max(self.start.x, self.end.x, self.controlPoint.x), 
                 max(self.start.y, self.end.y, self.controlPoint.y), 
                 max(self.start.z, self.end.z, self.controlPoint.z)]]


    def toString(self, printKey):
        """
        converting the curve to string format
        """
        s = ''
        if printKey and self.key is not None:
            s += str(self.key)+' '
        return s + '<'+str(self.start)+','+str(self.controlPoint)+','+str(self.end)+'>'
        
    def getValue(self):
        return str(self)

    def printVal(self):
        """
        print out a string representing the value of this individual, used in form.printE()
        """
        return '<'+str(self.start)+','+str(self.controlPoint)+','+str(self.end)+'>'

    def printEquation(self):
        return str(self.a)+'*t^2 + '+str(self.b)+'*t + '+str(self.c)

    def printRhino(self, dX = 0, dY = 0):
        """
        print the individual in a format that can be pasted into Rhinoceros for visualization
        """
        s = ''
        p1 = self.start
        p2 = self.controlPoint
        p3 = self.end
        s += '_Curve '+str(p1.x + dX)+','+str(p1.y + dY)+','+str(p1.z)+'  '+str(p2.x + dX)+','+str(p2.y + dY)+','+str(p2.z)+'  '+str(p3.x + dX)+','+str(p3.y + dY)+','+str(p3.z)+' EnterEnd'
        return s
        

    def equals(self, other):
        """
        check equality between two bezier curves

        :param other: <individual> - other object to be compared with
        :returns: boolean -- whether the curve and the other object are the same geometrically
        """
        if self.normal != other.normal or self.scalar != other.scalar:
            return False
        return self.a == other.a and self.b == other.b and self.c == other.c
        

    def isNil(self):
        """
        check if the curve is Nil (start == controlPoint == end)
        """
        return (self.localStart.x == self.localControlPoint.x == self.localEnd.x 
            and self.localStart.y == self.localControlPoint.y == self.localEnd.y)
        

    def compareValue(self, other, params = None):
        """
        Comparing the geometric value of this curve with another
        (consider the two curves as equal if they are on the exact same curve)
        
        :param other: other individual to be compared
        :type other: ~sortal.ind.bezierCurve.bezierCurve3D
        :returns: int -- -1 if this curve is smaller, 1 if bigger, 0 if equal
        """
        c = self.normal.compareValue(other.normal)
        if c != 0:
            return c
            
        c = self.scalar.compareValue(other.scalar)
        if c != 0:
            return c
            
        if params is None:
            l = 1/(((other.A**(3/2)) * (self.D**2))/((self.A**(3/2)) * (other.D**2)))
            phi = (self.A * other.D)/(other.A * self.D)
            v = (self.B * other.D - other.B * self.D)/(other.A * self.D * 2)
        else:
            l, phi, v = params
        c = self.pos(0).compareValue(other.pos(v).scale(l))
        if c != 0:
            return c
        return self.pos(1).compareValue(other.pos(phi+v).scale(l))
        
        
    def compareInterval(self, other):
        """
        Comparing the geometric value of this curve with another
        (compare the position of the start and end points of the two curves if they are on the exact same curve)

        :param other: other individual to be compared
        :type other: ~sortal.ind.bezierCurve.bezierCurve3D
        :returns: int -- -1 if this curve is smaller, 1 if bigger, 0 if equal
        """
        c = self.normal.compareValue(other.normal)
        if c != 0:
            return c
        c = self.scalar.compareValue(other.scalar)
        if c != 0:
            return c
            
        l = 1/(((other.A**(3/2)) * (self.D**2))/((self.A**(3/2)) * (other.D**2)))
        phi = (self.A * other.D)/(other.A * self.D)
        v = (self.B * other.D - other.B * self.D)/(other.A * self.D * 2)
        c = self.compareValue(other, params = (l,phi,v))
        if c != 0:
            return c
        c = v.compareValue(c0)
        if c != 0:
            return c
        c = (phi+v).compareValue(c1)
        if c != 0:
            return c
        return 0
        
    def pos(self, t):
        """
        find the position of a point on the curve given the t value
        
        :param t: other individual to be compared
        :type t: ~sortal.struct.coord.coordinate_float.coordinate
        :returns: :class:`~sortal.struct.vector.vector3D` -- position of the point on the curve with t value
        """
        tc = coordinate(t)
        return self.a.scale(tc**2).add(self.b.scale(tc)).add(self.c)
        
    def curvature(self, t):
        """
        find the curvature at a point on the curve given the t value

        :param t: t value
        :type t: ~sortal.struct.coord.coordinate_float.coordinate
        :returns: :class:`~sortal.struct.coord.coordinate_float.coordinate` -- curvature value
        """
        tc = coordinate(t)
        return self.D / ((self.A * (tc ** 2)+ self.B * tc + self.C)**(3/2))
                
    def gradient(self, t):
        """
        find the gradient at a point on the curve given the t value
        
        :param t: t value
        :type t: ~sortal.struct.coord.coordinate_float.coordinate
        :returns: :class:`~sortal.struct.coord.coordinate_float.coordinate` -- gradient value
        """
        dx = 2 * self.a.x * t + self.b.x
        dy = 2 * self.a.y * t + self.b.y
        if dx == 0:
            return None
        return dy/dx
        
    def vectGradient(self, t):
        """
        find the gradient vector at a point on the curve given the t value (local gradient)

        :param t: t value
        :type t: ~sortal.struct.coord.coordinate_float.coordinate
        :returns: :class:`~sortal.struct.vector.vector2D` -- gradient vector
        """
        dx = 2 * self.a.x * t + self.b.x
        dy = 2 * self.a.y * t + self.b.y
        return vector2D(dx, dy)
                
    def codescriptor(self):
        """
        codescriptor of the curve: the curve's vertex

        :returns: sortal.point3D - codescriptor
        """
        point = point3D(self.sort, self.position(self.vertex()))
        point.assoc = self
        return point
        
    def solveT(self, point):
        """
        solve for t value given its position vector (localized position)

        :param point: localized coordinate vector of a point on the curve
        :type point: ~sortal.struct.vector.vector2D
        :returns: :class:`~sortal.struct.coord.coordinate_float.coordinate` -- if point is not on the curve, returns None, else corresponding t value
        """
        a = self.a.x
        b = self.b.x
        c = self.c.x - point.x
        
        if a == 0:
            x1 = -c/b
            x2 = None
        else:
            d = b*b-4*a*c # discriminant
            if d < 0:
                return None
            elif d == 0:
                x1 = (-b)/(2*a)
                x2 = None
            else:
                x1 = (-b+d.sqrt())/(2*a)
                x2 = (-b-d.sqrt())/(2*a)
            
        a = self.a.y
        b = self.b.y
        c = self.c.y - point.y
        if a == 0:
            y1 = -c/b
            y2 = None
        else:
            d = b**2-4*a*c
            
            if d < 0:
                return None
            elif d == 0:
                y1 = (-b)/(2*a)
                y2 = None
            else:
                y1 = (-b+d.sqrt())/(2*a)
                y2 = (-b-d.sqrt())/(2*a)
        x1r = round(x1.value,3)
        y1r = round(y1.value,3)
        if x1r == y1r:
            return x1
        if y2 is not None and x1r == round(y2.value,3):
            return x1
        if x2 is not None:
            x2r = round(x2.value,3)
            if x2r == y1r:
                return x2
            if y2 is not None and x2r == round(y2.value,3):
                return x2
        return None
        
    def containsPoint(self, point):
        """
        check if the curve contains a point

        :param point: the vector coordinates of the point to be checked
        :type point: ~sortal.struct.vector.vector3D
        :returns: boolean -- result of the check
        """
        if not plane3D.containsPoint(self, point):
            return False
        if self.solveT(self.localize(point)) is not None:
            return True
        return False
    

    def compareEnd(self, other):
        """
        compare the endpoints of two curves

        :param other: the other curve
        :type other: ~sortal.ind.bezierCurve.bezierCurve3D
        :returns: int -- -1 if this curve's end point is smaller, 0 if equals and 1 if bigger
        """
        return self.localEnd.compareValue(other.localEnd) 
        
    def length(self):
        """
        # NOT IMPLEMENTED
        """
        return 0
        
    def touches(self,other):
        """
        check if two curves touch one another

        :param other: the other curve
        :type other: ~sortal.ind.bezierCurve.bezierCurve3D
        :returns: boolean -- result of the check
        """
        if not super(bezierCurve3D, self).equals(other):
            return False
        phi = (self.A * other.D)/(other.A * self.D)
        v = (self.B * other.D - other.B * self.D)/(other.A * self.D * 2)
        return self.compareValue(other) == 0 and ((v == 1) or (phi + v == 0))

    def contains(self,other):
        """
        check if this curve contains another

        :param other: the other curve
        :type other: ~sortal.ind.bezierCurve.bezierCurve3D
        :returns: boolean -- result of the check
        """
        if not super(bezierCurve3D, self).equals(other):
            return False
        phi = (self.A * other.D)/(other.A * self.D)
        v = (self.B * other.D - other.B * self.D)/(other.A * self.D * 2)
        return self.compareValue(other) == 0 and (v <= 0 and (phi + v >= 1))

    def disjoint(self,other):
        """
        check if the two curves are disjoint (but are on the same curve)

        :param other: the other curve
        :type other: ~sortal.ind.bezierCurve.bezierCurve3D
        :returns: boolean -- result of the check
        """
        if not super(bezierCurve3D, self).equals(other):
            return False
        phi = (self.A * other.D)/(other.A * self.D)
        v = (self.B * other.D - other.B * self.D)/(other.A * self.D * 2)
        return self.compareValue(other) == 0 and (v > 1 or (phi + v < 0))

    def align(self, other):
        """
        check if the two curves align with one another

        :param other: the other curve
        :type other: ~sortal.ind.bezierCurve.bezierCurve3D
        :returns: boolean -- result of the check
        """
        if not super(bezierCurve3D, self).equals(other):
            return False
        phi = (self.A * other.D)/(other.A * self.D)
        v = (self.B * other.D - other.B * self.D)/(other.A * self.D * 2)
        return self.compareValue(other) == 0 and (v == 0 or (phi + v == 1))

    def reconstruct(self, phi, v):
        """
        constructing a new curve segment (on the same curve but different start, end and control point) with phi and v
        For finding sum/difference/product of two curves

        :param phi: the difference between the t value of the new end point and t value of the new start point
        :type phi: ~sortal.struct.coord.coordinate_float.coordinate
        :param v: t value of the new start point
        :type v: ~sortal.struct.coord.coordinate_float.coordinate
        :returns: :class:`~sortal.ind.bezierCurve.bezierCurve3D` -- new bezier curve
        """
        phi = coordinate(phi)
        v = coordinate(v)
        newA = self.a.scale(phi **2)
        newB = (self.a.scale(2*v).add(self.b)).scale(phi)
        newC = self.a.scale(v*v).add(self.b.scale(v)).add(self.c)
        nCtrl = newB.scale(c0p5).add(newC)
        nEnd = newA.subtract(newC).add(nCtrl).add(nCtrl)
        newBezierCurve = bezierCurve3D(self.sort, self.position(newC), self.position(nCtrl), self.position(nEnd), True)
        return newBezierCurve
        
    def createCurve(self, startPoint, endPoint):
        """
        constructing a new curve segment (on the same curve but different start, end and control point) 
        with new localized start and localized end point For finding sum/difference/product of two curves

        :param startPoint: new localized start point
        :type startPoint: ~sortal.struct.vector.vector2D
        :param endPoint: new localized end point
        :type endPoint: ~sortal.struct.vector.vector2D
        :returns: :class:`~sortal.ind.bezierCurve.bezierCurve3D` -- new bezier curve
        """
        if self.localStart == startPoint:
            tS = coordinate(0)
        else:
            tS = self.solveT(startPoint)
        if self.localEnd == endPoint:
            tE = coordinate(1)
        else:
            tE = self.solveT(endPoint)
        phi = tE - tS
        v = tS
        return self.reconstruct(phi, v)

    def common(self,other):
        """
        Find the common curve segment of two curves

        :param other: the other curve
        :type other: ~sortal.ind.bezierCurve.bezierCurve3D
        :returns: :class:`~sortal.ind.bezierCurve.bezierCurve3D` -- common bezier curve
        """
        if self.disjoint(other) or self.touches(other):
            return False
        phi = (self.A * other.D)/(other.A * self.D)
        v = (self.B * other.D - other.B * self.D)/(other.A * self.D * 2)

        nStart = self.localStart
        nEnd = self.localEnd
        if phi+v > 1:
            nEnd = other.localEnd
        if v < 0:
            nStart = other.localStart
        return self.createCurve(nStart, nEnd)

    def combine(self,other):
        """
        Find the combination of two curves
        
        :param other: the other curve
        :type other: ~sortal.ind.bezierCurve.bezierCurve3D
        :returns: :class:`~sortal.ind.bezierCurve.bezierCurve3D` -- common bezier curve
        """
        if self.disjoint(other):
            return self
            
        phi = (self.A * other.D)/(other.A * self.D)
        v = (self.B * other.D - other.B * self.D)/(other.A * self.D * 2)

        nStart = self.localStart
        nEnd = self.localEnd
        if phi+v < 1:
            nEnd = other.localEnd
        if v > 0:
            nStart = other.localStart
        return self.createCurve(nStart, nEnd)
        
    def complement(self,other):
        """
        Find the complement of two curves
        
        :param other: the other curve
        :type other: ~sortal.ind.bezierCurve.bezierCurve3D
        :returns: list of :class:`~sortal.ind.bezierCurve.bezierCurve3D` -- list of complement bezier curve
        """
        result = []
        phi = (self.A * other.D)/(other.A * self.D)
        v = (self.B * other.D - other.B * self.D)/(other.A * self.D * 2)
        if v < 0 < phi+v:
            nStart = self.localStart
            nEnd = other.localStart
            new = self.createCurve(nStart, nEnd)
            if self.hasAttribute():
                new.addAttribute(self.attribute)
            new.key = self.key
            result.append(new)
        if v < 1 < phi+v:
            nStart = other.localEnd
            nEnd = self.localEnd
            new = self.createCurve(nStart, nEnd)
            if self.hasAttribute():
                new.addAttribute(self.attribute)
            new.key = self.key
            result.append(new)
        return result
        
    def intersectSegment(self, other):
        """
        # NOT IMPLEMENTED
        """
        pass      

    def replicateInd(self, sort = 0):
        """
        Create a duplicate of the individual without attributes
        
        :returns: the duplicate object
        """
        if sort == 0:
            sort = self.sort
        return self.__class__(sort, self.start, self.controlPoint, self.end)
        
    def transformInd(self, transformSet):
        """
        Transforming the individual with a transformation set (editor)

        :param transformSet: the transformation set to be applied
        :type transformSet: ~sortal.transform.transformSet.transformSet
        :returns: the transformed object
        """
        affine = transformSet.retrieveTransform(affineTransform3D)
        if affine is None:
            p0 = self.start
            p1 = self.controlPoint  
            p2 = self.end
            newCurve = self.duplicateInd()
        else:
            p0 = self.start.transform(affine)
            p1 = self.controlPoint.transform(affine)
            p2 = self.end.transform(affine)
            newCurve = self.__class__(self.sort, p0, p1, p2)
        
        transf = transformSet.retrieveTransform(graphTransform)
        if transf is not None:
            p0 = transf.apply(p0)
            p1 = transf.apply(p1)
            p2 = transf.apply(p2)
            newCurve = self.__class__(self.sort, p0, p1, p2)
        return newCurve
                
    def containsInd(self, ind):
        """
        Check if this individual contains another individual
        """
        if not self.contains(ind):
            return False
        if self.sort.type == 'sort-attribute':
            # if not self.hasAttribute():
            #     if ind.hasAttribute():
            #         return False
            #     else:
            #         return True
            if ind.hasAttribute():
                if ind.attribute.checkPartOf(self.attribute):
                    return True
                else:
                    return False
            else:
                return True
        return True
        
    @staticmethod
    def parse(reader, sort):
        """
        Parse the 3D bezier curve from a sortal parseReader (from string)

        :param reader: the reader of the string to be parsed
        :type reader: ~sortal.io.parseReader.parseReader
        :param sort: sort of the new object
        :type sort: ~sortal.sort.sort.sort
        :returns: :class:`~sortal.ind.lineSegment.lineSegment3D` -- the individual parsed
        """
        if reader.token() != '<':
            raise ParsingError(reader,'expected a "<" at the beginning of creating a line-segment')
        if reader.newToken() != '(': #bracket
            raise ParsingError(reader,'expected a "(" ')
        x,y,z = vector3D.readVector(reader)
        p0 = vector3D(x,y,z)
            
        reader.newToken() # bracket (current token would be ',')
        
        x,y,z = vector3D.readVector(reader)
        p1 = vector3D(x,y,z)
        
        reader.newToken() # bracket (current token would be ',')
        
        x,y,z = vector3D.readVector(reader)
        p2 = vector3D(x,y,z)
            
        ind = bezierCurve3D(sort, p0, p1, p2)

        if reader.token() != '>':
            raise ParsingError(reader, '">" expected to close the line-segment')
            
        reader.newToken() #next token
        
        return ind
        
        