# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import range
from builtins import super
from builtins import str

from .bezierCurve import bezierCurve2D, bezierCurve3D
from ..struct.vector import vector2D, vector3D
from ..transform.affineTransform import affineTransform2D, affineTransform3D
from ..transform.graphTransform import graphTransform, graphTransform3D
from ..io.parseReader import ParsingError

class bezierCurveP2D(bezierCurve2D):
    """
    2D bezier curve for parametric matching. The object is initialized similarly to
    normal bezierCurve2D and inherit all its parameters.

    **Class Properties:**

    * **Behavior**: :class:`~sortal.form.curve.curve`
    * **Sort Argument**: :class:`~sortal.sort.parameter.argument.argument.argumentType`.NONE
    * **Mapping**: :class:`~sortal.map.graphMap.graphMap`
    """

    type = 'ind-bezierCurveP2D'
    
    def __init__(self, sort, p1, p2, p3, isFunction = False, *args, **kwargs):
        """
        Initializing a 2D parametric bezier curve.

        :param sort: sort of the bezier curve
        :type sort: ~sortal.sort.sort.sort
        :param p1: start point of the curve (if isFunction is false) | value of a in the function (if isFunction is true)
        :type p1: ~sortal.struct.vector.vector2D
        :param p2: control point of the curve (if isFunction is false) | value of b in the function (if isFunction is true)
        :type p2: ~sortal.struct.vector.vector2D
        :param p3: end point of the curve (if isFunction is false) | value of c in the function (if isFunction is true)
        :type p3: ~sortal.struct.vector.vector2D
        :param isFunction: if False, the curve initialization takes 3 points on the curve (start, end and control point) if True, the curve initialization takes the value of a, b and c of the function
        :type isFunction: boolean
        """
        super(bezierCurveP2D, self).__init__(sort, p1, p2, p3, isFunction)
    
    def codescriptor(self):
        """
        codescriptor of the curve being itself
        """
        return self
        
    def parametricNode(self):
        """
        parametric node for parametric matching: a copy of the curve with no attribute
        """
        curve = bezierCurveP2D(self.sort, self.a, self.b, self.c, True)
        curve.assoc = self
        return curve
        

    def transformInd(self, transformSet):
        """
        Transforming the individual with a transformation set (editor)

        :param transformSet: the transformation set to be applied
        :type transformSet: ~sortal.transform.transformSet.transformSet
        :returns: the transformed object
        """
        transf = transformSet.retrieveTransform(graphTransform)
        if transf is None:
            return self.duplicateInd()
        p1 = transf.apply(self.start)
        p2 = transf.apply(self.controlPoint)
        p3 = transf.apply(self.end)
        return bezierCurveP2D(self.sort, p1, p2, p3)
        
    def checkTransform(self, transform, compareForm):
        """
        check if the transformed curve using a transformation set is contained within the shape's form

        :param transform: transformation set
        :type transform: ~sortal.transform.transformSet.transformSet
        :param compareForm: the shape's form to check whether the transformed curve is contained in
        :type compareForm: ~sortal.form.form.form
        :returns: boolean -- result of the check
        """
        try:
            comp = self.edit(transform)
        except ValueError:
            return False
        if compareForm.containsInd(comp):
            return True
        return False


    @staticmethod
    def parse(reader, sort):
        """
        Parse the 2D parametric bezier curve from a sortal parseReader (from string)

        :param reader: the reader of the string to be parsed
        :type reader: ~sortal.io.parseReader.parseReader
        :param sort: sort of the new object
        :type sort: ~sortal.sort.sort.sort
        :returns: :class:`~sortal.ind.lineSegment.lineSegment2D` -- the individual parsed
        """
        if reader.token() != '<':
            raise ParsingError(reader,'expected a "<" at the beginning of creating a line-segment')
        if reader.newToken() != '(': #bracket
            raise ParsingError(reader,'expected a "(" ')
        x,y = vector2D.readVector(reader)
        p0 = vector2D(x,y)
            
        reader.newToken() # bracket (current token would be ',')
        
        x,y = vector2D.readVector(reader)
        p1 = vector2D(x,y)
        
        reader.newToken() # bracket (current token would be ',')
        
        x,y = vector2D.readVector(reader)
        p2 = vector2D(x,y)
            
        ind = bezierCurveP2D(sort, p0, p1, p2)

        if reader.token() != '>':
            raise ParsingError(reader, '">" expected to close the line-segment')
            
        reader.newToken() #next token
        
        return ind

class bezierCurveP3D(bezierCurve3D):
    """
    3D bezier curve for parametric matching. The object is initialized similarly to
    normal bezierCurve3D and inherit all its parameters.

    **Class Properties:**

    * **Behavior**: :class:`~sortal.form.curve.curve`
    * **Sort Argument**: :class:`~sortal.sort.parameter.argument.argument.argumentType`.NONE
    * **Mapping**: :class:`~sortal.map.graphMap3D.graphMap3D`
    """

    type = 'ind-bezierCurveP3D'
    
    def __init__(self, sort, p1, p2, p3, *args, **kwargs):
        """
        Initializing a 3D bezier curve.

        :param sort: sort of the bezier curve
        :type sort: ~sortal.sort.sort.sort
        :param p1o: start point of the curve (if isFunction is false)
        :type p1o: ~sortal.struct.vector.vector3D
        :param p2o: control point of the curve (if isFunction is false)
        :type p2o: ~sortal.struct.vector.vector3D
        :param p3o: end point of the curve (if isFunction is false)
        :type p3o: ~sortal.struct.vector.vector3D
        """
        super(bezierCurveP3D, self).__init__(sort, p1, p2, p3)
    
    def codescriptor(self):
        """
        codescriptor of the curve being itself
        """
        return self
        
    def parametricNode(self):
        """
        parametric node for parametric matching: a copy of the curve with no attribute
        """
        curve = bezierCurveP3D(self.sort, self.start, self.controlPoint, self.end)
        curve.assoc = self
        return curve
        

    def transformInd(self, transformSet):
        """
        Transforming the individual with a transformation set (editor)

        :param transformSet: the transformation set to be applied
        :type transformSet: ~sortal.transform.transformSet.transformSet
        :returns: the transformed object
        """
        transf = transformSet.retrieveTransform(graphTransform3D)
        if transf is None:
            return self.duplicateInd()
        p1 = transf.apply(self.start)
        p2 = transf.apply(self.controlPoint)
        p3 = transf.apply(self.end)
        return bezierCurveP3D(self.sort, p1, p2, p3)
        
    def checkTransform(self, transform, compareForm):
        """
        check if the transformed curve using a transformation set is contained within the shape's form

        :param transform: transformation set
        :type transform: ~sortal.transform.transformSet.transformSet
        :param compareForm: the shape's form to check whether the transformed curve is contained in
        :type compareForm: ~sortal.form.form.form
        :returns: boolean -- result of the check
        """
        try:
            comp = self.edit(transform)
        except ValueError:
            return False
        if compareForm.containsInd(comp):
            return True
        return False


    @staticmethod
    def parse(reader, sort):
        """
        Parse the 3D parametric bezier curve from a sortal parseReader (from string)

        :param reader: the reader of the string to be parsed
        :type reader: ~sortal.io.parseReader.parseReader
        :param sort: sort of the new object
        :type sort: ~sortal.sort.sort.sort
        :returns: :class:`~sortal.ind.lineSegment.lineSegment3D` -- the individual parsed
        """
        if reader.token() != '<':
            raise ParsingError(reader,'expected a "<" at the beginning of creating a line-segment')
        if reader.newToken() != '(': #bracket
            raise ParsingError(reader,'expected a "(" ')
        x,y,z = vector3D.readVector(reader)
        p0 = vector3D(x,y,z)
            
        reader.newToken() # bracket (current token would be ',')
        
        x,y,z = vector3D.readVector(reader)
        p1 = vector3D(x,y,z)
        
        reader.newToken() # bracket (current token would be ',')
        
        x,y,z = vector3D.readVector(reader)
        p2 = vector3D(x,y,z)
            
        ind = bezierCurveP3D(sort, p0, p1, p2)

        if reader.token() != '>':
            raise ParsingError(reader, '">" expected to close the line-segment')
            
        reader.newToken() #next token
        
        return ind

