# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import super

from .plane import plane2D, plane3D
from ..struct.coordinate import c2, pi
from ..struct.vector import vector2D, vector3D
from ..transform.affineTransform import affineTransform2D, affineTransform3D
from ..transform.graphTransform import graphTransform, graphTransform3D
from ..io.parseReader import ParsingError

class circle2D(plane2D):
    """
    A circle is a closed circular planar curve with a discrete behavior.

    The circle2D class extends on the plane2D class.

    It defines the characteristic individual for circles.

    A circle is represented as a planar geometry with additionally one vector
    specifying the center of the circle and another vector specifying a position
    on the circle.

    **Object Parameters:** (not including inherited attributes)

    * **center** (:class:`~sortal.struct.vector.vector2D`) -- circle center.
    * **squareRadius** (coordinate) -- Circle's squared radius.
    * **radius** (coordinate) -- circle radius.

    **Class Properties:**

    * **Behavior**: :class:`~sortal.form.discrete.discrete`
    * **Sort Argument**: :class:`~sortal.sort.parameter.argument.argument.argumentType`.NONE
    * **Mapping**: :class:`~sortal.map.similarityMap.similarityMap2D`
    """

    type = 'ind-circle2D'
    def __init__(self, sort, center, p1, p2 = None, *args, **kwargs):
        """
        Initializing a 2D circle

        :param sort: sort of the circle
        :type sort: ~sortal.sort.sort.sort
        :param center: circle center
        :type center: ~sortal.struct.vector.vector2D
        :param p1: one point on the circle
        :type p1: ~sortal.struct.vector.vector2D
        :param p2: (optional) another point on the circle, for consistency with circle3D initialization
        :type p2: ~sortal.struct.vector.vector2D
        """
        super(circle2D, self).__init__(sort)     
        self.center = center
        self.squareRadius = p1.subtract(self.center).squareLength()
        self.radius = self.squareRadius.sqrt()
        
    def isNil(self):
        """
        Check if the circle is Nil
        """
        return False
        
    def getPoints(self):
        """
        Get 2 points on the circle that form a right angle with the center
        """
        s = self.center.add(vector2D(self.radius, 0))
        e = self.center.add(vector2D(0, self.radius))
        return  self.center, s, e
        
    def printVal(self):
        """
        print out a string representing the value of this individual, used in form.printE()
        """
        c, s, e = self.getPoints()
        return 'center:'+str(c)+', p1:'+str(s)+', p2:'+str(e)
        
    def printRhino(self, dX = 0, dY = 0):
        """
        print the individual in a format that can be pasted into Rhinoceros for visualization
        """
        return '_Circle '+str(self.center.x + dX)+','+str(self.center.y + dY)+'  '+str(self.radius)
        
    def toString(self, printKey):
        """
        string format of the individual (with no attribute)

        :param printKey: whether the tag of the individual is to be added to the string
        :type printKey: boolean
        :returns: str -- string format of the individual
        """
        st = ''
        if printKey and self.key is not None:
            st += str(self.key)+' '
        c, s, e = self.getPoints()
        return st + '<'+str(c)+','+str(s)+','+str(e)+'>'
        
    def equals(self, other):
        """
        check equality
        """
        if self.center == other.center:
            return self.radius == other.radius
        return False
        
    def containsPoint(self, point):
        """
        Check if the individual contains a specific point

        :param point: point to be checked
        :type point: ~sortal.struct.vector.vector2D
        :returns: boolean -- whether the circle contains the point
        """
        if point.subtract(self.center).squareLength() == self.squareRadius:
            return True
        return False
        
    def compareValue(self, other):
        """
        comparing value of 2 individuals
        """
        c = self.center.compareValue(other.center)
        if c != 0:
            return c
        return self.radius.compareValue(other.radius)
            
    def concentric(self, other):
        """
        check if 2 circles have the same center point
        """
        if other.type != 'ind-circle2D':
            return False
        return self.center == other.center
        
    def diameter(self):
        """
        get the diameter of the circle
        """
        return self.radius * 2
        
    def circumference(self):
        """
        get the circumference of the circle
        """
        return self.radius * 2 * pi
        
    def area(self):
        """
        get the area of the circle
        """
        return self.squareRadius * pi
        
    def transformInd(self, transformSet):        
        """
        Transforming the individual with a transformation set (editor)

        :param transformSet: the transformation set to be applied
        :type transformSet: ~sortal.transform.transformSet.transformSet
        :returns: the transformed object
        """
        affine = transformSet.retrieveTransform(affineTransform2D)
        if affine is None:
            newCircle = self.duplicateInd()
        else:
            c, s, e = self.getPoints()
            center = c.transform(affine)
            p1 = s.transform(affine)
            newCircle = circle2D(self.sort, center, p1)
            
        transf = transformSet.retrieveTransform(graphTransform)
        if transf is not None:
            center = transf.apply(self.center)
            newPts = []
            for i in transf.vects:
                if newCircle.containsPoint(i[0]):
                    newPts.append(i[1])
            newCircle = circle2D(self.sort, center, newPts[0], newPts[-1])
        return newCircle
        
    def replicateInd(self):
        """
        Create a duplicate of the individual without attributes
        
        :returns: the duplicate object
        """
        c, s, e = self.getPoints()
        return circle2D(self.sort, c, s)
        
    @staticmethod
    def parse(reader, sort):
        """
        Parse the 2D circle from a sortal parseReader (from string)

        :param reader: the reader of the string to be parsed
        :type reader: ~sortal.io.parseReader.parseReader
        :param sort: sort of the new object
        :type sort: ~sortal.sort.sort.sort
        :returns: :class:`~sortal.ind.circle.circle2D` -- the individual parsed
        """
        if reader.token() != '<':
            raise ParsingError(reader,'expected a "<" at the beginning of creating a circle')
        if reader.newToken() != '(': #bracket
            raise ParsingError(reader,'expected a "(" ')
        x,y = vector2D.readVector(reader)
        center = vector2D(x,y)
        if reader.token() != ',':
            raise ParsingError(reader,'"," expected')
        reader.newToken() # bracket (current token would be ',')
        x,y = vector2D.readVector(reader)
        p1 = vector2D(x,y)
        # if reader.token() != ',':
        #     raise ParsingError(reader,'"," expected')
        if reader.newToken() == ',': # bracket (current token would be ',')
            x,y = vector2D.readVector(reader)
            p2 = vector2D(x,y)
        else:
            p2 = None
        if reader.token() != '>':
            raise ParsingError(reader,'">" expected')
        reader.newToken() #next token
        return circle2D(sort, center, p1, p2)
        
class circle3D(plane3D):
    """
    A circle is a closed circular planar curve with a discrete behavior.

    The circle3D class extends on the plane3D class.

    It defines the characteristic individual for circles.

    A circle is represented as a planar geometry with additionally one vector
    specifying the center of the circle and another vector specifying a position
    on the circle.

    **Object Parameters:** (not including inherited attributes)

    * **center** (:class:`~sortal.struct.vector.vector3D`) -- circle center.
    * **squareRadius** (coordinate) -- Circle's squared radius.
    * **radius** (coordinate) -- circle radius.
    * **localCenter** (:class:`~sortal.struct.vector.vector2D`) -- localized circle center.

    **Class Properties:**

    * **Behavior**: :class:`~sortal.form.discrete.discrete`
    * **Sort Argument**: :class:`~sortal.sort.parameter.argument.argument.argumentType`.NONE
    * **Mapping**: :class:`~sortal.map.similarityMap3D.similarityMap3D`

    """
    type = 'ind-circle3D'
    
    def __init__(self, sort, center, p1, p2, *args, **kwargs):
        """
        Initializing a 3D circle

        :param sort: sort of the circle
        :type sort: ~sortal.sort.sort.sort
        :param center: circle center
        :type center: ~sortal.struct.vector.vector3D
        :param p1: one point on the circle
        :type p1: ~sortal.struct.vector.vector3D
        :param p2: another point on the circle, must not be colinear with center and p1
        :type p2: ~sortal.struct.vector.vector3D
        """
        super(circle3D, self).__init__(sort, [center, p1, p2])   
        self.squareRadius = p1.subtract(center).squareLength()
        self.radius = self.squareRadius.sqrt()
        self.center = center
        self.localCenter = self.localize(center)
        
    def isNil(self):
        """
        Check if the circle is Nil
        """
        return False
        
    def printVal(self):
        """
        print out a string representing the value of this individual, used in form.printE()
        """
        c, s, e = self.getPoints()
        return 'center:'+str(c)+', p1:'+str(s)+', p2:'+str(e)
    
    def printRhino(self, dX = 0, dY = 0):
        """
        print the individual in a format that can be pasted into Rhinoceros for visualization
        """
        pCenter = self.center
        pNormal = pCenter.add(self.normal)
        return '_Circle '+str(pCenter.x+dX)+','+str(pCenter.y + dY)+','+str(pCenter.z)+' O '+ str(pNormal.x+dX)+','+str(pNormal.y + dY)+','+str(pNormal.z)+' '+str(self.radius)
        
    def toString(self, printKey):
        """
        string format of the individual (with no attribute)

        :param printKey: whether the tag of the individual is to be added to the string
        :type printKey: boolean
        :returns: str -- string format of the individual
        """
        st = ''
        if printKey and self.key is not None:
            st += str(self.key)+' '
        c, s, e = self.getPoints()
        return st + '<'+str(c)+','+str(s)+','+str(e)+'>'
        
    def getPoints(self):
        """
        Get 2 points on the circle that form a right angle with the center
        """
        s = self.localCenter.add(vector2D(self.radius, 0))
        e = self.localCenter.add(vector2D(0, self.radius))
        return  self.center, self.position(s), self.position(e)

    def equals(self, other):
        """
        check equality
        """
        if self.normal == other.normal and self.localCenter == other.localCenter and self.radius == other.radius:
            return True
        return False
        
    def containsPoint(self, point):
        """
        Check if the individual contains a specific point

        :param point: point to be checked
        :type point: ~sortal.struct.vector.vector3D
        :returns: boolean -- whether the circle contains the point
        """
        if point == self.center or not point.subtract(self.center).dotproduct(self.normal) == 0:
            return False
        p = self.localize(point)
        if p.subtract(self.localCenter).squareLength() == self.squareRadius:
            return True
        return False
        
    def compareValue(self, other):
        """
        comparing value of 2 individuals
        """        
        c = super(circle3D, self).compareValue(other)
        if c != 0:
            return c
        c = self.localCenter.compareValue(other.localCenter)
        if c != 0:
            return c
        return self.radius.compareValue(other.radius)
            
    def concentric(self, other):
        """
        check if 2 circles have the same center point
        """
        if other.type != 'ind-circle3D':
            return False
        if not super(circle3D, self).__eq__(other):
            return False
        return self.localCenter == other.localCenter
        
    def diameter(self):
        """
        get the diameter of the circle
        """
        return self.radius * 2
        
    def circumference (self):
        """
        get the circumference of the circle
        """
        return self.radius * 2 * pi
        
    def area(self):
        """
        get the area of the circle
        """
        return self.squareRadius * pi
        
    def transformInd(self, transformSet):        
        """
        Transforming the individual with a transformation set (editor)

        :param transformSet: the transformation set to be applied
        :type transformSet: ~sortal.transform.transformSet.transformSet
        :returns: the transformed object
        """
        affine = transformSet.retrieveTransform(affineTransform3D)
        if affine is None:
            newCircle = self.duplicateInd()
        else:
            c,s,e = self.getPoints()
            center = c.transform(affine)
            p1 = s.transform(affine)
            p2 = e.transform(affine)    
            newCircle = circle3D(self.sort, center, p1, p2)
            
        transf = transformSet.retrieveTransform(graphTransform3D)
        if transf is not None:
            center = transf.apply(self.center)
            newPts = []
            for i in transf.vects:
                if newCircle.containsPoint(i[0]):
                    newPts.append(i[1])
            newCircle = circle3D(self.sort, center, newPts[0], newPts[-1])
        return newCircle
        
    def replicateInd(self):
        """
        Create a duplicate of the individual without attributes
        
        :returns: the duplicate object
        """
        c,s,e = self.getPoints()
        return circle3D(self.sort, c, s, e)
        
        
    @staticmethod
    def parse(reader, sort):
        """
        Parse the 3D circle from a sortal parseReader (from string)

        :param reader: the reader of the string to be parsed
        :type reader: ~sortal.io.parseReader.parseReader
        :param sort: sort of the new object
        :type sort: ~sortal.sort.sort.sort
        :returns: :class:`~sortal.ind.circle.circle3D` -- the individual parsed
        """
        if reader.token() != '<':
            raise ParsingError(reader,'expected a "<" at the beginning of creating a circle')
        if reader.newToken() != '(': #bracket
            raise ParsingError(reader,'expected a "(" ')
            
        x,y,z = vector3D.readVector(reader)
        center = vector3D(x,y,z)
        if reader.token() != ',':
            raise ParsingError(reader,'"," expected')
        reader.newToken() # bracket (current token would be ',')
        
        x,y,z = vector3D.readVector(reader)
        p1 = vector3D(x,y,z)
        if reader.token() != ',':
            raise ParsingError(reader,'"," expected')
        reader.newToken() # bracket (current token would be ',')
        
        x,y,z = vector3D.readVector(reader)
        p2 = vector3D(x,y,z)
        if reader.token() != '>':
            raise ParsingError(reader,'">" expected')
        reader.newToken() #next token
        
        return circle3D(sort, center, p1, p2)