# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import range
from builtins import super
from builtins import str

from .circle import circle2D, circle3D
from ..struct.coordinate import c2
from ..struct.vector import vector2D, vector3D
from ..transform.affineTransform import affineTransform2D, affineTransform3D
from ..transform.graphTransform import graphTransform, graphTransform3D
from ..io.parseReader import ParsingError


class circleP2D(circle2D):
    """
    Parametric Circle in 2D class. Inherit all its attributes from circle2D

    **Class Properties:**

    * **Behavior**: :class:`~sortal.form.discrete.discrete`
    * **Sort Argument**: :class:`~sortal.sort.parameter.argument.argument.argumentType`.NONE
    * **Mapping**: :class:`~sortal.map.graphMap.graphMap`

    """

    type = 'ind-circleP2D'
    
    def __init__(self, sort, center, p1, p2 = None, *args, **kwargs):
        """
        Initializing a parametric 2D circle

        :param sort: sort of the circle
        :type sort: ~sortal.sort.sort.sort
        :param center: circle center
        :type center: ~sortal.struct.vector.vector2D
        :param p1: one point on the circle
        :type p1: ~sortal.struct.vector.vector2D
        :param p2: (optional) another point on the circle, for consistency with circleP3D initialization
        :type p2: ~sortal.struct.vector.vector2D
        """
        super(circleP2D, self).__init__(sort, center, p1)
        
    def codescriptor(self):
        """
        codescriptor of the parametric 2D circle which is itself
        """
        return self
        
    def parametricNode(self):
        """
        node for parametric matching: a duplicate without attribute
        """
        circle = circleP2D(self.sort, self.center, self.center.add(vector2D(self.radius,0)))
        circle.assoc = self
        return circle
        

    def transformInd(self, transformSet):
        """
        Transforming the individual with a transformation set (editor)

        :param transformSet: the transformation set to be applied
        :type transformSet: ~sortal.transform.transformSet.transformSet
        :returns: the transformed object
        """
        transf = transformSet.retrieveTransform(graphTransform)
        if transf is None:
            return self.duplicateInd()
        center = transf.apply(self.center)
        newPts = []
        for i in transf.vects:
            if self.containsPoint(i[0]):
                newPts.append(i[1])
        if len(newPts) == 0:
            return self
        return circleP2D(self.sort, center, newPts[0])
        
    def checkTransform(self, transform, compareForm):
        """
        Check if the individual belongs to the compareForm after transformation with a transform set
        """
        try:
            comp = self.edit(transform)
        except ValueError:
            return False
        if compareForm.containsInd(comp):
            return True
        return False


    @staticmethod
    def parse(reader, sort):
        """
        Parse the parametric 2D circle from a sortal parseReader (from string)

        :param reader: the reader of the string to be parsed
        :type reader: ~sortal.io.parseReader.parseReader
        :param sort: sort of the new object
        :type sort: ~sortal.sort.sort.sort
        :returns: :class:`~sortal.ind.circleP.circleP2D` -- the individual parsed
        """
        if reader.token() != '<':
            raise ParsingError(reader,'expected a "<" at the beginning of creating a circle')
        if reader.newToken() != '(': #bracket
            raise ParsingError(reader,'expected a "(" ')
            
        x,y = vector2D.readVector(reader)
        center = vector2D(x,y)
        if reader.token() != ',':
            raise ParsingError(reader,'"," expected')
        reader.newToken() # bracket (current token would be ',')
        
        x,y = vector2D.readVector(reader)
        p1 = vector2D(x,y)
        if reader.token() != ',':
            raise ParsingError(reader,'"," expected')
        reader.newToken() # bracket (current token would be ',')

        if reader.newToken() == ',': # bracket (current token would be ',')
            x,y = vector2D.readVector(reader)
            p2 = vector2D(x,y)
        else:
            p2 = None

        if reader.token() != '>':
            raise ParsingError(reader,'">" expected')
        reader.newToken() #next token

        return circleP2D(sort, center, p1, p2)



class circleP3D(circle3D):
    """
    Parametric Circle in 3D class. Inherit all its attributes from circle3D

    **Class Properties:**

    * **Behavior**: :class:`~sortal.form.discrete.discrete`
    * **Sort Argument**: :class:`~sortal.sort.parameter.argument.argument.argumentType`.NONE
    * **Mapping**: :class:`~sortal.map.graphMap3D.graphMap3D`

    """

    type = 'ind-circleP3D'
    
    def __init__(self, sort, center, p1, p2, *args, **kwargs):
        """
        Initializing a parametric 3D circle

        :param sort: sort of the circle
        :type sort: ~sortal.sort.sort.sort
        :param center: circle center
        :type center: ~sortal.struct.vector.vector3D
        :param p1: one point on the circle
        :type p1: ~sortal.struct.vector.vector3D
        :param p2: another point on the circle, must not be colinear with center and p1
        :type p2: ~sortal.struct.vector.vector3D
        """
        super(circleP3D, self).__init__(sort, center, p1, p2)
        
    def codescriptor(self):
        """
        codescriptor of the parametric 3D circle which is itself
        """
        return self
        
    def parametricNode(self):
        """
        node for parametric matching: a duplicate without attribute
        """
        circle = circleP3D(self.sort, self.center, self.start, self.end)
        circle.assoc = self
        return circle
        

    def transformInd(self, transformSet):
        """
        Transforming the individual with a transformation set (editor)

        :param transformSet: the transformation set to be applied
        :type transformSet: ~sortal.transform.transformSet.transformSet
        :returns: the transformed object
        """
        transf = transformSet.retrieveTransform(graphTransform3D)
        if transf is None:
            return self.duplicateInd()
        center = transf.apply(self.center)
        newPts = []
        for i in transf.vects:
            if self.containsPoint(i[0]):
                newPts.append(i[1])
        if len(newPts) < 2:
            return self
        return circleP3D(self.sort, center, newPts[0], newPts[1])
        
        
    def checkTransform(self, transform, compareForm):
        """
        Check if the individual belongs to the compareForm after transformation with a transform set
        """
        try:
            comp = self.edit(transform)
        except ValueError:
            return False
        if compareForm.containsInd(comp):
            return True
        return False


    @staticmethod
    def parse(reader, sort):
        """
        Parse the parametric 3D circle from a sortal parseReader (from string)

        :param reader: the reader of the string to be parsed
        :type reader: ~sortal.io.parseReader.parseReader
        :param sort: sort of the new object
        :type sort: ~sortal.sort.sort.sort
        :returns: :class:`~sortal.ind.circleP.circleP3D` -- the individual parsed
        """
        if reader.token() != '<':
            raise ParsingError(reader,'expected a "<" at the beginning of creating a circle')
        if reader.newToken() != '(': #bracket
            raise ParsingError(reader,'expected a "(" ')
            
        x,y,z = vector3D.readVector(reader)
        center = vector3D(x,y,z)
        if reader.token() != ',':
            raise ParsingError(reader,'"," expected')
        reader.newToken() # bracket (current token would be ',')
        
        x,y,z = vector3D.readVector(reader)
        start = vector3D(x,y,z)
        if reader.token() != ',':
            raise ParsingError(reader,'"," expected')
        reader.newToken() # bracket (current token would be ',')
        
        x,y,z = vector3D.readVector(reader)
        end = vector3D(x,y,z)
        if reader.token() != '>':
            raise ParsingError(reader,'">" expected')
        reader.newToken() #next token
        
        return circleP3D(sort, center, start, end)