# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import range
from builtins import super


from copy import copy
from colorsys import rgb_to_hsv, hsv_to_rgb

from .special import special
from .description import description
from ..sort.parameter.colorMode import colorMode
from ..struct.coordinate import coordinate as co, c0
from ..io.parseReader import ParsingError, parseReader

class color(special):
    """
    color class, adhering to its sort's parameter including color mode (RGB or HSV/HSB),
    color behavior (RGB_MAX, RGB_AVG, RGB_SUM or RGB_ALPHA) and maximum color values.
    (see :class:`~sortal.sort.parameter.colorMode.colorMode`)

    **Object Parameters:** (not including inherited attributes)

    * **components** (list(float) or list(coordinate)) -- list of 3 float values representing the color in the object sort's mode
    * **alpha** (float) -- alpha value (if any)

    **Class Properties:**

    * **Behavior**: :class:`~sortal.form.specialForm.specialForm`
    * **Sort Argument**: :class:`~sortal.sort.parameter.argument.argument.argumentType`.COLORMODE -- :class:`~sortal.sort.parameter.colorMode.colorMode`
    * **Mapping**: :class:`~sortal.map.exactMap.exactMap`

    """

    type = 'ind-color'
    DEFAULT_MODE = colorMode('RGB,')
    RGB_NOTATION = "RGB"
    HSV_NOTATION = "HSV"
    
    def __init__(self, sort, c1, c2, c3, alpha = -1, RGB = True, *args, **kwargs):
        """
        Initializing a color object

        :param sort: sort of the individual
        :type sort: ~sortal.sort.sort.sort
        :param c1: first values representing the color (Red for RGB, Hue for HSV)
        :type c1: float or :class:`~sortal.struct.coord.coordinate_float.coordinate`
        :param c2: second values representing the color (Green for RGB, Saturation for HSV)
        :type c2: float or :class:`~sortal.struct.coord.coordinate_float.coordinate`
        :param c3: third values representing the color (Blue for RGB, Value for HSV)
        :type c3: float or :class:`~sortal.struct.coord.coordinate_float.coordinate`
        :param alpha: alpha value of the color
        :type alpha: float or :class:`~sortal.struct.coord.coordinate_float.coordinate`
        :param RGB: check indicating whether the color is under RGB model
        :type RGB: boolean
        """
        super(color, self).__init__(sort)
        mode = sort.baseSort().arguments
        if mode is None:
            mode = self.DEFAULT_MODE
        if mode.behavior == colorMode.Behavior.RGB_ALPHA:
            if alpha == -1:
                self.alpha = mode.max_alpha
            else:
                self.alpha = co(alpha)
        self.components = [0,0,0]
        
        c1 = co(c1)
        c2 = co(c2)
        c3 = co(c3)
        
        if RGB and mode.model == colorMode.Model.HSV:
            self.components = self.RGBtoHSV(c1, c2, c3)
            self.reduceValues(mode)
        elif (not RGB) and mode.model == colorMode.Model.RGB:
            self.components = self.HSVtoRGB(c1, c2, c3)
            self.reduceValues(mode)
        else:
            self.components[0] = c1
            self.components[1] = c2
            self.components[2] = c3
            self.checkValues(mode)
            
    def toString(self, printKey):
        """
        string format of the individual (with no attribute)

        :param printKey: whether the tag of the individual is to be added to the string
        :type printKey: boolean
        :returns: str -- string format of the individual
        """
        s = ''
        if printKey and self.key is not None:
            s += str(self.key)+' '
        if self.isNil():
            return '[NIL]'
        mode = self.sort.baseSort().arguments
        if mode is None:
            mode = self.DEFAULT_MODE
        if self.description is not None:
            s += str(self.description)
            return s
        s += '('+str(self.components[0])+', '+str(self.components[1])+', '+str(self.components[2])
        if mode.behavior == colorMode.Behavior.RGB_ALPHA:
            s += ', '+str(self.alpha)
        return s + ')'
        
    def getValue(self):
        return self.components + [self.alpha]
        
    def printVal(self):
        """
        print out a string representing the value of this individual, used in form.printE()
        """
        return str(self)
            
    def isNil(self):
        """
        Check if the Color is Nil
        """
        return self.components == [-1,-1,-1]
    
    def colorMode(self):
        """
        Get the color mode
        """
        return self.sort.baseSort().arguments
        
    def model(self):
        """
        Get the color model
        """
        if self.colorMode() is not None:
            return self.colorMode().model
        return self.DEFAULT_MODE.model
            
    def behavior(self):
        """
        Get the color behavior
        """
        if self.colorMode() is not None:
            return self.colorMode().behavior
        return self.DEFAULT_MODE.behavior
            
    def max_alpha(self):
        """
        Get the color maximum allowed alpha value
        """

        if self.colorMode() is not None:
            return self.colorMode().max_alpha
        return self.DEFAULT_MODE.max_alpha
            
    def max_cvalues(self):
        """
        Get the color maximum allowed color value(s)
        """
        if self.colorMode() is not None:
            return self.colorMode().max_cvalues
        return self.DEFAULT_MODE.max_cvalues
            
    @staticmethod
    def RGBtoHSV(c1, c2, c3):
        """
        converting RGB value to HSV
        """
        result = rgb_to_hsv(c1.value,c2.value,c3.value)
        comp = []
        for i in result:
            comp.append(co(i))
        return comp
        
    @staticmethod
    def HSVtoRGB(c1, c2, c3):
        """
        converting HSV value to RGB
        """
        result = hsv_to_rgb(c1.value,c2.value,c3.value)
        comp = []
        for i in result:
            comp.append(co(i))
        return comp
        
    def reduceValues(self, mode):
        """
        reduce values
        """
        if mode.behavior == colorMode.Behavior.RGB_ALPHA:
            if self.alpha <= 0 or self.alpha > colorMode.MAX_ALPHA:
                raise ValueError("Alpha value must lie between 0 and " + str(colorMode.MAX_ALPHA))
            if self.alpha == colorMode.MAX_ALPHA:
                self.alpha = self.alpha * colorMode.max_alpha / colorMode.MAX_ALPHA
        notation = self.RGB_NOTATION
        max_cvalues = colorMode.MAX_RGB_VALUES
        if mode.model == colorMode.Model.HSV:
            notation = self.HSV_NOTATION
            max_cvalues = colorMode.MAX_HSV_VALUES
        if len(max_cvalues) > 1:
            for i in range(3):
                if self.components[i] < 0 or self.components[i] > max_cvalues[i]:
                    raise ValueError(str(notation[i])+' value must lie between 0 and ' + str(max_cvalues[i]))
        else:
            for i in range(3):
                if self.components[i] < 0 or self.components[i] > max_cvalues[0]:
                    raise ValueError(str(notation[i])+' value must lie between 0 and ' + str(max_cvalues[0]))

        modeMaxCvalues = mode.max_cvalues
        if modeMaxCvalues != max_cvalues:
            if len(max_cvalues) > 1 and len(modeMaxCvalues)> 1:
                for i in range(3):
                    self.components[i] = round(self.components[i] * modeMaxCvalues[i] / max_cvalues[i])
            elif len(max_cvalues) > 1 and len(modeMaxCvalues) == 1:
                for i in range(3):
                    self.components[i] = round(self.components[i] * modeMaxCvalues[i] / max_cvalues[i])
            elif len(max_cvalues) == 1 and len(modeMaxCvalues) > 1:
                for i in range(3):
                    self.components[i] = round(self.components[i] * max_cvalues[0] / modeMaxCvalues[i])
            else:
                for i in range(3):
                    self.components[i] = round(self.components[i] * max_cvalues[0] / modeMaxCvalues[0])
                

    def checkValues(self, mode):
        """
        Check if the color values are valid, raise ValueError if not
        """
        if mode.behavior == colorMode.Behavior.RGB_ALPHA and (self.alpha <= 0 or self.alpha > mode.max_alpha):
            raise ValueError("Alpha value must lie between 0 and " + str(colorMode.MAX_ALPHA))
        notation = self.RGB_NOTATION
        if mode.model == colorMode.Model.HSV:
            notation = self.HSV_NOTATION
        max_cvalues = mode.max_cvalues
        if len(max_cvalues) > 1:
            for i in range(3):
                if self.components[i] < 0 or self.components[i] > max_cvalues[i]:
                    raise ValueError(str(notation[i])+' value must lie between 0 and ' + str(max_cvalues[i]))
        else:
            for i in range(3):
                if self.components[i] < 0 or self.components[i] > max_cvalues[0]:
                    raise ValueError(str(notation[i])+' value must lie between 0 and ' + str(max_cvalues[i]))
                
    def RGB(self):
        """
        Convert the color to RGB color code
        """
        mode = self.sort.baseSort().arguments
        if mode is None:
            mode = self.DEFAULT_MODE
        if mode.model == colorMode.Model.HSV:
            comps = self.components
            comps = self.HSVtoRGB(comps[0],comps[1],comps[2])
        else:
            comps = self.components
        return '0x'+format(int((comps[0] * 256 + comps[1]) * 256 + comps[2]), '06x')
        
    def equals(self, other):
        """
        check equality
        """
        if self.description is not None or other.description is not None:
            return False
        return self.components == other.components
        
    def compareValue(self, other):
        """
        Compare the value of two colors
        """
        for i in range(3):
            c = self.components[i] - other.components[i]
            if c < 0:
                return -1
            if c > 0:
                return 1
        mode = self.sort.baseSort().arguments
        if mode is None:
            mode = self.DEFAULT_MODE
        if mode.behavior == colorMode.Behavior.RGB_ALPHA:
            c = self.alpha - other.alpha
            if c < 0:
                return -1
            if c > 0:
                return 1
        return 0

    def checkContains(self, other):
        if self.sort != other.sort:
            raise ValueError('Colors must be of the same sort')
        if other.description is not None:
            other.description.registerDelay(self.assoc)
        return self.contains(other)
    
    def contains(self, other):
        """
        Check if this color contains another
        """
        if self.sort != other.sort:
            raise ValueError('Colors must be of the same sort')
        if self.isNil():
            return False
        if other.isNil():
            return True
        return (self.components[0] >= other.components[0] and 
                self.components[1] >= other.components[1] and
                self.components[2] >= other.components[2])
        
    def combine(self, other):
        """
        combine two colors based on the color behavior
        """
        if self.sort != other.sort:
            raise ValueError('Colors must be of the same sort')
        if self.isNil() and other.isNil():
            return False
        if other.isNil():
            return True
            
        mode = self.sort.baseSort().arguments
        if mode is None:
            mode = self.DEFAULT_MODE
            
        thisComps = self.components
        otherComps = other.components
        if mode.model == colorMode.Model.HSV:
            thisComps = self.HSVtoRGB(thisComps[0],thisComps[1],thisComps[2])
            otherComps = self.HSVtoRGB(otherComps[0],otherComps[1],otherComps[2])

        if mode.behavior == colorMode.Behavior.RGB_ALPHA:
            factor = 1 - other.alpha
            if self.alpha == 0 or other.alpha == 1:
                for i in range(3):
                    self.components[i] = otherComps[i]
                self.alpha = other.alpha
            elif self.alpha == 1:
                for i in range(3):
                    self.components[i] = round(otherComps[i] * other.alpha + thisComps[i] * factor)
            elif other.alpha != 0:
                _alpha = other.alpha + self.alpha * factor
                for i in range(3):
                    self.components[i] = round(otherComps[i] * other.alpha + thisComps[i] * factor * self.alpha) / _alpha
                self.alpha = _alpha
                
        elif mode.behavior == colorMode.Behavior.RGB_AVG:
            for i in range(3):
                self.components[i] = round((otherComps[i] + thisComps[i]) / 2)
                
        elif mode.behavior == colorMode.Behavior.RGB_SUM:
            for i in range(3):
                self.components[i] = otherComps[i] + thisComps[i]
            maxCVals = mode.max_cvalues
            if mode.model == colorMode.Model.HSV:
                maxCVals = colorMode.MAX_RGB_VALUES
            if len(maxCVals) > 1:
                for i in range(3):
                    if self.components[i] > maxCVals[i]:
                        self.components[i] = maxCVals[i]
            else:
                for i in range(3):
                    if self.components[i] > maxCVals[0]:
                        self.components[i] = maxCVals[0]
                        
        elif mode.behavior == colorMode.Behavior.RGB_MAX:
            for i in range(3):
                self.components[i] = max(otherComps[i], thisComps[i])
                
                
        if mode.model == colorMode.Model.HSV:
            self.components = self.RGBtoHSV(self.components[0], self.components[1], self.components[2])
            self.reduceValues(mode)
        
        return True


    def common(self, other):
        """
        Find the common color between two colors based on color behavior
        """
        if self.sort != other.sort:
            raise ValueError('Colors must be of the same sort')
        if self.isNil() or other.isNil():
            self.components = [-1,-1,-1]
            return False

        mode = self.sort.baseSort().arguments
        if mode is None:
            mode = self.DEFAULT_MODE
            
        thisComps = self.components
        otherComps = other.components
        if mode.model == colorMode.Model.HSV:
            thisComps = self.HSVtoRGB(thisComps[0],thisComps[1],thisComps[2])
            otherComps = self.HSVtoRGB(otherComps[0],otherComps[1],otherComps[2])
            
        if mode.behavior == colorMode.Behavior.RGB_ALPHA:
            self.alpha = min(self.alpha, other.alpha)
        else:
            for i in range(3):
                self.components[i] = min(otherComps[i], thisComps[i])
                
        if mode.model == colorMode.Model.HSV:
            self.components = self.RGBtoHSV(self.components[0], self.components[1], self.components[2])
            self.reduceValues(mode)
        return True
        
    def complement(self, other):
        """
        Find the complement color between this and another color based on their color behavior
        """
        if self.sort != other.sort:
            raise ValueError('Colors must be of the same sort')

        mode = self.sort.baseSort().arguments
        if mode is None:
            mode = self.DEFAULT_MODE
            
        thisComps = self.components
        otherComps = other.components
        if mode.model == colorMode.Model.HSV:
            thisComps = self.HSVtoRGB(thisComps[0],thisComps[1],thisComps[2])
            otherComps = self.HSVtoRGB(otherComps[0],otherComps[1],otherComps[2])
            
            
        if mode.behavior == colorMode.Behavior.RGB_ALPHA:
            factor = 1 - other.alpha
            if self.alpha == 0 or other.alpha == 1:
                for i in range(3):
                    self.components[i] = otherComps[i]
                self.alpha = other.alpha
            elif self.alpha == 1:
                for i in range(3):
                    self.components[i] = round(otherComps[i] * other.alpha + thisComps[i] * factor)
            elif other.alpha != 0:
                _alpha = other.alpha + self.alpha * factor
                for i in range(3):
                    self.components[i] = round(otherComps[i] * other.alpha + thisComps[i] * factor * self.alpha) / _alpha
                self.alpha = _alpha
                
        elif mode.behavior == colorMode.Behavior.RGB_AVG:
            for i in range(3):
                self.components[i] = round((otherComps[i] + thisComps[i]) / 2)
                
        elif mode.behavior == colorMode.Behavior.RGB_SUM:
            for i in range(3):
                if thisComps[i] < otherComps[i]:
                    self.components[i] = c0
                else:
                    self.components[i] = thisComps[i] - otherComps[i]
                
        elif mode.behavior == colorMode.Behavior.RGB_MAX:
            for i in range(3):
                if thisComps[i] <= otherComps[i]:
                     self.components[i] = c0
                
        if mode.model == colorMode.Model.HSV:
            self.components = self.RGBtoHSV(self.components[0], self.components[1], self.components[2])
            self.reduceValues(mode)
            
        if self.components == [0,0,0]:
            return False
            
        return True
        
        
    def replicateInd(self):
        """
        Create a duplicate of the individual without attributes
        
        :returns: the duplicate object
        """
        mode = self.sort.baseSort().arguments
        if mode is None:
            mode = self.DEFAULT_MODE
        if mode.behavior == colorMode.Behavior.RGB_ALPHA:
            duplicate = color(self.sort, self.components[0],self.components[1],self.components[2], self.alpha, mode.model == colorMode.Model.RGB)
        else:
            duplicate = color(self.sort, self.components[0],self.components[1],self.components[2], RGB = mode.model == colorMode.Model.RGB)
        if self.description is not None:
            duplicate.description = self.description.replicateInd()
        return duplicate

    def evaluateInd(self, paramSet):
        """
        Evaluate all the description within the individual with the paramSet
        """
        if self.description is not None:
            desc = self.description.evaluateInd(paramSet)
            root = None
            if desc.tree[0].type == 'description-tuple':
                root = desc.tree[0].tuple
            elif len(desc.tree) >= 3:
                root = desc.tree
            mode = self.sort.baseSort().arguments
            ind = None
            if mode.behavior == colorMode.Behavior.RGB_ALPHA:
                ind = color(self.sort, root[0].value, root[1].value, root[2].value, root[3].value)
            elif mode.model == colorMode.Model.HSV:
                ind = color(self.sort, root[0].value, root[1].value, root[2].value, False)
            else:
                ind = color(self.sort, root[0].value, root[1].value, root[2].value)
            return ind
        return self.replicateInd()

    @staticmethod
    def createDescriptionColor(sort, desc):
        """
        Create a color with a description

        :param sort: sort of the new color object
        :type sort: ~sortal.sort.sort.sort
        :param desc: the description representing the new color object
        :type desc: str
        :returns: :class:`~sortal.ind.color.color` -- the individual parsed
        """
        desc = description.parse(parseReader(desc), sort)
        ind = color(sort, 0, 0, 0)
        ind.description = desc
        return ind

    @staticmethod
    def parse(reader, sort):
        """
        Parse the color from a sortal parseReader (from string)

        :param reader: the reader of the string to be parsed
        :type reader: ~sortal.io.parseReader.parseReader
        :param sort: sort of the new object
        :type sort: ~sortal.sort.sort.sort
        :returns: :class:`~sortal.ind.color.color` -- the individual parsed
        """
        if reader.token() == '`':
            desc = description.parse(reader, sort)
            ind = color(sort, 0, 0, 0)
            ind.description = desc
            return ind
        mode = sort.baseSort().arguments
        if reader.token() != '(':
            color.parseError(reader,mode)
        r = reader.newValue()
        if reader.newToken() != ',':
            color.parseError(reader,mode)
        g = reader.newValue()
        if reader.newToken() != ',':
            color.parseError(reader,mode)
        b = reader.newValue()
        if mode.behavior == colorMode.Behavior.RGB_ALPHA:
            if reader.newToken() != ',' :
                color.parseError(reader,mode)
            alpha = reader.newValue()
        if reader.newToken() != ')':
            color.parseError(reader,mode)
        if mode.behavior == colorMode.Behavior.RGB_ALPHA:
            ind = color(sort, r, g, b, alpha)
        elif mode.model == colorMode.Model.HSV:
            ind = color(sort, r, g, b, False)
        else:
            ind = color(sort, r, g, b)
        reader.newToken()
        return ind

    def map(self, assignment, other):
        if self.sort != other.sort:
            raise ValueError('colors must be of the same sort')
        if self.description is not None:
            if other.description is not None:
                otherDesc = other.description
            else:
                otherDesc = description(other.sort, '`' + str(other) + '`')
            if len(self.description.tree) != len(otherDesc.tree):
                return False
            for i in range(len(self.description.tree)):
                if not self.description.tree[i].map(assignment, self.sort, otherDesc.tree[i]):
                    return False
            return True

        count = 0
        if self.assoc.hasAttribute():
            for i in self.assoc.attribute:
                if i.map(assignment, other):
                    count+=1
        if count > 0:
            return True
        return False 

        return True

    @staticmethod
    def parseError(reader, mode):
        """
        Specific color parsing error
        """
        if mode.behavior == colorMode.Behavior.RGB_ALPHA:
            raise ParsingError(reader, 'color has to be in format: "(Red, Green, Blue, Alpha)"')
        elif mode.model == colorMode.Model.HSV:
            raise ParsingError(reader, 'color has to be in format: "(Hue, Saturation, Value)"')
        else:
            raise ParsingError(reader, 'color has to be in format: "(Red, Green, Blue)"')
