# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import object
from ..planeSegment import planeSegment2D
from ..lineSegment import lineSegment2D
from ..line import line2D
from ...struct.planeStruct import bSegment
from ...struct.vector import vector2D
from ...form.coForm import coForm

class crossAlgebraPlane2D(object):
    """
    Cross algebra operations with plane (line vs plane, point vs plane, plane vs plane)

    A collection of static functions to check if geometric objects intersect or within a
    plane segment or to remove parts of objects that are not within a plane segment.

    These are to be used for limiting a shape down within a smaller area for rule detection
    """
    
    @staticmethod
    def freeSelection(vectorList, form):
        """
        create a new form from the elements of a form that are within a plane-segment formed
        by positions in a vectorList
        """
        plane = planeSegment2D(form.sort, vectorList)
        return crossAlgebraPlane2D.crossAlgebraForm(plane, form)
    
    @staticmethod
    def crossAlgebraForm(pl, form):
        """
        create a new form from the elements of a form that are within a plane-segment
        """
        newForm = form.sort.newForm()
        if form.type == 'form-metaForm':
            for i in form.elements:
                temp = crossAlgebraPlane2D.crossAlgebraForm(pl,i)
                if not temp.isEmpty():
                    newForm.insert(temp)
        else:
            temp = crossAlgebraPlane2D.assessForm(pl,form)
            if temp is not None:
                return temp
        return newForm
                
    
    @staticmethod
    def assessForm(pl, form):
        """
        create a new form from the elements of a form that are within a plane-segment.
        
        part of crossAlgebraForm function
        """
        if len(form.elements) == 0:
            return form
        if form.elements[0].type == 'ind-planeSegment2D' or form.elements[0].type == 'ind-planeSegmentP2D':
            return crossAlgebraPlane2D.plane_plane(pl,form)
        elif form.elements[0].type == 'ind-line2D':
            return crossAlgebraPlane2D.plane_line(pl,form)
        elif form.elements[0].type == 'ind-lineSegment2D' or form.elements[0].type == 'ind-lineSegmentP2D':
            return crossAlgebraPlane2D.plane_lineSeg(pl,form)
        elif form.elements[0].type == 'ind-point2D' or form.elements[0].type == 'ind-pointP2D':
            return crossAlgebraPlane2D.plane_point(pl,form)
        return form
    
    @staticmethod
    def plane_plane(pl, planeForm):
        """
        create a new form from the elements of a form that are within a plane-segment.
        
        specifically for form of plane segments
        """
        one = coForm(pl.sort)
        one.insert(pl)
        
        result = coForm(planeForm.sort)
        for i in planeForm.elements:
            two = coForm(planeForm.sort)
            two.insert(i)
            
            classes = planeSegment2D.classification(one, two)
            if classes.disjoint() or classes.touches():
                continue
            else:
                temp = coForm(planeForm.sort)
                i.construct(classes.productSet(), temp)
                if i.hasAttribute():
                    for j in temp.elements:
                        j.addAttribute(i.attribute)
                temp.move(result)
        return result
    
    @staticmethod
    def plane_line(pl, lineForm):
        """
        create a new form from the elements of a form that are within a plane-segment.
        
        specifically for form of lines
        """
        one = coForm(pl.sort)
        one.insert(pl)
        
        result = lineForm.sort.newForm()
        for line in lineForm.elements:
            intersectionList = []
            for i in pl.boundaries:
                for j in i.canonical:
                    pLine = line2D(line.sort, j.start, j.end)
                    if pLine.intersect(line):
                        intersection = line.intersection(pLine)
                    else:
                        continue
                    if intersection >= j.start and intersection <= j.end:
                        intersectionList.append(intersection)
                        
            if len(intersectionList) == 0:
                continue
            else:
                result.insert(line)
        result.maximalize()
        return result
        
    @staticmethod
    def planeLineIntersections(bound, lineForm):
        """
        find the intersections between a plane boundary and a form of lines
        """
        one = coForm(lineForm.sort)
        pl = planeSegment2D(lineForm.sort, bound)
        one.insert(pl)
        result = []
        for line in lineForm.elements:
            intersectionList = []
            for i in pl.boundaries:
                for j in i.canonical:
                    pLine = line2D(line.sort, j.start, j.end)
                    if pLine.intersect(line):
                        intersection = line.intersection(pLine)
                    else:
                        continue
                    if intersection >= j.start and intersection <= j.end and intersection not in intersectionList:
                        intersectionList.append(intersection)
            if len(intersectionList) == 0:
                continue
            else:
                result.append(intersectionList)
        return result
        
    @staticmethod
    def plane_lineSeg(pl, lineSegForm):
        """
        create a new form from the elements of a form that are within a plane-segment.
        
        specifically for form of line segments
        """
        one = coForm(pl.sort)
        one.insert(pl)
        bSeg = []
        
        for lineSeg in lineSegForm.elements:
            bSeg.append(bSegment(lineSeg.start,lineSeg.end,lineSeg))
        classes = planeSegment2D.classification_line(one, bSeg, False)
        if classes.disjoint():
            return lineSeg.sort.newForm()
        else:
            result = lineSeg.sort.newForm()
            for i in classes.innerSet():
                temp = lineSegForm.elements[0].__class__(lineSeg.sort,i.start, i.end)
                if lineSegForm.sort.type == 'sort-attribute':
                    temp.addAttribute(i.assoc.attribute)

                result.insert(temp)
            result.maximalize()
            return result
    
    @staticmethod
    def plane_point(pl, pointForm):
        """
        create a new form from the elements of a form that are within a plane-segment.
        
        specifically for form of points
        """
        one = coForm(pl.sort)
        one.insert(pl)
        
        result = pointForm.sort.newForm()
        
        bSeg = []
        for point in pointForm.elements:
            vec = None
            lenV = 0
            for i in pl.boundaries:
                stoploop = False
                for j in i.canonical:
                    pointVect = point.value
                    u = j.direction
                    v = pointVect.subtract(j.root())
                    if u.parallel(v):
                        if pointVect >= j.start and pointVect <= j.end: 
                            result.insert(point)
                            stoploop = True
                            break 
                        else:
                            continue
                    newVec = j.root().add(u.scale(v))
                    newLenV = newVec.subtract(pointVect)
                    if vec is None or newLenV.squareLength() < lenV.squareLength():
                        vec = newVec
                        lenV = newLenV
                if stoploop:
                    break
            if not stoploop:
                bSeg.append(bSegment(vec,pointVect, point))
            
        classes = planeSegment2D.classification_line(one, bSeg, False)
        for i in classes.containedSet():
            result.insert(i.assoc)
        result.maximalize()
        return result
        
    