# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import super
from builtins import str


from .individual import individual
from .descriptions.tree import DescriptionEvaluatingError
from .descriptions.symbols import symbols

from .descriptions.condition import condition
from .descriptions.function import function
from .descriptions.literal import literal
from .descriptions.numberLiteral import numberLiteral
from .descriptions.numericExpression import numericExpression
from .descriptions.parameter import parameter
from .descriptions.reference import reference
from .descriptions.stringExpression import stringExpression
from .descriptions.stringLiteral import stringLiteral
from .descriptions.tuple import tupleDesc
from .descriptions.tupleExpression import tupleExpression

from ..io.parseReader import parseReader, ParsingError
from ..sort.sort import sort as __sortClass__


class description(individual):
    """
    Description object

    **Object Parameters:** (not including inherited attributes)

    * **tree** (list of :class:`~sortal.ind.descriptions.tree.tree`) -- list of components contained in the description
    * **section** (str) -- sectioning symbol separating the contents of the description. Default to None

    **Class Properties:**

    * **Behavior**: :class:`~sortal.form.discrete.discrete`
    * **Sort Argument**: :class:`~sortal.sort.parameter.argument.argument.argumentType`.NONE
    * **Mapping**: :class:`~sortal.map.parametricMap.parametricMap`


    """

    condition = condition
    function = function
    literal = literal
    numberLiteral = numberLiteral
    numericExpression = numericExpression
    parameter = parameter
    reference = reference
    stringExpression
    stringLiteral = stringLiteral
    tupleDesc = tupleDesc
    tupleExpression = tupleExpression
    
    detectionDelayReg = []
    
    type = 'ind-description'
    
    def __init__(self, sort, s, section = None, dtree = None, *args, **kwargs):
        """
        Initialize a Description object

        :param sort: sort of the description
        :type sort: ~sortal.sort.sort.sort
        :param s: content of the description in string format (can be an empty string or None if dtree is not None)
        :type s: str
        :param section: (optional) sectioning symbol separating the contents of the description. Default to None
        :type section: str
        :param dtree: an existing description tree
        :type dtree: ~sortal.ind.descriptions.tree.tree
        """
        super(description, self).__init__(sort)
        if dtree is None:
            self.section = symbols.NONE
            self.tree = []
            self.descParse(parseReader(s))
        else:
            self.section = section
            self.tree = dtree

    def isNil(self):
        """
        Check whether the description is nil
        """
        if len(self.tree) == 0:
            return True
        return False
        
    def toString(self, printKey):
        """
        Converting the description to string format
        """
        s = ''
        if printKey and self.key is not None:
            s += str(self.key)+' '
        if self.isNil():
            return 'None'
        s += '`'+self.section
        for i in self.tree:
            s += str(i) + self.section
        s += '`'
        return s

    def printVal(self, p = 1):
        """
        Print out a string representing the value of this description, used in form.printE()
        """
        if p == 0:
            return self.printDesc()
        return str(self)
        
    def printDesc(self):
        """
        Print out a string representing the value of this description, this way of printing includes the type of each description sub-element
        """
        s = self.section
        for i in self.tree:
            s += i.printVal() + self.section
        return s
        
    def getValue(self):
        """
        get the content of the description (in string form)
        """
        return str(self)
        
    def content(self):
        """
        get the content of the description (in description tree form)
        """
        return self.tree
        
    def replicateInd(self):
        """
        Create a duplicate of the individual without attributes
        
        :returns: the duplicate object
        """
        return description(self.sort, str(self))
            
    def isKeyword(self, name):
        """
        Check if a string is a literal, a function name or an existing sort name
        """
        if literal.isLiteral(name): 
            return True
        if function.isFunctionName(name): 
            return True
        if __sortClass__.checkSort(name) == True: 
            return True
        return False
                
    def __len__(self):
        return len(self.tree)
        
    def checkPartOf(self, form):
        """
        Check if the description is a part of a form
        """
        # if len(self.tree) == 1 and self.tree[0].type == 'description-literal' and (self.tree[0].name == 'nil' or self.tree[0].name == 'e'):
        #     return True
        for other in form:
            if len(self.tree) != len(other.tree):
                continue
            i = 0
            check = False
            while i < len(self.tree):
                if self.tree[i].checkPartOf(other.tree[i]):
                    check = True
                    break
                i += 1
            if check:
                return True
        return False
        
    def compareValue(self, other):
        c = len(self.tree) - len(other.tree)
        i = 0
        while c == 0 and i < len(self.tree):
            c = self.tree[i].compareValue(other.tree[i])
            i+=1
        return c
        
    def getNumber(self, assignment):
        for i in self.tree:
            num = i.getNumber(assignment, self.sort)
        return num
            
    def equals(self, other):
        """
        Check if two descriptions are equal
        """
        #print('comparing', self,'<><>', other)
        c = len(self.tree) - len(other.tree)
        if c != 0:
            #print('false 1')
            return False
        for i in range(len(self.tree)):
            if self.tree[i] != other.tree[i]:
                #print('false')
                return False
        #print('true')
        return True
    
    def map(self, assignment, other):
        if self.sort != other.sort:
            raise ValueError('descriptions must be of the same sort')
        if len(self.tree) != len(other.tree):
            return False
        for i in range(len(self.tree)):
            if not self.tree[i].map(assignment, self.sort, other.tree[i]):
                return False
        return True
        
    def contains(self, other):
        """
        Check if this description contains another
        """
        if self.sort != other.sort:
            raise ValueError('descriptions must be of the same sort')
        for i in other.tree:
            check = False
            for j in self.tree:
                if i.type == j.type and j.contains(i):
                    check = True
                    break
            if check == False:
                return False 
        return True
        
    def evaluateInd(self, assignment):
        """
        Transform the description with a transformation set
        """
        result = description(self.sort, None, self.section, [])
        for i in self.tree:
            result.tree.append(i.evaluate(assignment, self.sort))

        # if len(result.tree) == 1 and assignment.retrieve(self.sort, symbols.VALUE) is None:
        #     assignment.add(self.sort, symbols.VALUE, result.tree[0])
            
        # if result.tree[0].type == 'description-tuple':
        #     print(self.printDesc())
        #     raise ValueError


        # print('-----------------')
        # print('desc:\t\t', self)
        # print('val:\t\t', result)

        return result
        

    def checkTransform(self, transform, compareForm):
        """
        Check if the description is part of a form after trasnforming with a transformation set
        """
        if compareForm.type == 'form-metaForm':
            for i in compareForm:
                if i.sort == self.sort:
                    self.registerDelay(i)
                    break
        elif compareForm.sort == self.sort:
            self.registerDelay(compareForm)

        # assignment = transform.retrieveTransform(paramSet)
        # evalDesc =  self.evaluateInd(assignment)
        # if self.attribute is not None:
        #     evalDesc.addAttribute(self.attribute)
        # if compareForm.containsInd(evalDesc):
        #     return True

        return True
        
    def registerDelay(self, otherForm):
        """
        Mark this description for delayed matching when there is not enough information to 
        solve the description conditions/references in the current state of detection
        """
        if not otherForm.isEmpty():
            self.detectionDelayReg.append([self, otherForm])

    @staticmethod
    def resolveDelays(transform):
        """
        Try to resolve the delayed description matchings with a specific transformation set.

        :returns: boolean -- True if the description is successfully solved, False otherwise
        """
        for i in description.detectionDelayReg:
            check = False
            try:
                desc = i[0].evaluateInd(transform)
            except DescriptionEvaluatingError:
                return False
            if i[1].type == 'ind-description' and desc == i[1]:
                check = True
            elif 'form' in i[1].type:
                for j in i[1].elements:
                    if desc == j:
                        check = True
            if check == False:
                return False
        return True
        
    @staticmethod
    def clearDelays():
        """
        Clear all the delayed descriptions
        """
        description.detectionDelayReg = []
        
    @staticmethod        
    def parse(reader, sort):
        """
        Parse the description from a sortal parseReader (from string)

        :param reader: the reader of the whole string of the individual
        :type reader: ~sortal.io.parseReader.parseReader
        :param sort: sort of the new object
        :type sort: ~sortal.sort.sort.sort
        :returns: :class:`~sortal.ind.description.description` -- the individual parsed
        """
        if reader.token() != '`':
            raise ParsingError(reader, "expected a '`'")
        startIndex = reader.index()
        while True:
            if reader.newToken() == '`':
                endIndex = reader.index() + 1
                break
        s = reader.text[startIndex:endIndex]
        reader.newToken()
        desc = description(sort, s)
        return desc
        
        
    def similarFragment(self, other):
        """
        Check if two descriptions have the exact same content
        """
        if len(self.tree) != len(other.tree):
            return False
        for i in range(len(self.tree)):
            if not self.tree[i].similarFragment(other.tree[i]):
                return False
        return True

    def mergeFragment(self, other):
        """
        Merge the content of two descriptions
        """
        for i in range(len(self.tree)):
            self.tree[i].mergeFragment(other.tree[i])
        return

    def descParse(self, reader):
        """
        Parse the content of the description. This function is the first one to be 
        called when parsing the content of the description. The description string
        must be in back-quotes(`).

        :param reader: the reader of the string of the description content to be parsed
        :type reader: ~sortal.io.parseReader.parseReader
        :returns: sortal.descriptions.tree.tree -- the description tree parsed
        """
        if reader.token() != symbols.DESCRIPTION:
            raise ParsingError(reader, "expected a '`'")
        ###############
        reader.newToken()
        entity = self.treeParse(reader) #######
        if entity is not None:
            self.tree.append(entity)
        else:
            # raise error if token is invalid symbol
            if reader.token() in [symbols.PAREN_CLOSE, symbols.ANGLE_CLOSE,symbols.SQUARE_CLOSE,
                                  symbols.SET_OPEN,symbols.SET_CLOSE,symbols.COMMA,symbols.SEMICOLON,
                                  symbols.CONCAT,symbols.CONDITIONAL,symbols.TIMES,symbols.DIVIDED,
                                  symbols.MODULO,symbols.POWER]:
                raise ParsingError(reader, 'unexpected character')
            else:
                self.section = reader.token()
        while self.section == reader.token():
            reader.newToken()
            entity = self.treeParse(reader)
            if entity is not None:
                self.tree.append(entity)

    def treeParse(self, reader):
        """
        Parse the content of the description. This is called within descParse as a continuation.

        :param reader: the reader of the string of the description content to be parsed
        :type reader: ~sortal.io.parseReader.parseReader
        :returns: sortal.descriptions.tree.tree -- the description tree parsed
        """
        # if token = '+' --> tuple expression
        if reader.token() == symbols.PLUS:
            return tupleExpression(None, tupleDesc.parse(self, reader))
        entity = self.parseEntity(self, reader)

        if entity is None:
            return None
        nextEntity = self.parseEntity(self, reader)
        if nextEntity is None:
            return entity
        return tupleExpression.parseContinue(self, entity, nextEntity, reader)
        
    def getID(self):
        return uuid.uuid1()

        
    @staticmethod
    def parseEntity(desc, reader, numExpr = False):
        """
        Parse one unknown fragment. Will branch off into more specific description fragment's parse function
        when there are enough information gathered.

        :param reader: the reader of the string of the description content to be parsed
        :type reader: ~sortal.io.parseReader.parseReader
        :returns: the description fragment
        """
        result = None
        ################ string --> string expression ################
        if reader.token() == parseReader.STRING: 
            result = stringExpression.parse(desc, reader)
        ################ number / -number --> numeric expression ################
        elif reader.token() == parseReader.NUMBER or reader.token() == parseReader.MINUS:
            result = numericExpression.parse(desc, reader)
        ################ identifier ################
        elif reader.token() == parseReader.IDENTIFIER or reader.token() == '#':
            if reader.token() == '#':
                result = reference.parse(desc, reader) # newToken() included
            else:
                name = reader.value()
                # if value is in literal.operators/literal.keywords --> literal
                if literal.isLiteral(name):
                    result = literal.getLiteral(name)
                    reader.newToken()
                # if value is in function.FNAME --> function
                elif function.isFunctionName(name) and not reader.previewToken() == '.':
                    result = function.parse(desc, reader) # newToken() included
                # if value is existing sort name --> sort
                elif __sortClass__.checkSort(name) or name == 'rhs':
                    result = reference.parse(desc, reader) # newToken() included
                # else --> parameter
                else:
                    result = parameter.parse(desc, reader, True) # newToken() included
            #------------------------------------------------#
            # if token == '.' --> continue as string expression
            if reader.token() == symbols.CONCAT:
                result = stringExpression.parseContinue(desc, result, reader)
            # if token == '-' '*' '/' '%' '^' --> continue as string expression
            elif reader.token() in [parseReader.MINUS, symbols.TIMES, 
                                    symbols.DIVIDED, symbols.MODULO, symbols.POWER]:
                result = numericExpression.parseContinue(desc, result, reader)
            # if token == '+' --> either tuple expression or numeric expression
            elif reader.token() == symbols.PLUS:
                # if result is parameter and preview token == <> or [] --> tuple expression
                if (result.type == 'description-parameter' and 
                    (reader.previewToken() == symbols.ANGLE_OPEN or 
                     reader.previewToken() == symbols.SQUARE_OPEN)):
                         reader.newToken()
                         result = tupleExpression(result, description.parseEntity(desc, reader))
                # if preview token == () --> also either tuple expression or numeric expression
                elif reader.previewToken() == symbols.PAREN_OPEN:
                    reader.newToken()
                    # parse the terms inside the brackets
                    entity = description.parseEntity(desc, reader) 
                    # if entity is tuple --> tuple expression
                    if entity.type == 'description-tuple':
                        result = tupleExpression(result, entity)
                    # if entity is not tuple --> numeric expression
                    else:
                        if result.type == 'description-parameter' and result.hasCondition():
                            expression = numericExpression(result.condition.operand, literal.getLiteral(symbols.PLUS), entity)
                            expression.parseNext(desc, reader)
                            result.condition.operand = expression
                        else:
                            result = numericExpression(result, literal.getLiteral(symbols.PLUS), entity)
                            result.parseNext(desc, reader)
                # if preview token is not () --> continue parsing as numeric expression
                else:
                    result = numericExpression.parseContinue(desc, result, reader)
        ################ <> / [] --> tuple expression ################
        elif reader.token() == symbols.ANGLE_OPEN or reader.token() == symbols.SQUARE_OPEN:
            result = tupleExpression.parse(desc, reader)
        ################ () --> parse the terms inside the bracket################
        elif reader.token() == symbols.PAREN_OPEN:
            reader.newToken()
            result = description.parseEntity(desc, reader)
            # if the bracket is closed --> complete the tuple with no separator
            if reader.token() == symbols.PAREN_CLOSE:
                reader.newToken()
                # check the result gotten
                # if numeric expression or reference --> continue parsing as numeric expression or string expression
                #if result is not None and (result.type == 'description-numericExpression' or result.type == 'description-reference'):
                if result is not None and result.type == 'description-reference':
                    # if token == '+', '-', '*', '/', '^' --> numeric expression
                    if reader.token() in [symbols.PLUS, parseReader.MINUS, 
                                          symbols.TIMES, symbols.DIVIDED, 
                                          symbols.MODULO, symbols.POWER]:
                        result = numericExpression.parseContinue(desc, result, reader)
                    # if token == '.' --> string expression
                    elif  reader.token() == symbols.CONCAT:
                        result = stringExpression.parseContinue(desc, result, reader)
                elif result is not None and result.type == 'description-numericExpression':
                    if numExpr:
                        return result
                    if reader.token() in [symbols.PLUS, parseReader.MINUS, 
                                          symbols.TIMES, symbols.DIVIDED, 
                                          symbols.MODULO, symbols.POWER]:
                        result = numericExpression.parseContinue(desc, result, reader)
                    # if token == '.' --> string expression
                    elif  reader.token() == symbols.CONCAT:
                        result = stringExpression.parseContinue(desc, result, reader)
                # if not numeric expression or reference --> parenthesis tuple with UNKNOWN(0) as separator and result as entity
                else: 
                    result = tupleDesc(symbols.PAREN_OPEN, symbols.UNKNOWN, result)
            # if the bracket is not closed but there is , or ; --> separator
            elif reader.token() == symbols.COMMA or reader.token() == symbols.SEMICOLON:
                result = tupleDesc(symbols.PAREN_OPEN, reader.token(), result)
                result.parseContinue(desc, reader)
            # raise error if neither separator or bracket close
            else:
                raise ParsingError(reader, "expected a ')'")
        return result