# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import str

from .tree import tree
from .symbols import symbols
from ...io.parseReader import ParsingError, parseReader
import uuid 


class function(tree):
    """
    description functions.

    List of functions:

    * Function for :class:`~sortal.ind.descriptions.numberLiteral.numberLiteral`:
        * abs(numLit): absolute value of the number
        * round(numLit): rounding the number to the nearest integer
        * ceil(numLit): rounding up the number to the nearest integer
        * floor(numLit): rounding down the number to the nearest integer
        * sqrt(numLit): square root value of the number
        * sin(numLit): sine value of the angle (angle value in radian)
        * cos(numLit): cosine value of the angle (angle value in radian)
        * tan(numLit): tan value of the angle (angle value in radian)
        * asin(numLit): angle value from sine value (in radian)
        * acos(numLit): angle value from cosine value (in radian)
        * atan(numLit): angle value from tan value (in radian)
        * atan2(numLit): angle value from tan value (in radian)
        * deg(numLit): convert radian to degree
        * rad(numLit): convert degree to radian
        * random(numLit) or random(numLit1, numLit2): generate a random number between two values, if only 1 value is given, find a random value between 0 and the given value
    * Function for :class:`~sortal.ind.descriptions.stringLiteral.stringLiteral`:
        * length(strLit): the length of the string
        * left(strLit): the leftmost character in the string
        * right(strLit): the rightmost character in the string
    * Function for :class:`~sortal.ind.descriptions.tuple.tupleDesc`:
        * length(tuple of objects): the number of objects in the tuple
        * first(tuple of objects): the first object in the tuple
        * last(tuple of objects): the last object in the tuple
        * unique(tuple of objects): get a tuple of unique objects in the tuple
        * min(tuple of numLits): the minimum number in the tuple
        * max(tuple of numLits): the maximum number in the tuple
        * round(tuple of numLits): rounding all the numbers to the nearest integer
        * random(tuple of 1-3 numLits): generate a random number between the first two values, third value to be the step increasement
        * vectoradd(tuple of 2 vector tuples): adding two vectors, the result would be a vector tuple
        * vectorsubtract(tuple of 2 vector tuples): subtracting two vectors, the result would be a vector tuple
        * vectorscale(tuple of 1 vector tuple and 1 numLit): scaling a vector with a factor, the result would be a vector tuple
        * mag(2 tuple vectors) or mag(tuple of 2 tuple vectors): calculate the distance magnitude between two positional vectors
        * proj(tuple of 3 tuple vectors): find the projection of a point on a line (with the line direction and root vectors)
            Input structure: (line_direction, line_root, point_coord)
        * angle(2 tuple vectors) or angle(tuple of 2 tuple vectors): calculate the angle between two vectors (in radian)
        * dotproduct(2 tuple vectors) or dotproduct(tuple of 2 tuple vectors): calculate the dot product of two vectors
        * crossproduct(2 tuple vectors) or crossproduct(tuple of 2 tuple vectors): calculate the cross product of two vectors
        * segments(tuple of objects): split the tuple into segments of two objects
            e.g. segments((a,b,c,d,e)) -> ((a,b),(b,c),(c,d),(d,e))
        * pairwise(tuple of objects): split the tuple into pairs of two objects
            e.g. pairwise((a,b,c,d,e)) -> ((a,b),(c,d))
        * loops(tuple of objects): find all the loops in a tuple
            e.g. loops((a,b,c,d,a,e,f,b)) -> ((a,b,c,d),(a,e,f,b))
        * adjacencies(tuple of 1 loop tuple and 1 door tuple): find the adjacencies matrix between loops for a tuple

    | \* vector tuple: a tuple of 2 or 3 number literal representing 2D or 3D vector
    | \* loop tuple: a tuple of a number of loops
    | \* door tuple: a tuple of a number of doors (tuple of 2 objects)
    """
    FNAME = {}
    type = 'description-function'
    def __init__(self, name):
        self.name = name
        self.argument = None
        self.assist = None
        self.id = uuid.uuid1()
    
    def __eq__(self, other):
        if other.type != 'description-function':
            return False
        if self.name != other.name or self.argument != other.argument:
            return False
        return self.assist is None or self.assist == other.assist
        
    def __lt__(self, other):
        return self.compareValue(other) == -1
        
    def __gt__(self, other):
        return self.compareValue(other) == 1

    def __hash__(self):
        hashNum = 5
        hashNum = 23 * hashNum + hash(self.name)
        hashNum = 23 * hashNum + hash(self.argument)
        hashNum = 23 * hashNum + hash(self.assist)
        return hashNum
    
    def compareValue(self, other):
        if other.type == 'description-function':
            if self.name > other.name:
                return 1
            elif self.name < other.name:
                return -1
            if self.argument > other.argument:
                return 1
            elif self.argument < other.argument:
                return -1
            if self.assist is None:
                if other.assist is None:
                    return 0
                else:
                    return -1
            if self.assist > other.assist:
                return 1
            elif self.assist < other.assist:
                return -1
            return 0
        elif (other.type == 'description-stringExpression' or
              other.type == 'description-numericExpression' or
              other.type == 'description-tupleExpression'):
                  return -other.compareValue(self)
        c = hash(self.__class__) - hash(other.__class__)
        if c > 0:
            return 1
        elif c < 0:
            return -1
        return 0
    
    # def __str__(self):
    #     result = ''
    #     result += self.name + symbols.PAREN_OPEN + str(self.argument)
    #     if self.assist is not None:
    #         result += symbols.COMMA + ' ' + str(self.assist)
    #     result += symbols.PAREN_CLOSE
    #     return result

    def __str__(self):
        if self.assist is None and self.argument.type == 'description-tuple' and self.argument.bracket == symbols.PAREN_OPEN:
            return self.name + '(' + str(self.argument) + ')'
        result = ''
        result += self.name + symbols.PAREN_OPEN + str(self.argument)
        if self.assist is not None:
            result += symbols.COMMA + ' ' + str(self.assist)
        result += symbols.PAREN_CLOSE
        return result

    def printVal(self):
        result = ''
        result += self.name + symbols.PAREN_OPEN + self.argument.printVal()
        if self.assist is not None:
            result += symbols.COMMA + ' ' + self.assist.printVal()
        result += symbols.PAREN_CLOSE
        return 'func:'+result
        
    def evaluate(self, assignment, sort):
        """
        evaluate the function. The result can be any type of description component based on the
        function itself
        """
        # if self.name == 'random':
        #     string = str(self)

        #     #####
        #     string = self.id
        #     # print('........',self.id)
        #     if string in assignment.results:
        #         return assignment.results[string]
        #     arg = self.argument.evaluate(assignment, sort)
        #     if self.assist is None:
        #         result = arg.apply(self.name)
        #     else:
        #         result = arg.apply(self.name, self.assist.evaluate(assignment, sort))
        #     assignment.results[string] = result
        #     return result
        arg = self.argument.evaluate(assignment, sort)
        if self.assist is None:
            return arg.apply(self.name)
        return arg.apply(self.name, self.assist.evaluate(assignment, sort))
        
    @staticmethod
    def reserve(name):
        """
        reserve the function name
        """
        name = name.lower()
        function.FNAME[name] = name
        
    @staticmethod
    def isFunctionName(name):
        """
        check if a string is a function name
        """
        name = name.lower()
        if name in function.FNAME:
            return True
        return False
    
    @staticmethod
    def parse(desc, reader):
        """
        parse a description function
        """
        name = reader.value().lower()
        # token must be identifier with parenthesis after e.g. adjacencies(a4, a5 a6), length(‘‘room’’)
        if reader.token() != parseReader.IDENTIFIER or reader.newToken() != symbols.PAREN_OPEN:
            raise ParsingError(reader, 'expected an identifier followed by a "("')
        
        if not function.isFunctionName(name):
            raise ParsingError(reader, "function '" + name + "' not recognized")
        # create function with function name
        result = function(name)
        
        reader.newToken()
        # parse the part inside the function
        result.argument = desc.parseEntity(desc, reader)

        # if token is not ')' or ',' --> could be part of a tuple expression --> parse again
        if reader.token() != symbols.PAREN_CLOSE and reader.token() != symbols.COMMA:
            nextEntity = desc.parseEntity(desc, reader)
            # if parse return None --> error
            if nextEntity is None:
                raise ParsingError(reader, "expected a ',' or ')'")
            # continue parsing as tuple expression, save it as argument
            result.argument = desc.tupleExpression.parseContinue(desc, result.argument, nextEntity, reader)
        # if token is ',' --> next parameter for the function --> parse again
        if reader.token() == symbols.COMMA:
            reader.newToken()
            # parse again as assist
            result.assist = desc.parseEntity(desc, reader)
            # if token is not ')' --> could be tuple expression --> parse again
            if reader.token() != symbols.PAREN_CLOSE:
                nextEntity = desc.parseEntity(desc, reader)
                # if next is None --> error
                if nextEntity is not None:
                    # set assist as the tuple expression
                    result.assist = desc.tupleExpression.parseContinue(desc, result.assist, nextEntity, reader)
                    # raise ParsingError(reader, "expected a ',' or ')'")
        while reader.token() == symbols.COMMA:
            reader.newToken()
            if result.assist is not None:
                result.argument = desc.tupleDesc(symbols.PAREN_OPEN, symbols.COMMA, result.argument)
                result.argument.addEntity(result.assist)
                result.assist = None
            nextEntity = desc.parseEntity(desc, reader)
            if nextEntity is None:
                raise ParsingError(reader, "expected a description component")
            result.argument.addEntity(nextEntity)
        if reader.token() != symbols.PAREN_CLOSE:
            raise ParsingError(reader, "expected a ',' or ')'")
        reader.newToken()
        return result
    
