# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import range
from builtins import str


from .tree import tree
from .symbols import symbols
from .function import function
from ...io.parseReader import ParsingError, parseReader
from ...struct.coordinate import coordinate, sqrt, sin, cos, acos, asin, tan, isint, pi, atan2, atan
from random import randint

class numberLiteral(tree):
    """
    description component representing number values, can exist on its own as a number in the description
    or the result of functions or numeric expressions.
    """
    ABS = 'abs'
    ROUND = 'round'
    CEIL = 'ceil'
    FLOOR = 'floor'
    SQRT = 'sqrt'
    SIN = 'sin'
    COS = 'cos'
    TAN = 'tan'
    ASIN = 'asin'
    ACOS = 'acos'
    ATAN = 'atan'
    ATAN2 = 'atan2'
    
    RANDOM = 'random'

    DEG = 'deg'
    RAD = 'rad'
    FUNCTIONS = [ABS,SQRT,SIN,COS,TAN,ASIN,ACOS,ATAN,ATAN2,DEG,RAD,ROUND,FLOOR,CEIL,RANDOM]
    ASSISTS = [None for i in range(len(FUNCTIONS))]
    
    for fName in FUNCTIONS:
        function.reserve(fName)

    type = 'description-numberLiteral'
    def __init__(self, value):
        self.value = value
        
    def __str__(self):
        if isinstance(self.value, coordinate):
            if isint(self.value.value):
                return str(int(self.value.value))
        return str(self.value)
        
    def printVal(self):
        return 'numLit:' + str(self)
        
    def __eq__(self, other):
        if other.type != 'description-numberLiteral':
            return False
        return self.value == other.value
        
    def __len__(self):
        return len(str(self.value))
        
    def __lt__(self, other):
        return self.compareValue(other) == -1
        
    def __gt__(self, other):
        return self.compareValue(other) == 1
        
    def getNumber(self):
        """
        get the numberLiteral value
        """
        return self.value
        
    def __hash__(self):
        hashNum = 5
        return  17 * hashNum + hash((self.value ** self.value).value)
        
    def apply(self, function, value = None):
        """
        resolving a numberLiteral function
        """
        if function == self.ABS:
            return numberLiteral(abs(self.value))
        elif function == self.ROUND:
            return numberLiteral(round(self.value, 0))
        elif function == self.CEIL:
            return numberLiteral(self.value.ceil())
        elif function == self.FLOOR:
            return numberLiteral(self.value.floor())
        elif function == self.SQRT:
            return numberLiteral(coordinate(sqrt(self.value.value)))
        elif function == self.SIN:
            return numberLiteral(coordinate(sin(self.value.value)))
        elif function == self.COS:
            return numberLiteral(coordinate(cos(self.value.value)))
        elif function == self.TAN:
            return numberLiteral(coordinate(tan(self.value.value)))
        elif function == self.ASIN:
            return numberLiteral(coordinate(asin(self.value.value)))
        elif function == self.ACOS:
            return numberLiteral(coordinate(acos(self.value.value)))
        elif function == self.ATAN:
            return numberLiteral(coordinate(atan(self.value.value)))
        elif function == self.ATAN2:
            return numberLiteral(coordinate(atan2(self.value.value, value.value.value)))
        elif function == self.DEG:
            return numberLiteral(self.value*180/pi)
        elif function == self.RAD:
            return numberLiteral(self.value*pi/180)
        elif function == self.RANDOM:
            if value is None:
                return numberLiteral(randint(0, 1000000000) * self.value / 1000000000)
            return numberLiteral(self.value + randint(0, 1000000000) * (value.value - self.value) / 1000000000)
        elif function == 'last':
            return numberLiteral(self.value)
        elif function == 'first':
            return numberLiteral(self.value)
        raise ValueError('function name '+function+' not recognized')
        
    
    def compareValue(self, other):
        """
        comparing two numberLiterals
        """
        if isinstance(other,numberLiteral):
            c = self.value - other.value
        elif (other.type == 'description-numericExpression' or 
              other.type == 'description-literal' or 
              other.type == 'description-stringExpression'):
            return -other.compareValue(self)
        else:
            c = hash(self.__class__) - hash(other.__class__)
        if c > 0:
            return 1
        elif c < 0:
            return -1
        return 0
        
    def evaluate(self, assignment, sort):
        """
        evaluating the numberLiteral: the result would be itself
        """
        return self
        
    @staticmethod
    def functionExists(name, assist = None):
        """
        check if there exist a function with a specific name for number literal
        """
        for i in range(len(numberLiteral.FUNCTIONS)):
            if numberLiteral.FUNCTIONS[i]  == name and numberLiteral.ASSISTS[i] is None:
                if assist is None:
                    if numberLiteral.ASSISTS[i] is None:
                        return True
                elif numberLiteral.ASSISTS[i] == assist:
                    return True
        return False
        
    @staticmethod
    def parse(desc, reader):
        """
        parse a number literal
        """
        if reader.token() != parseReader.NUMBER and reader.token() != parseReader.MINUS:
            raise ParsingError(reader, "expected a (positive or negative) number")
        if reader.token() == parseReader.MINUS and reader.previewToken() != parseReader.NUMBER:
            raise ParsingError(reader, "expected a number following the '-' sign")
        if reader.token() == parseReader.NUMBER:
            result = numberLiteral(coordinate(reader.value()))
        else:
            reader.newToken()
            result = numberLiteral(coordinate(reader.value())*(-1))
        reader.newToken()
        return result
        