# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import range
from builtins import str

from .tree import tree
from .symbols import symbols
from .numberLiteral import numberLiteral
from ...io.parseReader import parseReader
from ...struct.coordinate import coordinate, cn1, pi

class numericExpression(tree):
    """
    numeric expression, consisting of number literals, parameters, functions, 
    references or other numberic expressions.

    operations supported: plus(+), minus(-), multiply(*), divide(/), modulo(%), power(^)

    ** minus(-) must have space before and after the symbol to prevent it from being mistaken 
    as a part of a parameter: e.g. "a - 1" is read as a numeric expression, "a-1" is read as one
    parameter.

    numeric expression can only exists in rules' lhs and rhs
    """
    type = 'description-numericExpression'
    def __init__(self, left = None, operation = None, right = None):
        self.type = 'description-numericExpression'
        if left is None and operation is None and right is None:
            self.expr = []
        else:
            self.expr = [left, operation, right]
        
    def __eq__(self, other):
        if other.type != 'description-numericExpression':
            return False
        return self.expr == other.expr
    
    def __lt__(self, other):
        return self.compareValue(other) == -1
        
    def __gt__(self, other):
        return self.compareValue(other) == 1
        
    def __hash__(self):
        hashnum = 7
        return 19 * hashnum + hash(tuple(self.expr))
        
    def __str__(self):
        s = ''
        for i in self.expr:
            if isinstance(i, str):
                s += i
            elif i.type == 'description-numericExpression':
                s += '(' + str(i)+ ')'
            elif i.type == 'description-numberLiteral' or i.type == 'description-parameter':
                s += str(i)
            elif i.type == 'description-literal':
                if (i.name == symbols.PLUS or 
                    i.name == symbols.TIMES or
                    i.name == symbols.POWER or
                    i.name == symbols.DIVIDED or
                    i.name == symbols.MODULO or
                    (i.name == parseReader.MINUS and self.expr.index(i)>0)):
                    s += symbols.SPACE + i.name + symbols.SPACE
                else:
                    s += str(i)
            elif i.type == 'description-function':
                s += str(i)
            elif i.type == 'description-reference':
                s += str(i)
            else:
                s += str(i)
                #s += symbols.PAREN_OPEN + str(i) + symbols.PAREN_CLOSE
        return s
        
    def printVal(self):
        s = ''
        for i in self.expr:
            if isinstance(i, str):
                s += i
            elif i.type == 'description-numericExpression':
                s += i.printVal()
            elif i.type == 'description-numberLiteral' or i.type == 'description-parameter':
                s += i.printVal()
            elif i.type == 'description-literal':
                if (i.name == symbols.PLUS or 
                    i.name == symbols.TIMES or
                    i.name == symbols.POWER or
                    i.name == symbols.DIVIDED or
                    i.name == symbols.MODULO or
                    (i.name == parseReader.MINUS and self.expr.index(i)>0)):
                    s += symbols.SPACE + i.name + symbols.SPACE
                else:
                    s += i.printVal()
            elif i.type == 'description-function':
                s += i.printVal()
            elif i.type == 'description-reference':
                s += i.printVal()
            else:
                s += i.printVal()
                #s += symbols.PAREN_OPEN + i.printVal() + symbols.PAREN_CLOSE
        return 'numExpr: ('+s+')'
        
    def compareValue(self, other):
        if other.type == 'description-numericExpression':
            for i in range(len(self.expr)):
                if i >= len(other.expr):
                    break
                if self.expr[i] > other.expr[i]:
                    return 1
                if self.expr[i] < other.expr[i]:
                    return -1
            c = len(self.expr) - len(other.expr)
            return c
        elif other.type == 'description-stringExpression':
            return -other.compareValue(self)
        elif other.type == 'description-tuple' or other.type == 'description-stringLiteral':
            c = hash(self.__class__) - hash(other.__class__)
            if c > 0:
                return 1
            if c < 0:
                return -1
            return 0
        c = self.expr[0].compareValue(other)
        if c == 0:
            c = 1
        return c
        
    def evaluate(self, assignment, sort):
        """
        evaluate the value of the numeric expression (and all numeric expression within it
        by proxy), the result should be a numberLiteral.
        """
        if self.expr[0].type == 'description-literal' and self.expr[0].name != 'pi':
            result = self.expr[1].evaluate(assignment, sort).value * cn1
            index = 1
        else:
            result = self.expr[0].evaluate(assignment, sort)
            while result.type == 'description-tuple' and len(result.tuple) == 1:
                result = result.tuple[0]
            if result.type == 'description-stringLiteral':
                result = coordinate(result.s)
            elif result.type == 'description-numberLiteral':
                result = result.value
            elif result.type == 'description-literal' and result.name == 'pi':
                result = coordinate(pi)
            else:
                raise ValueError('cannot evaluate numberic expression with a component of type '+result.type)
            index = 0
        index += 1
        while index < len(self.expr)-2:
            op = self.expr[index].getChar()
            
            value = self.expr[index+1].evaluate(assignment, sort)
            while value.type == 'description-tuple' and len(value.tuple) == 1:
                value = value.tuple[0]
            if value.type == 'description-stringLiteral':
                value = coordinate(value.s)
            elif value.type == 'description-numberLiteral':
                value = value.value
            elif value.type == 'description-literal' and value.name == 'pi':
                value = coordinate(pi)
            else:
                raise ValueError('cannot evaluate numberic expression with a component of type '+value.type)
            result, index = self.evaluteRecursive(assignment, sort, result, op, value, index+2)
        if index == len(self.expr):
            #print(result)
            return numberLiteral(coordinate(result))
        op = self.expr[index].getChar()

        value = self.expr[index+1].evaluate(assignment, sort)
        while value.type == 'description-tuple' and len(value.tuple) == 1:
            value = value.tuple[0]
        if value.type == 'description-stringLiteral':
            value = coordinate(value.s)
        elif value.type == 'description-numberLiteral':
            value = value.value
        elif value.type == 'description-literal' and value.name == 'pi':
            value = coordinate(pi)
        else:
            raise ValueError('cannot evaluate numberic expression with a component of type '+value.type)
        
        result = self.calculate(result, op, value)
        #print(result)
        return numberLiteral(coordinate(result))

    def evaluteRecursive(self, assignment, sort, prev_v, prev_op, value, index):
        """
        recursively going through the numberic expressions within a numeric expression and evaluate
        their results.
        """
        while index < len(self.expr):
            op = self.expr[index].getChar()
            
            next_v = self.expr[index+1].evaluate(assignment, sort)
            while next_v.type == 'description-tuple' and len(next_v.tuple) == 1:
                next_v = next_v.tuple[0]
            if next_v.type == 'description-stringLiteral':
                next_v = coordinate(next_v.s)
            elif next_v.type == 'description-numberLiteral':
                next_v = next_v.value
            elif next_v.type == 'description-literal' and next_v.name == 'pi':
                next_v = coordinate(pi)
            elif next_v.type == 'description-tuple':
                raise ValueError('cannot evaluate numberic expression with a component of type '+next_v.type)
            
            if op == symbols.PLUS or op == parseReader.MINUS:
                return self.calculate(prev_v, prev_op, value), index
            if op == symbols.TIMES or op == symbols.DIVIDED or op == symbols.MODULO:
                if prev_op == symbols.PLUS or prev_op == parseReader.MINUS:
                    value, index = self.evaluteRecursive(assignment, sort, value, op, next_v, index + 2)
                else:
                    return self.calculate(prev_v, prev_op, value), index
            elif op == symbols.POWER:
                value, index = self.evaluteRecursive(assignment, sort, value, op,next_v, index + 2)
        return self.calculate(prev_v, prev_op, value), index
        
    @staticmethod
    def calculate(v1, op, v2):
        """
        perform the singular calculation
        """
        if op == symbols.PLUS:
            return v1 + v2
        if op == parseReader.MINUS:
            return v1 - v2
        if op == symbols.TIMES:
            return v1 * v2
        if op == symbols.DIVIDED:
            return v1 / v2
        if op == symbols.MODULO:
            return v1 % v2
        if op == symbols.POWER:
            return v1 ** v2
        raise ValueError('operation '+str(op)+' not reconized')
    
        
    @staticmethod
    def parse(desc, reader):
        """
        parse a numeric expression
        """
        #print('\n>>>>>>>>', reader.rText())
        parens = False
        if reader.token() == symbols.PAREN_OPEN:
            parens = True
            reader.newToken()
        
        result = numericExpression()
        result.parseNext(desc, reader)
        #print('   ', reader.rText(), parens)
        if parens and reader.token() == symbols.PAREN_CLOSE:
            reader.newToken()             
        if len(result.expr) == 1:
            return result.expr[0]
        return result
        
    def parseNext(self, desc, reader):
        """
        continuing the parsing of a numeric expression
        """
        while True:
            if reader.token() == '#':
                self.expr.append(desc.reference.parse(desc, reader))
            elif reader.token() == symbols.PAREN_OPEN:
                a = desc.parseEntity(desc, reader, True)
                self.expr.append(a)
            elif reader.token() == parseReader.NUMBER or (reader.token() == parseReader.MINUS and reader.previewToken() == parseReader.NUMBER):
                self.expr.append(desc.numberLiteral.parse(desc, reader))
            elif reader.token() == parseReader.IDENTIFIER:
                name = reader.value()
                if desc.literal.isLiteral(name):
                    self.expr.append(desc.literal.getLiteral(reader.value()))
                    reader.newToken()
                elif desc.function.isFunctionName(name) and not reader.previewToken() == '.':
                    self.expr.append(desc.function.parse(desc, reader))
                elif desc.sort.checkSort(name) or reader.value() == 'rhs':
                    self.expr.append(desc.reference.parse(desc, reader))
                else: 
                    self.expr.append(desc.parameter.parse(desc, reader, False))
                    #reader.newToken()
            elif reader.token() != parseReader.MINUS:
                break
            if reader.token() in set([symbols.PLUS, parseReader.MINUS, symbols.TIMES,
                                      symbols.DIVIDED, symbols.MODULO, symbols.POWER]):
                self.expr.append(desc.literal.getLiteral(reader.token()))
                reader.newToken()
            else:
                break
        
        
    @staticmethod
    def parseContinue(desc, entity, reader, bracket = False):
        """
        continuing the parsing of a numeric expression after reading the first component of the expression
        """
        #print('\n------>', reader.text[reader.index():])
        if entity.type == 'description-parameter' and entity.hasCondition():
            result = numericExpression()
            result.expr.append(entity.condition.operand)
            entity.condition.operand = result
            if reader.token() in set([symbols.PLUS, parseReader.MINUS, symbols.TIMES,
                                      symbols.DIVIDED, symbols.MODULO, symbols.POWER]):
                result.expr.append(desc.literal.getLiteral(reader.token()))
                reader.newToken()
                result.parseNext(desc, reader)
            return entity
            #return entity
            #raise ValueError
        result = numericExpression()
        result.expr.append(entity)
        if reader.token() in set([symbols.PLUS, parseReader.MINUS, symbols.TIMES,
                                  symbols.DIVIDED, symbols.MODULO, symbols.POWER]):
            result.expr.append(desc.literal.getLiteral(reader.token()))
            reader.newToken()
            result.parseNext(desc, reader)
        return result
        
