# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import str

from .tree import tree, DescriptionEvaluatingError
from .symbols import symbols
from .numberLiteral import numberLiteral
from .stringLiteral import stringLiteral
from .tuple import tupleDesc
from ...io.parseReader import parseReader, ParsingError
from ...struct.coordinate import coordinate
from ...struct.vector import vector2D, vector3D

VZ = vector3D(0, 0, 0)

class reference(tree):
    """
    reference to a value of another object in another sort. The reference can have 
    subcondition to determine which object in the sort is to be used to get the required
    value.

    reference syntax: 

    * `geometry_sort.property`: getting a property of a geometry object in a sort. e.g. lineSegmentSort.length
    * `description_sort.parameter`: getting a parameter value of another description with a different sort. e.g. descSort.a1
    * `geometry_sort.property:attribute_sort.property=value`
    * `geometry_sort.property:attribute_sort.property<value`
    * `geometry_sort.property:attribute_sort.property>value`
    * ...
    * `rhs.objSort.property`

    reference can only exists in rules' lhs or rhs
    """
    type = 'description-reference'
    def __init__(self, sort, rhs = False):
        self.sort = sort
        self.method = None
        self.subSort = None
        self.subMethod = None
        self.subCond = None
        self.value = None
        self.isTag = False
        self.rhs = rhs
        
        
    def __eq__(self, other):
        if other.type != 'description-reference':
            return False
        return self.sort == other.sort and self.method == other.method
    
    def __lt__(self, other):
        return self.compareValue(other) == -1
        
    def __gt__(self, other):
        return self.compareValue(other) == 1

    def compareValue(self, other):
        if other.type == 'description-reference':
            c = self.sort.compareValue(other.sort)
            if c > 0:
                return 1
            if c < 0:
                return -1
            if self.method > other.method:
                return 1
            if self.method < other.method:
                return -1
            return 0
        elif (other.type == 'description-stringExpression' or 
                    other.type == 'description-numericExpression' or 
                    other.type == 'description-tupleExpression'):
            return -other.compareValue(self)
        c = hash(self.__class__) - hash(other.__class__)
        if c > 0:
            return 1
        if c < 0:
            return -1
        return 0
    
    def __str__(self):
        if self.isTag:
            return '#'+str(self.sort)+'.'+self.method
        s = ''
        if self.rhs:
            s += 'rhs.'
        s += str(self.sort) + symbols.CONCAT + self.method
        if self.subSort is not None:
            s += ':' + str(self.subSort) + symbols.CONCAT + self.subMethod + '=' 
            if type(self.subCond) == str:# or type(self.subCond) == unicode:
                s += '"'+self.subCond+'"'
            else:
                s += str(self.subCond)
        return s

    def printVal(self):
        s = 'ref:'
        if self.rhs:
            s += 'rhs.'
        s += str(self.sort) + symbols.CONCAT + self.method
        if self.subSort is not None:
            s += ':' + str(self.subSort) + symbols.CONCAT + self.subMethod + '=' 
            if type(self.subCond) == str:
                s += '"'+self.subCond+'"'
            elif isinstance(self.subCond,int) or isinstance(self.subCond,float):
                s += str(self.subCond)
            else:
                s += self.subCond.printVal()
        return s
        
    def visitEnter(self, element, assignment):
        """
        visit enter for evaluating the reference value
        """
        subMethod = 'getValue' if self.subMethod == 'value' else self.subMethod
        method = 'getValue' if self.method == 'value' else self.method

        if 'ind' in element.type:
            if element.sort.name == self.sort.name or element.sort.baseSort().name == self.sort.name:
                if self.subSort is not None and not isinstance(self.subSort,str):
                    if isinstance(self.subCond, int) or isinstance(self.subCond, float) or isinstance(self.subCond, str):
                        cmpVal = self.subCond
                    else:
                        cmpVal = self.subCond.evaluate(assignment, self.sort).value
                    if element.hasAttribute():
                        for i in element.attribute:
                            if reference.checkS(i, self.subSort.name, subMethod, cmpVal, assignment):
                                if element.type == 'ind-description' and method == 'getValue':
                                    self.value = element.evaluateInd(assignment).getValue()[1:-1]
                                    if self.value[0] == '"':
                                        self.value = self.value[1:-1]
                                else:
                                    met = getattr(element, method)
                                    if hasattr(met, '__call__'):
                                        try:
                                            self.value = met(assignment)
                                        except TypeError:
                                            self.value = met()
                                    else:
                                        self.value = met
                                return True
                    else:
                        return False
                else:
                    if self.subSort is not None and self.subSort == 'tag':
                        if element.key is not None and str(element.key.name)==str(self.subCond):
                            pass
                        else:
                            return False
                    if element.type == 'ind-description' and method == 'getValue':
                        self.value = element.evaluateInd(assignment).getValue()[1:-1]
                        if self.value[0] == '"':
                            self.value = self.value[1:-1]
                    else:
                        met = getattr(element, method)
                        if hasattr(met, '__call__'):
                            try:
                                self.value = met(assignment)
                            except TypeError:
                                self.value = met()
                        else:
                            self.value = met
                    return True
            return False
        elif 'form' in element.type:
            return True
            # if (element.type == 'form-metaForm' 
            # or element.sort.name == self.sort.name 
            # or element.sort.baseSort().name == self.sort.name 
            # or element.weight is not None):
            #     return True
            # return False
        
    @staticmethod
    def checkS(ind, sortName, sortMethod, sortValue = None, assignment = None):
        if 'form' in ind.type:
            if ind.type == 'form-metaForm' or ind.sort.baseSort().name == sortName:
                for i in ind:
                    check = reference.checkS(i, sortName, sortMethod, sortValue, assignment)
                    if check:
                        return check
            return False
        if ind.sort.baseSort().name == sortName:
            if ind.type == 'ind-description' and sortMethod == 'getValue':
                val = ind.evaluateInd(assignment).getValue()[1:-1]
                if val[0] == '"':
                    val = val[1:-1]
                if val == sortValue:
                    return True
                else:
                    return False
            met = getattr(ind, sortMethod)
            val = met()
            if sortValue is not None:
                if val == sortValue:
                    return True
                else:
                    return False
            else:
                return [val]
        if ind.hasAttribute():
            for i in ind.attribute:
                check = reference.checkS(i, sortName, sortMethod, sortValue)
                if check:
                    return check
        return False
        
    def visitLeave(self, element):
        return
        
    def evaluate(self, assignment, sort):
        """
        evaluate the reference value
        """
        method = 'getValue' if self.method == 'value' else self.method
        if self.isTag: 
            if self.sort in assignment.keys:
                if method == 'direction' and 'ind-line' in assignment.keys[self.sort].ind.type:
                    resultDir = assignment.keys[self.sort].ind.transformDirection(assignment.transform)
                    extraInfo = { 'preTransform': assignment.keys[self.sort].ind.direction }
                    return self.constructValue(resultDir, extraInfo)
                ind = assignment.keys[self.sort].ind.transform(assignment.transform)
                met = getattr(ind, method)
                if hasattr(met, '__call__'):
                    try:
                        self.value = met(assignment)
                    except TypeError:
                        self.value = met()
                else:
                    self.value = met
                return self.constructValue(self.value)
            else:
                raise DescriptionEvaluatingError('Tag name '+self.sort+' does not exist in the rule')
        else:
            if self.sort.baseSort().characteristic.type == 'ind-description':
                retrv = assignment.retrieve(self.sort, method)
                if retrv is None:
                    if method == 'getValue':
                        pass
                    else:
                        raise DescriptionEvaluatingError('parameter name "'+self.method+'" of sort "'+self.sort.name+'" does not exists in the param set: '+str(assignment.parameters))
                else:
                    return retrv
            elif self.sort.baseSort().characteristic.type == 'ind-color':
                retrv = assignment.retrieve(self.sort, method)
                if retrv is not None:
                    return retrv
            if assignment.lhs is None and assignment.rhs is None:
                raise ValueError('lhs and rhs are not assigned to the paramSet yet')
            subMethod = 'getValue' if self.subMethod == 'value' else self.subMethod
            if self.subSort == 'tag' and subMethod == 'getValue':
                if self.subCond in assignment.keys:
                    ind = assignment.keys[self.subCond].ind.transform(assignment.transform)
                    met = getattr(ind, method)
                    if hasattr(met, '__call__'):
                        try:
                            self.value = met(assignment)
                        except TypeError:
                            self.value = met()
                    else:
                        self.value = met
            elif self.rhs == True:
                assignment.rhs.accept(self, assignment)
            else:
                assignment.lhs.accept(self, assignment)
        if self.value is not None:
            if self.sort.baseSort().characteristic.type == 'ind-description':
                try:
                    self.value = coordinate(self.value)
                except:
                    pass
            return self.constructValue(self.value)
        return numberLiteral(0)
        
    @staticmethod
    def constructValue(value, extraInfo = None):
        """
        construct the result component depending on its value
        """
        if isinstance(value, coordinate):
            return numberLiteral(value)
        elif isinstance(value, int) or isinstance(value, float):
            return numberLiteral(coordinate(value))
        elif isinstance(value, str):
            return numberLiteral(coordinate(value))
        elif isinstance(value, list) or isinstance(value, tuple):
            tup = tupleDesc(symbols.PAREN_OPEN, symbols.COMMA)
            for i in value:
                tup.tuple.append(reference.constructValue(i))
            return tup
        elif isinstance(value, vector2D):
            tup = tupleDesc(symbols.PAREN_OPEN, symbols.COMMA)
            tup.tuple.append(numberLiteral(value.x))
            tup.tuple.append(numberLiteral(value.y))
            tup.extraInfo = extraInfo
            return tup
        elif isinstance(value, vector3D):
            tup = tupleDesc(symbols.PAREN_OPEN, symbols.COMMA)
            tup.tuple.append(numberLiteral(value.x))
            tup.tuple.append(numberLiteral(value.y))
            tup.tuple.append(numberLiteral(value.z))
            tup.extraInfo = extraInfo
            return tup
        else:
            raise ValueError('Value '+ str(value) +': type not recognized.')
            
    @staticmethod
    def parse(desc, reader):
        """
        parsing a reference
        """
        if reader.token() == '#':
            name = reader.newValue()
            result = reference(name)
            reader.newToken()
            result.method = reader.newValue()
            result.isTag = True
            reader.newToken()
            return result
        rhs = False
        if reader.value() == 'rhs':
            rhs = True
            reader.newToken()
            reader.newToken()
        name = reader.value()
        if reader.token() != parseReader.IDENTIFIER or reader.newToken() != symbols.CONCAT:
            raise ParsingError(reader, "expected an identifier followed by a '.'")
        if desc.sort.checkSort(name) == False:
            raise ParsingError(reader, "sort '" + name + "' not recognized")
        result = reference(desc.sort.getSort(name), rhs)
        reader.newToken()
        result.method = reader.value()

        if reader.newToken() == ':':
            name = reader.newValue()
            if reader.token() != parseReader.IDENTIFIER or reader.newToken() != symbols.CONCAT:
                raise ParsingError(reader, "expected an identifier followed by a '.'")
            if name == 'tag':
                result.subSort =  name
            else:
                if desc.sort.checkSort(name) == False:
                    raise ParsingError(reader, "sort '" + name + "' not recognized")
                result.subSort = desc.sort.getSort(name)
            reader.newToken()
            result.subMethod = reader.value()
            
            if reader.newToken() != '=':
                raise ParsingError(reader, "expected a '='")

            if reader.newToken() == parseReader.STRING:
                result.subCond = str(reader.value()[1:-1])
            elif reader.token() == parseReader.NUMBER:
                try:
                    result.subCond = int(reader.value())
                except ValueError:
                    result.subCond = float(reader.value())
            elif (reader.token() == parseReader.MINUS and reader.previewToken() == parseReader.NUMBER):
                reader.newToken()
                try:
                    result.subCond = -int(reader.value())
                except ValueError:
                    result.subCond = -float(reader.value())
            else:
                result.subCond = desc.numericExpression.parse(desc, reader)
                return result
            reader.newToken()
        return result
