# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import range
from builtins import str


from .tree import tree
from .symbols import symbols
from ...io.parseReader import parseReader, ParsingError
from ...struct.vector import vector2D, vector3D
from ...struct.coordinate import coordinate, c0, c1, acos, pi

from .function import function
from .numberLiteral import numberLiteral
from .literal import literal
from random import randint

class tupleDesc(tree):
    """
    tuple: a list of description components
    """

    RANDOM = "random"
    ROUND = "round"
    MAG = "mag"
    PROJ = "proj"
    DISTANCE = "distance"
    VECTORADD = "vectoradd"
    VECTORSUBTRACT = "vectorsubtract"
    VECTORSCALE = "vectorscale"
    ANGLE = "angle"
    DOTPRODUCT = "dotproduct"
    CROSSPRODUCT = "crossproduct"
    LENGTH = "length"
    FIRST = "first"
    MINIMUM = "min"
    MAXIMUM = "max"
    LAST = "last"
    ITEM = "item"
    UNIQUE = "unique"
    SEGMENTS = "segments"
    PAIRWISE = "pairwise"
    LOOPS = "loops"
    ADJACENCIES = "adjacencies"
    
    EMPTY_TUPLE = []
    FUNCTIONS = [DISTANCE, ANGLE, MAG, PROJ,
                 VECTORADD, VECTORSUBTRACT, VECTORSCALE, LENGTH, CROSSPRODUCT, DOTPRODUCT,
                 FIRST, LAST, ITEM, MINIMUM, MAXIMUM, UNIQUE, SEGMENTS, PAIRWISE, RANDOM, ROUND,
                 LOOPS, ADJACENCIES]
    ASSISTS = [None for i in range(len(FUNCTIONS))]
    
    for fname in FUNCTIONS:
        function.reserve(fname)
    
    def __init__(self, bracket, separator, entity = None):
        self.type = 'description-tuple'
        self.tuple = []
        self.bracket = bracket
        if bracket == symbols.PAREN_OPEN:
            self.endBracket = symbols.PAREN_CLOSE
        elif bracket == symbols.ANGLE_OPEN:
            self.endBracket = symbols.ANGLE_CLOSE
        elif bracket == symbols.SQUARE_OPEN:
            self.endBracket = symbols.SQUARE_CLOSE
        elif bracket == symbols.SET_OPEN:
            self.endBracket = symbols.SET_CLOSE
        elif bracket == symbols.NONE:
            self.endBracket = symbols.NONE
        else:
            raise TypeError('bracket not recognized')
        if (separator != symbols.COMMA and separator != symbols.SEMICOLON and 
            separator != symbols.NONE and separator != symbols.UNKNOWN):
            raise TypeError('comma not recognized')
        self.separator = separator
        if entity is not None:
            self.tuple.append(entity)
        self.extraInfo = None

    @staticmethod
    def newTuple(first, second):
        tpl = tupleDesc(symbols.PAREN_OPEN, symbols.COMMA)
        tpl.tuple.append(first)
        tpl.tuple.append(second)
        return tpl
        
    def checkPartOf(self, other):
        """
        check if the tuple is part of another tuple
        """
        if self.type != other.type:
            return True

        if len(self.tuple) > len(other.tuple):
            return False

        for i in range(len(self.tuple)):
            if not self.tuple[i].checkPartOf(other.tuple[i]):
                return False
        return True
        
    def isEmpty(self):
        """
        check if the tuple has no component in it
        """
        return self.tuple is None or len(self.tuple) == 0
        
    def addEntity(self, entity, index = -1):
        """
        add a component into the tuple at a certain index
        """
        if index > len(self.tuple):
            raise KeyError('index exceeds length of tuple')
        if len(self.tuple) > 0 and self.bracket is not None and (self.separator is None or 
                                                                 self.separator is symbols.UNKNOWN):
            self.separator = symbols.COMMA
        if index == -1:
            self.tuple.append(entity)
        else: 
            self.tuple.insert(index, entity)
            
    def __hash__(self):
        hashNum = 5
        hashNum = 97 * hashNum
        for i in self.tuple:
            hashNum += hash(i)
        return hashNum

        
    def getEntity(self, index):
        """
        get the tuple's component at index position
        """
        if self.tuple is None or index >= len(self.tuple):
            return None
        return self.tuple[index]
        
    def shareDelimiters(self, other):
        return self.separator == other.separator and self.bracket == other.bracket
        
    def __len__(self):
        return len(self.tuple)
        
    def __eq__(self, other):
        if other.type != 'description-tuple':
            return False
        if len(self.tuple) != len(other.tuple):
            return False
        for i in range(len(self.tuple)):
            if self.tuple[i] != other.tuple[i]:
                return False
        return True
        
    def __lt__(self, other):
        return self.compareValue(other) == -1
        
    def __gt__(self, other):
        return self.compareValue(other) == 1

    def compareValue(self, other):
        """
        compare two tuples
        """
        if other.type == 'description-tuple':
            for i in range(len(self.tuple)):
                if i >= len(other.tuple):
                    break
                if self.tuple[i] > other.tuple[i]:
                    return 1
                elif self.tuple[i] < other.tuple[i]:
                    return -1
            c = len(self.tuple) - len(other.tuple)
            if c > 0:
                return 1
            elif c < 0:
                return -1
            return 0
        elif other.type == 'description-literal' or other.type == 'description-tupleExpression':
            return -other.compareValue(self)
        else:
            c = hash(self.__class__) - hash(other.__class__)
        if c > 0:
            return 1
        elif c < 0:
            return -1
        return 0

    def getNumber(self, assignment, sort):
        num = 0
        for i in self.tuple:
            num += i.getNumber(assignment, sort)
        return num
            
    def __str__(self):
        s = ''
        if self.bracket != symbols.NONE and self.bracket != symbols.UNKNOWN:
            s += self.bracket
        if not self.isEmpty():
            addComma = False
            for entity in self.tuple:
                if (self.separator == symbols.COMMA or self.separator == symbols.SEMICOLON) and addComma:
                    s += self.separator + symbols.SPACE
                elif not addComma:
                    addComma = True
                s += str(entity)
        if self.bracket != symbols.NONE and self.bracket != symbols.UNKNOWN:
            s += self.endBracket
        return s
        
    def printVal(self):
        s = ''
        if self.bracket != symbols.NONE and self.bracket != symbols.UNKNOWN:
            s += self.bracket
        if not self.isEmpty():
            addComma = False
            for entity in self.tuple:
                if (self.separator == symbols.COMMA or self.separator == symbols.SEMICOLON) and addComma:
                    s += self.separator + symbols.SPACE
                elif not addComma:
                    addComma = True
                s += entity.printVal()
        if self.bracket != symbols.NONE and self.bracket != symbols.UNKNOWN:
            s += self.endBracket
        else:
            s =  '(' + s + ')'
        return 'tup'+s
        
    def duplicate(self):
        result = tupleDesc(self.bracket, self.separator)
        for i in self.tuple:
            result.addEntity(i)
        return result
    
    def __iter__(self):
        return iter(self.tuple)
        
    def map(self, assignment, sort, onto):
        """
        mappping a tuple with another tuple
        """
        if onto.type != 'description-tuple' or len(self) != len(onto):
            return False
        for i in range(len(self.tuple)):
            if not self.tuple[i].map(assignment, sort, onto.tuple[i]):
                return False
        return True
        
    def evaluate(self, assignment, sort):
        """
        evaluate the value of the tuple: resolving the values of all its components
        """
        if self.isEmpty():
            return self
        result = tupleDesc(self.bracket, self.separator)
        for entity in self.tuple:
            result.addEntity(entity.evaluate(assignment, sort))
        return result
        
    def sum(self, other):
        """
        summing two tuples, only works with tuple of numberLiterals
        """
        if len(self) != len(other):
            raise KeyError('tuples must have the same length')
        for i in range(len(self.tuple)):
            if self.tuple[i].type == 'description-numberLiteral':
                self.tuple[i] = numberLiteral(self.tuple[i].value + other.tuple[i].value)
    
    def functionExists(self, name, assist= None):
        """
        check if there exist a function with a specific name for tuple
        """
        for i in range(len(tupleDesc.FUNCTIONS)):
            if tupleDesc.FUNCTIONS[i] == name and tupleDesc.ASSISTS[i] == assist:
                return True
                
        return False    
        
    def apply(self, function, assist = None):
        """
        resolving a tuple function
        """
        if assist is None:
            if function == self.MAXIMUM:
                return max(self.tuple).duplicate()
            elif function == self.MINIMUM:
                return min(self.tuple).duplicate()
            elif function == self.LENGTH:
                return numberLiteral(coordinate(len(self.tuple)))
            elif function == self.FIRST:
                return self.tuple[0].duplicate()
            elif function == self.LAST:
                return self.tuple[-1].duplicate()
            elif function == self.ROUND:
                result = tupleDesc(self.bracket,self.separator)
                for i in self.tuple:
                    if i.type == 'description-numberLiteral':
                        result.tuple.append(numberLiteral(coordinate(round(i.value.value,0))))
                    else:
                        result.tuple.append(i)
                return result
            elif function == self.ITEM:
                return self.calcItem(self.tuple[0], self.tuple[1])
            elif function == self.RANDOM:
                if len(self) < 1 or len(self) > 3:
                    return numberLiteral(c0)
                elif len(self) == 3:
                    start = self.tuple[0].value.value
                    stop = self.tuple[1].value.value
                    step = self.tuple[2].value.value
                                        
                    if start > stop:
                        start, stop = stop, start
                    randint(0, int((stop - start) / step)) * step + start
                    result = numberLiteral(randint(0, int((stop - start) / step)) * step + start)
                elif len(self) == 2:
                    start = self.tuple[0].value.value
                    stop = self.tuple[1].value.value
                    if start > stop:
                        start, stop = stop, start
                    result = numberLiteral(start + randint(0, 1000000000) * (stop - start) / 1000000000)
                else:
                    result = numberLiteral(randint(0, 1000000000) * self.tuple[0].value.value / 1000000000 )
                return result
            elif function == self.VECTORADD:
                if len(self) != 2:
                    raise ValueError('VectorAdd function requires exactly 2 vectors, '+ str(len(self)) +' vectors were given')
                return self.calcVectorAdd(self.tuple[0], self.tuple[1])
            elif function == self.VECTORSUBTRACT:
                if len(self) != 2:
                    raise ValueError('VectorSubtract function requires exactly 2 vectors, '+ str(len(self)) +' vectors were given')
                return self.calcVectorSubtract(self.tuple[0], self.tuple[1])
            elif function == self.VECTORSCALE:
                if len(self) != 2:
                    raise ValueError('VectorScale function requires 1 vector and 1 number, argument given: '+ str(self))
                return self.calcVectorScale(self.tuple[0], self.tuple[1])
            elif function == self.CROSSPRODUCT:
                if len(self) != 2:
                    raise ValueError('VectorScale function requires 1 vector and 1 number, argument given: '+ str(self))
                return self.calcCrossProduct(self.tuple[0], self.tuple[1])
            elif function == self.DOTPRODUCT:
                if len(self) != 2:
                    raise ValueError('VectorScale function requires 1 vector and 1 number, argument given: '+ str(self))
                return self.calcDotProduct(self.tuple[0], self.tuple[1])
            elif function == self.DISTANCE:
                if len(self) != 2:
                    raise ValueError('Distance function requires exactly 2 vectors, '+ str(len(self)) +' vectors were given')
                return self.calcDistance(self.tuple[0], self.tuple[1])
                
            elif function == self.MAG:
                if len(self) == 0:
                    raise ValueError('Magnitude function requires 1 or 2 vectors, '+ str(len(self)) +' vectors were given')
                if len(self) == 1:
                    v = self.tuple[0]
                    s = 0
                    for i in v.tuple:
                        s += i.value * i.value
                    return numberLiteral(s.sqrt())
                elif self.tuple[0].type != 'description-tuple':
                    s = 0
                    for i in self.tuple:
                        s += i.value * i.value
                    return numberLiteral(s.sqrt())
                return self.calcMag(self.tuple[0], self.tuple[1])
                
            # calculate the projection of a point on a line, syntax: proj((lineDirection, lineRoot, pointCoord))
            elif function == self.PROJ:
                if len(self) == 3:
                    direction = self.tuple[0]
                    root = self.tuple[1]
                    point = self.tuple[2]
                    if direction.type != root.type != point.type != 'description-tuple':
                        raise ValueError('Projection function requires exactly 3 tuples as vectors, received inputs: '+str(direction) +', '+str(root)+', '+str(point))
                    if not len(direction) == len(root) == len(point):
                        raise ValueError('vectors '+str(direction)+' and '+str(root)+' and '+str(point)+' are not of the same dimension')
                    v = []
                    for i in range(len(point)):
                        v.append(point.tuple[i].value - root.tuple[i].value)
                    d1 = 0
                    d2 = 0
                    for i in range(len(v)):
                        d1 += direction.tuple[i].value * v[i]
                        d2 += direction.tuple[i].value * direction.tuple[i].value
                    d1 = d1 / d2
                    result = tupleDesc(direction.bracket, direction.separator)
                    for i in range(len(direction)):
                        result.tuple.append(numberLiteral(root.tuple[i].value + direction.tuple[i].value * d1))
                    return result
                else:
                    raise ValueError('projection function requires exactly 2 vectors, '+ str(len(self)) +' vectors were given')
                    
            elif function == self.ANGLE:
                if len(self) != 2:
                    raise ValueError('Angle function requires exactly 2 vectors, '+ str(len(self)) +' vectors were given')
                return self.calcAngle(self.tuple[0], self.tuple[1])
            elif function == self.UNIQUE:
                result = tupleDesc(self.bracket, self.separator)
                for entity in self.tuple:
                    if len(result.tuple) == 0 or entity not in result.tuple:
                        result.addEntity(entity)
                return result
            elif function == self.SEGMENTS:
                result = tupleDesc(self.bracket, self.separator)
                for i in range(1,len(self.tuple)):
                    result.addEntity(tupleDesc.newTuple(self.tuple[i-1], self.tuple[i]))
                return result
            elif function == self.PAIRWISE:
                result = tupleDesc(self.bracket, self.separator)
                i = 1
                while i < len(self.tuple):
                    result.addEntity(tupleDesc.newTuple(self.tuple[i-1], self.tuple[i]))
                    i += 2
                return result
            elif function == self.LOOPS:
                result = tupleDesc(self.bracket, self.separator)
                loops = []
                for i in range(len(self.tuple)):
                    for j in range(i+1, len(self.tuple)):
                        if self.tuple[i] == self.tuple[j]:
                            loop = tupleDesc(self.bracket, self.separator)
                            for entity in range(i,j):
                                line = tupleDesc.newTuple(self.tuple[entity], self.tuple[entity+1])
                                loop.addEntity(line)
                            loops.append(loop)
                for i in loops:
                    j = 0
                    while j < len(i.tuple):
                        if i.tuple[j].tuple[0] == i.tuple[j-1].tuple[1] and i.tuple[j].tuple[1] == i.tuple[j-1].tuple[0]:
                            i.tuple.pop(j)
                            i.tuple.pop(j-1)
                            if j > 0:
                                j -= 1
                        else:
                            j += 1
                #loops.sort(key = lambda s: len(s.tuple))
                for rounds in range(2):
                    for i in range(len(loops)):
                        if loops[i] is None:
                            continue
                        if len(loops[i].tuple) <= 2:
                            loops[i] = None
                            continue
                        #print('....',loops)
                        del1 = False
                        for j in range(i+1, len(loops)):
                            if loops[j] is None:
                                continue
                            if len(loops[j].tuple) <= 2:
                                loops[j] = None
                                continue
                            common = []
                            lj = [False] * len(loops[j])
                            for ent1 in loops[i]:
                                for index, ent2 in enumerate(loops[j]):
                                    if lj[index]:
                                        continue
                                    if (ent1.tuple[0] == ent2.tuple[0] and ent1.tuple[1] == ent2.tuple[1]):
                                        common.append(ent1)
                                        lj[index] = True
                                        break
                                    elif (ent1.tuple[0] == ent2.tuple[1] and ent1.tuple[1] == ent2.tuple[0]):
                                        common.append(ent1)
                                        lj[index] = True
                                        break
    
                            if len(common) == 0:
                                continue
                            elif len(common) == len(loops[j]) == len(loops[i]):
                                loops[j] = None
                                continue
                            elif len(common) < min(len(loops[i]),len(loops[j]))/2:
                                continue
                            if len(loops[i]) > len(loops[j]):
                                loopRev = loops[i]
                                loopAdd = loops[j].duplicate()
                            else:
                                loopRev = loops[j]
                                loopAdd = loops[i].duplicate()
                            for c in common:
                                for lr in loopRev.tuple:
                                    if (c.tuple[0] == lr.tuple[0] and c.tuple[1] == lr.tuple[1]) or (c.tuple[0] == lr.tuple[1] and c.tuple[1] == lr.tuple[0]):
                                        loopRev.tuple.remove(lr)
                                        break
                                for la in loopAdd.tuple:
                                    if (c.tuple[0] == la.tuple[0] and c.tuple[1] == la.tuple[1]) or (c.tuple[0] == la.tuple[1] and c.tuple[1] == la.tuple[0]):
                                        loopAdd.tuple.remove(la)
                                        break
                            for add in loopAdd:
                                reverse = 0
                                normal = 0
                                for ent in range(len(loopRev)):
                                    if loopRev.tuple[ent].tuple[0] == add.tuple[0]:
                                        reverse += 1
                                        index = ent
                                    elif loopRev.tuple[ent].tuple[1] == add.tuple[1]:
                                        reverse += 1
                                        index = ent + 1
                                        if index == len(loopRev):
                                            index = 0
                                    elif loopRev.tuple[ent].tuple[1] == add.tuple[0]:
                                        normal += 1
                                        index = ent + 1
                                        if index == len(loopRev):
                                            index = 0
                                    elif loopRev.tuple[ent].tuple[0] == add.tuple[1]:
                                        normal += 1
                                        index = ent
                                if normal > reverse:
                                    loopRev.addEntity(add, index)
                                elif normal < reverse:
                                    loopRev.addEntity(tupleDesc.newTuple(add.tuple[1],add.tuple[0]), index)
                                else:
                                    raise ValueError('segment not recognized if reverse or not')
                            if len(loopRev.tuple) <= 2:
                                loopRev = None
                        if del1:
                            loops[i] = None
                            continue
                for i in range(len(loops)):
                    if loops[i] is None:
                        continue
                    for j in range(i+1, len(loops)):
                        if loops[j] is None:
                            continue
                        if len(loops[i].tuple) != len(loops[j].tuple):
                            continue
                        """
                        if set(loops[i].tuple) == set(loops[j].tuple):
                            loops[j] = None
                        """
                loops = filter(None, loops)
                for i in loops:
                    loop = tupleDesc(self.bracket, self.separator)
                    for j in i.tuple:
                        loop.addEntity(j.tuple[0])
                    result.addEntity(loop)
                if len(self.tuple) == 0:
                    return literal.getLiteral('e')
                return result
        elif function == self.DISTANCE:
            return self.calcDistance(self, assist)
        elif function == self.MAG:
            return self.calcMag(self, assist)
        elif function == self.PROJ:
            pass
        elif function == self.ITEM:
            return self.calcItem(self, assist)
        elif function == self.VECTORADD:
            return self.calcVectorAdd(self, assist)
        elif function == self.VECTORSUBTRACT:
            return self.calcVectorSubtract(self, assist)
        elif function == self.VECTORSCALE:
            return self.calcVectorScale(self, assist)
        elif function == self.CROSSPRODUCT:
            return self.calcCrossProduct(self, assist)
        elif function == self.DOTPRODUCT:
            return self.calcDotProduct(self, assist)
        elif function == self.ANGLE:
            return self.calcAngle(self, assist)
        elif function == self.ADJACENCIES:
            result = tupleDesc(self.bracket, self.separator)
            for i in range(len(self.tuple)+1):
                mat = tupleDesc(self.bracket, self.separator)
                for j in range(len(self.tuple)+1):
                    if i == j:
                        mat.addEntity(numberLiteral(c1))
                    else:
                        mat.addEntity(numberLiteral(c0))
                result.addEntity(mat)
            for door in assist:
                try:
                    for i in range(len(self.tuple)):
                        for wall1 in range(len(self.tuple[i].tuple)):
                            nex = wall1+1
                            if nex == len(self.tuple[i].tuple):
                                nex = 0
                            if (self.tuple[i].tuple[wall1] == door.tuple[0] and  
                                (self.tuple[i].tuple[wall1-1] == door.tuple[1] or 
                                self.tuple[i].tuple[nex] == door.tuple[1])):
                                    outside = True
                                    for j in range(i+1,len(self.tuple)):
                                        for wall2 in range(len(self.tuple[j].tuple)):
                                            nex1 = wall2+1
                                            if nex1 == len(self.tuple[j].tuple):
                                                nex1 = 0
                                            if (self.tuple[j].tuple[wall2] == door.tuple[0] and  
                                                (self.tuple[j].tuple[wall2-1] == door.tuple[1] or 
                                                 self.tuple[j].tuple[nex1] == door.tuple[1])):
                                                     outside = False
                                                     result.tuple[j+1].tuple[i+1] = c1
                                                     result.tuple[i+1].tuple[j+1] = c1
                                                     break
                                    if outside:
                                        result.tuple[0].tuple[i+1] = c1
                                        result.tuple[i+1].tuple[0] = c1
                                    raise StopIteration
                except StopIteration:
                    pass
            return result
        raise NotImplementedError('function '+function+' not recognized or not implemented')

    def calcVectorAdd(self, v0, v1):
        if v0.type != 'description-tuple' or v1.type != 'description-tuple':
            raise ValueError('VectorAdd function requires exactly 2 tuples as vectors, received inputs: '+str(v0) +', '+str(v1))
        if len(v0) != len(v1):
            raise ValueError('vectors '+str(v0)+' and '+str(v1)+' are of different dimensions')
        result = tupleDesc(symbols.PAREN_OPEN,symbols.COMMA)
        for i in range(len(v0.tuple)):
            result.tuple.append(numberLiteral(v0.tuple[i].value + v1.tuple[i].value))
        return result

    def calcVectorSubtract(self, v0, v1):
        if v0.type != 'description-tuple' or v1.type != 'description-tuple':
            raise ValueError('VectorSubtract function requires exactly 2 tuples as vectors, received inputs: '+str(v0) +', '+str(v1))
        if len(v0) != len(v1):
            raise ValueError('vectors '+str(v0)+' and '+str(v1)+' are of different dimensions')
        result = tupleDesc(symbols.PAREN_OPEN,symbols.COMMA)
        for i in range(len(v0.tuple)):
            result.tuple.append(numberLiteral(v0.tuple[i].value - v1.tuple[i].value))
        return result

    def calcVectorScale(self, v0, v1):
        if v0.type != 'description-tuple' or v1.type != 'description-numberLiteral':
            raise ValueError('VectorScale function requires 1 tuple as vector and 1 number literal as scaling number, received inputs: '+str(v0) +', '+str(v1))
        result = tupleDesc(symbols.PAREN_OPEN,symbols.COMMA)
        for i in range(len(v0.tuple)):
            result.tuple.append(numberLiteral(v0.tuple[i].value * v1.value))
        return result
                
    def calcCrossProduct(self, v0, v1):
        if v0.type != 'description-tuple' or v1.type != 'description-tuple':
            raise ValueError('CrossProduct function requires exactly 2 tuples as vectors, received inputs: '+str(v0) +', '+str(v1))
        if len(v0) != len(v1):
            raise ValueError('vectors '+str(v0)+' and '+str(v1)+' must be of the same dimensions in either 2D or 3D')
        if len(v0) == 2:
            result = numberLiteral(v0.tuple[0].value * v1.tuple[1].value - v0.tuple[1].value * v1.tuple[0].value)
        elif len(v0) == 3:
            result = tupleDesc(symbols.PAREN_OPEN,symbols.COMMA)
            result.tuple.append(numberLiteral(v0.tuple[1].value * v1.tuple[2].value - v0.tuple[2].value * v1.tuple[1].value))
            result.tuple.append(numberLiteral(v0.tuple[2].value * v1.tuple[0].value - v0.tuple[0].value * v1.tuple[2].value))
            result.tuple.append(numberLiteral(v0.tuple[0].value * v1.tuple[1].value - v0.tuple[1].value * v1.tuple[0].value))
        return result
                
    def calcDotProduct(self, v0, v1):
        if v0.type != 'description-tuple' or v1.type != 'description-tuple':
            raise ValueError('VectorSubtract function requires exactly 2 tuples as vectors, received inputs: '+str(v0) +', '+str(v1))
        if len(v0.tuple) != len(v1.tuple) and len(v0.tuple) < 2 or len(v0.tuple) > 3:
            raise ValueError('vectors '+str(v0)+' and '+str(v1)+' must be of the same dimensions in either 2D or 3D')
        result = 0
        for i in range(len(v0.tuple)):
            result += v0.tuple[i].value * v1.tuple[i].value
        return numberLiteral(result)

    def calcAngle(self, v0, v1):
        if v0.type != 'description-tuple' or v1.type != 'description-tuple':
            raise ValueError('Angle function requires exactly 2 tuples as vectors, received inputs: '+str(v0) +', '+str(v1))
        if len(v0) != len(v1):
            raise ValueError('Vectors '+str(v0)+' and '+str(v1)+' are of different dimensions')
        if len(v0) == 2:
            vec0 = vector2D(v0.tuple[0].value, v0.tuple[1].value)
            vec1 = vector2D(v1.tuple[0].value, v1.tuple[1].value)

            dem = (vec0.dotproduct(vec0) * vec1.dotproduct(vec1)).sqrt()
            if dem == c0:
                raise ZeroDivisionError('One of the vectors are (0,0) --> division by zero')

            if v0.extraInfo is not None or v1.extraInfo is not None:
                pre0 = vec0
                if v0.extraInfo is not None and v0.extraInfo['preTransform'] is not None:
                    pre0 = v0.extraInfo['preTransform']
                pre1 = vec1
                if v1.extraInfo is not None and v1.extraInfo['preTransform'] is not None:
                    pre1 = v1.extraInfo['preTransform']
                preCrss = pre0.product(pre1)
                postCrss = vec0.product(vec1)
                if (preCrss < 0) != (postCrss < 0):
                    vec0, vec1 = vec1, vec0

            nom = vec0.dotproduct(vec1)
            crss = vec0.product(vec1)
            cosTheta = nom / dem
            angle = coordinate(acos(cosTheta.value))
            if crss < 0:
                angle = -angle
            return numberLiteral(angle)
        if len(v0) == 3:
            vec0 = vector3D(v0.tuple[0].value, v0.tuple[1].value, v0.tuple[2].value)
            vec1 = vector3D(v1.tuple[0].value, v1.tuple[1].value, v1.tuple[2].value)

            dem = (vec0.dotproduct(vec0) * vec1.dotproduct(vec1)).sqrt()
            if dem == c0:
                raise ZeroDivisionError('One of the vectors is (0,0,0) --> division by zero')

            if v0.extraInfo is not None or v1.extraInfo is not None:
                pre0 = vec0
                if v0.extraInfo is not None and v0.extraInfo['preTransform'] is not None:
                    pre0 = v0.extraInfo['preTransform']
                pre1 = vec1
                if v1.extraInfo is not None and v1.extraInfo['preTransform'] is not None:
                    pre1 = v1.extraInfo['preTransform']
                preCrss = pre0.product(pre1)
                postCrss = vec0.product(vec1)
                if (preCrss.sign() < 0) != (postCrss.sign() < 0):
                    vec0, vec1 = vec1, vec0

            # if (v0.extraInfo['reflection'] < 0) != (v1.extraInfo['reflection'] < 0):
            #     vec0, vec1 = vec1, vec0
            nom = vec0.dotproduct(vec1)
            crss = vec0.product(vec1)
            cosTheta = nom / dem
            angle = coordinate(acos(cosTheta.value))
            if crss.sign() < 0:
                angle = -angle
            return numberLiteral(angle)
        return numberLiteral(coordinate(len(self.tuple)))

    def calcDistance(self, v0, v1):
        if v0.type != 'description-tuple' or v1.type != 'description-tuple':
            raise ValueError('Distance function requires exactly 2 tuples as vectors, received inputs: '+str(v0) +', '+str(v1))
        if len(v0) != len(v1):
            raise ValueError('vectors '+str(v0)+' and '+str(v1)+' are of different dimensions')
        s = 0
        for i in range(len(v0.tuple)):
            s += (v0.tuple[i].value - v1.tuple[i].value) * (v0.tuple[i].value - v1.tuple[i].value)
        return numberLiteral(s.sqrt())
                
    
    def calcMag(self, v0, v1):
        if v0.type != 'description-tuple' or v1.type != 'description-tuple':
            raise ValueError('Distance function requires exactly 2 tuples as vectors, received inputs: '+str(v0) +', '+str(v1))
        if len(v0) != len(v1):
            raise ValueError('vectors '+str(v0)+' and '+str(v1)+' are of different dimensions')
        s = 0
        for i in range(len(v0.tuple)):
            s += (v0.tuple[i].value - v1.tuple[i].value) * (v0.tuple[i].value - v1.tuple[i].value)
        return numberLiteral(s.sqrt())

    def calcItem(self, tupleObj, indexVal):
        try:
            index = int(indexVal.value)
        except Exception:
            index = int(indexVal)
        if index >= len(tupleObj.tuple):
            return tupleObj.tuple[-1].duplicate()
        return tupleObj.tuple[index].duplicate()

    def contains(self, other):
        """
        check if the tuple contain another
        """
        if other.isEmpty():
            return True
        if self.isEmpty():
            return False
        for i in other.tuple:
            check = False
            for j in self.tuple:
                if i.type == j.type and j.contains(i):
                    check = True
                    break
            if not check:
                return False
            
        return True
        
    def similarFragment(self, other):
        if self.type != other.type:
            return False
        if self.bracket != other.bracket or self.separator != other.separator:
            return False
        if len(self.tuple) != len(other.tuple):
            return False
        for i in range(len(self.tuple)):
            if not self.tuple[i].similarFragment(other.tuple[i]):
                return False
        return True
        
    def mergeFragment(self, other):
        for i in range(len(self.tuple)):
            self.tuple[i].mergeFragment(other.tuple[i])

    @staticmethod
    def parse(desc, reader):
        """
        parsing a tuple
        """
        try:
            result = tupleDesc(reader.token(), symbols.UNKNOWN)
        except TypeError:
            raise ParsingError(reader, "expected a '(', '<' or '['")
        if reader.newToken() == result.endBracket:
            reader.newToken()
            return result
        result.addEntity(desc.parseEntity(desc, reader))
        if reader.token() == result.endBracket:
            reader.newToken()
            return result
        if reader.token() != symbols.COMMA and reader.token() != symbols.SEMICOLON:
            raise ParsingError(reader, "expected a ',' or ';'")
        result.separator = reader.token()
        result.parseContinue(desc, reader)
        return result

    def parseContinue(self, desc, reader):
        """
        continually parsing the tuple's component
        """
        while reader.token() == self.separator:
            reader.newToken()
            self.addEntity(desc.parseEntity(desc, reader))
        if reader.token() == self.endBracket:
            reader.newToken()
        else: 
            raise ParsingError(reader, "expected a '" + self.endBracket +  "'")

tupleDesc.ASSISTS[-1] = tupleDesc