# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import range
from builtins import super
from builtins import str

from .plane import plane2D, plane3D
from ..struct.coordinate import coordinate, c2, c0, pi
from ..struct.vector import vector2D, vector3D
from ..transform.affineTransform import affineTransform2D, affineTransform3D
from ..io.parseReader import ParsingError


class ellipse2D(plane2D):
    """
    Ellipse object in 2D class

    An ellipse is a closed elliptical planar curve with a discrete
    behavior.

    The ellipse2D class extends on the plane2D class.
    It defines the characteristic individual for ellipses.
    An ellipse is represented as a planar geometry with additionally two vectors
    specifying the foci of the ellipse and another vector specifying a position
    on the ellipse. The foci vectors are naturally ordered.

    **Object Parameters:** (not including inherited attributes)

    * **center** (:class:`~sortal.struct.vector.vector2D`) -- center of the ellipse.
    * **foci** ([sortal.vector2D, sortal.vector2D]) -- the list of the two foci of the ellipse.
    * **ellipsePin** (:class:`~sortal.struct.vector.vector2D`) -- One point on the ellipse.
    * **radii** ([coordinate, coordinate]) -- list of 2 radius values, first value represents the long radius, second represents the short radius.

    **Class Properties:**

    * **Behavior**: :class:`~sortal.form.discrete.discrete`
    * **Sort Argument**: :class:`~sortal.sort.parameter.argument.argument.argumentType`.NONE
    * **Mapping**: :class:`~sortal.map.similarityMap.similarityMap2D`
    """

    type = 'ind-ellipse2D'
    def __init__(self, sort, focus1, focus2, pin, *args, **kwargs):
        """
        initializing a 2D ellipse

        :param sort: sort of the ellipse
        :type sort: ~sortal.sort.sort.sort
        :param focus1: one focus of the ellipse
        :type focus1: ~sortal.struct.vector.vector2D
        :param focus2: other focus of the ellipse
        :type focus2: ~sortal.struct.vector.vector2D
        :param pin: one point on the ellipse
        :type pin: ~sortal.struct.vector.vector2D
        """
        super(ellipse2D, self).__init__(sort)     
        
        if focus1 > focus2:
            self.foci = [focus2, focus1]
        else:
            self.foci = [focus1, focus2]
            
        self.center = focus1.midpoint(focus2)
        self.ellipsePin = pin
        dist1 = pin.subtract(self.foci[0])
        dist2 = pin.subtract(self.foci[1])
        
        # radii = [short radius, long radius]
        self.radii = [None,None]
        self.radii[0] = (dist1.dotproduct(dist1).sqrt() + dist2.dotproduct(dist2).sqrt()) / c2
        dist1 = self.foci[1].subtract(self.foci[0])
        self.radii[1] = (self.radii[0] ** c2 - dist1.dotproduct(dist1) / 4).sqrt()
                
    def getPoints(self):
        """
        Get the two foci and the pin point

        :returns: (focus1, focus2, pin)
        """
        return self.foci[0],self.foci[1], self.ellipsePin      
        
    def getQuadPoints(self):
        """
        Get the two quad points of the ellipse: one point on the ellipse having the lonngest 
        distance from the center and one point on the ellipse having the shortest distance 
        from the center

        :returns: (furthest_point, closest_point)
        """
        dir1 = self.foci[0].subtract(self.center).unit()
        dir2 = vector2D(dir1.y, c0 - dir1.x)
        return self.center.add(dir1.scale(self.radii[0])), self.center.add(dir2.scale(self.radii[1]))
        
        
    def isNil(self):
        """
        Check if the ellipse is Nil
        """
        return False
        
    def printVal(self):
        """
        print out a string representing the value of this individual, used in form.printE()
        """
        return 'foci:'+str(self.foci[0])+','+str(self.foci[1])+', pin:'+str(self.ellipsePin)
        
    def printRhino(self, dX = 0, dY = 0):
        """
        print the individual in a format that can be pasted into Rhinoceros for visualization
        """
        return '_Ellipse f '+str(self.foci[0].x + dX)+','+str(self.foci[0].y + dY)+' '+str(self.foci[1].x + dX)+','+str(self.foci[1].y + dY)+" "+str(self.ellipsePin.x + dX)+','+str(self.ellipsePin.y + dY)
        
    def toString(self, printKey):
        """
        string format of the individual (with no attribute)

        :param printKey: whether the tag of the individual is to be added to the string
        :type printKey: boolean
        :returns: str -- string format of the individual
        """
        s = ''
        if printKey and self.key is not None:
            s += str(self.key)+' '
        return s + '<'+str(self.foci[0])+','+str(self.foci[1])+','+str(self.ellipsePin)+'>'
        
    def equals(self, other):
        """
        check equality
        """
        if self.foci[0] == other.foci[0] and self.foci[1] == other.foci[1]:
            return self.radii[0] == other.radii[0] and self.radii[1] == other.radii[1]
        return False
        
    def containsPoint(self, point):
        """
        Check if the individual contains a specific point

        :param point: point to be checked
        :type point: ~sortal.struct.vector.vector2D
        :returns: boolean -- whether the individual contains the point
        """
        dist1 = point.subtract(self.foci[0])
        dist2 = point.subtract(self.foci[1])
        radius1 = ((dist1.dotproduct(dist1)).sqrt() + (dist2.dotproduct(dist2)).sqrt()) / c2
        distFoci = self.foci[1].subtract(self.foci[0])
        if ((radius1 ** c2 - distFoci.dotproduct(distFoci) / 4).sqrt() == self.radii[1] and
            radius1 == self.radii[0]):
                return True
        return False
        
    def area(self):
        """
        get the area of the ellipse
        """
        return self.radii[0] * self.radii[1] * pi
        
    def compareValue(self, other):
        """
        comparing value of 2 individuals
        """        
        c = self.foci[0].compareValue(other.foci[0])
        if c != 0:
            return c
        c = self.foci[1].compareValue(other.foci[1])
        if c != 0:
            return c
        c = self.radii[0].compareValue(other.radii[0])
        if c != 0:
            return c
        return self.radii[1].compareValue(other.radii[1])
            
    def concentric(self, other):
        """
        check if 2 ellipses have the same center point
        """
        if other.type != 'ind-ellipse2D':
            return False
        return self.foci[0] == other.foci[0] and self.foci[1] == other.foci[1]
        
    def transformInd(self, transformSet):
        """
        Transforming the individual with a transformation set (editor)

        :param transformSet: the transformation set to be applied
        :type transformSet: ~sortal.transform.transformSet.transformSet
        :returns: the transformed object
        """
        affine = transformSet.retrieveTransform(affineTransform2D)
        if affine is None:
            return self.duplicateInd()
        focus1 = self.foci[0].transform(affine)
        focus2 = self.foci[1].transform(affine)
        pin = self.ellipsePin.transform(affine)
        return ellipse2D(self.sort, focus1, focus2, pin)
        
    def replicateInd(self):
        """
        Create a duplicate of the individual without attributes
        
        :returns: the duplicate object
        """
        return ellipse2D(self.sort, self.foci[0], self.foci[1], self.ellipsePin)
        
    @staticmethod
    def parse(reader, sort):
        """
        Parse the 2D ellipse from a sortal parseReader (from string)

        :param reader: the reader of the string to be parsed
        :type reader: ~sortal.io.parseReader.parseReader
        :param sort: sort of the new object
        :type sort: ~sortal.sort.sort.sort
        :returns: :class:`~sortal.ind.ellipse.ellipse2D` -- the individual parsed
        """
        if reader.token() != '<':
            raise ParsingError(reader,'expected a "<" at the beginning of creating a circle')
        if reader.newToken() != '(': #bracket
            raise ParsingError(reader,'expected a "(" ')
        x,y = vector2D.readVector(reader)
        focus1 = vector2D(x,y)
        if reader.token() != ',':
            raise ParsingError(reader,'"," expected')
        reader.newToken() # bracket (current token would be ',')
        x,y = vector2D.readVector(reader)
        focus2 = vector2D(x,y)
        if reader.token() != ',':
            raise ParsingError(reader,'"," expected')
        reader.newToken() # bracket (current token would be ',')
        x,y = vector2D.readVector(reader)
        pin = vector2D(x,y)
        if reader.token() != '>':
            raise ParsingError(reader,'">" expected')
        reader.newToken() #next token
        return ellipse2D(sort, focus1, focus2, pin)
        
class ellipse3D(plane3D):
    """
    Ellipse object in 3D class

    An ellipse is a closed elliptical planar curve with a discrete
    behavior.

    The ellipse3D class extends on the plane3D class.
    It defines the characteristic individual for ellipses.
    An ellipse is represented as a planar geometry with additionally two vectors
    specifying the foci of the ellipse and another vector specifying a position
    on the ellipse. The foci vectors are naturally ordered.

    **Object Parameters:** (not including inherited attributes)

    * **center** (:class:`~sortal.struct.vector.vector3D`) -- center of the ellipse.
    * **localCenter** (:class:`~sortal.struct.vector.vector2D`) -- center of the ellipse localized onto the ellipse's plane.
    * **foci** ([sortal.vector3D, sortal.vector3D]) -- the list of the two foci of the ellipse.
    * **localFoci** ([sortal.vector2D, sortal.vector2D]) -- the two foci of the ellipse localized onto the ellipse's plane.
    * **ellipsePin** (:class:`~sortal.struct.vector.vector3D`) -- One point on the ellipse.
    * **localPin** (:class:`~sortal.struct.vector.vector2D`) -- the localized pin.
    * **radii** ([coordinate, coordinate]) -- list of 2 radius values, first value represents the long radius, second represents the short radius.

    **Class Properties:**

    * **Behavior**: :class:`~sortal.form.discrete.discrete`
    * **Sort Argument**: :class:`~sortal.sort.parameter.argument.argument.argumentType`.NONE
    * **Mapping**: :class:`~sortal.map.similarityMap3D.similarityMap3D`
    """

    type = 'ind-ellipse3D'
    
    def __init__(self, sort, focus1, focus2, pin, *args, **kwargs):
        """
        initializing a 3D ellipse

        :param sort: sort of the ellipse
        :type sort: ~sortal.sort.sort.sort
        :param focus1: one focus of the ellipse
        :type focus1: ~sortal.struct.vector.vector3D
        :param focus2: other focus of the ellipse
        :type focus2: ~sortal.struct.vector.vector3D
        :param pin: one point on the ellipse
        :type pin: ~sortal.struct.vector.vector3D
        """
        super(ellipse3D, self).__init__(sort, [focus1, focus2, pin])
        self.ellipsePin = pin
        self.localPin = self.localize(pin)
        
        localFocus1 = self.localize(focus1)
        localFocus2 = self.localize(focus2)
        
        if focus1 > focus2:
            self.localFoci = [localFocus2, localFocus1]
            self.foci = [focus2, focus1]
        else:
            self.localFoci = [localFocus1, localFocus2]
            self.foci = [focus1, focus2]
            
        self.center = focus1.midpoint(focus2)
        self.localCenter = localFocus1.midpoint(localFocus2)
        
        dist1 = self.localPin.subtract(self.localFoci[0])
        dist2 = self.localPin.subtract(self.localFoci[1])
        self.radii = [None,None]
        self.radii[0] = (dist1.dotproduct(dist1).sqrt() + dist2.dotproduct(dist2).sqrt()) / c2
        dist1 = self.localFoci[1].subtract(self.localFoci[0])
        self.radii[1] = (self.radii[0] ** c2 - dist1.dotproduct(dist1) / 4).sqrt()

    def getPoints(self):
        """
        Get the two foci and the pin point

        :returns: (focus1, focus2, pin)
        """
        return self.foci[0],self.foci[1], self.ellipsePin     
        
    def getQuadPoints(self):
        """
        Get the two quad points of the ellipse: one point on the ellipse having the lonngest 
        distance from the center and one point on the ellipse having the shortest distance 
        from the center

        :returns: (furthest_point, closest_point)
        """
        dir1 = self.foci[0].subtract(self.center)
        dir2 = dir1.product(self.normal).unit()
        dir1 = dir1.unit()
        return self.center.add(dir1.scale(self.radii[0])), self.center.add(dir2.scale(self.radii[1]))
                
    def isNil(self):
        """
        Check if the ellipse is Nil
        """
        return False
        
    def printVal(self):
        """
        print out a string representing the value of this individual, used in form.printE()
        """
        return ('foci:'+str(self.foci[0])+','+str(self.foci[1])+
                ', pin:'+str(self.ellipsePin))
    
    def printRhino(self, dX = 0, dY = 0):
        """
        print the individual in a format that can be pasted into Rhinoceros for visualization
        """
        f0 = self.foci[0]
        f1 = self.foci[1]
        s  = self.ellipsePin
        return '_Ellipse f '+str(f0.x + dX)+','+str(f0.y + dY)+','+str(f0.z)+' '+str(f1.x + dX)+','+str(f1.y + dY)+','+str(f1.z)+' '+str(s.x + dX)+','+str(s.y + dY)+','+str(s.z)
        
    def toString(self, printKey):
        """
        string format of the individual (with no attribute)

        :param printKey: whether the tag of the individual is to be added to the string
        :type printKey: boolean
        :returns: str -- string format of the individual
        """
        s = ''
        if printKey and self.key is not None:
            s += str(self.key)+' '
        return s + '<'+str(self.foci[0])+','+str(self.foci[1])+','+str(self.ellipsePin)+'>'
        
    def equals(self, other):
        """
        check equality
        """
        if self.normal == other.normal and self.localFoci[0] == other.localFoci[0] and self.localFoci[1] == other.localFoci[1]:
            return self.radii[0] == other.radii[0] and self.radii[1] == other.radii[1]
        return False
        
    def area(self):
        """
        get the area of the ellipse
        """
        return self.radii[0] * self.radii[1] * pi
    
    def containsPoint(self, point):
        """
        Check if the individual contains a specific point

        :param point: point to be checked
        :type point: ~sortal.struct.vector.vector3D
        :returns: boolean -- whether the individual contains the point
        """
        fc = self.foci[0]
        if point == fc or not point.subtract(fc).dotproduct(self.normal) == 0:
            return False
        p = self.localize(point)
        dist1 = p.subtract(self.localFoci[0])
        dist2 = p.subtract(self.localFoci[1])
        radius1 = (dist1.dotproduct(dist1).sqrt() + dist2.dotproduct(dist2).sqrt()) / c2
        distFoci = self.localFoci[1].subtract(self.localFoci[0])
        if ((radius1 ** c2 - distFoci.dotproduct(distFoci) / 4).sqrt() == self.radii[1] and
            radius1 == self.radii[0]):
                return True
        return False

        
        
    
    def compareValue(self, other):
        """
        comparing value of 2 individuals
        """        
        c = super(ellipse3D, self).compareValue(other)
        if c != 0:
            return c
        c = self.localFoci[0].compareValue(other.localFoci[0])
        if c != 0:
            return c
        c = self.localFoci[1].compareValue(other.localFoci[1])
        if c != 0:
            return c
        c = self.radii[0].compareValue(other.radii[0])
        if c != 0:
            return c
        return self.radii[1].compareValue(other.radii[1])
        
        
    def concentric(self, other):
        """
        check if 2 ellipses have the same center point
        """
        if other.type != 'ind-ellipse3D':
            return False
        if not super(ellipse3D, self).__eq__(other):
            return False
        return self.localFoci[0] == other.localFoci[0] and self.localFoci[1] == other.localFoci[1]
        
    def transformInd(self, transformSet):
        """
        Transforming the individual with a transformation set (editor)

        :param transformSet: the transformation set to be applied
        :type transformSet: ~sortal.transform.transformSet.transformSet
        :returns: the transformed object
        """
        affine = transformSet.retrieveTransform(affineTransform3D)
        if affine is None:
            return self.duplicateInd()        
        focus1 = self.foci[0].transform(affine)
        focus2 = self.foci[1].transform(affine)
        pin = self.ellipsePin.transform(affine)
        return ellipse3D(self.sort, focus1, focus2, pin)
        
    def replicateInd(self):
        """
        Create a duplicate of the individual without attributes
        
        :returns: the duplicate object
        """
        return ellipse3D(self.sort, self.foci[0], self.foci[1], self.ellipsePin)
        
    @staticmethod
    def parse(reader, sort):
        """
        Parse the 3D ellipse from a sortal parseReader (from string)

        :param reader: the reader of the string to be parsed
        :type reader: ~sortal.io.parseReader.parseReader
        :param sort: sort of the new object
        :type sort: ~sortal.sort.sort.sort
        :returns: :class:`~sortal.ind.ellipse.ellipse3D` -- the individual parsed
        """
        if reader.token() != '<':
            raise ParsingError(reader,'expected a "<" at the beginning of creating a circle')
        if reader.newToken() != '(': #bracket
            raise ParsingError(reader,'expected a "(" ')
        x,y,z = vector3D.readVector(reader)
        focus1 = vector3D(x,y,z)
        if reader.token() != ',':
            raise ParsingError(reader,'"," expected')
        reader.newToken() # bracket (current token would be ',')
        x,y,z = vector3D.readVector(reader)
        focus2 = vector3D(x,y,z)
        if reader.token() != ',':
            raise ParsingError(reader,'"," expected')
        reader.newToken() # bracket (current token would be ',')
        x,y,z = vector3D.readVector(reader)
        pin = vector3D(x,y,z)
        if reader.token() != '>':
            raise ParsingError(reader,'">" expected')
        reader.newToken() #next token
        return ellipse3D(sort, focus1, focus2, pin)