# -*- coding: utf-8 -*-


from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import range
from builtins import super
from builtins import str

from .ellipse import ellipse2D, ellipse3D
from ..struct.coordinate import c2
from ..struct.vector import vector2D, vector3D
from ..transform.affineTransform import affineTransform2D, affineTransform3D
from ..transform.graphTransform import graphTransform, graphTransform3D
from ..io.parseReader import ParsingError


class ellipseP2D(ellipse2D):
    """
    Parametric Ellipse in 2D class. Inherit all its attributes from ellipse2D

    **Class Properties:**

    * **Behavior**: :class:`~sortal.form.discrete.discrete`
    * **Sort Argument**: :class:`~sortal.sort.parameter.argument.argument.argumentType`.NONE
    * **Mapping**: :class:`~sortal.map.graphMap.graphMap`
    """

    type = 'ind-ellipseP2D'
    
    def __init__(self, sort, focus1, focus2, pin, *args, **kwargs):
        """
        initializing a parametric 2D ellipse

        :param sort: sort of the ellipse
        :type sort: ~sortal.sort.sort.sort
        :param focus1: one focus of the ellipse
        :type focus1: ~sortal.struct.vector.vector2D
        :param focus2: other focus of the ellipse
        :type focus2: ~sortal.struct.vector.vector2D
        :param pin: one point on the ellipse
        :type pin: ~sortal.struct.vector.vector2D
        """
        super(ellipseP2D, self).__init__(sort, focus1, focus2, pin)
        
    def codescriptor(self):
        """
        codescriptor of the parametric 2D circle which is itself
        """
        return self
        
    def parametricNode(self):
        """
        node for parametric matching: a duplicate without attribute
        """
        ellipse = ellipseP2D(self.sort, self.foci[0], self.foci[1], self.ellipsePin)
        ellipse.assoc = self
        return ellipse
        

    def transformInd(self, transformSet):
        """
        Transforming the individual with a transformation set (editor)

        :param transformSet: the transformation set to be applied
        :type transformSet: ~sortal.transform.transformSet.transformSet
        :returns: the transformed object
        """
        transf = transformSet.retrieveTransform(graphTransform)
        if transf is None:
            return self.duplicateInd()
        focus1 = transf.apply(self.foci[0])
        focus2 = transf.apply(self.foci[1])
        newPts = []
        for i in transf.vects:
            if self.containsPoint(i[0]):
                newPts.append(i[1])
        if len(newPts) == 0:
            return self
        return ellipseP2D(self.sort, focus1, focus2, newPts[0])
        
    def checkTransform(self, transform, compareForm):
        """
        Check if the individual belongs to the compareForm after transformation with a transform set
        """
        try:
            comp = self.edit(transform)
        except ValueError:
            return False
        if compareForm.containsInd(comp):
            return True
        return False


    @staticmethod
    def parse(reader, sort):
        """
        Parse the parametric 2D ellipse from a sortal parseReader (from string)

        :param reader: the reader of the string to be parsed
        :type reader: ~sortal.io.parseReader.parseReader
        :param sort: sort of the new object
        :type sort: ~sortal.sort.sort.sort
        :returns: :class:`~sortal.ind.ellipseP.ellipseP2D` -- the individual parsed
        """
        if reader.token() != '<':
            raise ParsingError(reader,'expected a "<" at the beginning of creating a ellipse')
        if reader.newToken() != '(': #bracket
            raise ParsingError(reader,'expected a "(" ')
            
        x,y = vector2D.readVector(reader)
        focus1 = vector2D(x,y)
        if reader.token() != ',':
            raise ParsingError(reader,'"," expected')
        reader.newToken() # bracket (current token would be ',')
        
        x,y = vector2D.readVector(reader)
        focus2 = vector2D(x,y)
        if reader.token() != ',':
            raise ParsingError(reader,'"," expected')
        reader.newToken() # bracket (current token would be ',')
        
        x,y = vector2D.readVector(reader)
        pin = vector2D(x,y)
        if reader.token() != '>':
            raise ParsingError(reader,'">" expected')
        reader.newToken() #next token
        return ellipseP2D(sort, focus1, focus2, pin)



class ellipseP3D(ellipse3D):
    """
    Parametric Ellipse in 3D class. Inherit all its attributes from ellipse3D

    **Class Properties:**

    * **Behavior**: :class:`~sortal.form.discrete.discrete`
    * **Sort Argument**: :class:`~sortal.sort.parameter.argument.argument.argumentType`.NONE
    * **Mapping**: :class:`~sortal.map.graphMap3D.graphMap3D`
    """

    type = 'ind-ellipseP3D'
    
    def __init__(self, sort, focus1, focus2, pin, *args, **kwargs):
        """
        initializing a parametric 3D ellipse

        :param sort: sort of the ellipse
        :type sort: ~sortal.sort.sort.sort
        :param focus1: one focus of the ellipse
        :type focus1: ~sortal.struct.vector.vector3D
        :param focus2: other focus of the ellipse
        :type focus2: ~sortal.struct.vector.vector3D
        :param pin: one point on the ellipse
        :type pin: ~sortal.struct.vector.vector3D
        """
        super(ellipseP3D, self).__init__(sort, focus1, focus2, pin)
        
    def codescriptor(self):
        """
        codescriptor of the parametric 3D circle which is itself
        """
        return self
        
    def parametricNode(self):
        """
        node for parametric matching: a duplicate without attribute
        """
        ellipse = ellipseP3D(self.sort, self.foci[0], self.foci[1], self.ellipsePin)
        ellipse.assoc = self
        return ellipse
        # return self
        

    def transformInd(self, transformSet):
        """
        Transforming the individual with a transformation set (editor)

        :param transformSet: the transformation set to be applied
        :type transformSet: ~sortal.transform.transformSet.transformSet
        :returns: the transformed object
        """
        transf = transformSet.retrieveTransform(graphTransform3D)
        if transf is None:
            return self.duplicateInd()
        focus1 = transf.apply(self.foci[0])
        focus2 = transf.apply(self.foci[1])
        newPts = []
        for i in transf.vects:
            if self.containsPoint(i[0]):
                newPts.append(i[1])
        if len(newPts) == 0:
            return self
        return ellipseP3D(self.sort, focus1, focus2, newPts[0])
        
        
    def checkTransform(self, transform, compareForm):
        """
        Check if the individual belongs to the compareForm after transformation with a transform set
        """
        try:
            comp = self.edit(transform)
        except ValueError:
            return False
        if compareForm.containsInd(comp):
            return True
        return False


    @staticmethod
    def parse(reader, sort):
        """
        Parse the parametric 3D ellipse from a sortal parseReader (from string)

        :param reader: the reader of the string to be parsed
        :type reader: ~sortal.io.parseReader.parseReader
        :param sort: sort of the new object
        :type sort: ~sortal.sort.sort.sort
        :returns: :class:`~sortal.ind.ellipseP.ellipseP3D` -- the individual parsed
        """
        if reader.token() != '<':
            raise ParsingError(reader,'expected a "<" at the beginning of creating a ellipse')
        if reader.newToken() != '(': #bracket
            raise ParsingError(reader,'expected a "(" ')
            
        x,y,z = vector3D.readVector(reader)
        focus1 = vector3D(x,y,z)
        if reader.token() != ',':
            raise ParsingError(reader,'"," expected')
        reader.newToken() # bracket (current token would be ',')
        
        x,y,z = vector3D.readVector(reader)
        focus2 = vector3D(x,y,z)
        if reader.token() != ',':
            raise ParsingError(reader,'"," expected')
        reader.newToken() # bracket (current token would be ',')
        
        x,y,z = vector3D.readVector(reader)
        pin = vector3D(x,y,z)
        if reader.token() != '>':
            raise ParsingError(reader,'">" expected')
        reader.newToken() #next token
        
        return ellipseP3D(sort, focus1, focus2, pin)