# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import range
from builtins import super
from builtins import str

from .plane import plane2D, plane3D
from .ellipse import ellipse2D, ellipse3D

from ..io.parseReader import ParsingError
from ..transform.affineTransform import affineTransform2D, affineTransform3D
from ..struct.vector import vector2D, vector3D
from ..struct.coordinate import coordinate, pi, acos, sqrt

class ellipticalArc2D(ellipse2D):
    """
    Elliptical Arc in 2D class

    An elliptical arc is a connected and bounded segment of an ellipse with an
    interval behavior. The ellipse defines the co-descriptor of the elliptical arc,
    the boundary of the elliptical arc is defined by the start and end directions
    of the arc segment.

    The ellipticalArc2D class extends on the ellipse2D class. It defines the characteristic
    individual for elliptical arcs. An elliptical arc is represented as a ellipse with
    additionally a start and end positions. The elliptical arc's start and end are ordered
    based on their relative position from the ellipse's reference point (a point
    on the ellipse that is horizontally to the right of the ellipse's center) with start being
    closer to the reference point than end (clockwise from reference point to start/end).
    Comparison between arcs on the same ellipse is performed in a similar manner.

    **Object Parameters:** (not including inherited attributes)

    * **start** (:class:`~sortal.struct.vector.vector2D`) -- counter-clockwise end point of the arc
    * **end** (:class:`~sortal.struct.vector.vector2D`) -- clockwise end point of the arc 
    * **pin** (:class:`~sortal.struct.vector.vector2D`) -- a point within the angle from the ellipse center that the arc defines. it is not required to be on the arc itself
    * **fullCircle** (boolean) -- boolean check indicating whether the arc covers the whole ellipse
    * **quadPoint** one of the quad points of the ellipse

    **Class Properties:**

    * **Behavior**: :class:`~sortal.form.cyclical.cyclical`
    * **Sort Argument**: :class:`~sortal.sort.parameter.argument.argument.argumentType`.NONE
    * **Mapping**: :class:`~sortal.map.similarityMap.similarityMap2D`

    """
    
    type = 'ind-ellipticalArc2D'
    
    def __init__(self, sort, focus1, focus2, start, end, pin, quadPoint = None, *args, **kwargs):
        """
        Initializing a new 2D Elliptical arc

        :param sort: sort of the arc
        :type sort: ~sortal.sort.sort.sort
        :param center: ellipse center
        :type center: ~sortal.struct.vector.vector2D
        :param start: counter-clockwise end point of the arc
        :type start: ~sortal.struct.vector.vector2D
        :param end: clockwise end point of the arc 
        :type end: ~sortal.struct.vector.vector2D
        :param pin: a point within the angle from the ellipse center that the arc defines. 
            Not required to be on the arc itself.
        :type pin: ~sortal.struct.vector.vector2D
        :param quadPoint: (optional) one of the quad points of the ellipse
        :type quadPoint: one of the quad points of the ellipse
        """
        if start == end:
            if pin is True:
                raise ValueError('For full elliptical arc, pin cannot be calculated')
            elif not start.colinear(focus1, pin):
                super(ellipticalArc2D, self).__init__(sort, focus1, focus2, start)
                self.quadPoint = quadPoint
                self.end = self.start = start
                self.pin = pin
                self.fullCircle = True
                return
            else:
                raise ValueError('Cannot create an arc with start point, end point and pin being the same')
        center = focus1.midpoint(focus2)
        
        if pin is True:
            pin = plane2D.findPin(center, start, end)
            
        if plane2D.angleContainsPin(center, start, pin, end) == 1:
            start, end = end, start

        super(ellipticalArc2D, self).__init__(sort, focus1, focus2, start)
        self.pin = pin
        
        self.start = start
        self.end = end
        
        self.fullCircle = False
        self.quadPoint = quadPoint
        
        
        
    def angle(self):
        """
        Find the angle that defines the arc from the ellipse center in radian
        """
        nom = self.start.dotproduct(self.end)
        denom = self.start.dotproduct(self.start) * self.end.dotproduct(self.end)
        compAngl = plane2D.compareAngle(self.center, self.start, self.end)
        if compAngl == 1:
            return coordinate(acos(nom.value / sqrt(denom.value)))
        elif compAngl == -1:
            return coordinate(2 * pi - acos(nom.value / sqrt(denom.value)))
        return coordinate(pi * (not self.start == self.end))
    
    def toString(self, printKey):
        """
        string format of the individual (with no attribute)

        :param printKey: whether the tag of the individual is to be added to the string
        :type printKey: boolean
        :returns: str -- string format of the individual
        """
        s = ''
        if printKey and self.key is not None:
            s += str(self.key)+' '
        return s + '<'+str(self.foci[0])+','+str(self.foci[1])+','+str(self.start)+','+str(self.end)+','+str(self.pin)+'>'

    def printVal(self):
        """
        print out a string representing the value of this individual, used in form.printE()
        """
        #return str(self.start)+','+str(self.end)
        return '<f:'+str(self.foci[0])+','+str(self.foci[1])+', s:'+str(self.start)+', e:'+str(self.end)+'>'+ str(self.radii)
        return 'center:'+str(self.center)+', start:'+str(self.start)+', end:'+str(self.end)

    def equals(self, other):
        """
        check equality
        """
        return (self.foci[0] == other.foci[0] and self.foci[1] == other.foci[1] 
                and self.start == other.start and self.end == other.end)

    def isNil(self):
        """
        Check if the start and end points of the arc are the same
        """
        if self.end is None or self.start is None:
            return True
        return self.start == self.end and not self.fullCircle

    def compareInterval(self, other):
        """
        Compare the value of two arcs:

        * First compare the ellipses they are on
        * If On the same ellipses, compare start points: point becomes bigger the more clockwise they are from the point that is horizontally to the right of the ellipse center
        * If starts points are equal, compare the end points with the same criteria

        # this function is named compareInterval and not compareValue since there
        is already a compareValue function in the ellipses 

        :param other: other elliptical arc to be compared
        :type other: ~sortal.ind.ellipticalArc.ellipticalArc2D
        :returns: int -- 0 if equals, 1 if this line is bigger, -1 if smaller
        """
        c = self.compareValue(other) 
        if c != 0:
            return c
        c = plane2D.compareRelAngle(self.center, self.center.add(vector2D(1,0)), self.start, other.start)
        if c != 0:
            return c
        return plane2D.compareRelAngle(self.center, self.start, self.end, other.end)

    def containsPoint(self, point):
        """
        Check if the arc contains a specific point

        :param point: point to be checked
        :type point: ~sortal.struct.vector.vector2D
        :returns: boolean -- whether the arc contains the point
        """
        if not super(ellipticalArc2D, self).containsPoint(point):
            return False
        return plane2D.angleContainsPin(self.center, self.start, point, self.end) != 1
        
    def compareEnd(self, other):
        """
        compare end points of two arcs
        """
        return plane2D.angleContainsPin(self.center, self.start, self.end, other.end)

    def length(self):
        """
        # NOT IMPLEMENTED
        """
        raise NotImplementedError
        
    def touches(self,other):
        """
        check if two arcs on the same ellipse touch

        :param other: other arc to be checked
        :type other: ~sortal.ind.ellipticalArc.ellipticalArc2D
        :returns: boolean -- whether the two arcs touch
        """
        if not self.compareValue(other) == 0:
            return False
        if self.fullCircle or other.fullCircle:
            return False
        return ((self.end == other.start and 
                 plane2D.angleContainsPin(self.center, self.start, other.end, self.end) != -1) or 
                (self.start == other.end and 
                 plane2D.angleContainsPin(self.center, self.start, other.start, self.end) != -1)) 

    def contains(self,other):
        """
        check if this arc contains another

        :param other: other arc to be checked
        :type other: ~sortal.ind.ellipticalArc.ellipticalArc2D
        :returns: boolean -- whether this arc contains other
        """
        if not self.compareValue(other) == 0:
            return False
        if self.fullCircle:
            return True
        if other.fullCircle:
            return False
        if self.start == other.start and self.end == other.end:
            return True
        selfContainsOtherStart = plane2D.angleContainsPin(self.center, self.start, other.start, self.end)
        selfContainsOtherEnd = plane2D.angleContainsPin(self.center, self.start, other.end, self.end)

        if (selfContainsOtherStart <= 0 and selfContainsOtherEnd <= 0 and 
            plane2D.angleContainsPin(self.center, other.start, self.start, other.end) <= 0 and 
            plane2D.angleContainsPin(self.center, other.start, self.end, other.end) <= 0):
                return False
        return (selfContainsOtherStart <= 0 and selfContainsOtherEnd <= 0 )
    
    def disjoint(self,other):
        """
        check if two arcs on the same ellipse are disjoint

        :param other: other arc to be checked
        :type other: ~sortal.ind.ellipticalArc.ellipticalArc2D
        :returns: boolean -- whether the two arcs are disjoint
        """
        if not self.compareValue(other) == 0:
            return False
        if self.fullCircle or other.fullCircle:
            return False
        return (plane2D.angleContainsPin(self.center, self.start, other.start, self.end) == 1 and 
                plane2D.angleContainsPin(self.center, self.start, other.end, self.end)  == 1 and 
                plane2D.angleContainsPin(self.center, other.start, self.start, other.end) == 1 and 
                plane2D.angleContainsPin(self.center, other.start, self.end, other.end)  == 1 )

    def align(self, other):
        """
        check if two arcs on the same ellipse align

        :param other: other arc to be checked
        :type other: ~sortal.ind.ellipticalArc.ellipticalArc2D
        :returns: boolean -- whether the two arcs align
        """
        if not self.compareValue(other) == 0:
            return False
        return self.end == other.end or self.start == other.start
        
    ####################################
    def common(self,other):
        """
        Find the common of two arcs, The two must be on the same ellipse
        
        :param other: other arc to find common of
        :type other: ~sortal.ind.ellipticalArc.ellipticalArc2D
        :returns: :class:`~sortal.ind.ellipticalArc.ellipticalArc2D` -- The common elliptical arc
        """
        result = []
        if self.fullCircle:
            new = ellipticalArc2D(self.sort, other.foci[0], other.foci[1], other.start, other.end, other.pin, other.quadPoint)
            if self.hasAttribute():
                new.addAttribute(self.attribute)
            result.append(new)
            return result
        if other.fullCircle:
            new = ellipticalArc2D(self.sort, self.foci[0], self.foci[1], self.start, self.end, self.pin, self.quadPoint)
            if self.hasAttribute():
                new.addAttribute(self.attribute)
            result.append(new)
            return result
            
        self_oStart = plane2D.angleContainsPin(self.center, self.start, other.start, self.end)
        self_oEnd = plane2D.angleContainsPin(self.center, self.start, other.end, self.end)
        other_sStart = plane2D.angleContainsPin(self.center, other.start, self.start, other.end)
        other_sEnd = plane2D.angleContainsPin(self.center, other.start, self.end, other.end)
        
        if self_oStart == 1 and self_oEnd == 1 and other_sStart == 1 and other_sEnd == 1:
            return result
            
        if (self.end == other.start and self_oEnd != -1) or (self.start == other.end and self_oStart != -1):
            return result
        
        if self_oStart == -1 and self_oEnd == -1 and other_sStart == -1 and other_sEnd == -1:
            result.append(ellipticalArc2D(self.sort, self.foci[0], self.foci[1], self.start, other.end, True, self.quadPoint))
            result.append(ellipticalArc2D(self.sort, self.foci[0], self.foci[1], other.start , self.end, True, self.quadPoint))
            if self.hasAttribute():
                for i in result:
                    i.addAttribute(self.attribute)
            return result
            
        if self_oStart == -1:
            start = other.start
        else:
            start = self.start
        if self_oEnd == -1:
            end = other.end
        else:
            end = self.end
        
        result.append(ellipticalArc2D(self.sort, self.foci[0], self.foci[1], start , end, True, self.quadPoint))
        if self.hasAttribute():
            result[0].addAttribute(self.attribute)
        return result

        
    def combine(self, other):
        """
        Find the combination of two arcs, The two must be on the same ellipse
        
        :param other: other arc to find combination of
        :type other: ~sortal.ind.ellipticalArc.ellipticalArc2D
        :returns: :class:`~sortal.ind.ellipticalArc.ellipticalArc2D` -- The combination elliptical arc
        """
        if self.fullCircle:
            return False
        if other.fullCircle:
            self.start = other.start
            self.end = other.end
            self.pin = other.pin
            return True
            
        self_oStart = plane2D.angleContainsPin(self.center, self.start, other.start, self.end)
        self_oEnd = plane2D.angleContainsPin(self.center, self.start, other.end, self.end)
        other_sStart = plane2D.angleContainsPin(self.center, other.start, self.start, other.end)
        other_sEnd = plane2D.angleContainsPin(self.center, other.start, self.end, other.end)
        
        if self_oStart == 1 and self_oEnd == 1 and other_sStart == 1 and other_sEnd == 1:
            return False
            
        if self_oStart != 1 and self_oEnd != 1 and other_sStart != 1 and other_sEnd != 1:
            if self.start == other.start and self.end == other.end:
                return False
            if self.pin.colinear(self.start, self.center):
                self.pin = self.end
                if self.pin.colinear(self.start, self.center):
                    self.pin = plane2D.findPin(self.center, self.start, self.end)
            self.end = self.start
            self.fullCircle = True
            return True
            
        if other_sStart < 1:
            self.start = other.start
        if other_sEnd < 1:
            self.end = other.end
            
        self.pin = plane2D.findPin(self.center, self.start, self.end)
        
        return True
        
    def complement(self,other):
        """
        Find the complement of this arc with another, The two must be on the same ellipse
        
        :param other: other arc to find complement with
        :type other: ~sortal.ind.ellipticalArc.ellipticalArc2D
        :returns: list of :class:`~sortal.ind.ellipticalArc.ellipticalArc2D` -- The list of complement elliptical arcs
        """
        result = []
        if other.fullCircle:
            return result
        if self.fullCircle:
            result.append(ellipticalArc2D(self.sort, other.foci[0], other.foci[1], other.end , other.start, other.pin, other.quadPoint))
            return result
            
        self_oStart = plane2D.angleContainsPin(self.center, self.start, other.start, self.end)
        self_oEnd = plane2D.angleContainsPin(self.center, self.start, other.end, self.end)
        other_sStart = plane2D.angleContainsPin(self.center, other.start, self.start, other.end)
        other_sEnd = plane2D.angleContainsPin(self.center, other.start, self.end, other.end)


        if self_oStart != 1 and self_oEnd != 1 and other_sStart != 1 and other_sEnd != 1:
            result.append(ellipticalArc2D(self.sort, self.foci[0], self.foci[1], other.end, other.start, True, other.quadPoint))
            
        else:
            if self_oStart == -1:
                result.append(ellipticalArc2D(self.sort, self.foci[0], self.foci[1], self.start, other.start, True, self.quadPoint))
            if self_oEnd == -1:
                result.append(ellipticalArc2D(self.sort, self.foci[0], self.foci[1], other.end, self.end, True, self.quadPoint))
            
        if self.hasAttribute():
            for i in result:
                i.addAttribute(self.attribute)
        return result
                
    def replicateInd(self, sort = 0):
        """
        Create a duplicate of the individual without attributes
        
        :returns: the duplicate object
        """
        if sort == 0:
            sort = self.sort
        return ellipticalArc2D(sort, self.foci[0], self.foci[1], self.start, self.end, self.pin, self.quadPoint) 
        
    def codescriptor(self):
        """
        Codescriptor of the arc: the 2D ellipse containing this arc

        :returns: sortal.ellipse2D
        """
        ellipse = ellipse2D(self.sort, self.foci[0], self.foci[1], self.start, self.end, self.pin)
        ellipse.assoc = self
        return ellipse
        
    def intersect(self, other):
        raise NotImplementedError
        
    def midpoint(self):
        raise NotImplementedError
        
    def borderPoint(self):
        """
        border point
        """
        a = vector2D(1,0) 
        b = vector2D(0,1)
        foci0 = self.foci[0].subtract(self.center)
        foci0Sqrt = foci0.dotproduct(foci0).sqrt()
        cosAlpha = foci0.dotproduct(a)/foci0Sqrt
        sinAlpha = foci0.dotproduct(b)/foci0Sqrt
        sq = (cosAlpha/self.radii[0])**2 + (sinAlpha/self.radii[1])**2
        return vector2D(1/(sq.sqrt()) + self.center.x, self.center.y)
        
    def splitArc(self):
        """
        Split the arc according to the reference-point: the point on the ellipse that is horizontally to
        the right of the ellipse center. If the arc contains the reference-point, it will be split to 
        become two arcs:

        * One arc will start with reference-point, end with the original end-point
        * Another arc will start with the original start point, end with the reference-point

        This splitting is performed for all arcs that are put into cyclical forms during so that they can be
        ordered according to their start points and end points more effectively without overlapping.

        :returns: list of :class:`~sortal.ind.ellipticalArc.ellipticalArc2D` -- The list of the two splitted elliptical arcs, None if the arc cannot be splitted
        """
        ref = self.center.add(vector2D(1,0))
        if plane2D.angleContainsPin(self.center, self.start, ref, self.end) == -1:
            ref = self.borderPoint()
            a = ellipticalArc2D(self.sort, self.foci[0], self.foci[1], ref, self.end, True, self.quadPoint)
            b = ellipticalArc2D(self.sort, self.foci[0], self.foci[1], self.start, ref, True, self.quadPoint)
            if self.hasAttribute():
                a.addAttribute(self.attribute)
                b.addAttribute(self.attribute)
            return [a,b]
        return None
        
    def crossingOver(self):
        """
        Check if the arc crosses over the reference-point.

        :returns: boolean
        """
        return plane2D.angleContainsPin(self.center, self.start, 
                                       self.center.add(vector2D(1,0)), 
                                       self.end)
        
    def transformInd(self, transformSet):
        """
        Transforming the individual with a transformation set (editor)

        :param transformSet: the transformation set to be applied
        :type transformSet: ~sortal.transform.transformSet.transformSet
        :returns: the transformed object
        """
        affine = transformSet.retrieveTransform(affineTransform2D)
        if affine is None:
            return self.duplicateInd()
        start = self.start.transform(affine)
        end = self.end.transform(affine)         
        pin = self.pin.transform(affine)
        focus1 = self.foci[0].transform(affine)    
        focus2 = self.foci[1].transform(affine)    
        quad = self.quadPoint.transform(affine)    
        

        check = 1
        for i in range(len(affine.matrix)):
            if affine.matrix[i][i] < 0:
                check *= -1
        if check == -1:
            temp = end
            end = start         
            start = temp

        return ellipticalArc2D(self.sort, focus1, focus2, start, end, pin, quad)
                
    def checkTransform(self, transform, compareForm):
        """
        Check if the individual belongs to the compareForm after transformation with a transform set
        """
        affine = transform.retrieveTransform(affineTransform2D)
        if affine is None:
            tfArc = self
        else:
            tfArc = self.transformInd(transform)
            if self.hasAttribute():
                tfArc.addAttribute(self.attribute)
        splits = tfArc.splitArc()
        if splits is None:
            if not compareForm.containsInd(tfArc):
                return False
        else:
            for arc in splits:
                if not compareForm.containsInd(arc):
                    return False
        return True
        
    def containsInd(self, ind):
        """
        Check if this individual contains another individual
        """
        if not self.contains(ind):
            return False
        if self.hasAttribute():
            if ind.hasAttribute():
                if ind.attribute.checkPartOf(self.attribute):
                    return True
                else:
                    return False
            else:
                return True
        return True
        
    @staticmethod
    def parse(reader, sort):
        """
        Parse the 2D elliptical arc from a sortal parseReader (from string)

        :param reader: the reader of the string to be parsed
        :type reader: ~sortal.io.parseReader.parseReader
        :param sort: sort of the new object
        :type sort: ~sortal.sort.sort.sort
        :returns: :class:`~sortal.ind.ellipticalArc.ellipticalArc2D` -- the individual parsed
        """
        if reader.token() != '<':
            raise ParsingError(reader,'expected a "<" at the beginning of creating a elliptical arc')
        if reader.newToken() != '(': #bracket
            raise ParsingError(reader,'expected a "(" ')
            
        x,y = vector2D.readVector(reader)
        focus1 = vector2D(x,y)
            
        reader.newToken() # bracket (current token would be ',')
        
        x,y = vector2D.readVector(reader)
        focus2 = vector2D(x,y)
            
        reader.newToken() # bracket (current token would be ',')
        
        x,y = vector2D.readVector(reader)
        start = vector2D(x,y)
            
        reader.newToken() # bracket (current token would be ',')
        
        x,y = vector2D.readVector(reader)
        end = vector2D(x,y)
            
        reader.newToken() # bracket (current token would be ',')
        
        x,y = vector2D.readVector(reader)
        pin = vector2D(x,y)
            
        ind = ellipticalArc2D(sort, focus1, focus2, start, end, pin)

        if reader.token() != '>':
            raise ParsingError(reader, '">" expected to close the line-segment')
            
        reader.newToken() #next token
        
        return ind
        
        
class ellipticalArc3D(ellipse3D):
    """
    Elliptical Arc in 3D class

    An elliptical arc is a connected and bounded segment of an ellipse with an
    interval behavior. The ellipse defines the co-descriptor of the elliptical arc,
    the boundary of the elliptical arc is defined by the start and end directions
    of the arc segment.

    The ellipticalArc3D class extends on the ellipse3D class. It defines the characteristic
    individual for elliptical arcs. An elliptical arc is represented as a ellipse with
    additionally a start and end positions. The elliptical arc's start and end are ordered
    based on their relative position from the ellipse's reference point (a point
    on the localized ellipse that is horizontally to the right of the ellipse's center) with 
    start being closer to the reference point than end (clockwise from reference point to start/end).
    Comparison between arcs on the same ellipse is performed in a similar manner.

    **Object Parameters:** (not including inherited attributes)

    * **start** (:class:`~sortal.struct.vector.vector3D`) -- counter-clockwise end point of the arc
    * **localStart** (:class:`~sortal.struct.vector.vector2D`) -- localized start point on the plane containing the arc
    * **end** (:class:`~sortal.struct.vector.vector3D`) -- clockwise end point of the arc 
    * **localEnd** (:class:`~sortal.struct.vector.vector2D`) -- localized end point on the plane containing the arc
    * **pin** (:class:`~sortal.struct.vector.vector3D`) -- a point within the angle from the circle center that the arc defines. it is not required to be on the arc itself
    * **localPin** (:class:`~sortal.struct.vector.vector2D`) -- localized pin on the plane containing the arc
    * **fullCircle** (boolean) -- boolean check indicating whether the arc covers the whole circle
    * **quadPoint** one of the quad points of the ellipse

    **Class Properties:**

    * **Behavior**: :class:`~sortal.form.cyclical.cyclical`
    * **Sort Argument**: :class:`~sortal.sort.parameter.argument.argument.argumentType`.NONE
    * **Mapping**: :class:`~sortal.map.similarityMap3D.similarityMap3D`
    """
    type = 'ind-ellipticalArc3D'
    
    def __init__(self, sort, focus1, focus2, start, end, pin, quadPoint = None, *args, **kwargs):
        """
        Initializing a new 3D Elliptical arc

        :param sort: sort of the arc
        :type sort: ~sortal.sort.sort.sort
        :param center: ellipse center
        :type center: ~sortal.struct.vector.vector3D
        :param start: counter-clockwise end point of the arc
        :type start: ~sortal.struct.vector.vector3D
        :param end: clockwise end point of the arc 
        :type end: ~sortal.struct.vector.vector3D
        :param pin: a point within the angle from the ellipse center that the arc defines. 
            Not required to be on the arc itself.
        :type pin: ~sortal.struct.vector.vector3D
        :param quadPoint: (optional) one of the quad points of the ellipse
        :type quadPoint: one of the quad points of the ellipse
        """
        if start == end:
            if pin is True:
                raise ValueError('For full elliptical arc, pin cannot be calculated')
            elif not start.colinear(focus1, pin):
                super(ellipticalArc3D, self).__init__(sort, focus1, focus2, start, pin)
                self.end = self.start = start
                self.localEnd = self.localStart = self.localize(start)
                self.pin = pin
                self.localPin = self.localize(pin)
                self.fullCircle = True
                self.quadPoint = quadPoint
                return
            else:
                raise ValueError('Cannot create an arc with start point, end point and pin being the same')

        if start.colinear(focus1, focus2):
            if end.colinear(focus1, focus2):
                center = focus2.midpoint(focus1)
                fDir = focus2.subtract(center)
                normal = fDir.product(pin.subtract(center))
                fDir = normal.product(fDir).unit()
                
                dist0 = start.subtract(end)
                dist1 = focus2.subtract(focus1)
                r1 = (dist0.dotproduct(dist0) - dist1.dotproduct(dist1)).sqrt() / 2
                nPin = center.add(fDir.scale(r1))
                super(ellipticalArc3D, self).__init__(sort, focus1, focus2, nPin)
            else:
                super(ellipticalArc3D, self).__init__(sort, focus1, focus2, end)
        else:
            super(ellipticalArc3D, self).__init__(sort, focus1, focus2, start)
            
        self.start = start
        self.end = end
        self.localStart = self.localize(start)
        self.localEnd = self.localize(end)
        
        self.pin = pin
        self.localPin = self.localize(pin)
            
        if plane2D.angleContainsPin(self.localCenter, self.localStart, self.localPin, self.localEnd) == 1:
            self.start, self.end, self.localStart, self.localEnd = self.end, self.start, self.localEnd, self.localStart
            
        self.fullCircle = False
        self.quadPoint = quadPoint
        
    def angle(self):
        """
        Find the angle that defines the arc from the ellipse center in radian
        """
        nom = self.localStart.dotproduct(self.localEnd)
        denom = self.localStart.dotproduct(self.localStart) * self.localEnd.dotproduct(self.localEnd)
        compAngl = plane2D.compareAngle(self.localCenter, self.localStart, self.localEnd)
        if compAngl == 1:
            return coordinate(acos(nom.value / sqrt(denom.value)))
        elif compAngl == -1:
            return coordinate(2 * pi - acos(nom.value / sqrt(denom.value)))
        return coordinate(pi * (not self.localStart == self.localEnd))
        
        
    def toString(self, printKey):
        """
        string format of the individual (with no attribute)

        :param printKey: whether the tag of the individual is to be added to the string
        :type printKey: boolean
        :returns: str -- string format of the individual
        """
        s = ''
        if printKey and self.key is not None:
            s += str(self.key)+' '
        return (s + '<'+str(self.foci[0])+','+str(self.foci[1])+
                ','+str(self.start)+','+str(self.end)+
                ','+str(self.pin)+'>')

    def printVal(self):
        """
        print out a string representing the value of this individual, used in form.printE()
        """
        #return str(self.start)+','+str(self.end)
        return ('<f:'+str(self.foci[0])+','+str(self.foci[1])+
                ', s:'+str(self.start)+', e:'+str(self.end)+'>')
                
        return ('foci:'+str(self.foci[0])+','+str(self.foci[1])+
                ', start:'+str(self.start)+', end:'+str(self.end))
    
    def equals(self, other):
        """
        check equality
        """
        return (self.localFoci[0] == other.localFoci[0] and self.localFoci[1] == other.localFoci[1] and 
                self.localStart == other.localStart and self.localEnd == other.localEnd)

    def isNil(self):
        """
        Check if the start and end points of the arc are the same
        """
        if self.localEnd is None or self.localStart is None:
            return True
        return self.localStart == self.localEnd and not self.fullCircle
        
    def compareInterval(self, other):
        """
        Compare the value of two arcs:

        * First compare the ellipses they are on
        * If On the same ellipses, compare start points: point becomes bigger the more clockwise they are from the point that is horizontally to the right of the ellipse center
        * If starts points are equal, compare the end points with the same criteria

        # this function is named compareInterval and not compareValue since there
        is already a compareValue function in the ellipses 

        :param other: other elliptical arc to be compared
        :type other: ~sortal.ind.ellipticalArc.ellipticalArc3D
        :returns: int -- 0 if equals, 1 if this line is bigger, -1 if smaller
        """
        c = self.compareValue(other)
        if c != 0:
            return c
        c = plane2D.compareRelAngle(self.localCenter, self.localCenter.add(vector2D(1,0)), self.localStart, other.localStart)
        if c != 0:
            return c
        return plane2D.compareRelAngle(self.localCenter, self.localStart, self.localEnd, other.localEnd)
                

    def containsPoint(self, point):
        """
        Check if the arc contains a specific point

        :param point: point to be checked
        :type point: ~sortal.struct.vector.vector3D
        :returns: boolean -- whether the arc contains the point
        """
        if not super(ellipticalArc3D, self).containsPoint(point):
            return False
        return plane2D.angleContainsPin(self.localCenter, self.localStart, self.localize(point), self.localEnd) != 1
        
    def compareEnd(self, other):
        """
        compare end points of two arcs
        """
        return plane2D.angleContainsPin(self.localCenter, self.localStart, self.localEnd, other.localEnd)
            
    def touches(self,other):
        """
        check if two arcs on the same ellipse touch

        :param other: other arc to be checked
        :type other: ~sortal.ind.ellipticalArc.ellipticalArc3D
        :returns: boolean -- whether the two arcs touch
        """
        if not self.compareValue(other) == 0:
            return False
        if self.fullCircle or other.fullCircle:
            return False
        return ((self.localEnd == other.localStart and 
                 plane2D.angleContainsPin(self.localCenter, self.localStart, other.localEnd, self.localEnd) != -1) or 
                (self.localStart == other.localEnd and 
                 plane2D.angleContainsPin(self.localCenter, self.localStart, other.localStart, self.localEnd) != -1)) 
    
    def contains(self,other):
        """
        check if this arc contains another

        :param other: other arc to be checked
        :type other: ~sortal.ind.ellipticalArc.ellipticalArc3D
        :returns: boolean -- whether this arc contains other
        """
        if not self.compareValue(other) == 0:
            return False
        if self.fullCircle:
            return True
        if other.fullCircle:
            return False
        if self.localStart == other.localStart and self.localEnd == other.localEnd:
            return True
            
        selfContainsOtherStart = plane2D.angleContainsPin(self.localCenter, self.localStart, other.localStart, self.localEnd)
        selfContainsOtherEnd = plane2D.angleContainsPin(self.localCenter, self.localStart, other.localEnd, self.localEnd)
                
        if (selfContainsOtherStart <= 0 and selfContainsOtherEnd <= 0 and 
            plane2D.angleContainsPin(self.localCenter, other.localStart, self.localStart, other.localEnd) <= 0 and 
            plane2D.angleContainsPin(self.localCenter, other.localStart, self.localEnd, other.localEnd) <= 0):
                return False
        return (selfContainsOtherStart <= 0 and selfContainsOtherEnd <= 0 )
    
        
    def disjoint(self,other):
        """
        check if two arcs on the same ellipse are disjoint

        :param other: other arc to be checked
        :type other: ~sortal.ind.ellipticalArc.ellipticalArc3D
        :returns: boolean -- whether the two arcs are disjoint
        """
        if not self.compareValue(other) == 0:
            return False
        if self.fullCircle or other.fullCircle:
            return False
        return (plane2D.angleContainsPin(self.localCenter, self.localStart, other.localStart, self.localEnd) == 1 and 
                plane2D.angleContainsPin(self.localCenter, self.localStart, other.localEnd, self.localEnd)  == 1 and 
                plane2D.angleContainsPin(self.localCenter, other.localStart, self.localStart, other.localEnd) == 1 and 
                plane2D.angleContainsPin(self.localCenter, other.localStart, self.localEnd, other.localEnd)  == 1 )
    
                
    def align(self, other):
        """
        check if two arcs on the same ellipse align

        :param other: other arc to be checked
        :type other: ~sortal.ind.ellipticalArc.ellipticalArc3D
        :returns: boolean -- whether the two arcs align
        """
        if not self.compareValue(other) == 0:
            return False
        return self.localEnd == other.localEnd or self.localStart == other.localStart
        
        
    def common(self,other):
        """
        Find the common of two arcs, The two must be on the same ellipse
        
        :param other: other arc to find common of
        :type other: ~sortal.ind.ellipticalArc.ellipticalArc3D
        :returns: :class:`~sortal.ind.ellipticalArc.ellipticalArc3D` -- The common elliptical arc
        """
        result = []
        if self.fullCircle:
            new = ellipticalArc3D(self.sort, other.foci[0], other.foci[1], other.start, other.end, other.pin, other.quadPoint)
            if self.sort.type == 'sort-attribute':
                new.addAttribute(self.attribute)
            result.append(new)
            return result
        if other.fullCircle:
            new = ellipticalArc3D(self.sort, self.foci[0], self.foci[1], self.start, self.end, self.pin, self.quadPoint)
            if self.sort.type == 'sort-attribute':
                new.addAttribute(self.attribute)
            result.append(new)
            return result
            
        self_oStart = plane2D.angleContainsPin(self.localCenter, self.localStart, other.localStart, self.localEnd)
        self_oEnd = plane2D.angleContainsPin(self.localCenter, self.localStart, other.localEnd, self.localEnd)
        other_sStart = plane2D.angleContainsPin(self.localCenter, other.localStart, self.localStart, other.localEnd)
        other_sEnd = plane2D.angleContainsPin(self.localCenter, other.localStart, self.localEnd, other.localEnd)
        
        if self_oStart == 1 and self_oEnd == 1 and other_sStart == 1 and other_sEnd == 1:
            return result
            
        if (self.localEnd == other.localStart and self_oEnd != -1) or (self.localStart == other.localEnd and self_oStart != -1):
            return result
        
        if self_oStart == -1 and self_oEnd == -1 and other_sStart == -1 and other_sEnd == -1:
            pin = self.position(plane2D.findPin(self.localCenter, self.localStart, other.localEnd))
            result.append(ellipticalArc3D(self.sort, self.foci[0], self.foci[1], self.start,  other.end, pin, self.quadPoint))
            pin = self.position(plane2D.findPin(self.localCenter, other.localStart, self.localEnd))
            result.append(ellipticalArc3D(self.sort, self.foci[0], self.foci[1],  other.start , self.end, pin, self.quadPoint))
            if self.sort.type == 'sort-attribute':
                for i in result:
                    i.addAttribute(self.attribute)
            return result
            
        if self_oStart == -1:
            start = other.localStart
        else:
            start = self.localStart
        if self_oEnd == -1:
            end = other.localEnd
        else:
            end = self.localEnd
        
        pin = self.position(plane2D.findPin(self.localCenter, start, end))
        
        result.append(ellipticalArc3D(self.sort, self.foci[0], self.foci[1], self.position(start), self.position(end), pin, self.quadPoint))
        if self.sort.type == 'sort-attribute':
            result[0].addAttribute(self.attribute)
        return result
        
    def combine(self, other):
        """
        Find the combination of two arcs, The two must be on the same ellipse
        
        :param other: other arc to find combination of
        :type other: ~sortal.ind.ellipticalArc.ellipticalArc3D
        :returns: :class:`~sortal.ind.ellipticalArc.ellipticalArc3D` -- The combination elliptical arc
        """
        if self.fullCircle:
            return False
        if other.fullCircle:
            self.localStart = other.localStart
            self.localEnd = other.localEnd
            self.localPin = other.localPin
            self.start = other.start
            self.end = other.end
            self.pin = other.pin
            return True
            
        self_oStart = plane2D.angleContainsPin(self.localCenter, self.localStart, other.localStart, self.localEnd)
        self_oEnd = plane2D.angleContainsPin(self.localCenter, self.localStart, other.localEnd, self.localEnd)
        other_sStart = plane2D.angleContainsPin(self.localCenter, other.localStart, self.localStart, other.localEnd)
        other_sEnd = plane2D.angleContainsPin(self.localCenter, other.localStart, self.localEnd, other.localEnd)
        
        if self_oStart == 1 and self_oEnd == 1 and other_sStart == 1 and other_sEnd == 1:
            return False
            
        if self_oStart != 1 and self_oEnd != 1 and other_sStart != 1 and other_sEnd != 1:
            if self.localStart == other.localStart and self.localEnd == other.localEnd:
                return False
            if self.localPin.colinear(self.localStart, self.localCenter):
                self.localPin = self.localEnd
                if self.localPin.colinear(self.localStart, self.localCenter):
                    self.localPin = plane2D.findPin(self.localCenter, self.localStart, self.localEnd)
            self.pin = self.position(self.localPin)
            self.localEnd = self.localStart
            self.end = self.start
            self.fullCircle = True
            return True
            
        if other_sStart < 1:
            self.localStart = other.localStart
            self.start = other.start
        if other_sEnd < 1:
            self.localEnd = other.localEnd
            self.end = other.end
                        
        self.localPin = plane2D.findPin(self.localCenter, self.localStart, self.localEnd)
        self.pin = self.position(self.localPin)
        return True
        
    def complement(self,other):
        """
        Find the complement of this arc with another, The two must be on the same ellipse
        
        :param other: other arc to find complement with
        :type other: ~sortal.ind.ellipticalArc.ellipticalArc3D
        :returns: list of :class:`~sortal.ind.ellipticalArc.ellipticalArc3D` -- The list of complement elliptical arcs
        """
        result = []
        if other.fullCircle:
            return result
        if self.fullCircle:
            new = ellipticalArc3D(self.sort, other.foci[0], other.foci[1], other.end, other.start, other.pin, self.quadPoint)
            new.addAttribute(self.attribute)            
            result.append(new)
            return result
            
        self_oStart = plane2D.angleContainsPin(self.localCenter, self.localStart, other.localStart, self.localEnd)
        self_oEnd = plane2D.angleContainsPin(self.localCenter, self.localStart, other.localEnd, self.localEnd)
        other_sStart = plane2D.angleContainsPin(self.localCenter, other.localStart, self.localStart, other.localEnd)
        other_sEnd = plane2D.angleContainsPin(self.localCenter, other.localStart, self.localEnd, other.localEnd)


        if self_oStart != 1 and self_oEnd != 1 and other_sStart != 1 and other_sEnd != 1:
            pin = self.position(plane2D.findPin(self.localCenter, other.localEnd, other.localStart))
            new = ellipticalArc3D(self.sort, self.foci[0], self.foci[1], other.end, other.start, pin, self.quadPoint)
            result.append(new)
            
        else:
            if self_oStart == -1:
                pin = self.position(plane2D.findPin(self.localCenter, self.localStart, other.localStart))
                new = ellipticalArc3D(self.sort, self.foci[0], self.foci[1], self.start, other.start, pin, self.quadPoint)
                result.append(new)
            if self_oEnd == -1:
                pin = self.position(plane2D.findPin(self.localCenter, other.localEnd, self.localEnd))
                new = ellipticalArc3D(self.sort, self.foci[0], self.foci[1], other.end, self.end, pin, self.quadPoint)
                result.append(new)
            
        if self.sort.type == 'sort-attribute':
            for i in result:
                i.addAttribute(self.attribute)
        return result
        
    def intersect(self, other):
        raise NotImplementedError
        
    def midpoint(self):
        raise NotImplementedError
        
    def borderPoint(self):
        """
        border point
        """
        a = vector2D(1,0) 
        b = vector2D(0,1)
        foci0 = self.localFoci[0].subtract(self.localCenter)
        foci0Sqrt = foci0.dotproduct(foci0).sqrt()
        cosAlpha = foci0.dotproduct(a)/foci0Sqrt
        sinAlpha = foci0.dotproduct(b)/foci0Sqrt
        sq = (cosAlpha/self.radii[0])**2 + (sinAlpha/self.radii[1])**2
        return vector2D(1/(sq.sqrt()) + self.localCenter.x, self.localCenter.y)
        
    def splitArc(self):
        """
        Split the arc according to the reference-point: the point on the ellipse that is horizontally to
        the right of the ellipse center. If the arc contains the reference-point, it will be split to 
        become two arcs:

        * One arc will start with reference-point, end with the original end-point
        * Another arc will start with the original start point, end with the reference-point

        This splitting is performed for all arcs that are put into cyclical forms during so that they can be
        ordered according to their start points and end points more effectively without overlapping.

        :returns: list of :class:`~sortal.ind.ellipticalArc.ellipticalArc3D` -- The list of the two splitted elliptical arcs, None if the arc cannot be splitted
        """
        ref = self.localCenter.add(vector2D(1,0))
        if plane2D.angleContainsPin(self.localCenter, self.localStart, ref, self.localEnd) == -1:
            ref = self.borderPoint()
            
            pin = self.position(plane2D.findPin(self.localCenter, ref, self.localEnd))
            a = ellipticalArc3D(self.sort, self.foci[0], self.foci[1], 
                              self.position(ref), self.end, pin, self.quadPoint)
                              
            pin = self.position(plane2D.findPin(self.localCenter, self.localStart, ref))
            b = ellipticalArc3D(self.sort, self.foci[0], self.foci[1], 
                              self.start, self.position(ref), pin, self.quadPoint)
                              
            if self.hasAttribute():
                a.addAttribute(self.attribute)
                b.addAttribute(self.attribute)
            return [a,b]
        return None
        
    def crossingOver(self):
        """
        Check if the arc crosses over the reference-point.

        :returns: boolean
        """
        return plane2D.angleContainsPin(self.localCenter, self.localStart, 
                                       self.localCenter.add(vector2D(1,0)), 
                                       self.localEnd)
        
    def replicateInd(self, sort = 0):
        """
        Create a duplicate of the individual without attributes
        
        :returns: the duplicate object
        """
        if sort == 0:
            sort = self.sort
        return ellipticalArc3D(sort, self.foci[0], self.foci[1], self.start, self.end, self.pin, self.quadPoint) 
        
    def codescriptor(self):
        """
        Codescriptor of the arc: the 3D ellipse containing this arc

        :returns: sortal.ellipse3D
        """
        ellipse = ellipse3D(self.sort, self.foci[0], self.foci[1], self.start, self.pin)
        ellipse.assoc = self
        return ellipse

    def transformInd(self, transformSet):
        """
        Transforming the individual with a transformation set (editor)

        :param transformSet: the transformation set to be applied
        :type transformSet: ~sortal.transform.transformSet.transformSet
        :returns: the transformed object
        """
        affine = transformSet.retrieveTransform(affineTransform3D)
        if affine is None:
            return self.duplicateInd()
        start = self.start.transform(affine)
        end = self.end.transform(affine)         
        pin = self.pin.transform(affine)
        
        focus1 = self.foci[0].transform(affine)         
        focus2 = self.foci[1].transform(affine)  
        
        if self.negNormal is True:
            temp = end
            end = start
            start = temp
                
        return ellipticalArc3D(self.sort, focus1, focus2, start, end, pin, self.quadPoint)
          
    def checkTransform(self, transform, compareForm):
        """
        Check if the individual belongs to the compareForm after transformation with a transform set
        """
        affine = transform.retrieveTransform(affineTransform3D)
        if affine is None:
            tfArc = self
        else:
            tfArc = self.transformInd(transform)
            if self.hasAttribute():
                tfArc.addAttribute(self.attribute)
                
        splits = tfArc.splitArc()
        if splits is None:
            if not compareForm.containsInd(tfArc):
                return False
        else:
            for arc in splits:
                if not compareForm.containsInd(arc):
                    return False
        return True
        
    def containsInd(self, ind):
        """
        Check if this individual contains another individual
        """
        if not self.contains(ind):
            return False
        if self.hasAttribute():
            if ind.hasAttribute():
                if ind.attribute.checkPartOf(self.attribute):
                    return True
                else:
                    return False
            else:
                return True
        return True
     
        
    @staticmethod
    def parse(reader, sort):
        """
        Parse the 3D elliptical arc from a sortal parseReader (from string)

        :param reader: the reader of the string to be parsed
        :type reader: ~sortal.io.parseReader.parseReader
        :param sort: sort of the new object
        :type sort: ~sortal.sort.sort.sort
        :returns: :class:`~sortal.ind.ellipticalArc.ellipticalArc3D` -- the individual parsed
        """
        if reader.token() != '<':
            raise ParsingError(reader,'expected a "<" at the beginning of creating an elliptical arc')
        if reader.newToken() != '(': #bracket
            raise ParsingError(reader,'expected a "(" ')
            
        x,y,z = vector3D.readVector(reader)
        focus1 = vector3D(x,y,z)
            
        reader.newToken() # bracket (current token would be ',')
        
        x,y,z = vector3D.readVector(reader)
        focus2 = vector3D(x,y,z)
            
        reader.newToken() # bracket (current token would be ',')
        
        x,y,z = vector3D.readVector(reader)
        start = vector3D(x,y,z)
            
        reader.newToken() # bracket (current token would be ',')
        
        x,y,z = vector3D.readVector(reader)
        end = vector3D(x,y,z)
            
        reader.newToken() # bracket (current token would be ',')
        
        x,y,z = vector3D.readVector(reader)
        pin = vector3D(x,y,z)
            
        ind = ellipticalArc3D(sort, focus1, focus2, start, end, pin)

        if reader.token() != '>':
            raise ParsingError(reader, '">" expected to close the line-segment')
            
        reader.newToken() #next token
        
        return ind
