# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import range
from builtins import super
from builtins import str

from .special import special

class enumerative(special):
    """
    enumerative class, adhering to its sort's parameter 
    (see :class:`~sortal.sort.parameter.enumeration.enumeration`).

    **Object Parameters:** (not including inherited attributes)

    * **value** (int) -- the index of the enumerative value in the sort's enumeration list

    **Class Properties:**

    * **Behavior**: :class:`~sortal.form.specialForm.specialForm`
    * **Sort Argument**: :class:`~sortal.sort.parameter.argument.argument.argumentType`.ENUMERATION -- :class:`~sortal.sort.parameter.enumeration.enumeration`
    * **Mapping**: :class:`~sortal.map.exactMap.exactMap`


    """
    type = 'ind-enumerative'
    def __init__(self, sort, value, *args, **kwargs):
        """
        Initializing a new enumerative.

        :param sort: sort of the description
        :type sort: ~sortal.sort.sort.sort
        :param value: the string name of the enumerative value. The enumerative value's index 
            value in the enumeration list is also accepted
        :type value: str or int

        """
        super(enumerative, self).__init__(sort)
        if isinstance(value, int):
            self.value = value
            return
        enums = self.sort.baseSort().arguments
        for i in range(len(enums.values)):
            if enums.values[i] == value:
                self.value = i
                return 
        raise ValueError('Argument is not a valid enumerative value')
    
        
    def strValue(self):
        """
        get the string value of the enumerative
        """
        return self.sort.baseSort().arguments.values[self.value]
        
    def ranking(self, other):
        """
        compare this enumerative with another using the ranking matrix
        """
        return self.sort.baseSort().arguments.rankings[self.value][other.value]
    
    def printVal(self):
        """
        Print out a string representing the value of this individual, used in form.printE()
        """
        return str(self.strValue())
        
    def getValue(self):
        """
        get the content of the individual (in string form)
        """
        return self.strValue()
        
    def isNil(self):
        """
        Check if the object value is invalid
        """
        if self.value < 0 or self.value >= len(self.sort.baseSort().arguments.values):
            return True
        return False
        
    def equals(self, other):
        """
        check equality
        """
        return self.value == other.value
        
    def compareValue(self, other):
        """
        Compare the value of two enumeratives
        """
        d = self.value - other.value
        if d < 0:
            return -1
        elif d > 0:
            return 1
        return 0
    
    def contains(self, other):
        """
        Check if this enumerative contains another
        """
        if self.sort != other.sort:
            raise ValueError('enumeratives must be of the same sort')
        return (not self.isNil()) and (not other.isNil()) and self.compareValue(other) == 0
        
    def combine(self, other):
        """
        combine two enumeratives ordinally using the ranking matrix
        """
        if other is None or other.isNil():
            return not self.isNil()
        if self.isNil():
            self.value = other.value
        else:
            self.value = self.ranking(other)
        return True
        
    def common(self, other):
        """
        Find the common enumerative between two enumeratives ordinally using the ranking matrix
        """
        if self.isNil() or other == None or other.isNil():
            return False
        return self.compareValue(other) == 0
        
    def complement(self, other):
        """
        Find the complement enumerative between this and another enumerative ordinally using the ranking matrix
        """
        if self.isNil():
            return False
        if other is None or other.isNil():
            return True
        return self.compareValue(other) != 0
        
    def getMinInd(self):
        return enumerative(self.sort)
        
    def symcomplement(self, other):
        if self.isNil():
            return False
        if other is None or other.isNil():
            return True
        return self.compareValue(other) != 0
        
    def replicateInd(self):
        """
        Create a duplicate of the individual without attributes
        
        :returns: the duplicate object
        """
        return enumerative(self.sort, self.value)

    @staticmethod
    def parse(reader, sort):
        """
        Parse the enumerative from a sortal parseReader (from string)

        :param reader: the reader of the string to be parsed
        :type reader: ~sortal.io.parseReader.parseReader
        :param sort: sort of the new object
        :type sort: ~sortal.sort.sort.sort
        :returns: :class:`~sortal.ind.enumerative.enumerative` -- the individual parsed
        """
        ind = enumerative(sort, int(reader.value()))
        reader.newToken()
        return ind