# -*- coding: utf-8 -*-

"""
An individual is the representational unit of a {Sort}.
An individual specifies a data element with a single value;
individuals may be collected into a :class:`~sortal.form.form.form`.
The value an individual may hold is dependent on the sort it belongs to.
Every {PrimitiveSort} is defined by its characteristic individual,
i.e., a representational abstraction of all individuals of the sort,
and by a description of the algebraic behavior of its forms. The algebraic
behavior of forms of individuals is derived from a partial order relationship
defined on the individuals and forms of a sort. This partial order
relationship defines the following operations and relations on individuals:
an individual may contain another individual, this is always the case
if they are equal; otherwise, two individuals may be disjoint; if two
individuals are not disjoint, they combine into another individual
that contains both individuals; two individuals that are not disjoint may
have a common part that is contained by both individuals; and an
individual may have a complement part wrt another individual, this is
always the case of the other individual does not contain this individual.
An individual may have an associate individual (under the object-attribute
relationship). This is the associate of the form this individual belongs to.

The Individual class extends on the Element class. It is an
abstract class specifying the common framework for all characteristic
individual classes. An instance of any characteristic individual class
defines an individual. Every individual belongs to a {PrimitiveSort};
the individual is necessarily an instance of the sort's characteristic
individual class. An individual is an Element with an optional
reference {Key} and an optional attribute Form.
The reference key serves to relate individuals. The optional attribute form
must be consistent with the sort this individual belongs to.
"""

from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import
from builtins import super
from builtins import str

from copy import deepcopy, copy

from ..element import element
from ..sort.attributeSort import attributeSort
from ..struct.planeStruct import PlaneTransformationError
from ..transform.transformSet import transformSet

class individual(element):
    """
    Individual class.

    **Object Parameters:** (not including inherited attributes)

    * **attribute** (:class:`~sortal.form.form.form`) -- attributes of the individual
    * **key** (:class:`~sortal.struct.key.key`) -- key/tag associated with the individual
    """
    
    def __init__(self, sort = None, *args, **kwargs):
        """
        Initialize individual. This function is called for all individual initializations.
        Create an attribute form if the individual's sort is an attribute sort.

        :param sort: sort of the individual
        :type sort: ~sortal.sort.sort.sort
        """
        super(individual, self).__init__(sort)
        if sort is not None and sort.type == 'sort-attribute':
            self.attribute = self.sort.weight.newForm()
            self.attribute.assoc = self
        else: 
            self.attribute = None
        self.key = None

    def __str__(self):
        """
        String format of the individual (with no attribute)
        """
        return self.toString(False)
        
    def toString(self, printKey):
        """
        String format of the individual (with no attribute)

        :param printKey: whether the tag of the individual is to be added to the string
        :type printKey: boolean
        :returns: str -- string format of the individual
        """
        if printKey and self.key is not None:
            return str(self.key)+' '+str(self.value)
        return str(self.value)
        
    def printE(self, string = ''):
        """
        Print the individual and its attribute

        :param string: a string to be added at the start of each printed line
        :type string: str
        """
        print(string, end = ' _ ')
        print(self.printVal())
        if self.hasAttribute():
            self.attribute.printE(string+'  ')
            
    def printRhino(self, dX = 0, dY = 0):
        """
        print the individual in a format that can be pasted into Rhinoceros for visualization
        """
        return None
        
    def geometricFrame(self):
        """
        Geometric frame of the individual
        """
        return None

        
    def getValue(self):
        return self.value

        
    def hasAttribute(self):
        """
        Check if an individual has attribute

        :returns: boolean -- whether the individual has attribute
        """
        return self.attribute is not None and not self.attribute.isEmpty()
        
    def getAttribute(self):
        """
        Get the attribute of the individual

        :returns: :class:`~sortal.form.form.form` -- attribute form of the individual
        """
        if (self.attribute is None) and isinstance(self.sort, attributeSort):
            print('ok')
            self.attribute = self.sort.weight.newForm()
            self.attribute.assoc = self
        else:
            print('not ok')
        return self.attribute
        
    def addAttribute(self, element):
        """
        Add an element as attribute to this individual. Must adhere to the sort of the individual attribute

        :param element: element to be added as attribute
        :type element: ~sortal.element
        """
        if 'form' in element.type:
            for i in element:
                if i is None:
                    pass
                elif 'form' in i.type:
                    self.addAttribute(i)
                elif not i.isNil():
                    self.attribute.insert(i)
        elif 'ind' in element.type:
            self.attribute.insert(element)
        
    def setAttribute(self, element):
        """
        Deprecated
        """
        raise NotImplementedError
        if self.sort.type == 'sort-primitive':
            if self.assoc is not None:
                print("Attribute form cannot be set if individual belongs to a form")
                return
            #
        else: 
            #if isinstance(element.sort, recursiveSort):
            #   sort = sort.instance()
            if not element.sort == self.sort.weight:
                print("Attribute of erroneous sort")
                return
        if isinstance(element, individual):
            self.attribute = self.sort.newForm()
            self.attribute.add
        element.assoc = self
        self.attribute = element
        return 

    def equals(self, other):
        """
        Check equality between two individuals

        :param other: other individual to be compared
        :type other: ~sortal.ind.individual.individual
        :returns: boolean -- whether two individuals are equal
        """
        if not self.value == other.value:
            return False
        return True
        
    def isNil(self):
        """
        Check if individual is Nil
        """
        return self.value is None
        
    def map(self, assignment, other):
        """
        Mapping a shape object to this individual (lhs object) for match detection
        """
        count = 0
        if self.assoc.hasAttribute():
            for i in self.assoc.attribute:
                if i.map(assignment, other):
                    count+=1
        if count > 0:
            return True
        return False 
    
    def compareValue(self, other):
        """
        Comparing value of 2 individuals. To be implemented in each of the individual class
        """
        raise NotImplementedError 
        
    def duplicateInd(self, sort = None):
        """
        Create a duplicate of the individual (with attribute). If a sort is not given, use the same sort as this individual
        """
        duplicate =self.replicateInd()
        if sort is not None:
            duplicate.sort = sort
        if self.hasAttribute():
            duplicate.addAttribute(self.attribute)
        duplicate.key = self.key
        return duplicate
        
    def replicateInd(self):
        """
        Create a duplicate of the individual without attributes
        
        :returns: the duplicate object
        """
        print(self.__class__)
        raise NotImplementedError 
        
    def updateAssoc(self):
        """
        Update the attribute associate individual to be this individual.
        """
        if self.hasAttribute():
            self.attribute.assoc = self
            self.attribute.updateAssoc()
        
    def accept(self, visitor, direction = -1):
        """
        Check the individual for rule analysis and rule detection
        """
        if self.isNil():
            return
        visitor.visitEnter(self, direction)
        if self.hasAttribute():
            self.attribute.accept(visitor, direction)
        visitor.visitLeave(self)
    
    def codescriptor(self) :
        """
        Return the codescriptor of the individual, usually the individual itself 
        except for line-segment which has the line containing it as codescriptor
        """
        codesc = self.duplicateInd()
        codesc.assoc = self
        return codesc
    
    def edit(self, editor):
        """
        Transforming the individual with a transformation set (editor)

        :param editor: the transformation set to be applied
        :type editor: ~sortal.transform.transformSet.transformSet
        :returns: the transformed object
        """
        if self.isNil():
            return self.duplicateInd()
        if editor is not None:
            result = self.editInd(editor)
            if result is None:
                return None
            if self.hasAttribute():
                result.sort = self.sort
                att = self.attribute.edit(editor)
                result.addAttribute(att)
        else:
            result = self.duplicateInd()
            if self.hasAttribute():
                attr = self.attribute.edit(None)
                attr.assoc = result
                result.addAttribute(attr)
        if self.key is not None:
            result.key = self.key
        return result
        
    def editInd(self, editor):
        """
        Transforming the individual with a transformation set (editor)

        :param editor: the transformation set or paramSet (description transformation set) to be applied
        :type editor: ~sortal.transform.transformSet.transformSet or sortal.paramSet
        :returns: the transformed object
        """
        if editor.type == 'transformSet':
            newInd = self.transformInd(editor)
            if newInd is not None:
                newInd.key = None
            return newInd
        elif editor.type == 'paramSet':
            return self.evaluateInd(editor)
        elif editor.type == 'transform-graphTransform3D' or editor.type == 'transform-graphTransform2D':
            trSet = transformSet()
            trSet.add(editor)
            newInd = self.transformInd(trSet)
            if newInd is not None:
                newInd.key = None
            return newInd
        return self.replicateInd()

    def transformInd(self, transformSet):
        """
        Transforming the individual with a transformation set (editor)

        :param transformSet: the transformation set to be applied
        :type transformSet: ~sortal.transform.transformSet.transformSet
        :returns: the transformed object
        """

        return copy(self)
        
    def evaluateInd(self, paramSet):
        """
        Evaluate all the description within the individual with the paramSet
        """
        return self.replicateInd()
        
    def checkTransform(self, transform, compareForm):
        """
        Check if the individual belongs to the compareForm after transformation with a transform set
        """
        try:
            comp = self.edit(transform)
        except PlaneTransformationError:
            return False
        except ZeroDivisionError:
            return True
        if comp.isNil():
            return False
        if compareForm.containsInd(comp):
            return True
        return False
        
    def containsInd(self, ind):
        """
        Check if this individual contains another individual
        """
        if self != ind:
            return False
        if self.sort.type == 'sort-attribute':
            # if not self.hasAttribute():
            #     if ind.hasAttribute():
            #         return False
            #     else:
            #         return True
            if ind.hasAttribute():
                if ind.attribute.checkPartOf(self.attribute):
                    return True
                else:
                    return False
            else:
                return True
        return True
        
    def clear(self):
        """
        Clear the attribute of the individual
        """
        if self.attribute is not None:
            self.attribute.clear()
            
    @staticmethod
    def parse(reader, sort):
        """
        Parse the individual from a sortal parseReader (from string)

        :param reader: the reader of the string to be parsed
        :type reader: ~sortal.io.parseReader.parseReader
        :param sort: sort of the new object
        :type sort: ~sortal.sort.sort.sort
        :returns: :class:`~sortal.ind.individual.individual` -- the individual parsed
        """
        raise NotImplementedError

    
