# -*- coding: utf-8 -*-
"""
A line is a linear, connected, non-bounded planar curve with a
discrete behavior.

The Line class extends on the Individual class.
It defines the characteristic individual for lines. A line is represented as
a direction vector and a position vector specifying the root of the line.
This characteristic individual accepts no parameters.
It specifies a {sortal.map.similarityMap} or {sortal.map.similarityMap3D} as default.
Forms of lines adhere to a discrete behavior.
"""

from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import absolute_import
from builtins import super
from builtins import str

from .individual import individual
from ..transform.affineTransform import affineTransform2D, affineTransform3D
from ..transform.graphTransform import graphTransform, graphTransform3D
from ..struct.vector import vector2D, vector3D
from ..struct.coordinate import c0
from ..io.parseReader import ParsingError
class line2D(individual):
    """
    Line in 2D space class.

    A line is a linear, connected, non-bounded planar curve with a
    discrete behavior.

    The line2D class extends on the individual class.
    It defines the characteristic individual for lines. A line is represented as
    a direction vector and a position vector specifying the root of the line.

    **Object Parameters:** (not including inherited attributes)

    * **direction** (:class:`~sortal.struct.vector.vector2D`) -- direction vector of the line.
    * **root** (:class:`~sortal.struct.vector.vector2D`) -- root vector of the line - projection of (0,0) onto the line.
    * **unitDir** (:class:`~sortal.struct.vector.vector2D`) -- unit direction vector of the line.

    **Class Properties:**

    * **Behavior**: :class:`~sortal.form.discrete.discrete`
    * **Sort Argument**: :class:`~sortal.sort.parameter.argument.argument.argumentType`.NONE
    * **Mapping**: :class:`~sortal.map.similarityMap.similarityMap2D`
    """

    type = 'ind-line2D'
    
    def __init__(self, sort, start, end, *args, **kwargs):
        """
        Initialize a line
        2 points (start and end) are required. the direction of the line will be calculated by
        having end subtracting start (the direction's vector value must always be positive)
        the unit direction is also calculated by dividing direction by its length
        root is the root vector of the line

        :param sort: sort of the line
        :type sort: ~sortal.sort.sort.sort
        :param start: first point on the line
        :type start: ~sortal.struct.vector.vector2D
        :param end: second point on the line
        :type end: ~sortal.struct.vector.vector2D
        """
        super(line2D, self).__init__(sort)
        if end.type == 'ind-point2D':
            end = end.value
        if start.type == 'ind-point2D':
            start = start.value
        self.oDirection = end.subtract(start)
        if end.compareValue(start) == -1:
            start, end = end, start
        self.direction = end.subtract(start)
        self.unitDir = self.direction.unit()
        self.root = start.subtract(self.direction.scale(start))
        
    def toString(self, printKey):
        """
        String format of the individual (with no attribute)

        :param printKey: whether the tag of the individual is to be added to the string
        :type printKey: boolean
        :returns: str -- string format of the individual
        """
        s = ''
        if printKey and self.key is not None:
            s += str(self.key)+' '
        #return s+ '<*'+str(self.root)+','+str(self.direction)+'*>'
        return s+'<*'+str(self.root)+','+str(self.root.add(self.direction))+'*>'
        #return '<'+str(self.root)+','+str(self.root.add(self.direction))+'>'
                
    def getValue(self):
        return str(self)

    def isNil(self):
        """
        Check if the line has no direction
        """
        return self.direction.x  == c0 and self.direction.y  == c0

    def printVal(self):
        """
        Print out a string representing the value of this individual, used in form.printE()
        """
        return ('direction: '+str(self.direction)+ ', unitDir: '+
                str(self.unitDir) + ', root: '+str(self.root))

    def printRhino(self, dX = 0, dY = 0):
        """
        Print the individual in a format that can be pasted into Rhinoceros for visualization
        """
        start = self.root.add(self.direction.scale(-10))
        end = self.root.add(self.direction.scale(10))
        return '_Line '+str(start.x + dX)+','+str(start.y + dY)+' '+str(end.x + dX)+','+str(end.y + dY)
        
    def equals(self, other):
        """
        Check equality between two lines

        :param other: other individual to be compared
        :type other: ~sortal.ind.line.line2D
        :returns: boolean -- whether two individuals are equal
        """
        return (self.unitDir == other.unitDir) and (self.root == other.root)
 
    def containsPoint(self, point):
        """
        Check if the line contains a specific point

        :param point: point to be checked
        :type point: ~sortal.struct.vector.vector2D
        :returns: boolean -- whether the line contains the point
        """
        return point.subtract(self.root).parallel(self.direction)
        
    def compareValue(self, other):
        """
        Comparing value of 2 lines.
        First compare

        :param other: other line to be compared
        :type other: ~sortal.ind.line.line2D
        :returns: int -- 0 if equals, 1 if this line is bigger, -1 if smaller
        """
        c = self.unitDir.compareValue(other.unitDir) 
        if c != 0:
            return c
        return self.root.compareValue(other.root) # c = self.root.compareValue(other.root)

    def projection(self, point):
        """
        Find the projection of a point on the line

        :param point: coordinates of the point to find projection of
        :type other: ~sortal.struct.vector.vector2D
        :returns: :class:`~sortal.struct.vector.vector2D` -- projection of the input point on the line
        """
        v = point.subtract(self.root)
        return self.root.add(self.direction.scale(v))


    def parallel(self, other):
        """
        Check if two lines are parallel

        :param other: other line to be checked
        :type other: ~sortal.ind.line.line2D
        :returns: boolean -- whether the lines are parallel
        """
        if other is None or 'ind-line' not in other.type :
            return False
        #return self.unitDir == other.unitDir and self.root != other.root
        return self.direction.parallel(other.direction)

    def perpendicular(self,other):
        """
        Check if two lines are perpendicular

        :param other: other line to be checked
        :type other: ~sortal.ind.line.line2D
        :returns: boolean -- whether the lines are perpendicular
        """
        if other is None or 'ind-line' not in other.type :
            return False
        return self.direction.perpendicular(other.direction)
        
    def intersect(self, other):
        """
        Check if two lines intersect

        :param other: other line to be checked
        :type other: ~sortal.ind.line.line2D
        :returns: boolean -- whether the lines intersect
        """
        if other is None or 'ind-line' not in other.type :
            return False
        return not self.direction.parallel(other.direction)
     
    def intersection(self, other):
        """
        Find the intersection of two lines

        :param other: other line to find intersection
        :type other: ~sortal.ind.line.line2D
        :returns: boolean -- whether the lines intersect
        """
        if not self.intersect(other):
            #print(self.printRhino(), other.printRhino(), self.intersect(other))
            raise ValueError('the two lines are not intersecting')
        if self.direction.x  == c0:
            a2 = other.direction.y / other.direction.x
            x = self.root.x
            b2 = other.root.y - a2*other.root.x
            y = a2*x+b2
        elif other.direction.x  == c0:
            a1 = self.direction.y / self.direction.x
            x = other.root.x
            b1 = self.root.y - a1*self.root.x
            y = a1*x+b1
        else:
            a1 = self.direction.y / self.direction.x
            a2 = other.direction.y / other.direction.x
            b1 = self.root.y - a1*self.root.x
            b2 = other.root.y - a2*other.root.x
            x = (b2-b1)/(a1-a2)
            y = a1*x+b1
        return vector2D(x,y)
        
    def replicateInd(self):
        """
        Create a duplicate of the individual without attributes
        
        :returns: the duplicate object
        """
        return line2D(self.sort, self.root.add(self.direction), self.root)
        
        
    def transformInd(self, transformSet):
        """
        Transforming the individual with a transformation set (editor)

        :param transformSet: the transformation set to be applied
        :type transformSet: ~sortal.transform.transformSet.transformSet
        :returns: the transformed object
        """
        affine = transformSet.retrieveTransform(affineTransform2D)
        if affine is None:
            newLine = self.duplicateInd()
        else:
            start = self.root.transform(affine)
            end = self.root.add(self.direction).transform(affine)
            newLine = line2D(self.sort, start, end)
        transf = transformSet.retrieveTransform(graphTransform)
        if transf is not None:
            vects = []
            for i in transf.vects:
                if self.containsPoint(i[0]):
                    vects.append(i[1])
                if len(vects) >=2 and vects[0] == vects[1]:
                    vects.pop()
                if len(vects) >=2:
                    break
            newLine = line2D(self.sort, vects[0], vects[1])
        return newLine

    def transformDirection(self, transformSet):
        """
        Find the transformed direction of a line with a transformation set 
        (the resulting direction would be based on the pre-transformed line
        direction i.e. the resulting direction might not be pointing towards
        positive x/y axis)

        :param transformSet: the transformation set to be applied
        :type transformSet: ~sortal.transform.transformSet.transformSet
        :returns: the transformed object
        """
        affine = transformSet.retrieveTransform(affineTransform2D)
        if affine is not None:
            start = self.root.transform(affine)
            end = self.root.add(self.direction).transform(affine)
            newDirection = end.subtract(start)
            return newDirection
        transf = transformSet.retrieveTransform(graphTransform)
        if transf is not None:
            vects = []
            for i in transf.vects:
                if self.containsPoint(i[0]):
                    vects.append(i)
                if len(vects) >=2 and vects[0] == vects[1]:
                    vects.pop()
                if len(vects) >=2:
                    break
            if (vects[0][0] < vects[1][0]):
                return vects[1][1].subtract(vects[0][1])
            return vects[0][1].subtract(vects[1][1])
        raise ValueError('Unable to find transformed direction')
        
    @staticmethod
    def parse(reader, sort):
        """
        Parse the 2D line a sortal parseReader (from string)

        :param reader: the reader of the string to be parsed
        :type reader: ~sortal.io.parseReader.parseReader
        :param sort: sort of the new object
        :type sort: ~sortal.sort.sort.sort
        :returns: :class:`~sortal.ind.line.line2D` -- the individual parsed
        """
        if reader.token() != '<':
            raise ParsingError(reader,'expected a "<" at the beginning of creating a line-segment')
        if reader.newToken() != '(': #bracket
            raise ParsingError(reader,'expected a "(" ')
        x,y = vector2D.readVector(reader)
        start = vector2D(x,y)
            
        reader.newToken() # bracket (current token would be ',')
        
        x,y = vector2D.readVector(reader)
        end = vector2D(x,y)
            
        ind = line2D(sort, start, end)

        if reader.token() != '>':
            raise ParsingError(reader, '">" expected to close the line-segment')
            
        reader.newToken() #next token
        
        return ind
        
        

class line3D(individual):
    """
    Line in 3D space class

    A line is a linear, connected, non-bounded planar curve with a
    discrete behavior.

    The line3D class extends on the individual class.
    It defines the characteristic individual for lines. A line is represented as
    a direction vector and a position vector specifying the root of the line.

    **Object Parameters:** (not including inherited attributes)

    * **direction** (:class:`~sortal.struct.vector.vector3D`) -- direction vector of the line.
    * **root** (:class:`~sortal.struct.vector.vector3D`) -- root vector of the line - projection of (0,0,0) onto the line.
    * **unitDir** (:class:`~sortal.struct.vector.vector3D`) -- unit direction vector of the line.

    **Class Properties:**

    * **Behavior**: :class:`~sortal.form.discrete.discrete`
    * **Sort Argument**: :class:`~sortal.sort.parameter.argument.argument.argumentType`.NONE
    * **Mapping**: :class:`~sortal.map.similarityMap3D.similarityMap3D`
    """

    type = 'ind-line3D'
    
    def __init__(self, sort, start, end, *args, **kwargs):
        """
        Initialize a line
        2 points (start and end) are required. the direction of the line will be calculated by
        having end subtracting start (the direction's vector value must always be positive)
        the unit direction is also calculated by dividing direction by its length
        root is the root vector of the line

        :param sort: sort of the line
        :type sort: ~sortal.sort.sort.sort
        :param start: first point on the line
        :type start: ~sortal.struct.vector.vector3D
        :param end: second point on the line
        :type end: ~sortal.struct.vector.vector3D
        """
        super(line3D, self).__init__(sort)     

        if end.type == 'ind-point3D':
            end = end.value
        if start.type == 'ind-point3D':
            start = start.value
        self.oDirection = end.subtract(start)
        if end.compareValue(start) == -1:
            start, end = end, start
        self.direction = end.subtract(start)
        self.unitDir = self.direction.unit()
        self.root = start.subtract(self.direction.scale(start))
        
    def toString(self, printKey):
        """
        String format of the individual (with no attribute)

        :param printKey: whether the tag of the individual is to be added to the string
        :type printKey: boolean
        :returns: str -- string format of the individual
        """
        s = ''
        if printKey and self.key is not None:
            s += str(self.key)+' '
        return s+'<*'+str(self.root)+','+str(self.root.add(self.direction))+'*>'
                
    def getValue(self):
        return str(self)
        
    def isNil(self):
        """
        Check if the line has no direction
        """
        return self.direction.x  == c0 and self.direction.y  == c0 and self.direction.z  == c0
        
    def printVal(self):
        """
        Print out a string representing the value of this individual, used in form.printE()
        """
        return ('direction: '+str(self.direction)+ ', unitDir: '+
                str(self.unitDir) + ', root: '+str(self.root))
        
    def printRhino(self, dX = 0, dY = 0):
        """
        Print the individual in a format that can be pasted into Rhinoceros for visualization
        """
        start = self.root.add(self.direction.scale(-1000))
        end = self.root.add(self.direction.scale(1000))
        return '_Line '+str(start.x + dX)+','+str(start.y + dY)+','+str(start.z)+' '+str(end.x + dX)+','+str(end.y + dY)+','+str(end.z)
        

    def equals(self, other):
        """
        Check equality between two lines

        :param other: other individual to be compared
        :type other: ~sortal.ind.line.line3D
        :returns: boolean -- whether two individuals are equal
        """
        return (self.unitDir == other.unitDir) and (self.root == other.root)
            
    def compareValue(self, other):
        """
        Comparing value of 2 lines:

        First compare unit directions

        If unit directions are equal, compare roots

        :param other: other line to be compared
        :type other: ~sortal.ind.line.line3D
        :returns: int -- 0 if equals, 1 if this line is bigger, -1 if smaller
        """
        c = self.unitDir.compareValue(other.unitDir) 
        if c != 0:
            return c
        return self.root.compareValue(other.root) # c = self.root.compareValue(other.root)
        
    def containsPoint(self,point):
        """
        Check if the line contains a specific point

        :param point: point to be checked
        :type point: ~sortal.struct.vector.vector3D
        :returns: boolean -- whether the line contains the point
        """
        return point.subtract(self.root).parallel(self.direction)
        
    def projection(self, point):
        """
        Find the projection of a point on the line

        :param point: coordinates of the point to find projection of
        :type other: ~sortal.struct.vector.vector3D
        :returns: :class:`~sortal.struct.vector.vector3D` -- projection of the input point on the line
        """
        v = point.subtract(self.root)
        return self.root.add(self.direction.scale(v))

    def parallel(self, other):
        """
        Check if two lines are parallel

        :param other: other line to be checked
        :type other: ~sortal.ind.line.line3D
        :returns: boolean -- whether the lines are parallel
        """
        if other is None or 'ind-line' not in other.type :
            return False
        return self.direction.parallel(other.direction)
        

    def perpendicular(self, other):
        """
        Check if two lines are perpendicular

        :param other: other line to be checked
        :type other: ~sortal.ind.line.line3D
        :returns: boolean -- whether the lines are perpendicular
        """
        if other is None or 'ind-line' not in other.type :
            return False
        return self.direction.dotproduct(other.direction)  == c0
     
    def skew(self, other):
        """
        Check if two lines are skew

        :param other: other line to be checked
        :type other: ~sortal.ind.line.line3D
        :returns: boolean -- whether the lines are skew
        """
        if other is None or 'ind-line' not in other.type :
            return False
        return not self.root.subtract(other.root).dotproduct(self.direction.product(other.direction))  == c0

    def coPlanar(self, other):
        """
        Check if two lines are co-planar

        :param other: other line to be checked
        :type other: ~sortal.ind.line.line3D
        :returns: boolean -- whether the lines are co-planar
        """
        if other is None or 'ind-line' not in other.type :
            return False
        return self.direction.product(other.direction).dotproduct(self.root.subtract(other.root))  == c0
        
    def intersect(self, other):
        """
        Check if two lines intersect

        :param other: other line to be checked
        :type other: ~sortal.ind.line.line3D
        :returns: boolean -- whether the lines intersect
        """
        if other is None or 'ind-line' not in other.type :
            return False
        return self.coPlanar(other) and not self.parallel(other)
        
    def intersection(self, other):
        """
        Find the intersection of two lines

        :param other: other line to find intersection
        :type other: ~sortal.ind.line.line3D
        :returns: boolean -- whether the lines intersect
        """
        if not self.intersect(other):
            raise ValueError('the two lines are not intersecting')
        denom = self.unitDir.product(other.unitDir)
        diff = other.root.subtract(self.root)
        numer1 = diff.product(other.unitDir)
        
        if denom.x != 0:
            x = numer1.x/denom.x
        elif denom.y != 0:
            x = numer1.y/denom.y
        else:
            x = numer1.z/denom.z
        return self.root.add(self.unitDir.scale(x))
        
    def replicateInd(self):
        """
        Create a duplicate of the individual without attributes
        
        :returns: the duplicate object
        """
        return line3D(self.sort, self.root.add(self.direction), self.root)
        
    def transformInd(self, transformSet):
        """
        Transforming the individual with a transformation set (editor)

        :param transformSet: the transformation set to be applied
        :type transformSet: ~sortal.transform.transformSet.transformSet
        :returns: the transformed object
        """
        affine = transformSet.retrieveTransform(affineTransform3D)
        if affine is None:
            newLine = self.duplicateInd()
        else:
            start = self.root.transform(affine)
            end = self.root.add(self.direction).transform(affine)
            newLine = line3D(self.sort, start, end)
        transf = transformSet.retrieveTransform(graphTransform3D)
        if transf is not None:
            vects = []
            for i in transf.vects:
                if self.containsPoint(i[0]):
                    vects.append(i[1])
                if len(vects) >=2 and vects[0] == vects[1]:
                    vects.pop()
                if len(vects) >=2:
                    break
            newLine = self.__class__(self.sort, vects[0], vects[1])
        return newLine

    def transformDirection(self, transformSet):
        """
        Find the transformed direction of a line with a transformation set 
        (the resulting direction would be based on the pre-transformed line
        direction i.e. the resulting direction might not be pointing towards
        positive x/y axis)

        :param transformSet: the transformation set to be applied
        :type transformSet: ~sortal.transform.transformSet.transformSet
        :returns: the transformed object
        """
        affine = transformSet.retrieveTransform(affineTransform3D)
        if affine is not None:
            start = self.root.transform(affine)
            end = self.root.add(self.direction).transform(affine)
            newDirection = end.subtract(start)
            return newDirection
        transf = transformSet.retrieveTransform(graphTransform3D)
        if transf is not None:
            vects = []
            for i in transf.vects:
                if self.containsPoint(i[0]):
                    vects.append(i)
                if len(vects) >=2 and vects[0] == vects[1]:
                    vects.pop()
                if len(vects) >=2:
                    break
            if (vects[0][0] < vects[1][0]):
                return vects[1][1].subtract(vects[0][1])
            return vects[0][1].subtract(vects[1][1])
        raise ValueError('Unable to find transformed direction')

    @staticmethod
    def parse(reader, sort):
        """
        Parse the 3D line a sortal parseReader (from string)

        :param reader: the reader of the string to be parsed
        :type reader: ~sortal.io.parseReader.parseReader
        :param sort: sort of the new object
        :type sort: ~sortal.sort.sort.sort
        :returns: :class:`~sortal.ind.line.line3D` -- the individual parsed
        """
        if reader.token() != '<':
            raise ParsingError(reader,'expected a "<" at the beginning of creating a line-segment')
        if reader.newToken() != '(': #bracket
            raise ParsingError(reader,'expected a "(" ')
        x,y,z = vector3D.readVector(reader)
        start = vector3D(x,y,z)
            
        reader.newToken() # bracket (current token would be ',')
        
        x,y,z = vector3D.readVector(reader)
        end = vector3D(x,y,z)
            
        ind = line3D(sort, start, end)

        if reader.token() != '>':
            raise ParsingError(reader, '">" expected to close the line-segment')
            
        reader.newToken() #next token
        
        return ind
