# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import super
from builtins import str

from .line import line2D, line3D
from ..struct.coordinate import c0p5, c2
from ..struct.vector import vector2D, vector3D
from ..transform.affineTransform import affineTransform2D, affineTransform3D
from ..transform.graphTransform import graphTransform, graphTransform3D
from ..io.parseReader import ParsingError


class lineSegment2D(line2D):
    """
    line-segment in 2D class.

    A line-segment is a connected and bounded segment of a line with an
    interval behavior. The line defines the co-descriptor of the line-segment,
    the boundary of the segment is defined by the start and end positions of the
    line-segment.

    The LineSegment2D class extends on the Line2D class and
    implements the Interval interface.
    It defines the characteristic individual for line-segments.
    A line-segment is represented as a line with two rational
    scalars specifying the start and end relative to the line's root.

    **Object Parameters:** (not including inherited attributes)

    * **start** (:class:`~sortal.struct.vector.vector2D`) -- lower end-point of the line.
    * **end** (:class:`~sortal.struct.vector.vector2D`) -- higher end-point of the line.
    * **reverse** (boolean) -- whether the inputted start and end was swapped during initialization.

    **Class Properties:**

    * **Behavior**: :class:`~sortal.form.interval.interval`
    * **Sort Argument**: :class:`~sortal.sort.parameter.argument.argument.argumentType`.NONE
    * **Mapping**: :class:`~sortal.map.similarityMap.similarityMap2D`
    """
    
    type = 'ind-lineSegment2D'
    
    def __init__(self, sort, start, end, *args, **kwargs):
        """
        Initialize a line-segment using its two end-points.
        The lower end-point will be saved as self.start; the higher end-point will be saved as self.end.

        :param sort: sort of the line
        :type sort: ~sortal.sort.sort.sort
        :param start: first end-point of the line
        :type start: ~sortal.struct.vector.vector2D
        :param end: second end-point of the line
        :type end: ~sortal.struct.vector.vector2D
        """
        if end.type == 'ind-point2D':
            end = end.value
        if start.type == 'ind-point2D':
            start = start.value        
        self.oStart = start
        self.oEnd = end
        if end.compareValue(start) == -1:
            self.start = end
            self.end = start
        else:
        #elif end.compareValue(start) == 1:
            self.end = end
            self.start = start
        super(lineSegment2D, self).__init__(sort, self.oStart, self.oEnd)

    def getValue(self):
        return str(self)

    def toString(self, printKey):
        """
        string format of the line-segment (with no attribute)

        :param printKey: whether the tag of the individual is to be added to the string
        :type printKey: boolean
        :returns: str -- string format of the individual
        """
        s = ''
        if printKey and self.key is not None:
            s += str(self.key)+' '
        return s+'<'+str(self.start)+','+str(self.end)+'>'
                
    def printVal(self):
        """
        print out a string representing the value of this individual, used in form.printE()
        """
        return (str(self.start)+ ','+ str(self.end)) # + ', direction: ('+str(self.direction.x)+','+str(self.direction.y)+ '), unitDir: ('+str(self.unitDir.x)+','+str(self.unitDir.y)+ '), root: ('+str(self.root.x)+','+str(self.root.y)+ ')'

    def printRhino(self, dX = 0, dY = 0):
        """
        print the individual in a format that can be pasted into Rhinoceros for visualization
        """
        return '_Line '+str(self.start.x + dX)+','+str(self.start.y + dY)+'  '+str(self.end.x + dX)+','+str(self.end.y + dY)

    def equals(self, other):
        """
        check equality between two line-segments

        :param other: other individual to be compared
        :type other: ~sortal.ind.lineSegment.lineSegment2D
        :returns: boolean -- whether two individuals are equal
        """
        return (self.start == other.start) and (self.end == other.end)
        
    def isNil(self):
        """
        Check if the start and end point of the line are the same
        """
        if self.end is None or self.start is None:
            return True
        return self.end.x == self.start.x and self.end.y == self.start.y
        return self.direction.isZero()
        
    def compareInterval(self, other):
        """
        Compare the value of two line-segments:

        * First compare unit directions
        * If unit directions are equal, compare roots
        * If roots are equal, compare the start points
        * If starts points are equal, compare the end points

        # this function is named compareInterval and not compareValue since there
        is already a compareValue function in the line 

        :param other: other line-segment to be compared
        :type other: ~sortal.ind.lineSegment.lineSegment2D
        :returns: int -- 0 if equals, 1 if this line is bigger, -1 if smaller
        """
        c = self.unitDir.compareValue(other.unitDir) 
        if c != 0:
            return c
        c = self.root.compareValue(other.root) 
        if c != 0:
            return c
        c = self.start.compareValue(other.start) 
        if c != 0:
            return c
        c = self.end.compareValue(other.end) 
        if c != 0:
            return c
        return 0
        
    def containsPoint(self,point):
        """
        Check if the line-segment contains a specific point

        :param point: point to be checked
        :type point: ~sortal.struct.vector.vector2D
        :returns: boolean -- whether the line contains the point
        """
        if not point.colinear(self.start,self.end):
            return False
        if self.start.compareValue(point) < 1 and self.end.compareValue(point) > -1:
            return True
        return False
        
    def compareEnd(self, other):
        """
        Check if the line-segment contains a specific point

        :param point: point to be checked
        :type point: ~sortal.struct.vector.vector2D
        :returns: boolean -- whether the line contains the point
        """
        return self.end.compareValue(other.end) 
            
    def length(self):
        """
        Find the length of the line-segment

        :returns: :class:`~sortal.struct.coord.coordinate_float.coordinate` -- the length of the line-segment
        """
        lengthVector = self.end.subtract(self.start)
        return lengthVector.dotproduct(lengthVector) ** c0p5
        
    def squareLength(self):
        """
        Find the square length of the line-segment

        :returns: :class:`~sortal.struct.coord.coordinate_float.coordinate` -- the square length of the line-segment
        """
        lengthVector = self.end.subtract(self.start)
        return lengthVector.dotproduct(lengthVector)
        
    def touches(self,other):
        """
        Check if two line-segments touch one another

        :param other: other line to be checked
        :type other: ~sortal.ind.lineSegment.lineSegment2D
        :returns: boolean -- whether the two line-segments touch
        """
        return self.compareValue(other) == 0 and (self.end == other.start or self.start == other.end) 
    
    def contains(self,other):
        """
        Check if this line-segment contains the other line-segment

        :param other: other line to be checked
        :type other: ~sortal.ind.lineSegment.lineSegment2D
        :returns: boolean -- whether this line-segment contains the other
        """
        return self.compareValue(other) == 0 and (self.end.compareValue(other.end) >= 0 and self.start.compareValue(other.start) <= 0) 
    
    def disjoint(self,other):
        """
        Check if two line-segments are disjoint

        :param other: other line to be checked
        :type other: ~sortal.ind.lineSegment.lineSegment2D
        :returns: boolean -- whether the two line-segments are disjoint
        """
        return self.compareValue(other) == 0 and (self.end.compareValue(other.start) == -1 or self.start.compareValue(other.end) == 1)
    
    def align(self, other):
        """
        Check if two line-segments align

        :param other: other line to be checked
        :type other: ~sortal.ind.lineSegment.lineSegment2D
        :returns: boolean -- whether the two line-segments align
        """
        return self.compareValue(other) == 0 and (self.end == other.end or self.start == other.start)
        
    def common(self,other):
        """
        Find the common of two line-segment and modify this line to be the common segment
        
        :param other: other line to find common segment of
        :type other: ~sortal.ind.lineSegment.lineSegment2D
        :returns: boolean -- whether there is a common segment between two lines
        """
        if self.disjoint(other) or self.touches(other):
            return False
        
        new = self.duplicateInd(self.sort)
        if self.end.compareValue(other.end) == 1:
            new.end = other.end
        if self.start.compareValue(other.start) == -1:
            new.start = other.start
            
        return new
        
        if self.end.compareValue(other.end) == 1:
            self.end = other.end
        if self.start.compareValue(other.start) == -1:
            self.start = other.start
        return True
    
    
    def combine(self,other):
        """
        Find the combination of two line-segments
        
        :param other: other line to find common segment of two line-segments
        :type other: ~sortal.ind.lineSegment.lineSegment2D
        :returns: :class:`~sortal.ind.lineSegment.lineSegment2D` -- the combination of the two line-segments
        """
        if self.disjoint(other):
            return self
        nStart = self.start
        nEnd = self.end
        if self.end.compareValue(other.end) == -1:
            nEnd = other.end
        if self.start.compareValue(other.start) == 1:
            nStart = other.start
        return self.__class__(self.sort, nStart, nEnd)

    def complement(self,other):
        """
        Find the complement of this line-segment on the other line-segment
        
        :param other: other line to find complement segments
        :type other: ~sortal.ind.lineSegment.lineSegment2D
        :returns: list of :class:`~sortal.ind.lineSegment.lineSegment2D` -- the list of complement segment of this line-segment on the other
        """
        result = []
        if self.start.compareValue(other.start) == -1 and self.end.compareValue(other.start) == 1 :
            new = self.duplicateInd(self.sort)
            new.end = other.start
            result.append(new)
        if self.start.compareValue(other.end) == -1 and self.end.compareValue(other.end) == 1 :
            new = self.duplicateInd(self.sort)
            new.start = other.end
            result.append(new)
        return result
                
        
            
    @staticmethod
    def cw(A,B,C):
        """
        Check the clockwise property from point A to point B to point C

        :param A: first point to check clockwise property of
        :type A: ~sortal.struct.vector.vector2D
        :param B: second point to check clockwise property of
        :type B: ~sortal.struct.vector.vector2D
        :param C: third point to check clockwise property of
        :type C: ~sortal.struct.vector.vector2D
        :returns: boolean -- True if A-B-C is a in a counterclockwise order, False if A-B-C is in a clockwise order
        """
        return (C.y-A.y) * (B.x-A.x) > (B.y-A.y) * (C.x-A.x)
    
    def intersectSegment(self, other):
        """
        Check if two line-segments intersect and both line-segments contains the intersection

        :param other: other line to be checked
        :type other: ~sortal.ind.lineSegment.lineSegment2D
        :returns: boolean -- whether the two line-segments intersect
        """
        if other is None or type(other) != type(self) :
            return False
        if self.direction.parallel(other.direction):
            return False
        if ((self.start.colinear(other.start, other.end) and other.start < self.start < other.end)
        or (self.end.colinear(other.start, other.end) and other.start < self.end < other.end)
        or (other.start.colinear(self.start, self.end) and self.start < other.start < self.end)
        or (other.end.colinear(self.start, self.end) and self.start < other.end < self.end)):
            return True
        A = self.start
        B = self.end
        C = other.start
        D = other.end
        return self.cw(A,C,D) != self.cw(B,C,D) and self.cw(A,B,C) != self.cw(A,B,D)
    
        
    def midpoint(self):
        """
        Find the midpoint of a line-segment

        :returns: :class:`~sortal.struct.vector.vector2D` -- the coordinates of the mid-point of the line-segment
        """
        return (float(((self.start.x+self.end.x)/c2).value),float(((self.start.y+self.end.y)/c2).value))
        
    def replicateInd(self, sort = 0):
        """
        Create a duplicate of the individual without attributes
        
        :returns: the duplicate object
        """
        if sort == 0:
            sort = self.sort
        return self.__class__(sort, self.start, self.end)    
        
    def codescriptor(self):
        """
        Codescriptor of the line-segment: the infinite 2D line that defines this line-segment's root and direction

        :returns: :class:`~sortal.ind.line.line2D` -- the infinite 2D line that defines this line-segment
        """
        line = line2D(self.sort, self.start, self.end)
        line.assoc = self
        return line
        

    def transformInd(self, transformSet):
        """
        Transforming the individual with a transformation set (editor)

        :param transformSet: the transformation set to be applied
        :type transformSet: ~sortal.transform.transformSet.transformSet
        :returns: the transformed object
        """
        affine = transformSet.retrieveTransform(affineTransform2D)
        start = self.oStart
        end = self.oEnd
        try:
            if affine is None:
                newLine = self.duplicateInd()
            else:
                start = start.transform(affine)
                end = end.transform(affine)         
                newLine = self.__class__(self.sort, start, end)
            
            transf = transformSet.retrieveTransform(graphTransform)
            if transf is not None:
                start = transf.apply(start)
                end = transf.apply(end)
                newLine = self.__class__(self.sort, start, end)
        except ZeroDivisionError:
            return None
        return newLine
                
                
    def containsInd(self, ind):
        """
        Check if this individual contains another individual
        """
        if not self.contains(ind):
            return False
        if self.sort.type == 'sort-attribute':
            # if not self.hasAttribute():
            #     if ind.hasAttribute():
            #         return False
            #     else:
            #         return True
            if ind.hasAttribute():
                if ind.attribute.checkPartOf(self.attribute):
                    return True
                else:
                    return False
            else:
                return True
        return True
        
    @staticmethod
    def parse(reader, sort):
        """
        Parse the 2D line-segment from a sortal parseReader (from string)

        :param reader: the reader of the string to be parsed
        :type reader: ~sortal.io.parseReader.parseReader
        :param sort: sort of the new object
        :type sort: ~sortal.sort.sort.sort
        :returns: :class:`~sortal.ind.lineSegment.lineSegment2D` -- the individual parsed
        """
        if reader.token() != '<':
            raise ParsingError(reader,'expected a "<" at the beginning of creating a line-segment')
        if reader.newToken() != '(': #bracket
            raise ParsingError(reader,'expected a "(" ')
        x,y = vector2D.readVector(reader)
        start = vector2D(x,y)
            
        reader.newToken() # bracket (current token would be ',')
        
        x,y = vector2D.readVector(reader)
        end = vector2D(x,y)
            
        ind = lineSegment2D(sort, start, end)

        if reader.token() != '>':
            raise ParsingError(reader, '">" expected to close the line-segment')
            
        reader.newToken() #next token
        
        return ind

class lineSegment3D(line3D):
    """
    line-segment in 3D class.

    A line-segment is a connected and bounded segment of a line with an
    interval behavior. The line defines the co-descriptor of the line-segment,
    the boundary of the segment is defined by the start and end positions of the
    line-segment.

    The LineSegment3D class extends on the Line3D class and
    implements the Interval interface.
    It defines the characteristic individual for line-segments.
    A line-segment is represented as a line with two rational
    scalars specifying the start and end relative to the line's root.

    **Object Parameters:** (not including inherited attributes)

    * **start** (:class:`~sortal.struct.vector.vector3D`) -- lower end-point of the line.
    * **end** (:class:`~sortal.struct.vector.vector3D`) -- higher end-point of the line.
    * **reverse** (boolean) -- whether the inputted start and end was swapped during initialization.

    **Class Properties:**

    * **Behavior**: :class:`~sortal.form.interval.interval`
    * **Sort Argument**: :class:`~sortal.sort.parameter.argument.argument.argumentType`.NONE
    * **Mapping**: :class:`~sortal.map.similarityMap3D.similarityMap3D`

    """
    type = 'ind-lineSegment3D'
    def __init__(self, sort, start = vector3D(0,0,0), end = vector3D(0,0,0), *args, **kwargs):
        """
        Initialize a line-segment using its two end-points.
        The lower end-point will be saved as self.start; the higher end-point will be saved as self.end.

        :param sort: sort of the line
        :type sort: ~sortal.sort.sort.sort
        :param start: first end-point of the line
        :type start: ~sortal.struct.vector.vector3D
        :param end: second end-point of the line
        :type end: ~sortal.struct.vector.vector3D
        """
        if end.type == 'ind-point3D':
            end = end.value
        if start.type == 'ind-point3D':
            start = start.value
        self.oStart = start
        self.oEnd = end
        if end.compareValue(start) == -1:
            self.start = end
            self.end = start
        else:
        #elif end.compareValue(start) == 1:
            self.end = end
            self.start = start
        """
        else:
            raise ValueError('start point and end point of the line-segment are the same')
        """
        super(lineSegment3D, self).__init__(sort, self.oStart, self.oEnd)     

    def toString(self, printKey):
        """
        string format of the line-segment (with no attribute)

        :param printKey: whether the tag of the individual is to be added to the string
        :type printKey: boolean
        :returns: str -- string format of the individual
        """
        s = ''
        if printKey and self.key is not None:
            s += str(self.key)+' '
        return s+'<'+str(self.start)+','+str(self.end)+'>'
                    
    def getValue(self):
        return str(self)

    def printVal(self):
        """
        print out a string representing the value of this individual, used in form.printE()
        """
        return str(self.start)+ ','+ str(self.end) # + ', direction: ('+str(self.direction.x)+','+str(self.direction.y)+ '), unitDir: ('+str(self.unitDir.x)+','+str(self.unitDir.y)+ '), root: ('+str(self.root.x)+','+str(self.root.y)+ ')'

    def printRhino(self, dX = 0, dY = 0):
        """
        print the individual in a format that can be pasted into Rhinoceros for visualization
        """
        return '_Line '+str(self.start.x + dX)+','+str(self.start.y + dY)+','+str(self.start.z)+'  '+str(self.end.x + dX)+','+str(self.end.y + dY)+','+str(self.end.z)

    def equals(self, other):
        """
        check equality between two line-segments

        :param other: other individual to be compared
        :type other: ~sortal.ind.lineSegment.lineSegment3D
        :returns: boolean -- whether two individuals are equal
        """
        return (self.start == other.start) and (self.end == other.end)
                
    def isNil(self):
        """
        Check if the start and end point of the line are the same
        """
        if self.end is None or self.start is None:
            return True
        return self.direction.isZero()
        
    def compareInterval(self, other):
        """
        Compare the value of two line-segments:

        First compare unit directions

        If unit directions are equal, compare roots

        If roots are equal, compare the start points

        If starts points are equal, compare the end points

        # this function is named compareInterval and not compareValue since there
        is already a compareValue function in the line 

        :param other: other line-segment to be compared
        :type other: ~sortal.ind.lineSegment.lineSegment3D
        :returns: int -- 0 if equals, 1 if this line is bigger, -1 if smaller
        """
        c = self.unitDir.compareValue(other.unitDir) 
        if c != 0:
            return c
        c = self.root.compareValue(other.root) 
        if c != 0:
            return c
        c = self.start.compareValue(other.start) 
        if c != 0:
            return c
        c = self.end.compareValue(other.end) 
        if c != 0:
            return c
        return 0
        
    def containsPoint(self,point):
        """
        Check if the line-segment contains a specific point

        :param point: point to be checked
        :type point: ~sortal.struct.vector.vector3D
        :returns: boolean -- whether the line contains the point
        """
        if not point.colinear(self.start,self.end):
            return False
        if self.start.compareValue(point) < 1 and self.end.compareValue(point) > -1:
            return True
        return False
        
    def compareEnd(self, other):
        return self.end.compareValue(other.end) 
            
    def length(self):
        """
        Find the length of the line-segment

        :returns: :class:`~sortal.struct.coord.coordinate_float.coordinate` -- the length of the line-segment
        """
        lengthVector = self.end.subtract(self.start)
        return lengthVector.dotproduct(lengthVector) ** c0p5
        
    def squareLength(self):
        """
        Find the square length of the line-segment

        :returns: :class:`~sortal.struct.coord.coordinate_float.coordinate` -- the square length of the line-segment
        """
        lengthVector = self.end.subtract(self.start)
        return lengthVector.dotproduct(lengthVector)
        
    def touches(self,other):
        """
        Check if two line-segments touch one another

        :param other: other line to be checked
        :type other: ~sortal.ind.lineSegment.lineSegment3D
        :returns: boolean -- whether the two line-segments touch
        """
        return self.compareValue(other) == 0 and (self.end == other.start or self.start == other.end) 
    
    def contains(self,other):
        """
        Check if this line-segment contains the other line-segment

        :param other: other line to be checked
        :type other: ~sortal.ind.lineSegment.lineSegment3D
        :returns: boolean -- whether this line-segment contains the other
        """
        return self.compareValue(other) == 0 and (self.end.compareValue(other.end) >= 0 and self.start.compareValue(other.start) <= 0) 
    
    def disjoint(self,other):
        """
        Check if two line-segments are disjoint

        :param other: other line to be checked
        :type other: ~sortal.ind.lineSegment.lineSegment3D
        :returns: boolean -- whether the two line-segments are disjoint
        """
        return self.compareValue(other) == 0 and (self.end.compareValue(other.start) == -1 or self.start.compareValue(other.end) == 1)
    
    def align(self, other):
        """
        Check if two line-segments align

        :param other: other line to be checked
        :type other: ~sortal.ind.lineSegment.lineSegment3D
        :returns: boolean -- whether the two line-segments align
        """
        return self.compareValue(other) == 0 and (self.end == other.end or self.start == other.start)

    def common(self,other):
        """
        Find the common of two line-segment and modify this line to be the common segment
        
        :param other: other line to find common segment of
        :type other: ~sortal.ind.lineSegment.lineSegment3D
        :returns: boolean -- whether there is a common segment between two lines
        """
        if self.disjoint(other) or self.touches(other):
            return False
        
        new = self.duplicateInd(self.sort)
        if self.end.compareValue(other.end) == 1:
            new.end = other.end
        if self.start.compareValue(other.start) == -1:
            new.start = other.start
            
        return new
            
        if self.end.compareValue(other.end) == 1:
            self.end = other.end
        if self.start.compareValue(other.start) == -1:
            self.start = other.start
        return True
        
    def combine(self,other):
        """
        Find the combination of two line-segments
        
        :param other: other line to find common segment of two line-segments
        :type other: ~sortal.ind.lineSegment.lineSegment3D
        :returns: :class:`~sortal.ind.lineSegment.lineSegment3D` -- the combination of the two line-segments
        """
        if self.disjoint(other):
            return self
        nStart = self.start
        nEnd = self.end
        if self.end.compareValue(other.end) == -1:
            nEnd = other.end
        if self.start.compareValue(other.start) == 1:
            nStart = other.start
        return self.__class__(self.sort, nStart, nEnd)
        
    def complement(self,other):
        """
        Find the complement of this line-segment on the other line-segment
        
        :param other: other line to find complement segments
        :type other: ~sortal.ind.lineSegment.lineSegment3D
        :returns: list of :class:`~sortal.ind.lineSegment.lineSegment3D` -- the list of complement segment of this line-segment on the other
        """
        result = []
        if self.start.compareValue(other.start) == -1 and self.end.compareValue(other.start) == 1 :
            new = self.duplicateInd(self.sort)
            new.end = other.start
            result.append(new)
        if self.start.compareValue(other.end) == -1 and self.end.compareValue(other.end) == 1 :
            new = self.duplicateInd(self.sort)
            new.start = other.end
            result.append(new)
        return result
        
    def midpoint(self):
        """
        Find the midpoint of a line-segment

        :returns: :class:`~sortal.struct.vector.vector3D` -- the coordinates of the mid-point of the line-segment
        """
        return (float(((self.start.x+self.end.x)/c2).value),float(((self.start.y+self.end.y)/c2).value),float(((self.start.z+self.end.z)/c2).value))
            
    def replicateInd(self, sort = 0):
        """
        Create a duplicate of the individual without attributes
        
        :returns: the duplicate object
        """
        if sort == 0:
            sort = self.sort
        return self.__class__(sort, self.start, self.end)    
        
    def codescriptor(self):
        """
        Codescriptor of the line-segment: the infinite 3D line that defines this line-segment's root and direction

        :returns: :class:`~sortal.ind.line.line3D` -- the infinite 3D line that defines this line-segment
        """
        line = line3D(self.sort, self.start, self.end)
        line.assoc = self
        return line
        

    def transformInd(self, transformSet):
        """
        Transforming the individual with a transformation set (editor)

        :param transformSet: the transformation set to be applied
        :type transformSet: ~sortal.transform.transformSet.transformSet
        :returns: the transformed object
        """
        affine = transformSet.retrieveTransform(affineTransform3D)
        start = self.oStart
        end = self.oEnd
        try:
            if affine is None:
                newLine = self.duplicateInd()
            else:
                start = start.transform(affine)
                end = end.transform(affine)         
                newLine = self.__class__(self.sort, start, end)
            
            transf = transformSet.retrieveTransform(graphTransform3D)
            if transf is not None:
                line = transf.apply(newLine)
                if line is not None:
                    return line
                start = transf.apply(start)
                end = transf.apply(end)
                newLine = self.__class__(self.sort, start, end)
        except ZeroDivisionError:
            return None
        return newLine

        
    def containsInd(self, ind):
        """
        Check if this individual contains another individual
        """
        if not self.contains(ind):
            return False
        if self.sort.type == 'sort-attribute':
            # if not self.hasAttribute():
            #     if ind.hasAttribute():
            #         return False
            #     else:
            #         return True
            if ind.hasAttribute():
                if ind.attribute.checkPartOf(self.attribute):
                    return True
                else:
                    return False
            else:
                return True
        return True
        
        
    @staticmethod
    def parse(reader, sort):
        """
        Parse the 3D line-segment from a sortal parseReader (from string)

        :param reader: the reader of the string to be parsed
        :type reader: ~sortal.io.parseReader.parseReader
        :param sort: sort of the new object
        :type sort: ~sortal.sort.sort.sort
        :returns: :class:`~sortal.ind.lineSegment.lineSegment3D` -- the individual parsed
        """
        if reader.token() != '<':
            raise ParsingError(reader,'expected a "<" at the beginning of creating a line-segment')
        if reader.newToken() != '(': #bracket
            raise ParsingError(reader,'expected a "(" ')
        x,y,z = vector3D.readVector(reader)
        start = vector3D(x,y,z)
            
        reader.newToken() # bracket (current token would be ',')
        
        x,y,z = vector3D.readVector(reader)
        end = vector3D(x,y,z)
            
        ind = lineSegment3D(sort, start, end)

        if reader.token() != '>':
            raise ParsingError(reader, '">" expected to close the line-segment')
            
        reader.newToken() #next token
        
        return ind
