# -*- coding: utf-8 -*-
"""
A parametric line-segment inherits most of its behavior from a normal line-segment with the
exception of its mapping for parametric match detection.

The LineSegmentP class extends on the LineSegment class.
It specifies a {sortal.map.graphMap} or {sortal.map.graphMap3D}.
"""

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import range
from builtins import super
from builtins import str

from .line import line2D, line3D
from .lineSegment import lineSegment2D, lineSegment3D
from .plane import plane2D, plane3D
from ..struct.vector import vector2D, vector3D
from ..io.parseReader import ParsingError
from ..transform.graphTransform import graphTransform, graphTransform3D
from ..transform.affineTransform import affineTransform2D, affineTransform3D


class lineSegmentP2D(lineSegment2D):
    """
    Parametric line-segment in 2D class. Inherit all its attributes from lineSegment2D.

    **Class Properties:**

    * **Behavior**: :class:`~sortal.form.interval.interval`
    * **Sort Argument**: :class:`~sortal.sort.parameter.argument.argument.argumentType`.NONE
    * **Mapping**: :class:`~sortal.map.graphMap.graphMap`

    """

    type = 'ind-lineSegmentP2D'
    
    def __init__(self, sort, start = vector2D(0,0), end = vector2D(0,0), *args, **kwargs):
        """
        Initialize a parametric line-segment using its two end-points. Same inputs and procedure as lineSegment2D.

        :param sort: sort of the line
        :type sort: ~sortal.sort.sort.sort
        :param start: first end-point of the line
        :type start: ~sortal.struct.vector.vector2D
        :param end: second end-point of the line
        :type end: ~sortal.struct.vector.vector2D
        """
        super(lineSegmentP2D, self).__init__(sort, start, end)
        
    
    def codescriptor(self):
        """
        Codescriptor of the line-segment for parametric matching which is itself
        """
        return self
        
    def parametricNode(self):
        """
        Parametric Node for parametric matching which is a line2D defining its direction and root
        """
        line = line2D(self.sort, self.start, self.end)
        line.assoc = self
        return line
        

    def transformInd(self, transformSet):
        """
        Transforming the individual with a transformation set (editor)

        :param transformSet: the transformation set to be applied
        :type transformSet: ~sortal.transform.transformSet.transformSet
        :returns: the transformed object
        """
        transf = transformSet.retrieveTransform(graphTransform)
        if transf is None:
            return self.duplicateInd()
        start = transf.apply(self.start)
        end = transf.apply(self.end)
        try:
            newLine = lineSegmentP2D(self.sort, start, end)
        except ZeroDivisionError:
            return None
        return newLine
        
    def checkTransform(self, transform, compareForm):
        """
        Check if the individual belongs to the compareForm after transformation with a transform set
        """
        try:
            comp = self.edit(transform)
        except ValueError:
            return False
        if comp is None:
            return True
        if compareForm.containsInd(comp):
            return True
        return False


    @staticmethod
    def parse(reader, sort):
        """
        Parse the 2D parametric line-segment a sortal parseReader (from string)

        :param reader: the reader of the string to be parsed
        :type reader: ~sortal.io.parseReader.parseReader
        :param sort: sort of the new object
        :type sort: ~sortal.sort.sort.sort
        :returns: :class:`~sortal.ind.lineSegment.lineSegment2D` -- the individual parsed
        """
        if reader.token() != '<':
            raise ParsingError(reader,'expected a "<" at the beginning of creating a line-segment')
        if reader.newToken() != '(': #bracket
            raise ParsingError(reader,'expected a "(" ')
        x,y = vector2D.readVector(reader)
        start = vector2D(x,y)
            
        reader.newToken() # bracket (current token would be ',')
        
        x,y = vector2D.readVector(reader)
        end = vector2D(x,y)
            
        ind = lineSegmentP2D(sort, start, end)

        if reader.token() != '>':
            raise ParsingError(reader, '">" expected to close the line-segment')
            
        reader.newToken() #next token
        
        return ind
    
    
class lineSegmentP3D(lineSegment3D):
    """
    Parametric line-segment in 3D class. Inherit all its attributes from lineSegment3D.

    **Class Properties:**

    * **Behavior**: :class:`~sortal.form.interval.interval`
    * **Sort Argument**: :class:`~sortal.sort.parameter.argument.argument.argumentType`.NONE
    * **Mapping**: :class:`~sortal.map.graphMap3D.graphMap3D`
    """

    type = 'ind-lineSegmentP3D'
    
    def __init__(self, sort, start = vector3D(0,0,0), end = vector3D(0,0,0), *args, **kwargs):
        """
        Initialize a parametric line-segment using its two end-points. Same inputs and procedure as lineSegment3D.

        :param sort: sort of the line
        :type sort: ~sortal.sort.sort.sort
        :param start: first end-point of the line
        :type start: ~sortal.struct.vector.vector3D
        :param end: second end-point of the line
        :type end: ~sortal.struct.vector.vector3D
        """
        super(lineSegmentP3D, self).__init__(sort, start, end)
        
    def codescriptor(self):
        """
        Codescriptor of the line-segment for parametric matching which is itself
        """
        return self
        
    def parametricNode(self):
        """
        Parametric Node for parametric matching which is a line3D defining its direction and root
        """
        line = line3D(self.sort.getSort('__dummyLine3D__'), self.start, self.end)
        line.assoc = self
        return line
        

    def transformInd(self, transformSet):
        """
        Transforming the individual with a transformation set (editor)

        :param transformSet: the transformation set to be applied
        :type transformSet: ~sortal.transform.transformSet.transformSet
        :returns: the transformed object
        """
        transf = transformSet.retrieveTransform(graphTransform3D)
        if transf is None:
            return self.duplicateInd()
        start = transf.apply(self.start)
        end = transf.apply(self.end)
        try:
            newLine = lineSegmentP3D(self.sort, start, end)
        except ZeroDivisionError:
            return None
        return newLine
        
    def checkTransform(self, transform, compareForm):
        """
        Check if the individual belongs to the compareForm after transformation with a transform set
        """
        try:
            comp = self.edit(transform)
        except ValueError:
            return False
        if comp is None:
            return True
        if compareForm.containsInd(comp):
            return True
        return False


    @staticmethod
    def parse(reader, sort):
        """
        Parse the 3D parametric line-segment a sortal parseReader (from string)

        :param reader: the reader of the string to be parsed
        :type reader: ~sortal.io.parseReader.parseReader
        :param sort: sort of the new object
        :type sort: ~sortal.sort.sort.sort
        :returns: :class:`~sortal.ind.lineSegment.lineSegment3D` -- the individual parsed
        """
        if reader.token() != '<':
            raise ParsingError(reader,'expected a "<" at the beginning of creating a line-segment')
        if reader.newToken() != '(': #bracket
            raise ParsingError(reader,'expected a "(" ')
        x,y,z = vector3D.readVector(reader)
        start = vector3D(x,y,z)
            
        reader.newToken() # bracket (current token would be ',')
        
        x,y,z = vector3D.readVector(reader)
        end = vector3D(x,y,z)
            
        ind = lineSegmentP3D(sort, start, end)

        if reader.token() != '>':
            raise ParsingError(reader, '">" expected to close the line-segment')
            
        reader.newToken() #next token
        
        return ind
    
    
    