# -*- coding: utf-8 -*-
"""
"""
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import super
from builtins import str

from .individual import individual
from ..struct.coordinate import cn1
from ..struct.vector import vector2D, vector3D
from ..transform.affineTransform import affineTransform3D
from ..transform.graphTransform import graphTransform, graphTransform3D

class plane2D(individual):
    """
    2D plane class. Since there is only exactly one plane in 2D space, all plane2D object would be the same

    **Class Properties:**

    * **Behavior**: :class:`~sortal.form.discrete.discrete`
    * **Sort Argument**: :class:`~sortal.sort.parameter.argument.argument.argumentType`.NONE
    * **Mapping**: :class:`~sortal.map.similarityMap.similarityMap2D`
    """

    type = 'ind-plane2D'
    
    def __init__(self, sort, *args, **kwargs):
        """
        Initialize a plane2D
        """
        super(plane2D, self).__init__(sort)
    
    def toString(self, printKey):
        """
        string format of the individual (with no attribute)

        :param printKey: whether the tag of the individual is to be added to the string
        :type printKey: boolean
        :returns: str -- string format of the individual
        """
        s = ''
        if printKey and self.key is not None:
            s += str(self.key)+' '
        return s+'plane'
        
    def equals(self, other):
        """
        check equality between two planes

        :param other: other individual to be compared
        :type other: ~sortal.ind.plane.plane2D
        :returns: boolean -- whether two individuals are equal
        """
        return True
        
    def isNil(self):
        """
        plane2D cannot be Nil
        """
        return False
    
    def compareValue(self, other):
        """
        Compare the value of two planes

        :param other: other planes to be compared
        :type other: ~sortal.ind.plane.plane2D
        :returns: int -- 0 if equals, 1 if this line is bigger, -1 if smaller
        """
        if other.type == 'ind-plane2D':
            return 0
        return False
        
    @staticmethod
    def compareAngle(center, first, second):
        return (first.product(second) + center.product(first.subtract(second))).sign()
        
    @staticmethod
    def compareRelAngle(center, ref, first, second):
        a = plane2D.compareAngle(center, ref, first)
        b = plane2D.compareAngle(center, ref, second)
        if a == b:
            if a != 0:
                return plane2D.compareAngle(center, first, second)
            if (ref < center) == (second < center) and (ref < center) != (first < center):
                return 1
            if (ref < center) != (second < center) and (ref < center) == (first < center):
                return -1
            return 0
        if a == 0:
            if (ref < center) == (first < center):
                return -1
            else:
                return -1 * b
        if b == 0:
            if (ref < center) == (second < center):
                return 1
        return a

    @staticmethod
    def angleContainsPin(center, p1, pin, p2):
        a = plane2D.compareAngle(center, p1, pin)
        b = plane2D.compareAngle(center, p1, p2)
        if a == b:
            if a != 0:
                return plane2D.compareAngle(center, pin, p2)
            if (p1 < center) == (p2 < center) and (p1 < center) != (pin < center):
                return 1
            return 0
        if a == 0:
            if (p1 < center) == (pin < center):
                return 0
            else:
                return -1 * b
        if b == 0:
            if (p1 < center) == (p2 < center):
                return 1
        return a
        
    @staticmethod
    def findPin(center, start, end):
        angleComp = plane2D.compareAngle(center, start, end)
        if angleComp == -1:
            return start.midpoint(end)
        mid = start.midpoint(end)
        if angleComp == 1:
            pin = center.subtract(mid.subtract(center))
            return pin
        if mid == center:
            a = vector2D(1,0)
            pin = center.add(a)
            if plane2D.compareAngle(center, start, pin) == 0:
                a = vector2D(0,1)
                pin = center.add(a)
            if plane2D.compareAngle(center, start, pin) == 1:
                pin = center.subtract(a)
            return pin
        raise ValueError('Backend error: start == end when find pin')
        
    def parallel(self, other):
        if other.type == 'ind-plane2D':
            return True
        return False
        
class plane3D(individual):
    """
    3D plane class. 

    A Plane is a connected, non-bounded planar geometry.

    The Plane class extends on the individual class.
    It defines the characteristic individual for planes.
    Each plane is defined by a normal vector and a plane root vector.
    All planes' normal vector is saved as positive direction vector. Whether the plane's normal
    is negative when initilized, it would be saved as negNormal: a boolean value which is True if 
    the initial normal is negative, False otherwise.

    **Object Parameters:** (not including inherited attributes)

    * **normal** (:class:`~sortal.struct.vector.vector3D`) -- normal vector of the plane.
    * **pRoot** (:class:`~sortal.struct.vector.vector3D`) -- root of the plane - the projection of the origin (0,0,0) on the plane.
    * **scalar** (:class:`~sortal.struct.coord.coordinate_float.coordinate`) -- 
    * **u** (:class:`~sortal.struct.vector.vector3D`) -- 3D vector representing the local direction vector (1,0) on the plane.
    * **v** (:class:`~sortal.struct.vector.vector3D`) -- 3D vector representing the local direction vector (0,1) on the plane.
    * **negNormal** (boolean) -- whether the normal first calculated is negative.

    **Class Properties:**

    * **Behavior**: :class:`~sortal.form.discrete.discrete`
    * **Sort Argument**: :class:`~sortal.sort.parameter.argument.argument.argumentType`.NONE
    * **Mapping**: :class:`~sortal.map.similarityMap3D.similarityMap3D`
    """

    type = 'ind-plane3D'
    
    def __init__(self, sort, boundary, normal_n_origin = False, normal = None, origin = None, *args, **kwargs):
        """
        Initialize a plane.

        :param sort: sort of the plane
        :type sort: ~sortal.sort.sort.sort
        :param boundary: list of points contained by the plane (must have at least 3 non-colinear points). If initialize with normal and origin,
        :type boundary: list of :class:`~sortal.struct.vector.vector3D`
        :param normal_n_origin: second end-point of the line
        :type normal_n_origin: boolean
        """
        if normal_n_origin is False:
            super(plane3D, self).__init__(sort)
            p0, p1, p2 = boundary[0], boundary[1], boundary[2]
            cc = 3
            while p0.colinear(p1,p2):
                #print('first 3 vectors of plane is colinear, getting next 1',p0.printRhino(), p1.printRhino(), p2.printRhino())
                p2 = boundary[cc]
                cc+=1
            self.normal = ((p2.subtract(p0)).product(p1.subtract(p0))).unit()
            if self.normal < vector3D(0,0,0):
                self.negNormal = True
                self.normal = self.normal.scale(cn1)
            else:
                self.negNormal = False
            self.scalar = self.normal.scalar(boundary[0])
            self.pRoot = self.normal.scale(self.scalar)
            
        else:
            super(plane3D, self).__init__(sort)
            self.normal = normal
            if self.normal < vector3D(0,0,0):
                self.negNormal = True
                self.normal = self.normal.scale(cn1)
            else:
                self.negNormal = False
            self.normal = self.normal.unit()
            self.scalar = self.normal.scalar(origin)
            self.pRoot = self.normal.scale(self.scalar)
            
        if self.normal.parallel(vector3D(0,0,1)):
            self.u = vector3D(1,0,0)
            self.v = vector3D(0,1,0)
        else:
            self.u = self.normal.product(vector3D(0,0,1)).unit()
            self.v = self.normal.product(self.u).unit()
            
    def toString(self, printKey):
        """
        string format of the individual (with no attribute)

        :param printKey: whether the tag of the individual is to be added to the string
        :type printKey: boolean
        :returns: str -- string format of the individual
        """
        s = ''
        if printKey and self.key is not None:
            s += str(self.key)+' '
        return s + 'normal:'+str(self.normal)+';pRoot:'+str(self.pRoot)
        
    def printRhino(self, dX = 0, dY = 0):
        """
        print the individual in a format that can be pasted into Rhinoceros for visualization
        """
        if not self.normal.x == 0:
            dir0 = self.normal.__class__(self.normal.y, -self.normal.x, 0)
            dir1 = self.normal.__class__(self.normal.z, 0, -self.normal.x)
        elif not self.normal.y == 0:
            dir0 = self.normal.__class__(self.normal.y, -self.normal.x, 0)
            dir1 = self.normal.__class__(0, self.normal.z, -self.normal.y)
        elif not self.normal.z == 0:
            dir0 = self.normal.__class__(self.normal.z, 0, -self.normal.x)
            dir1 = self.normal.__class__(0, self.normal.z, -self.normal.y)
        p = []
        p.append(dir0.scale(10).add(self.pRoot))
        p.append(dir1.scale(10).add(self.pRoot))
        p.append(dir0.scale(-10).add(self.pRoot))
        p.append(dir1.scale(-10).add(self.pRoot))
        s='_polyline '
        for j in p:
            s += str(j.x + dX) + ',' + str(j.y + dY) + ',' + str(j.z) + ' '
        s+= str(p[0].x + dX) + ',' + str(p[0].y + dY) + ',' + str(p[0].z) + '\n'+'   _SelPolyline _PlanarSrf _Delete'
        return s 
        
    def equals(self, other):
        """
        check equality between two planes

        :param other: other individual to be compared
        :type other: ~sortal.ind.plane.plane3D
        :returns: boolean -- whether two individuals are equal
        """
        return self.normal == other.normal and self.scalar == other.scalar
        
    def compareValue(self, other):
        """
        Comparing value of 2 planes:

        First compare normal vectors

        If normals are equal, compare scalar

        :param other: other line to be compared
        :type other: ~sortal.ind.line.line3D
        :returns: int -- 0 if equals, 1 if this line is bigger, -1 if smaller
        """
        c = self.normal.compareValue(other.normal)
        if c != 0:
            return c
        return self.scalar.compareValue(other.scalar)
                
    def isNil(self):
        """
        check if the normal vector is Nil
        """
        return self.normal.isZero()
        
    def direction(self, coords):
        """
        3D direction of a 2D local point on the plane from the plane's root

        :param coords: coordinates of the point to find direction of
        :type coords: ~sortal.struct.vector.vector2D
        :returns: :class:`~sortal.struct.vector.vector3D` -- direction from the 3D plane root to the point
        """
        return self.u.scale(coords.x).add(self.v.scale(coords.y))
        
    def position(self, coords):
        """
        3D position of a 2D local point on the plane 

        :param coords: coordinates of the point to find position of
        :type coords: ~sortal.struct.vector.vector2D
        :returns: :class:`~sortal.struct.vector.vector3D` -- position of the point in 3D space
        """
        return self.pRoot.add(self.direction(coords))
        
    def localize(self, point):
        """
        localize a 3D point into a 2D local point on the plane. If the point is not on the plane, the function
        would return the 2D local point of its projection

        :param point: coordinates of the point to find local position of
        :type point: ~sortal.struct.vector.vector3D
        :returns: :class:`~sortal.struct.vector.vector2D` -- local position of the point on the plane
        """
        diff = point.subtract(self.pRoot)
        return vector2D(self.u.scalar(diff), self.v.scalar(diff))
        
    def projection(self, point):
        """
        Find the 3D projection of a point on the plane

        :param point: coordinates of the point to find local position of
        :type point: ~sortal.struct.vector.vector3D
        :returns: :class:`~sortal.struct.vector.vector3D` -- projection of the point on the plane
        """
        return self.position(self.localize(point))
        
    def containsPoint(self, point):
        """
        Check if the plane contains a specific point

        :param point: point to be checked
        :type point: ~sortal.struct.vector.vector3D
        :returns: boolean -- whether the plane contains the point
        """
        if self.isNil():
            raise TypeError('Plane can not be nil')
        return self.normal.dotproduct(self.pRoot.subtract(point)) == 0
            
    def containsLine(self, line):
        """
        Check if the plane contains a specific line

        :param line: line to be checked
        :type line: ~sortal.ind.line.line3D
        :returns: boolean -- whether the plane contains the line
        """
        if self.isNil() or line.isNil():
            raise TypeError('Neither plane nor line may be nil')
        return self.normal.dotproduct(line.direction) == 0 and (self.pRoot == line.root 
        or line.root.subtract(self.pRoot).dotproduct(self.normal) == 0)
                
    def parallel(self, other):
        """
        Check if the plane is parallel to another plane

        :param other: other plane to be checked
        :type other: ~sortal.ind.plane.plane3D
        :returns: boolean -- whether the planes are parallel
        """
        if self.isNil() or other.isNil():
            raise TypeError('Neither plane may be nil')
        return self.normal == other.normal
    
    def perpendicular(self, other):
        """
        Check if the plane is perpendicular to another plane

        :param other: other plane to be checked
        :type other: ~sortal.ind.plane.plane3D
        :returns: boolean -- whether the planes are perpendicular
        """
        if self.isNil() or other.isNil():
            raise TypeError('Neither plane may be nil')
        return self.normal.perpendicular(other.normal)
            
    def parallelLine(self, line):
        """
        Check if the plane is parallel to a line

        :param line: line to be checked
        :type line: ~sortal.ind.line.line2D
        :returns: boolean -- whether the plane and line are parallel
        """
        if self.isNil() or line.isNil():
            raise TypeError('Neither plane nor line may be nil')
        return self.normal.dotproduct(line.direction) == 0
    
    def perpendicularLine(self, line):
        """
        Check if the plane is perpendicular to a line

        :param line: line to be checked
        :type line: ~sortal.ind.line.line2D
        :returns: boolean -- whether the plane and line are perpendicular
        """
        if self.isNil() or line.isNil():
            raise TypeError('Neither plane nor line may be nil')
        return self.normal == line.unitDir
        
    def replicateInd(self):
        """
        Create a duplicate of the individual without attributes
        
        :returns: the duplicate object
        """
        result = plane3D(self.sort, [], True, self.normal, self.pRoot)
        return result
        
        
    def transformInd(self, transformSet):
        """
        Transforming the individual with a transformation set (editor)

        :param transformSet: the transformation set to be applied
        :type transformSet: ~sortal.transform.transformSet.transformSet
        :returns: the transformed object
        """
        affine = transformSet.retrieveTransform(affineTransform3D)
        if affine is None:
            newPlane = self.duplicateInd()
        else:
            position0 = self.pRoot.transform(affine)
            position1 = self.pRoot.add(self.u).transform(affine)      
            position2 = self.pRoot.add(self.v).transform(affine)      
            newPlane =  plane3D(self.sort, [position0,position1,position2])
        transf = transformSet.retrieveTransform(graphTransform3D)
        if transf is not None:
            count = 0
            bounds = []
            for i in transf.vects:
                if self.containsPoint(i[0]):
                    if len(bounds) == 2 and i[1].colinear(bounds[0],bounds[1]):
                        continue
                    bounds.append(i[1])
                    count += 1
                if count > 3:
                    break
            if count < 3:
                count = 0
                bounds = []
                extraKey = str(self)
                for i in transf.extra:
                    if i[-1] == extraKey:
                        if len(bounds) == 2 and i[1].colinear(bounds[0],bounds[1]):
                            continue
                        bounds.append(i[1])
                        count += 1
                if count < 3:
                    return None
            newPlane = plane3D(self.sort, bounds)
        return newPlane
        
    def intersection(self, other):
        """
        find the line that is the intersection between this plane and another plane

        :param other: other plane to find intersection with
        :type other: ~sortal.ind.plane.plane3D
        :returns: (sortal.vector3D, sortal.vector3D) -- tuple of 2 vector3D representing the direction and root of the line that is the intersection
        """
        if self.isNil() or other.isNil():
            raise TypeError('Neither plane may be nil')
        if self.parallel(other):
            raise ValueError('The two planes are parallel to one another')
        prod = self.normal.product(other.normal)
        
        pDir = prod.product(self.normal)
        d = (other.pRoot.subtract(self.pRoot).dotproduct(other.normal)) / (pDir.dotproduct(other.normal))
        return prod, pDir.scale(d).add(self.pRoot)

    def intersectionLine(self, line):
        """
        find the intersection point between the plane and a line

        :param other: line to find intersection with
        :type other: ~sortal.ind.line.line3D
        :returns: :class:`~sortal.struct.vector.vector3D` -- coordinate of the intersection point
        """
        if self.isNil() or line.isNil():
            raise TypeError('Neither plane nor line may be nil')
        if self.parallelLine(line):
            raise ValueError('The two object are parallel to one another')
        d = (self.pRoot.subtract(line.root).dotproduct(self.normal)) / (line.direction.dotproduct(self.normal))
        return line.direction.scale(d).add(line.root)
