# -*- coding: utf-8 -*-
"""
Plane-segment for parametric matching.
The parametric plane-segment classes extends on the Plane-segment classes.
However, it specifies a {sortal.map.graphMap} or {sortal.map.graphMap3D} as default.
"""

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import range
from builtins import super
from builtins import str

from .plane import plane3D
from .planeSegment import planeSegment2D
from .planeSegment3D import planeSegment3D
from ..struct.planeStruct import bSegment, boundary
from ..struct.vector import vector2D, vector3D
from ..form.coForm import coForm
from ..io.parseReader import ParsingError

class planeSegmentP2D(planeSegment2D):
    """
    parametric plane-segment in 2D class.

    **Class Properties:**

    * **Behavior**: :class:`~sortal.form.coForm.coForm`
    * **Sort Argument**: :class:`~sortal.sort.parameter.argument.argument.argumentType`.NONE
    * **Mapping**: :class:`~sortal.map.graphMap.graphMap`
    """

    type = 'ind-planeSegmentP2D'
    
    def __init__(self, sort, boundr, boundaryList = False, *args, **kwargs):
        """
        2 ways to initialize:

        * With a list of vectors for the outer boundary: boundr = a list of vectors forming the outer boundary in a continuous sequence (repetition of the first vector is not required), boundaryList = False. The inner boundaries can be added after initialization with planeSegObj.addInnerBound(list_of_one_inner_bound_vectors).
        * With an established list of boundaries: boundr = a list of existing boundaries, boundaryList = True.
        
        :param sort: sort of the plane-segment
        :type sort: ~sortal.sort.sort.sort
        :param boundr: list of vectors or list of boundaries depending on the way the plane-segment is intended to be initialized
        :type boundr: list of :class:`~sortal.struct.vector.vector2D` or list of :class:`~sortal.struct.planeStruct.boundary`
        :param boundaryList: indication for which method is used to initialize the plane-segment, default to False (list of vectors for outer boundary)
        :type boundaryList: boolean
        """
        super(planeSegmentP2D, self).__init__(sort, boundr, boundaryList)

    def codescriptor(self):
        """
        codescriptor of a paramtetric plane-segment
        """
        return self
        
    @staticmethod
    def classifyBoundary(sort, boundary1, boundary2):
        """
        perform classification on two boundaries

        :param other: the plane-segment to be classified
        :type other: ~sortal.ind.planeSegmentP.planeSegmentP2D
        :returns: sortal.struct.classes.classes -- the classification of the two boundaries
        """
        form1 = coForm(sort)
        one = planeSegment2D(sort, [boundary1], True)
        form1.insertClassification(one)
        form2 = coForm(sort)
        two = planeSegment2D(sort, [boundary2], True)
        form2.insertClassification(two)
        result = planeSegment2D.classification(form1, form2)
        return result
        

    def replicateInd(self):
        """
        Create a duplicate of the individual without attributes
        
        :returns: the duplicate object
        """
        bdr = []
        for i in self.boundaries:
            bound = boundary()
            for j in i.canonical:
                bound.canonical.append(j)
            for j in i.polygonal:
                bound.polygonal.append(j)
            bdr.append(bound)
        result = planeSegmentP2D(self.sort, bdr, True)
        return result

    @staticmethod
    def parse(reader, sort):
        """
        Parse the 2D plane-segment from a sortal parseReader (from string)

        :param reader: the reader of the string to be parsed
        :type reader: ~sortal.io.parseReader.parseReader
        :param sort: sort of the new object
        :type sort: ~sortal.sort.sort.sort
        :returns: :class:`~sortal.ind.planeSegmentP.planeSegmentP2D` -- the individual parsed
        """
        if reader.token() != '<':
            raise ParsingError(reader,'expected a "<" at the beginning of creating a plane-segment')
        if reader.newToken() != '<': #bracket
            raise ParsingError(reader,'expected a "<" ')
            
        bounds = []
        index = 0
        while reader.token() == '<':
            bounds.append(boundary())
            while True:
                reader.newToken() #bracket
                x,y = vector2D.readVector(reader) # comma
                start = vector2D(x,y) 
                    
                reader.newToken() #bracket
                
                x,y = vector2D.readVector(reader) # '>'
                end = vector2D(x,y) 
    
                bounds[index].add(bSegment(start, end))
    
                reader.newToken() # ','/'>'/'<'
                if reader.token() == ',':
                    reader.newToken()
                elif reader.token() == '>' or reader.token() == '<':
                    while reader.previewToken() == '<':
                        reader.newToken()
                    break
            
            index += 1
            
        for bdr in bounds:
            vertices = 2
            bdr.add(bdr.canonical[0].start)
            bdr.add(bdr.canonical[0].end)
            vert = bdr.canonical[0].end
            seg = bdr.canonical[0]
            while vertices < len(bdr.canonical):
                for segment in bdr.canonical:
                    if vert == segment.start and seg != segment:
                        bdr.add(segment.end)
                        vert = segment.end
                        seg = segment
                        vertices += 1
                        break
                    if vert == segment.end and seg != segment:
                        bdr.add(segment.start)
                        vert = segment.start
                        seg = segment
                        vertices += 1
                        break
        reader.newToken()

        ind = planeSegmentP2D(sort, bounds, True)
        classes = ind.classify(None)
        result = coForm(sort)
        ind.construct(classes.sumSet(), result)
        ind.boundaries = result.elements[0].boundaries
        
        
        return ind
        
        

class planeSegmentP3D(planeSegment3D):
    """
    parametric plane-segment in 3D class.

    **Class Properties:**

    * **Behavior**: :class:`~sortal.form.areal.areal`
    * **Sort Argument**: :class:`~sortal.sort.parameter.argument.argument.argumentType`.NONE
    * **Mapping**: :class:`~sortal.map.graphMap3D.graphMap3D`
    """
    type = 'ind-planeSegmentP3D'
    
    def __init__(self, sort, boundr, boundaryList = False, *args, **kwargs):
        """
        2 ways to initialize:

        * With a list of vectors for the outer boundary: boundr = a list of vectors forming the outer boundary in a continuous sequence (repetition of the first vector is not required), boundaryList = False. The inner boundaries can be added after initialization with planeSegObj.addInnerBound(list_of_one_inner_bound_vectors).
        * With an established list of boundaries: boundr = a list of existing boundaries, boundaryList = True.
        
        :param sort: sort of the plane-segment
        :type sort: ~sortal.sort.sort.sort
        :param boundr: list of vectors or list of boundaries depending on the way the plane segment is intended to be initialized
        :type boundr: list of :class:`~sortal.struct.vector.vector3D` or list of :class:`~sortal.struct.planeStruct.boundary`
        :param boundaryList: indication for which method is used to initialize the plane-segment, default to False (list of vectors for outer boundary)
        :type boundaryList: boolean
        """
        super(planeSegmentP3D, self).__init__(sort, boundr, boundaryList)

    def codescriptor(self):
        """
        codescriptor of a paramtetric plane-segment
        """
        return self
        
    def parametricNode(self):
        """
        The parametric node for parametric matching: the plane3D that contains the plane segment
        """
        c = plane3D(self.sort.getSort('__dummyPlane3D__'), self.boundaries[0].polygonal)
        c.assoc = self
        return c
        
    @staticmethod
    def classifyBoundary(sort, boundary1, boundary2):
        """
        perform classification on two boundaries

        :param other: the plane-segment to be classified
        :type other: ~sortal.ind.planeSegmentP.planeSegmentP3D
        :returns: sortal.struct.classes.classes -- the classification of the two boundaries
        """
        form1 = coForm(sort)
        one = planeSegment3D(sort, [boundary1], True)
        form1.insertClassification(one)
        form2 = coForm(sort)
        two = planeSegment3D(sort, [boundary2], True)
        form2.insertClassification(two)
        result = planeSegment3D.classification(form1, form2)
        return result
        
    def compareValue(self, other):
        """
        Compare the plane that contain this plane segment with another plane segment

        :param other: other individual to be compared
        :type other: ~sortal.ind.planeSegmentP.planeSegmentP3D
        :returns: int -- -1 if this plane segment is smaller, 1 if bigger, 0 if equal
        """
        if 'ind-planeSegment' not in other.type:
            return False
        else:
            return self.parametricNode().compareValue(other.parametricNode())
            
    def replicateInd(self):
        """
        Create a duplicate of the individual without attributes
        
        :returns: the duplicate object
        """
        bdr = []
        for i in self.boundaries:
            bound = boundary()
            for j in i.canonical:
                bound.canonical.append(j)
            for j in i.polygonal:
                bound.polygonal.append(j)
            bdr.append(bound)
        result = planeSegmentP3D(self.sort, bdr, True)
        return result
            
    @staticmethod
    def parse(reader, sort):
        """
        Parse the 3D plane-segment from a sortal parseReader (from string)

        :param reader: the reader of the string to be parsed
        :type reader: ~sortal.io.parseReader.parseReader
        :param sort: sort of the new object
        :type sort: ~sortal.sort.sort.sort
        :returns: :class:`~sortal.ind.planeSegmentP.planeSegmentP3D` -- the individual parsed
        """
        if reader.token() != '<':
            raise ParsingError(reader,'expected a "<" at the beginning of creating a plane-segment')
        if reader.newToken() != '<': #bracket
            raise ParsingError(reader,'expected a "<" ')
            
        bounds = []
        index = 0
        while reader.token() == '<':
            bounds.append(boundary())
            while True:
                reader.newToken() #bracket
                x,y,z = vector3D.readVector(reader) # comma
                start = vector3D(x,y,z) 

                reader.newToken() #bracket
                x,y,z = vector3D.readVector(reader) # '>'
                end = vector3D(x,y,z) 
    
                bounds[index].add(bSegment(start, end))
    
                reader.newToken() # ','/'>'/'<'
                if reader.token() == ',':
                    reader.newToken()
                elif reader.token() == '>' or reader.token() == '<':
                    while reader.previewToken() == '<':
                        reader.newToken()
                    break
            
            index += 1
            
        for bdr in bounds:
            vertices = 2
            bdr.add(bdr.canonical[0].start)
            bdr.add(bdr.canonical[0].end)
            vert = bdr.canonical[0].end
            seg = bdr.canonical[0]
            while vertices < len(bdr.canonical):
                for segment in bdr.canonical:
                    if vert == segment.start and seg != segment:
                        bdr.add(segment.end)
                        vert = segment.end
                        seg = segment
                        vertices += 1
                        break
                    if vert == segment.end and seg != segment:
                        bdr.add(segment.start)
                        vert = segment.start
                        seg = segment
                        vertices += 1
                        break
        reader.newToken()
        ind = planeSegmentP3D(sort, bounds, True)
        classes = ind.classify(None)
        result = coForm(sort)
        ind.construct(classes.sumSet(), result)
        ind.boundaries = result.elements[0].boundaries
        
        
        return ind
