# -*- coding: utf-8 -*-
"""
A point is a 0-dimensional geometric data entity with a discrete
behavior for non-parametric mapping.

The Point class extends on the Individual class.
It defines the characteristic individual for points.
A point is represented as a position vector.
This characteristic individual accepts no parameters.
It specifies a {sortal.map.similarityMap} or {sortal.map.similarityMap3D} as default.
Forms of points adhere to a discrete behavior.
"""

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import super
from builtins import str

from .individual import individual
from copy import deepcopy
from ..transform.affineTransform import affineTransform2D, affineTransform3D
from ..transform.graphTransform import graphTransform, graphTransform3D
from ..struct.vector import vector2D, vector3D
from ..struct.coordinate import coordinate, c0, c1
from ..io.parseReader import parseReader, ParsingError

class point2D(individual):
    """
    The point in 2D class.

    A point is a 0-dimensional geometric data entity with a discrete
    behavior for non-parametric mapping.

    The Point class extends on the Individual class.
    It defines the characteristic individual for points.
    A point is represented as a position vector.

    **Object Parameters:** (not including inherited attributes)
    
    * **value** (:class:`~sortal.struct.vector.vector2D`) -- position vector of the point

    **Class Properties:**

    * **Behavior**: :class:`~sortal.form.discrete.discrete`
    * **Sort Argument**: :class:`~sortal.sort.parameter.argument.argument.argumentType`.NONE
    * **Mapping**: :class:`~sortal.map.similarityMap.similarityMap2D`
    """

    type = 'ind-point2D'
    
    def __init__(self, sort, x, y = None, *args, **kwargs):
        """
        Initialize a 2D point

        :param sort: sort of the point
        :type sort: ~sortal.sort.sort.sort
        :param x: x-coordinate of the point
        :type x: ~sortal.struct.coord.coordinate_float.coordinate
        :param y: y-coordinate of the point
        :type y: ~sortal.struct.coord.coordinate_float.coordinate
        """
        super(point2D, self).__init__(sort)
        if isinstance(x, vector2D):
            self.value = x
        else:
            vector = vector2D(x,y)
            self.value = vector
                   
    def isNil(self):
        """
        Check if the point's is None or (0,0,0)
        """
        return self.value is None

    def printVal(self):
        """
        Print out a string representing the value of this individual, used in form.printE()
        """
        return str(self.value)
        
    def printRhino(self, dX = 0, dY = 0):
        """
        Print the individual in a format that can be pasted into Rhinoceros for visualization
        """
        return '_Point '+str(self.value.x + dX)+','+str(self.value.y + dY)
         
    def geometricFrame(self):
        """
        Geometric frame of the individual
        """
        return [[self.value.x, self.value.y, None],[self.value.x, self.value.y, None]]
        
    def equals(self, other):
        """
        Check equality between two points

        :param other: other individual to be compared
        :type other: ~sortal.ind.point.point2D
        :returns: boolean -- whether two individuals are equal
        """
        return (self.value == other.value)
        
    def compareValue(self, other):
        """
        Comparing value of 2 points

        :param other: other individual to be compared
        :type other: ~sortal.ind.point.point2D
        :returns: int -- 0 if equals, 1 if this point is bigger, -1 if smaller
        """
        return self.value.compareValue(other.value)   
        
    def replicateInd(self):
        """
        Create a duplicate of the individual without attributes
        
        :returns: the duplicate object
        """
        return self.__class__(self.sort, self.value)    
        
    def transformInd(self, transformSet):
        """
        Transforming the individual with a transformation set (editor)

        :param transformSet: the transformation set to be applied
        :type transformSet: ~sortal.transform.transformSet.transformSet
        :returns: the transformed object
        """
        affine = transformSet.retrieveTransform(affineTransform2D)
        if affine is None:
            newPoint = self.duplicateInd()
        else:
            newPoint = self.__class__(self.sort, self.value.transform(affine))
        transf = transformSet.retrieveTransform(graphTransform)
        if transf is not None:
            newPoint = self.__class__(self.sort, transf.apply(self.value))
        return newPoint
                        

    @staticmethod
    def parse(reader, sort):
        """
        Parse the 2D point from a sortal parseReader (from string)

        :param reader: the reader of the string to be parsed
        :type reader: ~sortal.io.parseReader.parseReader
        :param sort: sort of the new object
        :type sort: ~sortal.sort.sort.sort
        :returns: :class:`~sortal.ind.point.point2D` -- the individual parsed
        """
        num = c1
        inf = False
        if reader.token() == parseReader.MINUS:
            if not reader.previewToken() == parseReader.NUMBER:
                raise ParsingError(reader,'expected a number after "-"')
            num = '-' + reader.newValue() # number
            reader.newToken() # bracket
        elif reader.token() == parseReader.NUMBER: # e.g. 3(2,2,0)
            num = reader.value()
            reader.newToken() # bracket
        
        if reader.token() == '/' and reader.previewToken() == parseReader.NUMBER:
            denom = reader.newValue() # number
            if denom == 0:
                inf = True
                num = c0
            else:
                num = coordinate(num)/coordinate(denom)
            reader.newToken() # bracket
            
        if reader.token() == '(':
            x,y = vector2D.readVector(reader)
            if not isinstance(num, coordinate):
                num = coordinate(num)
            if not inf:
                x = x*num
                y = y*num
            ind = point2D(sort,vector2D(x,y,inf))
        return ind


class point3D(individual):
    """
    The point in 3D class.

    A point is a 0-dimensional geometric data entity with a discrete
    behavior for non-parametric mapping.

    The Point class extends on the Individual class.
    It defines the characteristic individual for points.
    A point is represented as a position vector.

    **Object Parameters:** (not including inherited attributes)
    
    * **value** (:class:`~sortal.struct.vector.vector3D`) -- position vector of the point

    **Class Properties:**

    * **Behavior**: :class:`~sortal.form.discrete.discrete`
    * **Sort Argument**: :class:`~sortal.sort.parameter.argument.argument.argumentType`.NONE
    * **Mapping**: :class:`~sortal.map.similarityMap3D.similarityMap3D`
    """
    # :ivar sortal.vector3D: value : position vector of the point
    # :param value: position vector of the point
    # :type value: ~sortal.struct.vector.vector3D

    type = 'ind-point3D'
    
    def __init__(self, sort, x, y = None, z = None, *args, **kwargs):
        """
        Initialize a 3D point
        
        :param sort: sort of the point
        :type sort: ~sortal.sort.sort.sort
        :param x: x-coordinate of the point
        :type x: ~sortal.struct.coord.coordinate_float.coordinate
        :param y: y-coordinate of the point
        :type y: ~sortal.struct.coord.coordinate_float.coordinate
        :param z: z-coordinate of the point
        :type z: ~sortal.struct.coord.coordinate_float.coordinate
        """
        super(point3D, self).__init__(sort)
        if isinstance(x, vector3D):
            self.value = x
        else:
            vector = vector3D(x,y,z)
            self.value = vector
            
    def isNil(self):
        """
        Check if the point's is None or (0,0,0)
        """
        return self.value is None
        
    def printVal(self):
        """
        Print out a string representing the value of this point, used in form.printE()
        """
        return str(self.value)
        
    def printRhino(self, dX = 0, dY = 0):
        """
        Print the individual in a format that can be pasted into Rhinoceros for visualization
        """
        return '_Point '+str(self.value.x + dX)+','+str(self.value.y + dY)+','+str(self.value.z)
         
    def geometricFrame(self):
        """
        Geometric frame of the individual
        """
        return [[self.value.x, self.value.y, self.value.z],[self.value.x, self.value.y, self.value.z]]
        
    def equals(self, other):
        """
        Check equality between two points

        :param other: other individual to be compared
        :type other: ~sortal.ind.point.point3D
        :returns: boolean -- whether two individuals are equal
        """
        return (self.value == other.value)
        
    def compareValue(self, other):
        """
        Comparing value of 2 points

        :param other: other individual to be compared
        :type other: ~sortal.ind.point.point3D
        :returns: int -- 0 if equals, 1 if this point is bigger, -1 if smaller
        """
        return self.value.compareValue(other.value)   
        
    def replicateInd(self):
        """
        Create a duplicate of the individual without attributes
        
        :returns: the duplicate object
        """
        return self.__class__(self.sort, self.value)    
        
    def transformInd(self, transformSet):
        """
        Transforming the individual with a transformation set (editor)

        :param transformSet: the transformation set to be applied
        :type transformSet: ~sortal.transform.transformSet.transformSet
        :returns: the transformed object
        """

        affine = transformSet.retrieveTransform(affineTransform3D)
        if affine is None:
            newPoint = self.duplicateInd()
        else:
            newPoint = self.__class__(self.sort, self.value.transform(affine))
        transf = transformSet.retrieveTransform(graphTransform3D)
        if transf is not None:
            newPoint = self.__class__(self.sort, transf.apply(self.value))
        return newPoint
    

    @staticmethod
    def parse(reader, sort):
        """
        Parse the point from a sortal parseReader (from string)

        :param reader: the reader of the string to be parsed
        :type reader: ~sortal.io.parseReader.parseReader
        :param sort: sort of the new object
        :type sort: ~sortal.sort.sort.sort
        :returns: :class:`~sortal.ind.point.point2D` -- the individual parsed
        """
        num = c1
        inf = False
        if reader.token() == parseReader.MINUS:
            if not reader.previewToken() == parseReader.NUMBER:
                raise ParsingError(reader,'expected a number after "-"')
            num = '-' + reader.newValue() # number
            reader.newToken() # bracket
        elif reader.token() == parseReader.NUMBER: # e.g. 3(2,2,0)
            num = reader.value()
            reader.newToken() # bracket

        if reader.token() == '/' and reader.previewToken() == parseReader.NUMBER:
            denom = reader.newValue() # number
            if denom == 0:
                inf = True
                num = c0
            else:
                num = coordinate(num)/coordinate(denom)
            reader.newToken() # bracket

        if reader.token() == '(':
            x,y,z = vector3D.readVector(reader)
            if not isinstance(num, coordinate):
                num = coordinate(num)
            if not inf:
                x = x*num
                y = y*num
                z = z*num
            ind = point3D(sort,vector3D(x,y,z,inf))
        return ind
