# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import range
from builtins import super
from builtins import str

from .point import point2D, point3D
from ..struct.vector import vector2D, vector3D
from ..struct.coordinate import coordinate, c0, c1
from ..io.parseReader import parseReader, ParsingError

class pointP2D(point2D):
    """
    parametric point in 2D class.

    **Class Properties:**

    * **Behavior**: :class:`~sortal.form.discrete.discrete`
    * **Sort Argument**: :class:`~sortal.sort.parameter.argument.argument.argumentType`.NONE
    * **Mapping**: :class:`~sortal.map.graphMap.graphMap`
    """

    type = 'ind-pointP2D'
    
    def __init__(self, sort, x, y = None, *args, **kwargs):
        """
        initialize a parametric 2D point. Same as normal point.

        :param sort: sort of the point
        :type sort: ~sortal.sort.sort.sort
        :param x: x-coordinate of the point
        :type x: ~sortal.struct.coord.coordinate_float.coordinate
        :param y: y-coordinate of the point
        :type y: ~sortal.struct.coord.coordinate_float.coordinate
        """
        super(pointP2D, self).__init__(sort, x, y)
    
    def codescriptor(self):
        """
        codescriptor of the point for parametric mapping which is itself/
        """
        return self
        
    @staticmethod
    def parse(reader, sort):
        """
        parse the parametric 2D point from a sortal parseReader (from string)

        :param reader: the reader of the string to be parsed
        :type reader: ~sortal.io.parseReader.parseReader
        :param sort: sort of the new object
        :type sort: ~sortal.sort.sort.sort
        :returns: :class:`~sortal.ind.point.point2D` -- the individual parsed
        """
        num = c1
        inf = False
        if reader.token() == parseReader.MINUS:
            if not reader.previewToken() == parseReader.NUMBER:
                raise ParsingError(reader,'expected a number after "-"')
            num = '-' + reader.newValue() # number
            reader.newToken() # bracket
        elif reader.token() == parseReader.NUMBER: # e.g. 3(2,2,0)
            num = reader.value()
            reader.newToken() # bracket
        
        if reader.token() == '/' and reader.previewToken() == parseReader.NUMBER:
            denom = reader.newValue() # number
            if denom == 0:
                inf = True
                num = c0
            else:
                num = coordinate(num)/coordinate(denom)
            reader.newToken() # bracket
            
        if reader.token() == '(':
            x,y = vector2D.readVector(reader)
            if not isinstance(num, coordinate):
                num = coordinate(num)
            if not inf:
                x = x*num
                y = y*num
            ind = pointP2D(sort,vector2D(x,y,inf))
        return ind
        
class pointP3D(point3D):
    """
    parametric point in 3D class.

    **Class Properties:**

    * **Behavior**: :class:`~sortal.form.discrete.discrete`
    * **Sort Argument**: :class:`~sortal.sort.parameter.argument.argument.argumentType`.NONE
    * **Mapping**: :class:`~sortal.map.graphMap3D.graphMap3D`
    """

    type = 'ind-pointP3D'
    
    def __init__(self, sort, x, y = None, z = None, *args, **kwargs):
        """
        initialize a parametric 2D point. Same as normal point.

        :param sort: sort of the point
        :type sort: ~sortal.sort.sort.sort
        :param x: x-coordinate of the point
        :type x: ~sortal.struct.coord.coordinate_float.coordinate
        :param y: y-coordinate of the point
        :type y: ~sortal.struct.coord.coordinate_float.coordinate
        """
        super(pointP3D, self).__init__(sort, x, y, z)
    
    def codescriptor(self):
        """
        codescriptor of the point for parametric mapping which is itself/
        """
        return self
        
    @staticmethod
    def parse(reader, sort):
        """
        parse the parametric 2D point from a sortal parseReader (from string)

        :param reader: the reader of the string to be parsed
        :type reader: ~sortal.io.parseReader.parseReader
        :param sort: sort of the new object
        :type sort: ~sortal.sort.sort.sort
        :returns: :class:`~sortal.ind.point.point2D` -- the individual parsed
        """
        num = c1
        inf = False
        if reader.token() == parseReader.MINUS:
            if not reader.previewToken() == parseReader.NUMBER:
                raise ParsingError(reader,'expected a number after "-"')
            num = '-' + reader.newValue() # number
            reader.newToken() # bracket
        elif reader.token() == parseReader.NUMBER: # e.g. 3(2,2,0)
            num = reader.value()
            reader.newToken() # bracket
        
        if reader.token() == '/' and reader.previewToken() == parseReader.NUMBER:
            denom = reader.newValue() # number
            if denom == 0:
                inf = True
                num = c0
            else:
                num = coordinate(num)/coordinate(denom)
            reader.newToken() # bracket
            
        if reader.token() == '(':
            x,y,z = vector3D.readVector(reader)
            if not isinstance(num, coordinate):
                num = coordinate(num)
            if not inf:
                x = x*num
                y = y*num
                z = z*num
            ind = pointP3D(sort,vector3D(x,y,z,inf))
        return ind
