# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import


from builtins import super

from builtins import object
class parseReader(object):
    """
    parseReader class for parsing a string. 
    
    When a parseReader of a string is initialized,
    It will go over the string and separate the substrings based on their "tokens":
    the substring type. e.g. strings that start with an alphabet (followed by other alphabets
    or number or "_") would have an IDENTIFIER token. strings enclosed within double-quotes (")
    would have a STRING token etc...

    The substrings, their tokens and start and end indices are saved in a token list (self.tokens)
    in the following format:

    | self.tokens = [
    | [token1, start_index1, end_index1, substring1],
    | [token2, start_index2, end_index2, substring2]
    | [token3, start_index3, end_index3, substring3]
    | ...
    | ]

    to utilize the parseReader, the functions token(), value(), newToken(), newValue(), previewToken()
    and trailerToken() should be used.
    
    **Object Parameters:** 

    * **text** (str) -- the full string of the parse reader.
    * **tokens** -- the token list.

    """
    
    IDENTIFIER = 'a'
    NUMBER = '0'
    STRING = '"'
    LOGICALAND = 'n'
    LOGICALOR = 'o'
    GREATERQ = 'g'
    LESSQ = 'l'
    NOTEQUAL = 't'
    EQUALS = 's'
    MINUS = '-'
    
    formats = {}
    CURRENT = 0
    PREVIEW = 1
    TRAILER = 2    
    
    
    class parseToken(object):

        def __init__(self, reader):
            self.reader = reader
            self.token = 0
            self.line = 0
            self.index = 0
            self.endex = 0
            self.text = self.reader.text
            
        def __str__(self):
            return self.text[self.index - 1: self.endex]
            
        def parse(self):
            self.index = self.endex
            if self.index == len(self.reader.text):
                return None
            if self.advance():
                self.endex = self.span(self.endex)
            return self
            
        def advance(self):
            while (self.text[self.index] == ' ' or self.text[self.index] == '\t' or self.text[self.index] == '\n'):
                self.index += 1
            self.value = self.text[self.index]
            self.endex = self.index + 1
            if ((self.value >= 'a' and self.value <= 'z') or
                (self.value >= 'A' and self.value <= 'Z') or
                 self.value == '_'):
                     self.value = parseReader.IDENTIFIER
            elif self.value >= '0' and self.value <= '9':
                self.value = parseReader.NUMBER
            elif self.value == '"':
                self.value = parseReader.STRING
                
            if self.endex >= len(self.text):
                return True
                
            nextVal = self.text[self.endex]
            if nextVal == self.value:
                if self.value == '/':
                    if self.reader.readln() is None:
                        return False
                    return self.advance()
            
                if self.value == '=':
                    self.value = parseReader.EQUALS
                elif self.value == '&':
                    self.value = parseReader.LOGICALAND
                elif self.value == '|':
                    self.value = parseReader.LOGICALOR
                else:
                    return True
                self.endex += 1
                return True
            if nextVal == '=':
                if self.value == '<':
                    self.value = parseReader.LESSQ
                elif self.value == '>':
                    self.value = parseReader.GREATERQ
                elif self.value == '!':
                    self.value = parseReader.NOTEQUAL
                else:
                    return True
                self.endex += 1
            return True
                                
        def span(self, lead):
            if self.value == parseReader.STRING: # "abcd"
                lead = self.text[lead:].index('"') + lead + 1
                while lead > 0 and self.text[lead-2] == '\\':
                    lead = self.text[lead:].index('"') + lead + 1
                if lead < 1:
                    lead = len(self.text)
                    if self.reader.readln() == None:
                        return lead
                    self.text += '\n' + self.reader.string
                    return self.span(lead)
                return lead
                
            elif self.value == parseReader.IDENTIFIER or self.value == '_': # abcd
                if lead >= len(self.text):
                    return lead
                current = self.text[lead]
                while ((current >= 'a' and current <= 'z') or
                       (current >= 'A' and current <= 'Z') or
                       (current >= '0' and current <= '9') or
                       current == '_' ): # or current == '-'):
                    if lead == len(self.text):
                        return lead
                    lead += 1
                    try:
                        current = self.text[lead]  
                    except IndexError:
                        return lead
                return lead
            
            elif self.value == parseReader.NUMBER: # 1234
                if lead >= len(self.text):
                    return lead
                current = self.text[lead]
                while current >= '0' and current <= '9':
                    if lead == len(self.text):
                        return lead
                    lead += 1
                    try:
                        current = self.text[lead]  
                    except IndexError:
                        return lead
                if current == '.':
                    lead += 1
                    current = self.text[lead]
                    while current >= '0' and current <= '9':
                        if lead == len(self.text):
                            return lead
                        lead += 1
                        try:
                            current = self.text[lead]  
                        except IndexError:
                            return lead
                if current == 'e' or current == 'E':
                    lead += 1
                    current = self.text[lead]
                    if current == '+' or current == parseReader.MINUS:
                        lead += 1
                        current = self.text[lead]
                        while current >= '0' and current <= '9':
                            if lead == len(self.text):
                                return lead
                            lead += 1
                            try:
                                current = self.text[lead]  
                            except IndexError:
                                return lead
                return lead
                
            elif (self.value == parseReader.LOGICALAND or 
                  self.value == parseReader.LOGICALOR or 
                  #self.value == parseReader.LESSQ or 
                  #self.value == parseReader.GREATERQ or 
                  #self.value == parseReader.NOTEQUAL or 
                  self.value == parseReader.EQUALS):
                return lead + 1
            return lead
                            
            
    def __init__(self, text, numLines = 0): 
        self.text = text.strip()
        self.tokens = []
        self.updateTokens()
        self.numLines = numLines
                       
    def __str__(self):
        return self.text
    
    def __len__(self):
        return len(self.tokens)
        
    def token(self):
        """
        get the current token
        """
        try:
            return self.tokens[0][0]
        except:
            return None
        
    def previewToken(self):
        """
        get one token after the current token
        """
        try:
            return self.tokens[1][0]
        except IndexError:
            return None
        
    def trailerToken(self, trailing = 2):
        """
        get trailing token after the current token (trailing is default to 2)
        """
        try:
            return self.tokens[trailing][0]
        except IndexError:
            return None
                
    def value(self):
        """
        get the current substring
        """
        return self.tokens[0][3]
        
    def previewValue(self):
        """
        get the substring after the current token
        """
        try:
            return self.tokens[1][3]
        except IndexError:
            return None
        
    def trailerValue(self, trailing = 2):
        """
        get the trailing substring after the current token (trailing is default to 2)
        """
        try:
            return self.tokens[trailing][3]
        except IndexError:
            return None
                
    def stringVal(self):
        return self.tokens[0][3]
        
    def index(self):
        return self.tokens[0][1]
        
    def rText(self):
        return self.text[self.index():]
        
    def endex(self):
        return self.tokens[0][2]
    
    def newToken(self):
        """
        Advancing the token by removing the first set of token in the token list.
        Return the token of the new current token
        """
        try:
            self.tokens.pop(0)
            return self.tokens[0][0]
        except IndexError:
            return None
            
    def newValue(self):
        """
        Advancing the token by removing the first set of token in the token list.
        Return the substring of the new current token
        """
        try:
            self.tokens.pop(0)
            return self.tokens[0][3]
        except IndexError:
            return None
                            
    def updateTokens(self):
        """
        go through the string to analyze its tokens. Only used in initialization of a parseReader
        """
        token = self.parseToken(self)
        check = token.parse()
        while check is not None:
            self.tokens.append([token.value, token.index, token.endex, self.text[token.index:token.endex]])
            check = token.parse()
        
        
class ParsingError(Exception):
    
    def __init__(self, reader, message, line = None):
        subLines = reader.text[reader.index():].count('\n')
        if reader.numLines - subLines > 0:
            s = ('\n======================================\nPARSING ERROR: Line '+
                 str(reader.numLines - subLines)+'\n...'+reader.text[reader.index():]+ 
                 '\n======================================\nERROR MESSAGE: ' + message)
        else:
            s = ('\n======================================\nPARSING ERROR: Error String: '+
                 str(reader.text)+'\nError: ...'+reader.text[reader.index():]+ 
                 '\n======================================\nERROR MESSAGE: ' + message)
        super(ParsingError, self).__init__(s)


