# -*- coding: utf-8 -*-

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals
from __future__ import absolute_import

version = '__ver_27/11/2021__'

from ..sort.sort import sort
from ..sort.primitiveSort import primitiveSort
from ..sort.attributeSort import attributeSort
from ..sort.disjunctiveSort import disjunctiveSort
from ..sort.compoundSort import compoundSort

from ..struct.vector import vector2D, vector3D
from ..struct.coordinate import coordinate
from ..struct.key import key

#from ..form.form import form

from ..ind.point import point3D
from ..ind.lineSegment import lineSegment3D
from ..ind.planeSegment3D import planeSegment3D
from ..ind.circle import circle3D
from ..ind.ellipse import ellipse3D
from ..ind.circularArc import circularArc3D
from ..ind.ellipticalArc import ellipticalArc3D
from ..ind.bezierCurve import bezierCurve3D
from ..ind.rWeight import rWeight
from ..ind.color import color
from ..ind.description import description

import System
import Rhino
import Rhino.Geometry as rg
import Grasshopper.Kernel as gh
import rhinoscriptsyntax
import System.Drawing.Color as Color
import Grasshopper

from copy import deepcopy
import random
import math

# Defines a Rhino shape object as a combination of a form, an optional Rhino geometry and an optional set of descriptions
class RhinoShape(gh.IGH_BakeAwareData, gh.IGH_PreviewData):
    def __init__(self, sgi_setup, form = None):
        self.sgi_setup = sgi_setup
        self.sorts = self.sgi_setup['sgi_geo_sorts']
        self.labelSorts = self.sgi_setup['sgi_label_sorts']
        self.colorSorts = self.sgi_setup['sgi_color_sorts']
        self.thicknessSorts = self.sgi_setup['sgi_thickness_sorts']
        self.form = None
        self.formP = None
        if form is not None:
            if self.isParametric(form.sort):
                self.formP = form
            else:
                self.form = form
        self.formR = None
        self.geometry = None
        self.bbox = None
        self.colors = None
        self.materials = None
        self.thicknesses = None
        self.descriptions = None
        self.excluded = []
        self.remainder = None
        self.stypes = self.sgi_setup['sgi_geo_types']
        self.pointIndex = self.stypes.index('point')
        self.lineIndex = self.stypes.index('line segment')
        self.planeIndex = self.stypes.index('plane segment')
        self.circleIndex = self.stypes.index('circle')
        self.ellipseIndex = self.stypes.index('ellipse')
        self.arcIndex = self.stypes.index('circular arc')
        self.ellarcIndex = self.stypes.index('elliptical arc')
        self.bezierIndex = self.stypes.index('quadratic Bezier')

    def getForm(self, parametric = False):
        if parametric:
            if self.formP is None:
                self.formP, self.excluded = self.converse(self.form)
            return self.formP
        else:
            if self.form is None:
                self.form = self.converse(self.formP, self.excluded)[0]
            return self.form

    def getGeometry(self):
        if self.geometry is None:
            self.toGeometry()
        return self.geometry

    def getColors(self):
        if self.geometry is None:
            self.toGeometry()
        return self.colors

    def getMaterials(self):
        if self.materials is not None:
            return self.materials
        elif self.geometry is None:
            self.toGeometry()

        self.materials = []
        for colorValue in self.colors:
            material = Rhino.DocObjects.Material()
            if colorValue is not None:
                material.DiffuseColor = colorValue
                #material.SpecularColor = colorValue
                #material.emissionColor = colorValue
            material = Rhino.Render.RenderMaterial.CreateBasicMaterial(material, Rhino.RhinoDoc.ActiveDoc)
            material = gh.Types.GH_Material(material)
            self.materials.append(material)
        return self.materials

    @staticmethod
    def isParametric(sort):
        return sort.name[-1] == 'P'

    def adjustForm(self, sort):
        form = self.getForm(self.isParametric(sort))
        if sort is form.sort:
            return form
        if form.sort.contains(sort):
            rSort = form.sort.complementSort(sort)
            self.formR, self.remainder = form.splitSort(sort, rSort)
        else:
            self.formR = form.switchSort(sort)
        return self.formR

    def readjustForm(self, sort, rShape):
        form = self.getForm(self.isParametric(sort))
        if rShape.remainder is not None:
            form = rShape.remainder.mergeSort(form, sort)
        elif sort is not form.sort:
            rSort = form.sort.complementSort(sort)
            form = form.splitSort(sort, rSort)[0]
        if self.isParametric(sort):
            self.formP = form
        else:
            self.form = form

    # Obtains collective bounding box of list of geometry
    def get_Boundingbox(self, plane = rg.Plane.WorldXY):
        if self.bbox is not None:
            return self.bbox

        try:
            box = None
            for geom in self.getGeometry():
                try:
                    tempBox = geom.GetBoundingBox(plane)
                except:
                    if type(geom) is TextGoo:
                        tempBox =  geom.get_Boundingbox()
                    elif type(geom) is not rg.Point3d: # and type(geom) != type([]):
                        tempBox =  geom.get_Boundingbox()
                    else:
                        tempBox = None
                if box == None:
                    box = tempBox
                elif tempBox != None:
                    box = rg.BoundingBox.Union(box,tempBox)
            self.bbox = box.GetCorners()
        except:
            try:
                self.bbox = rhinoscriptsyntax.BoundingBox(geometries)
            except:
                self.bbox = None
        return self.bbox
        
    def get_ClippingBox(self):
        if self.bbox is None:
            self.get_Boundingbox()
        return self.bbox

    def BoundingBox(self):
        if self.bbox is None:
            self.get_Boundingbox()
        return self.bbox

    def GetBoundingBox(self, plane = rg.Plane.WorldXY):
        if self.bbox is None:
            self.get_Boundingbox(plane)
        return self.bbox

    def get_IsPreviewCapable(self):
        return True

    def getDisplacement(self, factor, Ydir = False):
        if self.bbox is None:
            self.get_Boundingbox()

        distance_x = 0.0
        if self.bbox is not None and not self.sgi_setup['displacementX']:
            distance_x = (self.bbox[1].X - self.bbox[0].X) * factor
        else:
            distance_x = self.sgi_setup['displacementX']
        if distance_x == 0.0:
            distance_x = 5
        if not Ydir:
            return distance_x

        distance_y = 0.0
        if self.bbox is not None and not self.sgi_setup['displacementY']:
            distance_y = (self.bbox[2].Y - self.bbox[1].Y) * factor
        else:
            distance_y = self.sgi_setup['displacementY']
        if distance_y == 0.0:
            distance_y = 5
        return distance_x, distance_y

    def DrawViewportWires(self, args): #args of type GH_PreviewWireArgs
        pipeline = args.Pipeline
        for index, item in enumerate(self.getGeometry()):
            itype = type(item)
            colorValue = self.colors[index]
            if colorValue is None:
                colorValue = args.Color
            lineWidth = self.thicknesses[index]
            if lineWidth is not None and lineWidth > 0:
                lineWidth = (int) (lineWidth * 10) ################
            else:
                lineWidth = args.Thickness
            if itype is rg.Point:
                pipeline.DrawPoint(item.Location, Rhino.Display.PointStyle.X, 5, colorValue)
            elif itype is rg.LineCurve or itype is rg.NurbsCurve or itype is rg.Curve:
                pipeline.DrawCurve(item, colorValue, lineWidth)
            elif itype is rg.Brep:
                pipeline.DrawBrepWires(item, colorValue)
            elif itype is rg.Circle:
                pipeline.DrawCircle(item, colorValue, lineWidth)
            #elif itype is rg.Ellipse:
            #    args.Pipeline.DrawCircle(item, colorValue, lineWidth)
            elif itype is rg.Arc:
                pipeline.DrawArc(item, colorValue, lineWidth)
            elif itype is TextGoo:
                pipeline.Draw3dText(item.m_value, args.Color)
            elif itype is Rhino.Display.Text3d:
                pipeline.Draw3dText(item, args.Color)
            #else:
            #    print("Not addressed: " + str(itype))
      
    def DrawViewportMeshes(self, args): #args of type GH_PreviewMeshArgs
        #display = args.Display
        pipeline = args.Pipeline
        for index, item in enumerate(self.getGeometry()):
            if type(item) is rg.Brep:
                colorValue = self.colors[index]
                if colorValue is not None:
                    material = Rhino.Display.DisplayMaterial(colorValue)
                else:
                    material = args.Material
                pipeline.DrawBrepShaded(item, material)

    def BakeGeometry(self, doc, att, id):
        pass

    # Conversion from Rhino objects to Sortal shape
    def toShape(self, objects, colors, thicknesses, descriptions, refpt, ghenv):
        
        # resulting form
        self.form = sort.register[self.sgi_setup['sort_name']].newForm()
        self.existingKeys = {}
        self.refpt = refpt
        defColor = Color.FromArgb(0, 0, 0)
        
        # Extracts coordinate data
        color_index = 0
        thick_index = 0
        for index, obj in enumerate(objects):
            if type(obj) is Rhino.Display.Text3d:
                continue

            geometricObject = obj.Geometry
            gtype = type(geometricObject)
            #print(gtype)

            color = None
            color_desc = None
            if color_index < len(colors) and type(colors[color_index]) is Color:
                color = colors[color_index]
            else:
                try:
                    if obj.Attributes.ColorSource == Rhino.DocObjects.ObjectColorSource.ColorFromObject:
                        color = obj.Attributes.ObjectColor
                except AttributeError:
                    pass
                if color_index < len(colors) and type(colors[color_index]) is type('a'):
                    color_desc = colors[color_index]
                    if color is None:
                        color = defColor
            thickness = 0
            if thick_index < len(thicknesses):
                try:
                    thickness = float(thicknesses[thick_index])
                except ValueError:
                    pass
            else:
                try:
                    if obj.Attributes.PlotWeightSource == Rhino.DocObjects.ObjectPlotWeightSource.PlotWeightFromObject:
                        thickness = obj.Attributes.PlotWeight
                except AttributeError:
                    pass

            # Points 
            if gtype is rg.Point:
                pt = self.Point3d_to_vector3D(geometricObject.Location)
                individual = point3D(self.sorts[self.pointIndex], pt)
                if geometricObject.UserDictionary.ContainsKey('label'):
                    label = geometricObject.UserDictionary.GetString('label')
                    self.addLabelAttr(individual, self.pointIndex, label)
                elif index + 1 < len(objects) and type(objects[index + 1]) is Rhino.Display.Text3d and geometricObject.Location == objects[index + 1].TextPlane.Origin:
                    self.addLabelAttr(individual, self.pointIndex, objects[index + 1].Text)
                self.form.insert(individual)

            # Line segments
            elif gtype is rg.Line:
                start = self.Point3d_to_vector3D(geometricObject.From)
                end = self.Point3d_to_vector3D(geometricObject.To)
                if start != end:
                    individual = lineSegment3D(self.sorts[self.lineIndex], start, end)
                    if color is not None:
                        self.addColorAttr(individual, self.lineIndex, color, color_desc)
                    if thickness > 0:
                        self.addThickAttr(individual, self.lineIndex, thickness)
                    self.form.insert(individual)
                color_index += 1
                thick_index += 1

            elif gtype is rg.LineCurve:
                start = self.Point3d_to_vector3D(geometricObject.PointAtStart)
                end = self.Point3d_to_vector3D(geometricObject.PointAtEnd)
                if start != end:
                    individual = lineSegment3D(self.sorts[self.lineIndex], start, end)
                    if geometricObject.UserDictionary.ContainsKey('label'):
                        label = geometricObject.UserDictionary.GetString('label')
                        self.addLabelAttr(individual, self.lineIndex, label)
                    if color is not None:
                        self.addColorAttr(individual, self.lineIndex, color, color_desc)
                    if thickness > 0:
                        self.addThickAttr(individual, self.lineIndex, thickness)
                    self.form.insert(individual)
                color_index += 1
                thick_index += 1
                
            # Polylines / PolyCurves
            elif gtype is rg.PolylineCurve or gtype is rg.PolyCurve or gtype is rg.NurbsCurve:
                label = None
                if geometricObject.UserDictionary.ContainsKey('label'):
                    label = geometricObject.UserDictionary.GetString('label')

                if gtype is rg.PolyCurve:
                    curves = geometricObject.Explode()
                else:
                    curves = geometricObject.DuplicateSegments()
                crvCount = 0   # Tracks number of successfully converted curves
                    
                # Separates straight and curved line segments
                for curve in curves:
                    # Handles straight line segments
                    if curve.IsLinear():
                        start = self.Point3d_to_vector3D(curve.PointAtStart)
                        end = self.Point3d_to_vector3D(curve.PointAtEnd)
                        if start != end:
                            individual = lineSegment3D(self.sorts[self.lineIndex], start, end)
                            if label is not None:
                                self.addLabelAttr(individual, self.lineIndex, label)
                            if color is not None:
                                self.addColorAttr(individual, self.lineIndex, color, color_desc)
                            if thickness > 0:
                                self.addThickAttr(individual, self.lineIndex, thickness)
                            self.form.insert(individual)
                        crvCount += 1

                    # Handles circles and ellipses
                    elif curve.IsClosed:
                        if curve.IsCircle:
                            newObj = rg.NurbsCurve.TryGetCircle(curve)[1]
                            center = self.Point3d_to_vector3D(newObj.Center)
                            c_pt1 = self.Point3d_to_vector3D(newObj.PointAt(0))
                            c_pt2 = self.Point3d_to_vector3D(newObj.PointAt(2))
                            #individual = circleSort.baseSort().characteristic(circleSort, center, c_pt1, c_pt2)
                            individual = circle3D(self.sorts[self.circleIndex], center, c_pt1, c_pt2)
                            if label is not None:
                                self.addLabelAttr(individual, self.circleIndex, label)
                            if color is not None:
                                self.addColorAttr(individual, self.circleIndex, color, color_desc)
                            if thickness > 0:
                                self.addThickAttr(individual, self.circleIndex, thickness)
                            self.form.insert(individual)
                            crvCount += 1

                        elif curve.IsEllipse:
                            newObj = rg.NurbsCurve.TryGetEllipse(curve)[1]
                            center = self.Point3d_to_vector3D(newObj.Plane.Origin)
                            normal = self.Point3d_to_vector3D(newObj.Plane.Normal).unit()
                            dist = coordinate(math.sqrt(math.abs(newObj.Radius1**2 - newObj.Radius2**2)))
                            pin_pt = self.Point3d_to_vector3D(curve.PointAt(curve.ClosestPoint(newObj.Plane.Origin)[1]))
                            shortV = center - pin_pt
                            longV = shortV.unit().product(normal)
                            longV = longV.unit().scale(dist)
                            foc1 = center + longV
                            foc2 = center - longV
                            #individual = ellipseSort.baseSort().characteristic((ellipseSort, foc1, foc2, pin_pt)
                            individual = ellipse3D(self.sorts[self.ellipseIndex], foc1, foc2, pin_pt)
                            if label is not None:
                                self.addLabelAttr(individual, self.ellipseIndex, label)
                            if color is not None:
                                self.addColorAttr(individual, self.ellipseIndex, color, color_desc)
                            if thickness > 0:
                                self.addThickAttr(individual, self.ellipseIndex, thickness)
                            self.form.insert(individual)
                            crvCount += 1

                    # Handles arcs
                    elif curve.IsArc():
                        if type(curve) is rg.ArcCurve:
                            arc = curve.Arc
                        else:
                            arc = curve.getTryArc()
                        start = self.Point3d_to_vector3D(arc.StartPoint)
                        end = self.Point3d_to_vector3D(arc.EndPoint)
                        center = self.Point3d_to_vector3D(arc.Center)
                        pin_pt = self.Point3d_to_vector3D(arc.MidPoint) #PointAtNormalizedLength(0.5)
                        #individual = arcSort.baseSort().characteristic(arcSort, center, start, end, pin_pt)
                        individual = circularArc3D(self.sorts[self.arcIndex], center, start, end, pin_pt)
                        if label is not None:
                            self.addLabelAttr(individual, self.arcIndex, label)
                        if color is not None:
                            self.addColorAttr(individual, self.arcIndex, color, color_desc)
                        if thickness > 0:
                            self.addThickAttr(individual, self.arcIndex, thickness)
                        self.form.insert(individual)
                        crvCount += 1

                    # Treats curved line segments as quadratic Bezier curves (parabolic only)
                    elif curve.HasBezierSpans:
                        if curve.Points.Count == 3:
                            try:
                                start = self.Point3d_to_vector3D(curve.PointAtStart)
                                other = self.Point3d_to_vector3D(curve.Points.Item[1].Location)
                                end = self.Point3d_to_vector3D(curve.PointAtEnd)
                                #individual = bezSort.baseSort().characteristic(bezSort, start, other, end)
                                individual = bezierCurve3D(self.sorts[self.bezierIndex], start, other, end)
                                if label is not None:
                                    self.addLabelAttr(individual, self.bezierIndex, label)
                                if color is not None:
                                    self.addColorAttr(individual, self.bezierIndex, color, color_desc)
                                if thickness > 0:
                                    self.addThickAttr(individual, self.bezierIndex, thickness)
                                self.form.insert(individual)
                                crvCount += 1
                            except:
                                pass
                        else:
                            ghenv.Component.AddRuntimeMessage(gh.GH_RuntimeMessageLevel.Warning, "Non-parabolic bezier curves are not supported")
    
                # checks if curves is empty, 
                # in Rhino 6, it happens that the obj type is given as PolyLineCurve when really it is just a LineCurve
                if len(curves) == 0:
                    start = self.Point3d_to_vector3D(geometricObject.PointAtStart)
                    end = self.Point3d_to_vector3D(geometricObject.PointAtEnd)
                    if start != end:
                        individual = lineSegment3D(self.sorts[self.lineIndex], start, end)
                        if label is not None:
                            self.addLabelAttr(individual, self.lineIndex, label)
                        if color is not None:
                            self.addColorAttr(individual, self.lineIndex, color, color_desc)
                        if thickness > 0:
                            self.addThickAttr(individual, self.lineIndex, thickness)
                        self.form.insert(individual)

                # Checks if any curve data was lost from processing straight vs curved line segments
                elif crvCount != len(curves):
                    ghenv.Component.AddRuntimeMessage(gh.GH_RuntimeMessageLevel.Warning, "Input contains unsupported geometries")
                color_index += 1
                thick_index += 1
                    
            # Rectangles
            elif gtype is rg.Rectangle3d:
                end = self.Point3d_to_vector3D(geometricObject.Corner(3))
                for i in range(4):
                    start = end
                    end = self.Point3d_to_vector3D(geometricObject.Corner(i))
                    if start != end:
                        individual = lineSegment3D(self.sorts[self.lineIndex], start, end)
                        if color is not None:
                            self.addColorAttr(individual, self.lineIndex, color, color_desc)
                        if thickness > 0:
                            self.addThickAttr(individual, self.lineIndex, thickness)
                        self.form.insert(individual)
                color_index += 1
                thick_index += 1
            
            # BREPs
            # allows for surfaces with holes (inner)
            elif gtype is rg.Brep:
                form = self.parseBrep(geometricObject, self.sorts[self.planeIndex])
                if geometricObject.UserDictionary.ContainsKey('label'):
                    label = geometricObject.UserDictionary.GetString('label')
                    for coForm in form.elements:
                        for individual in coForm:
                            self.addLabelAttr(individual, self.planeIndex, label)
                            #self.form.insert(individual)
                elif index + 1 < len(objects) and type(objects[index + 1]) is Rhino.Display.Text3d:
                    if geometricObject.GetBoundingBox(rg.Plane.WorldXY).Center == objects[index + 1].TextPlane.Origin:
                        label = objects[index + 1].Text
                        for coForm in form.elements:
                            for individual in coForm:
                                self.addLabelAttr(individual, self.planeIndex, label)
                if color is not None:
                    for coForm in form.elements:
                        for individual in coForm:
                            self.addColorAttr(individual, self.planeIndex, color, color_desc)
                self.form.insert(form)
                color_index += 1

            elif gtype is rg.Extrusion and geometricObject.HasBrepForm:
                form = self.parseBrep(rg.Brep.TryConvertBrep(geometricObject), self.sorts[self.planeIndex])
                if color is not None:
                    for coForm in form.elements:
                        for individual in coForm:
                            self.addColorAttr(individual, self.planeIndex, color, color_desc)
                self.form.insert(form)
                color_index += 1
                       
            # Meshes
            elif gtype is rg.Mesh:
                count =  geometricObject.Faces.Count
                for ind in range(count):
                    face = []
                    for vertex in geometricObject.Faces.GetFaceVertices(ind)[1:]:
                        face.append(self.Point3d_to_vector3D(vertex))
                    individual = planeSegment3D(self.sorts[self.planeIndex], face)
                    if color is not None:
                        self.addColorAttr(individual, self.planeIndex, color, color_desc)
                    self.form.insert(individual)
                color_index += 1
                    
            # Circles, Ellipses
            elif gtype is rg.Circle or gtype is rg.Ellipse: 

                if geometricObject.IsCircle and geometricObject.IsClosed:
                    newObj = rg.NurbsCurve.TryGetCircle(geometricObject)[1]
                    center = self.Point3d_to_vector3D(newObj.Center)
                    c_pt1 = self.Point3d_to_vector3D(newObj.PointAt(0))
                    c_pt2 = self.Point3d_to_vector3D(newObj.PointAt(2))
                    #individual = circleSort.baseSort().characteristic(circleSort, center, c_pt1, c_pt2)
                    individual = circle3D(self.sorts[self.circleIndex], center, c_pt1, c_pt2)
                    if geometricObject.UserDictionary.ContainsKey('label'):
                        label = geometricObject.UserDictionary.GetString('label')
                        self.addLabelAttr(individual, self.circleIndex, label)
                    if color is not None:
                        self.addColorAttr(individual, self.circleIndex, color, color_desc)
                    if thickness > 0:
                        self.addThickAttr(individual, self.circleIndex, thickness)
                    self.form.insert(individual)

                elif geometricObject.IsEllipse and geometricObject.IsClosed:
                    newObj = rg.NurbsCurve.TryGetEllipse(geometricObject)[1]
                    center = self.Point3d_to_vector3D(newObj.Plane.Origin)
                    normal = self.Point3d_to_vector3D(newObj.Plane.Normal).unit()
                    dist = coordinate(math.sqrt(math.abs(newObj.Radius1**2 - newObj.Radius2**2)))
                    pin_pt = self.Point3d_to_vector3D(curve.PointAt(curve.ClosestPoint(newObj.Plane.Origin)[1]))
                    shortV = center - pin_pt
                    longV = shortV.unit().product(normal)
                    longV = longV.unit().scale(dist)
                    foc1 = center + longV
                    foc2 = center - longV
                    #individual = ellipseSort.baseSort().characteristic((ellipseSort, foc1, foc2, pin_pt)
                    individual = ellipse3D(self.sorts[self.ellipseIndex], foc1, foc2, pin_pt)
                    if geometricObject.UserDictionary.ContainsKey('label'):
                        label = geometricObject.UserDictionary.GetString('label')
                        self.addLabelAttr(individual, self.ellipseIndex, label)
                    if color is not None:
                        self.addColorAttr(individual, self.ellipseIndex, color, color_desc)
                    if thickness > 0:
                        self.addThickAttr(individual, self.ellipseIndex, thickness)
                    self.form.insert(individual)
                color_index += 1
                thick_index += 1

            # Circular arcs
            elif gtype is rg.ArcCurve or gtype is rg.Arc:
                if geometricObject.IsCompleteCircle:
                    newObj = geometricObject.TryGetCircle()[1]
                    center = self.Point3d_to_vector3D(newObj.Center)
                    c_pt1 = self.Point3d_to_vector3D(newObj.PointAt(0))
                    c_pt2 = self.Point3d_to_vector3D(newObj.PointAt(2))
                    #individual = circleSort.baseSort().characteristic(circleSort, center, c_pt1, c_pt2)
                    individual = circle3D(self.sorts[self.circleIndex], center, c_pt1, c_pt2)
                    if geometricObject.UserDictionary.ContainsKey('label'):
                        label = geometricObject.UserDictionary.GetString('label')
                        self.addLabelAttr(individual, self.circleIndex, label)
                    if color is not None:
                        self.addColorAttr(individual, self.circleIndex, color, color_desc)
                    if thickness > 0:
                        self.addThickAttr(individual, self.circleIndex, thickness)
                    self.form.insert(individual)

                else:
                    if type(geometricObject) is rg.Arc:
                        start = self.Point3d_to_vector3D(geometricObject.StartPoint)
                        end = self.Point3d_to_vector3D(geometricObject.EndPoint)
                        center = self.Point3d_to_vector3D(geometricObject.Center)
                    else:
                        start = self.Point3d_to_vector3D(geometricObject.PointAtStart)
                        end = self.Point3d_to_vector3D(geometricObject.PointAtEnd)
                        center = self.Point3d_to_vector3D(geometricObject.Arc.Center)
                    pin_pt = self.Point3d_to_vector3D(geometricObject.PointAtNormalizedLength(0.5))
                    #individual = arcSort.baseSort().characteristic(arcSort, center, start, end, pin_pt)
                    individual = circularArc3D(self.sorts[self.arcIndex], center, start, end, pin_pt)
                    if geometricObject.UserDictionary.ContainsKey('label'):
                        label = geometricObject.UserDictionary.GetString('label')
                        self.addLabelAttr(individual, self.arcIndex, label)
                    if color is not None or color_desc is not None:
                        self.addColorAttr(individual, self.arcIndex, color, color_desc)
                    if thickness > 0:
                        self.addThickAttr(individual, self.arcIndex, thickness)
                    self.form.insert(individual)
                color_index += 1
                thick_index += 1

            #
            else:
                ghenv.Component.AddRuntimeMessage(gh.GH_RuntimeMessageLevel.Warning, "Input contains unsupported geometries. You have inputted a "+ str(gtype))

        # Extracts descriptions
        for desc in descriptions:
            i = desc.find(':')
            if i > 0:
                name = desc[0:i]
                if name in sort.register: #and sort.register[self.setup['sort_name']].contains(descSort)
                    descSort = sort.register[name]
                    for text in desc[i+1:].split('|'):
                        individual = description(descSort, '`'+text+'`')
                        self.form.insert(individual)
                else:
                    ghenv.Component.AddRuntimeMessage(gh.GH_RuntimeMessageLevel.Warning, ("The " + name + " description type is not recognized. Please add it to the SGI Setup component."))

        self.form.maximalize()
        self.toGeometry()
        return self
    
    # Extracts point data
    # If refPoint non-zero, then coordinates are extract w.r.t. refPoint
    def Point3d_to_vector3D(self, rhino_pt):
        try:
            x = rhino_pt.X 
            y = rhino_pt.Y 
            z = rhino_pt.Z 
        except:
            x = rhino_pt[0] 
            y = rhino_pt[1]
            z = rhino_pt[2] 

        if self.refpt is not None:
            x -= self.refpt.X
            y -= self.refpt.Y
            z -= self.refpt.Z

        return vector3D(x, y, z)
    
    def parseBrep(self, brep, sort):
        outerForm = sort.newForm()
        innerForm = sort.newForm()
        for curve in brep.Loops:
            # Holes, may be more than 1
            if curve.LoopType == rg.BrepLoopType.Inner:
                inner = []
                for seg in curve.To3dCurve().DuplicateSegments():
                    inner.append(self.Point3d_to_vector3D(seg.PointAtStart))
                inner.append(inner[0])
                innerForm.insert(planeSegment3D(sort, inner))
            # Outer bound, only 1 per brep
            elif curve.LoopType == rg.BrepLoopType.Outer:
                outer = []
                for seg in curve.To3dCurve().DuplicateSegments():
                    outer.append(self.Point3d_to_vector3D(seg.PointAtStart))
                outer.append(outer[0])
                outerForm.insert(planeSegment3D(sort, outer))
        outerForm.difference(innerForm)
        return outerForm
 
    def addLabelAttr(self, individual, spatialIndex, labelText):
        if '|' in labelText:
            labelText = labelText.split('|')
        else:
            labelText = [labelText]

        for label in labelText:
            label = label.strip()
            #if '#' in label:
            if label[0] == '#' and '.' not in label:
                labelKey = label.split('#')[-1]
                if label not in self.existingKeys:
                    key.createKey(labelKey, individual)
                    self.existingKeys[label] = 1
                elif label in self.existingKeys:
                    key.createKey(labelKey + str(self.existingKeys[label]), individual)
                    self.existingKeys[label] += 1
            else:
                newLabel = description(self.labelSorts[spatialIndex], '`'+label+'`')
                individual.addAttribute(newLabel)

    def addColorAttr(self, individual, spatialIndex, colorValue, colorDesc=None):
        colorSort = self.colorSorts[spatialIndex]
        if colorSort is not None:
            mode = self.sgi_setup['colorMode'][spatialIndex]
            if mode == 'graytone':
                tone = (colorValue.R + colorValue.G + colorValue.B)/3
                newColor = rWeight(colorSort, 255 - tone)
            elif mode == 'alphaRGB':
                apha = colorValue.A/255.0
                newColor = color(colorSort, colorValue.R, colorValue.G, colorValue.B, alpha)
            else:
                newColor = color(colorSort, colorValue.R, colorValue.G, colorValue.B)
            if colorDesc is not None:
                newColor.description = description(colorSort, '`'+colorDesc+'`')
            individual.addAttribute(newColor)

    def addThickAttr(self, individual, spatialIndex, thickness):
        thicknessSort = self.thicknessSorts[spatialIndex]
        if thicknessSort is not None:
            newThickness = rWeight(thicknessSort, thickness)
            individual.addAttribute(newThickness)

    
    # Draws the Rhino geometry corresponding to a form and considering the specified setup values
    def toGeometry(self):
        self.geometry = []
        self.colors = []
        self.thicknesses = []
        self.descriptions = []
        form = self.form
        if form is None:
            form = self.formP
        for element in form.elements:
            individuals = element.elements
            if len(individuals) == 0:
                continue 
            
            # Draws points, labeled points, circles, ellipses
            if element.type == 'form-discrete':
                # Draws points and labeled points
                if 'point' in individuals[0].type:
                    for ind in individuals:
                        objs = self.drawPoint(ind)
                        self.geometry.extend(objs)
                        self.colors.extend([None]*len(objs))
                        self.thicknesses.extend([None]*len(objs))
                elif 'circle' in individuals[0].type:
                    for ind in individuals:
                        objs, colorValue, thickness = self.drawCircle(ind)
                        self.geometry.extend(objs)
                        self.colors.extend([colorValue]*len(objs))
                        self.thicknesses.extend([thickness]*len(objs))
                elif 'ellipse' in individuals[0].type:
                    for ind in individuals:
                        objs, colorValue, thickness = self.drawEllipse(ind)
                        self.geometry.extend(objs)
                        self.colors.extend([colorValue]*len(objs))
                        self.thicknesses.extend([thickness]*len(objs))
                elif 'description' in individuals[0].type:
                    text = individuals[0].getValue()[1:-1]
                    for ind in individuals[1:]:
                        text += '|' + ind.getValue()[1:-1]
                    self.descriptions.append(((individuals[0].sort.name + ':' + text)))
            
            # Draws lines
            elif element.type == 'form-interval':
                for ind in individuals:
                    objs, colorValue, thickness = self.drawLineSeg(ind)
                    self.geometry.extend(objs)
                    self.colors.extend([colorValue]*len(objs))
                    self.thicknesses.extend([thickness]*len(objs))
            
            # Draws planes
            elif element.type == 'form-areal' or element.type == 'form-coForm':
                for ind in individuals:
                    if ind.type == 'form-coForm':
                        for object in ind.elements:
                            objs, colorValue = self.drawPlaneSegment(object)
                            for item in objs:
                                itype = type(item)
                                #print(itype)
                                if itype is TextGoo or itype is rg.Point3d or itype is rg.Brep:
                                    self.geometry.append(item)
                                    self.colors.append(colorValue)
                                    self.thicknesses.append(None)
                                else:
                                    self.geometry.extend(item)
                                    self.colors.extend([colorValue]*len(item))
                                    self.thicknesses.extend([None]*len(item))
                    #elif ind.type == 'ind-planeSegment3D':
                    #    self.geometry.extend(self.PlaneSegment(ind))
            
            # Draws circular and elliptical arcs
            elif element.type == 'form-cyclical':
                if 'circular' in individuals[0].type:
                    for ind in individuals:
                        objs, colorValue, thickness = self.drawArc(ind)
                        self.geometry.extend(objs)
                        self.colors.extend([colorValue]*len(objs))
                        self.thicknesses.extend([thickness]*len(objs))
                elif element.type == 'form-cyclical' and 'elliptical' in individuals[0].type:
                    for ind in individuals:
                        objs, colorValue, thickness = self.drawEArc(ind)
                        self.geometry.extend(objs)
                        self.colors.extend([colorValue]*len(objs))
                        self.thicknesses.extend([thickness]*len(objs))

            # Draws bezier curves
            elif element.type == 'form-curve':
                for ind in individuals:
                    objs, colorValue, thickness = self.drawBezier(ind)
                    self.geometry.extend(objs)
                    self.colors.extend([colorValue]*len(objs))
                    self.thicknesses.extend([thickness]*len(objs))

        self.get_ClippingBox()

    def Attributes(self, individual, spatialIndex):
        labels = []
        colorValue = None
        thickness = None
        if individual.key != None:
            labels.append('#'+individual.key.name)
        if individual.hasAttribute():
            attr = individual.attribute
            if attr.type == 'form-discrete':
                for element in attr.elements:
                    labels.append(element.getValue()[1:-1])
            else:
                for form in attr.elements:
                    if form.type == 'form-discrete':
                        for element in form.elements:
                            labels.append(element.getValue()[1:-1])
                    elif form.type == 'form-ordinal':
                        if 'Color' in form.sort.name:
                            tone = 255 - int(form.individual.value)
                            colorValue = Color.FromArgb(tone, tone, tone)
                        else:
                            thickness = form.individual.value
                    elif self.sgi_setup['colorMode'][spatialIndex] == 'alphaRGB':
                        rgb = form.individual.components
                        colorValue = Color.FromArgb(form.individual.alpha, rgb[0].value, rgb[1].value, rgb[2].value)
                    else:
                        rgb = form.individual.components
                        colorValue = Color.FromArgb(rgb[0].value, rgb[1].value, rgb[2].value)
        return '|'.join(labels), colorValue, thickness

    # Plots points and labeled points
    def drawPoint(self, individual):
        pt = rg.Point3d(float(individual.value.x.value), float(individual.value.y.value), float(individual.value.z.value))
        pt = rg.Point(pt)
        label, colorValue, thickness = self.Attributes(individual, self.pointIndex)
        if label != '':
            newPlane = rg.Plane(pt.Location, rg.Vector3d(0,0,1))
            textHt = self.sgi_setup['text_size']
            textPt = TextGoo(Rhino.Display.Text3d(label, newPlane, textHt))
            #pt.UserDictionary.Set('label', label)
            return [pt, textPt]
        else:
            return [pt]
    
    # Plots line segments and labeled line segments
    def drawLineSeg(self, individual):
        start = rg.Point3d(float(individual.start.x.value), float(individual.start.y.value), float(individual.start.z.value))
        end = rg.Point3d(float(individual.end.x.value), float(individual.end.y.value), float(individual.end.z.value))
        line = rg.LineCurve(start, end)
        label, colorValue, thickness = self.Attributes(individual, self.lineIndex)
        if label != '':
            newPlane = rg.Plane((start+end)/2, rg.Vector3d(0,0,1))
            textHt = self.sgi_setup['text_size']
            textPt = TextGoo(Rhino.Display.Text3d(label, newPlane, textHt))
            line.UserDictionary.Set('label', label)
            return [line, textPt], colorValue, thickness
        else:
            return [line], colorValue, thickness
    
    # Creates plane segment representation (BREP, allows for trims)
    def drawPlaneSegment(self, individual):
        #normal = rg.Vector3d(individual.normal.x, individual.normal.y, individual.normal.z)
        #origin = rg.Point3d(individual.pRoot.x, individual.pRoot.y, individual.pRoot.z)
        #planeSegment = rg.Plane(origin, normal)
        
        # Extracts outer bound (1 only)
        outer = []
        for point in individual.boundaries[0].polygonal:
            outer.append(rg.Point3d(point.x, point.y, point.z))
        outer.append(outer[0])
        outerBound = rg.PolylineCurve(outer)
        
        # Extracts inner loops (may be greater than 1)
        trims = []
        for bound in individual.boundaries[1:]:
            inner = []
            for point in bound.polygonal:
                inner.append(rg.Point3d(point.x, point.y, point.z))
            inner.append(inner[0])
            trims.append(rg.PolylineCurve(inner))

        if len(trims) > 0:
            trims.append(outerBound)
            planes = rg.Brep.CreatePlanarBreps(trims, 0.1)
        else:
            planes = rg.Brep.CreatePlanarBreps(outerBound, 0.1)

        label, colorValue, thickness = self.Attributes(individual, self.planeIndex)
        if label != '':
            bb = rg.BoundingBox(outer)
            newPlane = rg.Plane(bb.Center, rg.Vector3d(0,0,1))
            textHt = self.sgi_setup['text_size']
            textPt = TextGoo(Rhino.Display.Text3d(label, newPlane, textHt))
            for item in planes:
                if type(item) is rg.Brep:
                    item.UserDictionary.Set('label', label)
                else:
                    item.Geometry.UserDictionary.Set('label', label)
            return [planes, textPt], colorValue
        else:
            return [planes], colorValue
        
    def drawCircle(self, individual):
        center = rg.Point3d(individual.center.x, individual.center.y, individual.center.z)
        points = individual.getPoints()
        pin_1 = rg.Point3d(points[1].x, points[1].y, points[1].z)
        pin_2 = rg.Point3d(points[2].x, points[2].y, points[2].z)
        radius = individual.radius
        plane = rg.Plane(center, pin_1, pin_2)
        circle = rg.Circle(plane, radius)
        label, colorValue, thickness = self.Attributes(individual, self.circleIndex)
        if label != '':
            newPlane = rg.Plane(pin_1, rg.Vector3d(0,0,1))
            textHt = self.sgi_setup['text_size']
            textPt = TextGoo(Rhino.Display.Text3d(label, newPlane, textHt))
            circle.UserDictionary.Set('label', label)
            return [circle, textPt], colorValue, thickness
        else:
            return [circle], colorValue, thickness
        
    def drawEllipse(self, individual):
        center = rg.Point3d(individual.center.x, individual.center.y, individual.center.z)
        points = individual.getPoints()
        foc_1 = rg.Point3d(points[0].x, points[0].y, points[0].z)
        foc_2 = rg.Point3d(points[1].x, points[1].y, points[1].z)
        shortPt = rg.Point3d(points[2].x, points[2].y, points[2].z)
    
        cDist = foc_1.DistanceTo(foc_2)/2
        bDist =  center.DistanceTo(shortPt)
        aDist = math.sqrt(cDist**2 + bDist**2)
        shortVec = shortPt - center
        shortDist = math.sqrt(shortVec.X**2 + shortVec.Y**2 + shortVec.Z**2)
        shortUnit = rg.Vector3d(shortVec.X/shortDist, shortVec.Y/shortDist, shortVec.Z/shortDist)
        normVec = rg.Vector3d(eSource.normal.x, eSource.normal.y, eSource.normal.z)
        normDist = math.sqrt(normVec.X**2 + normVec.Y**2 + normVec.Z**2)
        normUnit = rg.Vector3d(normVec.X/normDist, normVec.Y/normDist, normVec.Z/normDist)
        longUnit = rg.Vector3d.CrossProduct(shortUnit, normUnit)
        longPt = center + rg.Vector3d.Multiply(longUnit, aDist)
        
        ellipse = rg.Ellipse(center, longPt, shortPt)
        label, colorValue, thickness = self.Attributes(individual, self.ellipseIndex)
        if label != '':
            newPlane = rg.Plane(longPt, rg.Vector3d(0,0,1))
            textHt = self.sgi_setup['text_size']
            textPt = TextGoo(Rhino.Display.Text3d(label, newPlane, textHt))
            ellipse.UserDictionary.Set('label', label)
            return [ellipse, textPt], colorValue, thickness
        else:
            return [ellipse], colorValue, thickness
        
    def drawArc(self, individual):
        start = rg.Point3d(individual.start.x, individual.start.y, individual.start.z)
        end = rg.Point3d(individual.end.x, individual.end.y, individual.end.z)
        pin = rg.Point3d(individual.pin.x, individual.pin.y, individual.pin.z)
        center = rg.Point3d(individual.center.x, individual.center.y, individual.center.z)
        angle = individual.angle()

        arc = rg.Arc(start, pin, end)
        if not arc.IsValid:
            pt = rg.Point3d.Add(center, rg.Point3d.Subtract(center, start))
            circle = rg.Circle(start, end, pt)
            arc = rg.Arc(circle, angle)
        label, colorValue, thickness = self.Attributes(individual, self.arcIndex)                    
        if label != '':
            newPlane = rg.Plane(arc.MidPoint, rg.Vector3d(0,0,1))
            textHt = self.sgi_setup['text_size']
            textPt = TextGoo(Rhino.Display.Text3d(label, newPlane, textHt))
            arc.UserDictionary.Set('label', label)
            return [arc, textPt], colorValue, thickness
        else:
            return [arc], colorValue, thickness
    
    def drawEArc(self, individual):
        # Preparation of data for computation and elliptical arc creation
        center = rg.Point3d(individual.center.x, individual.center.y, individual.center.z)
        start = rg.Point3d(individual.start.x, individual.start.y, individual.start.z)
        end = rg.Point3d(individual.end.x, individual.end.y, individual.end.z)
        shortPt = rg.Point3d(individual.quadPoint.x, individual.quadPoint.y, individual.quadPoint.z)
        points = individual.getPoints()
        foc_1 = rg.Point3d(points[0].x, points[0].y, points[0].z)
        foc_2 = rg.Point3d(points[1].x, points[1].y, points[1].z)
        
        # Computation of major, minor, foci distances    
        cDist = foc_1.DistanceTo(foc_2)/2
        bDist =  center.DistanceTo(shortPt)
        aDist = math.sqrt(cDist**2 + bDist**2)
    
        # Computation of minor axis (shortPt) and major axis (longPt) quad points
        shortVec = shortPt - center
        shortDist = math.sqrt(shortVec.X**2 + shortVec.Y**2 + shortVec.Z**2)
        shortUnit = rg.Vector3d(shortVec.X/shortDist, shortVec.Y/shortDist, shortVec.Z/shortDist)
        normVec = rg.Vector3d(eArcSource.normal.x, eArcSource.normal.y, eArcSource.normal.z)
        normDist = math.sqrt(normVec.X**2 + normVec.Y**2 + normVec.Z**2)
        normUnit = rg.Vector3d(normVec.X/normDist, normVec.Y/normDist, normVec.Z/normDist)
        longUnit = rg.Vector3d.CrossProduct(shortUnit, normUnit)
        longPt = center + rg.Vector3d.Multiply(longUnit, aDist)

        eArc = rg.Ellipse(center, longPt, shortPt)
        if not eArc.IsValid:
            return []
        eArc = ellipse.ToNurbsCurve()
        eArc.Trim(tMin, tMax)
        label, colorValue, thickness = self.Attributes(individual, self.ellarcIndex)                    
        if label != '':
            newPlane = rg.Plane(eArc.MidPoint, rg.Vector3d(0,0,1))
            textHt = self.sgi_setup['text_size']
            textPt = TextGoo(Rhino.Display.Text3d(label, newPlane, textHt))
            eArc.UserDictionary.Set('label', label)
            return [eArc, textPt], colorValue, thickness
        else:
            return [eArc], colorValue, thickness
    
    def drawBezier(self, individual):
        start = rg.Point3d(individual.start.x, individual.start.y, individual.start.z)
        end = rg.Point3d(individual.end.x, individual.end.y, individual.end.z)
        ctrl = rg.Point3d(individual.controlPoint.x, individual.controlPoint.y, individual.controlPoint.z)
        bez = rg.Curve.CreateControlPointCurve((start, ctrl, end))
        label, colorValue, thickness = self.Attributes(individual, self.bezierIndex)                    
        if label != '':
            newPlane = rg.Plane(bez.PointAtStart, rg.Vector3d(0,0,1))
            textHt = self.sgi_setup['text_size']
            textPt = TextGoo(Rhino.Display.Text3d(label, newPlane, textHt))
            bez.UserDictionary.Set('label', label)
            return [bez, textPt], colorValue, thickness
        else:
            return [bez], colorValue, thickness


    # Takes non-parametric sortal metaform as input and converts it into a parametric sortal object or the other way around
    def converse(self, metaform, reinsertion = []):
        exclusion = []
        targetSorts = {}

        if metaform.sort.name == self.sgi_setup['sort_name']:
            converseSort = sort.register[self.sgi_setup['sort_name'] + 'P']
            for index, a in enumerate(self.sgi_setup['sgi_geo_sorts']):
                targetSorts[a.name] = self.sgi_setup['sgi_pgeo_sorts'][index]
        else:
            converseSort = sort.register[self.sgi_setup['sort_name']]
            for index, a in enumerate(self.sgi_setup['sgi_pgeo_sorts']):
                if a is not None:
                    targetSorts[a.name] = self.sgi_setup['sgi_geo_sorts'][index]

        #for a in sort.register[sortName].component:
        #    if a.name not in self.objects and 'description' not in sort.register[a.name].baseSort().characteristic.type:
        #        self.objects[a.name] = []
             
        converseForm = converseSort.newForm()
        for form in metaform.elements:
            if len(form.elements) > 0:
                # For descriptions
                if form.elements[0].type == 'ind-description':
                    # Appends description individual to corresponding dictionary item inside self.objects['descriptions']
                    for ind in form.elements:
                        converseForm.insert(ind)

                elif targetSorts[form.sort.name] is None:
                    exclusion.append(form)

                else:
                    newSort = targetSorts[form.sort.name]
                    newForm = newSort.newForm()

                    # Converts points/ellipses/circles
                    if form.type == 'form-discrete':
                        if 'point' in form.elements[0].type:
                            for ind in form.elements:
                                newInd = newSort.baseSort().characteristic(newSort, ind.value)
                                if ind.hasAttribute():
                                    newInd.addAttribute(ind.attribute)
                                if ind.key is not None:
                                    key.createKey(ind.key.name, newInd)
                                newForm.insert(newInd)
                        elif 'circle' in form.elements[0].type:
                            for ind in form.elements:
                                points = ind.getPoints()
                                newInd = newSort.baseSort().characteristic(newSort, ind.center, points[1], points[2])
                                if ind.hasAttribute():
                                    newInd.addAttribute(ind.attribute)
                                if ind.key is not None:
                                    key.createKey(ind.key.name, newInd)
                                newForm.insert(newInd)
                        elif 'ellipse' in form.elements[0].type:
                            for ind in form.elements:
                                points = ind.getPoints()
                                newInd = newSort.baseSort().characteristic(newSort, points[0], points[1], points[2])
                                if ind.hasAttribute():
                                    newInd.addAttribute(ind.attribute)
                                if ind.key is not None:
                                    key.createKey(ind.key.name, newInd)
                                newForm.insert(newInd)

                    # Converts lines
                    elif form.type == 'form-interval':
                        for ind in form:
                            newInd = newSort.baseSort().characteristic(newSort, ind.start, ind.end)
                            if ind.hasAttribute():
                                newInd.addAttribute(ind.attribute)
                            if ind.key is not None:
                                key.createKey(ind.key.name, newInd)
                            newForm.insert(newInd)
            
                    elif form.type == 'form-curve':
                        for ind in form:
                            newInd = newSort.baseSort().characteristic(newSort, ind.start, ind.controlPoint, ind.end)
                            if ind.hasAttribute():
                                newInd.addAttribute(ind.attribute)
                            if ind.key is not None:
                                key.createKey(ind.key.name, newInd)
                            newForm.insert(newInd)

                    # Excludes elliptical/circular arcs
                    elif form.type == 'form-cyclical':
                        if 'circular' in form.elements[0].type:
                            for ind in form:
                                newInd = newSort.baseSort().characteristic(newSort, ind.center, ind.start, ind.end, ind.pin)
                                if ind.hasAttribute():
                                    newInd.addAttribute(ind.attribute)
                                if ind.key is not None:
                                    key.createKey(ind.key.name, newInd)
                                newForm.insert(newInd)
                        elif 'elliptical' in form.elements[0].type:
                            for ind in form:
                                newInd = newSort.baseSort().characteristic(newSort, ind.foci[0], ind.foci[1], ind.start, ind.end, ind.pin, ind.quadPoint)
                                if ind.hasAttribute():
                                    newInd.addAttribute(ind.attribute)
                                if ind.key is not None:
                                    key.createKey(ind.key.name, newInd)
                                newForm.insert(newInd)
                
                    # Converts plane segments
                    elif form.type == 'form-areal':
                        for coform in form.elements:
                            for ind in coform.elements:
                                newInd = newSort.baseSort().characteristic(newSort, ind.boundaries, True)
                                if ind.hasAttribute():
                                    newInd.addAttribute(ind.attribute)
                                if ind.key is not None:
                                    key.createKey(ind.key.name, newInd)
                                newForm.insert(newInd)
                
                    converseForm.insert(newForm)

        # Adds in circular/elliptical arcs if conversion is from parametric to non-parametric
        for form in reinsertion:
            converseForm.insert(form)

        return converseForm, exclusion


    # Extracts point data
    # If refPoint non-zero, then coordinates are extract w.r.t. refPoint
    @staticmethod
    def Point3d_to_triple(rhino_pt, refpt):
        try:
            x = rhino_pt.X 
            y = rhino_pt.Y 
            z = rhino_pt.Z 
        except:
            x = rhino_pt[0] 
            y = rhino_pt[1]
            z = rhino_pt[2] 

        if refpt is not None:
            x -= refpt.X
            y -= refpt.Y
            z -= refpt.Z

        return [x, y, z]

    @staticmethod
    def toPolygonal(objects, refpt, ghenv):
        polygons = []
        if len(objects) == 0:
            return polygons

        geometricObject = objects[0]
        if type(geometricObject) is rg.Point3d or type(geometricObject) is rg.LineCurve or type(geometricObject) is rg.Line:
            objects = [objects]

        # Extracts coordinate data
        for geometricObject in objects:
            viable = False
            
            # Polylines / Elliptical arcs
            if type(geometricObject) is rg.PolylineCurve:
                points = []
                crvCount = 0   # Tracks number of successfully converted curves
                curves = geometricObject.DuplicateSegments()
                for curve in curves:
                    # Handles straight line segments
                    if curve.IsLinear():
                        points.append(RhinoShape.Point3d_to_triple(curve.PointAtStart, refpt))
                        crvCount += 1
                # Checks if any curve data was lost from processing straight line segments
                if crvCount == len(curves):
                    polygons.append(points)
                    viable = True
                    
            # Rectangles OK
            elif type(geometricObject) is rg.Rectangle3d:
                points = []
                for i in range(4):
                    points.append(RhinoShape.Point3d_to_triple(geometricObject.Corner(i), refpt))
                polygons.append(points)
                viable = True
                
            # Surfaces, BREPs
            elif type(geometricObject) is rg.Brep:
                points = []
                for curve in geometricObject.Loops:
                    if curve.LoopType == rg.BrepLoopType.Outer:
                        thing = curve.To3dCurve()
                        curves = thing.DuplicateSegments()
                        for curve in curves:
                            points.append(RhinoShape.Point3d_to_triple(curve.PointAtStart, refpt))
                polygons.append(points)
                viable = True

            # points and Line segments
            elif type(geometricObject) is type([]) and len(geometricObject) > 0:
                points = []
                if type(geometricObject[0]) is rg.Point3d:
                    viable = True
                    for pt in geometricObject:
                        if type(pt) is rg.Point3d:
                            points.append(RhinoShape.Point3d_to_triple(pt, refpt))
                        else:
                            ghenv.Component.AddRuntimeMessage(gh.GH_RuntimeMessageLevel.Warning, "Expected a series of points. You have inputted a " + str(type(pt)))
                            break
                    if viable and points[0] == points[-1]:
                        points = points[:-1]
                    polygons.append(points)
                else:
                    viable = True
                    for line in geometricObject:
                        if type(geometricObject) is rg.LineCurve:
                            points.append(RhinoShape.Point3d_to_triple(geometricObject.PointAtStart, refpt))
                            points.append(RhinoShape.Point3d_to_triple(geometricObject.PointAtEnd, refpt))
                        elif type(geometricObject) is rg.Line:
                            points.append(RhinoShape.Point3d_to_triple(geometricObject.From, refpt))
                            points.append(RhinoShape.Point3d_to_triple(geometricObject.To, refpt))
                        else:
                            ghenv.Component.AddRuntimeMessage(gh.GH_RuntimeMessageLevel.Warning, "Expected a series of lines. You have inputted a " + str(type(line)))
                            break
                    pts = [points.pop(0)]
                    pt = points.pop(0)
                    while len(points) > 1:
                        i = points.index(pt)
                        pts.append(points.pop(i))
                        if i % 2 == 0:
                            pt = points.pop(i-1)
                        else:
                            pt = points.pop(i)
                    polygons.append(pts)
                
            if not viable:
                ghenv.Component.AddRuntimeMessage(gh.GH_RuntimeMessageLevel.Warning, "Input contains unsupported geometries. You have inputted a " + str(type(geometricObject)))
        
        return polygons


# Previews and bakes Text Objects
class TextGoo(gh.Types.GH_GeometricGoo[Rhino.Display.Text3d], gh.IGH_BakeAwareData, gh.IGH_PreviewData):
    #region construction
    def __init__(self, text):
        self.m_value = text
    
    @staticmethod
    def DuplicateText3d(original):
        if original is None: return None
        text = Rhino.Display.Text3d(original.Text, original.TextPlane, original.Height)
        #text.Bold = original.Bold,
        #text.Italic = original.Italic,
        text.FontFace = original.FontFace
        return text
    
    def DuplicateGeometry(self):
        return TextGoo(TextGoo.DuplicateText3d(self.m_value))
    
    #region properties
    def get_TypeName(self):
        return "3D Text"
        
    def get_TypeDescription(self):
        return "3D Text"
    
    def ToString(self):
        if self.m_value is None: return "<null>"
        return self.m_value.Text
        
    def get_Boundingbox(self):
        if self.m_value is None:
            return Rhino.Geometry.BoundingBox.Empty;
        return self.m_value.BoundingBox;
        
    def GetBoundingBox(self, xform):
        if self.m_value is None:
            return Rhino.Geometry.BoundingBox.Empty
        box = self.m_value.BoundingBox
        corners = xform.TransformList(box.GetCorners())
        return Rhino.Geometry.BoundingBox(corners)
    
    #region methods
    def Transform(self, xform):
        text = TextGoo.DuplicateText3d(self.m_value)
        if text is None: return TextGoo(None)
        
        plane = text.TextPlane
        point = plane.PointAt(1, 1)

        plane.Transform(xform)
        point.Transform(xform)
        dd = point.DistanceTo(plane.Origin)
        
        text.TextPlane = plane;
        text.Height *= dd / math.sqrt(2)
        return TextGoo(text)
        
    def Morph(self, xmorph):
        return self.DuplicateGeometry()

    #region preview
    def get_ClippingBox(self):
        return self.get_Boundingbox()
        
    def DrawViewportWires(self, args):
        if self.m_value is None: return
        args.Pipeline.Draw3dText(self.m_value, args.Color)
      
    def DrawViewportMeshes(self, args):
        # Do not draw in meshing layer.
        pass

    #region baking
    def BakeGeometry(self, doc, att, id):
        try:
            id = System.Guid.Empty
            
            if self.m_value is None:
                return false, id
            
            if att is None:
                att = doc.CreateDefaultAttributes()
            
            id = doc.Objects.AddText(self.m_value, att)
            
            return True, id
        except Exception as ex:
            System.Windows.Forms.MessageBox.Show(str(ex))
            return False, System.Guid.Empty
            draw
