# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import range
from builtins import super
from builtins import str

import itertools
from copy import copy
from random import randint

from .mapping import mapping
from .graphNode import graphNode

from ..io.parseReader import parseReader
from ..transform.graphTransform import graphTransform
from ..transform.paramSet import paramSet
from ..transform.transformSet import transformSet

from ..ind.description import description
from ..ind.line import line2D
from ..ind.planeSegment import planeSegment2D

from ..struct.planeStruct import bSegment, boundary
from ..struct.coordinate import coordinate, c0, c1, cn1, c0p5, pi, cos, sin, acos
from ..struct.vector import vector2D

vz = vector2D(0,0)

from ..struct.key import key

from ..sort.sort import sort as sortClass

ANY = 0
PERPENDICULAR = 1
PARALLEL = 2

OUTSIDE = 0
BOUND = 1
INSIDE = 2
STRICT_BOUND = 3

NONE = '[None]'

# ANY = '~'
# PARALLEL = '='
# PERPENDICULAR = '+'

# OUTSIDE = '0'
# BOUND = 'b'
# INSIDE = 'i'
# STRICT_BOUND = 's'

# NONE = '[--------]'




class graphMap(mapping):
    """
    GraphMap: mapping for 2D parametric Geometry matching. 
    
    The matching process is based on "nodes": codescriptor of all line-segments in 
    the lhs and the shape. Line-segments that are co-linear within a shape should be 
    under the same node. The relation between the nodes within the shapes (parallel, perpendicular, any) as well
    as the position of the nodes' intersections on their line-segments (outside, bound, inside) 
    are saved into a node matrix. All other Geometry Objects (point, plane-segment, etc ...) are added
    into the Graph as auxiliary objects to be matched in various specific point in the matching process.

    For matching procedure:

    #. Match all free points (points that are not contained by any node -> unconstrained)
    #. The node matrices of the rule lhs and the shape would be compared to find all combination of matching nodes -> all intersections are matched
    #. Check ordering of all matched intersections on their respective nodes
    #. Match all line end-points that are not node intersections as well as points that are contained by only one node (thus not intersection)
    #. Match all curves and plane-segments
    """
    type = 'map-graphMap'    

    def __init__(self, *args, **kwargs):
        """
        Initiating a new graphMap
        """
        super(graphMap, self).__init__()     
        self.nodeList = graphNode()
        self.edgeMatrices = []
        self.matchingNodeList = graphNode()
        self.matchingEdgeMatrices = []
        self.protoTransform = graphTransform()
        self.colinearInts = []

        self.rhsLines = []
        self.rhsSolvingOrder = []
        self.lhsResolved = None
        self.rhsResolved = []
        
        self.predicates = []
        self.directives = []
        self.postCheck = False

    def isValid(self):
        """
        Check if the mapping is valid
        """
        return True

    def clear(self):
        """
        Clear the previous detection data
        """
        self.matchingNodeList.clear()
        self.matchingEdgeMatrices = []
        self.matchingIntersections = []
        
    @staticmethod
    def getInstance():
        """
        Get a new graphMap instance
        """
        return graphMap()
        
        
    def addPredicate(self, cond, keyList):
        """
        Add predicate to the predicate List
        """
        cond = cond.lower()
        self.predicates.append([cond,keyList])
    
    def clearPredicates(self):
        """
        Clear the predicate list
        """
        self.predicates = []
        
    def addDirective(self, cond, keyList):
        """
        Add directive to the directive List
        """
        cond = cond.lower()
        self.directives.append([cond,keyList])
    
    def clearDirectives(self):
        """
        Clear the directive list
        """
        self.directives = []

    # ============================================================================================
    #                                      ANALYZING PROCESS
    # ============================================================================================

    def generateNodes(self, ind, matching = False):
        """
        add individual into the graph
        """
        if matching:
            nodeList = self.matchingNodeList
            # edgeMatrices = self.matchingEdgeMatrices
        else:
            nodeList = self.nodeList
            # edgeMatrices = self.edgeMatrices
            
        # for point, plane-segment and curves
        if ind.type == 'ind-pointP2D':
            nodeList.LHS_addPoint(ind)
            if not matching:
                self.protoTransform.addUnresolved(ind.value)
            return nodeList
        elif ind.type == 'ind-lineSegmentP2D':
            element = ind.parametricNode()
            nodeLen = len(self.nodeList.lines)
            nodeList.addInstance(element, ind)
            if not matching:
                self.protoTransform.addUnresolved(ind.start)
                self.protoTransform.addUnresolved(ind.end)
                if nodeLen < len(self.nodeList.lines):
                    newNode = self.nodeList.lines[-1]
                    for ln in range(len(self.nodeList.lines) - 1):
                        if not self.nodeList.lines[ln].nodeElement.intersect(newNode.nodeElement):
                            continue
                        intersection = self.nodeList.lines[ln].nodeElement.intersection(newNode.nodeElement)
                        notMin = False
                        notMax = False
                        for lnI in self.nodeList.lines[ln].forms:
                            if intersection > lnI.start:
                                notMin = True
                            if intersection < lnI.end:
                                notMax = True
                            if notMin and notMax:
                                break
                        if notMin and notMax:
                            self.protoTransform.addUnresolved(intersection)
                            continue
                        notMin = False
                        notMax = False
                        for lnJ in newNode.forms:
                            if intersection > lnJ.start:
                                notMin = True
                            if intersection < lnJ.end:
                                notMax = True
                            if notMin and notMax:
                                break
                        if notMin and notMax:
                            self.protoTransform.addUnresolved(intersection)



            return nodeList
        elif ind.type == 'ind-planeSegmentP2D':
            nodeList.LHS_addPlane(ind)
            if not matching:
                for b in ind.boundaries:
                    for v in b.polygonal: 
                        self.protoTransform.addUnresolved(v)
            return nodeList
        elif ind.type == 'ind-circleP2D' or ind.type == 'ind-ellipseP2D' or ind.type == 'ind-bezierCurveP2D':
            nodeList.LHS_addCurve(ind)
            return nodeList

        return nodeList

    def analyzeColinear(self):
        if len(self.nodeList.lines) < 3: return

        for i in range(len(self.nodeList.lines) - 2):
            li = self.nodeList.lines[i]
            for j in range(i + 1, len(self.nodeList.lines) - 1):
                if self.edgeMatrices[i][j][0] != PARALLEL:
                    continue
                lj = self.nodeList.lines[j]
                for k in range(j + 1, len(self.nodeList.lines)):
                    if self.edgeMatrices[i][k][0] != PARALLEL:
                        continue
                    lk = self.nodeList.lines[k]
                    for pti in li.intersections:
                        for ptj in lj.intersections:
                            for ptk in lk.intersections:
                                if not (pti.endP or ptj.endP or ptk.endP): continue
                                if pti.value.colinear(ptj.value, ptk.value):
                                    self.colinearInts.append([pti, ptj, ptk, li, lj, lk])

    def updateLineMatrix(self, matching = True):
        """
        update the line-line matrix
        """
        if matching is False:
            nodeList = self.nodeList
            self.edgeMatrices = [[None for i in range(len(nodeList.lines))] for j in range(len(nodeList.lines))]
            edgeMatrices = self.edgeMatrices
        else:
            nodeList = self.matchingNodeList
            self.matchingEdgeMatrices = [[None for i in range(len(nodeList.lines))] for j in range(len(nodeList.lines))]
            edgeMatrices = self.matchingEdgeMatrices
        for i in range(len(nodeList.lines)):
            li = nodeList.lines[i].nodeElement
            for j in range(i+1, len(nodeList.lines)):
                lj = nodeList.lines[j].nodeElement
                # analyze edge type
                edge = ANY
                if lj.parallel(li):
                    edge = PARALLEL
                elif lj.perpendicular(li):
                    edge = PERPENDICULAR
                    
                positionOne = OUTSIDE
                positionTwo = OUTSIDE

                # neither parallel nor skew -> analyze intersection
                if edge < PARALLEL:
                    intersection = lj.intersection(li)
                    newIntersection = graphNode.intersection(intersection)
                    # if not matching:
                    #     if (len(nodeList.lines[i].forms) > 0 
                    #     and nodeList.lines[i].forms[0].start < intersection < nodeList.lines[i].forms[-1].end):
                    #         self.protoTransform.addUnresolved(intersection)
                    #     if (len(nodeList.lines[j].forms) > 0 
                    #     and nodeList.lines[j].forms[0].start < intersection < nodeList.lines[j].forms[-1].end):
                    #         self.protoTransform.addUnresolved(intersection)
                    nodeList.lines[i].addIntersection(newIntersection)
                    nodeList.lines[j].addIntersection(newIntersection)

                    # intersection vs line at index i
                    for k in nodeList.lines[i].forms:
                        if k.start == intersection or k.end == intersection:
                            if positionOne == BOUND:
                                positionOne = INSIDE
                                newIntersection.iType = 'i'
                                break
                            else:
                                positionOne = BOUND
                                newIntersection.iType = 'i'
                        elif k.containsPoint(intersection):
                            positionOne = INSIDE
                            newIntersection.iType = 'i'
                            break

                    # intersection vs line at index j
                    for k in nodeList.lines[j].forms:
                        if k.start == intersection or k.end == intersection:
                            if positionTwo == BOUND:
                                positionTwo = INSIDE
                                newIntersection.iType = 'i'
                                break
                            else:
                                positionTwo = BOUND
                                newIntersection.iType = 'i'
                        elif k.containsPoint(intersection):
                            positionTwo = INSIDE
                            newIntersection.iType = 'i'
                            break
                # set respective line-line matrix value
                edgeMatrices[i][j] = [edge, positionOne]
                edgeMatrices[j][i] = [edge, positionTwo]

                # # analyze edge type
                # edge = ANY
                # if comparand.parallel(element):
                #     edge = PARALLEL
                # elif comparand.perpendicular(element):
                #     edge = PERPENDICULAR

                # positionOne = OUTSIDE
                # positionTwo = OUTSIDE

                # # if edge type is not parallel -> analyze intersection type
                # if edge != PARALLEL:
                #     intersection = comparand.intersection(element)
                #     newIntersection = graphNode.intersection(intersection)
                #     nodeList.lines[-1].addIntersection(newIntersection)
                #     nodeList.lines[j].addIntersection(newIntersection)
                #     for k in nodeList.lines[-1].forms:
                #         if k.start == intersection or k.end == intersection:
                #             if positionOne == BOUND:
                #                 positionOne = INSIDE
                #                 break
                #             else:
                #                 positionOne = BOUND
                #                 newIntersection.iType = 'i'
                #         elif k.containsPoint(intersection):
                #             positionOne = INSIDE
                #             newIntersection.iType = 'i'
                #             break
                #     for k in nodeList.lines[j].forms:
                #         if k.start == intersection or k.end == intersection:
                #             if positionTwo == BOUND:
                #                 positionTwo = INSIDE
                #                 break
                #             else:
                #                 positionTwo = BOUND
                #                 newIntersection.iType = 'i'
                #         elif k.containsPoint(intersection):
                #             positionTwo = INSIDE
                #             newIntersection.iType = 'i'
                #             break
                # edgeMatrices[i][j] = [edge, positionOne]
                # edgeMatrices[j][i] = [edge, positionTwo]

    
    def map(self, fromInd, toInd):
        """
        Mapping a shape individual to a lhs individual for detection
        """
        self.generateNodes(toInd, matching = True)
        return True
        
        
    def distinguish(self, ind):
        """
        Distinguishing a lhs individual for rule analysis
        """
        node = self.generateNodes(ind)
        return node

    def postAnalysis(self):
        """
        Extra analysis processes:

        * addEndPoint: go through the lists and add all the end points of lines that are not intersections into nodeList
        * classifyIntersections: classifying the relationship between all the intersections/end points on a line as connected/disconnected
        * organizeForm: order the line-segments of a node and combine+separate them according to their attributes
        """
        self.nodeList.generateLines(True)
        self.updateLineMatrix(matching=False)
        self.nodeList.classifyObjs(True)
        for node in self.nodeList:
            if len(node.forms) == 0:
                continue
            node.addEndPoints()
            #node.classifyIntersections()
            node.organizeForm()
        self.analyzeColinear()



    def analyzeRHS(self, rhs):
        # add all rhs object into the unresolved list
        if self.lhsResolved is None:
            self.lhsResolved = copy(self.protoTransform.unresolved)
        else:
            self.protoTransform.unresolved = copy(self.lhsResolved)

        self.rhsResolved = []
        self.nodeList.rhsLines = []
        self.nodeList.rhsPoints = []
        for node in self.nodeList.lines:
            node.rhsInfo = None
        self.rhsSolvingOrder = []
        
        self.addRHSObjects(rhs)

        # analyze all rhs lines relationship (parallel & perpendicular)
        for node1 in self.nodeList.rhsLines:
            for node2 in (self.nodeList.lines + self.nodeList.rhsLines):
                # ignore if same line
                if node1.nodeElement.equals(node2.nodeElement):
                    continue
                if node1.nodeElement.parallel(node2.nodeElement):
                    node1.rhsInfo["parallel"].append(node2)
                elif node1.nodeElement.perpendicular(node2.nodeElement):
                    node1.rhsInfo["perpendicular"].append(node2)
                else:
                    node1.rhsInfo["intersect"].append(node2)
                if node1.nodeElement.intersect(node2.nodeElement):
                    intn = node1.nodeElement.intersection(node2.nodeElement)
                    if intn not in self.protoTransform.unresolved:
                        node1.rhsInfo['extraInts'].add(intn)
        transformableVects = self.protoTransform.unresolved

        # find the order of resolving the lines - DIRECTIVE CASES
        for directive in self.directives:
            for dirStatement in directive[1]:
                # for point_on_line specifically, check if the individual is in any of the lhs node
                if directive[0] == 'point_on_line' :
                    if len(dirStatement) == 2:
                        arg = dirStatement[0]
                        if arg.type != 'struct-key':
                            raise ValueError('Error: lineSegment tag expected for the first argument of point_on_line directive.')
                        check = False
                        for node in self.nodeList.lines:
                            if node.nodeElement.equals(arg.ind) and arg.ind in node.forms:
                                if node.rhsInfo is None:
                                    node.rhsInfo = {
                                        'case': graphNode.node.line_casus.CASE_LHS,
                                        'directives': []
                                    }
                                    self.rhsSolvingOrder.insert(0, node)
                                node.rhsInfo['directives'].append(['point_on_line', arg.ind, dirStatement])
                                for vect in self.rhsResolved:
                                    if arg.ind.containsPoint(vect) and vect not in transformableVects:
                                        transformableVects.append(vect)
                                check = True
                                break
                        if check: continue
                        self.checkRHSDirectiveObject(arg, transformableVects, self.nodeList.rhsLines)
                    elif len(dirStatement) == 3:
                        targetObj = dirStatement[0]
                        if targetObj.type != 'struct-key' and not 'ind-lineSegment' in targetObj.ind.type and not 'ind-point' in targetObj.ind.type:
                            raise ValueError('Error: lineSegment or point tag expected for the first argument of point_on_line directive.')
                        arg = dirStatement[1]
                        if arg.type != 'struct-key':
                            raise ValueError('Error: lineSegment tag expected for the second argument of 3 arguments point_on_line directive.')
                        if 'ind-lineSegment' in targetObj.ind.type:
                            if line2D.containsPoint(arg.ind, targetObj.ind.start):
                                pointOnLinetarget = targetObj.ind.start
                            elif line2D.containsPoint(arg.ind, targetObj.ind.end):
                                pointOnLinetarget = targetObj.ind.end
                        elif 'ind-point' in targetObj.ind.type:
                                pointOnLinetarget = targetObj.ind.value
                        check = False
                        for node in self.nodeList.lines:
                            if node.nodeElement.equals(arg.ind) and arg.ind in node.forms:
                                if node.rhsInfo is None:
                                    node.rhsInfo = {
                                        'case': graphNode.node.line_casus.CASE_LHS,
                                        'directives': []
                                    }
                                    self.rhsSolvingOrder.insert(0, node)
                                node.rhsInfo['directives'].append(['point_on_line', arg.ind, dirStatement, pointOnLinetarget])
                                for vect in self.rhsResolved:
                                    if arg.ind.containsPoint(vect) and vect not in transformableVects:
                                        transformableVects.append(vect)
                                check = True
                                break
                        if check: continue
                        self.checkRHSDirectiveObject(arg, transformableVects, self.nodeList.rhsLines)
                    else:
                        raise ValueError('Error: 2 or 3 arguments expected for point_on_line directive.')
                    # for arg in dirStatement:
                    #     if arg.type == 'struct-key':
                    #         check = False
                    #         # if it for a line in the lhs, add a case for the node
                    #         for node in self.nodeList.lines:
                    #             if node.nodeElement.equals(arg.ind) and arg.ind in node.forms:
                    #                 if node.rhsInfo is None:
                    #                     node.rhsInfo = {
                    #                         'case': graphNode.node.line_casus.CASE_LHS,
                    #                         'directives': []
                    #                     }
                    #                     self.rhsSolvingOrder.insert(0, node)
                    #                 node.rhsInfo['directives'].append(['point_on_line', arg.ind, dirStatement])
                    #                 for vect in self.rhsResolved:
                    #                     if arg.ind.containsPoint(vect) and vect not in transformableVects:
                    #                         transformableVects.append(vect)
                    #                 check = True
                    #                 break
                    #         if check: continue
                    #         # if not for a lhs line segment, check for the rhs containing node case
                    #         self.checkRHSDirectiveObject(arg, transformableVects, self.nodeList.rhsLines)
                # any other type of directive, check RHS directive case
                else:
                    self.checkRHSDirectiveCase(dirStatement, directive[0], transformableVects, self.nodeList.rhsLines)
                # self.resetTempCase()
        
        # check the rhs node case for the rest of the nodes
        for rhsLineNode in self.nodeList.rhsLines:
            if self.checkRHSLineCase(rhsLineNode, transformableVects, self.nodeList.rhsLines):
                # self.finalizeCase(rhsLineNode)
                # self.resetTempCase()
                pass
        # add the rhs unresolved points into the protoTransform unresolved list
        for vect in self.rhsResolved:
            if vect not in self.protoTransform.unresolved:
                self.protoTransform.unresolved.append(vect)
        # print('\n\n\nResolving Order:')
        # for i in self.rhsSolvingOrder:
        #     print(i.nodeElement.printRhino())
        #     print(i.rhsInfo['case'],'\n')
        return


    def checkRHSDirectiveCase(self, dirStatement, directiveName, transformableVects, availableLineNodes):
        """
        check the case for a specific directive statement

        :param dirStatement: the directive to be analyzed
        :param directiveName: the directive name e.g. length, point_on_line, direction, angle, distance, ...
        :type directiveName: string
        :param transformableVects: the list of transformable vector by this point
        :type transformableVects: list of vector2D
        :param availableLineNodes: the list of avalable rhs line nodes that can be used for determining the directive case
        :type availableLineNodes: list of node
        """
        # length directive
        if directiveName == 'length' or directiveName == 'point_on_line':
            for arg in dirStatement:
                self.checkRHSDirectiveObject(arg, transformableVects, availableLineNodes)
        # CASE_DIRECTION: direction directive, 1 contained point
        elif directiveName == 'direction':
            if dirStatement[0].type != 'struct-key':
                raise ValueError('Error: the first arguments of direction directive must be an object reference key.')
            if dirStatement[0].ind.type != 'ind-lineSegmentP2D':
                return False
            objNode = None
            for i in range(len(availableLineNodes)):
                if availableLineNodes[i].nodeElement.equals(dirStatement[0].ind):
                    objNode = availableLineNodes[i]
                    break
            for vec in transformableVects:
                if objNode.nodeElement.containsPoint(vec):
                    self.assignCase(objNode, graphNode.node.line_casus.CASE_DIRECTION, 
                                    [dirStatement, vec], transformableVects)
                    return True
            return False
        else:
            if dirStatement[0].type != 'struct-key' or dirStatement[1].type != 'struct-key':
                raise ValueError('Error: the first and second arguments of ' + directiveName + ' directive must be an object reference key.')
            lineNodes = copy(availableLineNodes)
            objNode = None
            if dirStatement[0].ind.type == 'ind-lineSegmentP2D':
                for i in range(len(lineNodes)):
                    if lineNodes[i].nodeElement.equals(dirStatement[0].ind):
                        objNode = lineNodes.pop(i)
                        break
            elif dirStatement[0].ind.type == 'ind-pointP2D':
                objNode = graphNode.node(dirStatement[0].ind, dirStatement[0].ind)
            if objNode is None:
                return False
            check = self.checkRHSDirectiveObject(dirStatement[1], transformableVects, lineNodes)
            if not check:
                return False
            if dirStatement[0].ind.type == 'ind-lineSegmentP2D':
                # distance directive cases
                if directiveName == 'distance':
                    # CASE_LINE_LINE_DISTANCE_DIRECTION: line-line distance directive with direction
                    if dirStatement[1].ind.type == 'ind-lineSegmentP2D' and len(dirStatement) == 4:
                        self.assignCase(objNode, graphNode.node.line_casus.CASE_LINE_LINE_DISTANCE_DIRECTION, 
                                        [dirStatement], transformableVects)
                        return True
                    # CASE_POINT_LINE_DISTANCE: point-line distance directive, 1 contained point
                    if dirStatement[1].ind.type == 'ind-pointP2D':
                        for vec in transformableVects:
                            if objNode.nodeElement.containsPoint(vec):
                                anchorVect = None
                                dirCheck = None

                                projFromPt = dirStatement[0].ind.projection(dirStatement[1].ind.value)
                                if projFromPt == vec:
                                    self.assignCase(objNode, graphNode.node.line_casus.CASE_POINT_LINE_DISTANCE, 
                                                    [dirStatement, vec, None, None], transformableVects)
                                    return True
                                for anchor in transformableVects:
                                    if anchor is vec:
                                        continue
                                    # if the base line contains the anchor, reject
                                    if vec.colinear(anchor, dirStatement[1].ind.value):
                                        continue
                                    anchorVect = anchor
                                    # check the orientation: comparing clockwise property between:
                                    # _ anchor -> contained point -> from point
                                    # vs
                                    # _ projection of from point on new line -> contained point -> from point
                                    orientationAnchor = graphMap.cw(dirStatement[1].ind.value, vec, anchorVect)
                                    orientationNew = graphMap.cw(dirStatement[1].ind.value, vec, projFromPt)
                                    dirCheck = orientationAnchor == orientationNew
                                    break

                                self.assignCase(objNode, graphNode.node.line_casus.CASE_POINT_LINE_DISTANCE, 
                                                [dirStatement, vec, anchorVect, dirCheck], transformableVects)
                                return True
                    # CASE_LINE_LINE_DISTANCE: line-line directive, 1 extra vector for checking the new line's orientation
                    if dirStatement[1].ind.type == 'ind-lineSegmentP2D':
                        anchorVect = None
                        dirCheck = None
                        for anchor in transformableVects:
                            # if the base line contains the anchor, reject this
                            if line2D.containsPoint(dirStatement[1].ind, anchor):
                                continue
                            anchorVect = anchor

                            # find the projection of the anchor onto the base line
                            anchorProj = dirStatement[1].ind.projection(anchor)

                            # check the orientation 
                            orientationAnchor = anchor.subtract(anchorProj).unit()
                            orientationNew = dirStatement[0].ind.root.subtract(dirStatement[1].ind.root).unit()
                            dirCheck = orientationAnchor == orientationNew
                            break
                        self.assignCase(objNode, graphNode.node.line_casus.CASE_LINE_LINE_DISTANCE, 
                                        [dirStatement, anchorVect, dirCheck], transformableVects)
                        return True
                    return False
                # CASE_ANGLE: angle directive, 1 contained point
                elif directiveName == 'angle':
                    for vec in transformableVects:
                        if objNode.nodeElement.containsPoint(vec):
                            # anchor type 1: contained point is on base (intersection between base and new)
                            #         \ |
                            #          \| 
                            #           . (basePt2 - intersection/contained point)
                            #           |\ <-- angle
                            #           | \
                            #           |  \
                            #           |   \
                            #           |    \
                            #           |     \
                            #           |      \
                            #           |       \
                            #           |        \
                            #           |         \
                            #           |          \
                            #           |           \
                            #           |_           \
                            #  (projPt1)|_|_ _ _ _ _ _. (basePt1)
                            #           |              \
                            #           |               \
                            #        <<new>>         <<base>>

                            # anchor type 2: contained point is not on base ####### NEED THOROUGH TESTING #######
                            #         \ |
                            #          \| 
                            #           . 
                            #           |\ <-- angle
                            #           | \
                            #           |  \
                            #           |   \
                            #           |    .(basePt2 - nearer to intersection)
                            #           |     \
                            #           |      \
                            #           |       \
                            #   (contPt).        \
                            #           |         \
                            #           |          \
                            #           |           \
                            #           |_           \
                            #  (projPt1)|_|_ _ _ _ _ _.(basePt1 - further from intersection)
                            #           |              \
                            #           |               \
                            #        <<new>>         <<base>>
                            # cw check: basePt1 -> basePt2 -> projPt1
                            #        vs basePt1 -> basePt2 -> anchor 
                            basePt1, basePt2 = None, None
                            if line2D.containsPoint(dirStatement[1].ind, vec):
                                basePt1 = dirStatement[1].ind.start
                                basePt2 = vec
                                if basePt1 == basePt2:
                                    basePt1 = dirStatement[1].ind.end
                            else:
                                if not line2D.intersect(dirStatement[0].ind, dirStatement[1].ind):
                                    return False
                                intersection = line2D.intersection(dirStatement[0].ind, dirStatement[1].ind)
                                basePt1 = dirStatement[1].ind.start
                                basePt2 = dirStatement[1].ind.end

                                d1 = intersection.subtract(basePt1)
                                d2 = intersection.subtract(basePt2)
                                if d1 > vz == d2 > vz:
                                    if d1.squareLength() < d2.squareLength():
                                        basePt1, basePt2 = basePt2, basePt1
                            projPt1 = dirStatement[0].ind.projection(basePt1)
                            orientationNew = graphMap.cw(basePt1, basePt2, projPt1)

                            anchorVect = None
                            dirCheck = None
                            for anchor in transformableVects:
                                if anchor is vec:
                                    continue
                                # if the base line contains the anchor, reject this vec
                                if anchor.colinear(dirStatement[1].ind.start, dirStatement[1].ind.end):
                                    continue
                                anchorVect = anchor

                                # check the orientation: comparing clockwise property between:
                                # _ anchor_projection -> contained point -> from point
                                # vs
                                # _ projection of from point on new line -> contained point -> from point
                                orientationAnchor = graphMap.cw(basePt1, basePt2, anchorVect)
                                dirCheck = orientationAnchor == orientationNew
                                break

                            self.assignCase(objNode, graphNode.node.line_casus.CASE_ANGLE, 
                                            [dirStatement, vec, basePt1, basePt2, anchorVect, dirCheck], transformableVects)
                            return True
                    return False
            elif dirStatement[0].ind.type == 'ind-pointP2D':
                if directiveName == 'distance':
                    # CASE_LINE_LINE_DISTANCE_DIRECTION: line-line distance directive with direction
                    if dirStatement[1].ind.type == 'ind-pointP2D' and len(dirStatement) == 4:
                        self.assignPointCase(objNode, graphNode.node.point_casus.CASE_POINT_POINT_DISTANCE_DIRECTION, 
                                        [dirStatement], transformableVects)
                        return True

                    # all other cases: find a line containing the point
                    containingLine = None
                    for l in self.nodeList.lines + availableLineNodes:
                        if l.nodeElement.containsPoint(dirStatement[0].ind.value) and self.checkRHSLineCase(l, transformableVects, availableLineNodes):
                            containingLine = l.nodeElement
                            break
                    if containingLine is None:
                        return False

                    if dirStatement[1].ind.type == 'ind-pointP2D':
                        dirx = vector2D(c0 - containingLine.direction.y, containingLine.direction.x)
                        perpLine = line2D(containingLine.sort, dirStatement[1].ind.value, dirStatement[1].ind.value.add(dirx))
                        projNew = perpLine.projection(dirStatement[0].ind.value)
                        if projNew == dirStatement[0].ind.value:
                            self.assignPointCase(objNode, graphNode.node.point_casus.CASE_POINT_POINT_DISTANCE, 
                                            [dirStatement, vec, None, None], transformableVects)
                            return True
                        orientationNew = dirStatement[0].ind.value.subtract(projNew).unit()
                        anchorVect = None
                        dirCheck = None
                        for anchor in transformableVects:
                            # if the base line contains the anchor, reject this
                            if anchor == dirStatement[0].ind.value or perpLine.containsPoint(anchor):
                                continue
                            anchorVect = anchor

                            # find the projection of the anchor onto the base line
                            anchorProj = perpLine.projection(anchor)

                            # check the orientation 
                            orientationAnchor = anchor.subtract(anchorProj).unit()
                            dirCheck = orientationAnchor == orientationNew
                            break

                        self.assignPointCase(objNode, graphNode.node.point_casus.CASE_POINT_POINT_DISTANCE, 
                                        [dirStatement, containingLine, anchorVect, dirCheck], transformableVects)
                        return True

                    elif dirStatement[1].ind.type == 'ind-lineSegmentP2D':
                        anchorVect = None
                        dirCheck = None
                        projNew = dirStatement[1].ind.projection(dirStatement[0].ind.value)

                        if projNew == dirStatement[0].ind.value:
                            return False
                        orientationNew = dirStatement[0].ind.value.subtract(projNew).unit()

                        for anchor in transformableVects:
                            # if the base line contains the anchor, reject this
                            if anchor == dirStatement[0].ind.value or dirStatement[1].ind.containsPoint(anchor):
                                continue
                            anchorVect = anchor

                            # find the projection of the anchor onto the base line
                            anchorProj = dirStatement[1].ind.projection(anchor)

                            # check the orientation 
                            orientationAnchor = anchor.subtract(anchorProj).unit()
                            dirCheck = orientationAnchor == orientationNew
                            break

                        self.assignPointCase(objNode, graphNode.node.point_casus.CASE_LINE_POINT_DISTANCE, 
                                        [dirStatement, containingLine, anchorVect, dirCheck], transformableVects)

        return False

    def checkRHSDirectiveObject(self, arg, transformableVects, availableLineNodes):
        """
        check whether the line/point in a directive argument is 
        """
        if arg.type == 'struct-key':
            return self.findMatchingKeyNode(arg, transformableVects, availableLineNodes)
        # # check objects used in any description in the directive
        # elif arg.type == 'ind-description':
        #     pass

    def findMatchingKeyNode(self, key, transformableVects, availableLineNodes):
        keyObject = key.ind

        # check if there's a lhs or rhs node that contains the target object
        checkList = None
        if keyObject.type == 'ind-lineSegmentP2D':
            checkList = self.nodeList.rhsLines + self.nodeList.lines
        else:
            return keyObject.value in transformableVects

        for node in checkList:
            if node.nodeElement.equals(keyObject):
                if node.rhsInfo is None or node.rhsInfo['case'] is not None:
                    return True
                break

        # check if there's another directive that define the infinite line containing the target line segment
        for directive in self.directives:
            if directive[0] != 'distance' and directive[0] != 'angle' and directive[0] != 'direction':
                continue
            for dirStatement in directive[1]:
                if dirStatement[0].type == 'struct-key' and dirStatement[0].name == key.name:
                    check = self.checkRHSDirectiveCase(dirStatement, directive[0], transformableVects, 
                                                       availableLineNodes)
                    return check

        # if line segment
        if keyObject.type == 'ind-lineSegmentP2D':
            if keyObject.start in transformableVects and keyObject.end in transformableVects:
                return True
            containNode = None
            for node in (self.nodeList.lines + self.nodeList.rhsLines):
                if node.nodeElement.equals(keyObject):
                    containNode = node
                    break
            if containNode is None:
                return False
            return self.checkRHSLineCase(containNode, transformableVects, availableLineNodes)
        elif keyObject.type == 'ind-pointP2D':
            return keyObject.value in transformableVects
        return True



    def checkRHSLineCase(self, lnNode, transformableVects, ori_AvailableLineNodes):
        # if it is a lhs node (lnNode.rhsInfo is None) or under any case other than unsolvable => return True 
        if lnNode.rhsInfo is None or (lnNode.rhsInfo['case'] is not None 
        and lnNode.rhsInfo['case'] != graphNode.node.line_casus.CASE_UNSOLVABLE):
            return True
        # if under unsolvable case => return False
        if (lnNode.rhsInfo['case'] is not None
        and lnNode.rhsInfo['case'] == graphNode.node.line_casus.CASE_UNSOLVABLE):
            return False

        # leftover: is rhs node but has no case assigned => try to assign a cases

        # make a copy of available node list for use, remove the line node from it
        availableLineNodes = copy(ori_AvailableLineNodes)
        check = False
        for i in range(len(availableLineNodes)):
            if lnNode.nodeElement.equals(availableLineNodes[i].nodeElement):
                check = True
                availableLineNodes.pop(i)
                break
        if not check:
            return False

        # proxy check round 1
        # if the node contains 2 or more transformable vectors already => solved by proxy
        transformableCount = 0
        for vec in transformableVects:
            if lnNode.nodeElement.containsPoint(vec):
                transformableCount += 1
        if transformableCount >= 2:
            self.assignCase(lnNode, graphNode.node.line_casus.CASE_PROXY, 
                            [], transformableVects)
            return True

        # reorganize parallel and perpendicular lines list so that currently solvable lines are in front
        self.reorganizeRelatedNodes(lnNode.rhsInfo["perpendicular"])
        self.reorganizeRelatedNodes(lnNode.rhsInfo["parallel"])

        # ______ NORMAL CASE ______
        # cases with 1 perpendicular line
        for l in lnNode.rhsInfo["perpendicular"]:
            if not self.checkRHSLineCase(l, transformableVects, availableLineNodes):
                continue
            # CASE_1PERP_1PT: 1 perpendicular line, 1 contained point
            for vec in transformableVects:
                if lnNode.nodeElement.containsPoint(vec):
                    self.assignCase(lnNode, graphNode.node.line_casus.CASE_1PERP_1PT, 
                                    [l, vec], transformableVects)
                    return True

        # cases with 1 parallel line
        for l in lnNode.rhsInfo["parallel"]:
            if not self.checkRHSLineCase(l, transformableVects, availableLineNodes):
                continue
            # CASE_1PARALLEL_1PT: 1 perpendicular line, 1 contained point
            for vec in transformableVects:
                if lnNode.nodeElement.containsPoint(vec):
                    self.assignCase(lnNode, graphNode.node.line_casus.CASE_1PARALLEL_1PT, 
                                    [l, vec], transformableVects)
                    return True

        # proxy check round 2
        # if the node contains 2 or more transformable vectors already => solved by proxy
        transformableCount = 0
        for vec in transformableVects:
            if lnNode.nodeElement.containsPoint(vec):
                transformableCount += 1
        if transformableCount >= 2:
            self.assignCase(lnNode, graphNode.node.line_casus.CASE_PROXY, [], transformableVects)
            return True

        # lnNode.rhsInfo["case"] = graphNode.node.line_casus.CASE_UNSOLVABLE
        return False



    def assignCase(self, node, case, caseNodes, transformableVects, lhsNode = False):
        # print('     assign case', node.nodeElement.printRhino(), case)
        node.rhsInfo["case"] = case
        node.rhsInfo["caseNodes"] = caseNodes
        node.rhsInfo["caseTemp"] = True

        # if node.nodeElement.type == 'ind-line2D':
        #     self.rhsSolvingOrder.append(node)
        #     for l in node.forms:
        #         transformableVects.append(l.start)
        #         transformableVects.append(l.end)
        # else:
        #     self.rhsSolvingOrder.append(node)
        if not lhsNode:
            self.rhsSolvingOrder.append(node)

        if len(node.forms) == 0:
            return

        checkDirective = False

        for directive in self.directives:
            if directive[0] == 'point_on_line' :
                if 'ind-line' not in node.nodeElement.type: continue
                for dirStatement in directive[1]:
                    if len(dirStatement) == 2:
                        arg = dirStatement[0]
                        if arg.type != 'struct-key':
                            raise ValueError('Error: lineSegment tag expected for the first argument of point_on_line directive.')
                        if not node.nodeElement.equals(arg.ind): continue
                        for ind in node.forms:
                            if ind.key is not None and ind.key.name == argument.name:
                                for vect in self.rhsResolved:
                                    if ind.containsPoint(vect) and vect not in transformableVects:
                                        transformableVects.append(vect)
                                    node.rhsInfo['directives'].append([directive[0], ind, dirStatement])
                                    checkDirective = True
                                    break
                    elif len(dirStatement) == 3:
                        targetObj = dirStatement[0]
                        if targetObj.type != 'struct-key' and not 'ind-lineSegment' in targetObj.ind.type and not 'ind-point' in targetObj.ind.type:
                            raise ValueError('Error: lineSegment or point tag expected for the first argument of point_on_line directive.')
                        arg = dirStatement[1]
                        if arg.type != 'struct-key':
                            raise ValueError('Error: lineSegment tag expected for the second argument of 3 arguments point_on_line directive.')
                        if 'ind-lineSegment' in targetObj.ind.type:
                            if line2D.containsPoint(arg.ind, targetObj.ind.start):
                                target = targetObj.ind.start
                            elif line2D.containsPoint(arg.ind, targetObj.ind.end):
                                target = targetObj.ind.end
                        elif 'ind-point' in targetObj.ind.type:
                                target = targetObj.ind.value
                        if not node.nodeElement.equals(arg.ind): continue
                        for ind in node.forms:
                            if ind.key is not None and ind.key.name == argument.name:
                                for vect in self.rhsResolved:
                                    if ind.containsPoint(vect) and vect not in transformableVects:
                                        transformableVects.append(vect)
                                    node.rhsInfo['directives'].append([directive[0], ind, dirStatement, target])
                                    checkDirective = True
                                    break
                    else:
                        raise ValueError('Error: 2 or 3 arguments expected for point_on_line directive.')
            elif directive[0] == 'length':
                for dirStatement in directive[1]:
                    for argument in dirStatement:
                        if argument.type != 'struct-key':
                            continue
                        for ind in node.forms:
                            if ind.key is not None and ind.key.name == argument.name:
                                if ind.start not in transformableVects:
                                    transformableVects.append(ind.start)
                                if ind.end not in transformableVects:
                                    transformableVects.append(ind.end)
                                node.rhsInfo['directives'].append([directive[0], ind, dirStatement])
                                checkDirective = True
                                break

            # if directive[0] == 'length' or directive[0] == 'point_on_line':
            #     for dirStatement in directive[1]:
            #         for argument in dirStatement:
            #             if argument.type != 'struct-key':
            #                 continue
            #             for ind in node.forms:
            #                 if ind.key is not None and ind.key.name == argument.name:
            #                     if directive[0] == 'point_on_line':
            #                         for vect in self.rhsResolved:
            #                             if ind.containsPoint(vect) and vect not in transformableVects:
            #                                 transformableVects.append(vect)
            #                     else:
            #                         if ind.start not in transformableVects:
            #                             transformableVects.append(ind.start)
            #                         if ind.end not in transformableVects:
            #                             transformableVects.append(ind.end)
            #                     node.rhsInfo['directives'].append([directive[0], ind, dirStatement])
            #                     checkDirective = True
            #                     break
            else:
                for dirStatement in directive[1]:
                    if dirStatement[0].type != 'struct-key':
                        continue
                    for ind in node.forms:
                        if ind.key is not None and ind.key.name == dirStatement[0].name:
                            node.rhsInfo['directives'].append([directive[0], ind, dirStatement])
                            checkDirective = True

        if node.nodeElement.type == 'ind-line2D':
            for n in node.rhsInfo['parallel']:
                for ln in n.forms:
                    startProj = node.nodeElement.projection(ln.start)
                    endProj = node.nodeElement.projection(ln.end)
                    for i in node.forms:
                        if (startProj == i.start or endProj == i.start) and i.start not in transformableVects:
                            transformableVects.append(i.start)
                        if (startProj == i.end or endProj == i.end) and i.end not in transformableVects:
                            transformableVects.append(i.end)
        if lhsNode:
            if checkDirective:
                for i in range(len(self.rhsSolvingOrder)):
                    if len(self.rhsSolvingOrder[i].rhsInfo['directives']) == 0:
                        self.rhsSolvingOrder.insert(i, node)
                        return
            self.rhsSolvingOrder.append(node)
        # if checkDirective:
        #     self.finalizeCase(node)

    def assignPointCase(self, node, case, caseNodes, transformableVects):
        # print('     assign case', node.nodeElement.printRhino(), case)
        node.rhsInfo = {
            "case": case,
            "caseNodes": caseNodes,
            "caseTemp": True,
            "directives": []
        }

        if node.nodeElement.value not in transformableVects:
            transformableVects.append(node.nodeElement.value)
        if node not in self.rhsSolvingOrder:
            self.rhsSolvingOrder.append(node)

    def finalizeCase(self, targetNode):
        if targetNode.rhsInfo is None:
            return 
        # print('\nfinalize:\n',targetNode.nodeElement.printRhino())
        targetNode.rhsInfo['caseTemp'] = False
        for node in targetNode.rhsInfo['caseNodes']:
            if isinstance(node, vector2D) or isinstance(node, list) or targetNode.rhsInfo is None:
                continue
            self.finalizeCase(node)

    def resetTempCase(self):
        # print('\n!! reset temp cases')
        for node in self.nodeList.rhsLines:
            if node.rhsInfo['caseTemp'] == True:
                node.rhsInfo["case"] = None
                node.rhsInfo["caseNodes"] = []
                node.rhsInfo['caseTemp'] = False

    def reorganizeRelatedNodes(self, nodeList):
        i = 0
        while i < len(nodeList):
            node = nodeList[i]
            if node.rhsInfo is None or node.rhsInfo['case'] is not None:
                nodeList.pop(i)
                nodeList.insert(0, node)
            i += 1

    def addRHSObjects(self, form):
        """
        add all new rhs objects (not solvable with the lhs) into the graphNode

        :param form: form in the rhs
        :type form: ~sortal.form.form.form
        """
        if form.type == 'form-interval' and form.sort.baseSort().characteristic.type == 'ind-lineSegmentP2D':
            for i in form:
                self.addRHSLine(i)
        elif form.type == 'form-discrete' and form.sort.baseSort().characteristic.type == 'ind-pointP2D':
            for i in form:
                self.addRHSPoint(i)
        elif form.type == 'form-areal' or form.type == 'form-metaForm':
            for i in form:
                self.addRHSObjects(i)

    def addRHSLine(self, line):
        """
        add one rhs line segment into the graphNode

        :param line: a line segment in the rhs form
        :type line: ~sortal.ind.lineSegmentP.lineSegmentP2D
        :param rhsResolved: a list of all unresolved vectors in rhs
        :type rhsResolved: list of vector2D
        """
        # if the line's start and end are in the list of lhs unresolved (resolved after lhs matching)
        if line.start in self.protoTransform.unresolved and line.end in self.protoTransform.unresolved:
            # check whether there's a lhs line with the same direction + root, ignore the line if there is
            for node in self.nodeList.lines:
                if node.nodeElement.equals(line):
                    return
            # no line with the same direction + root => solve by proxy through lhs lines
            lineNode = self.nodeList.RHS_addLine(line.parametricNode(), ind = line)
            self.assignCase(lineNode, graphNode.node.line_casus.CASE_PROXY, lineNode.rhsInfo['caseNodes'], [], True)
            return
        if line.start not in self.protoTransform.unresolved and line.start not in self.rhsResolved:
            self.rhsResolved.append(line.start)
        if line.end not in self.protoTransform.unresolved and line.end not in self.rhsResolved:
            self.rhsResolved.append(line.end)
        
        lineNode = self.nodeList.RHS_addLine(line.parametricNode(), ind = line)
        if lineNode.rhsInfo['case'] == graphNode.node.line_casus.CASE_PROXY and lineNode not in self.rhsSolvingOrder:
            self.assignCase(lineNode, graphNode.node.line_casus.CASE_PROXY, lineNode.rhsInfo['caseNodes'], [], True)


    def addRHSPoint(self, point):
        """
        add one rhs point into the graphNode

        :param point: a point in the rhs form
        :type point: ~sortal.ind.pointP.pointP2D
        :param rhsResolved: a list of all unresolved vectors in rhs
        :type rhsResolved: list of vector2D
        """
        if point.value not in self.protoTransform.unresolved:
            self.rhsResolved.append(point.value)

    # ============================================================================================
    #                                    SHAPE MATCHING PROCESS
    # ============================================================================================

    def potentialTransforms(self):
        """
        Find all potential transform for the rule of the mapping with the saved shape
        """
            
        # analysis processes for the shape
        self.matchingNodeList.generateLines()
        self.updateLineMatrix()
        self.matchingNodeList.classifyObjs()
        for matchingNode in self.matchingNodeList:
            if len(matchingNode.forms) == 0:
                continue
            matchingNode.addEndPoints()
            #matchingNode.classifyIntersections()
            matchingNode.organizeForm()
        
        
        ### printing
        # print('\nlhs Matrix:')
        # for j in self.edgeMatrices:
        #     for k in j:
        #         if k is None:
        #             print(NONE, end=' ')
        #         else:
        #             print (k,end=' ')
        #     print()
        # print('\nMatching Matrix:')
        # for j in self.matchingEdgeMatrices:
        #     for k in j:
        #         if k is None:
        #             print(NONE, end=' ')
        #         else:
        #             print (k,end=' ')
        #     print()
        # print()
        ###

        transforms = []
        matchLists = []
        # detect transform --> find all possible transformation matches for nodes
        # e.g. node 0 in lhs is matched with node 4 in the shape
        #      node 1 in lhs is matched with node 2 in the shape
        #      node 2 in lhs is matched with node 0 in the shape
        #      node 3 in lhs is matched with node 1 in the shape
        #      --> match = [4,2,1,0]
        #      --> matchLists = list of all possible matches
        self.detectTransform(matchLists)
        noInt = True

        # start with an empty transform
        ptransf = [graphTransform()]

        # resolve points that are not on any line first
        if not self.resolveFreePoints(ptransf):
            return []

        # investigate each of the matches in detectTransform's result
        for match in matchLists:
            ### printRhino the matching lines with matching color
            ###
            # print('\n-----------------------------------------------------------')
            # print('*match:', match)
            # s = [0,0,0,2, int(765/len(match))]
            # for index, num in enumerate(match):
            #     print(self.nodeList.lines[index].nodeElement.printRhino(), end='')
            #     print(' SelLast -Properties o c o '+str(s[0])+','+str(s[1])+','+str(s[2])+' EnterEnd SelNone', end=' ')
            #     print(self.matchingNodeList.lines[num].nodeElement.printRhino(), end='')
            #     print(' SelLast -Properties o c o '+str(s[0])+','+str(s[1])+','+str(s[2])+' EnterEnd SelNone')
            #     s[s[3]] += s[4]
            #     if s[s[3]] > 255:
            #         s[3] -= 1
            #         s[s[3]] = s[s[3]+1] - 255
            #         s[s[3]+1] = 255
            # print()
            ###

            # initialize a graph transform object for this particular match
            transf = [] 
            for tr in ptransf:
                transf.append(tr.duplicate())

            # check: by the end of this process, if check is still True --> valid transformation, if False --> invalid
            check = True              
            singleIntNodes = []
            orientChecks = []
            # for each node in lhs:
            for nodeIndex in range(len(self.nodeList)):
                nodeI = self.nodeList.lines[nodeIndex]
                matchNodeI = self.matchingNodeList.lines[match[nodeIndex]]
                # find intersections transformation between the nodes and add it into the graph transform
                # e.g. node i in lhs is matched with node i0 in shape
                #      node j in lhs is matched with node j0 in shape
                #      node i  intersects j  at point vectFrom
                #      node i0 intersects j0 at point vectTo
                #      --> vectFrom is matched with vectTo
                    
                for j in range(nodeIndex+1, len(self.nodeList)):
                    # if the two lhs lines in the two matches are not parallel
                    if self.matchingEdgeMatrices[match[nodeIndex]][match[j]][0] != PARALLEL:
                        # marks as having intersection
                        noInt = False

                        # vectFrom: lhs intersection; vectTo: shape intersection
                        vectFrom = self.getIntersection(nodeI,self.nodeList.lines[j])
                        vectTo = self.getIntersection(matchNodeI,self.matchingNodeList.lines[match[j]])

                        # if vectFrom or vectTo is None (the two lhs/shape lines are parallel)
                        # -> the match is still accepted, move on to another line to compare
                        if vectFrom is None or vectTo is None:
                            continue

                        # if vectFrom is not contained by any line-segment in the lhs, add it as extra instead of normal transformation
                        if vectFrom.iType == 'o':
                            for tr in transf:
                                tr.addExtra(vectFrom.value, vectTo.value)
                            continue

                        trIndex = 0
                        while trIndex < len(transf):
                            tr = transf[trIndex]
                            # if the intersection match cannot be added to the graph transform
                            # (due to vectFrom already have a different match from VectTo, etc...),
                            # the graph transform is deemed invalid
                            if tr.contains(vectFrom.value) and tr.transform(vectFrom.value) == vectTo.value:
                                trIndex += 1
                                continue
                            if tr.containsMatch(vectTo.value) or not tr.add(vectFrom.value, vectTo.value, index = 0):
                                transf.pop(trIndex)
                            else:
                                trIndex += 1
                
                # if after comparing this one line with all other lines and there is no intersection
                # -> special cases: 1. only one line in the lhs
                #                   2. more than one line in the lhs but they are all parallel to one another
                if noInt:
                    break

                #check the ordering of the intersections
                if len(transf) == 0 or not self.orderingCheck(transf, nodeI):
                    check = False
                    break
                
                # if the node doesnt have any end points 
                # -> all points on the lhs line already have a match 
                # -> continue onto other nodes
                if (not nodeI.endP) and len(nodeI.linePoints) == 0:
                    continue
                
                # check if the node and matchingNode are matched in reverse
                # i.e.  node:           --A---B------C----
                #       matchingNode:   ----X--Y--Z----W------
                #       A matches with W, B matches with Z, C matches with X
                #       --> the two nodes match in reverse order
                startI = 0
                while (not transf[0].contains(nodeI.intersections[startI].value)):
                    startI += 1
                    if startI >= len(nodeI.intersections):
                        noInt = True
                        break
                if noInt:
                    break
                # start is to be the match of the smallest point that already has a matching (based on the lhs line)
                start = transf[0].apply(nodeI.intersections[startI].value)
                
                endI = -1
                while not transf[0].contains(nodeI.intersections[endI].value):
                    endI -= 1
                    if endI < -len(nodeI.intersections):
                        noInt = True
                        break
                if noInt:
                    break
                # end is to be the match of the biggest point that already has a matching (based on the lhs line)
                end = transf[0].apply(nodeI.intersections[endI].value)
                
                # if start == end -> the matched intersection points are the same
                if start == end:
                    # if the lhs start and end intersection points are not the same but their matches are 
                    # -> the match is discarded
                    if nodeI.intersections[startI] != nodeI.intersections[endI]:
                        check = False
                        break
                    # otherwise, there is only one intersection on the line 
                    # -> special case, add the matching to singleIntNodes and resolve later
                    singleIntNodes.append([nodeI, matchNodeI, startI])

                    # Orientation check: check the ordering of intersection relative to other intersections
                    for vec in tr.vects:
                        if nodeI.nodeElement.containsPoint(vec[0]):
                            break
                        
                    self.addOrientCheck(orientChecks, vec)
                    if not self.orientationChecking(orientChecks):
                        check = False
                        singleIntNodes = []
                        break
                    continue

                elif start > end:
                    reverse = True
                else:
                    reverse = False
                    
                # resolve points that are on the node but not an intersection
                if len(nodeI.linePoints) > 0:
                    if len(matchNodeI.linePoints) < len(nodeI.linePoints):
                        check = False
                        break
                    self.resolvePointsOnLine(nodeI, matchNodeI, reverse, transf)

                # if the node does not have any line-segment's start or end point unresolved, continue to the next node
                if not nodeI.endP:
                    continue
                if len(transf) == 0:
                    check = False
                    break

                ##### resolve all the unresolved start/end point on the node #####

                # get all existing resolved points on the node
                lineForm = []
                specialLineForm = []
                lineIndex = 0
                # for each of the line-segment on the node
                while lineIndex < len(nodeI.forms):
                    # check if the line-segment contains at least one intersection point in it
                    # if yes, add it to lineForm along with any other line-segments that touches it
                    if transf[0].hasCheckPointOf(nodeI.forms[lineIndex]):
                        lineForm.append(nodeI.forms[lineIndex])
                        prevCheck = lineIndex
                        while prevCheck > 0 and nodeI.forms[prevCheck-1].end == nodeI.forms[prevCheck].start:
                            lineForm.append(nodeI.forms[prevCheck-1])
                            prevCheck -= 1
                        while lineIndex < len(nodeI.forms)-1 and nodeI.forms[lineIndex].end == nodeI.forms[lineIndex+1].start:
                            lineForm.append(nodeI.forms[lineIndex+1])
                            lineIndex += 1
                    lineIndex += 1

                # any line-segments that are not in lineForm, add them to specialLineForm for special consideration
                for line in nodeI.forms:
                    if line not in lineForm:
                        specialLineForm.append(line)
                        
                # resolve line-segments in specialLineForm first
                for line in specialLineForm:
                    transfIndex = 0
                    # for each of the transformation
                    while transfIndex < len(transf):
                        tr = transf.pop(transfIndex)
                        # nodeInts = tr.getCheckPoints(nodeI.nodeElement)
                        nodeInts = self.getCheckPoints(tr, nodeI, True)
                        # resolve special line
                        posMatches = self.resolveSpecialLine(line, nodeInts, reverse, nodeI, matchNodeI)
                        if len(posMatches) == 0:
                            continue
                        # add the matches to the transformation sets
                        for posM in posMatches:
                            additionalTransf = tr.duplicate()
                            if not reverse:
                                additionalTransf.add(line.start, posM.start)
                                additionalTransf.add(line.end, posM.end)                            
                            else:
                                additionalTransf.add(line.start, posM.end)
                                additionalTransf.add(line.end, posM.start)
                            transf.insert(transfIndex, additionalTransf)
                            transfIndex += 1
                                
                # no matches for special lines -> discard the transformation
                if not check:
                    break

                # resolve all other normal line-segments
                ind = 0
                while ind < len(transf):
                    tr = transf[ind]
                    trCheck = True
                    # each of the normal line-segment
                    for line in lineForm:
                        # resolve normal line-segment
                        hasMatch = self.resolveLine(line, reverse, nodeI, matchNodeI, tr)
                        if not hasMatch:
                            transf.pop(ind)
                            trCheck = False
                            break
                    if not trCheck:
                        break
                    # check that each transformation set have transformation for all intersections on the line
                    for intersection in nodeI.intersections:
                        if intersection.iType != 'o' and not tr.contains(intersection.value):
                            transf.pop(ind)
                            ind -=1
                            break
                    graphMap.normalizeLineEndPoints(nodeI, tr, reverse)
                    ind += 1

            # resolve the nodes with no intersections
            if noInt:
                transf = self.resolveNoIntersectionNodes(match, transf)
                if transf is False:
                    continue

            # make sure there's one existing point match for each lhs point object that is on an intersection
            if not self.checkIntersectionPoints(transf):
                continue

            # if there are 2 lines with single intersection and they intersect each other
            if len(self.nodeList) == len(singleIntNodes) == 2 and singleIntNodes[0][1].nodeElement.intersect(singleIntNodes[1][1].nodeElement):
                self.resolveSingleIntersectionNodes([singleIntNodes[0]], transf, True)
                self.resolveSingleIntersectionNodes([singleIntNodes[1]], transf, True)

            # if there is only 1 line with 1 intersection --> there are at least 2 other lines that intersect this line at a single point
            elif len(singleIntNodes) == 1:
                if len(transf) == 0:
                    continue
                nodeI = singleIntNodes[0][0]
                matchNodeI = singleIntNodes[0][1]
                anchor = nodeI.intersections[singleIntNodes[0][2]].value
                vec = []
                for v in transf[0].vects:
                    if v[0] == anchor:
                        nAnchor = v[1]
                        continue
                    vec.append(v)

                while vec[0][0].colinear(vec[1][0], anchor):
                    vec.pop(1)

                p = nodeI.forms[0].end if nodeI.forms[0].start == anchor else nodeI.forms[0].start
                self.resolveSingleIntersectionNodes(singleIntNodes, transf)
                for index in range(len(transf)):
                    tr = transf.pop(0)
                    nP = tr.apply(p)
                    ccw1 = graphMap.cw(vec[0][0], vec[1][0], anchor)
                    ccw2 = graphMap.cw(vec[0][0], anchor, p)
                    ccw3 = graphMap.cw(vec[0][1], vec[1][1], nAnchor)
                    ccw4 = graphMap.cw(vec[0][1], nAnchor, nP)
                    if (ccw1 == ccw2) == (ccw3 == ccw4):
                        transf.append(tr)
            # all other cases
            else:
                self.resolveSingleIntersectionNodes(singleIntNodes, transf)

            self.resolveColinearEndPoints(transf)

            # resolve planes and curves
            self.resolvePlanes(transf, match)
            self.resolveCurves(transf)

            # register the transformation
            if check:
                for tr in transf:
                    #print('----------------')
                    #tr.printTransf()
                    tr.mergeExtra()
                    transforms.append(tr)
        i = 0
        while i < len(transforms):
            j = i+1
            while j < len(transforms):
                if transforms[i] == transforms[j]:
                    transforms.pop(j)
                else:
                    j += 1
            i += 1
        return transforms

    def detectTransform(self, matchLists):
        """
        detect all matches that satisfy the matrix matching
        """

        lhsCount = []
        for i in self.edgeMatrices:
            nodeCount = {'par': 0,
                         'per-o': 0, 'per-b': 0, 'per-i': 0,
                         'any-o': 0, 'any-b': 0, 'any-i': 0}
            for j in i:
                if j is None:
                    continue
                if j[0] == PARALLEL:
                    nodeCount['par'] += 1
                elif j[0] == PERPENDICULAR:
                    if j[1] == OUTSIDE:
                        nodeCount['per-o'] += 1
                    elif j[1] == BOUND:
                        nodeCount['per-b'] += 1
                    elif j[1] == INSIDE:
                        nodeCount['per-i'] += 1
                elif j[0] == ANY:
                    if j[1] == OUTSIDE:
                        nodeCount['any-o'] += 1
                    elif j[1] == BOUND:
                        nodeCount['any-b'] += 1
                    elif j[1] == INSIDE:
                        nodeCount['any-i'] += 1
            lhsCount.append(nodeCount)

        for i in range(len(self.nodeList.lines)):
            nodeCount = lhsCount[i]
            nodeCount['pts'] = len(self.nodeList.lines[i].containedPoints)

        shapeCount = []
        for i in self.matchingEdgeMatrices:
            nodeCount = {'par': 0,
                         'per-o': 0, 'per-b': 0, 'per-i': 0,
                         'any-o': 0, 'any-b': 0, 'any-i': 0}
            for j in i:
                if j is None:
                    continue
                if j[0] == PARALLEL:
                    nodeCount['par'] += 1
                elif j[0] == PERPENDICULAR:
                    if j[1] == OUTSIDE:
                        nodeCount['per-o'] += 1
                    elif j[1] == BOUND:
                        nodeCount['per-b'] += 1
                    elif j[1] == INSIDE:
                        nodeCount['per-i'] += 1
                elif j[0] == ANY:
                    if j[1] == OUTSIDE:
                        nodeCount['any-o'] += 1
                    elif j[1] == BOUND:
                        nodeCount['any-b'] += 1
                    elif j[1] == INSIDE:
                        nodeCount['any-i'] += 1
            shapeCount.append(nodeCount)
        for j in range(len(self.matchingNodeList.lines)):
            nodeCount = shapeCount[j]
            nodeCount['pts'] = len(self.matchingNodeList.lines[j].containedPoints)


        badMatches = []
        for i in range(len(lhsCount)):
            for j in range(len(shapeCount)):
                l = lhsCount[i]
                s = shapeCount[j]
                if (l['par'] > s['par']
                or  l['per-o'] > (s['per-o'] + s['per-b'] + s['per-i'])
                or  l['per-b'] > (             s['per-b'] + s['per-i'])
                or  l['per-i'] > (                          s['per-i'])
                or  l['any-o'] > (s['any-o'] + s['any-b'] + s['any-i'] + s['per-o'] + s['per-b'] + s['per-i'] + s['par'])
                or  l['any-b'] > (             s['any-b'] + s['any-i']              + s['per-b'] + s['per-i'])
                or  l['any-i'] > (                          s['any-i']                           + s['per-i'])
                or  l['pts'] > s['pts']):
                        badMatches.append(str(i) + ' ' + str(j))

        goodMatches = []
        for indI in range(len(self.edgeMatrices) - 1):
            for indJ in range(indI + 1, len(self.edgeMatrices[indI])):
                lhsEdge1 = self.edgeMatrices[indI][indJ]
                lhsEdge2 = self.edgeMatrices[indJ][indI]
                for i in range(len(self.matchingEdgeMatrices)):
                    if str(indI) + ' ' + str(i) in badMatches:
                        continue
                    for j in range(len(self.matchingEdgeMatrices[i])):
                        if i == j:
                            continue
                        if str(indJ) + ' ' + str(j) in badMatches:
                            continue

                        matchEdge1 = self.matchingEdgeMatrices[i][j]
                        matchEdge2 = self.matchingEdgeMatrices[j][i]

                        if ((lhsEdge1[0] == PARALLEL and lhsEdge1[0] != matchEdge1[0])
                        or (matchEdge1[0] == PARALLEL and (lhsEdge1[1] != OUTSIDE or lhsEdge2[1] != OUTSIDE))
                        or (lhsEdge1[0] > matchEdge1[0])
                        or lhsEdge1[1] > matchEdge1[1]
                        or lhsEdge2[1] > matchEdge2[1]):
                            continue

                        existing = False
                        for goodMatch in goodMatches:
                            if goodMatch[0][0] == indI and goodMatch[0][1] == i:
                                existing = True
                                goodMatch[1].append([indJ, j])
                        if not existing:
                            goodMatches.append([[indI, i],[[indJ, j]]])


        # for i in goodMatches:
        #     print(i[0])
        #     for j in i[1]:
        #         print('    ',j)
        index = len(self.edgeMatrices) - 1
        roughMatches = []
        for goodMatch in reversed(goodMatches):
            if index != goodMatch[0][0]:
                index -= 1
                matchIndex = 0
                while matchIndex < len(roughMatches):
                    if len(roughMatches[matchIndex]) < len(self.edgeMatrices) - goodMatch[0][0] - 1:
                        roughMatches.pop(matchIndex)
                    else:
                        matchIndex += 1
            if goodMatch[0][0] == len(self.edgeMatrices) - 2:
                for m in goodMatch[1]:
                    roughMatches.append([goodMatch[0][1], m[1]])
            else:
                for matchInd in range(len(roughMatches)):
                    match = roughMatches[matchInd]
                    if len(match) > len(self.edgeMatrices) - goodMatch[0][0] - 1:
                        break
                    checkIndex = goodMatch[0][0] + 1
                    matchIndex = 0
                    for gMatch in goodMatch[1]:
                        if checkIndex == gMatch[0] and match[matchIndex] == gMatch[1]:
                            checkIndex += 1
                            matchIndex += 1
                        elif checkIndex < gMatch[0]:
                            break
                    if checkIndex == len(self.edgeMatrices):
                        newMatch = copy(match)
                        newMatch.insert(0, goodMatch[0][1])
                        roughMatches.append(newMatch)
        matchIndex = 0
        while matchIndex < len(roughMatches):
            if len(roughMatches[matchIndex]) < len(self.edgeMatrices):
                roughMatches.pop(matchIndex)
            else:
                matchIndex += 1



        # roughMatches = []
        # for goodMatch in goodMatches:
        #     if goodMatch[0][0] != 0:
        #         break
        #     matches = [[goodMatch[0][1]]]
        #     for missingIndex in range(1, len(self.nodeList.lines)):
        #         matchesCount = len(matches)
        #         for i in range(matchesCount):
        #             match = matches.pop(0)
        #             for validMatch in goodMatch[1]:
        #                 if validMatch[0] == missingIndex and validMatch[1] not in match:
        #                     newMatch = copy(match)
        #                     newMatch.append(validMatch[1])
        #                     matches.append(newMatch)
        #                 elif validMatch[0] > missingIndex:
        #                     break
        #     roughMatches += matches

        # for i in range(1, len(self.nodeList.lines) - 1):
        #     valitatingMatches = []
        #     for goodMatch in goodMatches:
        #         if goodMatch[0][0] > i: break
        #         elif goodMatch[0][0] != i: continue
        #         matches = [[goodMatch[0][1]]]
        #         for missingIndex in range(i + 1, len(self.nodeList.lines)):
        #             matchesCount = len(matches)
        #             for j in range(matchesCount):
        #                 match = matches.pop(0)
        #                 for validMatch in goodMatch[1]:
        #                     if validMatch[0] == missingIndex and validMatch[1] not in match:
        #                         newMatch = copy(match)
        #                         newMatch.append(validMatch[1])
        #                         matches.append(newMatch)
        #                     elif validMatch[0] > missingIndex:
        #                         break
        #         valitatingMatches += matches
        #     matchI = 0
        #     while matchI < len(roughMatches):
        #         check = False
        #         for valMatch in valitatingMatches:
        #             if roughMatches[matchI][i:] == valMatch:
        #                 check = True
        #                 break
        #         if not check:
        #             roughMatches.pop(matchI)
        #         else:
        #             matchI += 1

        # find line matches
        for subset in roughMatches:
            check = True
            for indI, i in enumerate(subset):
                for indJ, j in enumerate(subset):
                    if i == j:
                        continue
                    lhsEdge = self.edgeMatrices[indI][indJ]
                    matchEdge = self.matchingEdgeMatrices[i][j]
                    # reject if:
                    # lhs lines are parallel but matching lines are not parallel
                    # lhs line intersection is inside/bound but matching lines are parallel
                    # lhs edge is bigger than match edge (lhsEdge[0] = perpedicular (1) but matchEdge[0] = any (0))
                    # lhs intersection type is bigger than match intersection type (lhsEdge[1] = inside (2) but matchEdge[1] = inside (0))
                    if ((lhsEdge[0] == PARALLEL and lhsEdge[0] != matchEdge[0])
                    or (matchEdge[0] == PARALLEL and lhsEdge[1] != OUTSIDE)
                    or (lhsEdge[0] > matchEdge[0]) 
                    or lhsEdge[1] > matchEdge[1]):
                        check = False
                        break
                if not check:
                    break
            if check:
                matchLists.append(subset)
                
    def resolveSpecialLine(self, line, nodeInts, reverse, nodeI, matchNodeI):
        """
        resolve end-points of special line-segment: line-segments that do not contain any intersection point on it
        """
        posMatch = []

        # if line.end is before the smallest intersection of the node
        if line.end < nodeInts[0][0]:
            checkEnd = False

            # check if there is any other line-segment in the node that contains the smallest intersection 
            # and the smallest intersection is not that line-segment's start point
            for lineCheck in nodeI.forms:
                if lineCheck.containsPoint(nodeInts[0][0]) and nodeInts[0][0] != lineCheck.start:
                    checkEnd = True

            # for each of the line-segment on the matching node
            for mLine in matchNodeI.forms:
                # remove lines that have different sort, have insufficient attribute, 
                # line starting point not smaller than smallest point match if not reverse or
                # line ending point not larger than smallest point match if reverse
                if line.sort.baseSort() != mLine.sort.baseSort() or (line.hasAttribute() 
                and not line.attribute.checkPartOf(mLine.attribute, flagDesc = False)):
                    continue
                if (not reverse) and mLine.start >= nodeInts[0][1]:
                    continue
                elif reverse and mLine.end <= nodeInts[0][1]:
                    continue

                # if there's no line that contain the smallest intersection, this line can match up to the smallest intersection match
                if not checkEnd:
                    if not reverse:
                        matchStart = mLine.start
                        matchEnd = nodeInts[0][1] if mLine.containsPoint(nodeInts[0][1]) else mLine.end
                    else:
                        matchStart = nodeInts[0][1] if mLine.containsPoint(nodeInts[0][1]) else mLine.start
                        matchEnd = mLine.end
                    newLine = mLine.__class__(mLine.sort, matchStart, matchEnd)
                # if there's another line that contain the smallest intersection, do not match if mLine contains the intersection
                else:
                    if checkEnd and mLine.containsPoint(nodeInts[0][1]):
                        continue
                    newLine = mLine
                posMatch.append(newLine)
        # if line.start is before the largest intersection of the node
        elif line.start > nodeInts[-1][0]:
            # check if there is any other line-segment in the node that contains the largest intersection 
            # and the largest intersection is not that line-segment's end point
            checkStart = False
            for lineCheck in nodeI.forms:
                if lineCheck.containsPoint(nodeInts[-1][0]) and nodeInts[-1][0] != lineCheck.end:
                    checkStart = True

            # for each of the line-segment on the matching node
            for mLine in matchNodeI.forms:
                # remove lines that have different sort, have insufficient attribute, 
                # line ending point not larger than largest point match if not reverse or
                # line starting point not smaller than largest point match if reverse
                if line.sort.baseSort() != mLine.sort.baseSort() or (line.hasAttribute() 
                and not line.attribute.checkPartOf(mLine.attribute, flagDesc = False)):
                    continue
                if (not reverse) and mLine.end <= nodeInts[-1][1]:
                    continue
                elif reverse and mLine.start >= nodeInts[-1][1]:
                    continue
                # if there's no line that contain the largest intersection, this line can match up to the largest intersection match
                if not checkStart:
                    if not reverse:
                        matchStart = nodeInts[-1][1] if mLine.containsPoint(nodeInts[-1][1]) else mLine.start
                        matchEnd = mLine.end
                    else:
                        matchStart = mLine.start
                        matchEnd = nodeInts[-1][1] if mLine.containsPoint(nodeInts[-1][1]) else mLine.end
                    newLine = mLine.__class__(mLine.sort, matchStart, matchEnd)
                # if there's another line that contain the largest intersection, do not match if mLine contains the intersection
                else:
                    if checkStart and mLine.containsPoint(nodeInts[-1][1]):
                        continue
                    newLine = mLine
                posMatch.append(newLine)
        # else -> line is between the smallest and biggest intersections
        else:
            index = 1
            while line.start > nodeInts[index][0]:
                index += 1
            # find the matches of the closest intersections that the line is between as lowerBound and upperBound
            if reverse:
                upperBound = nodeInts[0][1]
                for indexC in range(1,index):
                    if upperBound > nodeInts[indexC][1]:
                        upperBound = nodeInts[indexC][1]
                lowerBound = nodeInts[-1][1]
                for indexC in range(index,len(nodeInts)-1):
                    if lowerBound < nodeInts[indexC][1]:
                        lowerBound = nodeInts[indexC][1]
            else:
                lowerBound = nodeInts[0][1]
                for indexC in range(1,index):
                    if lowerBound < nodeInts[indexC][1]:
                        lowerBound = nodeInts[indexC][1]
                upperBound = nodeInts[-1][1]
                for indexC in range(index,len(nodeInts)-1):
                    if upperBound > nodeInts[indexC][1]:
                        upperBound = nodeInts[indexC][1]
                
            # check if there are any other lines that contains the two bounding intersections
            checkStart = False
            checkEnd = False
            for lineCheck in nodeI.forms:
                if lineCheck.containsPoint(nodeInts[index-1][0]) and nodeInts[index-1][0] != lineCheck.end:
                    checkStart = True
                if lineCheck.containsPoint(nodeInts[index][0]) and nodeInts[index][0] != lineCheck.start:
                    checkEnd = True
            # for each of the line-segment on the matching node
            for mLine in matchNodeI.forms:
                # remove lines that have different sort, have insufficient attribute, 
                # or doesn't have any part inside the bound
                if line.sort.baseSort() != mLine.sort.baseSort() or (line.hasAttribute() 
                and not line.attribute.checkPartOf(mLine.attribute, flagDesc = False)):
                    continue
                if mLine.end <= lowerBound or mLine.start >= upperBound:
                    continue

                # if mLine contains lowerBound
                if mLine.containsPoint(nodeInts[index-1][1]):
                    if checkStart:
                        continue
                    else:
                        matchStart = nodeInts[index-1][1]
                elif not reverse:
                    matchStart = mLine.start
                else:
                    matchStart = mLine.end

                if mLine.containsPoint(nodeInts[index][1]):
                    if checkEnd:
                        continue
                    else:
                        matchEnd = nodeInts[index][1]
                elif not reverse:
                    matchEnd = mLine.end
                else:
                    matchEnd = mLine.start


                # #### 
                # if not (checkStart or checkEnd):
                #     if not reverse:
                #         matchStart = nodeInts[index-1][1] if mLine.containsPoint(nodeInts[index-1][1]) else mLine.start
                #         matchEnd = nodeInts[index][1] if mLine.containsPoint(nodeInts[index][1]) else mLine.end
                #     else:
                #         matchStart = nodeInts[index-1][1] if mLine.containsPoint(nodeInts[index-1][1]) else mLine.end
                #         matchEnd = nodeInts[index][1] if mLine.containsPoint(nodeInts[index][1]) else mLine.start
                #     newLine = mLine.__class__(mLine.sort, matchStart, matchEnd)
                # else:
                #     if checkStart and mLine.containsPoint(nodeInts[index-1][1]):
                #         continue
                #     if checkEnd and mLine.containsPoint(nodeInts[index][1]):
                #         continue
                #     newLine = mLine
                # mergeable = False
                # if not mergeable:
                #     posMatch.append(newLine)
                #     #print(posMatch)

                newLine = mLine.__class__(mLine.sort, matchStart, matchEnd)
                posMatch.append(newLine)
        return posMatch
        
    def resolveLine(self, line, reverse, nodeI, matchNodeI, transf):
        """
        resolve end points of line-segments that contains at least one intersection in the node
        """

        # check if the line-segment start and end points already have matches and add the unmatched one(s) to unmatchedPts
        checkStart = transf.contains(line.start)
        checkEnd = transf.contains(line.end)
        if checkStart and checkEnd:
            return True
        if checkStart:
            unmatchedPts = [line.end]
        elif checkEnd:
            unmatchedPts = [line.start]
        else:
            unmatchedPts = [line.start, line.end]
        
        hasMatch = False

        # get all existing node intersections with matches
        segmentInts = []
        nodeInts = []
        for nInt in nodeI.intersections:
            if nInt.endP:
                if line.containsPoint(nInt.value) and transf.contains(nInt.value):
                    pass
                else:
                    continue
            matchingInt = transf.transform(nInt.value)
            if not matchingInt:
                continue
            if line.start == nInt.value or line.end == nInt.value:
                alignment = BOUND
                segmentInts.append([nInt.value, matchingInt, alignment])
            elif line.containsPoint(nInt.value):
                alignment = INSIDE
                segmentInts.append([nInt.value, matchingInt, alignment])
            nodeInts.append([nInt.value, matchingInt])

        if (len(nodeInts) < 2):
            segmentInts = []
            nodeInts = []
            for nInt in nodeI.intersections:
                if nInt.endP:
                    if line.containsPoint(nInt.value) and transf.contains(nInt.value):
                        pass
                    else:
                        continue
                matchingInt = transf.allTransform(nInt.value)
                if not matchingInt:
                    continue
                if line.start == nInt.value or line.end == nInt.value:
                    alignment = BOUND
                    segmentInts.append([nInt.value, matchingInt, alignment])
                elif line.containsPoint(nInt.value):
                    alignment = INSIDE
                    segmentInts.append([nInt.value, matchingInt, alignment])
                nodeInts.append([nInt.value, matchingInt])


        # checking against all line-segments in the matching node
        for mLine in matchNodeI.forms:
            # if line's attribute is not a part of matching line's attribute --> continue
            if line.sort.baseSort() != mLine.sort.baseSort() or (line.hasAttribute() 
                and not line.attribute.checkPartOf(mLine.attribute, flagDesc = False)):
                continue
            #print('   ____ mLine:', mLine, mLine.attribute)
            # check if the matching line contains the matching intersections of intersections on line
            lineCheck = True
            for segInt in segmentInts:
                if (mLine.containsPoint(segInt[1]) 
                or (segInt[2] == BOUND and (mLine.start == segInt[1] or mLine.end == segInt[1]))):
                    continue
                lineCheck = False
                break
            if not lineCheck:
                continue
            hasMatch = True
            #print('     ~~~~ match ~~~~ ')
            # for each unmatched points
            for pointIndex, point in enumerate(unmatchedPts):
                posMatch = None # possible match
                orderingPriority = False
                if point < segmentInts[0][0]:
                    if point > nodeInts[0][0]:
                        if reverse:
                            # point is smaller than first matched point on segment
                            # but bigger than first matched point on node
                            # and the node is matched in reverse
                            # --> matching point must be:
                            #       bigger than first matched point on matching segment
                            #       smaller or equals to first matched point on matching node
                            #       be on the matching segment
                            orderingPriority = '>'
                            for mInt in matchNodeI.intersections:
                                if (mInt.value > segmentInts[0][1]  
                                and mInt.value <= nodeInts[0][1]
                                and mLine.containsPoint(mInt.value)):
                                    posMatch = mInt.value
                        else:
                            # point is smaller than first matched point on segment
                            # but bigger than first matched point on node
                            # and the node is NOT matched in reverse
                            # --> matching point must be:
                            #       smaller than first matched point on matching segment
                            #       bigger or equals to first matched point on matching node
                            #       be on the matching segment
                            orderingPriority = '<'
                            for mInt in reversed(matchNodeI.intersections):
                                if (mInt.value < segmentInts[0][1]  
                                and mInt.value >= nodeInts[0][1]
                                and mLine.containsPoint(mInt.value)):
                                    posMatch = mInt.value
                    else:
                        if reverse:
                            # point is smaller than first matched point on segment,
                            # also smaller than first matched point on node 
                            # and the node is matched in reverse
                            # --> matching point must be:
                            #       bigger than first matched point on matching node
                            #       be on the matching segment
                            orderingPriority = '>'
                            for mInt in matchNodeI.intersections:
                                if (mInt.value > nodeInts[0][1]  
                                and mLine.containsPoint(mInt.value)):
                                    posMatch = mInt.value
                        else:
                            # point is smaller than first matched point on segment,
                            # also smaller than first matched point on node 
                            # and the node is NOT matched in reverse
                            # --> matching point must be:
                            #       smaller than first matched point on matching node
                            #       be on the matching segment
                            orderingPriority = '<'
                            for mInt in reversed(matchNodeI.intersections):
                                if (mInt.value < nodeInts[0][1]  
                                and mLine.containsPoint(mInt.value)):
                                    posMatch = mInt.value
                elif point > segmentInts[-1][0]:
                    if point < nodeInts[-1][0]:
                        if reverse:
                            # point is bigger than last matched point on segment,
                            # but smaller than last matched point on node 
                            # and the node is matched in reverse
                            # --> matching point must be:
                            #       smaller than last matched point on matching segment
                            #       bigger or equals to last matched point on matching node
                            #       be on the matching segment
                            orderingPriority = '<'
                            for mInt in reversed(matchNodeI.intersections):
                                if (mInt.value < segmentInts[-1][1]  
                                and mInt.value >= nodeInts[-1][1] ###
                                and mLine.containsPoint(mInt.value)):
                                    posMatch = mInt.value
                        else:
                            # point is bigger than last matched point on segment,
                            # but smaller than last matched point on node 
                            # and the node is NOT matched in reverse
                            # --> matching point must be:
                            #       bigger than last matched point on matching segment
                            #       smaller or equals to last matched point on matching node
                            #       be on the matching segment
                            orderingPriority = '>'
                            for mInt in matchNodeI.intersections:
                                if (mInt.value > segmentInts[-1][1]  
                                and mInt.value <= nodeInts[-1][1] ###
                                and mLine.containsPoint(mInt.value)):
                                    posMatch = mInt.value
                    else:
                        if reverse:
                            # point is bigger than last matched point on segment,
                            # also bigger than last matched point on node 
                            # and the node is matched in reverse
                            # --> matching point must be:
                            #       smaller than last matched point on matching node
                            #       be on the matching segment
                            orderingPriority = '<'
                            for mInt in reversed(matchNodeI.intersections):
                                if (mInt.value < nodeInts[-1][1]  
                                and mLine.containsPoint(mInt.value)):
                                    posMatch = mInt.value
                        else:
                            # point is bigger than last matched point on segment,
                            # also bigger than last matched point on node 
                            # and the node is NOT matched in reverse
                            # --> matching point must be:
                            #       bigger than last matched point on matching node
                            #       be on the matching segment
                            orderingPriority = '>'
                            for mInt in matchNodeI.intersections:
                                if (mInt.value > nodeInts[-1][1]  
                                and mLine.containsPoint(mInt.value)):
                                    posMatch = mInt.value
                else:
                    # point is in between the first and last matched points on segment
                    # narrow down the range in which the point can be 
                    index = 1
                    while point > segmentInts[index][0] and index < len(segmentInts):
                        index += 1
                    # the point must be now between segmentInts[index-1] and segmentInts[index]
                    # i.e. point is bigger than point on node at (index-1) 
                    #          and smaller than point on node at (index) 
                    if reverse:
                        # the node is matched in reverse
                        # --> matching point must be:
                        #       smaller than matched point on matching node at (index-1)
                        #       bigger than matched point on matching node at (index)
                        #       be on the matching segment
                        for mInt in reversed(matchNodeI.intersections):
                            if (mInt.value < segmentInts[index-1][1]  
                            and mInt.value > segmentInts[index][1]
                            and mLine.containsPoint(mInt.value)):
                                posMatch = mInt.value
                    else:
                        # the node is NOT matched in reverse
                        # --> matching point must be:
                        #       bigger than matched point on matching node at (index-1)
                        #       smaller than matched point on matching node at (index)
                        #       be on the matching segment
                        for mInt in matchNodeI.intersections:
                            if (mInt.value > segmentInts[index-1][1]  
                            and mInt.value < segmentInts[index][1]
                            and mLine.containsPoint(mInt.value)):
                                posMatch = mInt.value
                if posMatch is None:
                    continue
                transf.add(point, posMatch, orderingPriority)
        for i in unmatchedPts:
            if not transf.contains(i):
                hasMatch= False
        return hasMatch
        
        
    def resolveSingleIntersectionNodes(self, singleIntNodes, transf, noCheckPoint = False):
        """
        resolve nodes with only one single intersection
        """
        if len(singleIntNodes) == 0:
            return transf
        transf1 = copy(transf)
        #print(singleIntNodes)
        for sn in singleIntNodes:
            transfIndex = 0
            # find the matches for reverse order
            while transfIndex < len(transf):
                tr = transf.pop(transfIndex)
                newTransf = self.resolveSingleIntersectionLine(sn[0],sn[1],tr,sn[2], True, noCheckPoint)
                for i in newTransf:
                    transf.insert(transfIndex, i)
                    transfIndex += 1
            transfIndex1 = 0
            # find the matches for normal order
            while transfIndex1 < len(transf1):
                tr = transf1.pop(transfIndex1)
                newTransf = self.resolveSingleIntersectionLine(sn[0],sn[1],tr,sn[2], False, noCheckPoint)
                for i in newTransf:
                    transf1.insert(transfIndex1, i)
                    transfIndex1 += 1
        transf += transf1
        return transf
    
    
    def resolveSingleIntersectionLine(self, nodeI, matchNodeI, tr, anchorIndex, reverse, noCheckPoint = False):
        """
        resolve nodes with only one single intersection
        """
        if len(nodeI.forms) > len(matchNodeI.forms) or len(nodeI.linePoints) > len(matchNodeI.linePoints):
            return []

        # check the orientation of the external intersections if this line has any as compared to the reverse parameter
        if len(nodeI.intersections) > 1:
            transformableIntersections = []
            for i in nodeI.intersections:
                match = tr.getExtraMatch(i.value)
                if match:
                    transformableIntersections.append([i.value, match])
            if len(transformableIntersections) > 1:
                intReverse = ((transformableIntersections[0][0] < transformableIntersections[1][0])
                            != (tr.apply(transformableIntersections[0][1]) < tr.apply(transformableIntersections[1][1])))
                for i in range(1, len(transformableIntersections) - 1):
                    intCurr = transformableIntersections[i]
                    intNext = transformableIntersections[i + 1]
                    if ((intCurr[0] < intNext[0]) != (intCurr[1] < intNext[1])) != intReverse:
                        intReverse = None
                        break
                if intReverse is not None and intReverse != reverse:
                    return []

        # the single intersection on the line
        anchorFrom = nodeI.intersections[anchorIndex].value
        anchorTo = tr.apply(nodeI.intersections[anchorIndex].value)
    
        # check if the anchor is on any of the line-segments
        anchorCheck = False
        for line in nodeI.forms:
            if line.containsPoint(anchorFrom):
                anchorCheck = True
                break
        lineMatches = []
        
        
        # checkPoint: one point that is not contained by this line (for clockwise orientation check)
        checkPoint = None
        for vect in tr.vects:
            if (not nodeI.nodeElement.containsPoint(vect[0])) and (not matchNodeI.nodeElement.containsPoint(vect[1])):
                checkPoint = vect
                break
        # if no point that is not on the node has matches 
        if noCheckPoint or checkPoint is None:
            # find all possible matches for each of the line-segment in the node.
            # only consider the lines relative position compared to the only intersection the line has.
            # the ordering of the line-segments are not considered yet
            for line in nodeI.forms:
                # line before anchor -> lineCheck = -1
                # line contains anchor -> lineCheck = 0
                # line after anchor -> lineCheck = 1
                if line.containsPoint(anchorFrom):
                    lineCheck = 0
                elif line.start > anchorFrom:
                    lineCheck = 1
                else:
                    lineCheck = -1
                posMatch = []
                # for each line-segments in matching node
                for mLine in matchNodeI.forms:
                    # ignore lines with different sort and attribute
                    if line.sort.baseSort() != mLine.sort.baseSort() or (line.hasAttribute() 
                    and not line.attribute.checkPartOf(mLine.attribute, flagDesc = False)):
                        continue
                    
                    # find matching line-segment
                    if not reverse:
                        if lineCheck == 0 and mLine.containsPoint(anchorTo):
                            if line.start == anchorFrom:
                                if anchorTo == mLine.end: continue
                                newLine = mLine.__class__(mLine.sort, anchorTo, mLine.end)
                            elif line.end == anchorFrom:
                                if anchorTo == mLine.start: continue
                                newLine = mLine.__class__(mLine.sort, mLine.start, anchorTo)
                            else:
                                newLine = mLine
                        elif (not anchorCheck) and lineCheck == 1 and mLine.containsPoint(anchorTo):
                            if mLine.end == anchorTo: continue
                            newLine = mLine.__class__(mLine.sort, anchorTo, mLine.end)
                        elif (not anchorCheck) and lineCheck == -1 and mLine.containsPoint(anchorTo):
                            if mLine.start == anchorTo: continue
                            newLine = mLine.__class__(mLine.sort, mLine.start, anchorTo)
                        elif lineCheck == 1 and mLine.start > anchorTo:
                            newLine = mLine
                        elif lineCheck == -1 and mLine.end < anchorTo:
                            newLine = mLine
                        else:
                            continue
                    else:
                        if lineCheck == 0 and mLine.containsPoint(anchorTo):
                            if line.start == anchorFrom:
                                if anchorTo == mLine.start: continue
                                newLine = mLine.__class__(mLine.sort, mLine.start, anchorTo)
                            elif line.end == anchorFrom:
                                if anchorTo == mLine.end: continue
                                newLine = mLine.__class__(mLine.sort, anchorTo, mLine.end)
                            else:
                                newLine = mLine
                        elif (not anchorCheck) and lineCheck == 1 and mLine.containsPoint(anchorTo):
                            if mLine.start == anchorTo: continue
                            newLine = mLine.__class__(mLine.sort, mLine.start, anchorTo)
                        elif (not anchorCheck) and lineCheck == -1 and mLine.containsPoint(anchorTo):
                            if mLine.end == anchorTo: continue
                            newLine = mLine.__class__(mLine.sort, anchorTo, mLine.end)
                        elif lineCheck == 1 and mLine.end < anchorTo:
                            newLine = mLine
                        elif lineCheck == -1 and mLine.start > anchorTo:
                            newLine = mLine
                        else:
                            continue
                    posMatch.append(newLine)
                if len(posMatch) == 0:
                    return []
                # add the possible matches to lineMatches
                lineMatches.append((line, posMatch))
                
            indexing = [0] * len(lineMatches)
            transf = []
            transfIndex = 0
            # check all combinations of line matching to see whether they are valid: check the line-segments ordering
            while True:
                additionalTransf = tr.duplicate()
                addCheck = True
                for l in range(len(indexing)):
                    line = lineMatches[l][0]
                    posM = lineMatches[l][1][indexing[l]]
                    if not reverse:
                        additionalTransf.add(line.start, posM.start)
                        additionalTransf.add(line.end, posM.end)
                    else:
                        additionalTransf.add(line.start, posM.end)
                        additionalTransf.add(line.end, posM.start)

                # check the ordering of the resolved points on the node for additionalTransf
                if not self.orderingCheck([additionalTransf], nodeI, True):
                    addCheck = False
                
                if addCheck:
                    transf.insert(transfIndex, additionalTransf)

                # advance the indexing
                transfIndex += 1
                indexing[0] += 1
                for i in range(len(indexing)-1):
                    if indexing[i] == len(lineMatches[i][1]):
                        indexing[i] = 0
                        indexing[i+1] += 1
                    else:
                        break
                if indexing[-1] == len(lineMatches[-1][1]):
                    break

            # resolve points that are on the node
            self.resolvePointsOnLine(nodeI, matchNodeI, reverse, transf)
            return transf
                
        # if there are points not on the node that already have matches
        else:
            # find each line-segment matches based on their clockwise/counterclockwise
            # relation with checkPoint
            for line in nodeI.forms:
                posMatch = []

                # check if the line contains anchor
                containsAnchor = line.containsPoint(anchorFrom)
                # cPoint is the end point that is not anchor
                cPoint = line.end if line.start == anchorFrom else line.start

                # lhs orientation
                ccwCheck = graphMap.cw(cPoint, anchorFrom, checkPoint[0])

                # for each of the matching lines
                for mLine in matchNodeI.forms:
                    # ignore lines of different sorts & attributes
                    if line.sort.baseSort() != mLine.sort.baseSort() or (line.hasAttribute() 
                    and not line.attribute.checkPartOf(mLine.attribute, flagDesc = False)):
                        continue

                    # nCPoint matching line's end point that is not anchor
                    nCPoint = mLine.end if mLine.start == anchorTo else mLine.start
                    # if lhs line-segment contains anchor, the shape line-segment must contain anchor
                    # can match it with no further check
                    if containsAnchor:
                        if not mLine.containsPoint(anchorTo):
                            continue
                        # if anchorTo is either end-points
                        if anchorTo == mLine.end or anchorTo == mLine.start: 
                            if (graphMap.cw(nCPoint, anchorTo, checkPoint[1]) == ccwCheck) != reverse:
                                newLine = mLine.__class__(mLine.sort, nCPoint, anchorTo)
                            else:
                                continue
                        else:
                            newLine = mLine.__class__(mLine.sort, mLine.start, mLine.end)

                        # # if anchorTo is either end-points
                        # if anchorTo == mLine.end or anchorTo == mLine.start: 
                        #     if (graphMap.cw(nCPoint, anchorTo, checkPoint[1]) == ccwCheck) != reverse:
                        #         newLine = mLine.__class__(mLine.sort, nCPoint, anchorTo)
                        #     else:
                        #         continue
                        # elif (graphMap.cw(nCPoint, anchorTo, checkPoint[1]) == ccwCheck) != reverse:
                        #     newLine = mLine.__class__(mLine.sort, nCPoint, anchorTo)
                        # else:
                        #     newLine = mLine.__class__(mLine.sort, mLine.end, anchorTo)
                        # ####################
                        # if line.start == anchorFrom:
                        #     if anchorTo == mLine.end or anchorTo == mLine.start: 
                        #         if (graphMap.cw(nCPoint, anchorTo, checkPoint[1]) == ccwCheck) != reverse:
                        #             newLine = mLine.__class__(mLine.sort, nCPoint, anchorTo)
                        #         else:
                        #             continue
                        #     elif (graphMap.cw(nCPoint, anchorTo, checkPoint[1]) == ccwCheck) != reverse:
                        #         newLine = mLine.__class__(mLine.sort, nCPoint, anchorTo)
                        #     else:
                        #         newLine = mLine.__class__(mLine.sort, mLine.end, anchorTo)
                        # else:
                        #     if anchorTo == mLine.end or anchorTo == mLine.start: 
                        #         if (graphMap.cw(nCPoint, anchorTo, checkPoint[1]) == ccwCheck) != reverse:
                        #             newLine = mLine.__class__(mLine.sort, nCPoint, anchorTo)
                        #         else:
                        #             continue
                        #     elif (graphMap.cw(nCPoint, anchorTo, checkPoint[1]) == ccwCheck) != reverse:
                        #         newLine = mLine.__class__(mLine.sort, nCPoint, anchorTo)
                        #     else:
                        #         newLine = mLine.__class__(mLine.sort, mLine.end, anchorTo)

                        
                    # if lhs line-segment does not contain anchor but the shape line-segment does
                    elif mLine.containsPoint(anchorTo):
                        if (graphMap.cw(nCPoint, anchorTo, checkPoint[1]) == ccwCheck) != reverse:
                            newLine = mLine.__class__(mLine.sort, anchorTo, nCPoint)
                        elif anchorTo != mLine.end and anchorTo != mLine.start:
                            newLine = mLine.__class__(mLine.sort, anchorTo, mLine.end)
                        else:
                            continue
                    # if both do not contain anchor, check for orientation
                    elif (graphMap.cw(nCPoint, anchorTo, checkPoint[1]) == ccwCheck) != reverse:
                        newLine = mLine
                    else:
                        continue
                    posMatch.append(newLine)
                if len(posMatch) == 0:
                    return []
                # add to lineMatches
                lineMatches.append((line, posMatch))
                
            # check the ordering of
            indexing = [0] * len(lineMatches)
            transf = []
            transfIndex = 0
            # check all combinations of line matching to see whether they are valid: check the line-segments ordering
            while True:
                additionalTransf = tr.duplicate()
                addCheck = True
                for l in range(len(indexing)):
                    line = lineMatches[l][0]
                    posM = lineMatches[l][1][indexing[l]]
                    if anchorFrom == line.start:
                        if anchorTo == posM.start:
                            additionalTransf.add(line.end, posM.end)
                        else:
                            additionalTransf.add(line.end, posM.start)
                    elif anchorFrom == line.end:
                        if anchorTo == posM.start:
                            additionalTransf.add(line.start, posM.end)
                        else:
                            additionalTransf.add(line.start, posM.start)
                    elif line.start > anchorFrom:
                        if posM.start > anchorTo:
                            additionalTransf.add(line.start, posM.start)
                            additionalTransf.add(line.end, posM.end)
                        elif posM.end < anchorTo:
                            additionalTransf.add(line.start, posM.end)
                            additionalTransf.add(line.end, posM.start)
                        else:
                            raise NotImplementedError('Single Line Matching Special Case Not Implemented Yet')
                    elif line.end < anchorFrom:
                        if posM.end < anchorTo:
                            additionalTransf.add(line.start, posM.start)
                            additionalTransf.add(line.end, posM.end)
                        elif posM.start > anchorTo:
                            additionalTransf.add(line.start, posM.end)
                            additionalTransf.add(line.end, posM.start)
                    else:
                        if not reverse:
                            additionalTransf.add(line.start, posM.start)
                            additionalTransf.add(line.end, posM.end)
                        else:
                            additionalTransf.add(line.start, posM.end)
                            additionalTransf.add(line.end, posM.start)
                        
                    # ################
                    # ccwCheck0 = graphMap.cw(line.start, line.end, checkPoint[0])
                    # ccwCheck1 = graphMap.cw(posM.start, posM.end, checkPoint[1])
                    # if (ccwCheck0 == ccwCheck1) != reverse:
                    #     additionalTransf.add(line.start, posM.start)
                    #     additionalTransf.add(line.end, posM.end)
                    # else:
                    #     additionalTransf.add(line.start, posM.end)
                    #     additionalTransf.add(line.end, posM.start)

                if not self.orderingCheck([additionalTransf], nodeI, True):
                    addCheck = False
                
                if addCheck:
                    transf.insert(transfIndex, additionalTransf)
                transfIndex += 1
                indexing[0] += 1
                for i in range(len(indexing)-1):
                    if indexing[i] == len(lineMatches[i][1]):
                        indexing[i] = 0
                        indexing[i+1] += 1
                    else:
                        break
                if indexing[-1] == len(lineMatches[-1][1]):
                    break
            # resolve points that are on the node
            self.resolvePointsOnLine(nodeI, matchNodeI, reverse, transf)
            return transf
                
                

    # def resolveSingleIntersectionLine(self, nodeI, matchNodeI, tr, anchorIndex, reverse):
    #     if len(nodeI.forms) > len(matchNodeI.forms) or len(nodeI.linePoints) > len(matchNodeI.linePoints):
    #         return []
            
    #     # the single intersection on the line
    #     anchorFrom = nodeI.intersections[anchorIndex].value
    #     anchorTo = tr.apply(nodeI.intersections[anchorIndex].value)
        
    #     # check if the anchor is on the line
    #     anchorCheck = False
    #     for line in nodeI.forms:
    #         if line.containsPoint(anchorFrom):
    #             anchorCheck = True
    #             break
    #     lineMatches = []
        
    #     for line in nodeI.forms:
    #         if line.containsPoint(anchorFrom):
    #             lineCheck = 0
    #         elif line.start > anchorFrom:
    #             lineCheck = 1
    #         else:
    #             lineCheck = -1
    #         posMatch = []
    #         for mLine in matchNodeI.forms:
    #             if line.sort.baseSort() != mLine.sort.baseSort() or (line.hasAttribute() 
    #             and not line.attribute.checkPartOf(mLine.attribute, flagDesc = False)):
    #                 continue
    #             if not reverse:
    #                 if lineCheck == 0 and mLine.containsPoint(anchorTo):
    #                     if line.start == anchorFrom:
    #                         if anchorTo == mLine.end: continue
    #                         newLine = mLine.__class__(mLine.sort, anchorTo, mLine.end)
    #                     elif line.end == anchorFrom:
    #                         if anchorTo == mLine.start: continue
    #                         newLine = mLine.__class__(mLine.sort, mLine.start, anchorTo)
    #                     else:
    #                         newLine = mLine
    #                 elif (not anchorCheck) and lineCheck == 1 and mLine.containsPoint(anchorTo):
    #                     if mLine.end == anchorTo: continue
    #                     newLine = mLine.__class__(mLine.sort, anchorTo, mLine.end)
    #                 elif (not anchorCheck) and lineCheck == -1 and mLine.containsPoint(anchorTo):
    #                     if mLine.start == anchorTo: continue
    #                     newLine = mLine.__class__(mLine.sort, mLine.start, anchorTo)
    #                 elif lineCheck == 1 and mLine.start > anchorTo:
    #                     newLine = mLine
    #                 elif lineCheck == -1 and mLine.end < anchorTo:
    #                     newLine = mLine
    #                 else:
    #                     continue
    #             else:
    #                 if lineCheck == 0 and mLine.containsPoint(anchorTo):
    #                     if line.start == anchorFrom:
    #                         if anchorTo == mLine.start: continue
    #                         newLine = mLine.__class__(mLine.sort, mLine.start, anchorTo)
    #                     elif line.end == anchorFrom:
    #                         if anchorTo == mLine.end: continue
    #                         newLine = mLine.__class__(mLine.sort, anchorTo, mLine.end)
    #                     else:
    #                         newLine = mLine
    #                 elif (not anchorCheck) and lineCheck == 1 and mLine.containsPoint(anchorTo):
    #                     if mLine.start == anchorTo: continue
    #                     newLine = mLine.__class__(mLine.sort, mLine.start, anchorTo)
    #                 elif (not anchorCheck) and lineCheck == -1 and mLine.containsPoint(anchorTo):
    #                     if mLine.end == anchorTo: continue
    #                     newLine = mLine.__class__(mLine.sort, anchorTo, mLine.end)
    #                 elif lineCheck == 1 and mLine.end < anchorTo:
    #                     newLine = mLine
    #                 elif lineCheck == -1 and mLine.start > anchorTo:
    #                     newLine = mLine
    #                 else:
    #                     continue
    #             posMatch.append(newLine)
    #         if len(posMatch) == 0:
    #             return []
    #         lineMatches.append((line, posMatch))
    #     indexing = [0] * len(lineMatches)
    #     transf = []
    #     transfIndex = 0
    #     while True:
    #         additionalTransf = tr.duplicate()
    #         addCheck = True
    #         for l in range(len(indexing)):
    #             line = lineMatches[l][0]
    #             posM = lineMatches[l][1][indexing[l]]
    #             if not reverse:
    #                 additionalTransf.add(line.start, posM.start)
    #                 additionalTransf.add(line.end, posM.end)
    #             else:
    #                 additionalTransf.add(line.start, posM.end)
    #                 additionalTransf.add(line.end, posM.start)
    #         if not self.orderingCheck([additionalTransf], nodeI, True):
    #             addCheck = False
            
    #         if addCheck:
    #             transf.insert(transfIndex, additionalTransf)
    #         transfIndex += 1
    #         indexing[0] += 1
    #         for i in range(len(indexing)-1):
    #             if indexing[i] == len(lineMatches[i][1]):
    #                 indexing[i] = 0
    #                 indexing[i+1] += 1
    #             else:
    #                 break
    #         if indexing[-1] == len(lineMatches[-1][1]):
    #             break
    #     self.resolvePointsOnLine(nodeI, matchNodeI, reverse, transf)
    #     return transf

    
    def resolveNoIntersectionNodes(self, match, transf):
        """
        Resolve nodes with no intersection on them.
        """
        if len(self.nodeList) == 0:
            return transf


        transf1 = copy(transf)
        
        # # for each of the node
        # for nodeIndex in range(len(self.nodeList)):
        #     lineI = self.nodeList.lines[nodeIndex].nodeElement
        #     matchLineI = self.matchingNodeList.lines[match[nodeIndex]].nodeElement
        #     self.resolvePointsOnLine(lineI, matchLineI, True, transf)
        #     self.resolvePointsOnLine(lineI, matchLineI, False, transf1)

        line0 = self.nodeList.lines[0].nodeElement
        mLine0 = self.matchingNodeList.lines[match[0]].nodeElement

        # create a horizontal line <(0,0),(0,1)> as intersector
        intersector = matchIntersector = line0.__class__(line0.sort, vector2D(0,0), vector2D(0,1))

        # if lhs line parallel to the intersector -> make intersector
        if not intersector.intersect(line0):
            intersector = line0.__class__(line0.sort, vector2D(0,0), vector2D(1,0))
        if not matchIntersector.intersect(mLine0):
            matchIntersector = line0.__class__(line0.sort, vector2D(0,0), vector2D(1,0))
            
        orientChecks = []
        for nodeIndex in range(len(self.nodeList)):
            if (len(self.nodeList.lines[nodeIndex].forms) > len(self.matchingNodeList.lines[match[nodeIndex]].forms)
            or len(self.nodeList.lines[nodeIndex].linePoints) > len(self.matchingNodeList.lines[match[nodeIndex]].linePoints)):
                return False
            
            lineI = self.nodeList.lines[nodeIndex].nodeElement
            matchLineI = self.matchingNodeList.lines[match[nodeIndex]].nodeElement
            
            if lineI.parallel(intersector):
                continue

            # add the intersections between the intersector and the lhs/shape line to orientation check list
            v0 = lineI.intersection(intersector)
            v1 = matchLineI.intersection(matchIntersector)

            self.addOrientCheck(orientChecks, [v0,v1])
        # check orientation 
        if not self.orientationChecking(orientChecks):
            return False
            
        # resolve each line-segment
        for nodeIndex in range(len(self.nodeList)):
            nodeI = self.nodeList.lines[nodeIndex]
            matchNodeI = self.matchingNodeList.lines[match[nodeIndex]]
            transfIndex = 0
            # reverse order
            while transfIndex < len(transf):
                tr = transf.pop(transfIndex)
                newTransf = self.resolveNoIntersectionLine(nodeI,matchNodeI,tr,True)
                for i in newTransf:
                    transf.insert(transfIndex, i)
                    transfIndex += 1
            self.resolvePointsOnLine(nodeI, matchNodeI, True, transf)
            transfIndex1 = 0
            # normal order
            while transfIndex1 < len(transf1):
                tr = transf1.pop(transfIndex1)
                newTransf = self.resolveNoIntersectionLine(nodeI,matchNodeI,tr,False)
                for i in newTransf:
                    transf1.insert(transfIndex1, i)
                    transfIndex1 += 1
            self.resolvePointsOnLine(nodeI, matchNodeI, False, transf1)
        transf += transf1
        i = 0
        while i < len(transf):
            j = i+1
            while j <len(transf):
                if transf[i] == transf[j]:
                    transf.pop(j)
                else:
                    j += 1
            i += 1
        return transf
                    
    def resolveNoIntersectionLine(self, nodeI, matchNodeI, tr, reverse):
        """
        resolve a node with no intersection on it
        """

        # check the orientation of the external intersections if this line has any as compared to the reverse parameter
        if len(nodeI.intersections) > 1:
            transformableIntersections = []
            for i in nodeI.intersections:
                match = tr.getExtraMatch(i.value)
                if match:
                    transformableIntersections.append([i.value, match])
            if len(transformableIntersections) > 1:
                intReverse = ((transformableIntersections[0][0] < transformableIntersections[1][0])
                            != (tr.apply(transformableIntersections[0][1]) < tr.apply(transformableIntersections[1][1])))
                for i in range(1, len(transformableIntersections) - 1):
                    intCurr = transformableIntersections[i]
                    intNext = transformableIntersections[i + 1]
                    if ((intCurr[0] < intNext[0]) != (intCurr[1] < intNext[1])) != intReverse:
                        intReverse = None
                        break
                if intReverse is not None and intReverse != reverse:
                    return []


        transf = []
        # all permutation of line-segment combination
        for i in itertools.permutations(list(range(len(matchNodeI.forms))), len(nodeI.forms)):
            check = True

            # reject matches with mismatching sorts and attributes
            for j in range(len(i)):
                # reject matches with different sort
                if nodeI.forms[j].sort != matchNodeI.forms[i[j]].sort:
                    check = False
                    break
                # check attributes
                if nodeI.forms[j].sort.type == 'sort-attribute' :
                    if (not reverse) and not nodeI.forms[j].attribute.checkPartOf(matchNodeI.forms[i[j]].attribute, flagDesc = False):
                        check = False
                        break
                    elif reverse and not nodeI.forms[j].attribute.checkPartOf(matchNodeI.forms[i[len(i) - j - 1]].attribute, flagDesc = False):
                        check = False
                        break
            if not check:
                continue

            newtr = tr.duplicate()

            # add the line according to the orientation
            if not reverse:
                for j in range(len(i)):
                    newtr.add(nodeI.forms[j].start, matchNodeI.forms[i[j]].start)
                    newtr.add(nodeI.forms[j].end, matchNodeI.forms[i[j]].end)
            else:
                for j in range(len(i)):
                    k = len(i) - j - 1
                    newtr.add(nodeI.forms[j].start, matchNodeI.forms[i[k]].end)
                    newtr.add(nodeI.forms[j].end, matchNodeI.forms[i[k]].start)
            if self.orderingCheck([newtr], nodeI, True):
                transf.append(newtr)
        return transf
        
    def resolveFreePoints(self, transf):
        """
        resolve points that are not contained by any node(line)
        """

        # all lhs free points
        fPoints = []
        for p in self.nodeList.points:
            if p.oType == 'f':
                fPoints.append(p.value)
        if len(fPoints) == 0:
            return True

        # all shape free points
        mFPoints = []
        for p in self.matchingNodeList.points:
            if p.oType == 'f':
                mFPoints.append(p.value)
        if len(fPoints)>len(mFPoints):
            return False
            
        # match through permutation
        transfIndex = 0
        while transfIndex < len(transf):
            tr = transf.pop(transfIndex)
            for i in itertools.permutations(mFPoints, len(fPoints)):
                newTr = tr.duplicate()
                check = True
                for j in range(len(fPoints)):
                    if not newTr.add(fPoints[j], i[j]):
                        check = False
                if not check:
                    continue
                transf.insert(transfIndex, newTr)
                transfIndex += 1
        return True
        
        
    def resolvePointsOnLine(self, nodeI, matchNodeI, reverse, transf):
        """
        resolve points that are contained by a plane but not contained by any line
        """
        # for each point on the node
        for p in nodeI.linePoints:
            transfIndex = 0
            while transfIndex < len(transf):
                if transf[transfIndex].contains(p):
                    transfIndex += 1
                    continue
                tr = transf.pop(transfIndex)

                # get all matched points
                # nodeInts = tr.getCheckPoints(nodeI.nodeElement)
                nodeInts = self.getCheckPoints(tr, nodeI, True)

                # find the range in which the point's match should be
                if len(nodeInts) == 0:
                    posRange = (None, None)
                elif not reverse:
                    if p.value < nodeInts[0][0]:
                        posRange = (None, nodeInts[0][1])
                    elif p.value > nodeInts[-1][0]:
                        posRange = (nodeInts[-1][1], None)
                    else:
                        rangeIndex = 1
                        while p.value > nodeInts[rangeIndex][0]:
                            rangeIndex += 1
                        posRange = (nodeInts[rangeIndex-1][1], nodeInts[rangeIndex][1])
                else:
                    if p.value < nodeInts[0][0]:
                        posRange = (nodeInts[0][1], None)
                    elif p.value > nodeInts[-1][0]:
                        posRange = (None, nodeInts[-1][1])
                    else:
                        rangeIndex = 1
                        while p.value > nodeInts[rangeIndex][0]:
                            rangeIndex += 1
                        posRange = (nodeInts[rangeIndex][1], nodeInts[rangeIndex-1][1])
                if (posRange[0] is not None and posRange[1] is not None 
                and posRange[0] > posRange[1]):
                    posRange = [posRange[1], posRange[0]]
                    #continue

                # match the points
                posMatch = []
                for mP in matchNodeI.linePoints:
                    if ((posRange[0] is not None and mP.value<posRange[0])
                    or (posRange[1] is not None and mP.value>posRange[1])):
                        continue
                    elif ((posRange[1] is None and mP.value==posRange[0])
                    or (posRange[0] is None and mP.value==posRange[1])):
                        if tr.containsMatch(mP.value):
                            continue
                    posMatch.append(mP.value)

                # add the match
                for i in posMatch:
                    newTransf = tr.duplicate()
                    if not newTransf.add(p.value, i):
                        continue

                    ########
                    # if (newTransf.containsMatch(i) and not newTransf.contains(p.value)) or not newTransf.add(p.value, i):
                    #     continue

                    transf.insert(transfIndex, newTransf)
                    transfIndex+=1
                    
    def resolvePlanes(self, transf, match):
        """
        resolve plane-segments
        """ 

        # for each of the lhs planes
        for obj in self.nodeList.planes:
            plane = obj.value
            transfIndex = 0

            # go through the transformation sets
            while transfIndex < len(transf):
                tr = transf.pop(transfIndex)
                
                pCheck = True
                # check each plane polygon for existing matches
                for poly in plane.boundaries[0].polygonal:
                    if not tr.contains(poly):
                        pCheck = False
                        break

                # if all plane polygons already have matches, add the transformation set and continue to the next one
                if pCheck:
                    transf.insert(transfIndex, tr)
                    transfIndex += 1      
                    continue
                
                # if not all plane polygons have matches
                index = 0
                while index < len(plane.boundaries[0].polygonal):
                    if tr.contains(plane.boundaries[0].polygonal[index]):
                        pCheck = True
                        break
                    index += 1
                if pCheck:
                    mPoly = tr.apply(plane.boundaries[0].polygonal[index])
                    
                # for each of the shape planes, try to find a match
                for otherObj in self.matchingNodeList.planes:
                    otherPlane = otherObj.value

                    # compare sort, attribute and number of polygons
                    if (otherPlane.sort.baseSort() != plane.sort.baseSort() or (plane.hasAttribute() 
                    and (not plane.attribute == otherPlane.attribute))
                    or len(plane.boundaries[0].polygonal) != len(otherPlane.boundaries[0].polygonal)):
                        continue
                    
                    # if no point on the plane is matched -> continue on to other planes
                    if not pCheck:
                        continue
                    posMatch = []
                    check = False
                    
                    # find the match of the lhs point at index on the rhs plane
                    mIndex = 0
                    while mIndex < len(otherPlane.boundaries[0].polygonal):
                        if otherPlane.boundaries[0].polygonal[mIndex] == mPoly:
                            check = True
                            break
                        mIndex += 1

                    # if the match of the lhs point at index is not on this plane segment, continue to other plane segment
                    if not check:
                        continue
                    
                    # check if the lhs plane segment is matched in same order with the rhs plane segment
                    i = index + 1
                    j = mIndex + 1
                    for k in range(1, len(plane.boundaries[0].polygonal)):
                        if i >= len(plane.boundaries[0].polygonal):
                            i -= len(plane.boundaries[0].polygonal)
                        if j >= len(plane.boundaries[0].polygonal):
                            j -= len(plane.boundaries[0].polygonal)
                            
                        if tr.contains(plane.boundaries[0].polygonal[i]):
                            newPoly = tr.apply(plane.boundaries[0].polygonal[i])
                            if otherPlane.boundaries[0].polygonal[j] != newPoly:
                                check = False
                                break
                        else:
                            posMatch.append([plane.boundaries[0].polygonal[i], otherPlane.boundaries[0].polygonal[j]])
                        i += 1
                        j += 1

                    # check if the lhs plane segment is matched in reverse order with the rhs plane segment
                    if not check:
                        check = True
                        i = index + 1
                        j = mIndex - 1
                        for k in range(1, len(plane.boundaries[0].polygonal)):
                            if i >= len(plane.boundaries[0].polygonal):
                                i -= len(plane.boundaries[0].polygonal)
                            if j < 0:
                                j += len(plane.boundaries[0].polygonal)
                            if tr.contains(plane.boundaries[0].polygonal[i]):
                                newPoly = tr.apply(plane.boundaries[0].polygonal[i])
                                if otherPlane.boundaries[0].polygonal[j] != newPoly:
                                    check = False
                                    break
                            else:
                                posMatch.append([plane.boundaries[0].polygonal[i], otherPlane.boundaries[0].polygonal[j]])
                            i += 1
                            j -= 1

                    # if it does not match in either same or reverse order, continue
                    if not check:
                        continue

                    # add match
                    newTr = tr.duplicate()
                    for pM in posMatch:
                        if not newTr.add(pM[0], pM[1]):
                            check = False
                            break
                    if not check:
                        continue
                    
                    for i in range(len(self.nodeList.lines)):
                        nodeI = self.nodeList.lines[i]
                        # nodeInts = tr.getCheckPoints(nodeI.nodeElement)
                        nodeInts = self.getCheckPoints(tr, nodeI, False)
                        if len(nodeInts) < 2:
                            continue
                        
                        if nodeInts[0][1] > nodeInts[-1][1]:
                            reverse = True
                        else:
                            reverse = False
                            
                        polygons = []
                        for p in plane.boundaries[0].polygonal:
                            if nodeI.nodeElement.containsPoint(p):
                                polygons.append(p)
                        if len(polygons) == 0:
                            continue
                        
                        for i in polygons:
                            j = newTr.transform(i)
                            if j is False:
                                print('! ERROR: missing vector !',i)
                                newTr.printTransf()
                                raise ValueError('Missing vector in Transformation')
                            if i < nodeInts[0][0]:
                                if (reverse and j < nodeInts[0][1]) or (not reverse and j > nodeInts[0][1]):
                                    check = False
                                    break
                            elif i > nodeInts[-1][0]:
                                if (reverse and j > nodeInts[-1][1]) or (not reverse and j < nodeInts[-1][1]):
                                    check = False
                                    break
                            else:
                                index = 0
                                while index < len(nodeInts) and nodeInts[index][0] < i:
                                    index += 1
                                if not (nodeInts[index-1][1] <= j <= nodeInts[index][1]
                                or nodeInts[index][1] <= j <= nodeInts[index-1][1]):
                                    check = False
                                    break
                        if not check:
                            break
                    if not check:
                        continue
                    
                    transf.insert(transfIndex, newTr)
                    transfIndex += 1


    def resolveCurves(self, transf):
        """
        resolve curves (bezier curves, ellipses, circles only)
        """
        if len(self.nodeList.curves) == 0:
            return True
        if len(self.nodeList.curves) > len(self.matchingNodeList.curves):
            for i in range(len(transf)):
                transf.pop(0)
        tr = transf[0]
        circleList = []
        ellipseList = []
        bezierList = []
        for i in self.nodeList.curves:
            containedVects = []
            if i.value.type == 'ind-bezierCurveP2D':
                bezierList.append(i)
                continue
            for j in tr.vects:
                if i.value.containsPoint(j[0]):
                    containedVects.append(j[0])
            if len(containedVects) >= 1:
                i.oType = 't'
                if i.value.type == 'ind-ellipseP2D':
                    ellipseList.insert(0,i)
                if i.value.type == 'ind-circleP2D' and not tr.contains(i.value.center):
                    circleList.insert(0,i)
            else:
                if i.value.type == 'ind-ellipseP2D':
                    ellipseList.append(i)
                if i.value.type == 'ind-circleP2D':
                    circleList.append(i)
            i.assocObj = containedVects
            
        otherCircleList = []
        otherEllipseList = []
        otherBezierList = []
        for i in self.matchingNodeList.curves:
            if i.value.type == 'ind-bezierCurveP2D':
                otherBezierList.append(i)
            elif i.value.type == 'ind-circleP2D':
                otherCircleList.append(i)
            elif i.value.type == 'ind-ellipseP2D':
                otherEllipseList.append(i)
                
        transfIndex = 0
        while transfIndex < len(transf):
            tr = transf.pop(transfIndex)
            for i in itertools.permutations(otherBezierList, len(bezierList)):
                newTr = tr.duplicate()
                check = True
                for j in range(len(bezierList)):
                    mp1 = i[j].value.start
                    mp2 = i[j].value.controlPoint
                    mp3 = i[j].value.end
                    if newTr.contains(bezierList[j].value.controlPoint):
                        if newTr.apply(bezierList[j].value.controlPoint) != mp2:
                            check = False
                            break
                        mp2 = None
                    if newTr.contains(bezierList[j].value.start):
                        p1 = newTr.apply(bezierList[j].value.start)
                        if p1 == mp1:
                            mp1 = None
                        elif p1 == mp3:
                            mp3 = mp1
                            mp1 = None
                        else:
                            check = False
                            break
                    if newTr.contains(bezierList[j].value.end):
                        p3 = newTr.apply(bezierList[j].value.end)
                        if p3 == mp3:
                            mp3 = None
                        elif mp1 is not None and p3 == mp1:
                            mp1 = mp3
                            mp3 = None
                        else:
                            check = False
                            break
                    if mp1 is not None:
                        newTr.add(bezierList[j].value.start, mp1)
                    if mp2 is not None:
                        newTr.add(bezierList[j].value.controlPoint, mp2)
                    if mp3 is not None:
                        newTr.add(bezierList[j].value.end, mp3)
                if not check:
                    continue
                transf.insert(transfIndex, newTr)
                transfIndex += 1

        transfIndex = 0
        while transfIndex < len(transf):
            tr = transf.pop(transfIndex)
            for i in itertools.permutations(otherCircleList, len(circleList)):
                newTr = tr.duplicate()
                check = True
                for j in range(len(circleList)):
                    if circleList[j].oType == 't':
                        for v0 in circleList[j].assocObj:
                            v1 = tr.apply(v0)
                            if not i[j].value.containsPoint(v1):
                                check = False
                                break
                        if not check:
                            break
                        newTr.add(circleList[j].value.center, i[j].value.center, override = True)
                    else:
                        if (not newTr.add(circleList[j].value.center, i[j].value.center)
                        or not newTr.add(circleList[j].value.start, i[j].value.start) 
                        or not newTr.add(circleList[j].value.end, i[j].value.end)):
                            check = False
                            break
                if not check:
                    continue
                transf.insert(transfIndex, newTr)
                transfIndex += 1
        transfIndex = 0
        while transfIndex < len(transf):
            tr = transf.pop(transfIndex)
            for i in itertools.permutations(otherEllipseList, len(ellipseList)):
                newTr = tr.duplicate()
                check = True
                for j in range(len(ellipseList)):
                    if ellipseList[j].oType == 't':
                        for v0 in ellipseList[j].assocObj:
                            v1 = tr.apply(v0)
                            if not i[j].value.containsPoint(v1):
                                check = False
                                break
                        if not check:
                            break
                        newTr.add(ellipseList[j].value.foci[0], i[j].value.foci[0], override = True)
                        newTr.add(ellipseList[j].value.foci[1], i[j].value.foci[1], override = True)
                    else:
                        if (not newTr.add(ellipseList[j].value.foci[0], i[j].value.foci[0])
                        or not newTr.add(ellipseList[j].value.foci[1], i[j].value.foci[1])
                        or not newTr.add(ellipseList[j].value.start, i[j].value.start) 
                        or not newTr.add(ellipseList[j].value.end, i[j].value.end)):
                            check = False
                            break
                if not check:
                    continue
                transf.insert(transfIndex, newTr)
                transfIndex += 1
        return True
        
    def orderingCheck(self, transf, nodeI, outsideCheck = False):
        """
        Check the ordering of the intersections on a specific line

        For normal case (line has 2 or more intersections on it):

        * intersections contained by any line-segment on the line: must retain their ordering among one another after transformation
        * intersections not contained by any line-segment: 2 cases:
            * lhs intersections that are between the contained intersections: no restriction
            * lhs intersections that are outside of contained intersections (smaller/bigger than the smallest/biggest contained intersections)" their matches cannot be between the contained intersections (between the smallest/biggest contained intersections)
        
        For special cases (line has no intersection / line has only one intersection):

        * all intersections must retain their ordering
        """
        ind = 0
        containedInts = []
        externalInts = []
        # add intersections into containedInts and externalInts
        for ints in nodeI.intersections:
            if len(ints.nodes) < 2:
                continue
            if (not outsideCheck) and ints.iType == 'o':
                externalInts.append(ints.value)
            else:
                containedInts.append(ints.value)

        # if there are less than 2 contained intersections -> no ordering to be checked
        if len(containedInts) < 2:
            return True

        # sort the contained intersections
        containedInts = sorted(containedInts)

        # remove external intersections that are between the smallest and the biggest contained intersections
        exInd = 0
        while exInd < len(externalInts):
            if externalInts[exInd] > containedInts[0] and externalInts[exInd] < containedInts[-1]:
                externalInts.pop(exInd)
                continue
            exInd += 1

        # for each of the transformation set
        while ind < len(transf):
            tr = transf[ind]
            nodeInts = []
            # for each of the contained intersections, find the matching and add it to nodeInts
            for i in containedInts:
                if not tr.contains(i):
                    continue
                nodeInts.append([i,tr.apply(i)])

            # if there are fewer than 2 transformable intersections, there's no ordering to be checked 
            # -> continue to the next transformation set
            if len(nodeInts) < 2:
                ind += 1
                continue

            # check the first two contained intersections if the transformation reverse the intersection order or not
            reverse = False
            index = 1
            while index < len(nodeInts) and nodeInts[index][1] == nodeInts[index-1][1]:
                index += 1
            if index == len(nodeInts):
                ind += 1
                continue
            if nodeInts[index-1][1] < nodeInts[index][1]:
                reverse = False
            else:
                reverse = True

            # check the remaining intersection in pairs of 2 for the same ordering, if the ordering is violated,
            # remove the transformation
            check = True
            for index in range(1,len(nodeInts)):
                if ((not reverse and nodeInts[index-1][1] > nodeInts[index][1])
                or (reverse and nodeInts[index-1][1] < nodeInts[index][1])):
                    transf.pop(ind)
                    check = False
                    break
            if not check:
                continue
            
            # assign smallest, largest as smallest, largest contained intersections on the line
            smallest = nodeInts[0][1]
            largest = nodeInts[-1][1]
            if reverse:
                smallest, largest = largest, smallest

            # if any of the external intersection's match lies between the smallest and largest intersections 
            # -> discard the transformation
            for exInt in externalInts:
                exMatch = tr.getExtraMatch(exInt)
                if exMatch is None:
                    continue
                if exMatch > smallest and exMatch < largest:
                    transf.pop(ind)
                    check = False
                    break

            if check:
                ind += 1
        if len(transf) == 0 :
            return False
        return True
            
    def addOrientCheck(self, orientChecks, item):
        """
        add intersection points into orientationCheck list for ordering check
        """
        for i in range(len(orientChecks)):
            if orientChecks[i] == item:
                return
            elif orientChecks[i][0] > item[0]:
                orientChecks.insert(i, item)
                return
        orientChecks.append(item)
                
    def orientationChecking(self, orientChecks):
        """
        checking if intersection points on a line is in correct orientation relative to other intersection points
        """
        reverse = None
        ccwMatch = None
        if len(orientChecks) > 2:
            o = orientChecks[0]
            for index in range(2, len(orientChecks)):
                x = orientChecks[index-1]
                y = orientChecks[index]
                if o[0].colinear(x[0],y[0]):
                    if reverse is None:
                        a = o[0] < x[0]
                        b = o[1] < x[1]
                        reverse = (a!=b)
                    a = x[0] < y[0]
                    b = x[1] < y[1]
                    if reverse == (a==b):
                        return False
                else:
                    if ccwMatch is None:
                        a = graphMap.cw(o[0],x[0],y[0])
                        b = graphMap.cw(o[1],x[1],y[1])
                        ccwMatch = (a == b)
                    else:
                        a = graphMap.cw(o[0],x[0],y[0])
                        b = graphMap.cw(o[1],x[1],y[1])
                        if ccwMatch != (a == b):
                            return False
        return True
        
    def checkIntersectionPoints(self, transf):
        """
        Ensure all the existing points that lie on lines to have correct match 
        """
        ind = 0
        while ind < len(transf):
            tr = transf[ind]
            for p in self.nodeList.points:
                if p.oType != 'i':
                    continue
                transformedPoint = tr.allTransform(p.value)
                check = False
                for mP in self.matchingNodeList.points:
                    if mP.oType != 'i':
                        continue
                    if mP.value == transformedPoint:
                        check = True
                        break
                if not check:
                    transf.pop(ind)
                    ind -= 1
                    break
            ind += 1
        if len(transf) == 0:
            return False
        return True
        
    @staticmethod
    def normalizeLineEndPoints(nodeI, tr, reverse):
        for i in range(len(nodeI.forms)):
            li = nodeI.forms[i]
            trliEnd = tr.apply(li.end)
            for j in range(len(nodeI.forms)):
                if i == j:
                    continue
                lj = nodeI.forms[j]
                if li.attribute != lj.attribute or not li.end < lj.start:
                    continue
                trljStart = tr.apply(lj.start)
                if trliEnd > trljStart and not reverse:
                    nVect = trliEnd.add(trljStart).scale(c0p5)
                    tr.add(li.end, nVect, override = True)
                    tr.add(lj.start, nVect, override = True)
                if trliEnd < trljStart and reverse:
                    nVect = trliEnd.add(trljStart).scale(c0p5)
                    tr.add(li.end, nVect, override = True)
                    tr.add(lj.start, nVect, override = True)

    def resolveColinearEndPoints(self, transf):
        for colinearSet in self.colinearInts:
            pti, ptj, ptk, li, lj, lk = colinearSet
            for transfIndex in range(len(transf)):
                tr = transf.pop(0)
                trSet = transformSet()
                trSet.add(tr)
                trPti = tr.apply(pti.value)
                trPtj = tr.apply(ptj.value)
                trPtk = tr.apply(ptk.value)
                if trPti.colinear(trPtj, trPtk):
                    transf.append(tr)
                    continue
                if pti.endP and trPtj != trPtk:
                    self.adjustColinearEndPoints(tr, trSet, transf, trPtj, trPtk, pti.value, li)
                if ptj.endP and trPti != trPtk:
                    self.adjustColinearEndPoints(tr, trSet, transf, trPti, trPtk, ptj.value, lj)
                if ptk.endP and trPti != trPtj:
                    self.adjustColinearEndPoints(tr, trSet, transf, trPti, trPtj, ptk.value, lk)

    def adjustColinearEndPoints(self, tr, trSet, transf, anchor1, anchor2, movingPoint, movingNode):
        newLine = line2D(sortClass.getSort('__dummyLine2D__'), anchor1, anchor2)
        trLine = movingNode.nodeElement.edit(trSet)
        if newLine.intersect(trLine):
            newMovingPoint = newLine.intersection(trLine)
            check = True
            for lnSeg in movingNode.forms:
                if lnSeg.start == movingPoint or lnSeg.end == movingPoint:
                    trLnSeg = lnSeg.edit(trSet)
                    if not trLnSeg.containsPoint(newMovingPoint):
                        check = False
                        break
            if check:
                newTr = tr.duplicate()
                newTr.add(movingPoint, newMovingPoint, True)
                transf.append(newTr)

    def getCheckPoints(self, transf, lineNode, filtering = False):
        checkPoints = transf.getCheckPoints(lineNode.nodeElement)
        if filtering:
            i = 0
            while i < len(checkPoints):
                if len(checkPoints) <= 2:
                    break
                pt = checkPoints[i][0]
                isMin = True
                isMax = True
                for line in lineNode.forms:
                    if pt > line.start: isMin = False
                    if pt < line.end:   isMax = False
                    if not (isMin or isMax): break
                if not (isMin or isMax):
                    i += 1
                    continue
                for point in lineNode.containedPoints:
                    if pt > point.value: isMin = False
                    if pt < point.value: isMax = False
                    if not (isMin or isMax): break
                if isMin or isMax:
                    checkPoints.pop(i)
                    continue
                i += 1
        return checkPoints

    @staticmethod
    def getIntersection(line1, line2):
        """
        Get the intersection between two nodes
        """
        for i in line1.intersections:
            for j in line2.intersections:
                if i == j:
                # if i == j and (i.iType == 'i' or j.iType == 'i'):
                    return i
        return None

            


    # ============================================================================================
    #                            PREDICATE/DIRECTIVE SOLVING PROCESS
    # ============================================================================================

    
    def evaluatePredicate(self, transfSet, lhs, tlhs, form, param):   
        """
        resolve all predicates in the rule
        """
        graphTransf = self.getTransform(transfSet)
        if isinstance(param, dict):
            k = param
            param = paramSet()
            param.keys = k
        param.lhs = tlhs
        self.postCheck = True
        for cond in self.predicates:
            # maxline predicate
            if cond[0] == 'maxline':
                for keyI in cond[1]:
                    ind = keyI.ind
                    transfInd = ind.transform(transfSet)
                    form.accept(self, direction = [cond[0], transfInd])
            # bound predicate
            elif cond[0] == 'bound':
                for keyI in cond[1]:
                    ind = keyI.ind.transform(transfSet)
                    transfStart = graphTransf.apply(keyI.ind.start)
                    transfEnd = graphTransf.apply(keyI.ind.end)
                    form.accept(self, direction = [cond[0], ind, keyI.properties, transfStart, transfEnd])
                    
            # void predicate
            elif cond[0] == 'void':
                for bd in cond[1]:
                    # if there is a list of sort
                    if isinstance(bd[0], list):
                        sorts = bd[0]
                        bdr = bd[1:]
                        # create a plane boundary based on the vectors
                        bound = boundary()
                        if bdr[-1] == bdr[0]:
                            bdr.pop(-1)
                        for keyI in bdr:
                            newInd = graphTransf.transform(keyI)
                            if newInd is False:
                                graphTransf.printTransf()
                                raise ValueError('matching for vector '+str(keyI)+' is not found')
                            bound.add(newInd)
                        # remove duplicate vertices, add plane border segment
                        i = 0
                        while i < len(bound.polygonal):
                            if bound.polygonal[i] == bound.polygonal[i-1]:
                                bound.polygonal.pop(i)
                            else:
                                i += 1
                        for i in range(len(bound.polygonal)):
                            bound.add(bSegment(bound.polygonal[i],bound.polygonal[i-1]))
                        # create the plane-segment
                        planeSeg = planeSegment2D(lhs.sort, [bound], True)
                        # find all lines
                        lines = []
                        form.accept(self, direction = ['voidof', planeSeg, lines, sorts])
                        if not self.postCheck:
                            return False
                            
                        # check if any of the lines cut through the created plane-segment
                        check = planeSeg.classify_line(lines)
                        if not check.notInside():
                            self.postCheck = False
                    # if there's no sort list
                    else:
                        # create plane-segment boundary
                        bound = boundary()
                        if bd[-1] == bd[0]:
                            bd.pop(-1)
                        for keyI in bd:
                            newInd = graphTransf.transform(keyI)
                            if newInd is False:
                                graphTransf.printTransf()
                                raise ValueError('matching for vector '+str(keyI)+' is not found')
                            bound.add(newInd)
                        for i in range(len(bound.polygonal)):
                            bound.add(bSegment(bound.polygonal[i],bound.polygonal[i-1]))
                        # create plane-segment
                        planeSeg = planeSegment2D(lhs.sort, [bound], True)
                        # find all lines
                        lines = []
                        form.accept(self, direction = [cond[0], planeSeg, lines])
                        if not self.postCheck:
                            return False
                        # check if any of the lines cut through the created plane-segment
                        check = planeSeg.classify_line(lines)
                        if not check.notInside():
                            self.postCheck = False
            # inside predicate
            elif cond[0] == 'inside':
                for bd in cond[1]:
                    # if there is a list of sort
                    if isinstance(bd[0], list):
                        sorts = bd[0]
                        bdr = bd[1:]
                        # create a plane boundary based on the vectors
                        bound = boundary()
                        if bdr[-1] == bdr[0]:
                            bdr.pop(-1)
                        for keyI in bdr:
                            bound.add(keyI)
                        # remove duplicate vertices, add plane border segment
                        i = 0
                        while i < len(bound.polygonal):
                            if bound.polygonal[i] == bound.polygonal[i-1]:
                                bound.polygonal.pop(i)
                            else:
                                i += 1
                        for i in range(len(bound.polygonal)):
                            bound.add(bSegment(bound.polygonal[i],bound.polygonal[i-1]))
                        # create the plane-segment
                        planeSeg = planeSegment2D(lhs.sort, [bound], True)
                        # find all lines
                        lines = []
                        tlhs.accept(self, direction = ['insideof', planeSeg, lines, sorts])
                        if not self.postCheck:
                            return False
                            
                        # check if any of the lines cut through the created plane-segment
                        check = planeSeg.classify_line(lines)
                        if not check.isContained():
                            self.postCheck = False
                    # if there's no sort list
                    else:
                        # create plane-segment boundary
                        bound = boundary()
                        if bd[-1] == bd[0]:
                            bd.pop(-1)
                        for keyI in bd:
                            bound.add(keyI)
                        for i in range(len(bound.polygonal)):
                            bound.add(bSegment(bound.polygonal[i],bound.polygonal[i-1]))
                        # create plane-segment
                        planeSeg = planeSegment2D(lhs.sort, [bound], True)
                        # find all lines
                        lines = []
                        tlhs.accept(self, direction = [cond[0], planeSeg, lines])
                        if not self.postCheck:
                            return False
                        # check if any of the lines cut through the created plane-segment
                        check = planeSeg.classify_line(lines)
                        if not check.isContained():
                            self.postCheck = False
            # outside predicate
            elif cond[0] == 'outside':
                for bd in cond[1]:
                    # if there is a list of sort
                    if isinstance(bd[0], list):
                        sorts = bd[0]
                        bdr = bd[1:]
                        # create a plane boundary based on the vectors
                        bound = boundary()
                        if bdr[-1] == bdr[0]:
                            bdr.pop(-1)
                        for keyI in bdr:
                            bound.add(keyI)
                        # remove duplicate vertices, add plane border segment
                        i = 0
                        while i < len(bound.polygonal):
                            if bound.polygonal[i] == bound.polygonal[i-1]:
                                bound.polygonal.pop(i)
                            else:
                                i += 1
                        for i in range(len(bound.polygonal)):
                            bound.add(bSegment(bound.polygonal[i],bound.polygonal[i-1]))
                        # create the plane-segment
                        planeSeg = planeSegment2D(lhs.sort, [bound], True)
                        # find all lines
                        lines = []
                        tlhs.accept(self, direction = ['outsideof', planeSeg, lines, sorts])
                        if not self.postCheck:
                            return False
                            
                        # check if any of the lines cut through the created plane-segment
                        check = planeSeg.classify_line(lines)
                        if not check.notInside():
                            self.postCheck = False
                    # if there's no sort list
                    else:
                        # create plane-segment boundary
                        bound = boundary()
                        if bd[-1] == bd[0]:
                            bd.pop(-1)
                        for keyI in bd:
                            bound.add(keyI)
                        for i in range(len(bound.polygonal)):
                            bound.add(bSegment(bound.polygonal[i],bound.polygonal[i-1]))
                        # create plane-segment
                        planeSeg = planeSegment2D(lhs.sort, [bound], True)
                        # find all lines
                        lines = []
                        tlhs.accept(self, direction = [cond[0], planeSeg, lines])
                        if not self.postCheck:
                            return False
                        # check if any of the lines cut through the created plane-segment
                        check = planeSeg.classify_line(lines)
                        if not check.notInside():
                            self.postCheck = False
            # embeds predicate
            elif cond[0] == 'embeds':
                for keyI in cond[1]:
                    bigObj = keyI[0].ind.transform(transfSet)
                    for i in range(1, len(keyI)):
                        smallObj = keyI[i].ind.transform(transfSet)
                        if 'ind-planeSegment' in bigObj.type:
                            if 'ind-lineSegment' in smallObj.type:
                                check = bigObj.classify_line([smallObj])
                                if check.notInside():
                                    self.postCheck = False
                            elif 'ind-point' in smallObj.type:
                                check = bigObj.containsPoint(smallObj.value)
                                if not check:
                                    self.postCheck = False
                        elif 'ind-lineSegment' in bigObj.type and 'ind-point' in smallObj.type:
                            if not bigObj.containsPoint(smallObj.value):
                                self.postCheck = False
                        else:
                            self.postCheck = False
            # shortes_line/longest_line predicate
            elif cond[0] == 'shortest_line' or cond[0] == 'longest_line':
                for keyI in cond[1]:
                    if isinstance(keyI, list):
                        sorts = keyI[0]
                        if isinstance(sorts, str):
                            sorts = [sorts]
                        for i in keyI[1:]:
                            nInd = i.ind.transform(transfSet)
                            tlhs.accept(self, direction = [cond[0], nInd.squareLength(), sorts])
                    else:
                        ind = keyI.ind.transform(transfSet)
                        tlhs.accept(self, direction = [cond[0], ind.squareLength(), [ind.sort.baseSort().name]])
            # nolabel predicate
            elif cond[0] == 'nolabel':
                for keyI in cond[1]:
                    ind = keyI.ind.transform(transfSet)
                    form.accept(self, direction = [cond[0], ind])
            # *description* predicate
            elif cond[0] == 'description':
                for keyI in cond[1]:
                    val = description(sortClass.getSort('__dummyDesc__'), keyI[0])
                    try:
                        val = val.evaluateInd(param)
                    except Exception:
                        self.postCheck = False
                        return False
                    # '{}' -> equal to one of the values in the curly bracket
                    if keyI[1] == '{':
                        check = False
                        for comp in keyI[-1]:
                            
                            comparand = description(sortClass.getSort('__dummyDesc__'), comp)
                            comparand = comparand.evaluateInd(param)
                            if comparand == val:
                                check = True
                                break
                    # '[]' -> within the range of the square bracket
                    elif keyI[1] == '[':
                        if len(keyI[-1]) != 2:
                            raise ValueError('[ ] condition only accepts 2 arguments as min and max. e.g. [1,3] --> value between 1 and 3')
                        comp1 = description(sortClass.getSort('__dummyDesc__'), keyI[-1][0])
                        comp1 = comp1.evaluateInd(param)
                        comp2 = description(sortClass.getSort('__dummyDesc__'), keyI[-1][-1])
                        comp2 = comp2.evaluateInd(param)
                        if comp1 < val < comp2:
                            check = True
                        else:
                            check = False
                    # normal comparison
                    else:
                        comparand = description(sortClass.getSort('__dummyDesc__'), keyI[-1])
                        comparand = comparand.evaluateInd(param)
                        if keyI[1] == '=':
                            check = val == comparand
                        if keyI[1] == '<':
                            check = val < comparand
                        if keyI[1] == '>':
                            check = val > comparand
                        if keyI[1] == '<=':
                            check = val <= comparand
                        if keyI[1] == '>=':
                            check = val >= comparand
                        if keyI[1] == '<>':
                            check = not (val == comparand)
                    if not check:
                        self.postCheck = False
            if self.postCheck is False:
                return False
        return True
            
    # ============================================================================================
    #                               RHS/DIRECTIVES SOLVING PROCESS
    # ============================================================================================

    def solveDirectives(self, transfSet, lhs, tlhs, rhs, form, param):
        """
        solve the new lines/planes/points along with the directives in the rhs
        """
        allTransforms = [transfSet]
        if isinstance(param, dict):
            k = param
            param = paramSet()
            param.keys = k
        param.lhs = tlhs
        param.rhs = rhs
        # print('\n\n============================================\nResolving Order:')
        # for node in self.rhsSolvingOrder:
        #     print(node.nodeElement.printRhino())
        #     print(node.rhsInfo['case'])

        # print('\nRHS Solving Order:')
        # for i in self.rhsSolvingOrder:
        #     print()
        #     print(i.nodeElement.printRhino())
        #     print('  CASE:',i.rhsInfo['case'])
        # print('\n\n')

        for node in self.rhsSolvingOrder:
            numTransfSet = len(allTransforms)
            if numTransfSet == 0:
                return allTransforms
            for i in range(numTransfSet):
                transf = allTransforms.pop(0)
                if node.nodeElement.type == 'ind-line2D':
                    newTransfs = self.solveRHSLine(node, transf, param)
                elif node.nodeElement.type == 'ind-pointP2D':
                    newTransfs = self.solveRHSPoint(node, transf, param)
                else:
                    continue
                for new in newTransfs:
                    allTransforms.append(new)
        i = 0
        while i < len(allTransforms):
            tr = self.getTransform(allTransforms[i])
            if len(tr.unresolved) > 0:
                transfResolvable = True
                while len(tr.unresolved) > 0:
                    unresolvedVect = tr.unresolved[0]
                    vectResolvable = False
                    for lhsNode in self.nodeList.lines:
                        if not lhsNode.nodeElement.containsPoint(unresolvedVect):
                            continue
                        for lineSeg in lhsNode.forms:
                            if not lineSeg.containsPoint(unresolvedVect):
                                continue
                            if not (tr.contains(lineSeg.start) and tr.contains(lineSeg.end)):
                                continue
                            self.solvePointOnLine(tr, lineSeg, [coordinate(randint(0, 100000)/100000)], unresolvedVect)
                            vectResolvable = True
                            break
                        break
                    for rhsNode in self.nodeList.rhsLines:
                        if not rhsNode.nodeElement.containsPoint(unresolvedVect):
                            continue
                        for lineSeg in rhsNode.forms:
                            if not lineSeg.containsPoint(unresolvedVect):
                                continue
                            if not (tr.contains(lineSeg.start) and tr.contains(lineSeg.end)):
                                continue

                            self.solvePointOnLine(tr, lineSeg, [coordinate(randint(0, 100000)/100000)], unresolvedVect)
                            vectResolvable = True
                            break
                        break
                    if not vectResolvable:
                        transfResolvable = False
                        break
                if transfResolvable:
                    i += 1
                else:
                    print('\nOne or More Unresolved Vects Left.....')
                    for j in tr.vects:
                        print(j[0])
                    print(tr.unresolved)
                    allTransforms.pop(i)
            else: i += 1
            # if len(tr.unresolved) > 0:
            #     print('\nMore Than One Unresolved Vects Left.....')
            #     for j in tr.vects:
            #         print(j[0])
            #     print(tr.unresolved)
            #     allTransforms.pop(i)
            # else:
            #     i += 1
        return allTransforms

    def solveRHSLine(self, node, transf, param):
        """
        solve a specific line node with a specific transformation set and update it
        """
        tr = self.getTransform(transf)
        param.transform = transf
        resultLines = []
        parallelProjection = None
        reverseCheck = None

        ##################### SOLVE LINE DIRECTION + ROOT #####################
        # TODO: TO BE TESTED
        # CASE_1PARALLEL_1PT
        # CASE_1PERP_1PT
        # CASE_ANGLE
        # CASE_POINT_LINE_DISTANCE
        # CASE_LINE_LINE_DISTANCE
        # CASE_LINE_LINE_DISTANCE_DIRECTION
        # CASE_DIRECTION

        ## normal case ##
        if node.rhsInfo['case'] == graphNode.node.line_casus.CASE_1PARALLEL_1PT:
            ### case: 1 parallel line, 1 contained point
            ln = node.rhsInfo['caseNodes'][0]
            pt = node.rhsInfo['caseNodes'][1]
            transformedLn = ln.nodeElement.edit(transf)

            # dir = parallel line direction
            resultDir = transformedLn.direction
            resultLinePoint = tr.transform(pt)

            if not resultLinePoint:
                return []

            resultLines.append(line2D(node.nodeElement.sort, resultLinePoint, resultLinePoint.add(resultDir)))

        elif node.rhsInfo['case'] == graphNode.node.line_casus.CASE_1PERP_1PT:
            ### case: 1 perpendicular line, 1 contained point
            ln = node.rhsInfo['caseNodes'][0]
            pt = node.rhsInfo['caseNodes'][1]
            transformedLn = ln.nodeElement.edit(transf)

            # calculate result dir with the perpendicular line direction
            resultDir = vector2D(transformedLn.unitDir.y, c0 - transformedLn.unitDir.x)
            resultLinePoint = tr.transform(pt)

            if not resultLinePoint:
                return []
            resultLines.append(line2D(node.nodeElement.sort, resultLinePoint, resultLinePoint.add(resultDir)))

        ###### DIRECTIVE CASES ######
        elif node.rhsInfo['case'] == graphNode.node.line_casus.CASE_ANGLE:
            ### case: angle directive, solvable plane, 1 contained point
            # MULTIPLE RESULTS POSSIBLE
            dirStatement = node.rhsInfo['caseNodes'][0]
            bounds, check = self.evaluateDesc(dirStatement, param, special = True)
            if not check:
                return []
            new, base, angle = bounds # pylint: disable=unbalanced-tuple-unpacking

            while angle > pi:
                angle -= 2*pi

            new = new.ind
            base = base.ind
            transformedBase = base.edit(transf)

            pt = node.rhsInfo['caseNodes'][1]

            transformedPt = tr.transform(pt)

            basePt1, basePt2, anchorVect, dirCheck = node.rhsInfo['caseNodes'][2:]
            if anchorVect is not None:
                trBasePt1 = tr.transform(basePt1)
                trBasePt2 = tr.transform(basePt2)
                trAnchor = tr.transform(anchorVect)
                if trBasePt1 is None or trBasePt2 is None or trAnchor is None:
                    anchorVect = None
                else:
                    # localized direction of transformedBase
                    d1 = trBasePt2.subtract(trBasePt1).unit()

                    trAnchorVect = tr.transform(anchorVect)

                    if trAnchorVect is False:
                        return []

                    if trAnchorVect.colinear(trBasePt1, trBasePt2):
                        anchorVect = None

            # sin and cos of the angle
            cosAngle = cos(angle.value)
            sinAngle = sin(angle.value)


            if anchorVect is None:
                # localized direction of transformedBase
                d1 = transformedBase.orientationDir()

                newDir = vector2D(cosAngle*d1.x-sinAngle*d1.y, sinAngle*d1.x+cosAngle*d1.y).subtract(vz)
                resultLines.append(line2D(new.sort, transformedPt, transformedPt.add(newDir)))

                sinAngle = - sinAngle
                newDir = vector2D(cosAngle*d1.x-sinAngle*d1.y, sinAngle*d1.x+cosAngle*d1.y).subtract(vz)
                resultLines.append(line2D(new.sort, transformedPt, transformedPt.add(newDir)))
            else:
                cwCheck = graphMap.cw(trBasePt1, trBasePt2, trAnchorVect)
                if cwCheck == dirCheck:
                    sinAngle = -sinAngle
                newDir = vector2D(cosAngle*d1.x-sinAngle*d1.y, sinAngle*d1.x+cosAngle*d1.y).subtract(vz)
                resultLines.append(line2D(new.sort, transformedPt, transformedPt.add(newDir)))

        elif node.rhsInfo['case'] == graphNode.node.line_casus.CASE_POINT_LINE_DISTANCE:
            ### case: point-line distance directive, solvable plane, 1 contained point
            # MULTIPLE RESULTS POSSIBLE
            dirStatement = node.rhsInfo['caseNodes'][0]
            bounds, check = self.evaluateDesc(dirStatement, param, special = True)
            if not check:
                return []
            new = bounds[0].ind
            base = bounds[1].ind
            dist = bounds[2]
            transformedBase = base.edit(transf).value

            pt = node.rhsInfo['caseNodes'][1]

            transformedPt = tr.transform(pt)
            if transformedPt is False:
                return []

            # direction from transformed base point to transformed cPoint
            d1 = transformedBase.subtract(transformedPt)
            d1len = d1.length()

            #     (transformedPt)
            #           |\ <-- angle
            #           | \
            #           |  \
            #           |   \
            #           |    \
            #    result |     \
            #      line |      \ [d1len]
            #           |       \
            #           |        \
            #           |         \
            #           |          \
            #           |           \
            #           |_           \
            #           |_|_ _ _ _ _ _\ (transformedBase)
            #               [dist]

            # sin(angle) = dist/d1len
            sinAngle = dist / d1len

            # distance from projection of transformed base point onto the new line to transformed cPoint
            try:
                # cos(angle) = sqrt(d1len^2 + dist^2) / d1len ()
                cosAngle = (d1len*d1len - dist*dist).sqrt() / d1len
            except ValueError:
                return []
            

            anchorVect = node.rhsInfo['caseNodes'][2]
            dirCheck = node.rhsInfo['caseNodes'][3]

            if anchorVect is not None:
                trAnchorVect = tr.transform(anchorVect)

                if trAnchorVect is False:
                    return []

                if trAnchorVect.colinear(transformedBase, transformedPt):
                    anchorVect = None

            if anchorVect is None:
                newDir = vector2D(cosAngle*d1.x-sinAngle*d1.y, sinAngle*d1.x+cosAngle*d1.y).subtract(vz)
                resultLines.append(line2D(new.sort, transformedPt, transformedPt.add(newDir)))

                sinAngle = - sinAngle
                newDir = vector2D(cosAngle*d1.x-sinAngle*d1.y, sinAngle*d1.x+cosAngle*d1.y).subtract(vz)
                resultLines.append(line2D(new.sort, transformedPt, transformedPt.add(newDir)))
            else:
                cwCheck = graphMap.cw(transformedBase, transformedPt, trAnchorVect)
                if cwCheck != dirCheck:
                    sinAngle = -sinAngle

                # direction of the transformed new line
                newDir = vector2D(cosAngle*d1.x-sinAngle*d1.y, sinAngle*d1.x+cosAngle*d1.y).subtract(vz)

                resultLines.append(line2D(new.sort, transformedPt, transformedPt.add(newDir)))

        elif node.rhsInfo['case'] == graphNode.node.line_casus.CASE_LINE_LINE_DISTANCE:
            ### case: line-line distance directive with no direction vector
            # MULTIPLE RESULTS POSSIBLE
            dirStatement = node.rhsInfo['caseNodes'][0]
            bounds, check = self.evaluateDesc(dirStatement, param, special = True)
            if not check:
                return []

            new = bounds[0].ind
            base = bounds[1].ind
            dist = bounds[2]
            transformedBase = base.edit(transf)


            anchorVect = node.rhsInfo['caseNodes'][1]
            dirCheck = node.rhsInfo['caseNodes'][2]

            if anchorVect is not None:
                trAnchorVect = tr.transform(anchorVect)
                if trAnchorVect is False or transformedBase.projection(trAnchorVect) == trAnchorVect:
                    anchorCheck = self.replaceCaseAnchor(tr, node, 1, base, transformedBase)
                    if anchorCheck is None:
                        anchorVect = None
                    else:
                        anchorVect, trAnchorVect = anchorCheck
                        dirCheck = node.rhsInfo['caseNodes'][2]
                if anchorVect is not None:
                    projTrAnchor = transformedBase.projection(trAnchorVect)

            # if anchorVect and dirCheck ares None -> two result
            if anchorVect is None:
                distDir = vector2D(c0 - transformedBase.unitDir.y, transformedBase.unitDir.x0).scale(dist)
                resultLines.append(line2D(node.nodeElement.sort, transformedBase.start.add(distDir), transformedBase.end.add(distDir)))
                resultLines.append(line2D(node.nodeElement.sort, transformedBase.start.subtract(distDir), transformedBase.end.subtract(distDir)))
            else:
                distDir = trAnchorVect.subtract(projTrAnchor).unit().scale(dist)
                if not dirCheck:
                    distDir = distDir.scale(cn1)
                resultLines.append(line2D(node.nodeElement.sort, transformedBase.start.add(distDir), transformedBase.end.add(distDir)))

            # special line end-points matching: if the new line's start/end points are projection of the base line's start/end points
            parallelProjection = [base]

        elif node.rhsInfo['case'] == graphNode.node.line_casus.CASE_LINE_LINE_DISTANCE_DIRECTION:
            ### case: line-line distance directive with direction vector
            dirStatement = node.rhsInfo['caseNodes'][0]
            bounds, check = self.evaluateDesc(dirStatement, param, special = True)
            if not check:
                return []

            new = bounds[0].ind
            base = bounds[1].ind
            dist = bounds[2]
            distDirection = bounds[3]

            transformedBase = base.edit(transf)

            if not isinstance(distDirection, vector2D):
                distDirection = vector2D(*vector2D.readVector(parseReader(distDirection))).unit()

            resultLines.append(line2D(node.nodeElement.sort, 
                                transformedBase.start.add(distDirection.scale(dist)),
                                transformedBase.end.add(distDirection.scale(dist))))

            # special line end-points matching: if the new line's start/end points are projection of the base line's start/end points
            parallelProjection = [base]

        elif node.rhsInfo['case'] == graphNode.node.line_casus.CASE_DIRECTION:
            ### case: direction directive, 1 contained point
            dirStatement = node.rhsInfo['caseNodes'][0]
            bounds, check = self.evaluateDesc(dirStatement, param, special = True)

            if not check:
                return []

            if isinstance(bounds[1], vector2D):
                newDirection = bounds[1]
            else:
                newDirection = vector2D(*vector2D.readVector(parseReader(bounds[1]))).unit()

            containedPoint = node.rhsInfo['caseNodes'][1]
            trContainedPoint = tr.transform(containedPoint)
            if not trContainedPoint:
                return []
            reverseCheck = False
            resultLines.append(line2D(node.nodeElement.sort, trContainedPoint, trContainedPoint.add(newDirection)))

        ###### OTHER CASES ######s
        elif node.rhsInfo['case'] == graphNode.node.line_casus.CASE_PROXY:
            ### case: line already defined through other line cases
            try:
                resultLines.append(node.nodeElement.edit(transf))
            except IndexError:
                return []

        elif node.rhsInfo['case'] == graphNode.node.line_casus.CASE_LHS:
            ### case: line in the lhs -> no need to solve any more line segment, only need to solve point_on_line
            try:
                resultLines.append(node.nodeElement.edit(transf))
                for directive in node.rhsInfo['directives']:
                    if directive[0] == 'point_on_line':
                        ind = directive[1]
                        bounds = directive[2]
                        if len(directive) == 4:
                            pointOnLineTarget = directive[3]
                        else:
                            pointOnLineTarget = None
                        ratio, evalCheck = self.evaluateDesc(bounds, param)
                        if not evalCheck:
                            return []
                        self.solvePointOnLine(tr, ind, ratio, pointOnLineTarget)
                return [transf]
            except IndexError:
                return []

        if resultLines is None:
            return []

        ################### SOLVE INTERSECTIONS + END-POINTS ###################
        resultTransformations = []
        originalTransf = transf
        for resultLine in resultLines:
            transf = originalTransf.duplicate()
            tr = self.getTransform(transf)
            # intersection with lhs nodes
            for lhsNode in self.nodeList.lines:
                if lhsNode.nodeElement.intersect(node.nodeElement):
                    transformedLHSNode = lhsNode.nodeElement.edit(transf)
                    if not transformedLHSNode.intersect(resultLine):
                        continue
                    intersection = lhsNode.nodeElement.intersection(node.nodeElement)
                    transformedInt = transformedLHSNode.intersection(resultLine)
                    tr.add(intersection, transformedInt)

            # intersection with solved rhs nodes
            for rhsNode in self.nodeList.rhsLines:
                try:
                    transformedRHSNode = rhsNode.nodeElement.edit(transf)
                except IndexError:
                    continue
                if rhsNode.nodeElement.intersect(node.nodeElement):
                    if not transformedRHSNode.intersect(resultLine):
                        continue
                    intersection = rhsNode.nodeElement.intersection(node.nodeElement)
                    transformedInt = transformedRHSNode.intersection(resultLine)
                    tr.add(intersection, transformedInt)

            # find parallel projection if indicated by the case
            if parallelProjection is not None:
                for baseLine in parallelProjection:
                    self.parallelProjection(tr, baseLine, resultLine, node)


            # checkPoints = tr.getCheckPoints(node.nodeElement)
            checkPoints = self.getCheckPoints(tr, node, False)

            for directive in node.rhsInfo['directives']:
                if directive[0] == 'point_on_line':
                    ind = directive[1]
                    lengthDir = None
                    for i in node.rhsInfo['directives']:
                        if i[0] == 'length' and i[1] == ind:
                            lengthDir = i
                    if not self.solveOneRHSLineSeg(tr, param, node, resultLine, ind,
                    checkPoints, False, lengthDirective = lengthDir):
                        return []
                    bounds = directive[2]
                    if len(directive) == 4:
                        pointOnLineTarget = directive[3]
                    else:
                        pointOnLineTarget = None
                    ratio, evalCheck = self.evaluateDesc(bounds, param)
                    if not evalCheck:
                        return []
                    self.solvePointOnLine(tr, ind, ratio, pointOnLineTarget)


            # check if the line is matched in reverse order or not
            if reverseCheck is not None:
                reverse = reverseCheck
            else:
                reverse = None
                # _ if there are 2 transformable points or more
                while len(checkPoints) > 1:
                    if checkPoints[0][1] == checkPoints[1][1]:
                        checkPoints.pop(1)
                        continue
                    d0 = checkPoints[1][0].subtract(checkPoints[0][0]) > vz
                    d1 = checkPoints[1][1].subtract(checkPoints[0][1]) > vz
                    reverse = d0 != d1
                    break

                # _ if there's no transformable point on the line (this case is probably not possible)
                if len(checkPoints) == 0:
                    reverse  = None
                    # return []

                # _ if there's only one transformable point on the line
                elif len(checkPoints) == 1:
                    for vect in tr.vects:
                        # find a point projection on the line that is not
                        projvect0 = node.nodeElement.projection(vect[0])
                        if checkPoints[0][0] == projvect0:
                            continue
                        projvect1 = resultLine.projection(vect[1])
                        if checkPoints[0][1] == projvect1:
                            continue
                        d0 = projvect0.subtract(checkPoints[0][0]).unit() > vz
                        d1 = projvect1.subtract(checkPoints[0][1]).unit() > vz
                        # print(d0, d1)
                        # reverse = (d0.add(d1)).squareLength() == c0
                        reverse = d0 != d1
                        break

            # if reverse is None -> do both non-reverse and reverse transformation order
            # if reverse is False -> do only non-reverse
            # if reverse is True -> do only reverse
            if reverse is None or reverse == False:
                newTransf = transf.duplicate()
                newTr = self.getTransform(newTransf)
                param.transform = newTransf
                # solve the one with length directive first
                for directive in node.rhsInfo['directives']:
                    if directive[0] == 'length':
                        if not self.solveOneRHSLineSeg(newTr, param, node, resultLine, directive[1],
                        checkPoints, False, lengthDirective = directive):
                            return []
                for lineSeg in node.forms:
                    if not self.solveOneRHSLineSeg(newTr, param, node, resultLine, lineSeg, checkPoints, False):
                        return []
                resultTransformations.append(newTransf)

            if reverse is None or reverse == True:
                newTransf = transf.duplicate()
                newTr = self.getTransform(newTransf)
                param.transform = newTransf
                # solve the one with length directive first
                for directive in node.rhsInfo['directives']:
                    if directive[0] == 'length':
                        if not self.solveOneRHSLineSeg(newTr, param, node, resultLine, directive[1],
                        checkPoints, True, lengthDirective = directive):
                            return []
                for lineSeg in node.forms:
                    if not self.solveOneRHSLineSeg(newTr, param, node, resultLine, lineSeg, checkPoints, True):
                        return []
                resultTransformations.append(newTransf)

        # for transf in resultTransformations:
        #     tr = self.getTransform(transf)
        #     for directive in node.rhsInfo['directives']:
        #         if directive[0] == 'point_on_line':
        #             ind = directive[1]
        #             bounds = directive[2]
        #             ratio, evalCheck = self.evaluateDesc(bounds, param)
        #             if not evalCheck:
        #                 return []
        #             self.solvePointOnLine(tr, ind, ratio, singleRatio = True)
        return resultTransformations

    def solveRHSPoint(self, node, transf, param):
        """
        solve a specific point node with a specific transformation set and update it
        """
        tr = self.getTransform(transf)
        param.transform = transf
        if node.rhsInfo['case'] == graphNode.node.point_casus.CASE_POINT_POINT_DISTANCE_DIRECTION:
            ### case: point-point distance directive with direction vector
            dirStatement = node.rhsInfo['caseNodes'][0]
            bounds, check = self.evaluateDesc(dirStatement, param, special = True)
            if not check:
                return []

            new = bounds[0].ind.value
            base = bounds[1].ind.value
            dist = bounds[2]
            distDirection = bounds[3]

            transformedBase = tr.transform(base)

            if not transformedBase:
                return []

            if not isinstance(distDirection, vector2D):
                distDirection = vector2D(*vector2D.readVector(parseReader(distDirection))).unit()

            tr.add(new, transformedBase.add(distDirection.scale(dist)))
            return [transf]
        elif node.rhsInfo['case'] == graphNode.node.point_casus.CASE_POINT_POINT_DISTANCE:
            ### case: point-point distance directive with no direction vector
            dirStatement = node.rhsInfo['caseNodes'][0]
            bounds, check = self.evaluateDesc(dirStatement, param, special = True)
            if not check:
                return []

            new = bounds[0].ind.value
            base = bounds[1].ind.value
            dist = bounds[2]
            transformedBase = tr.transform(base)

            containingLine = node.rhsInfo['caseNodes'][1]
            trContainingLine = containingLine.edit(transf)

            anchorVect = node.rhsInfo['caseNodes'][2]
            dirCheck = node.rhsInfo['caseNodes'][3]

            if anchorVect is not None:
                trAnchorVect = tr.transform(anchorVect)

                dirx = vector2D(c0 - containingLine.direction.y, containingLine.direction.x)
                perpLine = line2D(containingLine.sort, base, base.add(dirx))
                trDirx = vector2D(c0 - trContainingLine.direction.y, trContainingLine.direction.x)
                trPerpLine = line2D(trContainingLine.sort, transformedBase, transformedBase.add(trDirx))
                if trAnchorVect is False or trPerpLine.projection(trAnchorVect) == trAnchorVect:

                    anchorCheck = self.replaceCaseAnchor(tr, node, 2, perpLine, trPerpLine)
                    if anchorCheck is None:
                        anchorVect = None
                    else:
                        anchorVect, trAnchorVect = anchorCheck
                        dirCheck = node.rhsInfo['caseNodes'][2]
                if anchorVect is not None:
                    projTrAnchor = trPerpLine.projection(trAnchorVect)

            projBase = trContainingLine.projection(transformedBase)
            distFromProjBase = (dist * dist - transformedBase.subtract(projBase).squareLength())
            if distFromProjBase < c0:
                return []
            distFromProjBase = distFromProjBase.sqrt()
            # if anchorVect and dirCheck ares None -> two result
            if anchorVect is None:
                distDir = trContainingLine.unitDir().scale(distFromProjBase)
                transf0 = transf.duplicate()
                tr0 = self.getTransform(transf0)
                tr0.add(new, projBase.add(distDir))
                transf1 = transf.duplicate()
                tr1 = self.getTransform(transf1)
                tr1.add(new, projBase.subtract(distDir))
                return [transf0, transf1]
            else:
                distDir = trAnchorVect.subtract(projTrAnchor).unit().scale(distFromProjBase)
                if not dirCheck:
                    distDir = distDir.scale(cn1)
                tr.add(new, projBase.add(distDir))
                return [transf]

        elif node.rhsInfo['case'] == graphNode.node.point_casus.CASE_LINE_POINT_DISTANCE:
            ### case: line-point distance directive
            dirStatement = node.rhsInfo['caseNodes'][0]
            bounds, check = self.evaluateDesc(dirStatement, param, special = True)
            if not check:
                return []

            new = bounds[0].ind.value
            base = bounds[1].ind
            dist = bounds[2]

            transformedBase = base.edit(transf)

            containingLine = node.rhsInfo['caseNodes'][1]
            trContainingLine = containingLine.edit(transf)

            if not trContainingLine.intersect(transformedBase):
                return []

            anchorVect = node.rhsInfo['caseNodes'][2]
            dirCheck = node.rhsInfo['caseNodes'][3]

            if anchorVect is not None:
                trAnchorVect = tr.transform(anchorVect)
                if trAnchorVect is False or transformedBase.projection(trAnchorVect) == trAnchorVect:
                    anchorCheck = self.replaceCaseAnchor(tr, node, 2, base, transformedBase)
                    if anchorCheck is None:
                        anchorVect = None
                    else:
                        anchorVect, trAnchorVect = anchorCheck
                        dirCheck = node.rhsInfo['caseNodes'][2]
                if anchorVect is not None:
                    projTrAnchor = transformedBase.projection(trAnchorVect)

            # if anchorVect and dirCheck ares None -> two result
            if anchorVect is None:
                moveDir = trContainingLine.unitDir().scale(dist)
                movedtrBase0 = line2D(base.sort, transformedBase.root.add(moveDir), transformedBase.root.add(transformedBase.direction).add(moveDir))
                transf0 = transf.duplicate()
                tr0 = self.getTransform(transf0)
                tr0.add(new, movedtrBase0.intersection(trContainingLine))
                movedtrBase1 = line2D(base.sort, transformedBase.root.subtract(moveDir), transformedBase.root.add(transformedBase.direction).subtract(moveDir))
                transf1 = transf.duplicate()
                tr1 = self.getTransform(transf1)
                tr0.add(new, movedtrBase1.intersection(trContainingLine))
                return [transf0, transf1]
            else:
                moveDir = trAnchorVect.subtract(projTrAnchor).unit().scale(dist)
                movedtrBase = line2D(base.sort, transformedBase.root.add(moveDir), transformedBase.root.add(transformedBase.direction).add(moveDir))
                if not dirCheck:
                    distDir = distDir.scale(cn1)
                tr.add(new, movedtrBase.intersection(trContainingLine))
                return [transf]

    def solveOneRHSLineSeg(self, tr, param, node, resultLine, lineSeg, checkPoints, reverse, lengthDirective = None):
        startSolved = tr.contains(lineSeg.start)
        endSolved   = tr.contains(lineSeg.end)
        # if both start and end points are already in the transformation set -> return
        if startSolved and endSolved:
            # if there's a length directive, check the matched line if it has the correct length 
            if lengthDirective is not None:

                # get the length
                dirStatement = lengthDirective[-1]
                bounds, check = self.evaluateDesc(dirStatement, param)
                if not check:
                    return False
                length = bounds[-1]

                # if the length matches -> return True
                matchStart = tr.transform(lineSeg.start)
                matchEnd = tr.transform(lineSeg.end)
                oriSqLength = matchEnd.subtract(matchStart).squareLength()
                if oriSqLength == length * length:
                    return True

                # if the length does not match -> readjust the point that is added last (or return False)
                revMod = cn1 if reverse else c1
                for i in tr.vects:
                    if i[0] == lineSeg.start:
                        # adjust end point
                        readjusted = lineSeg.end
                        newMatch = matchStart.add(resultLine.unitDir.scale(revMod).scale(length))
                        break
                    elif i[0] == lineSeg.end:
                        readjusted = lineSeg.start
                        newMatch = matchEnd.subtract(resultLine.unitDir.scale(revMod).scale(length))
                        break
                tr.add(readjusted, newMatch, override = True)
            return True

        # find the line length
        # line case 1: solve with length from length directive
        length = None
        if lengthDirective is not None:
            dirStatement = lengthDirective[-1]
            bounds, check = self.evaluateDesc(dirStatement, param)
            if not check:
                return False
            length = bounds[-1]
        if length is None and len(node.rhsInfo['parallel']) > 0:
            # line case 2: a parallel line that has both start and end points being the projections 
            #              of the new line start and end points
            for parallelNode in node.rhsInfo['parallel']:
                # if line is not transformable, continue
                try:
                    parallelNode.nodeElement.edit(tr)
                except IndexError:
                    continue
                # find a parallel line that has equivalent start and end points (same length)
                for parallelLine in parallelNode.forms:
                    if (lineSeg.projection(parallelLine.start) == lineSeg.start
                    and lineSeg.projection(parallelLine.end  ) == lineSeg.end):
                        nStart = tr.transform(parallelLine.start)
                        nEnd = tr.transform(parallelLine.end)
                        if not nStart or not nEnd:
                            continue
                        tr.add(lineSeg.start, resultLine.projection(nStart))
                        tr.add(lineSeg.end, resultLine.projection(nEnd))
                        return True

            # line case 3: the start and end points are projections of some other parallel lines start/end points
            for parallelNode in node.rhsInfo['parallel']:
                # if line is not transformable, continue
                try:
                    parallelNode.nodeElement.edit(tr)
                except IndexError:
                    continue
                # find a line end-point that its projection is the start/end point of the new line
                for parallelLine in parallelNode.forms:

                    projS = lineSeg.projection(parallelLine.start)
                    projE = lineSeg.projection(parallelLine.end)
                    nStart = tr.transform(parallelLine.start)
                    nEnd = tr.transform(parallelLine.end)

                    if projS == lineSeg.start and nStart != False:
                        tr.add(lineSeg.start, resultLine.projection(nStart))
                    elif projS == lineSeg.end and nStart != False:
                        tr.add(lineSeg.end, resultLine.projection(nStart))

                    if projE == lineSeg.start and nEnd != False:
                        tr.add(lineSeg.start, resultLine.projection(nEnd))
                    elif projE == lineSeg.end and nEnd != False:
                        tr.add(lineSeg.end, resultLine.projection(nEnd))
            if tr.contains(lineSeg.start) and tr.contains(lineSeg.end):
                return True
        # line case 4: a perpendicular line with a known end point intersecting at the unmatched point
        if length is None and len(node.rhsInfo['perpendicular']) > 0 and (startSolved or endSolved):
            tobeSolved = lineSeg.end if startSolved else lineSeg.start
            for pNode in node.rhsInfo['perpendicular']:
                if not pNode.nodeElement.containsPoint(tobeSolved):
                    continue
                # if it's a intersection with another line and that line has a legit case but is not yet matched
                # -> wait until that line is matched
                if (pNode.rhsInfo and pNode.rhsInfo['case'] != graphNode.node.line_casus.CASE_PROXY
                and pNode.rhsInfo['case'] != graphNode.node.line_casus.CASE_LHS 
                and node not in pNode.rhsInfo['caseNodes']):
                    return True

                for l in pNode.forms:
                    if tr.contains(l.start):
                        tr.add(tobeSolved, resultLine.projection(tr.transform(l.start)))
                        return True
                    if tr.contains(l.end):
                        tr.add(tobeSolved, resultLine.projection(tr.transform(l.end)))
                        return True
        # case 5: if start or end is between two checkpoints
        # if end is in between two checkpoints
        if length is None and not endSolved and len(checkPoints) > 1:
            for i in range(len(checkPoints) - 1):
                if checkPoints[i][0] < lineSeg.end and checkPoints[i + 1][0] > lineSeg.end:
                    checkPointDist0 = lineSeg.end.subtract(checkPoints[i][0]).length()
                    checkPointDist1 = lineSeg.end.subtract(checkPoints[i+1][0]).length()
                    ratio = checkPointDist0 / (checkPointDist0 + checkPointDist1)
                    # if checkPointDist0 == checkPointDist1:
                    #     ratio = c0p5
                    # else:
                    #     ratio = coordinate(randint(0, 100000)/100000)
                    tr.add(lineSeg.end, checkPoints[i][1].add(checkPoints[i+1][1].subtract(checkPoints[i][1]).scale(ratio)))
                    endSolved = True

        if length is None and not startSolved and len(checkPoints) > 1:
            for i in range(len(checkPoints) - 1):
                if checkPoints[i][0] < lineSeg.start and checkPoints[i + 1][0] > lineSeg.start:
                    checkPointDist0 = lineSeg.start.subtract(checkPoints[i][0]).length()
                    checkPointDist1 = lineSeg.start.subtract(checkPoints[i + 1][0]).length()
                    ratio = checkPointDist0 / (checkPointDist0 + checkPointDist1)
                    # if checkPointDist0 == checkPointDist1:
                    #     ratio = c0p5
                    # else:
                    #     ratio = coordinate(randint(0, 100000)/100000)
                    tr.add(lineSeg.start, checkPoints[i][1].add(checkPoints[i+1][1].subtract(checkPoints[i][1]).scale(ratio)))
                    startSolved = True
        if endSolved and startSolved:
            return True

        # line case 6: the start or end points are projections of some random points
        for vec in tr.vects:
            vecProj = lineSeg.projection(vec[0])
            if not startSolved and vecProj == lineSeg.start:
                tr.add(lineSeg.start, resultLine.projection(vec[1]))
                startSolved = True
            if not endSolved and vecProj == lineSeg.end:
                tr.add(lineSeg.end, resultLine.projection(vec[1]))
                endSolved = True
            if startSolved and endSolved:
                return True

        # leftover case: solve with length, same length as its pretransformed segment
        if length is None:
            length = lineSeg.length()

        # revMod = -1 if reverse, revMod = 1 if not reverse
        revMod = cn1 if reverse else c1
        # if only start has match -> find end
        if startSolved:
            matchStart = tr.transform(lineSeg.start)
            matchEnd = matchStart.add(resultLine.unitDir.scale(revMod).scale(length))
            tr.add(lineSeg.end, matchEnd, override = True)
        # if only end has match -> find start
        elif endSolved:
            matchEnd = tr.transform(lineSeg.end)
            matchStart = matchEnd.subtract(resultLine.unitDir.scale(revMod).scale(length))
            tr.add(lineSeg.start, matchStart, override = True)
        # if neither start or end has match
        else:
            containedCheckPoints = []
            minPoint = [None]
            maxPoint = [None]
            for i in checkPoints:
                if lineSeg.start < i[0] < lineSeg.end:
                    containedCheckPoints.append(i)
                elif i[0] < lineSeg.start:
                    if minPoint[0] is None or minPoint[0] < i[0]:
                        minPoint = i
                elif i[0] > lineSeg.end:
                    if maxPoint[0] is None or maxPoint[0] > i[0]:
                        maxPoint = i
            if reverse:
                length = c0 - length

            # if exactly 1 contained point -> that point is the midpoint of the new line 
            if len(containedCheckPoints) == 1:
                midPoint = containedCheckPoints[0][1]
            # if more than 1 contained points -> mid point of the first and last contained points
            elif len(containedCheckPoints) > 1:
                midPoint = containedCheckPoints[0][1].add(containedCheckPoints[-1][1]).scale(c0p5)
            # if there are both minPoint and maxPoint -> midpoint of minpoint and maxpoint
            elif minPoint[0] is not None and maxPoint[0] is not None:
                midPoint = minPoint[1].add(maxPoint[1]).scale(c0p5)
            elif len(checkPoints) > 1:
                midPoint = checkPoints[0][1].add(checkPoints[-1][1]).scale(c0p5)
            elif len(checkPoints) == 1:
                midPoint = checkPoints[0][1]
            else:
                return False
            tr.add(lineSeg.start, midPoint.subtract(resultLine.unitDir.scale(length/2)), override = True)
            tr.add(lineSeg.end, midPoint.add(resultLine.unitDir.scale(length/2)), override = True)
        return True

    def parallelProjection(self, tr, base, resultLine, lineNode):
        projS = lineNode.nodeElement.projection(base.start)
        projE = lineNode.nodeElement.projection(base.end)
        for ln in lineNode.forms:
            if projS == ln.start:
                tr.add(ln.start, resultLine.projection(tr.apply(base.start)))
            elif projS == ln.end:
                tr.add(ln.end, resultLine.projection(tr.apply(base.start)))
            if projE == ln.start:
                tr.add(ln.start, resultLine.projection(tr.apply(base.end)))
            elif projE == ln.end:
                tr.add(ln.end, resultLine.projection(tr.apply(base.end)))

    def replaceCaseAnchor(self, tr, node, anchorIndex, baseSet, trBaseSet):
        for vec in tr.vects:
            proj0 = baseSet.projection(vec[0])
            if proj0 == vec[0]:
                continue
            trproj0 = trBaseSet.projection(vec[1])
            if trproj0 == vec[1]:
                continue

            orientationAnchor = vec[0].subtract(proj0).unit()
            if 'line' in node.nodeElement.type:
                root0 = node.nodeElement.root
            else:
                root0 = node.nodeElement.value
            projRoot0 = baseSet.projection(root0)
            orientationNew = root0.subtract(projRoot0).unit()
            dirCheck = orientationAnchor == orientationNew

            node.rhsInfo['caseNodes'][anchorIndex] = vec[0]
            node.rhsInfo['caseNodes'][anchorIndex + 1] = dirCheck
            return vec
        return None

    # def replacePointCaseAnchor(self, tr, node, anchorIndex, baseSet, trBaseSet):
    #     for vec in tr.vects:
    #         if vec[0] == node.nodeElement.value:
    #             continue
    #         proj0 = baseSet.projection(vec[0])
    #         if proj0 == vec[0]:
    #             continue
    #         trproj0 = trBaseSet.projection(vec[1])
    #         if trproj0 == vec[1]:
    #             continue
    #         orientationAnchor = vec[0].subtract(proj0).unit()
    #         root0 = node.nodeElement.value
    #         projRoot0 = baseSet.projection(root0)
    #         orientationNew = root0.subtract(projRoot0).unit()
    #         dirCheck = orientationAnchor == orientationNew

    #         node.rhsInfo['caseNodes'][anchorIndex] = vec[0]
    #         node.rhsInfo['caseNodes'][anchorIndex + 1] = dirCheck
    #         return vec
    #     return None

    def getTransform(self, tfs):
        """
        get the graphTransform from the transformation set
        """
        string = self.id.split("=|=")
        if len(string) > 1:
            tfs.current =  string[0]
        return tfs.retrieveTransform(graphTransform)

    def visitEnter(self, obj, direction = 0):
        """
        Visit enter for GraphMap: for solving predicates and directives
        """
        if direction[0] == 'maxline':
            if (obj.__class__ == direction[1].__class__ 
            and obj.sort == direction[1].sort 
            and (not obj.equals(direction[1]))
            and (obj.compareValue(direction[1]) == 0)
            and ((obj.containsPoint(direction[1].start) and obj.start < direction[1].start)
            or (obj.containsPoint(direction[1].end) and obj.end > direction[1].end))):
                self.postCheck = False
                return False
        elif direction[0] == 'bound':
            if (obj.__class__ == direction[1].__class__ 
            and obj.sort == direction[1].sort 
            and obj.contains(direction[1])):
                if 0 in direction[2] and obj.start != direction[3] and obj.end != direction[3]:
                    self.postCheck = False
                    return False
                if 1 in direction[2] and obj.start != direction[4] and obj.end != direction[4]:
                    self.postCheck = False
                    return False
        elif direction[0] == 'void':
            if 'ind-planeSegment' in obj.type :
                if ('2D' in obj.type) == ('3D' in direction[1].type):
                    raise ValueError('There exists a mix of 2D and 3D objects in the shape')
                check = obj.classify(direction[1])
                if check.disjoint() or check.touches():
                    return True
                self.postCheck = False
                return False
            if 'ind-lineSegment' in obj.type:
                direction[2].append(bSegment(obj.start, obj.end))
            if 'ind-point' in obj.type:
                check = False
                for line in direction[1].boundaries[0].canonical:
                    if line.containsPoint(obj.value):
                        check = True
                if check:
                    return True
                if direction[1].boundaries[0].containsPointInside(obj.value):
                    self.postCheck = False
                    return False
        elif direction[0] == 'voidof':
            if 'ind-planeSegment' in obj.type :
                if not (obj.sort.name in direction[-1] or obj.sort.baseSort().name in direction[-1]):
                    return True
                if ('2D' in obj.type) == ('3D' in direction[1].type):
                    raise ValueError('There exists a mix of 2D and 3D objects in the shape')
                check = obj.classify(direction[1])
                if check.disjoint() or check.touches():
                    return True
                self.postCheck = False
                return False
            if 'ind-lineSegment' in obj.type:
                if not (obj.sort.name in direction[-1] or obj.sort.baseSort().name in direction[-1]):
                    return True
                direction[2].append(bSegment(obj.start, obj.end))
            if 'ind-point' in obj.type:
                if not (obj.sort.name in direction[-1] or obj.sort.baseSort().name in direction[-1]):
                    return True
                check = False
                for line in direction[1].boundaries[0].canonical:
                    if line.containsPoint(obj.value):
                        check = True
                if check:
                    return True
                if direction[1].boundaries[0].containsPointInside(obj.value):
                    self.postCheck = False
                    return False
        elif direction[0] == 'inside':
            if 'ind-planeSegment' in obj.type :
                if ('2D' in obj.type) == ('3D' in direction[1].type):
                    raise ValueError('There exists a mix of 2D and 3D objects in the shape')
                check = obj.classify(direction[1])
                if check.isContained():
                    return True
                self.postCheck = False
                return False
            if 'ind-lineSegment' in obj.type:
                direction[2].append(bSegment(obj.start, obj.end))
            if 'ind-point' in obj.type:
                check = False
                for line in direction[1].boundaries[0].canonical:
                    if line.containsPoint(obj.value):
                        check = True
                if check:
                    return True
                if direction[1].boundaries[0].containsPointInside(obj.value):
                    return True
                self.postCheck = False
                return False
        elif direction[0] == 'insideof':
            if 'ind-planeSegment' in obj.type :
                if not (obj.sort.name in direction[-1] or obj.sort.baseSort().name in direction[-1]):
                    return True
                if ('2D' in obj.type) == ('3D' in direction[1].type):
                    raise ValueError('There exists a mix of 2D and 3D objects in the shape')
                check = obj.classify(direction[1])
                if check.isContained():
                    return True
                self.postCheck = False
                return False
            if 'ind-lineSegment' in obj.type:
                if not (obj.sort.name in direction[-1] or obj.sort.baseSort().name in direction[-1]):
                    return True
                direction[2].append(bSegment(obj.start, obj.end))
            if 'ind-point' in obj.type:
                if not (obj.sort.name in direction[-1] or obj.sort.baseSort().name in direction[-1]):
                    return True
                check = False
                for line in direction[1].boundaries[0].canonical:
                    if line.containsPoint(obj.value):
                        check = True
                if check:
                    return True
                if direction[1].boundaries[0].containsPointInside(obj.value):
                    return True
                self.postCheck = False
                return False
        elif direction[0] == 'outside':
            if 'ind-planeSegment' in obj.type :
                if ('2D' in obj.type) == ('3D' in direction[1].type):
                    raise ValueError('There exists a mix of 2D and 3D objects in the shape')
                check = obj.classify(direction[1])
                if check.disjoint() or check.touches():
                    return True
                self.postCheck = False
                return False
            if 'ind-lineSegment' in obj.type:
                direction[2].append(bSegment(obj.start, obj.end))
            if 'ind-point' in obj.type:
                check = False
                for line in direction[1].boundaries[0].canonical:
                    if line.containsPoint(obj.value):
                        check = True
                if check:
                    return True
                if direction[1].boundaries[0].containsPointInside(obj.value):
                    self.postCheck = False
                    return False
        elif direction[0] == 'outsideof':
            if 'ind-planeSegment' in obj.type :
                if not (obj.sort.name in direction[-1] or obj.sort.baseSort().name in direction[-1]):
                    return True
                if ('2D' in obj.type) == ('3D' in direction[1].type):
                    raise ValueError('There exists a mix of 2D and 3D objects in the shape')
                check = obj.classify(direction[1])
                if check.disjoint() or check.touches():
                    return True
                self.postCheck = False
                return False
            if 'ind-lineSegment' in obj.type:
                if not (obj.sort.name in direction[-1] or obj.sort.baseSort().name in direction[-1]):
                    return True
                direction[2].append(bSegment(obj.start, obj.end))
            if 'ind-point' in obj.type:
                if not (obj.sort.name in direction[-1] or obj.sort.baseSort().name in direction[-1]):
                    return True
                check = False
                for line in direction[1].boundaries[0].canonical:
                    if line.containsPoint(obj.value):
                        check = True
                if check:
                    return True
                if direction[1].boundaries[0].containsPointInside(obj.value):
                    self.postCheck = False
                    return False
        elif direction[0] == 'shortest_line':
            if len(direction) == 3:
                if 'ind-lineSegment' not in obj.type:
                    return True
                if (obj.sort.name in direction[-1] or
                obj.sort.baseSort().name in direction[-1]) and obj.squareLength() < direction[1]:
                    self.postCheck = False
                    return False
            else:
                if 'ind-lineSegment' in obj.type and obj.squareLength() < direction[1]:
                    self.postCheck = False
                    return False
        elif direction[0] == 'longest_line':
            if len(direction) == 3:
                if 'ind-lineSegment' not in obj.type:
                    return True
                if (obj.sort.name in direction[-1] or
                obj.sort.baseSort().name in direction[-1]) and obj.squareLength() > direction[1]:
                    self.postCheck = False
                    return False
            else:
                if 'ind-lineSegment' in obj.type and obj.squareLength() > direction[1]:
                    self.postCheck = False
                    return False
        elif direction[0] == 'nolabel':
            if 'ind-lineSegment' in obj.type and obj.type == direction[1].type and obj.contains(direction[1]):
                if obj.hasAttribute():
                    obj.attribute.accept(self, direction = ['nolabel*'])
            if 'ind-planeSegment' in obj.type and obj.type == direction[1].type and obj.contains(direction[1]):
                if obj.hasAttribute():
                    obj.attribute.accept(self, direction = ['nolabel*'])
            if 'ind-point' in obj.type and obj.type == direction[1].type and obj == direction[1]:
                if obj.hasAttribute():
                    obj.attribute.accept(self, direction = ['nolabel*'])
        elif direction[0] == 'nolabel*':
            if obj.type == 'ind-label' or obj.type == 'ind-description':
                self.postCheck = False
                return False
        elif direction[0] == 'containsCheck':
            if 'ind-lineSegment' in obj.type:
                if len(direction) > 3 and (direction[3] is not None) and (obj == direction[3]):
                    return False
                elif obj.containsPoint(direction[1]):
                    direction[2][0] = True
            if 'ind-point' in obj.type:
                if obj.value == direction[1]:
                    direction[2][0] = True
        return True
        
    def visitLeave(self, obj, direction = 0):
        return
        
        
    @staticmethod
    def cw(A,B,C):
        """
        Check if ABC is a counter-clockwise angle
        """
        return (C.y-A.y) * (B.x-A.x) > (B.y-A.y) * (C.x-A.x)
        
    @staticmethod
    def evaluateDesc(boundList, param, special = False):
        """
        Evaluate all the descriptions in a predicate/directive 
        """
        newList = []
        check = True
        for i in boundList:
            if isinstance(i, description):
                try:
                    val = i.evaluateInd(param).getValue()[1:-1]
                except Exception:
                    check = False
                    break
                if special:
                    try:
                        newList.append(coordinate(val))
                    except:
                        newList.append(val)
                else:
                    newList.append(coordinate(val))
            elif isinstance(i,int) or isinstance(i,float):
                newList.append(coordinate(i))
            else:
                newList.append(i)
        return newList, check

        
    @staticmethod
    def evaluateBounds(bounds, tfSet):
        if isinstance(bounds[-1], list):
            raise NotImplementedError('......')
            # return graphMap.evaluateVariable(bounds[-1], tfSet)
        return bounds[-1]

        # OLD VERSION: MIN-MAX VALUES ALLOWED
        # #
        # if len(bounds) == 4:
        #     if isinstance(bounds[-1], list):
        #         b1 = graphMap.evaluateVariable(bounds[-1], tfSet)
        #     else:
        #         b1 = bounds[-1]
        #     if isinstance(bounds[-2], list):
        #         b0 = graphMap.evaluateVariable(bounds[-2], tfSet)
        #     else:
        #         b0 = bounds[-2]
        #     rand = randint(0, 100000)
        #     boundVal = (b1-b0)*rand/100000 + b0
        # elif len(bounds) == 3:
        #     if isinstance(bounds[-1], list):
        #         boundVal = graphMap.evaluateVariable(bounds[-1], tfSet)
        #     else:
        #         boundVal = bounds[-1]
        # return boundVal
        
    @staticmethod
    def solvePointOnLine(graphTransf, obj, bounds, target = None):        
        """
        Solve point_on_line directive
        """
        i = 0
        if target is not None:
            startP = obj.oStart
            endP = obj.oEnd
            if obj.containsPoint(target) and (not startP == target) and (not endP == target):
                if target.subtract(endP).length() < target.subtract(startP).length():
                    startP, endP = endP, startP
                nStart = graphTransf.apply(startP)
                nEnd = graphTransf.apply(endP)
                p = nStart.add(nEnd.subtract(nStart).scale(bounds[-1]))
                graphTransf.add(target, p)
            return
        while i < len(graphTransf.unresolved):
            u = graphTransf.unresolved[i]
            if obj.containsPoint(u) and (not obj.start == u) and (not obj.end == u):
                startP = obj.oStart
                endP = obj.oEnd
                if u.subtract(endP).length() < u.subtract(startP).length():
                    startP, endP = endP, startP
                # if adaptive:
                #     for j in graphTransf.vects:
                #         if obj.containsPoint(j[0]):
                #             if j[0] < u:
                #                 if j[0] > startP:
                #                     startP = j[0]
                #             else:
                #                 if j[0] < endP:
                #                     endP = j[0]
                graphTransf.unresolved.pop(i)
                nStart = graphTransf.apply(startP)
                nEnd = graphTransf.apply(endP)
                p = nStart.add(nEnd.subtract(nStart).scale(bounds[-1]))
                # if singleRatio:
                #     p = nStart.add(nEnd.subtract(nStart).scale(bounds[-1]))
                # else:
                #     rand = randint(0, 100000)
                #     ratio = (bounds[-1]-bounds[-2])*rand/100000 + bounds[-2]
                #     p = nStart.add(nEnd.subtract(nStart).scale(ratio))
                graphTransf.add(u, p)
            else:
                i += 1
