# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import range
from builtins import super
from builtins import str
from builtins import object

import itertools
from ..form.coForm import coForm
from ..ind.line import line2D, line3D
from ..ind.lineSegmentP import lineSegmentP3D
from ..ind.plane import plane3D
from ..sort.sort import sort as sortClass

from enum import Enum

CONNECTED = True
DISCONNECTED = False

class graphNode(object):
    """
    2D GraphNode: a graph containing all geometry elements inside a shape (can be a rule's lhs or a shape)
    structure:  

    **Object Parameters:**

    * **lines** (:class:`~graphNode.node`) -- list of nodes of the graph, containing the node element (line) and all the line-segments under that node.
    * **points** (:class:`~graphNode.auxiliaryObj`) -- points in the shape.
    * **planes** (:class:`~graphNode.auxiliaryObj`) -- plane-segments in the shape.
    * **curves** (:class:`~graphNode.auxiliaryObj`) -- curves in the shape.

    """

    class auxiliaryObj(object):
        """
        AuxiliaryObj: structure for other type of geometry that are not used in the main matching process in the graph
        """
        def __init__(self, objVal):
            self.value = objVal
            self.oType = 'f'
            self.misc = []
            self.assocObj = None
        
        def __str__(self):
            return self.oType+str(self.value)
            
    class intersection(object):
        """
        line intersection 

        **Object Parameters:**

        * **value** (:class:`~sortal.struct.vector.vector2D`) -- vector indicating the coordinates of the intersection.
        * **nodes** (list of :class:`~graphNode.node`) -- the nodes that contain this intersection.
        * **endP** (boolean) -- check if this intersection is the end point of a line.
        * **iType** (string) -- type of intersection:
            * 'o': intersection outside of all line-segments on a line; 
            * 'i': intersection inside one line-segment on a line;
            * 'b': intersection that is the end point of at least one line-segment on a line
        """

        type = 'graphNode-intersection'
        
        def __init__(self, coord):
            self.value = coord
            self.nodes = []
            self.endP = False
            self.iType = 'o'
            
        def __str__(self):
            return str(self.value)
            
        def __repr__(self):
            return str(self.value)
            
        def __eq__(self, other):
            if not isinstance(other, graphNode.intersection):
                return False
            return self.value == other.value
            
        def __ne__(self, other):
            if not isinstance(other, graphNode.intersection):
                return False
            return self.value != other.value
            
        def __lt__(self, other):
            if not isinstance(other, graphNode.intersection):
                return False
            return self.value < other.value
            
        def __gt__(self, other):
            if not isinstance(other, graphNode.intersection):
                return False
            return self.value > other.value
            
        def __le__(self, other):
            if not isinstance(other, graphNode.intersection):
                return False
            return self.value <= other.value
            
        def __ge__(self, other):
            if not isinstance(other, graphNode.intersection):
                return False
            return self.value >= other.value
            
        def addElm(self, elm):
            """
            add a node into this intersection nodes list
            """
            if elm not in self.nodes:
                self.nodes.append(elm)
            
    class node(object):
        """
        node object for line

        **Object Parameters:**

        * **nodeElement** (:class:`~sortal.ind.line.line2D`) -- the line that define the node.
        * **forms** (list of :class:`~sortal.ind.lineSegmentP.lineSegmentP2D`) -- the list of line-segments on the line
        * **intersections** (list of :class:`~graphNode.intersection`) -- list of line intersection
        * **intersectionMatrix** -- <<Unused>>
        * **endP** (boolean) -- True -> the point is an end point of a line; False -> the point is not an end point of a line
        * **linePoints** (list of :class:`~sortal.struct.vector.vector2D`) -- list of end points that are not intersections
        
        """
        class line_casus(Enum):
            ###### NORMAL CASES ######
            CASE_1PARALLEL_1PT  = '1 parallel line, 1 contained point'
            CASE_1PERP_1PT      = '1 perpendicular line, 1 contained point'

            ###### DIRECTIVE CASES ######
            CASE_ANGLE                          = 'angle directive, 1 contained point'
            CASE_POINT_LINE_DISTANCE            = 'point-line distance directive, 1 contained point'
            CASE_LINE_LINE_DISTANCE             = 'line-line distance directive no direction'
            CASE_LINE_LINE_DISTANCE_DIRECTION   = 'line-line distance directive with direction'
            CASE_DIRECTION                      = 'direction directive, 1 contained point'

            ###### OTHER CASES ######
            CASE_PROXY       = 'proxy'      # already been solved by proxy through other line nodes
            CASE_LHS         = 'lhs'        # node already in the lhs, just that one or more line(s) are not not accounted for
            CASE_UNSOLVABLE  = 'unsolvable' # unsolvable if cannot satisfy the above criteria

        class point_casus(Enum):
            ###### POINT CASES ######
            CASE_LINE_POINT_DISTANCE              = 'line-point distance directive no direction'
            CASE_POINT_POINT_DISTANCE             = 'point-point distance directive no direction'
            CASE_POINT_POINT_DISTANCE_DIRECTION   = 'point-point distance directive with direction'

        def __init__(self, nodeElement, ind):
            self.nodeElement = nodeElement
            self.forms = []
            if ind is not None:
                self.forms.append(ind)
            self.intersections = []
            self.intersectionMatrix = None
            self.endP = False
            self.linePoints = []
            self.containedPoints = []

            self.rhsInfo = None
            
        def __repr__(self):
            return str(self)
            
        def __str__(self):
            return str(self.nodeElement)+str(self.forms)
            
        def add(self, line):
            """
            add a line segment into the node
            """
            for i in range(len(self.forms)):
                if self.forms[i].compareInterval(line) == 1:
                    self.forms.insert(i, line)
                    return
            self.forms.append(line)
            return
            
        def addIntersection(self, intersection):
            """
            add an intersection to the node's intersection list in ascending order
            """
            intersection.addElm(self)
            for i in range(len(self.intersections)):
                if self.intersections[i] >= intersection:
                    if self.intersections[i] == intersection:
                        return
                    self.intersections.insert(i,intersection)
                    return
            self.intersections.append(intersection)
            
        def addEndPoints(self):
            """
            add an end-point into the intersection list
            """
            for indiv in self.forms:
                
                checkStart = False
                checkEnd = False
                for intsn in self.intersections:
                    if indiv.start == intsn.value:
                        checkStart = True
                    if indiv.end == intsn.value:
                        checkEnd = True
                if not checkStart:
                    start = graphNode.intersection(indiv.start)
                    start.endP = True
                    self.endP = True
                    self.addIntersection(start)
                if not checkEnd:
                    end = graphNode.intersection(indiv.end)
                    end.endP = True
                    self.endP = True
                    self.addIntersection(end)
                    
        def classifyIntersections(self):
            self.intersectionMatrix = [[None for i in range(len(self.intersections))] 
                                             for j in range(len(self.intersections))]
            for i in range(len(self.intersections)):
                for j in range(i+1,len(self.intersections)):
                    check = False
                    for line in self.forms:
                        if (line.containsPoint(self.intersections[i].value)
                        and line.containsPoint(self.intersections[j].value)):
                            check = True
                            self.intersectionMatrix[i][j] = CONNECTED
                            self.intersectionMatrix[j][i] = CONNECTED
                            break
                    if not check:
                        self.intersectionMatrix[i][j] = DISCONNECTED
                        self.intersectionMatrix[j][i] = DISCONNECTED
        
        def organizeForm(self):
            """
            organize the line-segments in the node according to their attribute for matching process
            """
            if self.forms[0].sort.type != 'sort-attribute':
                return
            attributes = []
            baseSorts = []

            # form combinations of attributes and add them to a list
            for i in self.forms:
                if i.sort.baseSort() not in baseSorts:
                    baseSorts.append(i.sort)
                graphNode.classifyAttribute(attributes, i.attribute, i.sort.baseSort())
            # add an empty attribute set for each line-segment sort
            for i in baseSorts:
                emptyAttribute = i.weight.newForm()
                attSet = [emptyAttribute, i.baseSort()]
                if attSet not in attributes:
                    attributes.append(attSet)
                    
            # create maximal line-segments for each attribute set
            currentForm = self.forms
            self.forms = []
            attributes = sorted(attributes)
            for i in reversed(attributes):
                form = []
                for line in currentForm:
                    if i[1] == line.sort.baseSort() and (i[0] == line.attribute or i[0].partOf(line.attribute)):
                        if len(form) == 0 or form[-1].end != line.start:
                            form.append(line)
                        else:
                            nLine = form[-1].combine(line)
                            nLine.addAttribute(i[0])
                            form[-1] = nLine
                self.forms += form

            # remove duplicates
            i = 0
            while i < len(self.forms):
                j = i + 1
                while j < len(self.forms):
                    if self.forms[j].sort == self.forms[i].sort and self.forms[j].equals(self.forms[i]):
                        self.forms.pop(j)
                    else:
                        j += 1
                i += 1
                
        
                
        def containsPoint(self, point):
            """
            check if any of the line-segments contains the point in them
            """
            for i in self.forms:
                if i.containsPoint(point):
                    return True
            return False

        def transfContainsPoint(self, transformSet, point):
            """
            check if any of the transformed line-segments contains the point in them
            """
            for i in self.forms:
                j = i.edit(transformSet)
                if j.containsPoint(point):
                    return True
            return False

            
        def __eq__(self, other):
            if not isinstance(other, graphNode.node):
                return False
            return self.nodeElement.equals(other.nodeElement)
    
    @staticmethod
    def classifyAttribute(attributeList, attribute, baseSort):
        attSet = [attribute, baseSort]
        if attSet not in attributeList:
            for j in attributeList:
                if baseSort != j[1]:
                    continue
                productAtt = attribute.duplicate()
                productAtt.product(j[0])
                attSetSub = [productAtt, baseSort]
                if attSetSub not in attributeList:
                    attributeList.append(attSetSub)
                    graphNode.classifyAttribute(attributeList, productAtt, baseSort)
            attributeList.append(attSet)
            
    def __init__(self):
        self.lines = []
        self.points = []
        self.planes = []
        self.curves = []

        self.rhsLines = []
        self.rhsPoints = []

    def __iter__(self):
        return iter(self.lines)
        
    def __len__(self):
        return len(self.lines)
        
    def __repr__(self):
        return str(self)
            
    def __str__(self):
        return str(self.lines)


    def clear(self):
        """
        clear all the line nodes, points, planes and curves in the graph
        """
        self.lines = []
        self.points = []
        self.planes = []
        self.curves = []
            
    def addInstance(self, codesc, ind):
        """
        add a new line-segment to the graph

        * if an existing node contains the line-segment, add the line-segment to the node
        * if there is no existing node containing the line-segment, create a new node and add it to the graph
        """
        # for j in range(len(self.lines)):
        #     if self.lines[j].nodeElement.equals(codesc):
        #         self.lines[j].add(ind)
        #         return j
        # newNode = graphNode.node(codesc, ind)
        # self.lines.append(newNode)
        # return True
        for j in range(len(self.lines)):
            #if self.lines[j].nodeElement == codesc:
            if self.lines[j].nodeElement.equals(codesc):
                if ind is not None:
                    self.lines[j].add(ind)
                return
        self.lines.append(graphNode.node(codesc, ind))
        return
        

    # ============================================================================================
    #                                         LHS ANALYSIS
    # ============================================================================================

    def LHS_addPoint(self, p):
        """
        add a new point to the graph
        """
        newP = self.auxiliaryObj(p.value)
        self.points.append(newP)
        
    def LHS_addPlane(self, p):
        """
        add a new plane to the graph
        """
        newP = self.auxiliaryObj(p)
        self.planes.append(newP)
        
    def LHS_addCurve(self, p):
        """
        add a new curve to the graph
        """
        newP = self.auxiliaryObj(p)
        self.curves.append(newP)

    def generateLines(self, ruleAnalysis = False):
        dummyLineSort = sortClass.getSort('__dummyLineSegP2D__')
        # for each of the point
        for i in range(len(self.points)):
            pti = self.points[i].value
            containingLines = []
            for ln in self.lines:
                if ln.nodeElement.containsPoint(pti):
                    containingLines.append(ln.nodeElement)
            if len(containingLines) > 0:
                for j in range(len(self.points)):
                    if i == j : continue
                    ptj = self.points[j].value
                    for ln in containingLines:
                        if ln.containsPoint(ptj): continue
                        if pti == ln.projection(ptj):
                            newLineNode = line2D(dummyLineSort, pti, ptj)
                            self.addInstance(newLineNode, None)

            if i >= len(self.points) - 2: continue
            for j in range(i + 1, len(self.points) - 1):
                ptj = self.points[j].value
                newLine1 = line2D(dummyLineSort, pti, ptj)
                for k in range(j + 1, len(self.points)):
                    ptk = self.points[k].value
                    if pti.colinear(ptj, ptk): continue
                    newLine2 = line2D(dummyLineSort, pti, ptk)
                    newLine3 = line2D(dummyLineSort, ptk, ptj)
                    if (newLine1.perpendicular(newLine2)):
                        self.addInstance(newLine1, None)
                        self.addInstance(newLine2, None)
                    if (newLine1.perpendicular(newLine3)):
                        self.addInstance(newLine1, None)
                        self.addInstance(newLine3, None)
                    if (newLine2.perpendicular(newLine3)):
                        self.addInstance(newLine2, None)
                        self.addInstance(newLine3, None)

    def classifyObjs(self, ruleAnalysis = False):
        """
        Classify each of the extra objects (not line) in the shape

        * for point:  
            * 'f': free points, do not lie on any line in the shape
            * 'l': line points, contained by exactly one line in the shape
            * 'i': intersection points, contained by at least 2 lines in the shape
        * for curves: assign the curves to the plane they belong to
        """

        # for each of the points, check how many lines they are contained by:
        # 0 lines -> free point - 'f'
        # 1 line -> line point - 'l'
        # 2+ lines -> intersection point - 'i's
        for p in self.points:
            check = 0
            prevNode = None
            for node in self.lines:
                if (prevNode is None or prevNode.unitDir != node.nodeElement.unitDir 
                or prevNode.root != node.nodeElement.root) and node.nodeElement.containsPoint(p.value):
                    check+=1
                    prevNode = node.nodeElement
                    node.containedPoints.append(p)
                    if not ruleAnalysis:
                        node.linePoints.append(p)
                    elif check == 1:
                        p.assocObj = node
            if check == 1:
                p.oType = 'l'
                if ruleAnalysis:
                    p.assocObj.linePoints.append(p)
            elif check > 1:
                p.oType = 'i'
                
                
        # only for lhs
        if ruleAnalysis:
            index = 0
            # for each of the plane-segment
            while index < len(self.planes):
                plane = self.planes[index]
                plane.oType = 'i'
                for bound in plane.value.boundaries:
                    for p in bound.polygonal:
                        check = 0
                        for node in self.lines:
                            if node.nodeElement.containsPoint(p):
                                check+=1
                                if check == 1:
                                    plane.assocObj = node
                        if check == 0:
                            plane.oType = 'f'
                            break
                        if check == 1:
                            plane.oType = 'l'
                    if plane.oType == 'f':
                        break
                if plane.oType == 'i':
                    self.planes.pop(index)
                    index -= 1
                index += 1
            return
        splitPlanes = []
        for plane in self.planes:
            if not plane.value.hasAttribute():
                splitPlanes.append(plane.value)
                continue
            splitAttributes = plane.value.attribute.splitElements()
            splitPlanes.append(plane.value.replicateInd())
            for i in range(1,len(splitAttributes)+1):
                for comp in itertools.combinations(splitAttributes,i):
                    newPlane = plane.value.replicateInd()
                    for att in comp:
                        newPlane.addAttribute(att)
                    splitPlanes.append(newPlane)
                    
        combinedPlanes = []
        while len(splitPlanes)!=0:
            plane1 = splitPlanes.pop(0)
            
            index = 0
            while index < len(combinedPlanes):
                plane = combinedPlanes[index]
                if (not plane1.sort == plane.sort) or (not plane1.attribute == plane.attribute):
                    index += 1
                    continue
                classes = plane1.classify(plane)
                if not classes.disjoint():
                    resultPlane = coForm(plane.sort)
                    plane.construct(classes.sumSet(), resultPlane)
                    plane1 = resultPlane.elements[0]
                    if plane.hasAttribute():
                        plane1.addAttribute(plane.attribute)
                    combinedPlanes.remove(plane)
                    index -= 1
                index += 1
            
            index = 0
            while index < len(splitPlanes):
                plane = splitPlanes[index]
                
                if (not plane1.sort == plane.sort) or (not plane1.attribute == plane.attribute):
                    index += 1
                    continue
                classes = plane1.classify(plane)
                
                if not classes.disjoint():
                    resultPlane = coForm(plane.sort)
                    plane.construct(classes.sumSet(), resultPlane)
                    plane1 = resultPlane.elements[0]
                    if plane.hasAttribute():
                        plane1.addAttribute(plane.attribute)
                    splitPlanes.remove(plane)
                    index -= 1
                index += 1
            combinedPlanes.append(plane1)
        self.planes = []
        for i in combinedPlanes:
            self.LHS_addPlane(i)

    # ============================================================================================
    #                                         RHS ANALYSIS
    # ============================================================================================

            
    def RHS_addLine(self, codesc, ind = None):
        """
        add a rhs line segment and its parametric node into the line-node list of the graph
        """

        # check if there's already the same line node in the rhs
        for j in range(len(self.rhsLines)):
            if self.rhsLines[j].nodeElement.equals(codesc):
                if ind is not None:
                    self.rhsLines[j].add(ind)
                return self.rhsLines[j]

        # check if the line is under an existing lhs node
        lhsNode = None
        for lhsN in self.lines:
            if lhsN.nodeElement.equals(codesc):
                # if it's in the lhs and there's no individual, return
                if ind is None:
                    return None
                lhsNode = lhsN
                break

        # create a new node for when the line node is not already in the rhs
        # a new node should still be created for the rhs even if it's in the lhs
        # to carry all the new rhs line segments in
        node = graphNode3D.lineNode(codesc, ind)
        node.rhsInfo = {
            "case": None,
            "caseNodes": [],
            "caseTemp": False,
            "extraInts": set(),
            "parallel": [],
            "perpendicular": [],
            "intersect": [],
            "directives": []
        }
        if lhsNode is not None:
            node.rhsInfo['case'] = graphNode.node.line_casus.CASE_PROXY
            node.rhsInfo['caseNodes'] = [lhsNode]

        self.rhsLines.append(node)
        return node
        
    def RHS_addPoint(self, p):
        """
        add a rhs point to the graph
        """
        pass

        


class graphNode3D(object):
    """
    3D GraphNode: a graph containing all geometry elements inside a shape (can be a rule's lhs or a shape)
    structure:  

    **Object Parameters:**

    * **lines** (:class:`~graphNode.lineNode`) -- list of line-nodes of the graph, containing the node element (line) and all the line-segments under that node.
    * **planes** (:class:`~graphNode.planeNode`) -- plane-segments in the shape.
    * **points** (:class:`~graphNode.auxiliaryObj`) -- points in the shape.
    * **curves** (:class:`~graphNode.auxiliaryObj`) -- curves in the shape.

    """

    class auxiliaryObj(object):
        """
        AuxiliaryObj: structure for other type of geometry that are not used in the main matching process in the graph
        """
        def __init__(self, coord):
            self.value = coord
            self.oType = 'f'
            self.assocObj = None
        
        def __str__(self):
            return self.oType+str(self.value)
            
            
    class lineIntersection(object):
        """
        line intersection 

        **Object Parameters:**

        * **value** (:class:`~sortal.struct.vector.vector3D`) -- vector indicating the coordinates of the intersection.
        * **assocObjs** (list of :class:`~graphNode.node`) -- the list of nodes that contain this intersection.
        * **endP** (boolean) -- check if this intersection is the end point of a line.
        * **iType** (string) -- type of intersection:
            * 'o': intersection outside of all line-segments on a line; 
            * 'i': intersection inside one line-segment on a line;
            * 'b': intersection that is the end point of at least one line-segment on a line
        """

        type = ''
        def __init__(self, coord):
            self.value = coord
            self.iType = 'o'
            self.assocObjs = []
            self.endP = False
            
        def __str__(self):
            return str(self.value)
            
        def __repr__(self):
            return str(self.value)
            
        def __eq__(self, other):
            if not isinstance(other, graphNode3D.lineIntersection):
                return False
            return self.value == other.value
            
        def __ne__(self, other):
            if not isinstance(other, graphNode3D.lineIntersection):
                return False
            return self.value != other.value
            
        def __lt__(self, other):
            if not isinstance(other, graphNode3D.lineIntersection):
                return False
            return self.value < other.value
            
        def __gt__(self, other):
            if not isinstance(other, graphNode3D.lineIntersection):
                return False
            return self.value > other.value
            
        def __le__(self, other):
            if not isinstance(other, graphNode3D.lineIntersection):
                return False
            return self.value <= other.value
            
        def __ge__(self, other):
            if not isinstance(other, graphNode3D.lineIntersection):
                return False
            return self.value >= other.value
            
        def addassocObj(self, line):
            """
            add a node into this intersection nodes list
            """
            if line not in self.assocObjs:
                self.assocObjs.append(line)
            
    class pointNode(object):
        """
        node object for point, only 

        **Object Parameters:**

        * **nodeElement** (:class:`~sortal.ind.line.line3D`) -- the point that define the node.
        
        """
        class point_casus(Enum):
            ###### POINT CASES ######
            CASE_PLANE_POINT_DISTANCE             = 'plane-point distance directive no direction'
            CASE_LINE_POINT_DISTANCE_SKEW         = 'line-point distance directive no direction, the base line is skew to the line containing the new point'
            CASE_LINE_POINT_DISTANCE_INTERSECT    = 'line-point distance directive no direction, the base line intersects the line containing the new point'
            CASE_POINT_POINT_DISTANCE             = 'point-point distance directive no direction'
            CASE_POINT_POINT_DISTANCE_DIRECTION   = 'point-point distance directive with direction'


        def __init__(self, nodeElement):
            self.nodeElement = nodeElement

            # rhsInfo structure:
            #   {
            #       "case": string indicating one of the solving cases based on the rhs
            #       "caseNodes": list of <planeNode> and <lineNode> and vector3D needed for solving the case (for solving purpose),
            #   }
            self.rhsInfo = None

        def __repr__(self):
            return str(self.nodeElement)
            
        def __str__(self):
            return str(self.nodeElement)

        def __eq__(self, other):
            if not isinstance(other, graphNode3D.pointNode):
                return False
            return self.nodeElement == other.nodeElement
            
    class lineNode(object):
        """
        node object for line

        **Object Parameters:**

        * **nodeElement** (:class:`~sortal.ind.line.line3D`) -- the line that define the node.
        * **forms** (list of :class:`~sortal.ind.lineSegmentP.lineSegmentP3D`) -- the list of line-segments on the line
        * **intersections** (list of :class:`~graphNode.intersection`) -- list of line intersection
        * **intersectionMatrix** -- <<Unused>>
        * **endP** (boolean) -- True -> at least one end points are not line intersection; False -> all end points are line intersections
        * **linePoints** (list of :class:`~sortal.struct.vector.vector3D`) -- list of end points that are not intersections
        
        """
        class line_casus(Enum):
            ###### NORMAL CASES ######
            CASE_1PL_1PARALLEL_1INT = 'solvable plane, 1 parallel line, 1 intersecting line'
            CASE_1PL_1PERP_1INT     = 'solvable plane, 1 perpendicular line, 1 intersecting line' # co-planar perpendicular line 
            CASE_1PL_1PERP_1PT      = 'solvable plane, 1 perpendicular line, 1 contained point' # co-planar perpendicular line 
            CASE_1PL_1PERP          = 'solvable plane, 1 perpendicular line' # not perpendicular to the plane, not co-planar

            CASE_1PARALLEL_1PT     = 'unknown plane, 1 parallel line, 1 contained point'
            CASE_2PERP             = 'unknown plane, 2 perpendicular lines' # the two lines not parallel to each other
            CASE_2PARALLEL         = 'unknown plane, 2 parallel lines' # direction from these two lines root to this line must be perpendicular

            ###### DIRECTIVE CASES ######
            CASE_ANGLE                          = 'angle directive, solvable plane, 1 contained point'
            CASE_POINT_LINE_DISTANCE            = 'point-line distance directive, solvable plane, 1 contained point'
            CASE_LINE_LINE_DISTANCE             = 'line-line distance directive no direction, solvable plane'
            CASE_LINE_LINE_DISTANCE_DIRECTION   = 'line-line distance directive with direction'
            CASE_DIRECTION                      = 'direction directive, 1 contained point'

            ###### OTHER CASES ######
            CASE_PROXY       = 'proxy'      # already been solved by proxy through other line nodes
            CASE_LHS         = 'lhs'        # node already in the lhs, just that one or more line(s) are not not accounted for
            CASE_UNSOLVABLE  = 'unsolvable' # unsolvable if cannot satisfy the above criteria


        def __init__(self, nodeElement, ind = None):
            self.nodeElement = nodeElement
            self.forms = []
            if ind is not None:
                self.forms.append(ind)
            self.intersections = []
            self.intersectionMatrix = None
            self.endP = False
            self.linePoints = []
            self.containedPoints = []

            # rhsInfo structure:
            #   {
            #       "case": string indicating one of the solving cases based on the rhs
            #       "caseNodes": list of <planeNode> and <lineNode> and vector3D needed for solving the case (for solving purpose),
            #       "extraInts": <lineIntersection>[] (line intersections + end-points not in the lhs, for solving and checking purposes),
            #       "parallel": <lineNode>[] (for checking purpose)
            #       "perpendicular": <lineNode>[] (for checking purpose)
            #       "intersect": <lineNode>[] (intersecting nodes but not parallel; for checking purpose)
            #   }
            self.rhsInfo = None

            
        def __repr__(self):
            return str(self)
            
        def __str__(self):
            return str(self.nodeElement)+str(self.forms)
            
        def add(self, line):
            for i in range(len(self.forms)):
                if self.forms[i].compareInterval(line) == 1:
                    self.forms.insert(i, line)
                    return
            self.forms.append(line)
            return
            
        def addIntersection(self, intersection):
            """
            add an intersection to the node's intersection list in ascending order
            """
            intersection.addassocObj(self)
            for i in range(len(self.intersections)):
                if self.intersections[i] >= intersection:
                    if self.intersections[i] == intersection:
                        return
                    self.intersections.insert(i,intersection)
                    return
            self.intersections.append(intersection)
            
        def addEndPoints(self):
            """
            add an end-point into the intersection list
            """
            for indiv in self.forms:
                checkStart = False
                checkEnd = False
                for intsn in self.intersections:
                    if indiv.start == intsn.value:
                        checkStart = True
                    if indiv.end == intsn.value:
                        checkEnd = True
                if not checkStart:
                    start = graphNode3D.lineIntersection(indiv.start)
                    start.endP = True
                    self.endP = True
                    self.addIntersection(start)
                if not checkEnd:
                    end = graphNode3D.lineIntersection(indiv.end)
                    end.endP = True
                    self.endP = True
                    self.addIntersection(end)
                                        
                    
        
        def classifyIntersections(self):
            self.intersectionMatrix = [[None for i in range(len(self.intersections))] 
                                             for j in range(len(self.intersections))]
            for i in range(len(self.intersections)):
                for j in range(i+1,len(self.intersections)):
                    check = False
                    for line in self.forms:
                        if (line.containsPoint(self.intersections[i].value)
                        and line.containsPoint(self.intersections[j].value)):
                            check = True
                            self.intersectionMatrix[i][j] = CONNECTED
                            self.intersectionMatrix[j][i] = CONNECTED
                            break
                    if not check:
                        self.intersectionMatrix[i][j] = DISCONNECTED
                        self.intersectionMatrix[j][i] = DISCONNECTED
        
        
        def organizeForm(self):
            """
            organize the line-segments in the node according to their attribute for matching process
            """
            if self.forms[0].sort.type != 'sort-attribute':
                return
            attributes = []
            baseSorts = []
            for i in self.forms:
                if i.sort not in baseSorts:
                    baseSorts.append(i.sort)
                graphNode.classifyAttribute(attributes, i.attribute, i.sort.baseSort())
            for i in baseSorts:
                emptyAttribute = i.weight.newForm()
                attSet = [emptyAttribute, i.baseSort()]
                if attSet not in attributes:
                    attributes.append(attSet)
                
            currentForm = self.forms
            self.forms = []
            attributes = sorted(attributes)
            for i in reversed(attributes):
                form = []
                for line in currentForm:
                    if i[1] == line.sort.baseSort() and (i[0] == line.attribute or i[0].partOf(line.attribute)):
                        if len(form) == 0 or form[-1].end != line.start:
                            form.append(line)
                        else:
                            nLine = form[-1].combine(line)
                            nLine.addAttribute(i[0])
                            form[-1] = nLine
                self.forms += form
            i = 0
            while i < len(self.forms):
                j = i + 1
                while j < len(self.forms):
                    if self.forms[j].sort == self.forms[i].sort and self.forms[j].equals(self.forms[i]):
                        self.forms.pop(j)
                    else:
                        j += 1
                i += 1
            
        def __eq__(self, other):
            if not isinstance(other, graphNode3D.lineNode):
                return False
            return self.nodeElement == other.nodeElement
            
    class planeNode(object):
        """
        node object for line

        **Object Parameters:**

        * **nodeElement** (:class:`~sortal.ind.plane.plane3D`) -- the plane that define the node.
        * **forms** (list of :class:`~sortal.ind.planeSegmentP.planeSegmentP3D`) -- the list of plane-segments on the plane
        * **points** (list of points) -- the list of all the points on the plane
        * **curves** (list of curves) -- the list of all the curves on the plane
        
        """

        class plane_casus(Enum):
            ###### OTHER PLANE CASES ######
            CASE_1PARALLEL_1PT          = '1 parallel plane, 1 contained point'
            CASE_2PERP_1PT              = '2 perpendicular plane, 1 contained point'

            ###### LINE CASES ######
            CASE_1PERPLN_1PT            = '1 perpendicular line, 1 contained point'
            CASE_2PARALLELLN_1PT        = '2 parallel lines, 1 contained point' # these 2 lines must intersect one another

            ###### DIRECTIVE CASES ######
            CASE_PLANE_PLANE_DISTANCE   = 'plane-plane distance directive' # only need the directive, no other requirements

            ###### OTHER CASES ######
            CASE_DEFAULT     = 'default'    # if there's no planeNode in the lhs and there's at least one in the rhs, 
                                            #   one of the rhs plane would be assigned this case: 
                                            #   the plane will be solved with matched normal being the same as the normal in the rule rhs
            CASE_PROXY       = 'proxy'      # already been solved by proxy through other plane nodes
            CASE_LHS         = 'lhs'        # already in the lhs, just that one or more plane-segment(s) are not accounted for
            CASE_UNSOLVABLE  = 'unsolvable' # unsolvable if cannot satisfy the above criterianode


        def __init__(self, nodeElement, ind = None):
            self.nodeElement = nodeElement
            self.forms = []
            if ind is not None:
                self.forms.append(ind)
            self.lines = []
            self.points = []
            self.curves = []

            # rhsInfo structure:
            #   {
            #       "case": string indicating one of the above cases
            #       "caseNodes": list of <planeNode> and <lineNode> and vector3D representing the case (for solving purpose)
            #       "required": boolean indicating if the plane is required to be solved (has a rhs plane segment or is required to solve a rhs line) 
            #   }
            self.rhsInfo = None

            
        def __repr__(self):
            return str(self)
            
        def __str__(self):
            return str(self.nodeElement)+str(self.forms)
            
        def add(self, plane):
            """
            add a plane segment into the plane list
            """
            self.forms.append(plane)
        
        def addLines(self, lines):
            """
            add a list of line nodes into the plane
            """
            for assocObj in lines:
                check = False
                for k in self.lines:
                    if k == assocObj:
                        check = True
                        break
                if not check:
                    self.lines.append(assocObj)
            
        def addLine(self, line):
            """
            add one line node into the plane's line node list
            """
            for k in self.lines:
                if k == line:
                    return
            self.lines.append(line)
            
        def addCurve(self, curve):
            """
            add a curve into the plane's curve list
            """
            self.curves.append(curve)
            
        def addIntersection(self, intersection):
            pass
                                
        def classifyIntersections(self):
            pass
        
        def organizeForm(self):
            pass
            
        def __eq__(self, other):
            if not isinstance(other, graphNode3D.planeNode):
                return False
            return self.nodeElement == other.nodeElement
                
    
    def __init__(self):
        self.lines = []
        self.points = []
        self.planes = []
        self.curves = []

        self.rhsLines = []
        self.rhsPoints = []
        self.rhsPlanes = []

        self.planeCase = False

    def __iter__(self):
        return iter(self.planes)
        
    def __len__(self):
        return len(self.planes)
        
    def __repr__(self):
        return str(self)
            
    def __str__(self):
        return str(self.planes)+str(self.lines)

    def clear(self):
        """
        clear all the saved geometry in the graph
        """
        self.lines = []
        self.points = []
        self.planes = []
        self.curves = []

        self.rhsLines = []
        self.rhsPoints = []
        self.rhsPlanes = []

    # ============================================================================================
    #                                         LHS ANALYSIS
    # ============================================================================================

        
    def LHS_addPlane(self, codesc, ind = None):
        """
        add a lhs plane segment and its parametric node into the plane-node list of the graph
        """
        for j in range(len(self.planes)):
            if self.planes[j].nodeElement == codesc:
                if ind is not None:
                    self.planes[j].add(ind)
                return
        newNode = graphNode3D.planeNode(codesc, ind)
        self.planes.append(newNode)
        return
            
    def LHS_addLine(self, codesc, ind = None):
        """
        add a lhs line segment and its parametric node into the line-node list of the graph
        """
        for j in range(len(self.lines)):
            #if self.lines[j].nodeElement == codesc:
            if self.lines[j].nodeElement.equals(codesc):
                if ind is not None:
                    self.lines[j].add(ind)
                return
        self.lines.append(graphNode3D.lineNode(codesc, ind))
        return
        
    def LHS_addPoint(self, p):
        """
        add a lhs point to the graph
        """
        newP = self.auxiliaryObj(p.value)
        self.points.append(newP)

    def LHS_addCurve(self, p):
        """
        add a new curve to the graph
        """
        newP = self.auxiliaryObj(p)
        if 'ind-bezierCurveP3D' == p.type:
            p1 = p.start
            p2 = p.controlPoint
            p3 = p.end
            newP.misc = [p1,p2,p3]
        elif 'ind-circleP3D' == p.type:
            p1,p2,p3 = p.getPoints()
            newP.misc = [p1,p2,p3]
        elif 'ind-ellipseP3D' == p.type:
            p1 = p.foci[0]
            p2 = p.foci[1]
            p3, p4 = p.getQuadPoints()
            newP.misc = [p1,p2,p3,p4]
        
        self.curves.append(newP)
        

    def generateLinesPlanes(self, ruleAnalysis = False):
        """
        Generating planes for matching matrices with existing lines. 
        Then generate lines for matching matrices with existing planes 
        """
        dummyLineSort = sortClass.getSort('__dummyLineSegP3D__')
        # for each of the point
        for i in range(len(self.points)):
            pti = self.points[i].value
            containingLines = []
            for ln in self.lines:
                if ln.nodeElement.containsPoint(pti):
                    containingLines.append(ln.nodeElement)
            if len(containingLines) > 0:
                for j in range(len(self.points)):
                    if i == j : continue
                    ptj = self.points[j].value
                    for ln in containingLines:
                        if ln.containsPoint(ptj): continue
                        if pti == ln.projection(ptj):
                            newLineNode = line3D(ln.sort, pti, ptj)
                            self.LHS_addLine(newLineNode, None)

            if i >= len(self.points) - 2: continue
            for j in range(i + 1, len(self.points) - 1):
                ptj = self.points[j].value
                newLine1 = line3D(dummyLineSort, pti, ptj)
                for k in range(j + 1, len(self.points)):
                    ptk = self.points[k].value
                    if pti.colinear(ptj, ptk): continue
                    newLine2 = line3D(dummyLineSort, pti, ptk)
                    newLine3 = line3D(dummyLineSort, ptk, ptj)
                    if (newLine1.perpendicular(newLine2)):
                        self.LHS_addLine(newLine1, None)
                        self.LHS_addLine(newLine2, None)
                    if (newLine1.perpendicular(newLine3)):
                        self.LHS_addLine(newLine1, None)
                        self.LHS_addLine(newLine3, None)
                    if (newLine2.perpendicular(newLine3)):
                        self.LHS_addLine(newLine2, None)
                        self.LHS_addLine(newLine3, None)


        # for each of the line
        for i in range(len(self.lines)-1):
            li = self.lines[i].nodeElement
            # first two points for the plane: (line.root) and (line.root + line.direction)
            p1, p2 = li.root, li.root.add(li.direction)
            # compare with another line
            for j in range(i+1, len(self.lines)):
                lj = self.lines[j].nodeElement

                # ignore *parallel* and skew lines
                # TODO: WARNING: it is set so that if there are more than 10 line nodes in the lhs, the library would not
                # generate plane nodes between parallel lines to improve matching speed. However, this is not the best
                # solution going forward and an alternative way to improve matching speed should be considered. 
                if ruleAnalysis and len(self.lines) > 10 and lj.parallel(li):
                    continue
                if lj.skew(li):
                    continue

                # third point for the plane: (other_line.root + other_line.direction)
                p3 = lj.root
                if p3.colinear(p1,p2):
                    p3 = p3.add(lj.direction)
                    if p3.colinear(p1,p2):
                        continue

                # create a plane with the 3 points
                plane = plane3D(sortClass.getSort('__dummyPlane3D__'), [p1,p2,p3])

                # add the plane to the plane list
                self.LHS_addPlane(plane)

        # # if there are too few line nodes, generate more from plane segment boundaries
        if ruleAnalysis and len(self.lines) < 2 :
            self.planeCase = True

        if self.planeCase:
            for i in range(len(self.planes)):
                for pl in self.planes[i].forms:
                    for b in pl.boundaries:
                        for ln in b.canonical:
                            boundLine = lineSegmentP3D(sortClass.getSort('__dummyLineSegP3D__'), ln.start, ln.end)
                            self.LHS_addLine(boundLine.parametricNode())
        # for i in range(len(self.planes)):
        #     for pl in self.planes[i].forms:
        #         for b in pl.boundaries:
        #             for ln in b.canonical:
        #                 boundLine = lineSegmentP3D(sortClass.getSort('__dummyLineSegP3D__'), ln.start, ln.end)
        #                 self.LHS_addLine(boundLine.parametricNode())

        # for each of the plane in the plane list
        for i in range(len(self.planes)):
            # if there's no plane-segment in the plane, continue
            if len(self.planes[i].forms) == 0:
                continue

            pi = self.planes[i].nodeElement

            # compare with all other planes
            for j in range(len(self.planes)):

                # if there's no plane-segment in this plane, continue
                if len(self.planes[j].forms) == 0:
                    continue

                pj = self.planes[j].nodeElement

                # ignore parallel planes
                if pj.parallel(pi):
                    continue
                
                # find the line intersection and add it to the line list
                direction, root = pj.intersection(pi)
                intersection = line3D(sortClass.getSort('__dummyLine3D__'), root, root.add(direction))
                self.LHS_addLine(intersection)

                
    def classifyObjs(self, ruleAnalysis = False):
        """
        Classify each of the extra objects (neither line nor plane) in the shape

        * for point:  
            * 'f': free points, do not lie on any line in the shape
            * 'l': line points, contained by exactly one line in the shape
            * 'i': intersection points, contained by at least 2 lines in the shape
        * for curves: assign the curves to the plane they belong to
        """

        # for each of the points, check how many lines they are contained by:
        # 0 lines -> free point - 'f'
        # 1 line -> line point - 'l'
        # 2+ lines -> intersection point - 'i's
        for p in self.points:
            check = 0
            prevNode = None

            for node in self.lines:
                if (prevNode is None or prevNode.unitDir != node.nodeElement.unitDir 
                or prevNode.root != node.nodeElement.root) and node.nodeElement.containsPoint(p.value):
                    check+=1
                    prevNode = node.nodeElement
                    node.containedPoints.append(p)
                    if not ruleAnalysis:
                        node.linePoints.append(p)
                    elif check == 1:
                        p.assocObj = node
            if check == 1:
                p.oType = 'l'
                if ruleAnalysis:
                    p.assocObj.linePoints.append(p)
            elif check > 1:
                p.oType = 'i'
                
            if p.oType == 'f':
                for node in self.planes:
                    if node.nodeElement.containsPoint(p.value):
                        p.oType = 'p'
                        node.points.append(p)
                        break

        # for each of the curve, find a plane that contain the curve and add the curve to the plane
        c = 0
        while c < len(self.curves):
            for j in self.planes:
                if j.nodeElement.compareValue(self.curves[c].value) == 0:
                    j.addCurve(self.curves[c])
                    if ruleAnalysis:
                        self.curves.pop(c)
                        c-=1
                    break
            c += 1
            

    # ============================================================================================
    #                                         RHS ANALYSIS
    # ============================================================================================

    def RHS_addPlane(self, codesc, ind = None):
        """
        add a rhs plane segment and its parametric node into the plane-node list of the graph
        """
        # check if there's already the same plane node in the rhs
        for j in range(len(self.rhsPlanes)):
            if self.rhsPlanes[j].nodeElement.equals(codesc):
                # add the ind if the plane already exists in the rhs and return
                if ind is not None:
                    self.rhsPlanes[j].add(ind)
                return self.rhsPlanes[j]

        # check if the plane is under an existing lhs node
        lhsNode = None
        for lhsN in self.planes:
            if lhsN.nodeElement.equals(codesc):
                # if it's in the lhs and there's no individual, return
                if ind is None:
                    return None
                lhsNode = lhsN
                break
        # create a new node for when the plane is not already in the rhs
        # a new node should still be created for the rhs even if it's in the lhs
        # to carry all the new rhs plane segments in
        node = graphNode3D.planeNode(codesc, ind)
        node.rhsInfo = {
            "case": None,
            "caseNodes": [],
            "caseTemp": False,
            "containedLines": [],
            "required": False,
            "directives": []
        }
        # assign case lhs if it's already in the lhs
        if lhsNode is not None:
            node.rhsInfo['case'] = graphNode3D.planeNode.plane_casus.CASE_PROXY
            node.rhsInfo['caseNodes'] = [lhsNode]

        self.rhsPlanes.append(node)
        return node
            
    def RHS_addLine(self, codesc, ind = None):
        """
        add a rhs line segment and its parametric node into the line-node list of the graph
        """

        # check if there's already the same line node in the rhs
        for j in range(len(self.rhsLines)):
            if self.rhsLines[j].nodeElement.equals(codesc):
                if ind is not None:
                    self.rhsLines[j].add(ind)
                return self.rhsLines[j]

        # check if the line is under an existing lhs node
        lhsNode = None
        for lhsN in self.lines:
            if lhsN.nodeElement.equals(codesc):
                # if it's in the lhs and there's no individual, return
                if ind is None:
                    return None
                lhsNode = lhsN
                break

        # create a new node for when the line node is not already in the rhs
        # a new node should still be created for the rhs even if it's in the lhs
        # to carry all the new rhs line segments in
        node = graphNode3D.lineNode(codesc, ind)
        node.rhsInfo = {
            "case": None,
            "caseNodes": [],
            "caseTemp": False,
            "extraInts": set(),
            "parallel": [],
            "perpendicular": [],
            "intersect": [],
            "directives": []
        }
        if lhsNode is not None:
            node.rhsInfo['case'] = graphNode3D.lineNode.line_casus.CASE_PROXY
            node.rhsInfo['caseNodes'] = [lhsNode]

        self.rhsLines.append(node)
        return node
        
    def RHS_addPoint(self, p):
        """
        add a rhs point to the graph
        """
        pass
        
