# -*- coding: utf-8 -*-

from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import absolute_import
from builtins import round
from builtins import range

from .mapping import mapping
from .distinguishable import distinguishable
from ..struct.coordinate import coordinate
from ..transform.transformSet import transformSet
from ..transform.affineTransform import affineTransform2D
from ..transform.transform import transform

from ..ind.point import point2D
from ..ind.circle import circle2D
from ..ind.ellipse import ellipse2D
from ..ind.line import line2D
from enum import Enum

p = coordinate.p
class similarityMap2D(mapping):
    """
    Similarity Mapping in 2D class
    """
    
    type = 'map-distinguishable-similarityMap2D'
    
        
    class casus(Enum):
        """
        Casus Enum class

        All the cases recognizable in 2D similarity mapping:
        
        * 2 points not coincident
        * 1 point 1 arc not coincident
        * 1 point 1 line not colinear
        * 2 arcs not coincident
        * 1 arc 1 line not colinear
        * 3 lines no two parallel
        * 3 lines first two parallel
        * 3 lines last two parallel
        """

        CASE_2PT_NOT_COINCIDENT = "2 points not coincident"
        CASE_1PT_1ARC_NOT_COINCIDENT = "1 point 1 arc not coincident"
        CASE_1PT_1LN_NOT_COLINEAR = "1 point 1 line not colinear"
        
        CASE_2ARC_NOT_COINCIDENT = "2 arcs not coincident"
        CASE_1ARC_1LN_NOT_COLINEAR = "1 arc 1 line not colinear"
        
        #CASE_1D_2LN_PARALLEL = 4
        CASE_3LN_NO_TWO_PARALLEL = "3 lines no two parallel"
        CASE_3LN_FIRST_TWO_PARALLEL = "3 lines first two parallel"
        CASE_3LN_LAST_TWO_PARALLEL = "3 lines last two parallel"
        REDUNDANT = 0
        
    class distinguishes(Enum):
        """
        Distinguishes class: representing the outcome of the rule analysis 
        based on all current distinguishables present.

        3 Different outcomes:

        * insufficient: Missing at least 1 more distinguishable to form a supported casus.
        * unsuitable: Does not match with any of the current supported casus.
        * success: successfully conform to a supported casus.
        """
        INSUFFICIENT = 1, 
        UNSUITABLE = 2, 
        SUCCESS = 3

    def __init__(self, *args, **kwargs):
        super(similarityMap2D, self).__init__()     
        self.distinguishables = []
        self.transforms = []
        self.fromSet = [None,None,None]
        #self.checkTree = None
        self.distinguishableSuccess = False
        self.reflection = 0
        self.transformset = transformSet()
        self.casus = similarityMap2D.casus.REDUNDANT
        self.subCase = []
        self.cosAngle = None
        self.ratio = None
        self.discardedDist = []
        
        self.predicates = []
        self.postCheck = False
        return
        
    @staticmethod
    def getInstance():
        """
        get a new 2D similarity map instance
        """
        return similarityMap2D()
        
        
    def addPredicate(self, cond, keyList):
        """
        Add predicate to the predicate List
        """
        cond = cond.lower()
        self.predicates.append([cond,keyList])
    
    def clearPredicates(self):
        """
        Clear the predicate list
        """
        self.predicates = []

    def clear(self):
        """
        Clear all the distinguishables in the mapping
        """
        for i in self.distinguishables:
            i.clear()
        self.transforms = []
        return
    
    def isValid(self):
        """
        Check if the mapping is valid. This function is called under ruleObject.isValid()
        
        :returns: boolean -- the validity of the mapping
        """
        if self.distinguishableSuccess:
            return True
        if len(self.discardedDist) > 0:
            return False
        if len(self.distinguishables) == 1:
            if (self.distinguishables[0].element.individual.type == 'ind-line2D' 
            or self.distinguishables[0].element.individual.type == 'ind-point2D'):
                return True
        if len(self.distinguishables) == 2:
            if (self.distinguishables[0].element.individual.type == 'ind-line2D' 
            and self.distinguishables[1].element.individual.type == 'ind-line2D'):
                return True
            if (self.distinguishables[0].element.individual.type == 'ind-point2D' 
            and self.distinguishables[1].element.individual.type == 'ind-line2D'):
                return True
        return False


    # ============================================================================================
    #                                      ANALYZING PROCESS
    # ============================================================================================



    def distinguish(self, ind):
        """
        Add an individual into the current list of distinguishables and analyze the rule's casus base on all 
        current distinguishables
        
        :param ind: the individual to be distinguished
        :type ind: ~sortal.ind.individual.individual
        :returns: :class:`~sortal.map.distinguishable.distinguishable` -- the distinguishable matching with the 
            ind input (existing distinguishable or newly created)
        """
        self.discardedDist = []
        # find ind's codescriptor
        element = ind.codescriptor()
        if ind.hasAttribute():
            element.addAttribute(ind.attribute)
        dist = None
        # find the distinguishable that contains ind
        for i in self.distinguishables:
            # if exists a distinguishable containing ind, return
            if element == i.element.individual:
                dist = i
                dist.addInstance(ind)
                return dist
        if self.distinguishableSuccess:
            return None
        # no exsiting distinguishable -> create a new one and add it to the list
        dist = distinguishable(element)
        self.distinguishables.append(dist)
        dist.addInstance(ind)

        # analyze all current distinguishable for casus
        success = self.assessDistinguishables()

        # The current distinguishables are unsuitable -> remove the newly created one but still save it as discard
        if success == similarityMap2D.distinguishes.UNSUITABLE:
            self.discardedDist = [self.distinguishables.pop(-1)]

        # if successfully find a casus, mark mapping as success
        if success == similarityMap2D.distinguishes.SUCCESS:
            self.distinguishableSuccess = True
        return dist

    def assessDistinguishables(self):
        """
        Assessing the rule's distinguishable and find an appropriate casus
        """
        size = len(self.distinguishables)

        last = self.distinguishables[-1].element.individual
        first = None
        second = None
        if size > 1:
            first = self.distinguishables[0].element.individual
            if size > 2:
                second = self.distinguishables[1].element.individual

        if size == 1:
            # 
            if last.__class__ in (point2D, circle2D, ellipse2D, line2D):
                return similarityMap2D.distinguishes.INSUFFICIENT
                
        if size == 2:
            # Case 6,7,8: 3 lines cases
            if last.__class__ == line2D and first.__class__ == line2D:
                return similarityMap2D.distinguishes.INSUFFICIENT

        if size >= 2:
        
            # Case 1: 2 points not coincident
            if last.__class__ == point2D and first.__class__ == point2D:
                if last == first:
                    return similarityMap2D.distinguishes.UNSUITABLE
                self.fromSet[1] = last.value
                self.fromSet[0] = first.value                
                self.casus = similarityMap2D.casus.CASE_2PT_NOT_COINCIDENT
                return similarityMap2D.distinguishes.SUCCESS
                
            # Case 2: 1 point + 1 arc not coincident (point and center of arc)
            elif first.__class__ == point2D and last.__class__ in (circle2D, ellipse2D):
                if first.value == last.center:
                    if second is not None:
                        if second.value == last.center:
                            return similarityMap2D.distinguishes.UNSUITABLE
                        self.fromSet[1] = last.center
                        self.fromSet[0] = second.value     
                    else:
                        return similarityMap2D.distinguishes.UNSUITABLE
                else:
                    self.fromSet[1] = last.center
                    self.fromSet[0] = first.value         
                
                self.casus = similarityMap2D.casus.CASE_1PT_1ARC_NOT_COINCIDENT
                return similarityMap2D.distinguishes.SUCCESS
                
            # Case 3: 1 point + 1 line not colinear
            elif first.__class__ == point2D and last.__class__ == line2D:
                self.fromSet[1] = first.value
                u = last.direction
                v = self.fromSet[1].subtract(last.root)
                self.fromSet[0] = last.root.add(u.scale(v))
                
                if self.fromSet[1] == self.fromSet[0]:
                    if second is not None:
                        self.fromSet[1] = second.value
                        u = last.direction
                        v = self.fromSet[1].subtract(last.root)
                        self.fromSet[0] = last.root.add(u.scale(v))
                        if self.fromSet[1] == self.fromSet[0]:
                            return similarityMap2D.distinguishes.UNSUITABLE
                        self.distinguishables.pop(0)
                    else:
                        return similarityMap2D.distinguishes.UNSUITABLE
                
                self.subCase.append(last.assoc.containsPoint(self.fromSet[0]))
                self.subCase.append(not(last.assoc.start == self.fromSet[0] or last.assoc.end == self.fromSet[0]))
                
                self.casus = similarityMap2D.casus.CASE_1PT_1LN_NOT_COLINEAR
                return similarityMap2D.distinguishes.SUCCESS
                
            # Case 4: 2 arcs with centers not coincident
            elif first.__class__ in (circle2D, ellipse2D) and last.__class__ in (circle2D, ellipse2D):
                if first.center == last.center:
                    return similarityMap2D.distinguishes.UNSUITABLE
                self.fromSet[1] = last.center
                self.fromSet[0] = first.center                
                self.casus = similarityMap2D.casus.CASE_2ARC_NOT_COINCIDENT
                return similarityMap2D.distinguishes.SUCCESS
                
            # Case 5: 1 arc + 1 line not colinear
            elif first.__class__ in (circle2D, ellipse2D) and last.__class__ == line2D:
                self.fromSet[1] = first.center
                u = last.direction
                v = self.fromSet[1].subtract(last.root)
                if u.parallel(v):
                    return similarityMap2D.distinguishes.UNSUITABLE
                self.fromSet[0] = last.root.add(u.scale(v))     
                
                self.subCase.append(last.assoc.containsPoint(self.fromSet[0]))
                self.subCase.append(not(last.assoc.start == self.fromSet[0] or last.assoc.end == self.fromSet[0]))
                
                self.casus = similarityMap2D.casus.CASE_1ARC_1LN_NOT_COLINEAR
                return similarityMap2D.distinguishes.SUCCESS
                
            elif last.__class__ == line2D and first.__class__ == line2D and second.__class__ == line2D:
                # case 6: 3 lines no two parallel
                fpl = last.parallel(first)
                fps = second.parallel(first)
                spl = second.parallel(last)
                if not (fpl or fps or spl):
                    self.fromSet[0] = first.intersection(last)
                    self.fromSet[1] = first.intersection(second)
                    self.fromSet[2] = second.intersection(last)
                    
                    if self.fromSet[0] == self.fromSet[1] == self.fromSet[2]:
                        return similarityMap2D.distinguishes.UNSUITABLE
                        
                    self.subCase.append(first.assoc.containsPoint(self.fromSet[0]))
                    self.subCase.append(first.assoc.containsPoint(self.fromSet[1]))
                    self.subCase.append(second.assoc.containsPoint(self.fromSet[1]))
                    self.subCase.append(second.assoc.containsPoint(self.fromSet[2]))
                    self.subCase.append(last.assoc.containsPoint(self.fromSet[0]))
                    self.subCase.append(last.assoc.containsPoint(self.fromSet[2]))

                    u = self.fromSet[1].subtract(self.fromSet[0])
                    v = self.fromSet[2].subtract(self.fromSet[1])
                    w = self.fromSet[2].subtract(self.fromSet[0])
                    self.ratio = [u.squareLength(),v.squareLength(),w.squareLength()]
                    
                    self.casus = similarityMap2D.casus.CASE_3LN_NO_TWO_PARALLEL
                    return similarityMap2D.distinguishes.SUCCESS
                    
                # unsuccessful case: 3 lines all parallel with one another
                if fpl and fps and spl:
                    return similarityMap2D.distinguishes.UNSUITABLE
                    
                # if 1st parallel last (highly unlikely) --> swap 2nd and last so that first two lines parallel
                if fpl:
                    self.distinguishables[1], self.distinguishables[-1] = self.distinguishables[-1], self.distinguishables[1]
                    second, last = last, second
                    fps = second.parallel(first)
                    
                # if 2nd parallel last --> swap 1st and last so that first two lines parallel
                if spl:
                    self.distinguishables[0], self.distinguishables[-1] = self.distinguishables[-1], self.distinguishables[0]
                    first, last = last, first
                    fps = True
                    
                # case 7: 3 lines first two parallel
                if fps:
                    self.fromSet[0] = first.intersection(last)
                    self.fromSet[1] = second.intersection(last)
                    
                    self.subCase.append(first.assoc.containsPoint(self.fromSet[0]))
                    self.subCase.append(second.assoc.containsPoint(self.fromSet[1]))
                    self.subCase.append(last.assoc.containsPoint(self.fromSet[0]))
                    self.subCase.append(last.assoc.containsPoint(self.fromSet[1]))
                    
                    u = first.unitDir
                    v = last.unitDir
                    self.cosAngle = abs(u.dotproduct(v))
                    
                    self.casus = similarityMap2D.casus.CASE_3LN_FIRST_TWO_PARALLEL
                    return similarityMap2D.distinguishes.SUCCESS
                    
        return similarityMap2D.distinguishes.UNSUITABLE
    
    def postAnalysis(self):
        self.reflection = 1
        # if the casus is redundant (so the matching would be a special case) and there is a discarded distinguishable
        if self.casus == self.__class__.casus.REDUNDANT and len(self.discardedDist) > 0:
            # for the case of 1 line and 1 point colinear to the line
            if (len(self.distinguishables) == 1 
            and 'point' in self.distinguishables[0].element.individual.type
            and 'line' in self.discardedDist[-1].element.individual.type):
                self.distinguishables.append(self.discardedDist[-1])
                self.discardedDist = []
            if (len(self.distinguishables) == 2
            and 'point' in self.distinguishables[0].element.individual.type
            and 'point' in self.distinguishables[1].element.individual.type):
                self.discardedDist = []



    # ============================================================================================
    #                                    SHAPE MATCHING PROCESS
    # ============================================================================================
    

    def map(self, fromInd, toInd):
        """
        Try to map an object from the shape to a lhs object and add it as distinguishable counterpart if they can be mapped
        
        :param fromInd: lhs object
        :type fromInd: ~sortal.ind.individual.individual
        :param toInd: shape object to be mapped
        :type toInd: ~sortal.ind.individual.individual
        :returns: boolean -- whether the two objects can be mapped
        """
        for i in self.distinguishables:
            elmInd = i.element.individual
            if elmInd == fromInd:
                if elmInd.hasAttribute():
                    if fromInd.hasAttribute() and elmInd.attribute.checkPartOf(toInd.attribute):
                        i.addCounterpart(toInd)
                        return True
                else:
                    i.addCounterpart(toInd)
                    return True
        return False
    

    def potentialTransforms(self): ########################
        """
        Detecting all possible matches in the shape based on the lhs casus and distinguishables
        """
        #print('CASUS:', self.casus)
        if len(self.transforms) != 0:
            return self.transforms

        if self.casus == similarityMap2D.casus.REDUNDANT:
            return self.specialPotentialTransforms()
        size = len(self.distinguishables)
        source = []
        counterparts = [] 
                        
        for elm in self.distinguishables:
            source.append(elm.element)
            counterparts.append(elm.getCounterparts())
        target = [None]*size
        for e0 in range(len(counterparts[0])):
            target[0] = counterparts[0][e0]
            if size > 1:
                for e1 in range(len(counterparts[1])):
                    if counterparts[1][e1].individual == target[0].individual:
                        continue
                    target[1] = counterparts[1][e1]
                    if not self.assessTargetSet(target):
                        continue
                    if size > 2:
                        for e2 in range(len(counterparts[2])):
                                target[2] = counterparts[2][e2]
                                self.detectTransforms(source, target)
                    else:
                        self.detectTransforms(source, target)
        return self.transforms
        
    def assessTargetSet(self, target):
        """
        Checking if a set of distinguishables from the shape is valid for matching (no duplicate individuals, ...)
        
        :param target: list of counterpart distinguishables that are going to be used for matching
        :type target: list of :class:`~sortal.map.distinguishable.distinguishable`
        :returns: boolean -- validity of the target set of distinguishables
        """
        if self.casus == similarityMap2D.casus.CASE_3LN_NO_TWO_PARALLEL:
            first = target[0].individual
            second = target[1].individual
            if first.parallel(second):
                return False
            return True
        if self.casus == similarityMap2D.casus.CASE_3LN_FIRST_TWO_PARALLEL:
            first = target[0].individual
            second = target[1].individual
            if not first.parallel(second):
                return False
            if second.compareValue(first) == 0:
                return False
            return True
        elif self.casus == similarityMap2D.casus.CASE_3LN_LAST_TWO_PARALLEL:
            first = target[0].individual
            second = target[1].individual
            if first.parallel(second):
                return False
            intersectingPoint = second.intersection(first)
            if not(second.containsPoint(intersectingPoint) and first.containsPoint(intersectingPoint)):
                return False
            return True
        return True
        
    
    def detectTransforms(self, source, target):
        """
        Finding the transformation matrix to transform the individuals in the source (lhs distinguishables) 
        into target (shape distinguishables) based on the mapping casus
        
        :param source: list of lhs distinguishable for matching
        :type source: list of :class:`~sortal.map.distinguishable.distinguishable`
        :param target: list of shape distinguishables that are counterparts of those in source for matching
        :type target: list of :class:`~sortal.map.distinguishable.distinguishable`
        """
        size = len(source)
        #leftLast = source[-1].codescriptor()
        last = target[-1].individual
        if size > 1:
            #leftFirst = source[0].codescriptor()
            first = target[0].individual
            if size > 2:
                second = target[1].individual
        #type = last.__class__
        toSet = [None, None, None]

        # case 1: 2 points not coincident
        # matching vectors: [point1, point2]
        if self.casus == similarityMap2D.casus.CASE_2PT_NOT_COINCIDENT:
            if last == first:
                return False
            toSet[1] = last.value
            toSet[0] = first.value
            tfs = affineTransform2D.similarityTransform(self.fromSet, toSet, self.reflection)
            for tf in tfs:
                if tf is not None:
                    self.transforms.append(tf)
            return True
            
        # Case 2: 1 point + 1 arc not coincident (point and center of arc)
        # matching vectors: [arc's center, point]
        if self.casus == similarityMap2D.casus.CASE_1PT_1ARC_NOT_COINCIDENT:
            if last.center == first:
                return False
            toSet[1] = last.center
            toSet[0] = first.value
            tfs = affineTransform2D.similarityTransform(self.fromSet, toSet, self.reflection)
            for tf in tfs:
                if tf is not None:
                    self.transforms.append(tf)
            return True
            

        # Case 3: 1 point + 1 line not colinear
        # matching vectors: [point, point projection on line]
        elif self.casus == similarityMap2D.casus.CASE_1PT_1LN_NOT_COLINEAR:
            if last.type == 'ind-point2D':
                temp = first
                first = last
                last = temp
            toSet[1] = first.value
            u = last.direction
            v = toSet[1].subtract(last.root)

            if u.parallel(v):
                return False                
            toSet[0] = last.root.add(u.scale(v))
            
            if toSet[0].subtract(toSet[1]).squareLength()==0:
                return False
                
            for i in target[1].forms[target[1].individual.sort.name]:
                if (i.containsPoint(toSet[0]) >= self.subCase[0] 
                and (not(i.start == toSet[0] or i.end == toSet[0])) >= self.subCase[1]):
                    check = True
                    break
                check = False
            if not check:
                return
                
            tfs = affineTransform2D.similarityTransform(self.fromSet, toSet, self.reflection)
            for tf in tfs:
                if tf is not None:
                    self.transforms.append(tf)
            return True
            
        # Case 4: 2 arcs with centers not coincident
        # matching vectors: [2st arc's center, 1st arc's center]
        if self.casus == similarityMap2D.casus.CASE_2ARC_NOT_COINCIDENT:
            if last.center == first.center:
                return False
            toSet[1] = last.center
            toSet[0] = first.center
            tfs = affineTransform2D.similarityTransform(self.fromSet, toSet, self.reflection)
            for tf in tfs:
                if tf is not None:
                    self.transforms.append(tf)
            return True

        # Case 5: 1 arc + 1 line not colinear
        # matching vectors: [projection of arc's center on line, arc's center]
        elif self.casus == similarityMap2D.casus.CASE_1ARC_1LN_NOT_COLINEAR:
            toSet[1] = first.center
            u = last.direction
            v = toSet[1].subtract(last.root)

            if u.parallel(v):
                return False                
            toSet[0] = last.root.add(u.scale(v))
            
            if toSet[0].subtract(toSet[1]).squareLength()==0:
                return False
                
            for i in target[1].forms[target[1].individual.sort.name]:
                if (i.containsPoint(toSet[0]) >= self.subCase[0] 
                and (not(i.start == toSet[0] or i.end == toSet[0])) >= self.subCase[1]):
                    check = True
                    break
                check = False
            if not check:
                return
                
            tfs = affineTransform2D.similarityTransform(self.fromSet, toSet, self.reflection)
            for tf in tfs:
                if tf is not None:
                    self.transforms.append(tf)
            return True
                    
        # case 6: 3 lines no two parallel
        # matching vectors: [intersection between first and second, intersection between first and last]
        elif self.casus == similarityMap2D.casus.CASE_3LN_NO_TWO_PARALLEL:
            fpl = last.parallel(first)
            fps = second.parallel(first)
            spl = second.parallel(last)
            if (fpl or fps or spl):
                return False
            is1 = first.intersection(last)
            is2 = first.intersection(second)
            is3 = second.intersection(last)
            
            
            if is1 == is2 or is1 == is3 or is2 == is3:
                return False
            
            u = is2.subtract(is1).squareLength()
            v = is3.subtract(is2).squareLength()
            w = is3.subtract(is1).squareLength()
            
            if u/self.ratio[0] == v/self.ratio[1] and u/self.ratio[0] == w/self.ratio[2]:
                toSet = [is1,is2,is3,target[0],target[1],target[2]]
                for i in toSet[3].forms[toSet[3].individual.sort.name]:
                    if (i.containsPoint(toSet[0]) >= self.subCase[0] and
                        i.containsPoint(toSet[1]) >= self.subCase[1]):
                            check = True
                            break
                    check = False
                if not check:
                    return False
                for i in toSet[4].forms[toSet[3].individual.sort.name]:
                    if (i.containsPoint(toSet[1]) >= self.subCase[2] and
                        i.containsPoint(toSet[2]) >= self.subCase[3]):
                            check = True
                            break
                    check = False
                if not check:
                    return False
                for i in toSet[5].forms[toSet[3].individual.sort.name]:
                    if (i.containsPoint(toSet[0]) >= self.subCase[4] and
                        i.containsPoint(toSet[2]) >= self.subCase[5]):
                            check = True
                            break
                    check = False
                if not check:
                    return False
                    
                tfs = affineTransform2D.similarityTransform(self.fromSet, toSet, self.reflection)
                for tf in tfs:
                    if tf is not None:
                        self.transforms.append(tf)
            
            
        # case 7: 3 lines first two parallel
        # matching vectors: [intersection between first and last, intersection between second and last]
        elif self.casus == similarityMap2D.casus.CASE_3LN_FIRST_TWO_PARALLEL:
            if last.parallel(first) or second.parallel(last) or not second.parallel(first):
                return False
                
            u = first.unitDir
            v = last.unitDir
            cosA = abs(u.dotproduct(v))
            if not self.cosAngle ==  cosA:
                return False
            
            toSet[0] = first.intersection(last)
            toSet[1] = second.intersection(last)
            
            for i in target[0].forms[target[0].individual.sort.name]:
                if i.containsPoint(toSet[0]) >= self.subCase[0]:
                    check1 = True
                    break
                check1 = False
            for i in target[1].forms[target[1].individual.sort.name]:
                if i.containsPoint(toSet[1]) >= self.subCase[1]:
                    check2 = True
                    break
                check2 = False
            for i in target[2].forms[target[2].individual.sort.name]:
                if (i.containsPoint(toSet[0]) >= self.subCase[2] and
                    i.containsPoint(toSet[1]) >= self.subCase[3]):
                    check3 = True
                    break
                check3 = False
                
            if check1 and check2 and check3:               
                tfs = affineTransform2D.similarityTransform(self.fromSet, toSet, self.reflection)
                for tf in tfs:
                    if tf is not None:
                        self.transforms.append(tf)
            for i in target[0].forms[target[0].individual.sort.name]:
                if i.containsPoint(toSet[0]) >= self.subCase[1]:
                    check1 = True
                    break
                check1 = False
            for i in target[1].forms[target[1].individual.sort.name]:
                if i.containsPoint(toSet[1]) >= self.subCase[0]:
                    check2 = True
                    break
                check2 = False
            for i in target[2].forms[target[2].individual.sort.name]:
                if (i.containsPoint(toSet[0]) >= self.subCase[3] and
                    i.containsPoint(toSet[1]) >= self.subCase[2]):
                    check3 = True
                    break
                check3 = False
                
            if check1 and check2 and check3:              
                tfs = affineTransform2D.similarityTransform(self.fromSet, [toSet[1],toSet[0],None], self.reflection)
                for tf in tfs:
                    if tf is not None:
                        self.transforms.append(tf)

        return False
        
    def specialPotentialTransforms(self):
        """
        Detect transformation for mapping that has redundant casus (not enough infomation 
        to be classified as a supported casus), matching can still be performed albeit less accurate.

        Valid special cases:

        * 1 single point 
        * 1 single line
        * 2 parallel lines
        * 1 line and 1 colinear point
        """
        if len(self.distinguishables) == 1:
            dist = self.distinguishables[0]
            # One single point -> only translation
            if dist.element.individual.type == 'ind-point2D':
                fromVect = dist.element.individual.value
                for i in dist.counterparts:
                    toVect = i.individual.value
                    self.transforms.append(affineTransform2D.translation(toVect.subtract(fromVect)))

            # One single line
            # matching vectors: [line.start, line.end]
            if dist.element.individual.type == 'ind-line2D':
                self.fromSet = [dist.element.individual.assoc.start, dist.element.individual.assoc.end]
                for i in dist.counterparts:
                    for line in i.forms[dist.element.individual.assoc.sort.name]:

                        toVects = [line.start, line.end]
                        tfs = affineTransform2D.similarityTransform(self.fromSet, toVects, 1)
                        for tf in tfs:
                            if tf is not None:
                                self.transforms.append(tf)
                        
                        toVects = [line.end, line.start]
                        tfs = affineTransform2D.similarityTransform(self.fromSet, toVects, 1)
                        for tf in tfs:
                            if tf is not None:
                                self.transforms.append(tf)
        if len(self.distinguishables) == 2:
            dist0 = self.distinguishables[0]
            dist1 = self.distinguishables[1]
            # 2 lines
            # matching vectors: [first line start, second line start]
            if dist0.element.individual.type == dist1.element.individual.type == 'ind-line2D':
                if dist1.element.individual.parallel(dist0.element.individual):
                    if self.fromSet[0] is None:
                        self.fromSet = []
                        self.fromSet.append(dist0.element.forms[dist0.element.individual.assoc.sort.name][0].start)
                        self.fromSet.append(dist1.element.forms[dist1.element.individual.assoc.sort.name][0].start)
                    for i in dist0.counterparts:
                        for j in dist1.counterparts:
                            if i.individual == j.individual:
                                continue
                            if not i.individual.parallel(j.individual):
                                continue
                            for line0 in i.forms[dist0.element.individual.assoc.sort.name]:
                                for line1 in j.forms[dist1.element.individual.assoc.sort.name]:
                                    toVects = [line0.start,line1.start]
                                    tfs = affineTransform2D.similarityTransform(self.fromSet, toVects, 1)
                                    for tf in tfs:
                                        if tf is not None:
                                            self.transforms.append(tf)
                                    toVects = [line0.end,line1.end]
                                    tfs = affineTransform2D.similarityTransform(self.fromSet, toVects, 1)
                                    for tf in tfs:
                                        if tf is not None:
                                            self.transforms.append(tf)
                                            
                else:
                    if self.fromSet[0] is None:
                        self.fromSet = [dist0.element.individual.intersection(dist1.element.individual), None]
                        checkContains = False
                        sqrlength = 0
                        for i in dist0.element.forms[dist0.element.individual.assoc.sort.name]:
                            if i.containsPoint(self.fromSet[0]):
                                checkContains = True

                                sl = i.start.subtract(self.fromSet[0]).squareLength()
                                if sqrlength == 0 or sqrlength > sl > 0:
                                    self.fromSet[1] = i.start
                                    sqrlength = sl

                                sl = i.end.subtract(self.fromSet[0]).squareLength()
                                if sqrlength == 0 or sqrlength > sl > 0:
                                    self.fromSet[1] = i.end
                                    sqrlength = sl

                                break
                        for i in dist1.element.forms[dist1.element.individual.assoc.sort.name]:
                            if i.containsPoint(self.fromSet[0]):
                                checkContains = True

                                sl = i.start.subtract(self.fromSet[0]).squareLength()
                                if sqrlength == 0 or sqrlength > sl > 0:
                                    self.fromSet[1] = i.start
                                    sqrlength = sl

                                sl = i.end.subtract(self.fromSet[0]).squareLength()
                                if sqrlength == 0 or sqrlength > sl > 0:
                                    self.fromSet[1] = i.end
                                    sqrlength = sl

                                break
                        if not checkContains:
                            for i in dist0.element.forms[dist0.element.individual.assoc.sort.name]:
                                sl = i.start.subtract(self.fromSet[0]).squareLength()
                                if sqrlength == 0 or sqrlength > sl > 0:
                                    self.fromSet[1] = i.start
                                    sqrlength = sl

                                sl = i.end.subtract(self.fromSet[0]).squareLength()
                                if sqrlength == 0 or sqrlength > sl > 0:
                                    self.fromSet[1] = i.end
                                    sqrlength = sl
                            for i in dist1.element.forms[dist1.element.individual.assoc.sort.name]:
                                sl = i.start.subtract(self.fromSet[0]).squareLength()
                                if sqrlength == 0 or sqrlength > sl > 0:
                                    self.fromSet[1] = i.start
                                    sqrlength = sl

                                sl = i.end.subtract(self.fromSet[0]).squareLength()
                                if sqrlength == 0 or sqrlength > sl > 0:
                                    self.fromSet[1] = i.end
                                    sqrlength = sl

                    for i in dist0.counterparts:
                        for j in dist1.counterparts:
                            if i.individual.parallel(j.individual):
                                continue
                            toVects = [i.individual.intersection(j.individual), None]
                            for line0 in i.forms[dist0.element.individual.assoc.sort.name]:
                                if not toVects[0] == line0.start:
                                    toVects[1] = line0.start
                                    tfs = affineTransform2D.similarityTransform(self.fromSet, toVects, 1)
                                    for tf in tfs:
                                        if tf is not None:
                                            self.transforms.append(tf)
                                if not toVects[0] == line0.end:
                                    toVects[1] = line0.end
                                    tfs = affineTransform2D.similarityTransform(self.fromSet, toVects, 1)
                                    for tf in tfs:
                                        if tf is not None:
                                            self.transforms.append(tf)
            # 1 point + 1 line, colinear
            # matching vectors: [point, line.start]
            elif dist0.element.individual.type == 'ind-point2D' and dist1.element.individual.type == 'ind-line2D':
                if dist0.element.individual.value != dist1.element.individual.assoc.start:
                    if self.fromSet[0] is None:
                        self.fromSet = [dist0.element.individual.value, dist1.element.individual.assoc.start]
                    for i in dist0.counterparts:
                        for j in dist1.counterparts:
                            if j.individual.containsPoint(i.individual.value):
                                for line in j.forms[dist1.element.individual.assoc.sort.name]:
                                    if i.individual.value != line.start:
                                        toVects = [i.individual.value, line.start]
                                        tfs = affineTransform2D.similarityTransform(self.fromSet, toVects, 1)
                                        for tf in tfs:
                                            if tf is not None:
                                                self.transforms.append(tf)
                                    if i.individual.value != line.end:
                                        toVects = [i.individual.value, line.end]
                                        tfs = affineTransform2D.similarityTransform(self.fromSet, toVects, 1)
                                        for tf in tfs:
                                            if tf is not None:
                                                self.transforms.append(tf)
                if dist0.element.individual.value != dist1.element.individual.assoc.end:
                    if self.fromSet[0] is None:
                        self.fromSet = [dist0.element.individual.value, dist1.element.individual.assoc.end]
                    for i in dist0.counterparts:
                        for j in dist1.counterparts:
                            if j.individual.containsPoint(i.individual.value):
                                for line in j.forms[dist1.element.individual.assoc.sort.name]:
                                    if i.individual.value != line.start:
                                        toVects = [i.individual.value, line.start]
                                        tfs = affineTransform2D.similarityTransform(self.fromSet, toVects, 1)
                                        for tf in tfs:
                                            if tf is not None:
                                                self.transforms.append(tf)
                                    if i.individual.value != line.end:
                                        toVects = [i.individual.value, line.end]
                                        tfs = affineTransform2D.similarityTransform(self.fromSet, toVects, 1)
                                        for tf in tfs:
                                            if tf is not None:
                                                self.transforms.append(tf)
        return self.transforms
        
    def checkTransform(self):
        return

    def getTransform(self, tfs):
        """
        get the affineTransform from the transformation set
        """
        string = self.id.split("=|=")
        if len(string) > 1:
            tfs.current = string[0]
        return tfs.retrieveTransform(affineTransform2D)

    def evaluatePredicate(self, transfSet, lhs, tlhs, form, param):   
        affineTransf = self.getTransform(transfSet)
        if isinstance(param, dict):
            k = param
            param = paramSet()
            param.keys = k
        param.lhs = tlhs
        self.postCheck = True
        for cond in self.predicates:
            # void predicate
            if cond[0] == 'void':
                for bd in cond[1]:
                    # if there is a list of sort
                    if isinstance(bd[0], list):
                        sorts = bd[0]
                        bdr = bd[1:]
                        # create a plane boundary based on the vectors
                        bound = boundary()
                        if bdr[-1] == bdr[0]:
                            bdr.pop(-1)
                        for keyI in bdr:
                            newInd = keyI.transform(affineTransf)
                            # newInd = affineTransf.transform(keyI)
                            if newInd is False:
                                affineTransf.printTransf()
                                raise ValueError('matching for vector '+str(keyI)+' is not found')
                            bound.add(newInd)
                        # remove duplicate vertices, add plane border segment
                        i = 0
                        while i < len(bound.polygonal):
                            if bound.polygonal[i] == bound.polygonal[i-1]:
                                bound.polygonal.pop(i)
                            else:
                                i += 1
                        for i in range(len(bound.polygonal)):
                            bound.add(bSegment(bound.polygonal[i],bound.polygonal[i-1]))
                        # create the plane-segment
                        planeSeg = planeSegment2D(lhs.sort, [bound], True)
                        # find all lines
                        lines = []
                        form.accept(self, direction = ['voidof', planeSeg, lines, sorts])
                        if not self.postCheck:
                            return False
                            
                        # check if any of the lines cut through the created plane-segment
                        check = planeSeg.classify_line(lines)
                        if not check.notInside():
                            self.postCheck = False
                    # if there's no sort list
                    else:
                        # create plane-segment boundary
                        bound = boundary()
                        if bd[-1] == bd[0]:
                            bd.pop(-1)
                        for keyI in bd:
                            newInd = keyI.transform(affineTransf)
                            # newInd = affineTransf.transform(keyI)
                            if newInd is False:
                                affineTransf.printTransf()
                                raise ValueError('matching for vector '+str(keyI)+' is not found')
                            bound.add(newInd)
                        for i in range(len(bound.polygonal)):
                            bound.add(bSegment(bound.polygonal[i],bound.polygonal[i-1]))
                        # create plane-segment
                        planeSeg = planeSegment2D(lhs.sort, [bound], True)
                        # find all lines
                        lines = []
                        form.accept(self, direction = [cond[0], planeSeg, lines])
                        if not self.postCheck:
                            return False
                        # check if any of the lines cut through the created plane-segment
                        check = planeSeg.classify_line(lines)
                        if not check.notInside():
                            self.postCheck = False
            # inside predicate
            elif cond[0] == 'inside':
                for bd in cond[1]:
                    # if there is a list of sort
                    if isinstance(bd[0], list):
                        sorts = bd[0]
                        bdr = bd[1:]
                        # create a plane boundary based on the vectors
                        bound = boundary()
                        if bdr[-1] == bdr[0]:
                            bdr.pop(-1)
                        for keyI in bdr:
                            bound.add(keyI)
                        # remove duplicate vertices, add plane border segment
                        i = 0
                        while i < len(bound.polygonal):
                            if bound.polygonal[i] == bound.polygonal[i-1]:
                                bound.polygonal.pop(i)
                            else:
                                i += 1
                        for i in range(len(bound.polygonal)):
                            bound.add(bSegment(bound.polygonal[i],bound.polygonal[i-1]))
                        # create the plane-segment
                        planeSeg = planeSegment2D(lhs.sort, [bound], True)
                        # find all lines
                        lines = []
                        tlhs.accept(self, direction = ['insideof', planeSeg, lines, sorts])
                        if not self.postCheck:
                            return False
                            
                        # check if any of the lines cut through the created plane-segment
                        check = planeSeg.classify_line(lines)
                        if not check.isContained():
                            self.postCheck = False
                    # if there's no sort list
                    else:
                        # create plane-segment boundary
                        bound = boundary()
                        if bd[-1] == bd[0]:
                            bd.pop(-1)
                        for keyI in bd:
                            bound.add(keyI)
                        for i in range(len(bound.polygonal)):
                            bound.add(bSegment(bound.polygonal[i],bound.polygonal[i-1]))
                        # create plane-segment
                        planeSeg = planeSegment2D(lhs.sort, [bound], True)
                        # find all lines
                        lines = []
                        tlhs.accept(self, direction = [cond[0], planeSeg, lines])
                        if not self.postCheck:
                            return False
                        # check if any of the lines cut through the created plane-segment
                        check = planeSeg.classify_line(lines)
                        if not check.isContained():
                            self.postCheck = False
            # outside predicate
            elif cond[0] == 'outside':
                for bd in cond[1]:
                    # if there is a list of sort
                    if isinstance(bd[0], list):
                        sorts = bd[0]
                        bdr = bd[1:]
                        # create a plane boundary based on the vectors
                        bound = boundary()
                        if bdr[-1] == bdr[0]:
                            bdr.pop(-1)
                        for keyI in bdr:
                            bound.add(keyI)
                        # remove duplicate vertices, add plane border segment
                        i = 0
                        while i < len(bound.polygonal):
                            if bound.polygonal[i] == bound.polygonal[i-1]:
                                bound.polygonal.pop(i)
                            else:
                                i += 1
                        for i in range(len(bound.polygonal)):
                            bound.add(bSegment(bound.polygonal[i],bound.polygonal[i-1]))
                        # create the plane-segment
                        planeSeg = planeSegment2D(lhs.sort, [bound], True)
                        # find all lines
                        lines = []
                        tlhs.accept(self, direction = ['outsideof', planeSeg, lines, sorts])
                        if not self.postCheck:
                            return False
                            
                        # check if any of the lines cut through the created plane-segment
                        check = planeSeg.classify_line(lines)
                        if not check.notInside():
                            self.postCheck = False
                    # if there's no sort list
                    else:
                        # create plane-segment boundary
                        bound = boundary()
                        if bd[-1] == bd[0]:
                            bd.pop(-1)
                        for keyI in bd:
                            bound.add(keyI)
                        for i in range(len(bound.polygonal)):
                            bound.add(bSegment(bound.polygonal[i],bound.polygonal[i-1]))
                        # create plane-segment
                        planeSeg = planeSegment2D(lhs.sort, [bound], True)
                        # find all lines
                        lines = []
                        tlhs.accept(self, direction = [cond[0], planeSeg, lines])
                        if not self.postCheck:
                            return False
                        # check if any of the lines cut through the created plane-segment
                        check = planeSeg.classify_line(lines)
                        if not check.notInside():
                            self.postCheck = False
            if self.postCheck is False:
                return False
        return True

    def visitEnter(self, obj, direction = 0):
        if direction[0] == 'void':
            if 'ind-planeSegment' in obj.type :
                if ('2D' in obj.type) == ('3D' in direction[1].type):
                    raise ValueError('There exists a mix of 2D and 3D objects in the shape')
                check = obj.classify(direction[1])
                if check.disjoint() or check.touches():
                    return True
                self.postCheck = False
                return False
            if 'ind-lineSegment' in obj.type:
                direction[2].append(bSegment(obj.start, obj.end))
            if 'ind-point' in obj.type:
                check = False
                for line in direction[1].boundaries[0].canonical:
                    if line.containsPoint(obj.value):
                        check = True
                if check:
                    return True
                if direction[1].boundaries[0].containsPointInside(obj.value):
                    self.postCheck = False
                    return False
        elif direction[0] == 'voidof':
            if 'ind-planeSegment' in obj.type :
                if not (obj.sort.name in direction[-1] or obj.sort.baseSort().name in direction[-1]):
                    return True
                if ('2D' in obj.type) == ('3D' in direction[1].type):
                    raise ValueError('There exists a mix of 2D and 3D objects in the shape')
                check = obj.classify(direction[1])
                if check.disjoint() or check.touches():
                    return True
                self.postCheck = False
                return False
            if 'ind-lineSegment' in obj.type:
                if not (obj.sort.name in direction[-1] or obj.sort.baseSort().name in direction[-1]):
                    return True
                direction[2].append(bSegment(obj.start, obj.end))
            if 'ind-point' in obj.type:
                if not (obj.sort.name in direction[-1] or obj.sort.baseSort().name in direction[-1]):
                    return True
                check = False
                for line in direction[1].boundaries[0].canonical:
                    if line.containsPoint(obj.value):
                        check = True
                if check:
                    return True
                if direction[1].boundaries[0].containsPointInside(obj.value):
                    self.postCheck = False
                    return False
        elif direction[0] == 'inside':
            if 'ind-planeSegment' in obj.type :
                if ('2D' in obj.type) == ('3D' in direction[1].type):
                    raise ValueError('There exists a mix of 2D and 3D objects in the shape')
                check = obj.classify(direction[1])
                if check.isContained():
                    return True
                self.postCheck = False
                return False
            if 'ind-lineSegment' in obj.type:
                direction[2].append(bSegment(obj.start, obj.end))
            if 'ind-point' in obj.type:
                check = False
                for line in direction[1].boundaries[0].canonical:
                    if line.containsPoint(obj.value):
                        check = True
                if check:
                    return True
                if direction[1].boundaries[0].containsPointInside(obj.value):
                    return True
                self.postCheck = False
                return False
        elif direction[0] == 'insideof':
            if 'ind-planeSegment' in obj.type :
                if not (obj.sort.name in direction[-1] or obj.sort.baseSort().name in direction[-1]):
                    return True
                if ('2D' in obj.type) == ('3D' in direction[1].type):
                    raise ValueError('There exists a mix of 2D and 3D objects in the shape')
                check = obj.classify(direction[1])
                if check.isContained():
                    return True
                self.postCheck = False
                return False
            if 'ind-lineSegment' in obj.type:
                if not (obj.sort.name in direction[-1] or obj.sort.baseSort().name in direction[-1]):
                    return True
                direction[2].append(bSegment(obj.start, obj.end))
            if 'ind-point' in obj.type:
                if not (obj.sort.name in direction[-1] or obj.sort.baseSort().name in direction[-1]):
                    return True
                check = False
                for line in direction[1].boundaries[0].canonical:
                    if line.containsPoint(obj.value):
                        check = True
                if check:
                    return True
                if direction[1].boundaries[0].containsPointInside(obj.value):
                    return True
                self.postCheck = False
                return False
        elif direction[0] == 'outside':
            if 'ind-planeSegment' in obj.type :
                if ('2D' in obj.type) == ('3D' in direction[1].type):
                    raise ValueError('There exists a mix of 2D and 3D objects in the shape')
                check = obj.classify(direction[1])
                if check.disjoint() or check.touches():
                    return True
                self.postCheck = False
                return False
            if 'ind-lineSegment' in obj.type:
                direction[2].append(bSegment(obj.start, obj.end))
            if 'ind-point' in obj.type:
                check = False
                for line in direction[1].boundaries[0].canonical:
                    if line.containsPoint(obj.value):
                        check = True
                if check:
                    return True
                if direction[1].boundaries[0].containsPointInside(obj.value):
                    self.postCheck = False
                    return False
        elif direction[0] == 'outsideof':
            if 'ind-planeSegment' in obj.type :
                if not (obj.sort.name in direction[-1] or obj.sort.baseSort().name in direction[-1]):
                    return True
                if ('2D' in obj.type) == ('3D' in direction[1].type):
                    raise ValueError('There exists a mix of 2D and 3D objects in the shape')
                check = obj.classify(direction[1])
                if check.disjoint() or check.touches():
                    return True
                self.postCheck = False
                return False
            if 'ind-lineSegment' in obj.type:
                if not (obj.sort.name in direction[-1] or obj.sort.baseSort().name in direction[-1]):
                    return True
                direction[2].append(bSegment(obj.start, obj.end))
            if 'ind-point' in obj.type:
                if not (obj.sort.name in direction[-1] or obj.sort.baseSort().name in direction[-1]):
                    return True
                check = False
                for line in direction[1].boundaries[0].canonical:
                    if line.containsPoint(obj.value):
                        check = True
                if check:
                    return True
                if direction[1].boundaries[0].containsPointInside(obj.value):
                    self.postCheck = False
                    return False
        return True

    def visitLeave(self, obj, direction = 0):
        return
