# -*- coding: utf-8 -*-

from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import absolute_import
from builtins import round
from builtins import range

from .mapping import mapping
from .distinguishable import distinguishable
from ..struct.coordinate import coordinate, c0p5, c2
from ..transform.transformSet import transformSet
from ..transform.affineTransform import affineTransform3D
from ..transform.transform import transform
from ..transform.paramSet import paramSet

from ..ind.plane import plane3D
from ..ind.planeSegment3D import planeSegment3D
from ..ind.line import line3D
from ..ind.ellipse import ellipse3D
from ..ind.circle import circle3D
from ..ind.point import point3D

from ..struct.planeStruct import bSegment, boundary
from ..sort.sort import sort as sortClass
from enum import Enum

p = coordinate.p
arcClasses = (circle3D, ellipse3D)

class similarityMap3D(mapping):
    """
    Similarity Mapping in 3D class
    """
    
    type = 'map-distinguishable-similarityMap3D'

    class casus(Enum):

        """
        Casus Enum class

        All the cases recognizable in 3D similarity mapping:
        
        * 3 points not coincident
        * 2 points 1 plane not perpendicular
        * 1 point 1 arc not coincident
        * 1 point 1 line not colinear
        * 1 point 2 planes not concurrent
        * 2 arcs not concentric
        * 2 arcs not parallel
        * 1 arc 1 line not colinear
        * 3 lines no two parallel
        * 3 lines first two parallel
        * 3 lines last two parallel
        * 2 lines skewed
        * 2 lines 1 plane not parallel nor concurrent
        * 2 lines 1 plane parallel first
        * 2 lines 1 plane parallel both
        * 1 lines 2 planes not all concurrent
        """

        CASE_3PT_NOT_COLINEAR = '3 points not colinear'
        CASE_2PT_1PL_NOT_PERPENDICULAR = '2 points 1 plane not perpendicular'
        
        CASE_1PT_1ARC_NOT_COINCIDENT = '1 point 1 arc not coincident'
        
        CASE_1PT_1LN_NOT_COLINEAR = '1 point 1 line not colinear'

        CASE_1PT_1LN_1PL_NOT_COINCIDENT_NOT_PERPENDICULAR = '1 point 1 line 1 plane not coincident nor perpendicular' # NOT IMPLEMENTED

        CASE_1PT_2PL_NOT_CONCURRENT = '1 point 2 planes not concurrent'
        
        CASE_2ARC_NOT_CONCENTRIC = '2 arcs not concentric'
        CASE_2ARC_NOT_PARALLEL = '2 arcs not parallel'
        CASE_1ARC_1LN_NOT_COLINEAR = '1 arc 1 line not colinear'
        
        CASE_3LN_NO_TWO_PARALLEL = '3 lines no two parallel'
        CASE_3LN_FIRST_TWO_PARALLEL = '3 lines first two parallel'
        CASE_3LN_LAST_TWO_PARALLEL = '3 lines last two parallel'
        CASE_2LN_SKEWED = '2 lines skewed'
        
        CASE_2LN_1PL_NOT_PARALLEL_NOR_CONCURRENT = '2 lines 1 plane not parallel nor concurrent'
        CASE_2LN_1PL_PARALLEL_FIRST = '2 lines 1 plane parallel first'
        CASE_2LN_1PL_PARALLEL_BOTH = '2 lines 1 plane parallel both'
        
        CASE_1LN_2PL_NOT_ALL_CONCURRENT = '1 lines 2 planes not all concurrent'

        CASE_4PL_NOT_ALL_PARALLEL = '4 planes not all parallel'

        REDUNDANT = 'redundant'
        
        

    class distinguishes(Enum):
        """
        Distinguishes class: representing the outcome of the rule analysis 
        based on all current distinguishables present.

        3 Different outcomes:

        * success: successfully conform to a supported casus.
        * insufficient: Missing at least 1 more distinguishable to form a supported casus.
        * unsuitable: Does not match with any of the current supported casus.
        * unsuitable keep discarded: Does not match with any of the current supported casus, but keep the discarded dist as is.
        """
        SUCCESS = 1,
        INSUFFICIENT = 2, 
        UNSUITABLE = 3

    def __init__(self, *args, **kwargs):
        super(similarityMap3D, self).__init__()     
        self.distinguishables = []
        self.transforms = []
        self.fromSet = [None,None,None]
        self.checkTree = None
        self.distinguishableSuccess = False
        self.reflection = 0
        self.transformset = transformSet()
        self.casus = similarityMap3D.casus.REDUNDANT
        self.subCase = []
        self.cosAngle = None
        self.ratio = None
        self.discardedDist = []

        self.predicates = []
        self.postCheck = False

    def addPredicate(self, cond, keyList):
        """
        Add predicate to the predicate List
        """
        cond = cond.lower()
        self.predicates.append([cond,keyList])
        
    def clearPredicates(self):
        """
        Clear the predicate list
        """
        self.predicates = []


    @staticmethod
    def getInstance():
        """
        get a new 3D similarity map instance
        """
        return similarityMap3D()
    
    def clear(self):
        """
        Clear all the distinguishables in the mapping
        """
        for i in self.distinguishables:
            i.clear()
        self.transforms = []
        return
        
    def isValid(self):
        """
        Check if the mapping is valid. This function is called under ruleObject.isValid()
        
        :returns: boolean -- the validity of the mapping
        """
        if self.distinguishableSuccess:
            return True
        if len(self.discardedDist) > 0 and len(self.distinguishables) == 2:
            if ((self.distinguishables[0].element.individual.type == 'ind-point3D' and
                self.distinguishables[1].element.individual.type == 'ind-point3D')):
                return True
            return False
        if len(self.distinguishables) == 1:
            if (self.distinguishables[0].element.individual.type == 'ind-point3D' 
            or self.distinguishables[0].element.individual.type == 'ind-line3D'):
                return True
        if len(self.distinguishables) == 2:
            if ((self.distinguishables[0].element.individual.type == 'ind-point3D' and
                self.distinguishables[1].element.individual.type == 'ind-point3D')
            or  (self.distinguishables[0].element.individual.type == 'ind-point3D' and
                self.distinguishables[1].element.individual.type == 'ind-line3D')
            or  (self.distinguishables[0].element.individual.type == 'ind-line3D' and
                self.distinguishables[1].element.individual.type == 'ind-line3D')):
                return True
        return False

    # ============================================================================================
    #                                      ANALYZING PROCESS
    # ============================================================================================



    def distinguish(self, ind):
        """
        Add an individual into the current list of distinguishables and analyze the rule's casus base on all 
        current distinguishables
        
        :param ind: the individual to be distinguished
        :type ind: ~sortal.ind.individual.individual
        :returns: :class:`~sortal.map.distinguishable.distinguishable` -- the distinguishable matching with the 
            ind input (existing distinguishable or newly created)
        """
        # self.discardedDist = []

        # find ind's codescriptor
        element = ind.codescriptor()
        if ind.hasAttribute():
            element.addAttribute(ind.attribute)
        dist = None
        # find the distinguishable that contains ind
        for i in self.distinguishables:
            # if exists a distinguishable containing ind, return
            if element == i.element.individual:
                dist = i
                dist.addInstance(ind)
                return dist
        if self.distinguishableSuccess:
            return None
        # no exsiting distinguishable -> create a new one and add it to the list
        dist = distinguishable(element)
        self.distinguishables.append(dist)
        dist.addInstance(ind)

        # analyze all current distinguishable for casus
        success = self.assessDistinguishables()

        # The current distinguishables are unsuitable -> remove the newly created one but still save it as discard
        if success == similarityMap3D.distinguishes.UNSUITABLE:
            # self.discardedDist = self.distinguishables.pop(-1)
            self.discardedDist.append(self.distinguishables.pop(-1))
        # if successfully find a casus, mark mapping as success
        elif success == similarityMap3D.distinguishes.SUCCESS:
            self.distinguishableSuccess = True

        return dist

    def assessDistinguishables(self):
        """
        Assessing the rule's distinguishable and find an appropriate casus
        """
        size = len(self.distinguishables)
        
        last = self.distinguishables[-1].element.individual
        first = None
        second = None
        third = None
        if size > 1:
            first = self.distinguishables[0].element.individual
            if size > 2:
                second = self.distinguishables[1].element.individual
                if size > 3:
                    third = self.distinguishables[2].element.individual

        if size == 1:
            if last.__class__ in (point3D, line3D, plane3D, circle3D, ellipse3D):
                return similarityMap3D.distinguishes.INSUFFICIENT
        if size == 2 : 
            #cases with 2 lines
            if last.__class__ == line3D and first.__class__ == line3D:
                if last.skew(first):
                    return self.assessSkew(first, last)
                return similarityMap3D.distinguishes.INSUFFICIENT

            #cases with 2 points
            elif last.__class__ == point3D and first.__class__ == point3D:
                if last == first:
                    return similarityMap3D.distinguishes.UNSUITABLE
                return similarityMap3D.distinguishes.INSUFFICIENT

            #cases with 1 point + 1 plane
            elif last.__class__ == plane3D and first.__class__ == point3D:
                if last.containsPoint(first.value):
                    return similarityMap3D.distinguishes.UNSUITABLE
                discarded = None
                for disc in self.discardedDist:
                    dInd = disc.element.individual
                    if (dInd.__class__ == line3D and not dInd.direction.parallel(last.normal) and not last.containsLine(dInd)):
                        discarded = disc
                if discarded is not None:
                    self.fromSet[0] = first.value
                    self.fromSet[1] = last.projection(first.value)

                    lineParallelPlane = discarded.element.individual.direction.perpendicular(last.normal)
                    self.subCase.append(lineParallelPlane)
                    if lineParallelPlane:
                        PtPlDistance = self.fromSet[0].subtract(self.fromSet[1]).length()
                        self.fromSet[2] = first.value.add(discarded.element.individual.unitDir.scale(PtPlDistance))
                    else:
                        self.fromSet[2] = last.intersectionLine(discarded.element.individual)

                    # the discarded distinguishable must be added back to the active distinguishables list
                    self.distinguishables.insert(1, discarded)

                    self.casus = similarityMap3D.casus.CASE_1PT_1LN_1PL_NOT_COINCIDENT_NOT_PERPENDICULAR
                    return similarityMap3D.distinguishes.SUCCESS
                return similarityMap3D.distinguishes.INSUFFICIENT
                
            #cases with 1 line + 1 plane
            elif last.__class__ == plane3D and first.__class__ == line3D:
                if last.containsLine(first):
                    return similarityMap3D.distinguishes.UNSUITABLE
                return similarityMap3D.distinguishes.INSUFFICIENT

                
        if size >= 2:
            #case: 1 point 1 arc not coincident
            if first.__class__ == point3D and last.__class__ in arcClasses:
                lastCenter = last.center
                
                point = first.value
                if first.value == lastCenter:
                    if second is None or second.value == lastCenter:
                        return similarityMap3D.distinguishes.UNSUITABLE
                    point = second.value
                    self.distinguishables.pop(0)
                elif second is not None:
                    self.distinguishables.pop(1)
                                        
                self.fromSet[0] = point
                self.fromSet[1] = lastCenter
                
                u = last.normal
                v = point.subtract(lastCenter)
                
                if u.parallel(v):
                    return similarityMap3D.distinguishes.UNSUITABLE
                                    
                self.fromSet[2] = self.fromSet[1].add(u.scale(((v.dotproduct(v)/u.dotproduct(u))**c0p5)*c2))
                
                self.cosAngle = u.unit().dotproduct(v.unit())
                self.subCase.append(last.__class__)
                if self.cosAngle == 0:
                    self.subCase.append(last.assoc.containsPoint(self.fromSet[0]))
                else:
                    self.subCase.append(False)
                                
                self.casus = similarityMap3D.casus.CASE_1PT_1ARC_NOT_COINCIDENT
                return similarityMap3D.distinguishes.SUCCESS         
                
            #case: 1 point + 1 line not colinear
            if first.__class__ == point3D and last.__class__ == line3D:
                self.fromSet[1] = first.value
                u = last.direction
                v = self.fromSet[1].subtract(last.root)
                
                if u.parallel(v):
                    if second is None:
                        return similarityMap3D.distinguishes.UNSUITABLE
                    self.fromSet[1] = second.value
                    v = self.fromSet[1].subtract(last.root)
                    if u.parallel(v):
                        return similarityMap3D.distinguishes.UNSUITABLE
                    self.distinguishables.pop(0)
                elif second is not None:
                    self.distinguishables.pop(1)
                    
                self.fromSet[0] = last.root.add(u.scale(v))
                self.subCase.append(last.assoc.containsPoint(self.fromSet[0]))
                self.subCase.append(not(last.assoc.start == self.fromSet[0] or last.assoc.end == self.fromSet[0]))
                
                d = self.fromSet[1].subtract(self.fromSet[0])
                self.fromSet[2] = self.fromSet[0].add(u.scale(((d.dotproduct(d)/u.dotproduct(u))**c0p5)*c2))

                if not last.assoc.containsPoint(self.fromSet[2]):
                    self.fromSet[2] = self.fromSet[0].subtract(u.scale(((d.dotproduct(d)/u.dotproduct(u))**c0p5)*c2))
                
                
                self.subCase.append(last.assoc.containsPoint(self.fromSet[2]))
                
                self.casus = similarityMap3D.casus.CASE_1PT_1LN_NOT_COLINEAR
                return similarityMap3D.distinguishes.SUCCESS         
                
            #case: 2 arcs not concentric / 2 concentric arcs with normals not parallel
            if first.__class__ in arcClasses and last.__class__ in arcClasses:
                firstCenter = first.center
                lastCenter = last.center
                
                #case: 2 arcs not concentric
                if firstCenter != lastCenter:
                    firstArc = first
                    lastArc = last
                    self.fromSet[0] = firstCenter
                    self.fromSet[1] = lastCenter
                    u = lastArc.normal
                    v = firstCenter.subtract(lastCenter)
                    
                    if u.parallel(v):
                        firstArc = last
                        lastArc = first
                        self.fromSet[0] = lastCenter
                        self.fromSet[1] = firstCenter
                        u = lastArc.normal
                        v = lastCenter.subtract(firstCenter)
                        if u.parallel(v):
                            return similarityMap3D.distinguishes.UNSUITABLE
                            
                    self.fromSet[2] = self.fromSet[1].add(u.scale(((v.dotproduct(v)/u.dotproduct(u))**c0p5)*c2))
                    
                    self.cosAngle = u.unit().dotproduct(v.unit())

                    distanceVec = self.fromSet[1].subtract(self.fromSet[0])
                    distanceSqr = distanceVec.dotproduct(distanceVec)

                    if firstArc.__class__ == circle3D:
                        firstRadius = firstArc.radius
                    else:
                        firstRadius = firstArc.radius[0]
                    
                    if lastArc.__class__ == circle3D:
                        lastRadius = lastArc.radius
                    else:
                        lastRadius = lastArc.radius[0]
                    
                    self.ratio = [firstRadius/lastRadius, (firstRadius**2) / distanceSqr]
                    
                    self.subCase.append(firstArc.__class__)
                    self.subCase.append(lastArc.__class__)
                                        
                    self.casus = similarityMap3D.casus.CASE_2ARC_NOT_CONCENTRIC
                    return similarityMap3D.distinguishes.SUCCESS
                    
                #case: 2 concentric arcs with normals not parallel
                elif not first.normal.parallel(last.normal):
                    self.fromSet[0] = firstCenter
                    
                    if first.__class__ == circle3D:
                        firstRadius = first.radius
                    else:
                        firstRadius = first.radius[0]
                    
                    if last.__class__ == circle3D:
                        lastRadius = last.radius
                    else:
                        lastRadius = last.radius[0]
                        
                    self.ratio = firstRadius/lastRadius
                        
                    self.subCase.append(first.__class__)
                    self.subCase.append(last.__class__)
                    
                    firstUnit = first.normal.unit()
                    lastUnit = last.normal.unit()
                    
                    self.cosAngle = firstUnit.dotproduct(lastUnit)
                    
                    self.fromSet[1] = self.fromSet[0].add(firstUnit.scale(firstRadius))
                    self.fromSet[2] = self.fromSet[0].add(lastUnit.scale(lastRadius))
                    
                    #self.checkTree = True
                    
                    self.casus = similarityMap3D.casus.CASE_2ARC_NOT_PARALLEL
                    return similarityMap3D.distinguishes.SUCCESS       
                else:
                    return similarityMap3D.distinguishes.UNSUITABLE
                    
                    
            #case: 1 arc + 1 line not colinear                    
            elif first.__class__ in arcClasses and last.__class__ == line3D:
                self.fromSet[1] = first.center
                u = last.direction
                v = self.fromSet[1].subtract(last.root)
                
                if u.parallel(v):
                    if second is None:
                        return similarityMap3D.distinguishes.UNSUITABLE
                    self.fromSet[1] = second.center
                    v = self.fromSet[1].subtract(last.root)
                    if u.parallel(v):
                        return similarityMap3D.distinguishes.UNSUITABLE
                    self.distinguishables.pop(0)
                elif second is not None:
                    self.distinguishables.pop(1)
                    
                self.fromSet[0] = last.root.add(u.scale(v))
                self.subCase.append(last.assoc.containsPoint(self.fromSet[0]))
                self.subCase.append(not(last.assoc.start == self.fromSet[0] or last.assoc.end == self.fromSet[0]))
                
                d = self.fromSet[1].subtract(self.fromSet[0])
                self.fromSet[2] = self.fromSet[0].add(u.scale((d.dotproduct(d)/u.dotproduct(u))**c0p5*c2))

                if not last.assoc.containsPoint(self.fromSet[2]):
                    self.fromSet[2] = self.fromSet[0].subtract(u.scale((d.dotproduct(d)/u.dotproduct(u))**c0p5*c2))
                
                
                self.subCase.append(last.assoc.containsPoint(self.fromSet[2]))
                
                self.casus = similarityMap3D.casus.CASE_1ARC_1LN_NOT_COLINEAR
                return similarityMap3D.distinguishes.SUCCESS         
        
            #case: 3 points not colinear
            elif last.__class__ == point3D and first.__class__ == point3D and second.__class__ == point3D:
                if last == first or last == second:
                    return similarityMap3D.distinguishes.UNSUITABLE
                if last.value.colinear(first.value, second.value):
                    return similarityMap3D.distinguishes.UNSUITABLE
                    
                self.fromSet[0] = first.value      
                self.fromSet[1] = second.value                
                self.fromSet[2] = last.value
                
                self.casus = similarityMap3D.casus.CASE_3PT_NOT_COLINEAR
                return similarityMap3D.distinguishes.SUCCESS
            
            #cases: 3 lines
            elif last.__class__ == line3D and first.__class__ == line3D and second.__class__ == line3D:
                if last.skew(first):
                    del self.distinguishables[1]
                    return self.assessSkew(first, last)
                if last.skew(second):
                    del self.distinguishables[0]
                    return self.assessSkew(second, last)
                    
                fpl = last.parallel(first)
                fps = second.parallel(first)
                spl = second.parallel(last)
                if not (fpl or fps or spl):
                    self.fromSet[0] = first.intersection(last)                        
                    self.fromSet[1] = first.intersection(second)
                    self.fromSet[2] = second.intersection(last)
                        
                    if self.fromSet[0] == self.fromSet[1] == self.fromSet[2]:
                        return similarityMap3D.distinguishes.UNSUITABLE
                        
                    self.subCase.append(first.assoc.containsPoint(self.fromSet[0]))
                    self.subCase.append(first.assoc.containsPoint(self.fromSet[1]))
                    self.subCase.append(second.assoc.containsPoint(self.fromSet[1]))
                    self.subCase.append(second.assoc.containsPoint(self.fromSet[2]))
                    self.subCase.append(last.assoc.containsPoint(self.fromSet[0]))
                    self.subCase.append(last.assoc.containsPoint(self.fromSet[2]))
                    
                    u = self.fromSet[1].subtract(self.fromSet[0])
                    v = self.fromSet[2].subtract(self.fromSet[1])
                    w = self.fromSet[2].subtract(self.fromSet[0])
                    
                    self.ratio = [u.squareLength(),v.squareLength(),w.squareLength()]
                    
                    self.casus = similarityMap3D.casus.CASE_3LN_NO_TWO_PARALLEL
                    return similarityMap3D.distinguishes.SUCCESS

                if fpl and fps and spl:
                    return similarityMap3D.distinguishes.UNSUITABLE
                    
                if fpl:
                    self.distinguishables[1], self.distinguishables[-1] = self.distinguishables[-1], self.distinguishables[1]
                    second, last = last, second
                    fps = second.parallel(first)
                    
                if spl:
                    self.distinguishables[0], self.distinguishables[-1] = self.distinguishables[-1], self.distinguishables[0]
                    first, last = last, first
                    fps = True
                    
                if fps:
                    try:
                        self.fromSet[0] = first.intersection(last)
                        self.fromSet[1] = second.intersection(last)
                    except ValueError:
                        return similarityMap3D.distinguishes.UNSUITABLE
                        
                    u = first.direction
                    d = self.fromSet[1].subtract(self.fromSet[0])
                    v = u.scale((d.dotproduct(d)/(u.dotproduct(u)))**c0p5*c2)
                    self.fromSet[2] = self.fromSet[0].add(v)
                    if not first.assoc.containsPoint(self.fromSet[2]):
                        self.fromSet[2] = self.fromSet[0].subtract(v)
                    
                    self.subCase.append(first.assoc.containsPoint(self.fromSet[0]))
                    self.subCase.append(second.assoc.containsPoint(self.fromSet[1]))
                    self.subCase.append(last.assoc.containsPoint(self.fromSet[0]))
                    self.subCase.append(last.assoc.containsPoint(self.fromSet[1]))
                    self.subCase.append(first.assoc.containsPoint(self.fromSet[2]))
                    
                    u = first.unitDir
                    v = last.unitDir
                    self.cosAngle = abs(u.dotproduct(v))
                
                    self.casus = similarityMap3D.casus.CASE_3LN_FIRST_TWO_PARALLEL
                    return similarityMap3D.distinguishes.SUCCESS

            #cases: 2 points, 1 plane
            elif last.__class__ == plane3D and first.__class__ == point3D and second.__class__ == point3D:
                foot = last.projection(first.value)
                if foot == first:
                    foot = last.projection(second.value)
                    if foot == second:
                        d = second.subtract(first)
                        u = last.normal.product(d)
                        foot = first.add(u.scale((d.dotProduct(d) / (u.dotProduct(u)))**c0p5*c2))
                        self.subCase = 0
                    else:
                        self.subCase = 2
                else:
                    self.subCase = 1

                if (foot.colinear(second.value, first.value)):
                    return similarityMap3D.distinguishes.UNSUITABLE
                self.fromSet[0] = first.value
                self.fromSet[1] = second.value
                self.fromSet[2] = foot
                self.casus = similarityMap3D.casus.CASE_2PT_1PL_NOT_PERPENDICULAR
                return similarityMap3D.distinguishes.SUCCESS
            
            #cases: 1 points, 2 plane
            elif last.__class__ == plane3D and first.__class__ == point3D and second.__class__ == plane3D:
                if last.parallel(second):
                    return similarityMap3D.distinguishes.UNSUITABLE
                self.fromSet[0] = first.value
                self.fromSet[1] = second.projection(first.value)
                self.fromSet[2] = last.projection(first.value)
                if self.fromSet[0] == self.fromSet[1] or self.fromSet[2] == self.fromSet[1] or self.fromSet[0] == self.fromSet[2]:
                    return similarityMap3D.distinguishes.UNSUITABLE
                self.casus = similarityMap3D.casus.CASE_1PT_2PL_NOT_CONCURRENT
                return similarityMap3D.distinguishes.SUCCESS
            
            #cases: 2 lines, 1 plane
            elif last.__class__ == plane3D and first.__class__ == line3D and second.__class__ == line3D:
                if last.containsLine(first) or last.containsLine(second):
                    return similarityMap3D.distinguishes.UNSUITABLE
                if last.parallelLine(first) and last.parallelLine(second):
                    if first.parallel(second):
                        return similarityMap3D.distinguishes.UNSUITABLE
                    self.fromSet[0] = first.intersection(second)
                    self.fromSet[1] = last.projection(self.fromSet[0])
                    d0 = last.normal.product(first.unitDir)
                    l0 = self.fromSet[1].subtract(self.fromSet[0]).length()
                    self.fromSet[2] = self.fromSet[1].add(d0.scale(l0))
                    self.casus = similarityMap3D.casus.CASE_2LN_1PL_PARALLEL_BOTH
                    return similarityMap3D.distinguishes.SUCCESS
                if last.parallelLine(first) or last.parallelLine(second):
                    if last.parallelLine(second):
                        first, second = second, first
                    self.fromSet[0] = last.intersectionLine(second)
                    self.fromSet[1] = first.intersection(second)
                    l0 = self.fromSet[1].subtract(self.fromSet[0]).length()
                    self.fromSet[2] = self.fromSet[1].add(first.unitDir.scale(l0))
                    self.casus = similarityMap3D.casus.CASE_2LN_1PL_PARALLEL_FIRST
                    return similarityMap3D.distinguishes.SUCCESS
                self.fromSet[0] = last.intersectionLine(first)
                self.fromSet[1] = last.intersectionLine(second)
                if self.fromSet[0] == self.fromSet[1]:
                    return similarityMap3D.distinguishes.UNSUITABLE
                self.fromSet[2] = self.fromSet[0].add(last.normal.scale(self.fromSet[0].subtract(self.fromSet[1]).length()))
                self.casus = similarityMap3D.casus.CASE_2LN_1PL_NOT_PARALLEL_NOR_CONCURRENT
                return similarityMap3D.distinguishes.SUCCESS
                
            #cases: 1 line, 2 planes
            elif last.__class__ == plane3D and first.__class__ == line3D and second.__class__ == plane3D:
                if last.parallelLine(first) and second.parallelLine(first):
                    return similarityMap3D.distinguishes.UNSUITABLE
                if second.parallelLine(first):
                    last, second = second, last
                self.fromSet[0] = second.intersectionLine(first)
                self.fromSet[1] = last.projection(self.fromSet[0])
                if self.fromSet[0] == self.fromSet[1]:
                    return similarityMap3D.distinguishes.UNSUITABLE
                l0 = self.fromSet[1].subtract(self.fromSet[0]).length()
                self.fromSet[2] = self.fromSet[0].add(first.unitDir.scale(l0))
                
                self.subCase = True if last.parallelLine(first) else False                
                self.casus = similarityMap3D.casus.CASE_1LN_2PL_NOT_ALL_CONCURRENT
                return similarityMap3D.distinguishes.SUCCESS

            elif last.__class__ == plane3D and first.__class__ == plane3D:
                if second is None:
                    return similarityMap3D.distinguishes.INSUFFICIENT
                
                # if there are two (or more) set of parallel planes, discard the latest plane: there can only be at 
                # most one pair of parallel planes
                if third is None:
                    if first.parallel(second) and first.parallel(last):
                        return similarityMap3D.distinguishes.UNSUITABLE
                    return similarityMap3D.distinguishes.INSUFFICIENT
                if (second.parallel(third) and second.parallel(last)) or (first.parallel(second) and third.parallel(last)):
                    return similarityMap3D.distinguishes.UNSUITABLE

                # rearrange the planes if there's a parallel so that first parallel second
                # -> first 2 parallel, and both intersect third and last
                # rearraging the distinguishables themselves is not necessary since the parallel property would be checked
                # during the matching process with subcase rather than the distinguishables themselves
                self.subCase = [True]
                if first.parallel(second):
                    pass
                elif second.parallel(third):
                    first, third = third, first
                elif third.parallel(last):
                    first, second, third, last = third, last, first, second
                else:
                    self.subCase[0] = False
                lineSort = sortClass.getSort('__dummyLine3D__')
                p13 = first.intersection(third)
                p14 = first.intersection(last)

                p23 = second.intersection(third)
                p24 = second.intersection(last)

                check_1_34 = p13[0].parallel(p14[0])
                check_2_34 = p23[0].parallel(p24[0])

                if (self.subCase[0] and (check_1_34 or check_2_34)) or (check_1_34 and check_2_34):
                    return similarityMap3D.distinguishes.UNSUITABLE
                elif check_1_34:
                    second, last = last, second
                    p14 = first.intersection(last)
                    p23 = second.intersection(third)
                    p24 = second.intersection(last)
                elif check_2_34:
                    first, last = last, first
                    p13 = first.intersection(third)
                    p14 = first.intersection(last)
                    p24 = second.intersection(last)

                l13 = line3D(lineSort, p13[1], p13[1].add(p13[0]))
                l14 = line3D(lineSort, p14[1], p14[1].add(p14[0]))
                l23 = line3D(lineSort, p23[1], p23[1].add(p23[0]))
                l24 = line3D(lineSort, p24[1], p24[1].add(p24[0]))
                self.fromSet[0] = l13.intersection(l14)
                self.fromSet[1] = l23.intersection(l24)

                distance = self.fromSet[0].subtract(self.fromSet[1]).length()
                self.fromSet[2] = self.fromSet[0].add(third.normal.scale(distance))

                self.casus = similarityMap3D.casus.CASE_4PL_NOT_ALL_PARALLEL
                return similarityMap3D.distinguishes.SUCCESS

        return similarityMap3D.distinguishes.UNSUITABLE

    def assessSkew(self, first, last):
        """
        Assessing the case where there are two line distinguisables that are skew
        """
        n = last.direction.product(first.direction)
        p1 = first.root
        p2 = last.root
        d1 = first.direction
        d2 = last.direction
        n1 = d1.product(n)
        n2 = d2.product(n)
        self.fromSet[0] = p1.add(d1.scale((p2.subtract(p1).dotproduct(n2))/(d1.dotproduct(n2))))
        self.fromSet[1] = p2.add(d2.scale((p1.subtract(p2).dotproduct(n1))/(d2.dotproduct(n1))))
        
        
        self.subCase.append(first.assoc.containsPoint(self.fromSet[0]))
        self.subCase.append(last.assoc.containsPoint(self.fromSet[1]))
        
        d = self.fromSet[1].subtract(self.fromSet[0])
        v = d1.scale((d.dotproduct(d)/(d1.dotproduct(d1)))**c0p5*c2)
        self.fromSet[2] = self.fromSet[0].add(v)
        check = first.assoc.containsPoint(self.fromSet[2])
        
        if check:
            self.subCase.append(check)
        else:
            self.fromSet[2] = self.fromSet[0].subtract(v)
            self.subCase.append(first.assoc.containsPoint(self.fromSet[2]))
            
        #self.checkTree = True
        
        self.casus = similarityMap3D.casus.CASE_2LN_SKEWED
        return similarityMap3D.distinguishes.SUCCESS

    def postAnalysis(self):
        self.reflection = 1
        # if the casus is redundant (so the matching would be a special case) and there is a discarded distinguishable
        if self.casus == self.__class__.casus.REDUNDANT and len(self.discardedDist) > 0:
            # for the case of 1 line and 1 point colinear to the line
            if (len(self.distinguishables) == 1 
            and 'point' in self.distinguishables[0].element.individual.type
            and 'line' in self.discardedDist[-1].element.individual.type):
                self.distinguishables.append(self.discardedDist[-1])
                self.discardedDist = []
            if (len(self.distinguishables) == 2
            and 'point' in self.distinguishables[0].element.individual.type
            and 'point' in self.distinguishables[1].element.individual.type):
                self.discardedDist = []

    # ============================================================================================
    #                                    SHAPE MATCHING PROCESS
    # ============================================================================================
    
    def map(self, fromInd, toInd):
        """
        Try to map an object from the shape to a lhs object and add it as distinguishable counterpart if they can be mapped
        
        :param fromInd: lhs object
        :type fromInd: ~sortal.ind.individual.individual
        :param toInd: shape object to be mapped
        :type toInd: ~sortal.ind.individual.individual
        :returns: boolean -- whether the two objects can be mapped
        """
        # print('\n\n.....', toInd.type, toInd)
        # print('\n\n.....', toInd.type, toInd)
        for i in self.distinguishables:
            elmInd = i.element.individual
            if elmInd == fromInd:
                if elmInd.hasAttribute():
                    if fromInd.hasAttribute() and elmInd.attribute.checkPartOf(toInd.attribute):
                        i.addCounterpart(toInd)
                        return True
                else:
                    i.addCounterpart(toInd)
                    return True
        return False
        
    def potentialTransforms(self): ########################
        """
        Detecting all possible matches in the shape based on the lhs casus and distinguishables
        """
        # print('CASUS:', self.casus)
        if len(self.transforms) != 0:
            return self.transforms
            
        if self.casus == similarityMap3D.casus.REDUNDANT:
            return self.specialPotentialTransforms()
            
        size = len(self.distinguishables)
        source = []
        counterparts = [] 
                        
        for elm in self.distinguishables:
            source.append(elm.element)
            counterparts.append(elm.getCounterparts())
        target = [None]*size
        for e0 in range(len(counterparts[0])):
            target[0] = counterparts[0][e0]
            if size > 1:
                for e1 in range(len(counterparts[1])):
                    if counterparts[1][e1].individual == target[0].individual:
                        continue
                    target[1] = counterparts[1][e1]
                    if not self.assessTargetSet(target, 2):
                        continue
                    if size > 2:
                        for e2 in range(len(counterparts[2])):
                            target[2] = counterparts[2][e2]
                            if size > 3:
                                if not self.assessTargetSet(target, 3):
                                    continue
                                for e3 in range(len(counterparts[3])):
                                    target[3] = counterparts[3][e3]
                                    self.detectTransforms(source, target)
                            else:
                                self.detectTransforms(source, target)
                    else:
                        self.detectTransforms(source, target)
        return self.transforms
    
    def assessTargetSet(self, target, size):
        """
        Checking if a set of distinguishables from the shape is valid for matching (no duplicate individuals, ...)
        
        :param target: list of counterpart distinguishables that are going to be used for matching
        :type target: list of :class:`~sortal.map.distinguishable.distinguishable`
        :returns: boolean -- validity of the target set of distinguishables
        """
        if self.casus == similarityMap3D.casus.CASE_3PT_NOT_COLINEAR:
            first = target[0].individual
            second = target[1].individual
            if first == second:
                return False
            return True
        elif self.casus == similarityMap3D.casus.CASE_2PT_1PL_NOT_PERPENDICULAR:
            first = target[0].individual
            second = target[1].individual
            if first.type == second.type and first == second:
                return False
        elif self.casus == similarityMap3D.casus.CASE_1PT_1LN_1PL_NOT_COINCIDENT_NOT_PERPENDICULAR:
            first = target[0].individual
            second = target[1].individual
            if not second.containsPoint(first.value):
                return False
        elif self.casus == similarityMap3D.casus.CASE_3LN_NO_TWO_PARALLEL:
            first = target[0].individual
            second = target[1].individual
            if first.parallel(second):
                return False
            return True
        elif self.casus == similarityMap3D.casus.CASE_3LN_FIRST_TWO_PARALLEL:
            first = target[0].individual
            second = target[1].individual
            if not first.parallel(second):
                return False
            if second.compareValue(first) == 0:
                return False
            return True
        elif self.casus == similarityMap3D.casus.CASE_3LN_LAST_TWO_PARALLEL:
            first = target[0].individual
            second = target[1].individual
            if first.parallel(second):
                return False
            return True
        elif self.casus == similarityMap3D.casus.CASE_2LN_1PL_PARALLEL_BOTH:
            first = target[0].individual
            second = target[1].individual
            if first.parallel(second):
                return False
            return True
        elif self.casus == similarityMap3D.casus.CASE_1LN_2PL_NOT_ALL_CONCURRENT:
            first = target[0].individual
            second = target[1].individual
            if second.parallelLine(first):
                return False
        elif self.casus == similarityMap3D.casus.CASE_4PL_NOT_ALL_PARALLEL:
            first = target[0].individual
            second = target[1].individual
            if size == 3:
                third = target[2].individual
                if third.parallel(first) or third.parallel(second):
                    return False
            elif self.subCase[0] != second.parallel(first):
                return False
        return True

    def detectTransforms(self, source, target):
        """
        Finding the transformation matrix to transform the individuals in the source (lhs distinguishables) 
        into target (shape distinguishables) based on the mapping casus
        
        :param source: list of lhs distinguishable for matching
        :type source: list of :class:`~sortal.map.distinguishable.distinguishable`
        :param target: list of shape distinguishables that are counterparts of those in source for matching
        :type target: list of :class:`~sortal.map.distinguishable.distinguishable`
        """
        size = len(source)
        last = target[-1].individual
        if size > 1:
            first = target[0].individual
            if size > 2:
                second = target[1].individual
                if size > 3:
                    third = target[2].individual
        toSet = [None, None, None]
        
        # 1 point + 1 line, not colinear
        # matching vectors: [point projection on line, 
        #                    point,
        #                    a point on the plane's normal that is the same distance away from the projection as point]
        if self.casus == similarityMap3D.casus.CASE_1PT_1LN_NOT_COLINEAR:
            toSet[1] = first.value
            u = last.direction
            v = toSet[1].subtract(last.root)

            if u.parallel(v):
                return                
            toSet[0] = last.root.add(u.scale(v))
            
            if toSet[0].subtract(toSet[1]).squareLength() == 0:
                return
            
            for i in target[-1].forms[target[-1].individual.sort.name]:
                check1 = False
                if (i.containsPoint(toSet[0]) >= self.subCase[0] and
                    (not(i.start == toSet[0] or i.end == toSet[0])) >= self.subCase[1]):
                    check1 = True
                    break
            if check1:
                d = toSet[1].subtract(toSet[0])
                
                toSet[2] = toSet[0].add(u.scale((d.dotproduct(d)/u.dotproduct(u))**c0p5*c2))
                
                for i in target[-1].forms[target[-1].individual.sort.name]:
                    if i.containsPoint(toSet[2]) >= self.subCase[2]:
                        tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection)
                        for tf in tfs:
                            if tf is not None:
                                self.transforms.append(tf)
                        break
    
                        
                toSet[2] = toSet[0].subtract(u.scale((d.dotproduct(d)/u.dotproduct(u))**c0p5*c2))
                for i in target[-1].forms[target[-1].individual.sort.name]:
                    if i.containsPoint(toSet[2]) >= self.subCase[2]:
                        tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection)
                        for tf in tfs:
                            if tf is not None:
                                self.transforms.append(tf)
                        break

        # 1 point + 1 line + 1 plane, point lies on line; line and plane are not coincident and not perpendicular
        # matching vectors: [point, 
        #                    point projection on plane,
        #                    line intersection with plane] (if line and plane are not parallel)
        #                   [point, 
        #                    point projection on plane,
        #                    a point on the plane that is the same distance away from the projection as point following the line's direction] (if line and plane are parallel)
        if self.casus == similarityMap3D.casus.CASE_1PT_1LN_1PL_NOT_COINCIDENT_NOT_PERPENDICULAR:
            if last.containsLine(second) or last.perpendicularLine(second) or self.subCase[0] != last.parallelLine(second):
                return
            toSet[0] = first.value
            toSet[1] = last.projection(first.value)
            # if line parallel plane
            if self.subCase[0]:
                PtPlDistance = toSet[0].subtract(toSet[1]).length()

                # positive direction
                toSet[2] = first.value.add(second.unitDir.scale(PtPlDistance))
                tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection, True)

                for tf in tfs:
                    if tf is not None:
                        self.transforms.append(tf)
                
                # negative direction
                toSet[2] = first.value.subtract(second.unitDir.scale(PtPlDistance))
                tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection, True)
                for tf in tfs:
                    if tf is not None:
                        self.transforms.append(tf)

            # if line intersects plane
            else:
                toSet[2] = last.intersectionLine(second)
                tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection, True)
                for tf in tfs:
                    if tf is not None:
                        self.transforms.append(tf)

            return

        # 3 points, not colinear
        # matching vectors: [point 1, 
        #                    point 2,
        #                    point 3]
        elif self.casus == similarityMap3D.casus.CASE_3PT_NOT_COLINEAR:
            if last == first or last == second or first == second:
                return
            if last.value.colinear(first.value, second.value):
                return
            toSet[0] = first.value
            toSet[1] = second.value
            toSet[2] = last.value

            tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection, True)
            for tf in tfs:
                if tf is not None:
                    self.transforms.append(tf)
            

        # 1 point + 1 arc, point must not be on the arc
        # matching vectors: [point, 
        #                    arc's center,
        #                    a point on the arc's normal that is the same distance away from arc's center as point]
        elif self.casus == similarityMap3D.casus.CASE_1PT_1ARC_NOT_COINCIDENT:
            if last.__class__ != self.subCase[0]:
                return
            lastCenter = last.center
            if first.value == lastCenter:
                return
                
            toSet[0] = first.value
            toSet[1] = lastCenter

            u = last.normal
            v = first.value.subtract(lastCenter)
            
            if u.parallel(v):
                return
            
            cosAngle1 = u.unit().dotproduct(v.unit())    
            check1 = check2 = False
            if cosAngle1 == self.cosAngle:
                check1 = True
                if cosAngle1 == 0:
                    u1 = u.scale(-1)
                    check2 = True
            else:
                u1 = u.scale(-1)
                cosAngle2 = u1.unit().dotproduct(v.unit())
                if cosAngle2 == self.cosAngle:
                    check2 = True
                    
            if cosAngle1 == 0:
                for i in target[-1].forms[target[-1].individual.sort.name]:
                    containsCheck = False
                    if i.containsPoint(toSet[0]) >= self.subCase[1] :
                        containsCheck = True
                        break
                if not containsCheck:
                    return
                                
            if check1:
                toSet[2] = toSet[1].add(u.scale((v.dotproduct(v)/u.dotproduct(u))**c0p5*c2))
                tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection)
                for tf in tfs:
                    if tf is not None:
                        self.transforms.append(tf)

            if check2:
                toSet[2] = toSet[1].add(u1.scale((v.dotproduct(v)/u1.dotproduct(u1))**c0p5*c2))
                tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection)
                for tf in tfs:
                    if tf is not None:
                        self.transforms.append(tf)
        # 2 arcs, not concentric
        # matching vectors: [1st arc's center, 
        #                    2nd arc's center,
        #                    a point on the 2nd arc's normal that is the same distance away from 2nd arc's center as 1st arc's center]
        elif self.casus == similarityMap3D.casus.CASE_2ARC_NOT_CONCENTRIC:
            if first.__class__ != self.subCase[0]:
                return
            if last.__class__ != self.subCase[1]:
                return
            firstCenter = first.center
            lastCenter = last.center
            
            toSet[0] = firstCenter
            toSet[1] = lastCenter
            u = last.normal
            v = firstCenter.subtract(lastCenter)
            
            if u.parallel(v):
                return
            
            toSet[2] = self.fromSet[1].add(u.scale((v.dotproduct(v)/u.dotproduct(u))**c0p5*c2))
            
            if first.__class__ == circle3D:
                firstRadius = first.radius
            else:
                firstRadius = first.radius[0]
            
            if last.__class__ == circle3D:
                lastRadius = last.radius
            else:
                lastRadius = last.radius[0]
            
            if not (firstRadius / lastRadius) == self.ratio[0]:
                return
                
            distanceVec = toSet[0].subtract(toSet[1])
            distanceSqr = distanceVec.dotproduct(distanceVec)
            
            if not ((firstRadius**2) / distanceSqr) == self.ratio[1]:
                return
                
            cosAngle1 = u.unit().dotproduct(v.unit())
            
            if cosAngle1 == self.cosAngle:
                check1 = True
                if cosAngle1 == 0:
                    u1 = u.scale(-1)
                    check2 = True
            else:
                u1 = u.scale(-1)
                cosAngle2 = u1.unit().dotproduct(v.unit())
                if cosAngle2 == self.cosAngle:
                    check2 = True
                                
            if check1:
                toSet[2] = toSet[1].add(u.scale((v.dotproduct(v)/u.dotproduct(u))**c0p5*c2))
                tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection)
                for tf in tfs:
                    if tf is not None:
                        self.transforms.append(tf)

            if check2:
                toSet[2] = toSet[1].add(u1.scale((v.dotproduct(v)/u1.dotproduct(u1))**c0p5*c2))
                tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection)
                for tf in tfs:
                    if tf is not None:
                        self.transforms.append(tf)
                        
        # 2 arcs, not parallel (must be concentric, otherwise the casus would be CASE_2ARC_NOT_CONCENTRIC)
        # matching vectors: [1st arc's center, 
        #                    a point on the 1st arc's normal that is 1st arc's radius away from 1st arc's center,
        #                    a point on the 2nd arc's normal that is 2nd arc's radius away from 2nd arc's center]
        # *if any arc is an ellipse, the radius used will be the big radius
        elif self.casus == similarityMap3D.casus.CASE_2ARC_NOT_PARALLEL:
            if first.__class__ != self.subCase[0]:
                return
            if last.__class__ != self.subCase[1]:
                return            
            firstCenter = first.center

            if not firstCenter == last.center:
                return
            if first.normal.parallel(last.normal):
                return
            toSet[0] = firstCenter
            
            if first.__class__ == circle3D:
                firstRadius = first.radius
            else:
                firstRadius = first.radius[0]
            
            if last.__class__ == circle3D:
                lastRadius = last.radius
            else:
                lastRadius = last.radius[0]
            
            if not (firstRadius / lastRadius) == self.ratio :
                if (lastRadius / firstRadius) == self.ratio:
                    first, last = last, first
                    firstRadius, lastRadius = lastRadius, firstRadius
                else:
                    return

            firstUnit = first.normal.unit()
            lastUnit = last.normal.unit()
            revFirstUnit = firstUnit.scale(-1)
            revLastUnit = lastUnit.scale(-1)
            
            cosAngle1 = firstUnit.dotproduct(lastUnit)
            
            toSet[1] = firstUnit.scale(firstRadius)
            toSet[2] = lastUnit.scale(lastRadius)

            check1 = False
            check2 = False
            if cosAngle1 == self.cosAngle:
                check1 = True
                if cosAngle1 == 0:
                    check2 = True
            else:
                cosAngle2 = revFirstUnit.dotproduct(lastUnit)
                if cosAngle2 == self.cosAngle:
                    check2 = True
                    
            if check1:
                toSet[1] = toSet[0].add(firstUnit.scale(firstRadius))
                toSet[2] = toSet[0].add(lastUnit.scale(lastRadius))
                tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection, True)
                for tf in tfs:
                    if tf is not None:
                        self.transforms.append(tf)
                        
                toSet[1] = toSet[0].add(revFirstUnit.scale(firstRadius))
                toSet[2] = toSet[0].add(revLastUnit.scale(lastRadius))
                tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection, True)
                for tf in tfs:
                    if tf is not None:
                        self.transforms.append(tf)

            if check2:
                toSet[1] = toSet[0].add(revFirstUnit.scale(firstRadius))
                toSet[2] = toSet[0].add(lastUnit.scale(lastRadius))
                tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection, True)
                for tf in tfs:
                    if tf is not None:
                        self.transforms.append(tf)
                        
                toSet[1] = toSet[0].add(firstUnit.scale(firstRadius))
                toSet[2] = toSet[0].add(revLastUnit.scale(lastRadius))
                tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection, True)
                for tf in tfs:
                    if tf is not None:
                        self.transforms.append(tf)
                        
        # 1 arc + 1 line, the arc's center must not be colinear to the line
        # matching vectors: [arc's center, 
        #                    projection of arc's center on the line,
        #                    a point on the line that is the same distance away from the projection point as the arc's center]
        if self.casus == similarityMap3D.casus.CASE_1ARC_1LN_NOT_COLINEAR:
            toSet[1] = first.center
            u = last.direction
            v = toSet[1].subtract(last.root)

            if u.parallel(v):
                return                
            toSet[0] = last.root.add(u.scale(v))
            
            if toSet[0].subtract(toSet[1]).squareLength() == 0:
                return
            
            for i in target[-1].forms[target[-1].individual.sort.name]:
                check1 = False
                if (i.containsPoint(toSet[0]) >= self.subCase[0] and
                    (not(i.start == toSet[0] or i.end == toSet[0])) >= self.subCase[1]):
                    check1 = True
                    break
            if check1:
                d = toSet[1].subtract(toSet[0])
                
                toSet[2] = toSet[0].add(u.scale((d.dotproduct(d)/u.dotproduct(u))**c0p5*c2))
                
                for i in target[-1].forms[target[-1].individual.sort.name]:
                    if i.containsPoint(toSet[2]) >= self.subCase[2]:
                        tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection)
                        for tf in tfs:
                            if tf is not None:
                                self.transforms.append(tf)
                        break
    
                        
                toSet[2] = toSet[0].subtract(u.scale((d.dotproduct(d)/u.dotproduct(u))**c0p5*c2))
                for i in target[-1].forms[target[-1].individual.sort.name]:
                    if i.containsPoint(toSet[2]) >= self.subCase[2]:
                        tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection)
                        for tf in tfs:
                            if tf is not None:
                                self.transforms.append(tf)
                        break
        # 3 lines, all 3 lines intersect one another
        # matching vectors: [line 1 intersection with line 3, 
        #                    line 1 intersection with line 2,
        #                    line 2 intersection with line 3]
        elif self.casus == similarityMap3D.casus.CASE_3LN_NO_TWO_PARALLEL:
            if (last.parallel(first) or second.parallel(first) or second.parallel(last)):
                return
            try:
                is1 = first.intersection(last)
                is2 = first.intersection(second)
                is3 = second.intersection(last)
            except:
                return
            
            if is1 == is2 or is1 == is3 or is2 == is3:
                return
                
            u = is2.subtract(is1).squareLength()
            v = is3.subtract(is2).squareLength()
            w = is3.subtract(is1).squareLength()
            if (u/self.ratio[0] == v/self.ratio[1] and u/self.ratio[0] == w/self.ratio[2]):
                toSet = [is1,is2,is3,target[0],target[1],target[2]]
                for i in toSet[3].forms[toSet[3].individual.sort.name]:
                    if (i.containsPoint(toSet[0]) >= self.subCase[0] and
                        i.containsPoint(toSet[1]) >= self.subCase[1]):
                            check = True
                            break
                    check = False
                if not check:
                    return
                for i in toSet[4].forms[toSet[3].individual.sort.name]:
                    if (i.containsPoint(toSet[1]) >= self.subCase[2] and
                        i.containsPoint(toSet[2]) >= self.subCase[3]):
                            check = True
                            break
                    check = False
                if not check:
                    return
                for i in toSet[5].forms[toSet[3].individual.sort.name]:
                    if (i.containsPoint(toSet[0]) >= self.subCase[4] and
                        i.containsPoint(toSet[2]) >= self.subCase[5]):
                            check = True
                            break
                    check = False
                if not check:
                    return
                tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection)
                for tf in tfs:
                    if tf is not None:
                        self.transforms.append(tf)
                        
        # 3 lines, first 2 lines parallel one another and intersect the last line
        # matching vectors: [line 1 intersection with line 3 (intersection 1), 
        #                    line 2 intersection with line 3 (intersection 2),
        #                    point on line 1 that is at the same distance from intersection 1 as intersection 2]
        elif self.casus == similarityMap3D.casus.CASE_3LN_FIRST_TWO_PARALLEL:
            if (last.parallel(first) or second.parallel(last)) or not second.parallel(first):
                return
            if second == first:
                return
                
            u = first.unitDir
            v = last.unitDir
            cosA = abs(u.dotproduct(v))
            if not self.cosAngle ==  cosA:
                return
            
            try:
                toSet[0] = first.intersection(last)
                toSet[1] = second.intersection(last)
            except ValueError:
                return
            
            for i in target[0].forms[target[0].individual.sort.name]:
                if i.containsPoint(toSet[0]) >= self.subCase[0]:
                    check1 = True
                    break
                check1 = False
            for i in target[1].forms[target[1].individual.sort.name]:
                if i.containsPoint(toSet[1]) >= self.subCase[1]:
                    check2 = True
                    break
                check2 = False
            for i in target[2].forms[target[2].individual.sort.name]:
                if (i.containsPoint(toSet[0]) >= self.subCase[2] and
                    i.containsPoint(toSet[1]) >= self.subCase[3]):
                    check3 = True
                    break
                check3 = False

            if check1 and check2 and check3:        
                
                u = first.direction
                d = toSet[1].subtract(toSet[0])
                v = u.scale((d.dotproduct(d)/(u.dotproduct(u)))**c0p5*c2)
                toSet[2] = toSet[0].add(v)
                
                for i in target[0].forms[target[0].individual.sort.name]:
                    if i.containsPoint(toSet[2]) >= self.subCase[4]:                
                        tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection)
                        for tf in tfs:
                            if tf is not None:
                                self.transforms.append(tf) 
                        break
                    
                toSet[2] = toSet[0].subtract(v)
                for i in target[0].forms[target[0].individual.sort.name]:
                    if i.containsPoint(toSet[2]) >= self.subCase[4]:                
                        tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection)
                        for tf in tfs:
                            if tf is not None:
                                self.transforms.append(tf)
                        break

            
        # 2 skew lines
        # matching vectors: [point on line 1 that is closest to line 2 (point1), 
        #                    point on line 2 that is closest to line 1 (point2),
        #                    point on line 1 that is at the same distance from point1 as point2]
        elif self.casus == similarityMap3D.casus.CASE_2LN_SKEWED:
            if first.parallel(last):
                return
            elif not first.skew(last):
                return
            n = last.direction.product(first.direction)
            p1 = first.root
            p2 = last.root
            d1 = first.direction
            d2 = last.direction
            n1 = d1.product(n)
            n2 = d2.product(n)
            toSet[0] = p1.add(d1.scale((p2.subtract(p1).dotproduct(n2))/(d1.dotproduct(n2))))
            toSet[1] = p2.add(d2.scale((p1.subtract(p2).dotproduct(n1))/(d2.dotproduct(n1))))
            
            check1 = False
            for i in target[0].forms[target[0].individual.sort.name]:
                check = i.containsPoint(toSet[0])
                if check >= self.subCase[0]:
                    check1 = True
                    
            check2 = False
            for i in target[1].forms[target[1].individual.sort.name]:
                check = i.containsPoint(toSet[1])
                if check >= self.subCase[1]:
                    check2 = True
                       

            if not (check1 and check2):
                return
            
            d = toSet[1].subtract(toSet[0])
            
            allToSet = []
            
            if check1 and check2:
                v1 = d1.scale((d.dotproduct(d)/(d1.dotproduct(d1)))**c0p5*c2)
                toSet2_1 = toSet[0].add(v1)
                toSet2_2 = toSet[0].subtract(v1)
                for i in target[0].forms[target[0].individual.sort.name]:
                    if i.containsPoint(toSet2_1) >= self.subCase[2]:
                        allToSet.append([toSet[0], toSet[1], toSet2_1])
                        break
                for i in target[0].forms[target[0].individual.sort.name]:
                    if i.containsPoint(toSet2_2) >= self.subCase[2]:
                        allToSet.append([toSet[0], toSet[1], toSet2_2])
                        break
                    
            for i in allToSet:
                tfs = affineTransform3D.similarityTransform(self.fromSet, i, self.reflection)
                for tf in tfs:
                    if tf is not None:
                        self.transforms.append(tf)


        # 2 points + 1 plane, the 2 points must not form a line that is perpendicular to the plane
        # matching vectors: [point 1, 
        #                    point 2,
        #                    projection of point 1 on the plane]
        elif self.casus == similarityMap3D.casus.CASE_2PT_1PL_NOT_PERPENDICULAR:
            foot = last.projection(first.value)
            if foot == first: 
                if self.subCase == 1:
                    return
                else:
                    foot = last.projection(second.value)
                    if foot == second:
                        if self.subCase == 2:
                            return
                        else:
                            d = second.subtract(first)
                            u = last.normal.product(d)
                            foot = first.add(u.scale((d.dotProduct(d) / (u.dotProduct(u)))**c0p5*c2))

            if (foot.colinear(second.value, first.value)):
                return
                
            toSet[0] = first.value
            toSet[1] = second.value
            toSet[2] = foot
            
            tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection)
            for tf in tfs:
                if tf is not None:
                    self.transforms.append(tf)


        # 2 lines + 1 plane, the 2 lines must intersect one another but parallel to the plane
        # matching vectors: [intersection of the two lines, 
        #                    projection of the intersection onto the plane,
        #                    point on the plane on a line with perpendicular direction to the first line and same distance from the projection as the intersection]
        elif self.casus == similarityMap3D.casus.CASE_2LN_1PL_PARALLEL_BOTH:
            if not (last.parallelLine(first) and last.parallelLine(second)):
                return
                
            toSet[0] = first.intersection(second)
            toSet[1] = last.projection(toSet[0])
            d0 = last.normal.product(first.unitDir)
            l0 = toSet[1].subtract(toSet[0]).length()
            
            toSet[2] = toSet[1].add(d0.scale(l0))
            tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection)
            for tf in tfs:
                if tf is not None:
                    self.transforms.append(tf)
                    
            toSet[2] = toSet[1].subtract(d0.scale(l0))
            tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection)
            for tf in tfs:
                if tf is not None:
                    self.transforms.append(tf)
        # 2 lines + 1 plane, the first line parallel to the plane while the second line intersect the plane, the two lines cannot be skew
        # matching vectors: [intersection of the two lines, 
        #                    intersection of the second line and the plane,
        #                    projection of the second line - plane intersection onto the first line]
        elif self.casus == similarityMap3D.casus.CASE_2LN_1PL_PARALLEL_FIRST:
            if (not last.parallelLine(first)) or last.parallelLine(second) or first.skew(second):
                return
            toSet[0] = last.intersectionLine(second)
            toSet[1] = first.intersection(second)
            l0 = toSet[1].subtract(toSet[0]).length()
            
            toSet[2] = toSet[1].add(first.unitDir.scale(l0))
            tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection)
            for tf in tfs:
                if tf is not None:
                    self.transforms.append(tf)
                    
            toSet[2] = toSet[1].subtract(first.unitDir.scale(l0))
            tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection)
            for tf in tfs:
                if tf is not None:
                    self.transforms.append(tf)
        # 2 lines + 1 plane, the plane intersects both lines
        # matching vectors: [intersection of the first line and the plane, 
        #                    intersection of the second line and the plane,
        #                    point on the plane normal at the same distance away from first line-plane intersection as the other intersection]
        elif self.casus == similarityMap3D.casus.CASE_2LN_1PL_NOT_PARALLEL_NOR_CONCURRENT:
            if last.parallelLine(first) or last.parallelLine(second):
                return
            toSet[0] = last.intersectionLine(first)
            toSet[1] = last.intersectionLine(second)
            
            if toSet[0] == toSet[1]:
                return
                
            toSet[2] = toSet[0].add(last.normal.scale(toSet[0].subtract(toSet[1]).length()))
            tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection)
            for tf in tfs:
                if tf is not None:
                    self.transforms.append(tf)
                    
            toSet[2] = toSet[0].subtract(last.normal.scale(toSet[0].subtract(toSet[1]).length()))
            tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection)
            for tf in tfs:
                if tf is not None:
                    self.transforms.append(tf)
                    
        # 1 lines + 2 planes, not all concurrent
        # matching vectors: [intersection of the line and the first plane, 
        #                    projection of the intersection onto the second plane,
        #                    point on the line at the same distance away from the intersection as the projection]
        elif self.casus == similarityMap3D.casus.CASE_1LN_2PL_NOT_ALL_CONCURRENT:
            if last.parallelLine(first) != self.subCase:
                return

            toSet[0] = second.intersectionLine(first)
            toSet[1] = last.projection(toSet[0])
                
            if toSet[0] == toSet[1]:
                return similarityMap3D.distinguishes.UNSUITABLE
            l0 = toSet[1].subtract(toSet[0]).length()
            
            toSet[2] = toSet[0].add(first.unitDir.scale(l0))
            tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection)
            for tf in tfs:
                if tf is not None:
                    self.transforms.append(tf)
                    
            toSet[2] = toSet[0].subtract(first.unitDir.scale(l0))
            tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection)
            for tf in tfs:
                if tf is not None:
                    self.transforms.append(tf)
        # 1 point + 2 planes, not all concurrent
        # matching vectors: [point, 
        #                    point projection onto the first plane,
        #                    point projection onto the second plane]
        elif self.casus == similarityMap3D.casus.CASE_1PT_2PL_NOT_CONCURRENT:
            if last.parallel(second):
                return similarityMap3D.distinguishes.UNSUITABLE

            toSet[0] = first.value
            toSet[1] = second.projection(first.value)
            toSet[2] = last.projection(first.value)

            if toSet[0] == toSet[1] or toSet[0] == toSet[2] or toSet[1] == toSet[2]:
                return
            tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection)
            for tf in tfs:
                if tf is not None:
                    self.transforms.append(tf)

        elif self.casus == similarityMap3D.casus.CASE_4PL_NOT_ALL_PARALLEL:
            if last.parallel(first) or last.parallel(second)or last.parallel(third):
                return

            lineSort = sortClass.getSort('__dummyLine3D__')

            p13 = first.intersection(third)
            p14 = first.intersection(last)


            p23 = second.intersection(third)
            p24 = second.intersection(last)

            if p13[0].parallel(p14[0]) or p23[0].parallel(p24[0]):
                return

            l13 = line3D(lineSort, p13[1], p13[0].add(p13[1]))
            l14 = line3D(lineSort, p14[1], p14[0].add(p14[1]))
            l23 = line3D(lineSort, p23[1], p23[0].add(p23[1]))
            l24 = line3D(lineSort, p24[1], p24[0].add(p24[1]))

            toSet[0] = l13.intersection(l14)
            toSet[1] = l23.intersection(l24)

            distance = toSet[0].subtract(toSet[1]).length()
            toSet[2] = toSet[0].add(third.normal.scale(distance))

            tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection)
            for tf in tfs:
                if tf is not None:
                    self.transforms.append(tf)

            toSet[2] = toSet[0].subtract(third.normal.scale(distance))

            tfs = affineTransform3D.similarityTransform(self.fromSet, toSet, self.reflection)
            for tf in tfs:
                if tf is not None:
                    self.transforms.append(tf)
        return
        
    def specialPotentialTransforms(self):
        """
        Detect transformation for mapping that has redundant casus (not enough infomation 
        to be classified as a supported casus), matching can still be performed albeit less accurate.

        Valid special cases:    

        * 1 single point 
        * 1 single line
        * 2 colinear points
        * 2 parallel lines
        * 1 line and 1 colinear point
        """
        if len(self.distinguishables) == 1:
            dist = self.distinguishables[0]

            # One single point -> only translation
            if dist.element.individual.type == 'ind-point3D':
                # if a line is discarded, use the line as distinguishable instead
                if len(self.discardedDist) > 0 and self.discardedDist[-1].element.individual.type == 'ind-line3D':
                    dist = self.discardedDist[-1]
                else:
                    if self.fromSet[0] is None:
                        self.fromSet[0] = dist.element.individual.value
                    for i in dist.counterparts:
                        toVect = i.individual.value
                        self.transforms.append(affineTransform3D.translation(toVect.subtract(self.fromSet[0])))
            
            # One single line
            # matching vectors: [line.start, line.end]
            # *limited transform, only use 2 vectors for finding transformation
            if dist.element.individual.type == 'ind-line3D':
                for sortName in dist.element.forms:
                    sortLines = dist.element.forms[sortName]
                    for baseLine in sortLines:
                        if self.fromSet[0] is None:
                            self.fromSet = [baseLine.start, baseLine.end]
                        for i in dist.counterparts:
                            for l in i.forms[sortName]:
                                toVects = [l.start, l.end]
                                tfs = affineTransform3D.limitedTransform(self.fromSet, toVects)
                                for tf in tfs:
                                    if tf is not None:
                                        self.transforms.append(tf)
                                toVects = [l.end, l.start]
                                tfs = affineTransform3D.limitedTransform(self.fromSet, toVects)
                                for tf in tfs:
                                    if tf is not None:
                                        self.transforms.append(tf)
        if len(self.distinguishables) == 2:
            dist0 = self.distinguishables[0]
            dist1 = self.distinguishables[1]

            # 2 points, not concurrent
            # matching vectors: [point 1, point 2]
            # *limited transform, only use 2 vectors for finding transformation
            if dist0.element.individual.type == dist1.element.individual.type == 'ind-point3D':
                if self.fromSet[0] is None:
                    self.fromSet = [dist0.element.individual.value, dist1.element.individual.value]
                for i in dist0.counterparts:
                    for j in dist1.counterparts:
                        if i == j:
                            continue
                        toVects = [i.individual.value, j.individual.value]

                        tfs = affineTransform3D.limitedTransform(self.fromSet, toVects)
                        for tf in tfs:
                            if tf is not None:
                                self.transforms.append(tf)

            # 2 lines, not skew
            # matching vectors: [first line start, 
            #                    first line end,
            #                    second line start]
            elif dist0.element.individual.type == dist1.element.individual.type == 'ind-line3D':
                if dist1.element.individual.parallel(dist0.element.individual):
                    if self.fromSet[0] is None:
                        self.fromSet = []
                        self.fromSet.append(dist0.element.forms[dist0.element.individual.assoc.sort.name][0].start)
                        self.fromSet.append(dist1.element.forms[dist1.element.individual.assoc.sort.name][0].start)
                        self.fromSet.append(dist1.element.forms[dist1.element.individual.assoc.sort.name][0].end)
                    for i in dist0.counterparts:
                        for j in dist1.counterparts:
                            if i.individual == j.individual:
                                continue
                            if not i.individual.parallel(j.individual):
                                continue
                            for line0 in i.forms[dist0.element.individual.assoc.sort.name]:
                                for line1 in j.forms[dist1.element.individual.assoc.sort.name]:
                                    toVects = [line0.start,line1.start,line1.end]
                                    tfs = affineTransform3D.similarityTransform(self.fromSet, toVects, 1)
                                    for tf in tfs:
                                        if tf is not None:
                                            self.transforms.append(tf)
                                    toVects = [line0.end,line1.end,line1.start]
                                    tfs = affineTransform3D.similarityTransform(self.fromSet, toVects, 1)
                                    for tf in tfs:
                                        if tf is not None:
                                            self.transforms.append(tf)
                                            
                else:
                    if self.fromSet[0] is None:
                        self.fromSet = [dist0.element.individual.intersection(dist1.element.individual), None, None]
                        checkContains = False
                        sqrlength = 0

                        for i in dist0.element.forms[dist0.element.individual.assoc.sort.name]:
                            if i.containsPoint(self.fromSet[0]):
                                checkContains = True

                                sl = i.start.subtract(self.fromSet[0]).squareLength()
                                if sqrlength == 0 or sqrlength > sl > 0:
                                    self.fromSet[1] = i.start
                                    sqrlength = sl

                                sl = i.end.subtract(self.fromSet[0]).squareLength()
                                if sqrlength == 0 or sqrlength > sl > 0:
                                    self.fromSet[1] = i.end
                                    sqrlength = sl
                                break
                        for i in dist1.element.forms[dist1.element.individual.assoc.sort.name]:
                            if i.containsPoint(self.fromSet[0]):
                                checkContains = True

                                sl = i.start.subtract(self.fromSet[0]).squareLength()
                                if sqrlength == 0 or sqrlength > sl > 0:
                                    self.fromSet[1] = i.start
                                    sqrlength = sl
                                    dist0, dist1 = dist1, dist0

                                sl = i.end.subtract(self.fromSet[0]).squareLength()
                                if sqrlength == 0 or sqrlength > sl > 0:
                                    self.fromSet[1] = i.end
                                    sqrlength = sl
                                    dist0, dist1 = dist1, dist0
                                break

                        if not checkContains:
                            for i in dist0.element.forms[dist0.element.individual.assoc.sort.name]:
                                sl = i.start.subtract(self.fromSet[0]).squareLength()
                                if sqrlength == 0 or sqrlength > sl > 0:
                                    self.fromSet[1] = i.start
                                    sqrlength = sl

                                sl = i.end.subtract(self.fromSet[0]).squareLength()
                                if sqrlength == 0 or sqrlength > sl > 0:
                                    self.fromSet[1] = i.end
                                    sqrlength = sl
                            swapDist = False
                            for i in dist1.element.forms[dist1.element.individual.assoc.sort.name]:
                                sl = i.start.subtract(self.fromSet[0]).squareLength()
                                if sqrlength == 0 or sqrlength > sl > 0:
                                    self.fromSet[1] = i.start
                                    sqrlength = sl
                                    swapDist = True

                                sl = i.end.subtract(self.fromSet[0]).squareLength()
                                if sqrlength == 0 or sqrlength > sl > 0:
                                    self.fromSet[1] = i.end
                                    sqrlength = sl
                                    swapDist = True
                            if swapDist is True:
                                dist0, dist1 = dist1, dist0
                        distance = self.fromSet[1].subtract(self.fromSet[0]).length()
                        self.fromSet[2] = self.fromSet[0].add(dist1.element.individual.unitDir.scale(distance))

                    for i in dist0.counterparts:
                        for j in dist1.counterparts:
                            if not i.individual.intersect(j.individual):
                                continue
                            toVects = [i.individual.intersection(j.individual), None, None]
                            for line0 in i.forms[dist0.element.individual.assoc.sort.name]:
                                if not toVects[0] == line0.start:
                                    toVects[1] = line0.start
                                    distance = toVects[1].subtract(toVects[0]).length()
                                    toVects[2] = toVects[0].add(j.individual.unitDir.scale(distance))
                                    tfs = affineTransform3D.similarityTransform(self.fromSet, toVects, 1)
                                    for tf in tfs:
                                        if tf is not None:
                                            self.transforms.append(tf)
                                    toVects[2] = toVects[0].subtract(j.individual.unitDir.scale(distance))
                                    tfs = affineTransform3D.similarityTransform(self.fromSet, toVects, 1)
                                    for tf in tfs:
                                        if tf is not None:
                                            self.transforms.append(tf)
                                if not toVects[0] == line0.end:
                                    toVects[1] = line0.end
                                    distance = toVects[1].subtract(toVects[0]).length()
                                    toVects[2] = toVects[0].add(j.individual.unitDir.scale(distance))
                                    tfs = affineTransform3D.similarityTransform(self.fromSet, toVects, 1)
                                    for tf in tfs:
                                        if tf is not None:
                                            self.transforms.append(tf)
                                    toVects[2] = toVects[0].subtract(j.individual.unitDir.scale(distance))
                                    tfs = affineTransform3D.similarityTransform(self.fromSet, toVects, 1)
                                    for tf in tfs:
                                        if tf is not None:
                                            self.transforms.append(tf)

            # 1 point + 1 line, colinear
            # matching vectors: [point, line.start]
            # *limited transform, only use 2 vectors for finding transformation
            elif dist0.element.individual.type == 'ind-point3D' and dist1.element.individual.type == 'ind-line3D':
                for sortName in dist1.element.forms:
                    sortLines = dist1.element.forms[sortName]
                    for baseLine in sortLines:
                        if dist0.element.individual.value != baseLine.start:
                            if self.fromSet[0] is None:
                                self.fromSet = [dist0.element.individual.value, baseLine.start]
                            for i in dist0.counterparts:
                                for j in dist1.counterparts:
                                    if j.individual.containsPoint(i.individual.value):
                                        for line in j.forms[baseLine.sort.name]:
                                            if i.individual.value != line.start:
                                                toVects = [i.individual.value, line.start]
                                                tfs = affineTransform3D.limitedTransform(self.fromSet, toVects)
                                                for tf in tfs:
                                                    if tf is not None:
                                                        self.transforms.append(tf)
                                            if i.individual.value != line.end:
                                                toVects = [i.individual.value, line.end]
                                                tfs = affineTransform3D.limitedTransform(self.fromSet, toVects)
                                                for tf in tfs:
                                                    if tf is not None:
                                                        self.transforms.append(tf)
                        if dist0.element.individual.value != baseLine.end:
                            if self.fromSet[0] is None:
                                self.fromSet = [dist0.element.individual.value, baseLine.end]
                            for i in dist0.counterparts:
                                for j in dist1.counterparts:
                                    if j.individual.containsPoint(i.individual.value):
                                        for line in j.forms[baseLine.sort.name]:
                                            if i.individual.value != line.start:
                                                toVects = [i.individual.value, line.start]
                                                tfs = affineTransform3D.limitedTransform(self.fromSet, toVects)
                                                for tf in tfs:
                                                    if tf is not None:
                                                        self.transforms.append(tf)
                                            if i.individual.value != line.end:
                                                toVects = [i.individual.value, line.end]
                                                tfs = affineTransform3D.limitedTransform(self.fromSet, toVects)
                                                for tf in tfs:
                                                    if tf is not None:
                                                        self.transforms.append(tf)
        return self.transforms

    def getTransform(self, tfs):
        """
        get the affineTransform from the transformation set
        """
        string = self.id.split("=|=")
        if len(string) > 1:
            tfs.current = string[0]
        return tfs.retrieveTransform(affineTransform3D)

    # ============================================================================================
    #                            PREDICATE SOLVING PROCESS
    # ============================================================================================

    def evaluatePredicate(self, transfSet, lhs, tlhs, form, param):
        if isinstance(param, dict):
            k = param
            param = paramSet()
            param.keys = k
        param.lhs = tlhs
        param.transform = transfSet
        affineTransf = self.getTransform(transfSet)
        self.postCheck = True

        for cond in self.predicates:
            # void predicate
            if cond[0] == 'void':
                for bd in cond[1]:
                    # if there is a list of sort
                    if isinstance(bd[0], list):
                        sorts = bd[0]
                        bdr = bd[1:]

                        # create a plane boundary based on the vectors
                        bound = boundary()
                        if bdr[-1] == bdr[0]:
                            bdr.pop(-1)
                        for keyI in bdr:
                            newInd = keyI.transform(affineTransf)
                            # newInd = affineTransf.transform(keyI)
                            if newInd is False:
                                affineTransf.printTransf()
                                raise ValueError('matching for vector '+str(keyI)+' is not found')
                            bound.add(newInd)
                        
                        # remove duplicate vertices, add plane border segment
                        i = 0
                        while i < len(bound.polygonal):
                            if bound.polygonal[i] == bound.polygonal[i-1]:
                                bound.polygonal.pop(i)
                            else:
                                i += 1
                        for i in range(len(bound.polygonal)):
                            bound.add(bSegment(bound.polygonal[i],bound.polygonal[i-1]))
                        
                        # create the plane-segment
                        planeSeg = planeSegment3D(lhs.sort, [bound], True)

                        # find all lines that are co-planar with the plane
                        lines = []
                        form.accept(self, direction = ['voidof', planeSeg, lines, sorts])
                        if not self.postCheck:
                            return False
                        # check if any of the lines cut through the created plane-segment
                        check = planeSeg.classify_line(lines)
                        if not check.notInside():
                            self.postCheck = False
                    # if there's no sort list
                    else:
                        # create plane-segment boundary
                        bound = boundary()
                        if bd[-1] == bd[0]:
                            bd.pop(-1)
                        for keyI in bd:
                            newInd = keyI.transform(affineTransf)
                            # newInd = affineTransf.transform(keyI)
                            if newInd is False:
                                affineTransf.printTransf()
                                raise ValueError('matching for vector '+str(keyI)+' is not found')
                            bound.add(newInd)
                        for i in range(len(bound.polygonal)):
                            bound.add(bSegment(bound.polygonal[i],bound.polygonal[i-1]))
                        # create plane-segment
                        planeSeg = planeSegment3D(lhs.sort, [bound], True)

                        # find all lines that are co-planar with the plane
                        lines = []
                        form.accept(self, direction = [cond[0], planeSeg, lines])
                        if not self.postCheck:
                            return False

                        # check if any of the lines cut through the created plane-segment
                        check = planeSeg.classify_line(lines)
                        if not check.notInside():
                            self.postCheck = False
            # inside predicate
            elif cond[0] == 'inside':
                for bd in cond[1]:
                    # if there is a list of sort
                    if isinstance(bd[0], list):
                        sorts = bd[0]
                        bdr = bd[1:]

                        # create a plane boundary based on the vectors
                        bound = boundary()
                        if bdr[-1] == bdr[0]:
                            bdr.pop(-1)
                        for keyI in bdr:
                            bound.add(keyI)
                        
                        # remove duplicate vertices, add plane border segment
                        i = 0
                        while i < len(bound.polygonal):
                            if bound.polygonal[i] == bound.polygonal[i-1]:
                                bound.polygonal.pop(i)
                            else:
                                i += 1
                        for i in range(len(bound.polygonal)):
                            bound.add(bSegment(bound.polygonal[i],bound.polygonal[i-1]))
                        
                        # create the plane-segment
                        planeSeg = planeSegment3D(lhs.sort, [bound], True)

                        # find all lines that are co-planar with the plane
                        lines = []
                        tlhs.accept(self, direction = ['insideof', planeSeg, lines, sorts])
                        if not self.postCheck:
                            return False
                        
                        # check if any of the lines cut through the created plane-segment
                        check = planeSeg.classify_line(lines)
                        if not check.isContained():
                            self.postCheck = False
                    # if there's no sort list
                    else:
                        # create plane-segment boundary
                        bound = boundary()
                        if bd[-1] == bd[0]:
                            bd.pop(-1)
                        for keyI in bd:
                            bound.add(keyI)
                        for i in range(len(bound.polygonal)):
                            bound.add(bSegment(bound.polygonal[i],bound.polygonal[i-1]))
                        # create plane-segment
                        planeSeg = planeSegment3D(lhs.sort, [bound], True)
                        # find all lines that are co-planar with the plane
                        lines = []
                        
                        tlhs.accept(self, direction = [cond[0], planeSeg, lines])
                        if not self.postCheck:
                            return False

                        # check if any of the lines cut through the created plane-segment
                        check = planeSeg.classify_line(lines)
                        if not check.isContained():
                            self.postCheck = False
            # outside predicate
            elif cond[0] == 'outside':
                for bd in cond[1]:
                    # if there is a list of sort
                    if isinstance(bd[0], list):
                        sorts = bd[0]
                        bdr = bd[1:]

                        # create a plane boundary based on the vectors
                        bound = boundary()
                        if bdr[-1] == bdr[0]:
                            bdr.pop(-1)
                        for keyI in bdr:
                            bound.add(keyI)

                        # remove duplicate vertices, add plane border segment
                        i = 0
                        while i < len(bound.polygonal):
                            if bound.polygonal[i] == bound.polygonal[i-1]:
                                bound.polygonal.pop(i)
                            else:
                                i += 1
                        for i in range(len(bound.polygonal)):
                            bound.add(bSegment(bound.polygonal[i],bound.polygonal[i-1]))
                        
                        # create the plane-segment
                        planeSeg = planeSegment3D(lhs.sort, [bound], True)

                        # find all lines that are co-planar with the plane
                        lines = []
                        tlhs.accept(self, direction = ['outsideof', planeSeg, lines, sorts])
                        if not self.postCheck:
                            return False
                        
                        # check if any of the lines cut through the created plane-segment
                        check = planeSeg.classify_line(lines)
                        if not check.notInside():
                            self.postCheck = False
                    # if there's no sort list
                    else:
                        # create plane-segment boundary
                        bound = boundary()
                        if bd[-1] == bd[0]:
                            bd.pop(-1)
                        for keyI in bd:
                            bound.add(keyI)
                        for i in range(len(bound.polygonal)):
                            bound.add(bSegment(bound.polygonal[i],bound.polygonal[i-1]))
                        # create plane-segment
                        planeSeg = planeSegment3D(lhs.sort, [bound], True)

                        # find all lines that are co-planar with the plane
                        lines = []
                        tlhs.accept(self, direction = [cond[0], planeSeg, lines])
                        if not self.postCheck:
                            return False
                        
                        # check if any of the lines cut through the created plane-segment
                        check = planeSeg.classify_line(lines)
                        if not check.notInside():
                            self.postCheck = False
            if self.postCheck is False:
                return False
        return True

    def visitEnter(self, obj, direction = 0):
        if direction[0] == 'void':
            if 'ind-planeSegment' in obj.type :
                if ('3D' in obj.type) == ('2D' in direction[1].type):
                    raise ValueError('There exists a mix of 2D and 3D objects in the shape')
                if not plane3D.compareValue(obj,direction[1]) == 0:
                    return True
                check = obj.classify(direction[1])
                if check.disjoint() or check.touches():
                    return True
                self.postCheck = False
                return False
            if 'ind-lineSegment' in obj.type:
                if not plane3D.containsLine(direction[1], obj):
                    return True
                direction[2].append(bSegment(obj.start, obj.end))
            if 'ind-point' in obj.type:
                if not plane3D.containsPoint(direction[1], obj.value):
                    return True
                check = False
                for line in direction[1].boundaries[0].canonical:
                    if line.containsPoint(obj.value):
                        check = True
                if check:
                    return True
                if direction[1].boundaries[0].containsPointInside(obj.value):
                    self.postCheck = False
                    return False
        elif direction[0] == 'voidof':
            if 'ind-planeSegment' in obj.type :
                if not (obj.sort.name in direction[-1] or obj.sort.baseSort().name in direction[-1]):
                    return True
                if ('3D' in obj.type) == ('2D' in direction[1].type):
                    raise ValueError('There exists a mix of 2D and 3D objects in the shape')
                if not plane3D.compareValue(obj,direction[1]) == 0:
                    return True
                check = obj.classify(direction[1])
                if check.disjoint() or check.touches():
                    return True
                self.postCheck = False
                return False
            if 'ind-lineSegment' in obj.type:
                if not (obj.sort.name in direction[-1] or obj.sort.baseSort().name in direction[-1]):
                    return True
                if not plane3D.containsLine(direction[1], obj):
                    return True
                direction[2].append(bSegment(obj.start, obj.end))
            if 'ind-point' in obj.type:
                if not (obj.sort.name in direction[-1] or obj.sort.baseSort().name in direction[-1]):
                    return True
                if not plane3D.containsPoint(direction[1], obj.value):
                    return True
                check = False
                for line in direction[1].boundaries[0].canonical:
                    if line.containsPoint(obj.value):
                        check = True
                if check:
                    return True
                if direction[1].boundaries[0].containsPointInside(obj.value):
                    self.postCheck = False
                    return False
        elif direction[0] == 'inside':
            if 'ind-planeSegment' in obj.type :
                if ('3D' in obj.type) == ('2D' in direction[1].type):
                    raise ValueError('There exists a mix of 2D and 3D objects in the shape')
                if not plane3D.compareValue(obj,direction[1]) == 0:
                    self.postCheck = False
                    return False
                check = obj.classify(direction[1])
                if check.isContained():
                    return True
                self.postCheck = False
                return False
            if 'ind-lineSegment' in obj.type:
                if not plane3D.containsLine(direction[1], obj):
                    self.postCheck = False
                    return False
                direction[2].append(bSegment(obj.start, obj.end))
            if 'ind-point' in obj.type:
                if not plane3D.containsPoint(direction[1], obj.value):
                    self.postCheck = False
                    return False
                check = False
                for line in direction[1].boundaries[0].canonical:
                    if line.containsPoint(obj.value):
                        check = True
                if check:
                    return True
                if direction[1].boundaries[0].containsPointInside(obj.value):
                    return True
                self.postCheck = False
                return False
        elif direction[0] == 'insideof':
            if 'ind-planeSegment' in obj.type :
                if not (obj.sort.name in direction[-1] or obj.sort.baseSort().name in direction[-1]):
                    return True
                if ('3D' in obj.type) == ('2D' in direction[1].type):
                    raise ValueError('There exists a mix of 2D and 3D objects in the shape')
                if not plane3D.compareValue(obj,direction[1]) == 0:
                    self.postCheck = False
                    return False
                check = obj.classify(direction[1])
                if check.isContained():
                    return True
                self.postCheck = False
                return False
            if 'ind-lineSegment' in obj.type:
                if not (obj.sort.name in direction[-1] or obj.sort.baseSort().name in direction[-1]):
                    return True
                if not plane3D.containsLine(direction[1], obj):
                    self.postCheck = False
                    return False
                direction[2].append(bSegment(obj.start, obj.end))
            if 'ind-point' in obj.type:
                if not (obj.sort.name in direction[-1] or obj.sort.baseSort().name in direction[-1]):
                    return True
                if not plane3D.containsPoint(direction[1], obj.value):
                    self.postCheck = False
                    return False
                check = False
                for line in direction[1].boundaries[0].canonical:
                    if line.containsPoint(obj.value):
                        check = True
                if check:
                    return True
                if direction[1].boundaries[0].containsPointInside(obj.value):
                    return True
                self.postCheck = False
                return False
        elif direction[0] == 'outside':
            if 'ind-planeSegment' in obj.type :
                if ('3D' in obj.type) == ('2D' in direction[1].type):
                    raise ValueError('There exists a mix of 2D and 3D objects in the shape')
                if not plane3D.compareValue(obj,direction[1]) == 0:
                    return True
                check = obj.classify(direction[1])
                if check.disjoint() or check.touches():
                    return True
                self.postCheck = False
                return False
            if 'ind-lineSegment' in obj.type:
                if not plane3D.containsLine(direction[1], obj):
                    return True
                direction[2].append(bSegment(obj.start, obj.end))
            if 'ind-point' in obj.type:
                if not plane3D.containsPoint(direction[1], obj.value):
                    return True
                check = False
                for line in direction[1].boundaries[0].canonical:
                    if line.containsPoint(obj.value):
                        check = True
                if check:
                    return True
                if direction[1].boundaries[0].containsPointInside(obj.value):
                    self.postCheck = False
                    return False
        elif direction[0] == 'outsideof':
            if 'ind-planeSegment' in obj.type :
                if not (obj.sort.name in direction[-1] or obj.sort.baseSort().name in direction[-1]):
                    return True
                if ('3D' in obj.type) == ('2D' in direction[1].type):
                    raise ValueError('There exists a mix of 2D and 3D objects in the shape')
                if not plane3D.compareValue(obj,direction[1]) == 0:
                    return True
                check = obj.classify(direction[1])
                if check.disjoint() or check.touches():
                    return True
                self.postCheck = False
                return False
            if 'ind-lineSegment' in obj.type:
                if not (obj.sort.name in direction[-1] or obj.sort.baseSort().name in direction[-1]):
                    return True
                if not plane3D.containsLine(direction[1], obj):
                    return True
                direction[2].append(bSegment(obj.start, obj.end))
            if 'ind-point' in obj.type:
                if not (obj.sort.name in direction[-1] or obj.sort.baseSort().name in direction[-1]):
                    return True
                if not plane3D.containsPoint(direction[1], obj.value):
                    return True
                check = False
                for line in direction[1].boundaries[0].canonical:
                    if line.containsPoint(obj.value):
                        check = True
                if check:
                    return True
                if direction[1].boundaries[0].containsPointInside(obj.value):
                    self.postCheck = False
                    return False
        return True

    def visitLeave(self, obj, direction = 0):
        return
