# -*- coding: utf-8 -*-

from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import
from builtins import super

from .sort import sort

class attributeSort(sort):
    """
    An Attribute Sort specifies a composition of a simple :class:`~sortal.ind.sort.sort`
    (its base sort) with any other :class:`~sortal.ind.sort.sort` (its weigth sort)
    under the object-attribute relationship. An individual of the
    attribute sort is an individual of the base sort that has a form of the
    weight sort assigned as an attribute.
    An attribute sort may have a name assigned.

    The AttributeSort class represents an attribute sort additionally by
    its base and weight sorts. The canonical version of an attribute sort is the
    unnamed attribute sort of the canonical versions of the base and weight sort.

    **Object Parameters:** (not including inherited attributes)

    * **base** (:class:`~sortal.sort.sort.sort`) -- base sort object.
    * **weight** (:class:`~sortal.sort.sort.sort`) -- weight sort object.
    * **description** (str) -- sort description in sdl format.
    """
    separator = ' ^ '
    type = 'sort-attribute'
    
    def __init__(self, name, base, weight, *args, **kwargs):
        """
        initialize the attribute sort with name, base sort and weight sort

        :param name: name of the sort
        :type name: str
        :param base: the base sort object
        :type base: ~sortal.sort.sort.sort
        :param weight: the weight sort object
        :type weight: ~sortal.sort.sort.sort
        """
        if self.checkSort(name) == True:
            raise ValueError('A sort with name '+name+' already exists')
        super(attributeSort, self).__init__(name)
        self.base = base
        self.weight = weight
        self.description = self.description(base, weight)
        return

    @classmethod
    def create(cls, name, base, weight):
        """
        Try to initialize a new attribute sort with name, base sort and weight sort, 
        if a sort with the same name, base and weight already exists, that sort 
        will be returned instead of initializing a new sort

        :param name: name of the sort
        :type name: str
        :param base: the base sort object
        :type base: ~sortal.sort.sort.sort
        :param weight: the weight sort object
        :type weight: ~sortal.sort.sort.sort
        """
        if sort.checkSort(name) == True: 
            s = sort.getSort(name)
            if s.type != 'sort-attribute' or base != s.base or weight != s.weight:
                raise ValueError('A sort with name '+name+' already exists')
            return s
        return attributeSort(name, base, weight)

    def baseSort(self):
        """
        returning the base sort
        """
        return self.base

    def description(self, base, weight):
        """
        creating the description of the sort template: BaseSort ^ WeightSort

        e.g. point ^ description, line ^ weight, lineSegment ^ (description + weight + color)
        """
        desc = base.name + attributeSort.separator
        desc += weight.name
        return desc
        
    def contains(self, other):
        """
        check if this sort contains other sort:

        check both base sort and weight sort if this sort is not the same as other sort
        """
        if other.name == self.name:
            return True
        if other.type == 'sort-attribute' and self.base.name == other.base.name:
            return self.weight.contains(other.weight)
        return False
        
    def newForm(self):
        """
        creating a new form of this sort. The type of the sort is determined 
        according to the base sort individual's behavior.

        i.e. If the base sort is of lineSegment2D, the form created would be an 
        interval form
        """
        form = self.base.behavior(self)
        return form
    