# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import range
from builtins import super
from builtins import str

from .sort import sort
from ..form.metaForm import metaForm

class compoundSort(sort):
    """
    A Compound Sort specifies a disjunctive composition of
    :class:`~sortal.sort.disjunctiveSort.disjunctiveSort` under the sum operation. 
    The compound sort cannot directly contain any other sort types.

    The form corresponding for this sort would always be a
    :class:`~sortal.form.metaForm.metaForm`

    The compoundSort class represents a compound sort additionally
    as an ordered set of component sorts. 
    In the definition of a compound sort, the operation of sum is denoted by '+'.

    **Object Parameters:** (not including inherited attributes)

    * **component** (list of :class:`~sortal.sort.sort.sort`) -- a list of all the component sorts.
    * **description** (str) -- sort description in sdl format.
    """
    separator = ' + '
    type = 'sort-compound'
    
    def __init__(self, name, component = [], *args, **kwargs):
        """
        initialize the sort, a compound sort should not contain any sort types 
        other than disjunctive sort.
        
        :param name: name of the sort
        :type name: str
        :param component: the list of component disjunctive sorts
        :type component: list of :class:`~sortal.sort.disjunctiveSort.disjunctiveSort`
        """
        if self.checkSort(name) == True:
            raise ValueError('A sort with name '+name+' already exists')
        super(compoundSort, self).__init__(name)
        for i in component:
            if i.type != 'sort-disjunctive':
                raise ValueError('Compound sort '+name+' only take disjunctive sorts as components. A '+i.type[5:]+' sort named '+i.name+' is inputted')
        self.component = component
        self.description = self.description(self.component)
        return
        
    @classmethod
    def create(cls, name, component = []):
        """
        Try to initialize a new compound sort with name and component, 
        if a sort with the same name and component already exists, that sort 
        will be returned instead of initializing a new sort

        :param name: name of the sort
        :type name: str
        :param component: the list of component sorts
        :type component: list of :class:`~sortal.sort.disjunctiveSort.disjunctiveSort`
        """
        if sort.checkSort(name) == True: 
            s = sort.getSort(name)
            if s.type != 'sort-compound' or component != s.component:
                for i in component:
                    print(i)
                print()
                for i in s.component:
                    print(i)
                raise ValueError('A sort with name '+name+' already exists')
            return s
        return compoundSort(name, component)
        
    def __eq__(self,other):
        if self.type != other.type:
            return False
        for i in self.component:
            check = False
            for j in other.component:
                if i == j:
                    check = True
            if check == False:
                return False
        return True
    
    def addComponent(self, sort):
        """
        add a component sort to this sort
        """
        for i in self.component:
            if sort.name == i.name:
                return
        self.component.append(sort)

    @staticmethod
    def sameComponent(component):
        uniqueComp = compoundSort.manageComp(component)
        for name in sort.register:
            s = sort.register[name]
            if s.type != 'sort-disjunctive':
                continue
            if uniqueComp == s.component:
                return s
        return False
    
    def description(self, component):
        """
        creating the description for this sort
        """
        if len(component) <= 1:
            desc = ''
            for i in component:
                desc += i.description
        else:
            desc = '('
            for i in component:
                if desc == '(':
                    desc += i.description
                else:
                    desc += compoundSort.separator + i.description
            desc += ')'
        return desc
        
    def contains(self, other):
        """
        checking if this sort contain other sort
        """
        if other.type == 'sort-compound':
            for i in other.component:
                Check = False
                for j in self.component:
                    if j.contains(i):
                        Check = True
                if Check == False:
                    return False
            return True
        else:
            for i in self.component:
                if i.contains(other):
                    return True
        return False
        
    def complementSort(self, other):
        """
        find the result when subtracting this sort by another sort
        """
        if not self.contains(other):
            raise ValueError("Compound sort "+self.name+" does not contain the sort named "+other.name)
        comp = []
        if other.type != 'sort-compound':
            for i in self.component:
                if i == other:
                    continue
                comp.append(i)
        else:
            for i in self.component:
                check = False
                for j in other.component:
                    if i == j:
                        check = True
                        break
                if check:
                    continue
                comp.append(i)
        nName = newName = self.name+'_'+other.name+'_complement'
        count = 0
        while self.checkSort(nName):
            nName = newName+str(count)
            count+=1
        return compoundSort(nName, comp)
        
    def dimension(self):
        """
        dimension of this sort is 0
        """
        return 0
        
    def newForm(self):
        """
        creating a new metaform of this sort
        """    
        try:
            form = metaForm(self)
            return form
        except:
            print('exception')
    
    