# -*- coding: utf-8 -*-

from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import

from builtins import super

from .sort import sort
from ..form.metaForm import metaForm


class disjunctiveSort(sort):
    """
    A Disjunctive Sort specifies a disjunctive composition of
    :class:`~sortal.sort.sort.sort` under the sum operation. Under the sum operation,
    a form of the resulting sort may contain a form of every operand sort.
    If any component sort is an unnamed disjunctive sort, instead,
    its components become part of the disjunctive composition.

    A disjunctive sort may have a name assigned.

    The form corresponding for this sort would always be a
    :class:`~sortal.form.metaForm.metaForm`

    The DisjunctiveSort class represents a disjunctive sort additionally
    as an ordered set of component sorts. 
    In the definition of a disjunctive sort, the operation of sum is denoted by '+'.

    **Object Parameters:** (not including inherited attributes)

    * **component** (list of :class:`~sortal.sort.sort.sort`) -- a list of all the component sorts.
    * **description** (str) -- sort description in sdl format.
    """

    separator = ' + '
    type = 'sort-disjunctive'
    
    def __init__(self, name, component = [], *args, **kwargs):
        """
        initialize the sort, a disjunctive sort should not contain any other disjunctive sort
        or compound sort. All the subordinate sorts with the same base sort would be combined

        | e.g.
        | (a + b) + (a + c) -> a + b + c
        | a ^ b + a ^ c + d -> a ^ (b + c) + d

        The description will be based on the updated component

        :param name: name of the sort
        :type name: str
        :param component: the list of component sorts
        :type component: list of :class:`~sortal.sort.sort.sort`
        """
        if self.checkSort(name) == True:
            raise ValueError('A sort with name '+name+' already exists')
        super(disjunctiveSort, self).__init__(name)
        
        if 'absoluteComp' in kwargs and kwargs['absoluteComp'] is True:
            self.component = component
        else:
            #compiling all sorts from the component
            uniqueComp = disjunctiveSort.manageComp(component)
        
            #assign to self.component
            self.component = uniqueComp

        self.dimensional = None
        for i in self.component:
            dim = i.baseSort().characteristic.type[-2:]
            if dim != '2D' and dim != '3D':
                continue
            if self.dimensional is None:
                self.isParametric = True if i.baseSort().characteristic.type[-3] == 'P' else False
                self.dimensional = dim
            else:
                if self.isParametric != (i.baseSort().characteristic.type[-3] == 'P'):
                    print(i.baseSort().characteristic.type)
                    raise ValueError('Cannot have both Parametric and Non-Parametric objects under the same disjunctive sort '+name)
                if self.dimensional != dim:
                    print(i.baseSort().characteristic.type)
                    raise ValueError('Cannot have both 2D and 3D objects under the same disjunctive sort '+name)

        self.description = self.description(self.component)
        return
        
    @classmethod
    def create(cls, name, component = []):
        """
        Try to initialize a new disjunctive sort with name and component, 
        if a sort with the same name and component already exists, that sort 
        will be returned instead of initializing a new sort

        :param name: name of the sort
        :type name: str
        :param component: the list of component sorts
        :type component: list of :class:`~sortal.sort.sort.sort`
        """
        uniqueComp = disjunctiveSort.manageComp(component)
        if sort.checkSort(name) == True: 
            s = sort.getSort(name)
            if s.type != 'sort-disjunctive' or uniqueComp != s.component:
                for i in uniqueComp:
                    print(i)
                print()
                for i in s.component:
                    print(i)
                raise ValueError('A sort with name '+name+' already exists')
            return s
        return disjunctiveSort(name, uniqueComp, absoluteComp = True)
        
    @staticmethod
    def manageComp(component):
        """
        Remove duplicate component sorts, combine sorts with the same base sorts
        and reorder the sorts in the component.

        :param component: the list of component sorts
        :type component: list of :class:`~sortal.sort.sort.sort`
        """
        #compiling all sorts from the component
        uniqueComp = []
        for i in component:
            if i.type == 'sort-disjunctive':
                for j in i.component:
                    if j not in uniqueComp:
                        uniqueComp.append(j)
            elif i not in uniqueComp:
                uniqueComp.append(i)
                
        #combining attribute sorts with the same base sort
        iIndex = 0
        while iIndex < len(uniqueComp):
            i = uniqueComp[iIndex]
            jIndex = iIndex + 1
            while jIndex < len(uniqueComp):
                j = uniqueComp[jIndex]
                if i.type == 'sort-attribute' and j.type == 'sort-attribute' and i.base.name == j.base.name:
                    nameList = []
                    if i.weight.type == 'sort-disjunctive':
                        for comp in i.weight.component:
                            if comp.name not in nameList:
                                nameList.append(comp.name)
                    else:
                        nameList.append(i.weight.name)
                        
                    if j.weight.type == 'sort-disjunctive':
                        for comp in j.weight.component:
                            if comp.name not in nameList:
                                nameList.append(comp.name)
                    else:
                        nameList.append(j.weight.name)
                        
                    
                    newDisj = disjunctiveSort(disjunctiveSort.separator.join(nameList), [i.weight, j.weight])
                    i.weight = newDisj
                    i.name= i.base.name+'_'+'('+i.weight.name.replace(" ", "")+')'
                    uniqueComp.remove(j)
                else:
                    jIndex += 1
            iIndex += 1
            
        #sort the list
        uniqueComp.sort() 
        return uniqueComp
        
    def __eq__(self,other):
        if self.__class__ != other.__class__:
            return False
        for i in self.component:
            check = False
            for j in other.component:
                if i == j:
                    check = True
            if check == False:
                return False
        return True
    
    def addComponent(self, sort):
        """
        add a component sort to this sort
        """
        for i in self.component:
            if sort.name == i.name:
                return
        self.component.append(sort)

    @staticmethod
    def sameComponent(component):
        """
        Get the sort from the sort list that have the same component as 
        the input component list
        """
        uniqueComp = disjunctiveSort.manageComp(component)
        for name in sort.register:
            s = sort.register[name]
            if s.type != 'sort-disjunctive':
                continue
            if uniqueComp == s.component:
                return s
        return False
    
    def description(self, component):
        """
        creating the description for this sort
        """
        if len(component) <= 1:
            desc = ''
            for i in component:
                desc += i.description
        else:
            desc = '('
            for i in component:
                if desc == '(':
                    desc += i.description
                else:
                    desc += disjunctiveSort.separator + i.description
            desc += ')'
        return desc
        
    def contains(self, other):
        """
        checking if this sort contain other sort
        """
        if other.type == 'sort-disjunctive':
            for i in other.component:
                Check = False
                for j in self.component:
                    if j.contains(i):
                        Check = True
                if Check == False:
                    return False
            return True
        else:
            for i in self.component:
                if i.contains(other):
                    return True
        return False
        
    def complementSort(self, other):
        """
        find the result when subtracting this sort by another sort
        """
        if not self.contains(other):
            raise ValueError("Disjunctive sort "+self.name+" does not contain the sort named "+other.name)
        comp = []
        if other.type != 'sort-disjunctive':
            for i in self.component:
                if i == other:
                    continue
                comp.append(i)
        else:
            for i in self.component:
                check = False
                for j in other.component:
                    if i == j:
                        check = True
                        break
                if check:
                    continue
                comp.append(i)
        nName = newName = self.name+'_'+other.name+'_complement'
        count = 0
        while self.checkSort(nName):
            nName = newName+str(count)
            count+=1
        return disjunctiveSort(nName, comp)
        
    def dimension(self):
        """
        dimension of this sort: 0
        """
        return 0
        
    def newForm(self):
        """
        creating a new metaform of this sort
        """    
        try:
            form = metaForm(self)
            return form
        except:
            print('exception')
    
    