# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import object
from enum import Enum
from .colorMode import colorMode
from .enumeration import enumeration
from ...io.parseReader import parseReader, ParsingError

class argument(object):
    """
    Argument class: Extra argument passed into a :class:`~sortal.sort.primitiveSort.primitiveSort`
    initialization for determining the behavior of the individuals under the sort. e.g. an argument indicating
    the upper-bound value of a weight individual should be passed into the primitive sort initialization. The
    argument type for each individual is fixed and is registered in the primitiveSort class' parameter dictionary
    in sortal/__init__.py

    An argument object is used for all sorts of individuals that are not color and enumerative.
    Color and enumerative sorts would have :class:`~sortal.sort.parameter.colorMode.colorMode` 
    and :class:`~sortal.sort.parameter.enumeration.enumeration` in place of argument that serve
    the same purpose.

    **Object Parameters:**

    * **argType** (:class:`~sortal.sort.parameter.argument.argument.argumentType`) -- the argument type of the sort.
    * **value** (str) -- the value of the argument.

    """

    class argumentType(Enum):
        """
        Argument type enum class.
        
        Each individual characteristic class should be assigned an argument type
        to indicate the extra arguments to be passed into its sort initialization

        * NONE: No extra parameter required
        * IDENTIFIER: 
        * UPPERBOUND: one string of one float/int acting as the maximum limit of the value of the individual under the sort
        * LOCALE:
        * SORTSLINK:
        * COLORMODE: one string of multiple values indicating the color model, color behavior and the maximum individual color values
        * ENUMERATION: one string indicating the enumaration values as well as the ranking matrix for enumerative individuals

        """
        NONE = [False,0]
        IDENTIFIER = [True,1]
        UPPERBOUND = [True,2]
        LOCALE = [True,3]
        SORTSLINK = [False,4]
        COLORMODE = [True,5]
        ENUMERATION = [False,6]
        
    def __init__(self, argType, value):
        self.type = 'sortParam-argument'
        self.argType = argType
        self.value = value
        
    def argValue(self):
        """
        get the argument value
        """
        return self.value
                
    def cardinality(self,argType):
        """
        get the cardinality of the argument
        """
        a = self.argumentType
        if argType == a.IDENTIFIER or argType == a.UPPERBOUND:
            return 1
        elif argType == a.SORTSLINK or argType == a.LOCALE:
            return 2
        elif argType == a.COLORMODE:
            return 6
        else:
            return 0

    @staticmethod
    def defineArgument(argType, argList, characteristic):
        """
        Create an argument/colorMode/enumeration object based on the individual's characteristic,
        argument type and argument value
        """
        if argType == argument.argumentType.NONE:
            raise ValueError("This characteristic individual allows no arguments")
        result = None
        if argType == argument.argumentType.IDENTIFIER:
            pass
        elif argType == argument.argumentType.UPPERBOUND:
            if argList is None and characteristic.type in ('ind-rWeight', 'ind-weight'):
                return None
            expression = parseReader(argList)
            if expression.token() != parseReader.NUMBER:
                raise ParsingError(expression,'Expected a numeric argument')
            value = float(expression.value())
            result = argument(argType, value)
        elif argType == argument.argumentType.LOCALE:
            pass
        elif argType == argument.argumentType.SORTSLINK:
            pass
        elif argType == argument.argumentType.COLORMODE:
            result = colorMode(argList)
        elif argType == argument.argumentType.ENUMERATION:
            result = enumeration(argList)
        return result
