# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import object
from enum import Enum

from ...io.parseReader import parseReader
from ...struct.coordinate import coordinate as coord

class colorMode(object):
    """
    Color mode class: extra argument passed into a :class:`~sortal.sort.primitiveSort.primitiveSort`
    initialization for determining the behavior of all color individuals under the sort. The extra
    argument would contain:

    * Color model:
        * HSV/HSB: hue-saturation-value/brightness model
        * RGB: red-green-blue model
    * Molor behavior (optional)
        * RGB_MAX (default): the color exhibit an ordinal behavior
        * RGB_AVG: combination and complement of colors are done by finding the average values of the colors, otherwise behave ordinally
        * RGB_SUM: the color exhibit an interval behavior 
        * RGB_ALPHA:the color have an alpha value which is relied on for color operations.
    * Maximum color values (optional): 1 or 3 values representing the maximum of the corresponding color value 
        * For RGB, the default maximum values are 255 for each of the color. Declared maximum value should be between 0 and 255
        * For HSV/HSB, the default maximum values are 1 for Hue and saturation, 255 for value
        * If color model is RGB and there is only one value given, all colors would share the maximum value
    * Maximum alpha value (only for RGB_ALPHA -- optional): number between 0 and 1 indicating the maximum alpha value allowed (default to 0 of not specified)
    * The arguments should be separated by space (" ") and the argument string should end with a comma (",")

    E.g. arg = 'RGB RGB_MAX 100 100 100,'

    **Object Parameters:** 

    * **model** (:class:`~colorMode.Model`) -- model of the color individual under this sort.
    * **behavior** (:class:`~colorMode.Behavior`) -- behavior of the color individual under this sort.
    * **max_alpha** (float) -- maximum alpha value.
    * **max_cvalues** (list of 1 or 3 floats) -- maximum alpha value.

    """

    class Model(Enum):
        """
        the color model enum class: RGB and HSV
        """
        RGB = 0
        HSV = 1
        
    class Behavior(Enum):
        """
        the color behavior class: RGB_MAX, RGB_ALPHA, RGB_AVG and RGB_SUM
        """
        RGB_MAX = 0
        RGB_ALPHA = 1
        RGB_AVG = 2
        RGB_SUM = 3
    
    MAX_ALPHA = coord(1) #transparency of color
    MAX_RGB_VALUES = [coord(255)]
    MAX_HSV_VALUES = [coord(1),coord(1),coord(255)]
        
        
    def __init__(self, argList):
        m = colorMode.Model
        b = colorMode.Behavior
        reader = parseReader(argList)
        if reader.token() != parseReader.IDENTIFIER:
            raise ValueError(reader, "Expected an identifier argument ('a')")
        modelStr = reader.value()
        if modelStr == 'HSV' or modelStr == 'HSB':
            model = m.HSV
        elif modelStr == 'RGB':
            model = m.RGB
        else:
            raise ValueError(reader, "Expected 'RGB', 'HSV' or 'HSB'")
        behavior = b.RGB_MAX
        if reader.newToken() == ',':
            self.model = model
            self.behavior = behavior
            self.max_alpha = 0
            self.max_cvalues = None
            return
        
        if reader.token() != "a":
            raise ValueError(reader, "Expected an identifier argument ('a')")
        behaviorstr = reader.value()
        if behaviorstr == 'RGB_ALPHA' or behaviorstr == 'ALPHA':
            behavior = b.RGB_ALPHA
        elif behaviorstr == 'RGB_AVG' or behaviorstr == 'AVG':
            behavior = b.RGB_AVG
        elif behaviorstr == 'RGB_SUM' or behaviorstr == 'SUM':
            behavior = b.RGB_SUM
        elif behaviorstr == 'RGB_MAX' or behaviorstr == 'MAX':
            behavior = b.RGB_MAX
        else: 
            raise ValueError(reader, "Expected 'RGB_MAX', 'RGB_ALPHA', 'RGB_AVG' or 'RGB_SUM'")
                    
        if reader.newToken() == ',':
            self.model = model
            self.behavior = behavior
            self.max_alpha = 0
            self.max_cvalues = None
            return

        if reader.token() != "0":
            raise ValueError(reader, "Number expected ('0')")
        alpha = 0
        value = reader.value()

        if reader.previewToken() != ',' :
            values = [value,None,None,None]
            i = 1
            while i < 4 and reader.newToken() != ',':
                if reader.token() != "0":
                    raise ValueError(reader, "Number expected ('0')")
                values[i] = reader.value()
                i+=1
            
            #if i == 4:
            if behavior == b.RGB_ALPHA:
                alpha = values[0]
            values = values[1:i]
        else:
            values = [value]

        for i in range(len(values)):
            if values[i] is not None:
                values[i] = coord(values[i])
        self.model = model
        self.behavior = behavior
        self.max_alpha = alpha
        self.max_cvalues = values
    @property
    def max_cvalues(self):
        """
        get the max color values
        """
        if self.__max_cvalues is not None:
            return self.__max_cvalues
        if self.model == colorMode.Model.HSV:
            return self.MAX_HSV_VALUES
        return self.MAX_RGB_VALUES
            
    @max_cvalues.setter
    def max_cvalues(self, val):
        """
        set the max color values
        """
        self.__max_cvalues = val

    
    @property
    def max_alpha(self):
        """
        get the max alpha value
        """
        if self.__max_alpha > 0:
            return self.__max_alpha
        return self.MAX_ALPHA
    

    @max_alpha.setter
    def max_alpha(self, val):
        """
        set the max alpha values
        """
        self.__max_alpha = coord(val)

    