# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import range
from builtins import str
from builtins import object

from ...io.parseReader import parseReader, ParsingError

class enumeration(object):
    """
    Color mode class: extra argument passed into a :class:`~sortal.sort.primitiveSort.primitiveSort`
    initialization for determining the behavior of all color individuals under the sort. The extra
    argument would contain two parts separated by a space:

    * First part: a list of all enumeration values joined by ','. The list should be enclosed within a set of curly brackets ("{}")
    * Second part: a ranking matrix indicating the comparison result between any two enumerations enclosed within a set of square brackets ("[]")

    | i.e. 
    | argList = '{identifier_0, identifier_1, identifier_2 ,..., identifier_n} 
    |   [rank_0-0 rank_0-1 rank_0-2 ... rank_0-n 
    |   rank_1-0 rank_1-1 rank_1-2 ... rank_1-n 
    |   ......
    |   rank_n-0 rank_n-1 rank_n-2 ... rank_n-n]'

    rank_x-y should be the value of the enumeration that is bigger in the comparison. 
    For example, if identifier_0 (index being 0) is smaller than identifier_5 (index 
    being 5), rank_0-5 and rank_5-0 would both be 5.

    | Example: (the whole string should be in one line)
    | argList = '{red green blue white black} 
    |   [red green blue white black
    |   green blue white black black
    |   blue white black black black
    |   white black black black black
    |   black black black black black]'

    **Object Parameters:**

    * **values** (list of string) -- model of the color individual under this sort.
    * **rankings** (:class:`~colorMode.Behavior`) -- behavior of the color individual under this sort.

    """
    def __init__(self, argList):
        if isinstance(argList, str) or isinstance(argList, unicode):
            reader = parseReader(argList)
            identifiers = []
            if reader.token() != '{':
                raise ParsingError(reader, 'Expected an enumeration of identifiers, separated with commas, and enclosed with curly brackets')
            if reader.newToken() != reader.IDENTIFIER:
                raise ParsingError(reader, 'Expected at least one identifier')
            identifiers.append(reader.value())
            while reader.newToken() == ',':
                if reader.newToken() != reader.IDENTIFIER:
                    raise ParsingError(reader, 'Expected identifier')
                identifiers.append(reader.value())
            if reader.token() != '}':
                raise ParsingError(reader, 'expected "}"')
                
            identifiers = sorted(identifiers)
            size = len(identifiers)

            if reader.newToken() != '[':
                raise ParsingError(reader, "Expected a list of " + (size * size) + " values taken from the previous list, separated with spaces, and enclosed with square brackets")
            
            ranking = [[None for i in range(size)] for j in range(size)]
            for i in range(size):
                for j in range(size):
                    if reader.newToken() != reader.IDENTIFIER:
                        raise ParsingError(reader, 'Expected identifier')
                    ranking[i][j] = identifiers.index(reader.value())
            self.values = identifiers
            self.rankings = ranking

        elif isinstance(argList, list):
            raise TypeError('newer way of reading enumeration has been implemented, enumeration argument should be string instead of list')
        else:
            raise TypeError('enumeration argument should be string')
        
    def argValue(self):
        """
        get all the enumeration in the argument, joined by ','
        """
        string = ''
        for i in range(len(self.values)-1):
            string+=self.values[i]+','
        string += self.values[-1]
        return string
        
    def ranking(self, first, second):
        """
        check the ranking matrix for the relation between object at first index compared to the object at second index
        """
        if first <0 or second < 0 or first >= len(self.values) or second >= len(self.values):
            raise IndexError('enumeration index out of bounds')
        return self.rankings[first][second]