# -*- coding: utf-8 -*-

from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import
from builtins import super
from builtins import str

from .sort import sort
from .parameter.argument import argument
class primitiveSort(sort):
    """
    A Primitive Sort is a :class:`~sortal.sort.sort.sort` defined by
    a characteristic :class:`~sortal.ind.individual.individual`
    and a behavioral category that specifies the operational behavior of
    its :class:`~sortal.form.form.form`, that is, collections of individuals.

    A primitive sort also always has a name assigned.

    The primitiveSort class represents a primitive sort additionally by
    its characteristic individual class, behavior class, corresponding arguments,
    default :class:`~sortal.map.mapping.mapping`, and dimensionality.
    The canonical version of a primitive sort is the sort itself, unless
    the sort has been created as a duplicate of another primitive sort,
    in which case it takes its canonical version from that other sort.

    It extends on the :class:`~sortal.sort.sort.sort` abstract class.

    **Object Parameters:** (not including inherited attributes)

    * **characteristic** (python class) -- the specific individual class of the individual represented by the sort. e.g. point2D class, lineSegmentP3D class, etc.
    * **behavior** (python class) -- the specific form class of represented individual' behavior. e.g. interval, discrete, etc.
    * **parameter** (:class:`~sortal.sort.parameter.argument.argument.argumentType`) -- the type of the extra parameters the sort has for certain individual classes.
    * **arguments** () -- the extra parameters the sort has for certain individual classes.
    * **description** (str) -- sort description in sdl format.
    """
    characteristicList = {}
    categories = {}
    parameter = {}
    defaultMapping = {}
    dimensionality = {}
    type = 'sort-primitive'
    
    def __init__(self, name, characteristic, arg = None, dummy = False, *args, **kwargs):
        """
        Initializing the primitive sort

        :param name: name of the sort
        :type name: str
        :param characteristic: the class of the individual represented by the sort
        :type characteristic: python class
        :param arg: (optional) the extra parameters required for initializing the sort of certain individual types
            (e.g. color, enumerative, weight, rWeight)
        :type arg: str
        """
        if self.checkSort(name) == True: 
            raise ValueError('A sort with name '+name+' already exists')
        super(primitiveSort, self).__init__(name, dummy = dummy)
        self.characteristic = characteristic
        self.behavior = primitiveSort.categories[characteristic]
        self.parameter = primitiveSort.parameter[characteristic]
        self.description = '['+characteristic.type[4:]+']'
        if self.parameter != argument.argumentType.NONE:
            self.arguments = argument.defineArgument(self.parameter, arg, characteristic)
            if self.arguments is not None:
                self.description += ' ('+arg+')'
                
    @classmethod
    def create(cls, name, characteristic, arg = None):
        """
        Try to initialize a new compound sort with name, characteristic and argument, 
        if a sort with the same name, characteristic and argument already exists, that sort 
        will be returned instead of initializing a new sort

        :param name: name of the sort
        :type name: str
        :param characteristic: the class of the individual represented by the sort
        :type characteristic: python class
        :param arg: (optional) the extra parameters required for initializing the sort of certain individual types
            (e.g. color, enumerative, weight, rWeight)
        :type arg: str
        """
        if sort.checkSort(name) == True: 
            s = sort.getSort(name)
            if s.type != 'sort-primitive' or characteristic != s.characteristic:
                raise ValueError('A sort with name '+name+' already exists')
            return s
        return primitiveSort(name, characteristic, arg)
        
    @staticmethod
    def register(characteristic, b, p, m, d):
        """

        Registering each individual with its category, parameter, defaultMapping and dimensionality.

        * a point is of discrete category, no parameter, similarity mapping and discrete dimensionality
        * a line-segment is of interval category, no parameter, similarity mapping and linear dimensionality
        * a label is of discrete category, no parameter, exact mapping and discrete dimensionality
        * a weight is of ordinal category, 'upperbound' parameter, exact mapping and linear dimensionality
        * etc...

        This function should only be performed in sortal/__init__.py and register all available individuals for
        rule analysis and detection
        """
        primitiveSort.characteristicList[characteristic.type[4:].lower()] = characteristic
        primitiveSort.categories[characteristic] = b
        primitiveSort.parameter[characteristic] = p
        primitiveSort.defaultMapping[characteristic] = m
        primitiveSort.dimensionality[characteristic] = d
      
    @staticmethod
    def getBehavior(characteristic):
        """
        get respective behavior(category) of an individual
        """
        return primitiveSort.categories[characteristic]

    def getParameter(self):
        """
        get respective parameter of an individual
        """
        return primitiveSort.parameter[self.characteristic]

    def mapping(self):
        """
        get respective mapping of an individual
        """
        return primitiveSort.defaultMapping[self.characteristic]

    def dimension(self):
        """
        get respective dimensionality of an individual
        """
        return primitiveSort.dimensionality[self.characteristic]
        
    def contains(self, other):
        """
        checking if this sort contains another sort i.e. the two sorts are the same
        """
        if other.type == 'sort-primitive' and other.name == self.name:
            return True
        return False
        
    def newForm(self):
        """
        creating a new form (discrete form / ordinal form / interval form, ...)
        corresponding to the individual of this primitive sort
        """    
        form = self.behavior(self)
        return form

