# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import str

from builtins import object

class sort(object):
    """
    A Sort can be considered as a class structure, specifying either
    a single data type or a composition of other class structures.
    Conceptually, a sort is a set of similar values or models, 
    i.e., :class:`~sortal.ind.individual.individual`, that can be unifiably 
    described. The specification of a sort includes a specification of the 
    algebraic behavior of its :class:`~sortal.form.form.form`, i.e., collections 
    of individuals.

    A sort may also be assigned a name; naming sorts serves for referencing
    as well as to distinguish sorts that may otherwise be identical.

    A :class:`~sortal.sort.primitiveSort.primitiveSort` specifies a single data type, and is defined by
    its characteristic individual and its behavioral category.

    An :class:`~sortal.sort.attributeSort.attributeSort` specifies a composition of a primitive sort
    (its base) with any other (primitive, attribute or disjunctive) sort (its weight) under an object-attribute
    relationship.
    
    A :class:`~sortal.sort.disjunctiveSort.disjunctiveSort` specifies a co-ordinate composition of any number
    of primitive and attribute sorts.
    
    A :class:`~sortal.sort.compoundSort.compoundSort` specifies a co-ordinate composition of any number
    of disjunctive sorts.
    
    The sort class is an abstract class specifying the common ground
    for all sortal classes. This common representation includes the context
    in which the sort is registered, the description of
    the sort, the semantic base of the sort, an optional name, and the
    following counts: the number of named components, the number of simple
    components, and the number of all components.

    The description of a sort is a string in sdl format that
    defines this sort, without naming it. The semantic base of a sort is a sort
    that is (syntactically identical and) semantically equivalent and is
    determined for the purpose of comparing sorts for semantic equivalence.
    Two sorts are semantically equivalent if they are both semantically derived,
    through one or more (re)naming operations, from the same semantic base.
    A semantic base is itself not semantically derived from any other sort.
    In the case of a simple sort (or a recursive sort), it is any sort that is
    not defined through a renaming operation. In the case of a (non-recursive)
    composite sort, it is any unnamed sort.

    The name of a sort must be an identifier, i.e., a composition of letters,
    digits and/or underscores, not starting with a digit.

    **Object Parameters:** (not including inherited attributes)

    * **name** (str) -- name of the sort.

    """
    register = {}
    dummyRegister = {}

    def __init__(self, name=None, dummy = False):
        """
        Initializing a new sort, registering the sort in the sort register
        """
        self.name = name
        if dummy:
            sort.dummyRegister[name] = self
        else:
            sort.register[name] = self
        return
        
    def __repr__(self):
        return str(self)
        
    @staticmethod
    def checkSort(name):
        """
        check if a sort name already exists in the register
        """
        if name in sort.dummyRegister:
            return True
        if name not in sort.register:
            return False
        else:
            return True
            
    def registerSort(self, name):
        sort.register[name] = self
        
    @staticmethod
    def getSort(name):
        """
        Retrieve a sort object from the sort register with the sort name
        """
        if name in sort.dummyRegister:
            return sort.dummyRegister[name]
        try:
            return sort.register[name]
        except:
            return None            
        
    @staticmethod
    def cleanup():
        """
        Clear all the sorts from the sort register
        """
        for i in list(sort.register.keys()):
            del sort.register[i]
            
    def __str__(self):
        return str(self.name)
            
    def __eq__(self, other):
        """
        checking if two sorts are the same. Two sorts are the same if 
        they have the same name and same description
        """
        if self is other:
            return True
        if other is None or not isinstance(other, sort):
            return False
        return (self.name == other.name) and (self.description == other.description)
        
    def __ne__(self, other):
        return not self == other
 
    
    def compareValue(self, other):
        """
        Comparing two sorts
        """
        if self.name == other.name:
            if str(self.__dict__) > str(other.__dict__):
                return 1
            elif str(self.__dict__) < str(other.__dict__):
                return -1
            else: 
                return 0
        elif self.name > other.name:
            return 1
        else: 
            return -1

    def baseSort(self):
        """
        getting the base sort of the sort.

        most sorts (primitive sort and disjunctive sort) have itself as the base sort,
        however attribute sorts have a different base sort.
    
        This function will only be available for primitive sorts and disjunctive sorts
        """
        return self
        

    def __lt__(self, other):
        """
        less than
        """
        if self.name < other.name:
            return True
        if str(self.__dict__) < str(other.__dict__):
            return True
        return False
        
    def __le__(self, other):
        """
        less than or equal to
        """
        if self.name <= other.name:
            return True
        if str(self.__dict__) <= str(other.__dict__):
            return True
        return False
        
    def __gt__(self, other):
        """
        greater than
        """
        if self.name > other.name:
            return True
        if str(self.__dict__) > str(other.__dict__):
            return True
        return False
        
    def __ge__(self, other):
        """
        greater than or equal to
        """
        if self.name >= other.name:
            return True
        if str(self.__dict__) >= str(other.__dict__):
            return True
        return False    
        
    def newForm(self):
        """
        Abstract function for creating a new form of this sort. 
        To be implemented by all sort classes.
        """
        raise NotImplementedError 
