# -*- coding: utf-8 -*-

from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import absolute_import
from __future__ import with_statement

from builtins import int
from builtins import str

from builtins import object
from mpmath import sqrt, pi, sin, cos, tan, asin, acos, atan, atan2
from mpmath import isint, mp, nstr, almosteq
from math import floor, ceil

class coordinate(object):
    """
    Coordinate (mpf) is the representational unit for geometric.
    Coordinate (mpf) is saved and represented as mpmath mpf objects.

    **Object Parameters:**

    * **value** (mpmath.mpf) -- Value of the coordinate.

    """

    """
    The precision of the floating point conversion is set to 40.
    i.e. the value of the coordinate is rounded to  40 decimal places when converted to float value
    """
    mp.dps = 40
    p = 9
    defaultP = 9
    cp = mp.mpf('1e-'+str(p))
    c0 = mp.mpf(0)
    c1 = mp.mpf(1)
    type = 'struct-coordinate'
    
    def __init__(self, num):
        """
        Initializing the coordinate.

        :param num: the value of the new coordinate
        :type num: str or int or float or mpmath.mpf
        """
        if isinstance(num, mp.mpf):
            self.value = num
        elif isinstance(num, coordinate):
            # raise ValueError('initiating an already existing coordinate object')
            self.value = num.value
        elif isinstance(num, int) or isinstance(num, str):
            self.value = mp.mpf(num)
        else:
            try:
                self.value = mp.mpf(num)
            except:
                print('Cannot create a coordinate with '+str(num))
    
    def __float__(self):
        """
        convert the coordinate to float
        """
        return float(self.value)
        
    def __int__(self):
        """
        convert the coordinate to int
        """
        return int(self.value)
        
    def __repr__(self):
        """
        string representation of the coordinate
        """
        return nstr(self.value,3)
        
    def __str__(self):
        """
        string representation of the coordinate
        """
        return nstr(self.value,16)
        return str(self.value)
        
    @staticmethod
    def precision():
        """
        get the current precision setting
        """
        return coordinate.p
        
    @staticmethod
    def setPrecision(prec = None):
        """
        set the current precision setting
        """
        if prec is None:
            coordinate.p = coordinate.defaultP
        else:
            coordinate.p = prec
        
    @staticmethod
    def setDefaultPrecision():
        """
        set the current precision setting back to default
        """
        coordinate.p = coordinate.defaultP
        
    def __round__(self, n = None):
        """
        round the coodinate to the closest interger (or a specified decimal place)
        """
        if n is None:
            return coordinate(round(self.value))
        return coordinate(round(self.value,n))
                
    def ceil(self):
        """
        round the coodinate up to the closest interger
        """
        return coordinate(ceil(self.value))
                
    def floor(self):
        """
        round the coodinate down to the closest interger
        """
        return coordinate(floor(self.value))
        
    def sign(self):
        """
        find the sign of the coordinate: 1 if positive, -1 if negative, 0 if 0
        """
        if self == 0:
            return 0
        if self > 0:
            return 1
        return -1
        
    def isZero(self):
        """
        check if the coordinate is equal to 0
        """
        return almosteq(self.value, 0, self.cp)
        
    def isOne(self):
        """
        check if the coordinate is equal to 1
        """
        return round(self.value, self.p) == 1
    
    def sqrt(self):
        """
        find square root of the coordinate
        """
        return coordinate(sqrt(self.value))
        
    def __add__(self, other):
        """
        addition
        """
        if not isinstance(other, coordinate):
            other = coordinate(other)
        return coordinate(self.value + other.value)   

    def __sub__(self, other):
        """
        subtraction
        """
        if not isinstance(other, coordinate):
            other = coordinate(other)
        return coordinate(self.value - other.value)

    def __mul__(self, other):
        """
        multiplication
        """
        if not isinstance(other, coordinate):
            other = coordinate(other)
        return coordinate(self.value * other.value)

    def __truediv__(self, other):
        """
        division
        """
        if not isinstance(other, coordinate):
            other = coordinate(other)
        return coordinate(self.value / other.value)

    def __floordiv__(self, other):
        """
        floor dividsion (only taking the integer part of the result)
        e.g. 23 // 2 = 11
        """
        if not isinstance(other, coordinate):
            other = coordinate(other)
        return coordinate(self.value // other.value)  

    def __pow__(self, other):
        """
        power
        """
        if not isinstance(other, coordinate):
            other = coordinate(other)
        return coordinate(self.value ** other.value)

    def __mod__(self, other):
        """
        modulo
        """
        if not isinstance(other, coordinate):
            other = coordinate(other)
        return coordinate(self.value % other.value)   
        
    def __radd__(self, other):
        """
        reverse addition: other + self (only happens if the operation (other + self) gives an error)
        """
        other = coordinate(other)
        return coordinate(other.value + self.value)   

    def __rsub__(self, other):
        """
        reverse subtraction: other - self (only happens if the operation (other - self) gives an error)
        """
        other = coordinate(other)
        return coordinate(other.value - self.value)

    def __rmul__(self, other):
        """
        reverse multiplication: other * self (only happens if the operation (other * self) gives an error)
        """
        other = coordinate(other)
        return coordinate(other.value * self.value)

    def __rtruediv__(self, other):
        """
        reverse division: other / self (only happens if the operation (other / self) gives an error)
        """
        other = coordinate(other)
        return coordinate(other.value / self.value)

    def __rfloordiv__(self, other):
        """
        reverse floor division: other // self (only happens if the operation (other // self) gives an error)
        """
        other = coordinate(other)
        return coordinate(other.value // self.value)  

    def __rpow__(self, other):
        """
        reverse power: other ** self (only happens if the operation (other ** self) gives an error)
        """
        raise NotImplementedError('NOT IMPLEMENTED')
        other = coordinate(other)
        return coordinate(other.value ** self.value) 

    def __rmod__(self, other):
        """
        reverse modulo; other % self (only happens if the operation (other % self) gives an error)
        """
        other = coordinate(other)
        return coordinate(other.value % self.value)   

    def __neg__ (self):
        """
        negation
        """
        return coordinate(-self.value)   

    def __pos__ (self):
        """
        positive (unary +)
        """
        return  coordinate(self.value)

    def __abs__(self):
        """
        absolute value
        """
        return coordinate(abs(self.value))

    def invert (self):
        """
        inverting the coordinate: 1 / self
        """
        return coordinate(mp.mpf('1') / self.value)   

    #@workdps(30)
    
    def __eq__(self, other):
        """
        check equality
        """
        if isinstance(other, coordinate):
            return almosteq(self.value, other.value, self.cp)
        elif isinstance(other, mp.mpf):
            return almosteq(self.value, other, self.cp)
        elif isinstance(other, int) or isinstance(other, str):
            return almosteq(self.value, mp.mpf(other), self.cp)
        raise TypeError('not supported type')
        
        
    def __ne__(self, other):
        """
        not equal
        """
        return not self == other

    
    def __lt__(self,other):
        """
        <
        """
        # if isinstance(other, coordinate):
        #     if almosteq(self.value, other.value, self.cp):
        #         return False
        #     return self.value < other.value
        # if almosteq(self.value, mp.mpf(other), self.cp):
        #     return False
        # return self.value < mp.mpf(other)
        
        # if isinstance(other, coordinate):
        #     if self.value < other.value and not almosteq(self.value, other.value, self.cp):
        #         return True
        #     return False
        # elif other is 0:
        #     t = self.c0
        # elif other is 1:
        #     t = self.c1
        # if self.value < t and not almosteq(self.value, t, self.cp):
        #     return True
        # return False
        t = None
        if isinstance(other, coordinate):
            if self.value < other.value and not almosteq(self.value, other.value, self.cp):
                return True
            return False
        elif other is 0:
            t = self.c0
        elif other is 1:
            t = self.c1
        else:
            t = mp.mpf(other)
        if self.value < t and not almosteq(self.value, t, self.cp):
            return True
        return False
        

    
    def __gt__(self,other):
        """
        >
        """
        # if isinstance(other, coordinate):
        #     if almosteq(self.value, other.value, self.cp):
        #         return False
        #     return self.value > other.value
        # if almosteq(self.value, mp.mpf(other), self.cp):
        #     return False
        # return self.value > mp.mpf(other)
        t = None
        if isinstance(other, coordinate):
            if self.value > other.value and not almosteq(self.value, other.value, self.cp):
                return True
            return False
        elif other is 0:
            t = self.c0
        elif other is 1:
            t = self.c1
        else:
            t = mp.mpf(other)
        if self.value > t and not almosteq(self.value, t, self.cp):
            return True
        return False

    def __le__(self,other):
        """
        <=
        """
        if isinstance(other, coordinate):
            if almosteq(self.value, other.value, self.cp):
                return True
            return self.value <= other.value
        if almosteq(self.value, mp.mpf(other), self.cp):
            return True
        return self.value <= mp.mpf(other)

    def __ge__(self,other):
        """
        >=
        """
        if isinstance(other, coordinate):
            if almosteq(self.value, other.value, self.cp):
                return True
            return self.value >= other.value
        if almosteq(self.value, mp.mpf(other), self.cp):
            return True
        return self.value >= mp.mpf(other)

    def __rshift__(self, other):
        """
        >>
        """
        if isinstance(other, coordinate):
            return coordinate(int(self.value) >> int(other.value))
        return coordinate(int(self.value) >> int(other))
        

    def compareValue(self,other):
        """
        comparing values: returns -1 if smaller, 0 if equals, 1 if bigger
        """
        if isinstance(other, coordinate):
            if almosteq(self.value, other.value, self.cp):
                return 0
            elif self.value > other.value:
                return 1
            return -1
        if almosteq(self.value, mp.mpf(other), self.cp):
            return True
        elif self.value > mp.mpf(other):
            return 1
        return -1

c0 = coordinate(0)
c1 = coordinate(1)
c2 = coordinate(2)
cn1 = coordinate(-1)
c0p5 = coordinate('0.5')
