# -*- coding: utf-8 -*-

from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import absolute_import
from builtins import str

from builtins import object
from ..struct.coordinate import c0p5
from ..struct.vector import vector2D, vector3D

class boundary():
    """
    plane segment boundary, defining one boundary of a plane segment, either inner or outer.

    **Object Parameters:** (not including inherited attributes)

    * **canonical** (list of :class:`~sortal.struct.planeStruct.bSegment`) -- list of bSegment defining the boundary, arranged in ascending order.
    * **polygonal** (list of :class:`~sortal.struct.vector.vector2D` or :class:`~sortal.struct.vector.vector3D`) -- list of vetices forming the boundary, arranged in a continuous ordering forming the boundary

    """
    def __init__(self):
        self.canonical = []
        self.polygonal = []
    
    def __len__(self):
        return len(self.canonical)
    
    def __eq__(self, other):
        if len(self.canonical) != len(other.canonical):
            return False
        for i in self.canonical:
            check = False
            for j in other.canonical:
                if i == j:
                    check =True
            if check == False:
                return False
        return True
    
    def printRhino(self, displacement = 0):
        s = ''
        s+='_polyline '
        for j in self.polygonal:
            s += str(j.x + displacement) + ',' + str(j.y) + ' '
        s+= str(self.polygonal[0].x + displacement) + ',' + str(self.polygonal[0].y) + '\n '
        s+='_SelPolyline _PlanarSrf _Delete'
        return s 
        
    def isNil(self):
        pass
        
    def add(self, obj):
        """
        add a vector/bSegment to the boundary according to its type
        """
        if obj.type == 'struct-vector2D' or obj.type == 'struct-vector3D':
            self.polygonal.append(obj)
        else:
            self.canonical.append(obj)
    
    def merge(self, other):
        """
        merge two boundaries
        """
        self.canonical += other.canonical
        self.polygonal += other.polygonal
        
    def sort(self):
        """
        sort the canonical list
        """
        self.canonical.sort()
    
    def partOf(self, other):
        """
        not implemented
        """
        raise ValueError()
        for i in self.canonical:
            check = False
            for j in other.canonical:
                if i == j or i.partOf(j):
                    check = True
            if check == False:
                return False
        return True
       
    def containsBoundary(self, other):
        """
        check if this boundary contains another
        """
        otherCanon = []
        for line in other.canonical:
            otherCanon.append([line, 0])
        
        for point in other.polygonal:
            for oLines in otherCanon:
                if oLines[0].start == point or oLines[0].end == point:
                    oLines[1] += 1
            check = False
            for line in self.canonical:
                if line.containsPoint(point):
                    check = True
                    break
            if check:
                continue
            if not self.containsPointInside(point):
                return False
        for oLine in otherCanon:
            otherLine = oLine[0]

            if oLine[1] == 1:
                continue
            elif oLine[1] == 2:
                midp = otherLine.start.midpoint(otherLine.end)
                check = False
                for line in self.canonical:
                    if line.containsPoint(midp):
                        check = True
                        break
                if check:
                    continue
                if not self.containsPointInside(midp):
                    return False
                continue

            for line in self.canonical:
                if (line.direction.parallel(otherLine.direction)
                or line.start.colinear(otherLine.start, otherLine.end) 
                or line.end.colinear(otherLine.start, otherLine.end)
                or otherLine.start.colinear(line.start, line.end) 
                or otherLine.end.colinear(line.start, line.end)):
                    continue
                if self.intersect2D(line.start, line.end, otherLine.start, otherLine.end):
                    return False

        return True
        
    def containsPointInside(self, point):
        """
        check if a point is contained within the boundary

        \* assumption: point not on any bSegments of the boundary
        """
        maxX = self.polygonal[0].x
        minX = self.polygonal[0].x
        maxY = self.polygonal[0].y
        minY = self.polygonal[0].y
        for i in self.polygonal:
            if maxX < i.x:
                maxX = i.x
            elif minX > i.x:
                minX = i.x
            if maxY < i.y:
                maxY = i.y
            elif minY > i.y:
                minY = i.y
        if minX > point.x or maxX < point.x or minY > point.y or maxY < point.y:
            return False
        start = point
        end = vector2D(point.x, maxY+1)
        intersectionCount = 0
        for i in self.canonical:
            if boundary.intersect2D(start, end, i.start, i.end):
                intersectionCount += 1
        if intersectionCount % 2 == 1:
            return True
        return False
        
            
    @staticmethod
    def cw(A,B,C):
        """
        check if 3 points A, B, C form a counter-clockwise angles (A,B,C are not colinear)
        """
        return (C.y-A.y) * (B.x-A.x) > (B.y-A.y) * (C.x-A.x)
    
    # Return true if line-segments AB and CD intersect
    @staticmethod
    def intersect2D(A, B, C, D):
        """
        check if lines AB and CD intersect each other (no 3 points colinear)
        """
        dir1 = B.subtract(A)
        dir2 = D.subtract(C)
        if dir1.parallel(dir2): 
            return False
        return boundary.cw(A,C,D) != boundary.cw(B,C,D) and boundary.cw(A,B,C) != boundary.cw(A,B,D)
    
    def maximalize(self):
        """
        maximalize the boundary: merging touching colinear lines etc...
        """
        i = 1
        while i < len(self.canonical):
            if (self.canonical[i].start == self.canonical[i-1].end
            and self.canonical[i].start.colinear(self.canonical[i].end, self.canonical[i-1].start)):
                for j in self.polygonal:
                    if j == self.canonical[i].start:
                        self.polygonal.remove(j)
                        break
                newLine = bSegment(self.canonical[i-1].start, self.canonical[i].end)                
                self.canonical[i-1] = newLine
                self.canonical.pop(i)
                continue
            i+=1
        return
        
    def __iter__(self):
        return iter(self.canonical)
    
    def transform(self, matrix):
        """
        transform the boundary with a transformation matrix (similarityMap -- affineTransform)
        """
        result = boundary()
        for i in self.canonical:
            result.canonical.append(i.transform(matrix))
        for i in self.polygonal:
            result.polygonal.append(i.transform(matrix))
        minIndex = result.polygonal.index(min(result.polygonal))
        result.polygonal = result.polygonal[minIndex:] + result.polygonal[:minIndex]
        return result
        
    def graphTransform(self, graphTrans):
        """
        transform the boundary with a graph transform
        """
        result = boundary()
        for i in self.polygonal:
            check = graphTrans.transform(i)
            if check is False:
                return False
            result.polygonal.append(check)
        i = 0
        while i < len(result.polygonal):
            if result.polygonal[i] == result.polygonal[i-1]:
                result.polygonal.pop(i)
            else:
                i += 1
        for i in self.canonical:
            bs = i.graphTransform(graphTrans)
            if bs.start == bs.end:
                continue
            addCheck = True
            for j in result.canonical:
                if bs == j:
                    addCheck = False
                    break
                if bs.start == j.start or bs.start == j.end or bs.end == j.start or bs.end == j.end:
                    continue
                if bs.intersect(j):
                    pass
                    #raise PlaneTransformationError(self)
            if addCheck:
                result.canonical.append(bs)
        i = 0
        while len(result.polygonal) > 2 and i < len(result.polygonal):
            if result.polygonal[i].colinear(result.polygonal[i-1], result.polygonal[i-2]):
                bI = 0
                while bI < len(result.canonical):
                    bs = result.canonical[bI]
                    if bs.start == result.polygonal[i-1] or bs.end == result.polygonal[i-1]:
                        result.canonical.pop(bI)
                        continue
                    bI += 1
                result.canonical.append(bSegment(result.polygonal[i-2], result.polygonal[i]))
                result.polygonal.pop(i-1)
            else:
                i+=1
        if len(result.canonical) < 3:
            raise PlaneTransformationError(self)
        return result
    
    @staticmethod
    def parse(reader, sort):
        raise NotImplementedError
        
        
class bSegment(object):
    """
    boundary segment class. Similar to lineSegment2D and lineSegment3D in representing line-segment.

    bSegment should only be used in boundary objects and the classification process of plane segments.

    **Object Parameters:** (not including inherited attributes)

    * **start** -- start point of the segment.
    * **end** -- end point of the segment.
    * **direction** -- direction vector of the segment.
    * **assoc** -- the associated boundary.
    * **alignedPositive** (boolean) -- alignment with another plane segment, used in classification process.

    """
    POS_X_INF = vector2D(1,0,True)
    POS_Y_INF = vector2D(0,1,True)
    NEG_Y_INF = vector2D(0,-1,True)
    
    POS_X_INF_3D = vector3D(1,0,0,True)
    POS_Y_INF_3D = vector3D(0,1,0,True)
    NEG_Y_INF_3D = vector3D(0,-1,0,True)
    
    INFINITE = -1
    CLOSED = 0
    OPEN = 1

    type = 'struct-bSegment'

    def __init__(self, start, end, assoc = None):
        if 'ind' in start.type or 'ind' in end.type:
            raise ValueError('start and end must be vectors')
        if start.compareValue(end) == 1:
            self.start = end
            self.end = start
        else:
            self.start = start
            self.end = end
        self.direction = self.end.subtract(self.start)
        self.assoc = assoc
        self.alignedPositive = False
        
    def __str__(self):
        return 'bs'+str(self.start)+str(self.end)
        
    def __repr__(self):
        return 'bs'+str(self.start)+str(self.end)
        
    def printRhino(self, string = '', displacement = 0):
        s = '_line '
        s+= self.start.printRhino().split(' ')[1]+' '+self.end.printRhino().split(' ')[1]
        return s
        
    def root(self):
        """
        find the root of the line
        """
        return self.start.subtract(self.direction.scale(self.start))
    
    def length(self):
        """
        find the length of the segment
        """
        direction = self.end.subtract(self.start)
        return direction.dotproduct(direction) ** c0p5
        
    def squareLength(self):
        """
        find the square length of the segment
        """
        direction = self.end.subtract(self.start)
        return direction.dotproduct(direction)
        
    def isInfinite(self):
        """
        check whether the segment is infinite
        """
        return self.start.isInfinite() or self.end.isInfinite()
    
    def aligns(self, other):
        """
        check whether two bSegments align
        """
        return (self.end.colinear(other.start, other.end) and 
                self.direction.parallel(other.direction))
                
    def partOf(self, other):
        """
        check whether the bSegment is part of another
        """
        return (self.end.colinear(other.start, other.end) and 
                self.direction.parallel(other.direction) and
                self.start >= other.start and self.end <= other.end)
                
    def complement(self, other):
        """
        find the complement segment with another bSegment
        """
        result = []
        if self.start < other.start and self.end > other.start:
            result.append(bSegment(self.start, other.start))
        if self.start < other.end and self.end > other.end:
            result.append(bSegment(other.end, self.end))
        return result
    
    def __eq__(self, other):
        if other is None or other.__class__ != self.__class__:
            return False
        return (self.start == other.start) and (self.end == other.end)
            
    def __lt__(self, other):
        return self.compareValue(other) == -1
        
    def __gt__(self, other):
        return self.compareValue(other) == 1
        
    def __le__(self, other):
        return self.compareValue(other) < 1
        
    def __ge__(self, other):
        return self.compareValue(other) > -1
    
    def compareValue(self, other):
        c = self.direction.unit().compareValue(other.direction.unit()) 
        if c != 0:
            return c
        c = self.root().compareValue(other.root()) 
        if c != 0:
            return c
        c = self.start.compareValue(other.start) 
        if c != 0:
            return c
        return self.end.compareValue(other.end) 
        
    def disjoint(self, other):
        """
        check whether two bSegments are disjoint
        """
        return self.direction.parallel(other.direction) and self.root() == other.root() and (self.end.compareValue(other.start) == -1 or self.start.compareValue(other.end) == 1)
        
    def transform(self, matrix):
        """
        transform the bSegment with an affineTransform
        """
        return bSegment(self.start.transform(matrix), self.end.transform(matrix))
        
    def graphTransform(self, transf):
        """
        transform the bSegment with a graphTransform
        """
        return bSegment(transf.apply(self.start), transf.apply(self.end))
        
    def containsPoint(self, point):
        """
        check if the bSegment contains a point
        """
        if self.start == point or self.end == point:
            return True
        if not point.colinear(self.start,self.end):
            return False
        if self.start.compareValue(point) < 1 and self.end.compareValue(point) > -1:
            return True
        elif self.start.compareValue(point) > -1 and self.end.compareValue(point) < 1:
            return True
        return False
    
    @staticmethod
    def cw(A,B,C):
        """
        check if 3 points A, B and C forms a counter-clockwise angle (cannot be colinear)
        """
        return (C.y-A.y) * (B.x-A.x) > (B.y-A.y) * (C.x-A.x)
    
    def intersect(self, other):
        """
        check if two bSegments intersect
        """
        if self.start.type == 'struct-vector2D':
            if other is None or type(other) != type(self) :
                return False
            if self.direction.parallel(other.direction):
                return False
            if ((self.start.colinear(other.start, other.end) and other.start < self.start < other.end)
            or (self.end.colinear(other.start, other.end) and other.start < self.end < other.end)
            or (other.start.colinear(self.start, self.end) and self.start < other.start < self.end)
            or (other.end.colinear(self.start, self.end) and self.start < other.end < self.end)):
                return True
            A = self.start
            B = self.end
            C = other.start
            D = other.end
            return self.cw(A,C,D) != self.cw(B,C,D) and self.cw(A,B,C) != self.cw(A,B,D)
        if self.direction.parallel(other.direction):
            return False
        i = self.intersectionLine(other.root(), other.direction)
        if self.start < i < self.end and other.start < i < other.end:
            return True
        return False
    
    def intersection(self, other, type = CLOSED):
        """
        find the intersection between two bSegments. bSegment.intersect must be checked first
        """
        if self.isInfinite() or other.isInfinite():
            return None
        dir1 = self.direction
        dir2 = other.direction

        if dir1.parallel(dir2): 
            return None

        cross = dir1.product(dir2)
        diff = other.start.subtract(self.start)
        
        if self.start.type == 'struct-vector2D' or self.start.type == 'struct-vector2Dfloat':
            par1 = diff.product(dir2) / cross
            par2 = diff.product(dir1) / cross
        else:
            denom = cross.dotproduct(cross)
            par1 = diff.product(dir2).dotproduct(cross) / denom
            par2 = diff.product(dir1).dotproduct(cross) / denom
            
        if (par2>=0 and par2<=1 and (type == bSegment.INFINITE or 
            (type == bSegment.CLOSED and par1 >= 0 and par1 <= 1) or
            (type == bSegment.OPEN and par1 > 0 and par1< 1))):
            return self.start.add(dir1.scale(par1))
        return None


    def intersectionLine(self, root, direction):
        """
        check if the bSegment intersect another infinite line
        """
        selfUnit = self.direction.unit()
        otherUnit = direction.unit()
        selfRoot = self.root()
        denom = selfUnit.product(otherUnit)
        diff = root.subtract(selfRoot)
        numer1 = diff.product(otherUnit)
        if denom.x != 0:
            x = numer1.x/denom.x
        elif denom.y != 0:
            x = numer1.y/denom.y
        else:
            x = numer1.z/denom.z
        return selfRoot.add(selfUnit.scale(x))

class PlaneTransformationError(Exception):
    def __init__(self, boundary):
        s = ('Error while transforming plane-segment: '+ str(boundary.printRhino()))
        super(PlaneTransformationError, self).__init__(s)

     


