# -*- coding: utf-8 -*-

from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import

from builtins import super
from builtins import str
from builtins import round

from builtins import object

from .coordinate import coordinate
from .vector import vector2D, vector3D

v0_2D = vector2D(0,0)
v0_3D = vector3D(0,0,0) 

class sSeg(object):
    """
    splay segment for plane segment classification process.

    contains a bSegment defining the line-segment along with other properties for alignment checking

    **Object Parameters:** (not including inherited attributes)

    * **segment** (list of :class:`~sortal.struct.planeStruct.bSegment`) -- bSegment defining the line.
    * **membership** (boolean) -- whether the splay segment is on coform1 or coform2: True -> sSeg belongs to coform1; False -> sSeg belongs to coform2
    * **alignedWithOne** (boolean) -- whether the splay segment align with coform1
    * **alignedWithOther** (boolean) -- whether the splay segment align with coform2
    
    """
    type = 'struct-sSeg'
    def __init__(self, segment, other = None):
        """
        :param segment: bSegment defining the line
        :type segment: ~sortal.struct.planeStruct.bSegment
        :param other: (optional) the original sSeg this sSeg is part of; this initialization is for when an sSeg is split into two separate sSegs
        :type other: sSeg
        """
        if segment.type != 'struct-bSegment':
            print(segment.type)
            raise ValueError('segment is not a bSegment')
        if other is not None:
            self.segment = segment
            self.membership = other.membership
            self.alignedWithOne = other.alignedWithOne
            self.alignedWithOther = other.alignedWithOther
        else:
            self.segment = segment
            self.membership = False
            self.alignedWithOne = False
            self.alignedWithOther = False
            
    def __str__(self):
        return 'ss-'+str(self.segment)
            
    def splayCompare(self, other):
        """
        comparing two splay segments
        """
        if self.segment.start.type == 'struct-vector2D':
            
            if (other.type == 'struct-vector2D'):
                return self.segment.direction.product(other.subtract(self.segment.start))
            elif not other.type == 'struct-sSeg':
                raise TypeError('other must be either splaySegment or vector')
            c = (self.segment.direction.product(other.segment.start.subtract(self.segment.start)))
            if c != 0:
                return c
            return (self.segment.direction.product(other.segment.end.subtract(self.segment.start)))
        else:
            if other.type == 'struct-vector3D':
                return (self.segment.direction.productSign(other.subtract(self.segment.start)))
            elif not other.type == 'struct-sSeg':
                raise TypeError('other must be either splaySegment or vector')
            c = self.segment.direction.productSign(other.segment.start.subtract(self.segment.start))
            if c != 0:
                return c
            return self.segment.direction.productSign(other.segment.end.subtract(self.segment.start))
            

class sRef(object):
    """
    splay reference class
    """
    START_POSITION = 0
    END_POSITION = 1
    INTERSECTION = 2
    NOT_PART_OF_CYCLE = 0
    PART_OF_CYCLE = 1

    type = 'struct-sRef'
    
    def __init__(self, position, segment, tp):
        if 'ind' in position.type:
            raise ValueError('position must not be an individual')
        self.position = position
        self.segment = segment
        self.posType = tp
        
    def __str__(self):
        return 'ref'+str(self.position) +'..'+ str(self.segment)
        
    def isStartPosition(self):
        return self.posType == sRef.START_POSITION
        
    def isEndPosition(self):
        return self.posType == sRef.END_POSITION
        
    def isIntersection(self):
        return self.posType == sRef.INTERSECTION
        
    def isPartOfCycle(self):
        return self.posType == sRef.PART_OF_CYCLE

    @staticmethod
    def startPosition(seg):
        return sRef(seg.segment.start, seg, sRef.START_POSITION)
        
        
    def setPartOfCycle(self, partOfCycle):
        if (partOfCycle):
            self.posType = sRef.PART_OF_CYCLE
        else: self.posType = sRef.NOT_PART_OF_CYCLE

    @staticmethod
    def endPosition(seg):
        return sRef(seg.segment.end, None, sRef.END_POSITION)

    @staticmethod
    def intersection(vect):
        return sRef(vect, None, sRef.INTERSECTION)

    def splayCompare(self, other):
        if 'struct-s' in self.position.type:
            if other.type == 'struct-sRef':
                return self.position.splayCompare(other.position)
            else: 
                return self.position.splayCompare(other)
        if other.type == 'struct-sRef':
            return self.position.compareValue(other.position)
        return self.position.compareValue(other)
        
class vertex(sRef):
    vertices = {}
    type = 'struct-sVertex'
    
    def __init__(self, vect, *args, **kwargs):
        if (vect.type != 'struct-vector2D' and vect.type != 'struct-vector3D'):
            raise ValueError('vect is not a vector')
        super(vertex, self).__init__(vect, None, 0)
        self.edges = []
        if vect.type == 'struct-vector2D' :
            string = str(vect.x)+'|'+str(vect.y)
        else:
            string = str(vect.x)+'|'+str(vect.y)+'|'+str(vect.z)
        vertex.vertices[string] = self
        #vertex.vertices[vect] = self

    def __str__(self):
        s = str(self.position)+'-{'
        for i in self.edges:
            s+=str(i)+','
        if s[-1] == ',':
            s = s[:-1]+'}'
        else:
            s = s+'}'
        return s
    
    def __repr__(self):
        return str(self)
        
    def lastVisited(self):
        return self.segment
        
    def setLastVisited(self, interval):
        self.segment = interval
        
    def __eq__(self, other):
        if other.type != 'struct-sVertex':
            return False
        return self.position == other.position
        
    def compareValue(self, other):
        return 0
