# -*- coding: utf-8 -*-

from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import absolute_import
from builtins import str

from builtins import object
from .coordinate import coordinate, c0, c1, c0p5, c2, acos, pi, atan2
from ..io.parseReader import parseReader, ParsingError

class vector2D(object):
    """
    A 2D vector specifies a position in a two-dimensional cartesian space.
    If normalized, it only specifies a direction.

    The 2D Vector class defines a vector as a pair of coordinates
    and a w factor to reflect the vector's infinity characteristic
    A 2D Vector object is never modified after creation, thus,
    it can be used multiple times within multiple different objects.

    **Object Parameters:** (not including those inherited from plane3D)

    * **x** (:class:`~sortal.struct.coord.coordinate_float.coordinate`) -- x-coordinate of the vector.
    * **y** (:class:`~sortal.struct.coord.coordinate_float.coordinate`) -- y-coordinate of the vector.
    * **w** (int) -- indicator of infinity: 1 -> not an infinity vector; 0 -> infinity vector with direction(x,y).

    """

    type = 'struct-vector2D'
    
    def __init__(self, x, y, infinity = False):
        """
        initialize the 2D vector

        :param x: x-coordinate of the vector
        :type x: int or float or :class:`~sortal.struct.coord.coordinate_float.coordinate`
        :param y: y-coordinate of the vector
        :type y: int or float or :class:`~sortal.struct.coord.coordinate_float.coordinate`
        :param infinity: whether the vector is infinite
        :type infinity: boolean
        """
        if infinity:
            self.w = 0
        else:
            self.w = 1
        # if c0 == x:
        #     self.x = c0
        # else:
        #     self.x = coordinate(x)
        # if c0 == y:
        #     self.y = c0
        # else:
        #     self.y = coordinate(y)
        self.x = coordinate(x)
        self.y = coordinate(y)

    def __str__(self):
        """
        return the string that represents this vector's values
        """       
        if self.w == 0:
            return ' inf('+str(self.x)+','+str(self.y)+')'
        return '('+str(self.x)+','+str(self.y)+')'
        
    def __repr__(self):
        """
        representation of the vector
        """       
        return self.__str__()
        
    def __hash__(self):
        s = str(self.x)+' '+str(self.y)
        return hash(s)
        
    def __abs__(self):
        """
        Absolute value of the vector
        """
        if self.sign() >= 0:
            return self
        else:
            return vector2D(-self.x, -self.y)
        
    def printVal(self, displacement = 0):
        if self.w == 0:
            return 'inf('+self.x+','+self.y+')'
        return '('+str(self.x+displacement)+','+str(self.y)+')'
        
    def printRhino(self, string = '', displacement = 0):
        return '_Point '+str(self.x + displacement)+','+str(self.y)
    
    def sign(self):
        """
        find the sign of the vector
        the vector is positive if the x value is positive
        the vector is negative if the x value is negative
        if x value is 0, the sign of the vector is decided by the y value

        :returns: int -- 1 if positive, -1 if negative and 0 if both x and y are 0
        """
        if self.x > 0:
            return 1
        if self.x < 0:
            return -1
        if self.y > 0:
            return 1
        if self.y < 0:
            return -1
        return 0
    
    def isZero(self):
        """
        check if this vector is a zero vector: (0,0)
        """
        return self.x == c0 and self.y == c0
        
    def parallel(self,other):
        """
        checking if two vectors are parallel
        two vectors are parallel if: this.x / other.x = this.y / other.y --> this.x * other.y = this.y * other.x
        """
        if other.type != 'struct-vector2D':
            return False
        return (self.x * other.y) == (self.y * other.x)
        
    def perpendicular(self,other):
        """
        checking if two vectors are perpendicular with one another
        two vectors are perpendicular if their dot product value is 0
        """
        if other.type != 'struct-vector2D':
            return False
        return self.dotproduct(other) == 0
        
    def isInfinite(self):
        return self.w == 0
        
    def __eq__(self, other):
        """
        checking if this vector is equal to other vector
        two vectors are equal if both have equal x values, equal y values and equal z values
        """
        if other is None or other.__class__ != self.__class__:
            return False
        return self.x == other.x  and self.y  == other.y and self.w == other.w
        
    def __ne__(self, other):
        return not self == other
 
        
    def __lt__(self, other):
        return self.compareValue(other) == -1
        
    def __le__(self, other):
        return self.compareValue(other) < 1
        
    def __gt__(self, other):
        return self.compareValue(other) == 1
        
    def __ge__(self, other):
        return self.compareValue(other) > -1

    def compareValue(self, other):
        """
        comparing the value of this vector with another vector
        a vector is greater if its x value is greater and vice versa.
        If the x values of the two vectors are the same, the vector with the 
        greater y value is greater
        
        :returns: int -- 1 if this vector is greater, 0 if equals, -1 if lesser
        """
        if self.w == other.w:
            if self.x > other.x:
                return 1
            elif self.x < other.x:
                return -1
            elif self.y > other.y:
                return 1
            elif self.y < other.y:
                return -1
            return 0
        elif self.w == 0:
            if self.x < 0:
                return -1
            elif self.x > 0:
                return 1
            elif other.x < 0:
                return 1
            elif other.x > 0:
                return -1
            else:
                if self.y < 0:
                    return -1
                elif self.y > 0:
                    return 1
                elif other.y < 0:
                    return 1
                elif other.y > 0:
                    return -1
                else:
                    return 0
        elif other.w == 0:
            if other.x < 0:
                return 1
            elif other.x > 0:
                return -1
            elif self.x < 0:
                return -1
            elif self.x > 0:
                return 1
            else:
                if other.y < 0:
                    return 1
                elif other.y > 0:
                    return -1
                elif self.y < 0:
                    return -1
                elif self.y > 0:
                    return 1
                else:
                    return 0
    
    def scale(self, other):
        """
        scaling the vector by a certain factor.
        
        * If factor is a number: multiply the vector's x and y with the factor.
        * If factor is another vector: multiply the vector's x and y with the projection length of the other vector
        """    
        if isinstance(other, coordinate) or isinstance(other, int):
            return vector2D(self.x*other, self.y*other)
        else:
            scalar = (other.dotproduct(self)) / (self.dotproduct(self))
            return vector2D(self.x*scalar, self.y*scalar)
            
    def abs(self):
        """
        find the vector with absolute values of this vector's x and y
        """
        return self.scale(self.sign())
                
    def inverse(self):
        """
        find the vector that is a reflection of this vector by the origin
        """
        return vector2D(-self.x, -self.y)
        
    def length(self):
        """
        find the length of the vector
        """
        return (self.x * self.x + self.y *self.y) ** c0p5
        
    def squareLength(self):
        """
        find the squared length of the vector.

        # if length equality is to be checked, use squared length since it is more 
        time efficient (no square root calculation)
        """
        return (self.x * self.x + self.y *self.y)
        
    def unit(self):
        """
        find the unit vector of the vector: dividing the vector's x and y by the length of the vector
        """
        a = c1 / self.length()
        return vector2D(self.x * a, self.y * a)
    
    def add(self,other):
        """
        adding two vector

        :returns: :class:`~sortal.struct.vector.vector2D` -- new result vector
        """
        if self.w == 1 and other.w == 1:
            return vector2D(self.x+other.x, self.y+other.y)
        if self.w == 0:
            if other.w == 0:
                raise ValueError('Adding two infinite vectors is undefined')
            else:
                return self
        return other
    
    def subtract(self, other):
        """
        subtracting the vector by another vector

        :returns: :class:`~sortal.struct.vector.vector2D` -- new result vector
        """
        if self.w == 1 and other.w == 1:
            return vector2D(self.x-other.x, self.y-other.y)
        elif self.w == 0:
            if other.w == 0 and self != other.inverse():
                return vector2D(self.x-other.x, self.y-other.y, True)
            else:
                return self
        return other
        
    def dotproduct(self, other):
        """
        find the dot product of the vector with another

        :returns: coordinate -- resulting dot product
        """
        return self.x * other.x + self.y * other.y
        
    def product(self, other):
        """
        Find the cross product of the vector with another.
        Since the vectors are in 2D space, the vectors will be (x0,y0,0) and (x1,y1,0)
        in 3D space --> the product vector will be (0, 0, x0*y1-x1*y0).

        This function returns only the z value of the product vector since the vector
        itself cannot be represented in the 2D space

        :returns: coordinate -- resulting z-coordinate of the product vector
        """
        return self.x * other.y - self.y * other.x
    
    def scalar(self, other):
        """    
        """
        return other.dotproduct(self) / self.dotproduct(self)
    
    def colinear(self, other1, other2):
        """
        check if this vector and 2 other vectors are colinear
        """
        v1 = other1.subtract(self)
        v2 = other2.subtract(self)
        return (v1.x * v2.y) == (v1.y * v2.x)
        
    def midpoint(self, other):
        """
        find the midpoint vector between two vectors
        """
        return vector2D((self.x+other.x)/c2,(self.y+other.y)/c2)
    
    def angle(self, other):
        """
        find the clockwise angle from this vector to another vector
        """
        a = self.unit()
        b = other.unit()
        return (atan2(b.y.value, b.x.value) - atan2(a.y.value, a.x.value))
        c = atan2(b.y.value, b.x.value) - atan2(a.y.value, a.x.value)
        if (c < 0 and c > -pi) or (c > pi and c < 2*pi):            
            return - acos(self.unit().dotproduct(other.unit()).value)
        return acos(self.unit().dotproduct(other.unit()).value)
                
    def transform(self, transf):
        """
        transforming the vector with a transformation matrix

        matrix: [[a, b, c], [d, e, f]]

        this vector: (x, y)

        => vector after transforming: (a*x + b*y + c, d*x + e*y + f)
        """
        return vector2D((self.x*transf.matrix[0][0] + self.y*transf.matrix[0][1] + transf.matrix[0][2]),
                        (self.x*transf.matrix[1][0] + self.y*transf.matrix[1][1] + transf.matrix[1][2]))
                 
    @staticmethod
    def readVector(reader):
        """
        parsing vector from a string

        :param reader: the reader of the string to be parsed
        :type reader: ~sortal.io.parseReader.parseReader
        :returns: (coordinate, coordinate) -- tuple of 2 coordinate objects representing the x and y coordinates of the vector
        """
        if reader.newToken() == parseReader.MINUS:
            neg = '-'
            reader.newToken()
        else:
            neg = ''
        x = coordinate(neg+reader.value())
        
        reader.newToken() # comma
        
        if reader.newToken() == parseReader.MINUS:
            neg = '-'
            reader.newToken()
        else:
            neg = ''
        y = coordinate(neg+reader.value())
        
        reader.newToken() # bracket
        
        reader.newToken() # next token
        return x,y
                                                

    
class vector3D(object):
    """
    A 3D vector specifies a position in a three-dimensional cartesian space.
    If normalized, it only specifies a direction.

    The 3D Vector class defines a vector as a set of three coordinates
    and a w factor to reflect the vector's infinity characteristic
    A 3D Vector object is never modified after creation, thus,
    it can be used multiple times.

    **Object Parameters:** (not including those inherited from plane3D)

    * **x** (:class:`~sortal.struct.coord.coordinate_float.coordinate`) -- x-coordinate of the vector.
    * **y** (:class:`~sortal.struct.coord.coordinate_float.coordinate`) -- y-coordinate of the vector.
    * **z** (:class:`~sortal.struct.coord.coordinate_float.coordinate`) -- z-coordinate of the vector.
    * **w** (int) -- indicator of infinity: 1 -> not an infinity vector; 0 -> infinity vector with direction(x,y).
    """

    type = 'struct-vector3D'
    def __init__(self, x, y, z=0, infinity = False):
        """
        initialize the 3D vector

        :param x: x-coordinate of the vector
        :type x: int or float or :class:`~sortal.struct.coord.coordinate_float.coordinate`
        :param y: y-coordinate of the vector
        :type y: int or float or :class:`~sortal.struct.coord.coordinate_float.coordinate`
        :param z: z-coordinate of the vector
        :type z: int or float or :class:`~sortal.struct.coord.coordinate_float.coordinate`
        :param infinity: whether the vector is infinite
        :type infinity: boolean
        """
        if infinity:
            self.w = 0
        else:
            self.w = 1
        # if c0 == x:
        #     self.x = c0
        # else:
        #     self.x = coordinate(x)
        # if c0 == y:
        #     self.y = c0
        # else:
        #     self.y = coordinate(y)
        # if c0 == z:
        #     self.z = c0
        # else:
        #     self.z = coordinate(z)
        self.x = coordinate(x)
        self.y = coordinate(y)
        self.z = coordinate(z)
            
    def __str__(self):
        """
        return the string that represents this vector's values
        """       
        if self.w == 0:
            return ' inf('+str(self.x)+','+str(self.y)+','+str(self.z)+')'
        return '('+str(self.x)+','+str(self.y)+','+str(self.z)+')'
        
    def __repr__(self):
        """
        representation of the vector
        """       
        return self.__str__()
        
    def __hash__(self):
        s = str(self.x)+' '+str(self.y)+' '+str(self.z)
        return hash(s)
        
    def __abs__(self):
        """
        representation of the vector
        """       
        if self.sign() >= 0:
            return self
        else:
            return vector3D(-self.x, -self.y, -self.z)
        
    def printVal(self, displacement = 0):
        if self.w == 0:
            return 'inf('+self.x+','+self.y+','+self.z+')'
        return '('+str(self.x+displacement)+','+self.y+','+self.z+')'
        
    def printRhino(self, string = '', displacement = 0):
        return '_Point '+str(self.x + displacement)+','+str(self.y)+','+str(self.z)
    

    def sign(self):
        """
        find the sign of the vector
        the vector is positive if the x value is positive
        the vector is negative if the x value is negative
        if x value is 0, the sign of the vector is decided by the y value
        if x and y values are 0, the sign of the vector is decided by the z value

        :returns: int -- 1 if positive, -1 if negative and 0 if both x and y are 0
        """
        if self.x > 0:
            return 1
        if self.x < 0:
            return -1
        if self.y > 0:
            return 1
        if self.y < 0:
            return -1
        if self.z > 0:
            return 1
        if self.z < 0:
            return -1
        return 0
    
    def isZero(self):
        """
        check if this vector is a zero vector: (0,0,0)
        """
        return self.x.isZero() and self.y.isZero() and self.z.isZero()
        
    def parallel(self,other):
        """
        checking if two vectors are parallel
        """
        if other.type != 'struct-vector3D':
            return False
        if self.x == 0 and other.x == 0:
            lhs = (self.z * other.y)
            rhs = (self.y * other.z)
            return lhs == rhs
            
        lhs1 = (self.x * other.y)
        rhs1 = (self.y * other.x)
        
        lhs2 = (self.x * other.z)
        rhs2 = (self.z * other.x)
        
        return lhs1 == rhs1 and lhs2 == rhs2
        
    def perpendicular(self,other):
        """
        Checking if two vectors are perpendicular with one another.
        Two vectors are perpendicular if their dot product value is 0
        """
        if other.type != 'struct-vector3D':
            return False
        return self.dotproduct(other) == 0
        
    def isInfinite(self):
        return self.w == 0
        
    def __eq__(self, other):
        """
        Checking if this vector is equal to other vector.
        Two vectors are equal if both have equal x values, equal y values, equal z values and equal w values
        """
        if other is None or other.__class__ != self.__class__:
            return False
        return (self.x == other.x  and self.y  == other.y and 
                self.z  == other.z and self.w == other.w)
        
    def __ne__(self, other):
        return not self == other
 
    def __lt__(self, other):
        return self.compareValue(other) == -1
        
    def __le__(self, other):
        return self.compareValue(other) < 1

    def __gt__(self, other):
        return self.compareValue(other) == 1
        
    def __ge__(self, other):
        return self.compareValue(other) > -1

    def compareValue(self, other):
        """
        comparing the value of this vector with another vector
        a vector is greater if its x value is greater and vice versa
        if the x values of the two vectors are the same, the vector with the 
        greater y value is greater, if y values are the same, the vector with the 
        greater z value is greater
        
        :returns: int -- 1 if this vector is greater, 0 if equals, -1 if lesser
        """
        if self.w == other.w:
            if self.x > other.x:
                return 1
            elif self.x < other.x:
                return -1
            elif self.y > other.y:
                return 1
            elif self.y < other.y:
                return -1
            elif self.z > other.z:
                return 1
            elif self.z < other.z:
                return -1
            return 0
        elif self.w == 0:
            if self.x < 0:
                return -1
            elif self.x > 0:
                return 1
            elif other.x < 0:
                return 1
            elif other.x > 0:
                return -1
            else:
                if self.y < 0:
                    return -1
                elif self.y > 0:
                    return 1
                elif other.y < 0:
                    return 1
                elif other.y > 0:
                    return -1
                else:
                    if self.z < 0:
                        return -1
                    elif self.z > 0:
                        return 1
                    elif other.z < 0:
                        return 1
                    elif other.z > 0:
                        return -1
                    else:
                        return 0
        elif other.w == 0:
            if other.x < 0:
                return 1
            elif other.x > 0:
                return -1
            elif self.x < 0:
                return -1
            elif self.x > 0:
                return 1
            else:
                if other.y < 0:
                    return 1
                elif other.y > 0:
                    return -1
                elif self.y < 0:
                    return -1
                elif self.y > 0:
                    return 1
                else:
                    if other.z < 0:
                        return 1
                    elif other.z > 0:
                        return -1
                    elif self.z < 0:
                        return -1
                    elif self.z > 0:
                        return 1
                    else:
                        return 0
    
    def scale(self, other):
        """
        scaling the vector by a certain factor

        * if factor is a number: multiply the vector's x, y and z with the factor
        * if factor is another vector: multiply the vector's x, y and z with the projection length of the other vector
        """    
        if isinstance(other, coordinate) or isinstance(other, int):
            return vector3D(self.x*other, self.y*other, self.z*other)
        else:
            scalar = (other.dotproduct(self)) / (self.dotproduct(self))
            return vector3D(self.x*scalar, self.y*scalar, self.z*scalar)
            
    def abs(self):
        """
        find the vector with absolute values of this vector's x, y and z
        """
        return self.scale(self.sign())
        
        
    def inverse(self):
        """
        find the vector that is a reflection of this vector by the origin
        """
        return vector3D(-self.x, -self.y, -self.z)
        
    def length(self):
        """
        find the length of the vector
        """
        return (self.x * self.x + self.y *self.y + self.z *self.z) ** c0p5
        
    def squareLength(self):
        """
        find the squared length of the vector

        # if length equality is to be checked, use squared length since it is more 
        time efficient (no need to calculate square root)
        """
        return (self.x * self.x + self.y *self.y + self.z *self.z)
        
    def unit(self):
        """
        find the unit vector of the vector: dividing the vector's x and y by the length of the vector
        """
        a = c1 / self.length()
        return vector3D(self.x * a, self.y * a, self.z * a)
    
    def add(self,other):
        """
        adding two vector
        """
        if self.w == 1 and other.w == 1:
            return vector3D(self.x+other.x, self.y+other.y, self.z+other.z)
        if self.w == 0:
            if other.w == 0:
                raise ValueError('Adding two infinite vectors is undefined')
            else:
                return self
        return other
    
    def subtract(self, other):
        """
        find the subtraction between this vector by another vector
        """
        if self.w == 1 and other.w == 1:
            return vector3D(self.x-other.x, self.y-other.y, self.z-other.z)
        if self.w == 0:
            if other.w == 0 and self != other.inverse():
                return vector3D(self.x-other.x, self.y-other.y, self.z-other.z, True)
            else:
                return self
        return other
        
    def dotproduct(self, other):
        """
        find the dot product of the vector with another
        """
        return self.x * other.x + self.y * other.y + self.z * other.z
        
    def product(self, other):
        """
        find the product of the vector with another
        """
        return vector3D(self.y * other.z - self.z * other.y,
                        self.z * other.x - self.x * other.z,
                        self.x * other.y - self.y * other.x)
                        
    def productSign(self, other):
        """
        find the sign of the cross product of this vector with another vector
        """
        c = self.y * other.z - self.z * other.y
        if c > 0:
            return 1
        elif c < 0:
            return -1
        c = self.z * other.x - self.x * other.z
        if c > 0:
            return 1
        elif c < 0:
            return -1
        c = self.x * other.y - self.y * other.x
        if c > 0:
            return 1
        elif c < 0:
            return -1
        return 0
    
    def scalar(self, other):
        """    
        """
        return other.dotproduct(self) / self.dotproduct(self)
    
    def colinear(self, other1, other2):
        """
        check if this vector and 2 other vectors are colinear
        """
        return other1.subtract(self).parallel(other2.subtract(self))
        
    def midpoint(self, other):
        """
        find the midpoint vector between two vectors
        """
        return vector3D((self.x+other.x)/c2, (self.y+other.y)/c2, (self.z+other.z)/c2)
        
    
    
    # def angle(self, other):
    #     a = self.unit()
    #     b = other.unit()
    #     #return (atan2(b.y, b.x) - atan2(a.y, a.x))
        
    #     c = atan2(b.y, b.x) - atan2(a.y, a.x)
    #     if (c < 0 and c > -pi) or (c > pi and c < 2*pi):            
    #         return - acos(self.unit().dotproduct(other.unit()).value)
    #     return acos(self.unit().dotproduct(other.unit()).value)

    def transform(self, transf):
        """
        transforming the vector with a transformation matrix.

        matrix: [[a, b, c, d], [e, f, g, h], [i, j, k, l]]

        this vector: (x, y, z)
        => vector after transforming: (a*x + b*y + c*z +d, e*x + f*y + g*z +h, i*x + j*y + k*z +l)
        """
        return vector3D((self.x*transf.matrix[0][0] + self.y*transf.matrix[0][1] + self.z*transf.matrix[0][2] + transf.matrix[0][3]),
                        (self.x*transf.matrix[1][0] + self.y*transf.matrix[1][1] + self.z*transf.matrix[1][2] + transf.matrix[1][3]),
                        (self.x*transf.matrix[2][0] + self.y*transf.matrix[2][1] + self.z*transf.matrix[2][2] + transf.matrix[2][3]))
            
    @staticmethod
    def readVector(reader):
        """
        parsing vector from a string

        :param reader: the reader of the string to be parsed
        :type reader: ~sortal.io.parseReader.parseReader
        :returns: (coordinate, coordinate, coordinate) -- tuple of 3 coordinate objects representing the x, y and z coordinates of the vector
        """
        if reader.newToken() == parseReader.MINUS:
            neg = '-'
            reader.newToken()
        else:
            neg = ''
            
        x = coordinate(neg+reader.value())
        
        reader.newToken() # comma
        
        if reader.newToken() == parseReader.MINUS:
            neg = '-'
            reader.newToken()
        else:
            neg = ''
        y = coordinate(neg+reader.value())
        
        if reader.newToken() == ')': # should be comma, if it's ) => 2D vector
            return x, y, None
        
        if reader.newToken() == parseReader.MINUS:
            neg = '-'
            reader.newToken()
        else:
            neg = ''
        z = coordinate(neg+reader.value())
        
        reader.newToken() # bracket

        reader.newToken() # next token
        return x,y,z
            