# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import range
from builtins import super
from builtins import str
from builtins import object


from ..struct.vector import vector2D, vector3D
class graphTransform(object):
    """
    Transformation for parametric shape grammar

    GraphTransform consists of vector transformations which are sets of 
    fromVects and toVects in the following structure:

    vects = [[fromVect1, toVect1], [fromVect2, toVect2], [fromVect3, toVect3], ...]

    When applying a graphTransform on a shape, any vector in points/lines/planes in
    the shape that is equals to a fromVect in graphTransform will be transformed into
    the coresponding toVect.

    | e.g. 
    | transformation:
    |   from     |    to
    |   (0,0)   -->   (3,3)
    |   (0,1)   -->   (5,5)
    |   (1,0)   -->   (6,0)
    |   (1,1)   -->   (8,2)
    | applying on shape: <(0,0),(0,1)>,<(0,0),(1,0)>,<(0,1),(1,1)>,<(1,0),(1,1)>
    | resulting shape:   <(3,3),(5,5)>,<(3,3),(6,0)>,<(5,5),(8,2)>,<(6,0),(8,2)>

    **Object Parameters:**

    * **vects** (list of list of two :class:`~sortal.struct.vector.vector2D`) -- the transformation list: list of vector mapping.
    * **unresolved** (list of :class:`~sortal.struct.vector.vector2D`) -- the list of rhs vectors that are not yet resolved in the detection process, first generated in analyzeRHS, subsequent matched vectors will be removed (in directive solving process).
    * **directiveResults** (list of :class:`~sortal.struct.vector.vector2D`) -- the transformation list for existing directive results.
    * **extra** (list of list of two :class:`~sortal.struct.vector.vector2D`) -- the list of intersection transformations that are not contained within the shape that are not to be included in the main transformation list.
    * **id** (str) -- the name of the disjunctive sort applicable for this transformation.
    """

    type = 'transform-graphTransform'
    
    def __init__(self):
        self.vects = []
        self.unresolved = []
        self.directiveResults = []
        self.extra = []
        self.id = None
        
    def __eq__(self, other):
        return self.vects == other.vects
    
    def __repr__(self):
        return str(self)
    
    def __str__(self):
        s = '\n* graph_transf: {\n'
        for i in self.vects:
            s+= '  '+str(i[0])+' -> '+str(i[1])+'\n'
        for i in self.unresolved:
            s+= '  '+str(i)+' -> ?\n'
        s = s[:-1] + '\n}'
        return s
        
    def __len__(self):
        return len(self.vects)
            
            
    def add(self, fromVect, toVect, override = False, index = None):
        """
        add a set of vector transformation (fromVect and its corresponding toVect)

        :param fromVect: source vector
        :type fromVect: `~sortal.struct.vector.vector2D`
        :param toVect: target vector
        :type toVect: `~sortal.struct.vector.vector2D`
        :param override: whether this is to override the current match if there is one
        :type override: boolean
        :param index: the specific index for the transformation to be inserted in
        :type index: int
        
        \* override parameter details: 

        * False: do not override the vector transformation if fromVect already exists in the current graphTransform
        * '>' : override the vector transformation only if the new toVect is bigger than the current toVect corresponding with fromVect in the current graphTransform
        * '<' : override the vector transformation only if the new toVect is smaller than the current toVect corresponding with fromVect in the current graphTransform
        * True: override the vector transformation if fromVect already exists in the current graphTransform
        """
        # print('      ', fromVect.printRhino(), toVect.printRhino(), override)
        for i in self.vects:
            if i[0] == fromVect:
                if not i[1] == toVect:
                    if override == '>' and toVect > i[1]:
                        i[1] = toVect
                        return True
                    if override == '<' and toVect < i[1]:
                        i[1] = toVect
                        return True
                    elif override is True:
                        i[1] = toVect
                        return True
                    else:
                        return False
                return True
        if index is None:
            self.vects.append([fromVect, toVect])
        else:
            self.vects.insert(index, [fromVect, toVect])
        for i in range(len(self.unresolved)):
            if self.unresolved[i] == fromVect:
                self.unresolved.pop(i)
                break
        return True
        
    def addExtra(self, fromVect, toVect, key = ''):
        """
        add a match to extra under a key name. The key is used for identifying the purpose of the extra vector
        """
        for i in self.extra:
            if i[0] == fromVect:
                i[1] = toVect
                return
        self.extra.append([fromVect, toVect, key])
        for i in range(len(self.unresolved)):
            if self.unresolved[i] == fromVect:
                self.unresolved.pop(i)
                break
    
    def getExtraMatch(self, fromVect, key = ''):
        """
        get an extra match on fromVect under a key
        """
        for i in self.extra:
            if i[0] == fromVect and i[-1] == key:
                return i[1]
        return None

    def addUnresolved(self, fromVect):
        """
        add a vector to unresolved
        """
        if fromVect not in self.unresolved:
            self.unresolved.append(fromVect)
        
    def checkUnresolved(self, fromVect):
        """
        check if a vector is in the unresolved list
        """
        return fromVect in self.unresolved
        
    def addDirectiveResult(self, result):
        """
        add a vector transformation into directiveResults list
        """
        if len(result) < 2:
            return
        self.directiveResults.append(result)
        
    def getDirectiveResults(self, coordMatch):
        """
        get a match from directiveResults
        """
        result = []
        for i in self.directiveResults:
            if i[0][0] == coordMatch[0] or i[1][0] == coordMatch[0]:
                result.append(i)
        return result
        
    def duplicate(self):
        """
        create a duplicate transformation
        """
        new = graphTransform()
        for i in self.vects:
            new.vects.append(i)
        for i in self.unresolved:
            new.unresolved.append(i)
        for i in self.extra:
            new.extra.append(i)
        for i in self.directiveResults:
            new.directiveResults.append(i)
        return new
        
    def printTransf(self):
        """
        print the transformation for debugging
        """
        for i in self.vects:
            print('  ',i[0],'->',i[1])
        for i in self.unresolved:
            print('  ',i,'-> ?')
        print()

    def printRhino(self):
        fromLine = '  _polyline '
        toLine =   '  _polyline '
        for i in self.vects:
            if '2D' in i[0].type:
                fromLine += str(i[0].x) + ',' + str(i[0].y) + ' '
                toLine += str(i[1].x) + ',' + str(i[1].y) + ' '
            else:
                fromLine += str(i[0].x) + ',' + str(i[0].y) + ',' + str(i[0].z) + ' '
                toLine += str(i[1].x) + ',' + str(i[1].y) + ',' + str(i[1].z) + ' '
        fromLine += 'EnterEnd\n  SelLast -Properties o c o 0,255,0 EnterEnd SelNone\n'
        toLine   += 'EnterEnd\n  SelLast -Properties o c o 0,0,255 EnterEnd SelNone\n'
        string = fromLine + toLine
        string += self.vects[0][0].printRhino() + '\n SelLast -Properties o c o 255,0,0 EnterEnd SelNone\n'
        string += self.vects[0][1].printRhino() + '\n SelLast -Properties o c o 255,0,0 EnterEnd SelNone\n'
        return string

            
    def apply(self, vect):
        """
        apply the transformation on a vector, return the vector itself if there is no match
        """
        for i in self.vects:
            if i[0] == vect:
                return i[1]
        return vect
        
    def transform(self, vect):
        """
        apply the transformation on a vector, return False if there is no match
        """
        for i in self.vects:
            if i[0] == vect:
                return i[1]
        return False
        
    def allTransform(self, vect):
        """
        apply the transformation on a vector, using the extra match as well as 
        the main transformation list, return False if there is no match
        """
        for i in self.vects:
            if i[0] == vect:
                return i[1]
        return self.getExtraMatch(vect)
        
    def mergeExtra(self, key = ''):
        """
        merge extra vector matches of the same key back into the main transformation list
        """
        i = 0
        while i < len(self.extra):
            if self.extra[i][-1] == key:
                self.add(self.extra[i][0], self.extra[i][1], True)
                self.extra.pop(i)
            else:
                i += 1

    def contains(self, vect, checkUnresolved = False):
        """
        check if a vector has an existing transformation in this graphTransform
        """
        for i in self.vects:
            if i[0] == vect:
                return True
        if checkUnresolved:
            for i in self.unresolved:
                if i == vect:
                    return True
        return False     
        
    def containsMatch(self, vect):
        """
        check if a vector is already a transformation for a from vector
        """
        for i in self.vects:
            if i[1] == vect:
                return True
        return False        
        
    def hasCheckPointOf(self, line):
        """
        check if a line-segment contains at least one existing match
        """
        for i in self.vects:
            if line.containsPoint(i[0]):
                return True
        return False
        
    def getCheckPoints(self, line):
        """
        get all vector matches contained within the line-segment
        """
        ints = []
        for i in self.vects:
            if line.containsPoint(i[0]):
                check = False
                for j in range(len(ints)):
                    if i[0] < ints[j][0]:
                        ints.insert(j, i)
                        check = True
                        break
                if check is False:
                    ints.append(i)
        return ints
        
    def findOrigin(self, vect):
        """
        find the fromVect that would be transformed into this vector
        """
        for i in self.vects:
            if i[1] == vect:
                return i[0]
        return False
        
    def transformable(self, ind, residuals):
        """
        check if an individual can be transformed fully by this transformation. If not, add the individual to residuals list
        """
        if 'ind' in ind.type:
            if 'ind-point' in ind.type:
                for i in self.vects:
                    if i[0] == ind.value:
                        return True
                residuals.append(ind)
                return False
            elif 'ind-lineSegment' in ind.type:
                cS = cE = False
                for i in self.vects:
                    if i[0] == ind.start:
                        cS = True
                    if i[0] == ind.end:
                        cE = True
                if cS and cE:
                    return True
                residuals.append(ind)
                return False
            elif 'ind-planeSegment' in ind.type:
                for b in ind.boundaries:
                    for p in b.polygonal:
                        check = False
                        for i in self.vects:
                            if i[0] == p:
                                check = True
                                break
                        if not check:
                            residuals.append(ind)
                            return False
            return True
        else:
            check = True
            if len(ind.elements) == 0:
                return check
            if ind.type in ('form-metaForm','form-interval','form-areal','form-coForm'):
                for i in ind.elements:
                    if not self.transformable(i, residuals):
                        check = False
            elif ind.type == 'form-discrete' and 'ind-point' in ind.elements[0].type:
                for i in ind.elements:
                    if not self.transformable(i, residuals):
                        check = False
            return check
        
class graphTransform3D(graphTransform):
    """
    graph Transform for 3D class, inherit all its parameters and functions from graphTransform
    """
    type = 'transform-graphTransform3D'
    def __init__(self):
        super(graphTransform3D, self).__init__()
        
    def duplicate(self):
        """
        create a duplicate transformation
        """
        new = graphTransform3D()
        for i in self.vects:
            new.vects.append(i)
        for i in self.extra:
            new.extra.append(i)
        for i in self.unresolved:
            new.unresolved.append(i)
        for i in self.directiveResults:
            new.directiveResults.append(i)
        return new
        
