# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import range
from builtins import super
from builtins import str


import itertools
import time

from copy import copy
from random import randint


from .mapping import mapping
from .graphNode import graphNode3D

from ..transform.graphTransform import graphTransform3D
from ..transform.affineTransform import affineTransform3D
from ..transform.paramSet import paramSet
from ..transform.transformSet import transformSet

from ..ind.line import line2D, line3D
from ..ind.description import description
from ..ind.plane import plane3D

#from ..ind.planeSegmentP import planeSegmentP3D
from ..ind.planeSegment3D import planeSegment3D

from ..io.parseReader import parseReader

from ..sort.sort import sort as sortClass

from ..struct.planeStruct import bSegment, boundary
from ..struct.vector import vector3D
from ..struct.coordinate import coordinate, c0, c1, c0p5, cn1, pi, cos, sin, sqrt, acos
from ..struct.key import key



ANY = 0
PERPENDICULAR = 1
PARALLEL = 2
SKEW = 3
EMBEDDED = 4

OUTSIDE = 0
BOUND = 1
INSIDE = 2
STRICT_BOUND = 3

NONE = '[None]'

vz = vector3D(0,0,0)

# ANY = '~'
# PARALLEL = '='
# PERPENDICULAR = '+'

# OUTSIDE = '0'
# BOUND = 'b'
# INSIDE = 'i'
# STRICT_BOUND = 's'

# NONE = '[--------]'




class graphMap3D(mapping):
    """
    graphMap3D: mapping for 3D parametric Geometry matching. 
    
    The matching process is based on "line-nodes" and "plane-nodes": codescriptor of all line-segments and 
    plane-segments in the lhs and the shape. The relationship between these line-nodes and plane-nodes are
    checked and added into 4 matrices: plane-plane, plane-line, line-plane, line-line matrices.

    For matching procedure:

    #. Match all free points (points that are not contained by any node -> unconstrained)
    #. The node matrices of the rule lhs and the shape would be compared to find all combination of matching nodes -> all point intersections are matched
    #. Check ordering of all matched intersections on their respective nodes
    #. Match all line end-points that are not node intersections as well as points that are contained by only one node (thus not intersection)
    #. Match all curves and plane-segments

    """
    # **Parameters:**
    # nodeList <graphNode3D>
    #     the 3D graph containing all the lhs line nodes and plane nodes as well as other objects in the shape for matching
    # matrices <tuple of 4 matrices (list of list)>
    #     the set of 4 lhs matrices (line-line, line-plane, plane-line and plane-plane matrix)
    # matchingNodeList <graphNode3D>
    #     the 3D graph containing all the shape line nodes and plane
    # matchingMatrices <tuple of 4 matrices (list of list)>
    #     the set of 4 shape matrices (line-line, line-plane, plane-line and plane-plane matrix)
    # rhsLines <list of line3D>
    #     list of lines in the rhs that are not present in the lhs
    # predicates <predicates in list form>
    #     a list of the rule's predicates to be performed for every detection
    # directives <directives in list form>
    #     a list of the rule's directives to be performed for every detection

    type = 'map-graphMap'


    
    def __init__(self, *args, **kwargs):
        """
        Initilization of graphMap3D, should only be used during initialization of ruleAnalysis
        """
        super(graphMap3D, self).__init__()     
        self.nodeList = graphNode3D()
        self.matrices = [[],[],[],[]]
        self.matchingNodeList = graphNode3D()
        self.matchingMatrices = [[],[],[],[]]
        self.protoTransform = graphTransform3D()
        self.colinearInts = []

        
        self.rhsLines = []
        self.rhsSolvingOrder = []
        self.lhsResolved = None
        self.rhsResolved = []
        
        self.predicates = []
        self.directives = []
        self.postCheck = False


    def isValid(self):
        """
        Check if the mapping is valid
        """
        return True

    def clear(self):
        """
        Clear the previous detection data
        """
        self.matchingNodeList.clear()
        self.matchingMatrices = [[],[],[],[]]
        
    @staticmethod
    def getInstance():
        """
        Get a new graphMap instance
        """
        return graphMap3D()
        
    def addPredicate(self, cond, keyList):
        """
        Add predicate to the predicate List
        """
        cond = cond.lower()
        self.predicates.append([cond,keyList])
        
    def clearPredicates(self):
        """
        Clear the predicate list
        """
        self.predicates = []
        
    def addDirective(self, cond, keyList):
        """
        Add directive to the directive List
        """
        cond = cond.lower()
        self.directives.append([cond,keyList])
    
    def clearDirectives(self):
        """
        Clear the directive list
        """
        self.directives = []

    def findContainingPlane(self, lhsVects, rhsVects, tfSet):
        """
        find an appropriate lhs plane (that is not in the plane node list) that contains the lhsVects and the rhsVects 
        for directive solving (distance + angle directives)
        """
        tr = self.getTransform(tfSet)
        if len(lhsVects) == 1:
            for i in range(len(tr.vects)):
                if tr.vects[i] == lhsVects[0]:
                    continue
                for j in range(i + 1, len(tr.vects)):
                    if tr.vects[j] == lhsVects[0] or lhsVects[0].colinear(tr.vects[i][0], tr.vects[j][0]):
                        continue
                    pl = plane3D(sortClass.getSort('__dummyPlane3D__'), [lhsVects[0], tr.vects[i][0], tr.vects[j][0]])
                    check = True
                    for checkVect in rhsVects:
                        if not pl.containsPoint(checkVect):
                            check = False
                            break
                    if not check:
                        continue
                    plVects = [tr.transform(lhsVects[0]), tr.vects[i][1], tr.vects[j][1]]
                    return plane3D(sortClass.getSort('__dummyPlane3D__'), plVects)
        elif len(lhsVects) == 2:
            pl = plane3D(sortClass.getSort('__dummyPlane3D__'), [rhsVects[0], lhsVects[0], lhsVects[1]])
            tr_lhsVect0 = tr.transform(lhsVects[0])
            tr_lhsVect1 = tr.transform(lhsVects[1])
            for vect in tr.vects:
                if (vect[0] == lhsVects[0] or vect[0] == lhsVects[1]
                or vect[1] == tr_lhsVect0 or vect[0] == tr_lhsVect1
                or vect[0].colinear(lhsVects[0], lhsVects[1])
                or vect[1].colinear(tr_lhsVect0, tr_lhsVect1)
                or not pl.containsPoint(vect[0])):
                    continue
                # check = True
                # for checkVect in rhsVects:
                #     if not pl.containsPoint(checkVect):
                #         check = False
                #         break
                # if not check:
                #     continue
                plVects = [vect[1], tr_lhsVect0, tr_lhsVect1]
                return plane3D(sortClass.getSort('__dummyPlane3D__'), plVects)
            # for vect in tr.vects:
            #     if (vect[0] == lhsVects[0] or vect[0] == lhsVects[1]
            #     or vect[0].colinear(lhsVects[0], lhsVects[1])):
            #         continue
            #     pl = plane3D(sortClass.getSort('__dummyPlane3D__'), [vect[0], lhsVects[0], lhsVects[1]])
            #     check = True
            #     for checkVect in rhsVects:
            #         if not pl.containsPoint(checkVect):
            #             check = False
            #             break
            #     if not check:
            #         continue
            #     plVects = [vect[1], tr.transform(lhsVects[0]), tr.transform(lhsVects[1])]
            #     return plane3D(sortClass.getSort('__dummyPlane3D__'), plVects)
        return None

    # ============================================================================================
    #                                      ANALYZING PROCESS
    # ============================================================================================

    def generateNodes(self, ind, matching = False):
        """
        add individual object into the graph
        """
        if matching:
            nodeList = self.matchingNodeList
        else:
            nodeList = self.nodeList
        if ind.type == 'ind-pointP3D':
            nodeList.LHS_addPoint(ind)
            if not matching:
                self.protoTransform.addUnresolved(ind.value)
        elif ind.type == 'ind-lineSegmentP3D':
            element = ind.parametricNode()
            nodeLen = len(self.nodeList.lines)
            nodeList.LHS_addLine(element, ind)
            if not matching:
                self.protoTransform.addUnresolved(ind.start)
                self.protoTransform.addUnresolved(ind.end)
                if nodeLen < len(self.nodeList.lines):
                    newNode = self.nodeList.lines[-1]
                    for ln in range(len(self.nodeList.lines) - 1):
                        if not self.nodeList.lines[ln].nodeElement.intersect(newNode.nodeElement):
                            continue
                        intersection = self.nodeList.lines[ln].nodeElement.intersection(newNode.nodeElement)
                        notMin = False
                        notMax = False
                        for lnI in self.nodeList.lines[ln].forms:
                            if intersection > lnI.start:
                                notMin = True
                            if intersection < lnI.end:
                                notMax = True
                            if notMin and notMax:
                                break
                        if notMin and notMax:
                            self.protoTransform.addUnresolved(intersection)
                            continue
                        notMin = False
                        notMax = False
                        for lnJ in newNode.forms:
                            if intersection > lnJ.start:
                                notMin = True
                            if intersection < lnJ.end:
                                notMax = True
                            if notMin and notMax:
                                break
                        if notMin and notMax:
                            self.protoTransform.addUnresolved(intersection)

        elif ind.type == 'ind-planeSegmentP3D':
            element = ind.parametricNode()
            nodeList.LHS_addPlane(element, ind)
            if not matching:
                for b in ind.boundaries:
                    for v in b.polygonal: 
                        self.protoTransform.addUnresolved(v)
        elif ind.type == 'ind-circleP3D' or ind.type == 'ind-ellipseP3D' or ind.type == 'ind-bezierCurveP3D':
            nodeList.LHS_addCurve(ind)
        return nodeList
        
    def updateMatrices(self, matching = True):
        """
        update the three matrices
        """
        self.updatePlaneMatrix(matching)
        self.updateLineMatrix(matching)
        self.updatePlaneLineMatrix(matching)
        
    def updatePlaneMatrix(self, matching):
        """
        update the plane-plane matrix
        """
        if matching is False:
            nodeList = self.nodeList
            matrices = self.matrices
        else:
            nodeList = self.matchingNodeList
            matrices = self.matchingMatrices

        matrices[0] = [[None for i in range(len(nodeList.planes))] for j in range(len(nodeList.planes))]

        # for each of the plane
        for i in range(len(nodeList.planes)):
            pli = nodeList.planes[i].nodeElement
            for j in range(i+1, len(nodeList.planes)):
                plj = nodeList.planes[j].nodeElement

                # analyze edge type
                edge = ANY
                if plj.parallel(pli):
                    edge = PARALLEL
                elif plj.perpendicular(pli):
                    edge = PERPENDICULAR
                    
                positionOne = OUTSIDE
                positionTwo = OUTSIDE

                # not parallel -> analyze intersection type 
                if edge < PARALLEL:
                    iDir, iRoot = plj.intersection(pli)

                    # intersection vs plane at index i
                    for k in nodeList.planes[i].forms:
                        check = False
                        for line in k.boundaries[0].canonical:
                            if iDir.parallel(line.direction) and iRoot == line.root():
                                if positionOne == BOUND:
                                    positionOne = INSIDE
                                    check = True
                                    break
                                else:
                                    positionOne = BOUND
                                    continue
                            elif iDir.parallel(line.direction):
                                continue
                            else:
                                intsn = line.intersectionLine(iRoot, iDir)
                                if line.containsPoint(intsn):
                                    positionOne = INSIDE
                                    check = True
                                    break
                        if check:
                            break

                    # intersection vs plane at index j
                    for k in nodeList.planes[j].forms:
                        check = False
                        for line in k.boundaries[0].canonical:
                            if iDir.parallel(line.direction) and iRoot == line.root():
                                if positionTwo == BOUND:
                                    positionTwo = INSIDE
                                    check = True
                                    break
                                else:
                                    positionTwo = BOUND
                                    continue
                            elif iDir.parallel(line.direction):
                                continue
                            else:
                                intsn = line.intersectionLine(iRoot, iDir)
                                if line.containsPoint(intsn):
                                    positionTwo = INSIDE
                                    check = True
                                    break
                        if check:
                            break
                # set respective plane-plane matrix value
                matrices[0][i][j] = [edge, positionOne]
                matrices[0][j][i] = [edge, positionTwo]

                
    def updateLineMatrix(self, matching):
        """
        update the line-line matrix
        """
        if matching is False:
            nodeList = self.nodeList
            matrices = self.matrices
        else:
            nodeList = self.matchingNodeList
            matrices = self.matchingMatrices
        matrices[3] = [[None for i in range(len(nodeList.lines))] for j in range(len(nodeList.lines))]
        for i in range(len(nodeList.lines)):
            li = nodeList.lines[i].nodeElement
            for j in range(i+1, len(nodeList.lines)):
                lj = nodeList.lines[j].nodeElement

                # analyze edge type
                edge = ANY
                if lj.skew(li):
                    edge = SKEW
                elif lj.parallel(li):
                    edge = PARALLEL
                elif lj.perpendicular(li):
                    edge = PERPENDICULAR
                    
                positionOne = OUTSIDE
                positionTwo = OUTSIDE

                # neither parallel nor skew -> analyze intersection
                if edge < PARALLEL:
                    intersection = lj.intersection(li)
                    newIntersection = graphNode3D.lineIntersection(intersection)
                    # if not matching:
                    #     if (len(nodeList.lines[i].forms) > 0 
                    #     and nodeList.lines[i].forms[0].start < intersection < nodeList.lines[i].forms[-1].end):
                    #         self.protoTransform.addUnresolved(intersection)
                    #     if (len(nodeList.lines[j].forms) > 0 
                    #     and nodeList.lines[j].forms[0].start < intersection < nodeList.lines[j].forms[-1].end):
                    #         self.protoTransform.addUnresolved(intersection)
                    nodeList.lines[i].addIntersection(newIntersection)
                    nodeList.lines[j].addIntersection(newIntersection)

                    # intersection vs line at index i
                    for k in nodeList.lines[i].forms:
                        if k.start == intersection or k.end == intersection:
                            if positionOne == BOUND:
                                positionOne = INSIDE
                                newIntersection.iType = 'i'
                                break
                            else:
                                positionOne = BOUND
                                newIntersection.iType = 'i'
                        elif k.containsPoint(intersection):
                            positionOne = INSIDE
                            newIntersection.iType = 'i'
                            break

                    # intersection vs line at index j
                    for k in nodeList.lines[j].forms:
                        if k.start == intersection or k.end == intersection:
                            if positionTwo == BOUND:
                                positionTwo = INSIDE
                                newIntersection.iType = 'i'
                                break
                            else:
                                positionTwo = BOUND
                                newIntersection.iType = 'i'
                        elif k.containsPoint(intersection):
                            positionTwo = INSIDE
                            newIntersection.iType = 'i'
                            break
                # set respective line-line matrix value
                matrices[3][i][j] = [edge, positionOne]
                matrices[3][j][i] = [edge, positionTwo]


    def analyzeColinear(self):
        if len(self.nodeList.lines) < 3: return

        for i in range(len(self.nodeList.lines) - 2):
            li = self.nodeList.lines[i]
            for j in range(i + 1, len(self.nodeList.lines) - 1):
                if self.matrices[3][i][j][0] != PARALLEL:
                    continue
                lj = self.nodeList.lines[j]
                for k in range(j + 1, len(self.nodeList.lines)):
                    if self.matrices[3][i][k][0] != PARALLEL:
                        continue
                    lk = self.nodeList.lines[k]
                    for pti in li.intersections:
                        for ptj in lj.intersections:
                            for ptk in lk.intersections:
                                if not (pti.endP or ptj.endP or ptk.endP): continue
                                if pti.value.colinear(ptj.value, ptk.value):
                                    self.colinearInts.append([pti, ptj, ptk, li, lj, lk])


    def updatePlaneLineMatrix(self, matching):
        """
        update the plane-line and line-plane matrices
        """
        if matching is False:
            nodeList = self.nodeList
            matrices = self.matrices
        else:
            nodeList = self.matchingNodeList
            matrices = self.matchingMatrices
        matrices[1] = [[None for i in range(len(nodeList.lines))] for j in range(len(nodeList.planes))]
        matrices[2] = [[None for i in range(len(nodeList.planes))] for j in range(len(nodeList.lines))]

        for i in range(len(nodeList.planes)):
            pli = nodeList.planes[i].nodeElement
            for j in range(len(nodeList.lines)):
                lj = nodeList.lines[j].nodeElement

                # analyze edge
                edge = ANY
                if pli.parallelLine(lj):
                    edge = PARALLEL
                    if lj.root == pli.pRoot or lj.root.subtract(pli.pRoot).dotproduct(pli.normal) == 0:
                        edge = EMBEDDED
                        nodeList.planes[i].addLine(nodeList.lines[j])
                elif pli.perpendicularLine(lj):
                    edge = PERPENDICULAR
                
                positionOne = OUTSIDE
                positionTwo = OUTSIDE
                
                # if any or perpendicular -> analyze intersection
                if edge < PARALLEL:
                    intersection = pli.intersectionLine(lj)
                    newIntersection = graphNode3D.lineIntersection(intersection)
                    nodeList.planes[i].addIntersection(newIntersection)
                    for k in nodeList.planes[i].forms:
                        if k.containsPoint(intersection):
                            positionOne = INSIDE
                            break
                    for k in nodeList.lines[j].forms:
                        if k.start == intersection or k.end == intersection:
                            if positionTwo == BOUND:
                                positionTwo = INSIDE
                                newIntersection.iType = 'i'
                                break
                            else:
                                positionTwo = BOUND
                                newIntersection.iType = 'i'
                        elif k.containsPoint(intersection):
                            positionTwo = INSIDE
                            newIntersection.iType = 'i'
                            break
                # add to respective line-plane and plane-line matrix
                matrices[1][i][j] = [edge, positionOne]
                matrices[2][j][i] = [edge, positionTwo]
        
    def map(self, fromInd, toInd):
        """
        Mapping a shape individual to a lhs individual for detection
        """
        self.generateNodes(toInd, matching = True)
        return True
        
    def distinguish(self, ind):
        """
        Distinguishing a lhs individual for rule analysis
        """
        node = self.generateNodes(ind)
        return node

    def postAnalysis(self):
        """
        Extra analysis processes:

        * generateLinesPlanes: generate plane-nodes based on existing lines
        * updateMatrices: update the 4 matrices
        * classifyObjs: classify the objects that are not line-segments for later matching
        * addEndPoint: go through the lists and add all the end points of lines that are not intersections into nodeList
        * classifyIntersections: classifying the relationship between all the intersections/end points on a line as connected/disconnected
        * organizeForm: order the line-segments of a node and combine+separate them according to their attributes
        """
        # generate planes based on the existing lines
        self.nodeList.generateLinesPlanes(True)
        self.updateMatrices(matching=False)
        self.nodeList.classifyObjs(True)

        for node in self.nodeList.lines:
            if len(node.forms) == 0:
                continue
            node.addEndPoints()
            #node.classifyIntersections()
            node.organizeForm()
        self.analyzeColinear()

    def analyzeRHS(self, rhs):
        """
        analyze and get all information for solving all new objects in the rhs

        :param rhs: rhs form
        :type rhs: ~sortal.form.form.form
        """
        # print(self.protoTransform.unresolved)

        # add all rhs object into the unresolved list
        if self.lhsResolved is None:
            self.lhsResolved = copy(self.protoTransform.unresolved)
        else:
            self.protoTransform.unresolved = copy(self.lhsResolved)

        self.rhsResolved = []
        self.nodeList.rhsLines = []
        self.nodeList.rhsPlanes = []
        self.nodeList.rhsPoints = []
        for node in self.nodeList.lines:
            node.rhsInfo = None
        for node in self.nodeList.planes:
            node.rhsInfo = None
        self.rhsSolvingOrder = []
        
        # go through the rhs and add rhs line node/plane node
        self.addRHSObjects(rhs)
        # for j in self.nodeList.rhsLines:
        #     print('root:',j.nodeElement.root,' direction:', j.nodeElement.direction)
        #     for k in j.forms:
        #         print(k)


        # create new rhs planes from lines
        for i in range(len(self.nodeList.rhsLines)):
            rhsLine1 = self.nodeList.rhsLines[i]
            # new planes between a RHS line and a LHS line
            for lhsLine in self.nodeList.lines:
                if not (lhsLine.nodeElement.equals(rhsLine1.nodeElement) or lhsLine.nodeElement.skew(rhsLine1.nodeElement)):
                    vecList = [lhsLine.nodeElement.root, lhsLine.nodeElement.root.add(lhsLine.nodeElement.direction)]
                    if rhsLine1.nodeElement.root.colinear(vecList[0], vecList[1]):
                        vecList.append(rhsLine1.nodeElement.root.add(rhsLine1.nodeElement.direction))
                    else:
                        vecList.append(rhsLine1.nodeElement.root)
                    extraPlane = plane3D(sortClass.getSort('__dummyPlane3D__'), vecList)
                    planeNode = self.nodeList.RHS_addPlane(extraPlane)
                    if (planeNode is not None 
                    and planeNode.rhsInfo['case'] == graphNode3D.planeNode.plane_casus.CASE_PROXY
                    and planeNode not in self.rhsSolvingOrder):
                        self.rhsSolvingOrder.append(planeNode)
            # new planes between two RHS lines
            for j in range(i+1, len(self.nodeList.rhsLines)):
                rhsLine2 = self.nodeList.rhsLines[j]
                if not (rhsLine2.nodeElement.equals(rhsLine1.nodeElement) or rhsLine2.nodeElement.skew(rhsLine1.nodeElement)):
                    vecList = [rhsLine2.nodeElement.root, rhsLine2.nodeElement.root.add(rhsLine2.nodeElement.direction)]
                    if rhsLine1.nodeElement.root.colinear(vecList[0], vecList[1]):
                        vecList.append(rhsLine1.nodeElement.root.add(rhsLine1.nodeElement.direction))
                    else:
                        vecList.append(rhsLine1.nodeElement.root)
                    extraPlane = plane3D(sortClass.getSort('__dummyPlane3D__'), vecList)
                    planeNode = self.nodeList.RHS_addPlane(extraPlane)
                    if (planeNode is not None 
                    and planeNode.rhsInfo['case'] == graphNode3D.planeNode.plane_casus.CASE_PROXY 
                    and planeNode not in self.rhsSolvingOrder):
                        self.rhsSolvingOrder.append(planeNode)
        # analyze all rhs lines relationship (parallel & perpendicular)
        for node1 in self.nodeList.rhsLines:
            for node2 in (self.nodeList.lines + self.nodeList.rhsLines):
                # ignore if same line or skew
                if node1.nodeElement.equals(node2.nodeElement) or node1.nodeElement.skew(node2.nodeElement):
                    continue
                if node1.nodeElement.parallel(node2.nodeElement):
                    node1.rhsInfo["parallel"].append(node2)
                elif node1.nodeElement.perpendicular(node2.nodeElement):
                    node1.rhsInfo["perpendicular"].append(node2)
                else:
                    node1.rhsInfo["intersect"].append(node2)
                if node1.nodeElement.intersect(node2.nodeElement):
                    intn = node1.nodeElement.intersection(node2.nodeElement)
                    if intn not in self.protoTransform.unresolved:
                        node1.rhsInfo['extraInts'].add(intn)
        # for j in self.nodeList.rhsPlanes:
        #     print(j.nodeElement.printRhino())
        #     # print('root:',j.nodeElement.pRoot,' normal:', j.nodeElement.normal)
        #     # for k in j.forms:
        #     #     print(k)
        for plNode in self.nodeList.rhsPlanes:
            for lnNode in self.nodeList.rhsLines:
                if plNode.nodeElement.containsLine(lnNode.nodeElement):
                    plNode.rhsInfo['containedLines'].append(lnNode)
        
        if len(self.nodeList.planes) == 0 and len(self.nodeList.rhsPlanes) > 0:
            maxVecs = []
            maxPlane = None
            for planeNode in self.nodeList.rhsPlanes:
                containedVecs = []
                for vec in self.protoTransform.unresolved:
                    if planeNode.nodeElement.containsPoint(vec):
                        containedVecs.append(vec)
                        if len(containedVecs) == 3 and containedVecs[0].colinear(containedVecs[1], containedVecs[2]):
                            containedVecs.pop()
                if len(containedVecs) > len(maxVecs):
                    maxPlane = planeNode
                    maxVecs = containedVecs
                if len(containedVecs) == 3:
                    break
            if 0 < len(maxVecs) < 3:
                self.assignCase(maxPlane, graphNode3D.planeNode.plane_casus.CASE_DEFAULT, [maxVecs], self.protoTransform.unresolved)

        transformableVects = self.protoTransform.unresolved

        # find the order of resolving the lines/planes - DIRECTIVE CASES
        for directive in self.directives:
            for dirStatement in directive[1]:
                # for point_on_line specifically, check if the individual is in any of the lhs node
                if directive[0] == 'point_on_line' :
                    if len(dirStatement) == 2:
                        arg = dirStatement[0]
                        if arg.type != 'struct-key':
                            raise ValueError('Error: lineSegment tag expected for the first argument of point_on_line directive.')
                        check = False
                        for node in self.nodeList.lines:
                            if node.nodeElement.equals(arg.ind) and arg.ind in node.forms:
                                if node.rhsInfo is None:
                                    node.rhsInfo = {
                                        'case': graphNode3D.lineNode.line_casus.CASE_LHS,
                                        'directives': []
                                    }
                                    self.rhsSolvingOrder.insert(0, node)
                                node.rhsInfo['directives'].append(['point_on_line', arg.ind, dirStatement])
                                for vect in self.rhsResolved:
                                    if arg.ind.containsPoint(vect) and vect not in transformableVects:
                                        transformableVects.append(vect)
                                check = True
                                break
                        if check: continue
                        self.checkRHSDirectiveObject(arg, transformableVects, self.nodeList.rhsLines, self.nodeList.rhsPlanes)
                    elif len(dirStatement) == 3:
                        targetObj = dirStatement[0]
                        if targetObj.type != 'struct-key' and not 'ind-lineSegment' in targetObj.ind.type and not 'ind-point' in targetObj.ind.type:
                            raise ValueError('Error: lineSegment or point tag expected for the first argument of point_on_line directive.')
                        arg = dirStatement[1]
                        if arg.type != 'struct-key':
                            raise ValueError('Error: lineSegment tag expected for the second argument of 3 arguments point_on_line directive.')
                        if 'ind-lineSegment' in targetObj.ind.type:
                            if line3D.containsPoint(arg.ind, targetObj.ind.start):
                                pointOnLinetarget = targetObj.ind.start
                            elif line3D.containsPoint(arg.ind, targetObj.ind.end):
                                pointOnLinetarget = targetObj.ind.end
                        elif 'ind-point' in targetObj.ind.type:
                                pointOnLinetarget = targetObj.ind.value
                        check = False
                        for node in self.nodeList.lines:
                            if node.nodeElement.equals(arg.ind) and arg.ind in node.forms:
                                if node.rhsInfo is None:
                                    node.rhsInfo = {
                                        'case': graphNode3D.lineNode.line_casus.CASE_LHS,
                                        'directives': []
                                    }
                                    self.rhsSolvingOrder.insert(0, node)
                                node.rhsInfo['directives'].append(['point_on_line', arg.ind, dirStatement, pointOnLinetarget])
                                for vect in self.rhsResolved:
                                    if arg.ind.containsPoint(vect) and vect not in transformableVects:
                                        transformableVects.append(vect)
                                check = True
                                break
                        if check: continue
                        self.checkRHSDirectiveObject(arg, transformableVects, self.nodeList.rhsLines, self.nodeList.rhsPlanes)
                    else:
                        raise ValueError('Error: 2 or 3 arguments expected for point_on_line directive.')
                    # for arg in dirStatement:
                    #     if arg.type == 'struct-key':
                    #         check = False
                    #         for node in self.nodeList.lines:
                    #             if node.nodeElement.equals(arg.ind) and arg.ind in node.forms:
                    #                 if node.rhsInfo is None:
                    #                     node.rhsInfo = {
                    #                         'case': graphNode3D.lineNode.line_casus.CASE_LHS,
                    #                         'directives': []
                    #                     }
                    #                     self.rhsSolvingOrder.insert(0, node)
                    #                 node.rhsInfo['directives'].append(['point_on_line', arg.ind, dirStatement])
                    #                 for vect in self.rhsResolved:
                    #                     if arg.ind.containsPoint(vect) and vect not in transformableVects:
                    #                         transformableVects.append(vect)
                    #                 check = True
                    #                 break
                    #         if check: continue
                    #         self.checkRHSDirectiveObject(arg, transformableVects, self.nodeList.rhsLines, self.nodeList.rhsPlanes)
                # any other type of directive, check RHS directive case
                else:
                    self.checkRHSDirectiveCase(dirStatement, directive[0], transformableVects, self.nodeList.rhsLines,
                                            self.nodeList.rhsPlanes)
                # self.resetTempCase()
        for rhsLineNode in self.nodeList.rhsLines:
            if self.checkRHSLineCase(rhsLineNode, transformableVects, self.nodeList.rhsLines, self.nodeList.rhsPlanes):
                # self.finalizeCase(rhsLineNode)
                # self.resetTempCase()
                pass
        for rhsPlaneNode in self.nodeList.rhsPlanes:
            if rhsPlaneNode.rhsInfo['required'] == False: continue
            if self.checkRHSPlaneCase(rhsPlaneNode, transformableVects, self.nodeList.rhsLines, self.nodeList.rhsPlanes):
                # self.finalizeCase(rhsPlaneNode)
                # self.resetTempCase()
                pass
        for vect in self.rhsResolved:
            if vect not in self.protoTransform.unresolved:
                self.protoTransform.unresolved.append(vect)
        # print('\n\n\nResolving Order:')
        # for i in self.rhsSolvingOrder:
        #     print(i.nodeElement.printRhino())
        #     print(i.rhsInfo['case'],'\n')
        #     print(i.rhsInfo['directives'],'\n')
        return


    def checkRHSDirectiveCase(self, dirStatement, directiveName, transformableVects, availableLineNodes,
                              availablePlaneNodes):
        """
        check the case for a specific directive statement

        :param dirStatement: the directive to be analyzed
        :param directiveName: the directive name e.g. length, point_on_line, direction, angle, distance, ...
        :type directiveName: string
        :param transformableVects: the list of transformable vector by this point
        :type transformableVects: list of vector3D
        :param availableLineNodes: the list of avalable rhs line nodes that can be used for determining the directive case
        :type availableLineNodes: list of node
        :param availablePlaneNodes: the list of avalable rhs plane nodes that can be used for determining the directive case
        :type availablePlaneNodes: list of node
        """
        # length directive
        if directiveName == 'length' or directiveName == 'point_on_line':
            for arg in dirStatement:
                self.checkRHSDirectiveObject(arg, transformableVects, availableLineNodes, availablePlaneNodes)
        # CASE_DIRECTION: direction directive, 1 contained point
        elif directiveName == 'direction':
            if dirStatement[0].type != 'struct-key':
                raise ValueError('Error: the first arguments of direction directive must be an object reference key.')
            if dirStatement[0].ind.type != 'ind-lineSegmentP3D':
                return False
            objNode = None
            for i in range(len(availableLineNodes)):
                if availableLineNodes[i].nodeElement.equals(dirStatement[0].ind):
                    objNode = availableLineNodes[i]
                    break
            for vec in transformableVects:
                if objNode.nodeElement.containsPoint(vec):
                    self.assignCase(objNode, graphNode3D.lineNode.line_casus.CASE_DIRECTION, 
                                    [dirStatement, vec], transformableVects)
                    return True
            return False
        else:
            if dirStatement[0].type != 'struct-key' or dirStatement[1].type != 'struct-key':
                raise ValueError('Error: the first and second arguments of ' + directiveName + ' directive must be an object reference key.')
            lineNodes = copy(availableLineNodes)
            planeNodes = copy(availablePlaneNodes)
            objNode = None

            if dirStatement[0].ind.type == 'ind-lineSegmentP3D':
                for i in range(len(lineNodes)):
                    if lineNodes[i].nodeElement.equals(dirStatement[0].ind):
                        objNode = lineNodes.pop(i)
                        break
            elif dirStatement[0].ind.type == 'ind-planeSegmentP3D':
                for i in range(len(planeNodes)):
                    if planeNodes[i].nodeElement.equals(dirStatement[0].ind):
                        objNode = planeNodes.pop(i)
                        break
            elif dirStatement[0].ind.type == 'ind-pointP3D':
                objNode = graphNode3D.pointNode(dirStatement[0].ind)
            if objNode is None:
                return False
            check = self.checkRHSDirectiveObject(dirStatement[1], transformableVects, lineNodes, planeNodes)
            if not check:
                return False
            if dirStatement[0].ind.type == 'ind-lineSegmentP3D':
                # distance directive cases, requires a solvable plane
                if directiveName == 'distance':
                    if dirStatement[1].ind.type == 'ind-lineSegmentP3D' and len(dirStatement) == 4:
                        self.assignCase(objNode, graphNode3D.lineNode.line_casus.CASE_LINE_LINE_DISTANCE_DIRECTION, 
                                        [dirStatement], transformableVects)
                        return True
                    rejectedPlanes = []
                    # find a solvable plane
                    # Distance directive cases where the plane contains the from-object
                    for planeNode in (self.nodeList.planes + self.nodeList.rhsPlanes):
                        if (not planeNode.nodeElement.containsLine(objNode.nodeElement)
                        or not self.checkRHSPlaneCase(planeNode, transformableVects, availableLineNodes, availablePlaneNodes)):
                            continue
                        # CASE_POINT_LINE_DISTANCE: point-line directive, 1 plane, 1 contained point
                        if dirStatement[1].ind.type == 'ind-pointP3D':
                            if dirStatement[1].ind.type == 'ind-pointP3D' and not planeNode.nodeElement.containsPoint(dirStatement[1].ind.value):
                                rejectedPlanes.append(planeNode)
                                continue
                            for vec in transformableVects:
                                if objNode.nodeElement.containsPoint(vec):
                                    anchorVect = None
                                    dirCheck = None
                                    projFromPt = dirStatement[0].ind.projection(dirStatement[1].ind.value)
                                    if projFromPt == vec:
                                        self.assignCase(objNode, graphNode3D.lineNode.line_casus.CASE_POINT_LINE_DISTANCE, 
                                                        [dirStatement, planeNode, vec, None, None], transformableVects)
                                        return True
                                    for anchor in transformableVects:
                                        if anchor is vec:
                                            continue
                                        # find the projection of vec onto the plane
                                        projAnchor = planeNode.nodeElement.projection(anchor)
                                        # if the base line contains the projection, reject this vec
                                        if vec.colinear(projAnchor, dirStatement[1].ind.value):
                                            continue
                                        anchorVect = anchor

                                        # check the orientation: comparing clockwise property between:
                                        # _ anchor_projection -> contained point -> from point
                                        # vs
                                        # _ projection of from point on new line -> contained point -> from point
                                        orientationAnchor = graphMap3D.cw(planeNode.nodeElement, dirStatement[1].ind.value, vec, projAnchor)
                                        orientationNew = graphMap3D.cw(planeNode.nodeElement, dirStatement[1].ind.value, vec, projFromPt)
                                        dirCheck = orientationAnchor == orientationNew
                                        break
                                    self.assignCase(objNode, graphNode3D.lineNode.line_casus.CASE_POINT_LINE_DISTANCE, 
                                                    [dirStatement, planeNode, vec, anchorVect, dirCheck], transformableVects)
                                    return True
                        # CASE_LINE_LINE_DISTANCE: line-line directive, 1 plane, 1 extra vector for checking the new line's orientation
                        if dirStatement[1].ind.type == 'ind-lineSegmentP3D':
                            if dirStatement[1].ind.type == 'ind-lineSegmentP3D' and not planeNode.nodeElement.containsLine(dirStatement[1].ind):
                                rejectedPlanes.append(planeNode)
                                continue

                            extraVect = None
                            dirCheck = None
                            for vec in transformableVects:
                                # find the projection of vec onto the plane
                                projAnchor0 = planeNode.nodeElement.projection(vec)
                                # if the base line contains the projection, reject this vec
                                if line3D.containsPoint(dirStatement[1].ind, projAnchor0):
                                    continue

                                extraVect = vec

                                # find the projection of the vec onto the base line
                                projAnchor1 = dirStatement[1].ind.projection(extraVect)

                                # check the orientation 
                                orientationAnchor = projAnchor0.subtract(projAnchor1).unit()
                                orientationNew = dirStatement[0].ind.root.subtract(dirStatement[1].ind.root).unit()
                                dirCheck = orientationAnchor == orientationNew
                                break
                            # there have to be at least one possible extraVect since the planeNode is solvable
                            self.assignCase(objNode, graphNode3D.lineNode.line_casus.CASE_LINE_LINE_DISTANCE, 
                                            [dirStatement, planeNode, extraVect, dirCheck], transformableVects)
                            return True
                    # Distance directive cases where the plane doesnt contain the from-object
                    for planeNode in rejectedPlanes:
                        # CASE_POINT_LINE_DISTANCE: point-line directive, 1 plane, 1 contained point
                        if dirStatement[1].ind.type == 'ind-pointP3D':
                            for vec in transformableVects:
                                if objNode.nodeElement.containsPoint(vec):
                                    anchorVect = None
                                    dirCheck = None
                                    projFromPt = dirStatement[0].ind.projection(dirStatement[1].ind.value)
                                    if projFromPt == vec:
                                        self.assignCase(objNode, graphNode3D.lineNode.line_casus.CASE_POINT_LINE_DISTANCE, 
                                                        [dirStatement, planeNode, vec, None, None], transformableVects)
                                        return True
                                    for anchor in transformableVects:
                                        if anchor is vec:
                                            continue
                                        # find the projection of vec onto the plane
                                        projAnchor = planeNode.nodeElement.projection(anchor)
                                        # if the base line contains the projection, reject this vec
                                        if vec.colinear(projAnchor, dirStatement[1].ind.value):
                                            continue
                                        anchorVect = anchor

                                        # check the orientation: comparing clockwise property between:
                                        # _ anchor_projection -> contained point -> from point
                                        # vs
                                        # _ projection of from point on new line -> contained point -> from point
                                        orientationAnchor = graphMap3D.cw(planeNode.nodeElement, dirStatement[1].ind.value, vec, projAnchor)
                                        orientationNew = graphMap3D.cw(planeNode.nodeElement, dirStatement[1].ind.value, vec, projFromPt)
                                        dirCheck = orientationAnchor == orientationNew
                                        break
                                    self.assignCase(objNode, graphNode3D.lineNode.line_casus.CASE_POINT_LINE_DISTANCE, 
                                                    [dirStatement, planeNode, vec, anchorVect, dirCheck], transformableVects)
                                    return True
                        # CASE_LINE_LINE_DISTANCE: line-line directive, 1 plane, 1 extra vector for checking the new line's orientation
                        if dirStatement[1].ind.type == 'ind-lineSegmentP3D':
                            extraVect = None
                            dirCheck = None
                            baseLine = line3D(sortClass.getSort('__dummyLine3D__'),
                                              planeNode.nodeElement.projection(dirStatement[1].ind.start), 
                                              planeNode.nodeElement.projection(dirStatement[1].ind.end))
                            for vec in transformableVects:
                                # find the projection of vec onto the plane
                                projAnchor0 = planeNode.nodeElement.projection(vec)
                                # if the base line contains the projection, reject this vec
                                if line3D.containsPoint(baseLine, projAnchor0):
                                    continue

                                extraVect = vec

                                # find the projection of the vec onto the base line
                                projAnchor1 = baseLine.projection(extraVect)

                                # check the orientation 
                                orientationAnchor = projAnchor0.subtract(projAnchor1).unit()
                                orientationNew = dirStatement[0].ind.root.subtract(baseLine.root).unit()
                                dirCheck = orientationAnchor == orientationNew
                                break
                            # there have to be at least one possible extraVect since the planeNode is solvable
                            self.assignCase(objNode, graphNode3D.lineNode.line_casus.CASE_LINE_LINE_DISTANCE, 
                                            [dirStatement, planeNode, extraVect, dirCheck], transformableVects)
                            return True
                    return False
                # CASE_ANGLE: angle directive, 1 plane, 1 contained point
                elif directiveName == 'angle':
                    # find a solvable plane
                    for planeNode in (self.nodeList.planes + self.nodeList.rhsPlanes):
                        if (not planeNode.nodeElement.containsLine(objNode.nodeElement)
                        or not self.checkRHSPlaneCase(planeNode, transformableVects, availableLineNodes, availablePlaneNodes)):
                            continue
                        for vec in transformableVects:
                            if objNode.nodeElement.containsPoint(vec):
                                # anchor type 1: contained point is on base (intersection between base and new)
                                #         \ |
                                #          \| 
                                #           . (basePt2 - intersection/contained point)
                                #           |\ <-- angle
                                #           | \
                                #           |  \
                                #           |   \
                                #           |    \
                                #           |     \
                                #           |      \
                                #           |       \
                                #           |        \
                                #           |         \
                                #           |          \
                                #           |           \
                                #           |_           \
                                #  (projPt1)|_|_ _ _ _ _ _. (basePt1)
                                #           |              \
                                #           |               \
                                #        <<new>>         <<base>>

                                # anchor type 2: contained point is not on base ####### NEED THOROUGH TESTING #######
                                #         \ |
                                #          \| 
                                #           . 
                                #           |\ <-- angle
                                #           | \
                                #           |  \
                                #           |   \
                                #           |    .(basePt2 - nearer to intersection)
                                #           |     \
                                #           |      \
                                #           |       \
                                #   (contPt).        \
                                #           |         \
                                #           |          \
                                #           |           \
                                #           |_           \
                                #  (projPt1)|_|_ _ _ _ _ _.(basePt1 - further from intersection)
                                #           |              \
                                #           |               \
                                #        <<new>>         <<base>>
                                # cw check: basePt1 -> basePt2 -> projPt1
                                #        vs basePt1 -> basePt2 -> anchor (projection of anchor on plane)
                                basePt1, basePt2 = None, None
                                if line3D.containsPoint(dirStatement[1].ind, vec):
                                    basePt1 = dirStatement[1].ind.start
                                    basePt2 = vec
                                    if basePt1 == basePt2:
                                        basePt1 = dirStatement[1].ind.end
                                else:
                                    if not line3D.intersect(dirStatement[0].ind, dirStatement[1].ind):
                                        return False
                                    intersection = line3D.intersection(dirStatement[0].ind, dirStatement[1].ind)
                                    basePt1 = dirStatement[1].ind.start
                                    basePt2 = dirStatement[1].ind.end

                                    d1 = intersection.subtract(basePt1)
                                    d2 = intersection.subtract(basePt2)
                                    if d1 > vz == d2 > vz:
                                        if d1.squareLength() < d2.squareLength():
                                            basePt1, basePt2 = basePt2, basePt1
                                projPt1 = dirStatement[0].ind.projection(basePt1)
                                orientationNew = graphMap3D.cw(planeNode.nodeElement, basePt1, basePt2, projPt1)

                                anchorVect = None
                                dirCheck = None
                                for anchor in transformableVects:
                                    if anchor is vec:
                                        continue
                                    # find the projection of vec onto the plane
                                    projAnchor = planeNode.nodeElement.projection(anchor)
                                    # if the base line contains the projection, reject this vec
                                    if projAnchor.colinear(dirStatement[1].ind.start, dirStatement[1].ind.end):
                                        continue
                                    anchorVect = anchor

                                    # check the orientation: comparing clockwise property between:
                                    # _ anchor_projection -> contained point -> from point
                                    # vs
                                    # _ projection of from point on new line -> contained point -> from point
                                    orientationAnchor = graphMap3D.cw(planeNode.nodeElement, basePt1, basePt2, projAnchor)
                                    dirCheck = orientationAnchor == orientationNew
                                    break

                                self.assignCase(objNode, graphNode3D.lineNode.line_casus.CASE_ANGLE, 
                                                [dirStatement, planeNode, vec, basePt1, basePt2, anchor, dirCheck], transformableVects)
                                return True
                    return False
            # CASE_PLANE_PLANE_DISTANCE: plane-plane directive
            elif dirStatement[0].ind.type == 'ind-planeSegmentP3D':
                if directiveName == 'distance' and dirStatement[1].ind.type == 'ind-planeSegmentP3D':
                    extraVect = None
                    dirCheck = None
                    for vec in transformableVects:
                        if plane3D.containsPoint(dirStatement[1].ind, vec):
                            continue
                        extraVect = vec
                        # find the projection of vec onto the plane
                        projAnchor0 = dirStatement[1].ind.projection(vec)
                        # find the projection of the vec onto the base line
                        projAnchor1 = dirStatement[1].ind.projection(objNode.nodeElement.pRoot)

                        # check the orientation 
                        orientationAnchor = vec.subtract(projAnchor0).unit()
                        orientationNew = objNode.nodeElement.pRoot.subtract(projAnchor1).unit()
                        dirCheck = orientationAnchor == orientationNew
                        break
                    self.assignCase(objNode, graphNode3D.planeNode.plane_casus.CASE_PLANE_PLANE_DISTANCE, 
                                    [dirStatement, extraVect, dirCheck], transformableVects)
                    return True
                return False
            elif dirStatement[0].ind.type == 'ind-pointP3D':
                if directiveName == 'distance':
                    # CASE_POINT_POINT_DISTANCE_DIRECTION: pt-pt distance directive with direction
                    if dirStatement[1].ind.type == 'ind-pointP3D' and len(dirStatement) == 4:
                        self.assignPointCase(objNode, graphNode3D.pointNode.point_casus.CASE_POINT_POINT_DISTANCE_DIRECTION, 
                                        [dirStatement], transformableVects)
                        return True

                    # all other cases: find a line containing the point
                    containingLine = None
                    for l in self.nodeList.lines + availableLineNodes:
                        if l.nodeElement.containsPoint(dirStatement[0].ind.value) and self.checkRHSLineCase(l, transformableVects, availableLineNodes, availablePlaneNodes):
                            containingLine = l.nodeElement
                            break
                    if containingLine is None:
                        return False

                    if dirStatement[1].ind.type == 'ind-pointP3D':
                        perpPlane = plane3D(containingLine.sort, [], True, containingLine.direction, dirStatement[1].ind.value)
                        projNew = perpPlane.projection(dirStatement[0].ind.value)
                        if projNew == dirStatement[0].ind.value:
                            self.assignPointCase(objNode, graphNode3D.pointNode.point_casus.CASE_POINT_POINT_DISTANCE, 
                                            [dirStatement, vec, None, None], transformableVects)
                            return True
                        orientationNew = dirStatement[0].ind.value.subtract(projNew).unit()
                        anchorVect = None
                        dirCheck = None
                        for anchor in transformableVects:
                            # if the base line contains the anchor, reject this
                            if anchor == dirStatement[0].ind.value or perpPlane.containsPoint(anchor):
                                continue
                            anchorVect = anchor

                            # find the projection of the anchor onto the base line
                            anchorProj = perpPlane.projection(anchor)

                            # check the orientation 
                            orientationAnchor = anchor.subtract(anchorProj).unit()
                            dirCheck = orientationAnchor == orientationNew
                            break

                        self.assignPointCase(objNode, graphNode3D.pointNode.point_casus.CASE_POINT_POINT_DISTANCE, 
                                        [dirStatement, containingLine, anchorVect, dirCheck], transformableVects)
                        return True

                    elif dirStatement[1].ind.type == 'ind-lineSegmentP3D':
                        if dirStatement[1].ind.parallel(containingLine):
                            return False
                        # dirStatement[1].ind intersects containingLine case -> CASE_LINE_POINT_DISTANCE_INTERSECT
                        anchorVect = None
                        dirCheck = None
                        projNew = dirStatement[1].ind.projection(dirStatement[0].ind.value)

                        if projNew == dirStatement[0].ind.value:
                            return False
                        orientationNew = dirStatement[0].ind.value.subtract(projNew).unit()

                        for anchor in transformableVects:
                            # if the base line contains the anchor, reject this
                            if anchor == dirStatement[0].ind.value or dirStatement[1].ind.containsPoint(anchor):
                                continue
                            anchorVect = anchor

                            # find the projection of the anchor onto the base line
                            anchorProj = dirStatement[1].ind.projection(anchor)

                            # check the orientation 
                            orientationAnchor = anchor.subtract(anchorProj).unit()
                            dirCheck = orientationAnchor == orientationNew
                            break

                        if dirStatement[1].ind.skew(containingLine):
                            self.assignPointCase(objNode, graphNode3D.pointNode.point_casus.CASE_LINE_POINT_DISTANCE_SKEW, 
                                            [dirStatement, containingLine, anchorVect, dirCheck], transformableVects)
                        else:
                            self.assignPointCase(objNode, graphNode3D.pointNode.point_casus.CASE_LINE_POINT_DISTANCE_INTERSECT, 
                                            [dirStatement, containingLine, anchorVect, dirCheck], transformableVects)
                        return True

    def checkRHSDirectiveObject(self, arg, transformableVects, availableLineNodes, availablePlaneNodes):
        """
        check whether the line/plane/point in a directive argument is 
        """
        if arg.type == 'struct-key':
            return self.findMatchingKeyNode(arg, transformableVects, availableLineNodes, availablePlaneNodes)
        # # check objects used in any description in the directive
        # elif arg.type == 'ind-description':
        #     pass

    def findMatchingKeyNode(self, key, transformableVects, availableLineNodes, availablePlaneNodes):
        keyObject = key.ind

        # check if there's a lhs or rhs node that contains the target object
        checkList = None
        if keyObject.type == 'ind-lineSegmentP3D':
            checkList = self.nodeList.rhsLines + self.nodeList.lines
        elif keyObject.type == 'ind-planeSegmentP3D':
            checkList = self.nodeList.rhsPlanes + self.nodeList.planes
        else:
            return keyObject.value in transformableVects
        for node in checkList:
            if node.nodeElement.equals(keyObject):
                if node.rhsInfo is None or node.rhsInfo['case'] is not None:
                    return True
                break

        # check if there's another directive that define the infinite line/plane containing the target line/plane segment
        for directive in self.directives:
            if directive[0] != 'distance' and directive[0] != 'angle' and directive[0] != 'direction':
                continue
            for dirStatement in directive[1]:
                if dirStatement[0].type == 'struct-key' and dirStatement[0].name == key.name:
                    check = self.checkRHSDirectiveCase(dirStatement, directive[0], transformableVects, 
                                                       availableLineNodes, availablePlaneNodes)
                    return check

        # if line segment
        if keyObject.type == 'ind-lineSegmentP3D':
            if keyObject.start in transformableVects and keyObject.end in transformableVects:
                return True
            containNode = None
            for node in (self.nodeList.lines + self.nodeList.rhsLines):
                if node.nodeElement.equals(keyObject):
                    containNode = node
                    break
            if containNode is None:
                return False
            return self.checkRHSLineCase(containNode, transformableVects, availableLineNodes, availablePlaneNodes)
        elif keyObject.type == 'ind-planeSegmentP3D':
            check = True
            for b in keyObject.boundaries:
                for p in b.polygonal:
                    if p not in transformableVects:
                        check = False
            if check:
                return True
            containNode = None
            for node in (self.nodeList.planes + self.nodeList.rhsPlanes):
                if node.nodeElement.equals(keyObject):
                    containNode = node
                    break
            if containNode is None:
                return False
            return self.checkRHSPlaneCase(containNode, transformableVects, availableLineNodes, availablePlaneNodes)
        elif keyObject.type == 'ind-pointP3D':
            return keyObject.value in transformableVects
        return True



    def checkRHSLineCase(self, lnNode, transformableVects, ori_AvailableLineNodes, ori_AvailablePlaneNodes):

        # if it is a lhs node (lnNode.rhsInfo is None) or under any case other than unsolvable => return True 
        if lnNode.rhsInfo is None or (lnNode.rhsInfo['case'] is not None 
        and lnNode.rhsInfo['case'] != graphNode3D.lineNode.line_casus.CASE_UNSOLVABLE):
            return True
        # if under unsolvable case => return False
        if (lnNode.rhsInfo['case'] is not None
        and lnNode.rhsInfo['case'] == graphNode3D.lineNode.line_casus.CASE_UNSOLVABLE):
            return False
        # leftover: is rhs node but has no case assigned => try to assign a cases

        # make a copy of available node list for use, remove the line node from it
        availableLineNodes = copy(ori_AvailableLineNodes)
        availablePlaneNodes = copy(ori_AvailablePlaneNodes)
        check = False
        for i in range(len(availableLineNodes)):
            if lnNode.nodeElement.equals(availableLineNodes[i].nodeElement):
                check = True
                availableLineNodes.pop(i)
                break
        if not check:
            return False

        # proxy check round 1
        # if the node contains 2 or more transformable vectors already => solved by proxy
        transformableCount = 0
        for vec in transformableVects:
            if lnNode.nodeElement.containsPoint(vec):
                transformableCount += 1
        if transformableCount >= 2:
            self.assignCase(lnNode, graphNode3D.lineNode.line_casus.CASE_PROXY, 
                            [], transformableVects)
            return True

        # reorganize parallel and perpendicular lines list so that currently solvable lines are in front
        self.reorganizeRelatedNodes(lnNode.rhsInfo["perpendicular"])
        self.reorganizeRelatedNodes(lnNode.rhsInfo["parallel"])
        # ______ NO PLANE CASES _______
        # CASE_2PERP: unknown plane, 2 perpendicular lines, these two lines must not be parallel to one another
        if len(lnNode.rhsInfo["perpendicular"]) >= 2:
            perp1 = None
            for l in lnNode.rhsInfo["perpendicular"]:
                # check if the line is not parallel to perp1 if it's already asigned and the line is solvable
                if ((perp1 is not None and perp1.nodeElement.parallel(l.nodeElement)) or not
                self.checkRHSLineCase(l, transformableVects, availableLineNodes, availablePlaneNodes)):
                    continue
                # assign to perp1 if it is None
                if perp1 is None:
                    perp1 = l
                    continue

                # assign the case if perp1 already exists
                self.assignCase(lnNode, graphNode3D.lineNode.line_casus.CASE_2PERP, 
                                [perp1, l], transformableVects)
                return True

        # CASE_1PARALLEL_1PT: solvable plane, 1 parallel line, 1 contained point
        if len(lnNode.rhsInfo["parallel"]) > 0:
            for vec in transformableVects:
                if lnNode.nodeElement.containsPoint(vec):
                    for l in lnNode.rhsInfo["parallel"]:
                        if self.checkRHSLineCase(l, transformableVects, availableLineNodes, availablePlaneNodes):
                            self.assignCase(lnNode, graphNode3D.lineNode.line_casus.CASE_1PARALLEL_1PT, 
                                            [l, vec], transformableVects)
                            return True

        # CASE_2PARALLEL: unknown plane, 2 parallel lines, direction from these two lines root to this line root must be perpendicular
        if len(lnNode.rhsInfo["parallel"]) >= 2:
            for i in range(len(lnNode.rhsInfo["parallel"]) - 1):
                l1 = lnNode.rhsInfo["parallel"][i]
                dir1 = l1.nodeElement.root.subtract(lnNode.nodeElement.root)
                for j in range(i + 1, len(lnNode.rhsInfo["parallel"])):
                    l2 = lnNode.rhsInfo["parallel"][j]
                    dir2 = l2.nodeElement.root.subtract(lnNode.nodeElement.root)
                    if (dir1.perpendicular(dir2)
                    and self.checkRHSLineCase(l1, transformableVects, availableLineNodes, availablePlaneNodes)
                    and self.checkRHSLineCase(l2, transformableVects, availableLineNodes, availablePlaneNodes)):
                        pl = plane3D(sortClass.getSort('__dummyPlane3D__'), 
                                    [l1.nodeElement.root,
                                     l1.nodeElement.root.add(l1.nodeElement.direction),
                                     l2.nodeElement.root])

                        # find a vector not on the plane containing the two parallel lines to check for the new line orientation
                        extraVect = None
                        dirCheck = None
                        for vec in transformableVects:
                            if pl.containsPoint(vec):
                                continue
                            extraVect = vec
                            projVec = pl.projection(extraVect)

                            lnRoot = lnNode.nodeElement.root
                            projRoot = pl.projection(lnRoot)

                            orientationAnchor = vec.subtract(projVec).unit()
                            orientationNew = lnRoot.subtract(projRoot).unit()
                            dirCheck = orientationAnchor == orientationNew
                            break

                        self.assignCase(lnNode, graphNode3D.lineNode.line_casus.CASE_2PARALLEL, 
                                        [l1, l2, extraVect, dirCheck], transformableVects)

                        return True
        # ______ SOLVABLE PLANE CASES ______
        for planeNode in (self.nodeList.planes + self.nodeList.rhsPlanes):
            # find a solvable plane
            if (not planeNode.nodeElement.containsLine(lnNode.nodeElement)
            or not self.checkRHSPlaneCase(planeNode, transformableVects, availableLineNodes, availablePlaneNodes)):
                continue
            # cases with 1 perpendicular line
            if len(lnNode.rhsInfo["perpendicular"]) >= 1:
                for l in lnNode.rhsInfo["perpendicular"]:
                    if (planeNode.nodeElement.normal.parallel(l.nodeElement.unitDir)
                    or not self.checkRHSLineCase(l, transformableVects, availableLineNodes, availablePlaneNodes)):
                        continue
                    if (planeNode.nodeElement.containsLine(l.nodeElement)):
                        # CASE_1PL_1PERP_1PT: solvable plane, 1 perpendicular line, 1 contained point
                        for vec in transformableVects:
                            if lnNode.nodeElement.containsPoint(vec):
                                self.assignCase(lnNode, graphNode3D.lineNode.line_casus.CASE_1PL_1PERP_1PT, 
                                                [planeNode, l, vec], transformableVects)
                                return True

                        # CASE_1PL_1PERP_1INT: solvable plane, 1 perpendicular line, 1 intersecting line
                        # *** not considering other perpendicular lines as the intersecting line since 
                        #     it would already be counted as 2 perpendicular lines case
                        if len(lnNode.rhsInfo["intersect"]) > 0:
                            for intLine in lnNode.rhsInfo["intersect"]:
                                if not planeNode.nodeElement.containsLine(intLine.nodeElement):
                                    self.assignCase(lnNode, graphNode3D.lineNode.line_casus.CASE_1PL_1PERP_1INT, 
                                                    [planeNode, l, lnNode.rhsInfo["intersect"][0]], transformableVects)
                                    return True

                    # CASE_1PL_1PERP: solvable plane, 1 perpendicular line
                    else:
                        self.assignCase(lnNode, graphNode3D.lineNode.line_casus.CASE_1PL_1PERP, 
                                        [planeNode, l], transformableVects)
                        return True

            # cases with 1 parallel line
            if len(lnNode.rhsInfo["parallel"]) >=1:
                for l in lnNode.rhsInfo["parallel"]:
                    if not self.checkRHSLineCase(l, transformableVects, availableLineNodes, availablePlaneNodes):
                        continue
                    # CASE_1PL_1PARALLEL_1INT: solvable plane, 1 parallel line, 1 intersecting line
                    # *** check both perpendicular lines and intersecting lines for the last intersecting line for the case
                    if len(lnNode.rhsInfo["intersect"]) > 0:
                        for intLine in lnNode.rhsInfo["intersect"]:
                            if not planeNode.nodeElement.containsLine(intLine.nodeElement):
                                self.assignCase(lnNode, graphNode3D.lineNode.line_casus.CASE_1PL_1PARALLEL_1INT, 
                                                [planeNode, l, lnNode.rhsInfo["intersect"][0]], transformableVects)
                                return True
                    if len(lnNode.rhsInfo["perpendicular"]) > 0:
                        for intLine in lnNode.rhsInfo["perpendicular"]:
                            if not planeNode.nodeElement.containsLine(intLine.nodeElement):
                                self.assignCase(lnNode, graphNode3D.lineNode.line_casus.CASE_1PL_1PARALLEL_1INT, 
                                                [planeNode, l, lnNode.rhsInfo["perpendicular"][0]], transformableVects)
                                return True

        # proxy check round 2
        # if the node contains 2 or more transformable vectors already => solved by proxy
        transformableCount = 0
        for vec in transformableVects:
            if lnNode.nodeElement.containsPoint(vec):
                transformableCount += 1
        if transformableCount >= 2:
            self.assignCase(lnNode, graphNode3D.lineNode.line_casus.CASE_PROXY, [], transformableVects)
            return True
        # lnNode.rhsInfo["case"] = graphNode3D.lineNode.line_casus.CASE_UNSOLVABLE
        return False

    def checkRHSPlaneCase(self, plNode, transformableVects, ori_AvailableLineNodes, ori_AvailablePlaneNodes):
        # if it is a lhs node (plNode.rhsInfo is None) or under any case other than unsolvable => return True 
        if plNode.rhsInfo is None or (plNode.rhsInfo['case'] is not None 
        and plNode.rhsInfo['case'] != graphNode3D.planeNode.plane_casus.CASE_UNSOLVABLE):
            return True
        # if under unsolvable case => return False
        if (plNode.rhsInfo['case'] is not None
        and plNode.rhsInfo['case'] == graphNode3D.planeNode.plane_casus.CASE_UNSOLVABLE):
            return False
        # leftover: is rhs node but has no case assigned => try to assign a cases
        # make a copy of available node list for use, remove the line node from it
        availableLineNodes = copy(ori_AvailableLineNodes)
        availablePlaneNodes = copy(ori_AvailablePlaneNodes)
        check = False
        for i in range(len(availablePlaneNodes)):
            if plNode.nodeElement.equals(availablePlaneNodes[i].nodeElement):
                check = True
                availablePlaneNodes.pop(i)
                break
        if not check:
            return False
        
        # # check all contained lines to see if any of them are solvable and update transformableVects accordingly
        # for lnNode in plNode.rhsInfo['containedLines']:
        #     self.checkRHSLineCase(lnNode, transformableVects, availableLineNodes, availablePlaneNodes)

        # proxy check round 1
        # if the node contains 3 or more transformable vectors already => solved through proxy
        containedPoints = []
        for vec in transformableVects:
            if plNode.nodeElement.containsPoint(vec):
                containedPoints.append(vec)
        while len(containedPoints) >= 3 and containedPoints[2].colinear(containedPoints[0], containedPoints[1]):
            containedPoints.pop(2)
        if len(containedPoints) >= 3:
            self.assignCase(plNode, graphNode3D.planeNode.plane_casus.CASE_PROXY, [], transformableVects)
            return True
        
        # 0 contained transformable point => not possible to solve
        if len(containedPoints) == 0:
            return False

        parallelLines = []
        ###### CASES WITH LINES ######
        for lnNode in (self.nodeList.lines + self.nodeList.rhsLines):
            # CASE_2PARALLELLN_1PT: 2 parallel lines (these 2 lines not parallel to one another), 1 contained point
            if plNode.nodeElement.parallelLine(lnNode.nodeElement) or plNode.nodeElement.containsLine(lnNode.nodeElement):
                if not self.checkRHSLineCase(lnNode, transformableVects, availableLineNodes, availablePlaneNodes):
                    continue
                if len(parallelLines) > 0 and lnNode.nodeElement.parallel(parallelLines[0].nodeElement):
                    continue
                parallelLines.append(lnNode)
                if len(parallelLines) == 2:
                    self.assignCase(plNode, graphNode3D.planeNode.plane_casus.CASE_2PARALLELLN_1PT,
                                    [parallelLines[0], parallelLines[1], containedPoints[0]], transformableVects)
                    return True
            # CASE_1PERPLN_1PT: 1 perpendicular line, 1 contained point
            if plNode.nodeElement.perpendicularLine(lnNode.nodeElement):
                if not self.checkRHSLineCase(lnNode, transformableVects, availableLineNodes, availablePlaneNodes):
                    continue
                self.assignCase(plNode, graphNode3D.planeNode.plane_casus.CASE_1PERPLN_1PT,
                                [lnNode, containedPoints[0]], transformableVects)
                return True

        ###### CASES WITH PLANES ######
        perp1 = None
        for plNode1 in (self.nodeList.planes + self.nodeList.rhsPlanes):
            # CASE_1PARALLEL_1PT: 1 parallel plane, 1 contained point
            if plNode1.nodeElement.parallel(plNode.nodeElement):
                if not self.checkRHSPlaneCase(plNode1, transformableVects, availableLineNodes, availablePlaneNodes):
                    continue
                self.assignCase(plNode, graphNode3D.planeNode.plane_casus.CASE_1PARALLEL_1PT,
                                [plNode1, containedPoints[0]], transformableVects)
                return True
            # CASE_2PERP_1PT: 2 perpendicular plane, 1 contained point
            if plNode1.nodeElement.perpendicular(plNode.nodeElement):
                if perp1 is not None and perp1.nodeElement.parallel(plNode1.nodeElement):
                    continue
                if not self.checkRHSPlaneCase(plNode1, transformableVects, availableLineNodes, availablePlaneNodes):
                    continue
                if perp1 is None:
                    perp1 = plNode1
                    continue
                self.assignCase(plNode, graphNode3D.planeNode.plane_casus.CASE_2PERP_1PT,
                                [perp1, plNode1, containedPoints[0]], transformableVects)
                return True

        # proxy check round 2
        # if the node contains 3 or more transformable vectors already => solved by proxy
        containedPoints = []
        for vec in transformableVects:
            if plNode.nodeElement.containsPoint(vec):
                containedPoints.append(vec)
        while len(containedPoints) >= 3 and containedPoints[2].colinear(containedPoints[0], containedPoints[1]):
            containedPoints.pop(2)
        if len(containedPoints) >= 3:
            self.assignCase(plNode, graphNode3D.planeNode.plane_casus.CASE_PROXY, [], transformableVects)
            return True

        return False



    def assignCase(self, node, case, caseNodes, transformableVects, lhsNode = False):
        node.rhsInfo["case"] = case
        node.rhsInfo["caseNodes"] = caseNodes
        node.rhsInfo["caseTemp"] = True

        # if node.nodeElement.type == 'ind-line3D':
        #     self.rhsSolvingOrder.append(node)
        #     for l in node.forms:
        #         transformableVects.append(l.start)
        #         transformableVects.append(l.end)
        # else:
        #     self.rhsSolvingOrder.append(node)
        if not lhsNode:
            self.rhsSolvingOrder.append(node)


        if len(node.forms) == 0:
            return

        checkDirective = False
        for directive in self.directives:
            # for point_on_line specifically, check if the individual is in any of the lhs node
            if directive[0] == 'point_on_line' :
                if 'ind-line' not in node.nodeElement.type: continue
                for dirStatement in directive[1]:
                    if len(dirStatement) == 2:
                        arg = dirStatement[0]
                        if arg.type != 'struct-key':
                            raise ValueError('Error: lineSegment tag expected for the first argument of point_on_line directive.')
                        if not node.nodeElement.equals(arg.ind): continue
                        for ind in node.forms:
                            if ind.key is not None and ind.key.name == arg.name:
                                for vect in self.rhsResolved:
                                    if ind.containsPoint(vect) and vect not in transformableVects:
                                        transformableVects.append(vect)
                                    node.rhsInfo['directives'].append([directive[0], ind, dirStatement])
                                    checkDirective = True
                                    break
                    elif len(dirStatement) == 3:
                        targetObj = dirStatement[0]
                        if targetObj.type != 'struct-key' and not 'ind-lineSegment' in targetObj.ind.type and not 'ind-point' in targetObj.ind.type:
                            raise ValueError('Error: lineSegment or point tag expected for the first argument of point_on_line directive.')
                        arg = dirStatement[1]
                        if arg.type != 'struct-key':
                            raise ValueError('Error: lineSegment tag expected for the second argument of 3 arguments point_on_line directive.')
                        if 'ind-lineSegment' in targetObj.ind.type:
                            if line3D.containsPoint(arg.ind, targetObj.ind.start):
                                target = targetObj.ind.start
                            elif line3D.containsPoint(arg.ind, targetObj.ind.end):
                                target = targetObj.ind.end
                        elif 'ind-point' in targetObj.ind.type:
                                target = targetObj.ind.value
                        if not node.nodeElement.equals(arg.ind): continue
                        for ind in node.forms:
                            if ind.key is not None and ind.key.name == arg.name:
                                for vect in self.rhsResolved:
                                    if ind.containsPoint(vect) and vect not in transformableVects:
                                        transformableVects.append(vect)
                                    node.rhsInfo['directives'].append([directive[0], ind, dirStatement, target])
                                    checkDirective = True
                                    break
                    else:
                        raise ValueError('Error: 2 or 3 arguments expected for point_on_line directive.')
            elif directive[0] == 'length':
                for dirStatement in directive[1]:
                    for argument in dirStatement:
                        if argument.type != 'struct-key':
                            continue
                        for ind in node.forms:
                            if ind.key is not None and ind.key.name == argument.name:
                                if ind.start not in transformableVects:
                                    transformableVects.append(ind.start)
                                if ind.end not in transformableVects:
                                    transformableVects.append(ind.end)
                                node.rhsInfo['directives'].append([directive[0], ind, dirStatement])
                                checkDirective = True
                                break
            else:
                for dirStatement in directive[1]:
                    if dirStatement[0].type != 'struct-key':
                        continue
                    for ind in node.forms:
                        if ind.key is not None and ind.key.name == dirStatement[0].name:
                            node.rhsInfo['directives'].append([directive[0], ind, dirStatement])
                            checkDirective = True

        if node.nodeElement.type == 'ind-line3D':
            for n in node.rhsInfo['parallel']:
                for ln in n.forms:
                    startProj = node.nodeElement.projection(ln.start)
                    endProj = node.nodeElement.projection(ln.end)
                    for i in node.forms:
                        if (startProj == i.start or endProj == i.start) and i.start not in transformableVects:
                            transformableVects.append(i.start)
                        if (startProj == i.end or endProj == i.end) and i.end not in transformableVects:
                            transformableVects.append(i.end)
            for n in self.nodeList.lines:
                if n.nodeElement.intersect(node.nodeElement):
                    intersection = n.nodeElement.intersection(node.nodeElement)
                    if intersection not in transformableVects:
                        check = False
                        for f in node.forms:
                            if f.start == intersection or f.end == intersection or f.containsPoint(intersection):
                                check = True
                                break
                        if check:
                            transformableVects.append(intersection)
                            continue
                        for f in n.forms:
                            if f.start == intersection or f.end == intersection or f.containsPoint(intersection):
                                check = True
                                break
                        if check:
                            transformableVects.append(intersection)
            for n in self.nodeList.rhsLines:
                if n.rhsInfo['case'] is not None and n.nodeElement.intersect(node.nodeElement):
                    intersection = n.nodeElement.intersection(node.nodeElement)
                    if intersection not in transformableVects:
                        check = False
                        for f in node.forms:
                            if f.start == intersection or f.end == intersection or f.containsPoint(intersection):
                                check = True
                                break
                        if check:
                            transformableVects.append(intersection)
                            continue
                        for f in n.forms:
                            if f.start == intersection or f.end == intersection or f.containsPoint(intersection):
                                check = True
                                break
                        if check:
                            transformableVects.append(intersection)

        elif node.nodeElement.type == 'ind-plane3D':
            for pl in node.forms:
                for b in pl.boundaries:
                    for p in b.polygonal:
                        if p not in transformableVects:
                            transformableVects.append(p)

        if lhsNode:
            if checkDirective:
                for i in range(len(self.rhsSolvingOrder)):
                    if len(self.rhsSolvingOrder[i].rhsInfo['directives']) == 0:
                        self.rhsSolvingOrder.insert(i, node)
                        return
            self.rhsSolvingOrder.append(node)
        # if checkDirective:
        #     self.finalizeCase(node)

    def assignPointCase(self, node, case, caseNodes, transformableVects):
        # print('     assign case', node.nodeElement.printRhino(), case)
        node.rhsInfo = {
            "case": case,
            "caseNodes": caseNodes,
            "caseTemp": True,
            "directives": []
        }

        if node.nodeElement.value not in transformableVects:
            transformableVects.append(node.nodeElement.value)
        if node not in self.rhsSolvingOrder:
            self.rhsSolvingOrder.append(node)

    def finalizeCase(self, targetNode):
        if targetNode.rhsInfo is None:
            return 
        targetNode.rhsInfo['caseTemp'] = False
        for node in targetNode.rhsInfo['caseNodes']:
            if isinstance(node, vector3D) or isinstance(node, list) or targetNode.rhsInfo is None:
                continue
            self.finalizeCase(node)

    def resetTempCase(self):
        for node in self.nodeList.rhsLines:
            if node.rhsInfo['caseTemp'] == True:
                node.rhsInfo["case"] = None
                node.rhsInfo["caseNodes"] = []
                node.rhsInfo['caseTemp'] = False

    def reorganizeRelatedNodes(self, nodeList):
        i = 0
        while i < len(nodeList):
            node = nodeList[i]
            if node.rhsInfo is None or node.rhsInfo['case'] is not None:
                nodeList.pop(i)
                nodeList.insert(0, node)
            i += 1

    def updateResolvedVecs(self, transformableVects):
        for i in range(len(self.nodeList.lines) - 1):
            nodeI = self.nodeList.lines[i]
            for j in range(i, len(self.nodeList.lines)):
                nodeJ = self.nodeList.lines[j]
                if not nodeI.nodeElement.intersect(nodeJ.nodeElement):
                    continue
                intersection = nodeI.nodeElement.intersection(nodeJ.nodeElement)
                notMin = False
                notMax = False
                for lnI in nodeI.forms:
                    if intersection > lnI.start:
                        notMin = True
                    if intersection < lnI.end:
                        notMax = True
                    if notMin and notMax:
                        break
                if notMin and notMax:
                    if intersection not in transformableVects:
                        transformableVects.append(intersection)
                    continue
                notMin = False
                notMax = False
                for lnJ in nodeJ.forms:
                    if intersection > lnJ.start:
                        notMin = True
                    if intersection < lnJ.end:
                        notMax = True
                    if notMin and notMax:
                        break
                if notMin and notMax:
                    if intersection not in transformableVects:
                        transformableVects.append(intersection)


    def addRHSObjects(self, form):
        """
        add all new rhs objects (not solvable with the lhs) into the graphNode

        :param form: form in the rhs
        :type form: ~sortal.form.form.form
        """
        if form.type == 'form-coForm':
            for i in form:
                self.addRHSPlane(i)
        elif form.type == 'form-interval' and form.sort.baseSort().characteristic.type == 'ind-lineSegmentP3D':
            for i in form:
                self.addRHSLine(i)
        elif form.type == 'form-discrete' and form.sort.baseSort().characteristic.type == 'ind-pointP3D':
            for i in form:
                self.addRHSPoint(i)
        elif form.type == 'form-areal' or form.type == 'form-metaForm':
            for i in form:
                self.addRHSObjects(i)

    def addRHSPlane(self, plane):
        """
        add one rhs plane segment into the graphNode

        :param plane: a plane segment in the rhs form
        :type plane: ~sortal.ind.planeSegmentP.planeSegmentP3D
        :param rhsResolved: a list of all unresolved vectors in rhs
        :type rhsResolved: list of vector3D
        """
        addCheck = False
        for b in plane.boundaries:
            for v in b.polygonal:
                if v not in self.protoTransform.unresolved:
                    self.rhsResolved.append(v)
                    addCheck = True
        
        # if there's no new vector not resolved in the LHS in the plane, no need to add it to the RHS plane node list
        if not addCheck:
            return
        planeNode = self.nodeList.RHS_addPlane(plane.parametricNode(), ind = plane)
        planeNode.rhsInfo["required"] = True

        if planeNode.rhsInfo['case'] == graphNode3D.planeNode.plane_casus.CASE_PROXY and planeNode not in self.rhsSolvingOrder:
            self.assignCase(planeNode, graphNode3D.planeNode.plane_casus.CASE_PROXY, planeNode.rhsInfo['caseNodes'], [])
            # self.rhsSolvingOrder.append(planeNode)

    def addRHSLine(self, line):
        """
        add one rhs line segment into the graphNode

        :param line: a line segment in the rhs form
        :type line: ~sortal.ind.lineSegmentP.lineSegmentP3D
        :param rhsResolved: a list of all unresolved vectors in rhs
        :type rhsResolved: list of vector3D
        """
        # if the line's start and end are in the list of lhs unresolved (resolved after lhs matching)
        if line.start in self.protoTransform.unresolved and line.end in self.protoTransform.unresolved:
            # check whether there's a lhs line with the same direction + root, ignore the line if there is
            for node in self.nodeList.lines:
                if node.nodeElement.equals(line):
                    return
            # no line with the same direction + root => solve by proxy through lhs lines
            lineNode = self.nodeList.RHS_addLine(line.parametricNode(), ind = line)
            self.assignCase(lineNode, graphNode3D.lineNode.line_casus.CASE_PROXY, lineNode.rhsInfo['caseNodes'], [], True)
            # lineNode.rhsInfo['case'] == graphNode3D.lineNode.line_casus.CASE_PROXY
            # self.rhsSolvingOrder.append(lineNode)
            return

        if line.start not in self.protoTransform.unresolved and line.start not in self.rhsResolved:
            self.rhsResolved.append(line.start)
        if line.end not in self.protoTransform.unresolved and line.end not in self.rhsResolved:
            self.rhsResolved.append(line.end)
        
        lineNode = self.nodeList.RHS_addLine(line.parametricNode(), ind = line)

        if lineNode.rhsInfo['case'] == graphNode3D.lineNode.line_casus.CASE_PROXY and lineNode not in self.rhsSolvingOrder:
            self.assignCase(lineNode, graphNode3D.lineNode.line_casus.CASE_PROXY, lineNode.rhsInfo['caseNodes'], [], True)
            # self.rhsSolvingOrder.append(lineNode)


    def addRHSPoint(self, point):
        """
        add one rhs point into the graphNode

        :param point: a point in the rhs form
        :type point: ~sortal.ind.pointP.pointP3D
        :param rhsResolved: a list of all unresolved vectors in rhs
        :type rhsResolved: list of vector3D
        """
        if point.value not in self.protoTransform.unresolved:
            self.rhsResolved.append(point.value)


    # ============================================================================================
    #                                    SHAPE MATCHING PROCESS
    # ============================================================================================

    def potentialTransforms(self):
        """
        Find all potential transform for the rule of the mapping with the saved shape
        """
        # analysis processes for the shape
        self.matchingNodeList.planeCase = self.nodeList.planeCase
        self.matchingNodeList.generateLinesPlanes()
        self.updateMatrices()
        self.matchingNodeList.classifyObjs()
        for matchingNode in self.matchingNodeList.lines:
            if len(matchingNode.forms) == 0:
                continue
            matchingNode.addEndPoints()
            # matchingNode.classifyIntersections()
            matchingNode.organizeForm()

        # ####################
        # print('\n0. Plane lines')
        # for j in self.nodeList.planes:
        #     print(len(j.lines))
            
        # print('\n1. Plane matrix')
        # for j in self.matrices[0]:
        #     for k in j:
        #         if k is None:
        #             print(NONE, end=' ')
        #         else:
        #             print (k,end=' ')
        #     print()
        # print('\n2. Plane Line matrix')
        # for j in self.matrices[1]:
        #     for k in j:
        #         if k is None:
        #             print(NONE, end=' ')
        #         else:
        #             print (k,end=' ')
        #     print()
        # print('\n3. Line Plane matrix')
        # for j in self.matrices[2]:
        #     for k in j:
        #         if k is None:
        #             print(NONE, end=' ')
        #         else:
        #             print (k,end=' ')
        #     print()
        # print('\n4. Line matrix')
        # for j in self.matrices[3]:
        #     for k in j:
        #         if k is None:
        #             print(NONE, end=' ')
        #         else:
        #             print (k,end=' ')
        #     print()
        # print()
        # print('\n\n-----------------------------------------------------------------\n\n')
        # print('\n0. Plane lines')
        # for j in self.matchingNodeList.planes:
        #     print(len(j.lines))
        # print('\n1. Plane matrix')
        # for j in self.matchingMatrices[0]:
        #     for k in j:
        #         if k is None:
        #             print(NONE, end=' ')
        #         else:
        #             print (k,end=' ')
        #     print()
        # print('\n2. Plane Line matrix')
        # for j in self.matchingMatrices[1]:
        #     for k in j:
        #         if k is None:
        #             print(NONE, end=' ')
        #         else:
        #             print (k,end=' ')
        #     print()
        # print('\n3. Line Plane matrix')
        # for j in self.matchingMatrices[2]:
        #     for k in j:
        #         if k is None:
        #             print(NONE, end=' ')
        #         else:
        #             print (k,end=' ')
        #     print()
        # print('\n4. Line matrix')
        # for j in self.matchingMatrices[3]:
        #     for k in j:
        #         if k is None:
        #             print(NONE, end=' ')
        #         else:
        #             print (k,end=' ')
        #     print()
        # print()


        # detect general matches using the matrices
        # things being taken into account: 
        #   line-line: orientation type (parallel/intersect/skew) & intersection type (outside, bound, inside)
        #   line-plane: orientation type (parallel/intersect) & intersection type (outside, bound, inside)
        #   plane-line: orientation type (parallel/intersect) & intersection type (outside, bound, inside)
        #   plane-plane: orientation type (parallel/intersect) & intersection type (outside, bound, inside)
        transforms = []
        matchLists = []
        self.detectTransform(matchLists)

        # start with an empty transform
        ptransf = [self.protoTransform.duplicate()]
        
        # resolve points that are not on any line first
        if not self.resolveFreePoints(ptransf):
            return []
        # investigate each of the matches in detectTransform's result
        for matches in matchLists:
            planeMatch = matches[0]
            lineMatch = matches[1]

            # print('\n-----------------------------------------------------------')
            # print('*match:', lineMatch)
            # s = [0,0,0,2, int(765/len(lineMatch))]
            # for index, num in enumerate(lineMatch):
            #     print(self.nodeList.lines[index].nodeElement.printRhino(-1000), end='')
            #     print(' SelLast -Properties o c o '+str(s[0])+','+str(s[1])+','+str(s[2])+' EnterEnd SelNone', end=' ')
            #     print(self.matchingNodeList.lines[num].nodeElement.printRhino(), end='')
            #     print(' SelLast -Properties o c o '+str(s[0])+','+str(s[1])+','+str(s[2])+' EnterEnd SelNone')
            #     s[s[3]] += s[4]
            #     if s[s[3]] > 255:
            #         s[3] -= 1
            #         s[s[3]] = s[s[3]+1] - 255
            #         s[s[3]+1] = 255
            # print()

            # initialize a graph transform object for this particular match
            transf = [] 
            for tr in ptransf:
                transf.append(tr.duplicate())

            check = True
            noIntNodes = []
            singleIntNodes = []
            orientChecks = []
            
            # plane-segments match: only resolve points that are contained by plane-segments for now
            for nodeIndex in range(len(self.nodeList.planes)):
                nodeI = self.nodeList.planes[nodeIndex]
                matchNodeI = self.matchingNodeList.planes[planeMatch[nodeIndex]]
                self.resolvePointsOnPlane(nodeI, matchNodeI, transf)
            
            # x = [ [8, 9, 1, 2, 18, 19, 11, 10],
            #     [8, 6, 1, 4, 16, 17, 13, 12],
            #     [7, 6, 0, 2, 17, 16, 10, 11],
            #     [7, 5, 0, 4, 15, 14, 12, 13],
            #     [1, 4, 8, 6, 13, 12, 16, 17],
            #     [0, 4, 7, 5, 12, 13, 15, 14],
            #     [1, 2, 8, 9, 11, 10, 18, 19],
            #     [0, 2, 7, 6, 10, 11, 17, 16]]
            # checkcheck = False
            # for xxxxxxx in x:
            #     if xxxxxxx == lineMatch:
            #         checkcheck = True
            #         break
            # if checkcheck:
            #     print('\n\n\n-----------------------')
            #     print(lineMatch)
            # line-segments match:
            for nodeIndex in range(len(self.nodeList.lines)):

                nodeI = self.nodeList.lines[nodeIndex]
                matchNodeI = self.matchingNodeList.lines[lineMatch[nodeIndex]]
                noInt = True

                # if checkcheck:
                #     print('\n')
                #     print(nodeI.nodeElement.printRhino())
                #     print(matchNodeI.nodeElement.printRhino())
                # for each of the other lines in the match, 
                for j in range(len(self.nodeList.lines)):
                    if j == nodeIndex: continue
                    # if the two lhs lines in the two matches are not parallel or skew (parallel = 2 & skew = 3)
                    if self.matchingMatrices[3][lineMatch[nodeIndex]][lineMatch[j]][0] < PARALLEL:
                        # marks as having intersection
                        noInt = False

                        if j <= nodeIndex:
                            continue

                        # vectFrom: lhs intersection; vectTo: shape intersection
                        vectFrom = self.getIntersection(nodeI, self.nodeList.lines[j])
                        vectTo = self.getIntersection(matchNodeI,self.matchingNodeList.lines[lineMatch[j]])

                        # if vectFrom or vectTo is None (the two lhs/shape lines are parallel)
                        # -> the match is still accepted, move on to another line to compare
                        if vectFrom is None or vectTo is None:
                            continue


                        # if vectFrom is not contained by any line-segment in the lhs, add it as extra instead of normal transformation
                        if vectFrom.iType == 'o':
                            for tr in transf:
                                tr.addExtra(vectFrom.value, vectTo.value)
                            continue

                        # if both vectFrom and VectTo are defined, add them to the transformation
                        trIndex = 0
                        while trIndex < len(transf):
                            tr = transf[trIndex]
                            # if the intersection match cannot be added to the graph transform
                            # (due to vectFrom already have a different match from VectTo, etc...),
                            # the graph transform is deemed invalid
                            if tr.contains(vectFrom.value) and tr.transform(vectFrom.value) == vectTo.value:
                                trIndex += 1
                                continue
                            if tr.containsMatch(vectTo.value) or not tr.add(vectFrom.value, vectTo.value, index = 0):
                                transf.pop(trIndex)
                            else:
                                trIndex += 1

                # if after comparing this one line with all other lines and there is no intersection
                # -> special cases: 1. only one line in the lhs
                #                   2. more than one line in the lhs but they are all parallel to one another
                if noInt:
                    noIntNodes.append([nodeI, matchNodeI])
                    continue
                # check the ordering of the intersections
                if len(transf) == 0 or not self.orderingCheck(transf, nodeI):
                    check = False
                    break

                # if the node doesnt have any end points 
                # -> all points on the lhs line already have a match 
                # -> continue onto other nodes
                if (not nodeI.endP) and len(nodeI.linePoints) == 0:
                    continue

                # check if the node and matchingNode are matched in reverse
                # i.e.  node:           --A---B------C----
                #       matchingNode:   ----X--Y--Z----W------
                #       A matches with W, B matches with Z, C matches with X
                #       --> the two nodes match in reverse order
                startI = 0
                while (not transf[0].contains(nodeI.intersections[startI].value)):
                    startI += 1
                    if startI >= len(nodeI.intersections):
                        noInt = True
                        break
                if noInt:
                    noIntNodes.append([nodeI, matchNodeI])
                    continue

                # start is to be the match of the smallest point that already has a matching (based on the lhs line)
                start = transf[0].apply(nodeI.intersections[startI].value)
                
                endI = -1
                while not transf[0].contains(nodeI.intersections[endI].value):
                    endI -= 1
                    if endI < -len(nodeI.intersections):
                        noInt = True
                        break
                if noInt:
                    noIntNodes.append([nodeI, matchNodeI])
                    continue

                # end is to be the match of the biggest point that already has a matching (based on the lhs line)
                end = transf[0].apply(nodeI.intersections[endI].value)

                # if start == end -> the matched intersection points are the same
                if start == end:
                    # if the lhs start and end intersection points are not the same but their matches are 
                    # -> the match is discarded
                    if nodeI.intersections[startI] != nodeI.intersections[endI]:
                        check = False
                        break
                    # otherwise, there is only one intersection on the line 
                    # -> special case, add the matching to singleIntNodes and resolve later
                    singleIntNodes.append([nodeI, matchNodeI, startI])

                    # Orientation check: check the ordering of intersection relative to other intersections
                    self.addOrientCheck(orientChecks, [nodeI.intersections[startI].value, start])
                    if not self.orientationChecking(orientChecks):
                        check = False
                        singleIntNodes = []
                        break
                    continue
                
                elif start > end:
                    reverse = True
                else:
                    reverse = False

                # resolve points that are on the node but not an intersection
                if len(nodeI.linePoints) > 0:
                    if len(matchNodeI.linePoints) < len(nodeI.linePoints):
                        check = False
                        break
                    self.resolvePointsOnLine(nodeI, matchNodeI, reverse, transf)

                # if the node does not have any line-segment's start or end point unresolved, continue to the next node
                if not nodeI.endP:
                    continue
                ##### resolve all the unresolved start/end point on the node #####

                # get all existing resolved points on the node
                # nodeInts = transf[0].getCheckPoints(nodeI.nodeElement)
                nodeInts = self.getCheckPoints(transf[0], nodeI, True)
                lineForm = []
                specialLineForm = []
                lineIndex = 0
                # for each of the line-segment on the node
                while lineIndex < len(nodeI.forms):
                    # check if the line-segment contains at least one intersection point in it
                    # if yes, add it to lineForm along with any other line-segments that touches it
                    if transf[0].hasCheckPointOf(nodeI.forms[lineIndex]):
                        lineForm.append(nodeI.forms[lineIndex])
                        prevCheck = lineIndex
                        while prevCheck > 0 and nodeI.forms[prevCheck-1].end == nodeI.forms[prevCheck].start:
                            lineForm.append(nodeI.forms[prevCheck-1])
                            prevCheck -= 1
                        while lineIndex < len(nodeI.forms)-1 and nodeI.forms[lineIndex].end == nodeI.forms[lineIndex+1].start:
                            lineForm.append(nodeI.forms[lineIndex+1])
                            lineIndex += 1
                    lineIndex += 1

                # any line-segments that are not in lineForm, add them to specialLineForm for special consideration
                for line in nodeI.forms:
                    if line not in lineForm:
                        specialLineForm.append(line)

                # resolve line-segments in specialLineForm first
                for line in specialLineForm:
                    # for each of the transformation
                    transfIndex = 0
                    while transfIndex < len(transf):
                        tr = transf.pop(transfIndex)
                        # nodeInts = tr.getCheckPoints(nodeI.nodeElement)
                        nodeInts = self.getCheckPoints(tr, nodeI, True)
                        # resolve special line
                        posMatches = self.resolveSpecialLine(line, nodeInts, reverse, nodeI, matchNodeI)
                        if len(posMatches) == 0:
                            continue
                        # add the matches to the transformation sets
                        for posM in posMatches:
                            additionalTransf = tr.duplicate()
                            if not reverse:
                                additionalTransf.add(line.start, posM.start)
                                additionalTransf.add(line.end, posM.end)                            
                            else:
                                additionalTransf.add(line.start, posM.end)
                                additionalTransf.add(line.end, posM.start)
                            transf.insert(transfIndex, additionalTransf)
                            transfIndex += 1
                # no matches for special lines -> discard the transformation
                if not check:
                    break
                # resolve all other normal line-segments
                ind = 0
                while ind < len(transf):
                    tr = transf[ind]
                    trCheck = True
                    # each of the normal line-segment
                    for line in lineForm:

                        # resolve normal line-segment
                        hasMatch = self.resolveLine(line, reverse, nodeI, matchNodeI, tr)
                        if not hasMatch:
                            transf.pop(ind)
                            trCheck = False

                            break
                    if not trCheck:
                        break
                    for intersection in nodeI.intersections:
                        if intersection.iType != 'o' and not tr.contains(intersection.value):
                            transf.pop(ind)
                            ind -=1
                            break

                    graphMap3D.normalizeLineEndPoints(nodeI, tr, reverse)
                    ind += 1
            if len(transf) == 0:
                continue
            
            # make sure there's one existing point match for each lhs point object that is on an intersection
            if not self.checkIntersectionPoints(transf):
                continue
            # resolve the nodes with no intersections
            if len(noIntNodes) > 0:
                transf = self.resolveNoIntersectionNodes(lineMatch, noIntNodes, transf)
                if transf is False:
                    continue
            # if there are 2 lines with single intersection and they intersect each other
            if len(singleIntNodes) == 2 and singleIntNodes[0][0].nodeElement.intersect(singleIntNodes[1][0].nodeElement):
                self.resolveSingleIntersectionNodes([singleIntNodes[0]], transf, True)
                self.resolveSingleIntersectionNodes([singleIntNodes[1]], transf, True)

            # if there is only 1 line with 1 intersection in the whole shape 
            # -> 2 cases: 1. there are at least 2 other lines that intersect this line at a single point
            #                and there are another line parallel to this that intersect the other 2 lines at different points 
            #             2. the line is on a whole different plane compared with the rest of the shape
            elif len(singleIntNodes) == 1:
                if len(transf) == 0:
                    continue
                
                nodeI = singleIntNodes[0][0]
                matchNodeI = singleIntNodes[0][1]

                # anchor: the single intersection; nAnchor: the match of anchor
                anchor = nodeI.intersections[singleIntNodes[0][2]].value
                nAnchor = transf[0].apply(anchor)
                vec = [[anchor, nAnchor]]

                # add all other vect
                for v in transf[0].vects:
                    if v[0] == anchor:
                        nAnchor = v[1]
                        continue
                    vec.append(v)
                p = nodeI.forms[0].end if nodeI.forms[0].start == anchor else nodeI.forms[0].start

                vecCheck = True if len(vec) > 1 else False

                while len(vec) > 1 and not vec[1][0].subtract(vec[0][0]).parallel(anchor.subtract(p)):
                    vec.pop(1)
                    if len(vec) == 1:
                        vecCheck = False
                        break

                # if there is another line that is parallel to this line -> case 1
                if vecCheck:
                    cPlane1 = plane3D(sortClass.getSort('__dummyPlane3D__'), [vec[0][0], vec[1][0],anchor])
                    cPlane2 = plane3D(sortClass.getSort('__dummyPlane3D__'), [vec[0][1], vec[1][1],nAnchor])
                    self.resolveSingleIntersectionNodes(singleIntNodes, transf)
                    for index in range(len(transf)):
                        tr = transf.pop(0)
                        nP = tr.apply(p)
                        cw1 = graphMap3D.cw(cPlane1, vec[0][0], vec[1][0], anchor)
                        cw2 = graphMap3D.cw(cPlane1, vec[0][0], anchor, p)
                        cw3 = graphMap3D.cw(cPlane2, vec[0][1], vec[1][1], nAnchor)
                        cw4 = graphMap3D.cw(cPlane2, vec[0][1], nAnchor, nP)
                        if (cw1 == cw2) == (cw3 == cw4):
                            transf.append(tr)
                # no other line parallel to this line -> case 2
                else:
                    self.resolveSingleIntersectionNodes([singleIntNodes[0]], transf, True)

            # all other cases
            else:
                self.resolveSingleIntersectionNodes(singleIntNodes, transf)

            # merge all the extra vector transformations back into the main list
            for tr in transf:
                tr.mergeExtra()

            self.resolveColinearEndPoints(transf)


            # resolve planes and curves
            for i in range(len(self.nodeList.planes)):
                planeNode = self.nodeList.planes[i]
                matchingPlaneNode = self.matchingNodeList.planes[planeMatch[i]]
                self.resolvePlanes(transf, planeNode, matchingPlaneNode, lineMatch)
                self.resolveCurves(transf, planeNode, matchingPlaneNode)
            # resolve curves that are not on any existing plane
            self.resolveCurves(transf, self.nodeList, self.matchingNodeList)

            # register the transformation
            for tr in transf:
                transforms.append(tr)
        i = 0
        while i < len(transforms):
            j = i+1
            while j < len(transforms):
                if transforms[i] == transforms[j]:
                    transforms.pop(j)
                else:
                    j += 1
            i += 1

        return transforms
        
    def detectTransform(self, matchLists):
        """
        Find all the line-node and plane-node matches by comparing the line-line/line-plane/plane-line/plane-plane matrices
        """
        # if there are no plane -> no entry in plane-plane matrix
        # -> only need to consider the line-line matrix
        # NOT USED IN ACTUALITY SINCE PLANE NODES ARE GENERATED IF THERE ARE ONLY LINE NODES
        if len(self.matrices[0]) == 0:
            for subset in itertools.permutations([k for k in range(len(self.matchingNodeList.lines))], len(self.nodeList.lines)):
                check = True
                for indI, i in enumerate(subset):
                    for indJ, j in enumerate(subset):
                        if i == j:
                            continue
                        lhsEdge = self.matrices[3][indI][indJ]
                        matchEdge = self.matchingMatrices[3][i][j]
                        if (lhsEdge[0] == SKEW and matchEdge[0] == SKEW):
                            continue
                        if (((lhsEdge[0] == PARALLEL or lhsEdge[0] == SKEW) and lhsEdge[0] != matchEdge[0])
                        or (matchEdge[0] == PARALLEL and lhsEdge[1] != OUTSIDE)
                        or (lhsEdge[0] > matchEdge[0]) 
                        or lhsEdge[1] > matchEdge[1]):
                            check = False
                            break
                    if not check:
                        break
                if check:
                    matchLists.append([(),subset])
                # matchLists.append([(),subset])
            return matchLists


        # find plane matches, check plane-plane matrix
        # planeMatches = self.findRoughMatches('plane')
        planeMatches = []
        roughMatches = self.findRoughMatches('plane')
        for subset in roughMatches:
            check = True
            for indI, i in enumerate(subset):
                numLines = len(self.nodeList.planes[indI].lines)
                matchingNumLines = len(self.matchingNodeList.planes[i].lines)
                if numLines > matchingNumLines:
                    check = False
                    break
            if not check:
                continue
            for indI, i in enumerate(subset):
                for indJ, j in enumerate(subset):
                    if i == j:
                        continue
                    lhsEdge = self.matrices[0][indI][indJ]
                    matchEdge = self.matchingMatrices[0][i][j]
                    # reject if:
                    # lhs planes are parallel but matching planes are not
                    # lhs plane intersection is inside/bound but matching planes are parallel
                    # lhs edge is bigger than match edge (lhsEdge[0] = perpedicular (1) but matchEdge[0] = any (0))
                    # lhs intersection type is bigger than match intersection type (lhsEdge[1] = inside (2) but matchEdge[1] = inside (0))
                    if ((lhsEdge[0] == PARALLEL and lhsEdge[0] != matchEdge[0])
                    or (matchEdge[0] == PARALLEL and lhsEdge[1] != OUTSIDE)
                    or (lhsEdge[0] > matchEdge[0]) 
                    or lhsEdge[1] > matchEdge[1]):
                        # print('--2 ',subset, i, j, lhsEdge, matchEdge)
                        check = False
                        break
                if not check:
                    break
            ###########################################################
            if check:
                planeMatches.append(subset)
        planeMatches = sorted(planeMatches)

        # find line matches, check line-line matrix
        # lineMatches = self.findRoughMatches('line')
        lineMatches = []
        roughMatches = self.findRoughMatches('line')
        for subset in roughMatches:
            check = True
            # print('~~~',subset)
            for indI, i in enumerate(subset):
                for indJ, j in enumerate(subset):
                    if i == j:
                        continue
                    lhsEdge = self.matrices[3][indI][indJ]
                    matchEdge = self.matchingMatrices[3][i][j]
                    if (lhsEdge[0] == SKEW and matchEdge[0] == SKEW):
                        continue
                    # reject if:
                    # lhs lines are parallel/skew but matching lines are not the same
                    # lhs line intersection is inside/bound but matching lines are parallel
                    # lhs edge is bigger than match edge (lhsEdge[0] = perpedicular (1) but matchEdge[0] = any (0))
                    # lhs intersection type is bigger than match intersection type (lhsEdge[1] = inside (2) but matchEdge[1] = inside (0))
                    if (((lhsEdge[0] == PARALLEL or lhsEdge[0] == SKEW) and lhsEdge[0] != matchEdge[0])
                    or (matchEdge[0] == PARALLEL and lhsEdge[1] != OUTSIDE)
                    or (lhsEdge[0] > matchEdge[0]) 
                    or lhsEdge[1] > matchEdge[1]):
                        check = False
                        break
                if not check:
                    break
            if check:
                lineMatches.append(subset)





        # # find plane matches, check plane-plane matrix
        # planeMatches = []
        # # badMatches = []
        # for subset in itertools.permutations([k for k in range(len(self.matchingNodeList.planes))], len(self.nodeList.planes)):
        #     check = True
        #     # for badMatch in badMatches:
        #     #     if subset[badMatch[0]] == badMatch[1] and subset[badMatch[2]] == badMatch[3]:
        #     #         check = False
        #     #         break
        #     # if not check:
        #     #     continue
        #     for indI, i in enumerate(subset):
        #         numLines = len(self.nodeList.planes[indI].lines)
        #         matchingNumLines = len(self.matchingNodeList.planes[i].lines)
        #         if numLines > matchingNumLines:
        #             # print('--1 ',subset)
        #             check = False
        #             break
        #     if not check:
        #         continue
        #     for indI, i in enumerate(subset):
        #         for indJ, j in enumerate(subset):
        #             if i == j:
        #                 continue
        #             lhsEdge = self.matrices[0][indI][indJ]
        #             matchEdge = self.matchingMatrices[0][i][j]
        #             # reject if:
        #             # lhs planes are parallel but matching planes are not
        #             # lhs plane intersection is inside/bound but matching planes are parallel
        #             # lhs edge is bigger than match edge (lhsEdge[0] = perpedicular (1) but matchEdge[0] = any (0))
        #             # lhs intersection type is bigger than match intersection type (lhsEdge[1] = inside (2) but matchEdge[1] = inside (0))
        #             if ((lhsEdge[0] == PARALLEL and lhsEdge[0] != matchEdge[0])
        #             or (matchEdge[0] == PARALLEL and lhsEdge[1] != OUTSIDE)
        #             or (lhsEdge[0] > matchEdge[0]) 
        #             or lhsEdge[1] > matchEdge[1]):
        #                 # badMatches.append([indI, i, indJ, j])
        #                 check = False
        #                 break
        #         if not check:
        #             break
        #     ###########################################################
        #     if check:
        #         planeMatches.append(subset)

        # # find line matches, check line-line matrix
        # lineMatches = []
        # # badMatches = []
        # for subset in itertools.permutations([k for k in range(len(self.matchingNodeList.lines))], len(self.nodeList.lines)):
        #     check = True
        #     # for badMatch in badMatches:
        #     #     if subset[badMatch[0][0]] == badMatch[0][1]:
        #     #         for badPairing in badMatch[1]:
        #     #             if subset[badPairing[0]] == badPairing[1]:
        #     #                 check = False
        #     #                 break
        #     #         if not check: break
        #     # if not check:
        #     #     continue


        #     for indI, i in enumerate(subset):
        #         for indJ, j in enumerate(subset):
        #             if i == j:
        #                 continue
        #             lhsEdge = self.matrices[3][indI][indJ]
        #             matchEdge = self.matchingMatrices[3][i][j]
        #             if (lhsEdge[0] == SKEW and matchEdge[0] == SKEW):
        #                 continue
        #             # reject if:
        #             # lhs lines are parallel/skew but matching lines are not the same
        #             # lhs line intersection is inside/bound but matching lines are parallel
        #             # lhs edge is bigger than match edge (lhsEdge[0] = perpedicular (1) but matchEdge[0] = any (0))
        #             # lhs intersection type is bigger than match intersection type (lhsEdge[1] = inside (2) but matchEdge[1] = inside (0))
        #             if (((lhsEdge[0] == PARALLEL or lhsEdge[0] == SKEW) and lhsEdge[0] != matchEdge[0])
        #             or (matchEdge[0] == PARALLEL and lhsEdge[1] != OUTSIDE)
        #             or (lhsEdge[0] > matchEdge[0]) 
        #             or lhsEdge[1] > matchEdge[1]):
        #                 check = False
        #                 # if (indI < indJ):
        #                 #     existing = False
        #                 #     for badMatch in badMatches:
        #                 #         if badMatch[0][0] == indI and badMatch[0][1] == i:
        #                 #             existing = True
        #                 #             badMatch[1].append([indJ, j])
        #                 #     if not existing:
        #                 #         badMatches.append([[indI, i],[[indJ, j]]])
        #                 break
        #         if not check:
        #             break
        #     if check:
        #         lineMatches.append(subset)

        # check the line-plane/plane-line matrices
        for plMatch in planeMatches:
            for lnMatch in lineMatches:
                check = True

                for indI, i in enumerate(plMatch):
                    for indJ, j in enumerate(lnMatch):
                        lhsPL = self.matrices[1][indI][indJ]
                        matchPL = self.matchingMatrices[1][i][j]
                        
                        if ((lhsPL[0] == EMBEDDED and matchPL[0] != EMBEDDED)
                        or (lhsPL[0] == PARALLEL and lhsPL[0] != matchPL[0])
                        or (matchPL[0] == PARALLEL and lhsPL[1] != OUTSIDE)
                        or (lhsPL[0] > matchPL[0]) 
                        or lhsPL[1] > matchPL[1]):
                            check = False
                            break

                        lhsLP = self.matrices[2][indJ][indI]
                        matchLP = self.matchingMatrices[2][j][i]
                        
                        if ((lhsLP[0] == EMBEDDED and matchLP[0] != EMBEDDED)
                        or (lhsLP[0] == PARALLEL and lhsLP[0] != matchLP[0])
                        or (matchLP[0] == PARALLEL and lhsLP[1] != OUTSIDE)
                        or (lhsLP[0] > matchLP[0]) 
                        or lhsLP[1] > matchLP[1]):
                            check = False
                            break
                        
                    if not check:
                        break
                if check:
                    #matchLists.append(lnMatch)
                    matchLists.append([plMatch, lnMatch])
        # print('\n\n\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~')
        # for i in matchLists:
        #     print(i[0], i[1])
        # matchLists = sorted(matchLists)
        # print('\n\n\n............................')
        # for i in matchLists:
        #     print(i[0], i[1])
        return matchLists

    def findRoughMatches(self, matchType):
        if matchType == 'line':
            lhsMatrix = self.matrices[3]
            shapeMatrix = self.matchingMatrices[3]
            # nodeListObjs = self.nodeList.lines
        elif matchType == 'plane':
            lhsMatrix = self.matrices[0]
            shapeMatrix = self.matchingMatrices[0]
            # nodeListObjs = self.nodeList.planes
        if len(lhsMatrix) <= 1:
            return [[i] for i in range(len(shapeMatrix))]
            
        # count all the relationship of each node in the lhs matrix
        lhsCount = []
        for i in lhsMatrix:
            nodeCount = {'par': 0, 
                         'skw': 0, 
                         'per-o': 0, 'per-b': 0, 'per-i': 0, 
                         'any-o': 0, 'any-b': 0, 'any-i': 0,
                         'pts': 0}
            for j in i:
                if j is None:
                    continue
                if j[0] == SKEW:
                    nodeCount['skw'] += 1
                elif j[0] == PARALLEL:
                    nodeCount['par'] += 1
                elif j[0] == PERPENDICULAR:
                    if j[1] == OUTSIDE:
                        nodeCount['per-o'] += 1
                    elif j[1] == BOUND:
                        nodeCount['per-b'] += 1
                    elif j[1] == INSIDE:
                        nodeCount['per-i'] += 1
                elif j[0] == ANY:
                    if j[1] == OUTSIDE:
                        nodeCount['any-o'] += 1
                    elif j[1] == BOUND:
                        nodeCount['any-b'] += 1
                    elif j[1] == INSIDE:
                        nodeCount['any-i'] += 1
            lhsCount.append(nodeCount)
        if matchType == 'line':
            for i in range(len(self.nodeList.lines)):
                nodeCount = lhsCount[i]
                nodeCount['pts'] = len(self.nodeList.lines[i].containedPoints)
        
        # count all the relationship of each node in the shape matrix
        shapeCount = []
        for i in shapeMatrix:
            nodeCount = {'skw': 0,
                         'par': 0,
                         'per-o': 0, 'per-b': 0, 'per-i': 0,
                         'any-o': 0, 'any-b': 0, 'any-i': 0,
                         'pts': 0}
            for j in i:
                if j is None:
                    continue
                if j[0] == SKEW:
                    nodeCount['skw'] += 1
                elif j[0] == PARALLEL:
                    nodeCount['par'] += 1
                elif j[0] == PERPENDICULAR:
                    if j[1] == OUTSIDE:
                        nodeCount['per-o'] += 1
                    elif j[1] == BOUND:
                        nodeCount['per-b'] += 1
                    elif j[1] == INSIDE:
                        nodeCount['per-i'] += 1
                elif j[0] == ANY:
                    if j[1] == OUTSIDE:
                        nodeCount['any-o'] += 1
                    elif j[1] == BOUND:
                        nodeCount['any-b'] += 1
                    elif j[1] == INSIDE:
                        nodeCount['any-i'] += 1
            shapeCount.append(nodeCount)
        if matchType == 'line':
            for j in range(len(self.matchingNodeList.lines)):
                nodeCount = shapeCount[j]
                nodeCount['pts'] = len(self.matchingNodeList.lines[j].containedPoints)

        # find all the bad matches in the lhs and shape matrix. bad matches are matches where the shape node does not 
        # have the required amount of corresponding edge/half-edge compared to the lhs node (i.e. lhs node has 3 parallel
        # nodes but shape node only has 2, lhs node has 3 perpendicular inside but shape node has 3 perpendicular with 2 
        # outside and 1 inside, ...)
        badMatches = []
        for i in range(len(lhsCount)):
            for j in range(len(shapeCount)):
                l = lhsCount[i]
                s = shapeCount[j]
                if (l['skw'] > s['skw']
                or  l['par'] > s['par']
                or  l['per-o'] > (s['per-o'] + s['per-b'] + s['per-i'])
                or  l['per-b'] > (             s['per-b'] + s['per-i'])
                or  l['per-i'] > (                          s['per-i'])
                or  l['any-o'] > (s['any-o'] + s['any-b'] + s['any-i'] + s['per-o'] + s['per-b'] + s['per-i'] + s['par'])
                or  l['any-b'] > (             s['any-b'] + s['any-i']              + s['per-b'] + s['per-i'])
                or  l['any-i'] > (                          s['any-i']                           + s['per-i'])
                or  l['pts'] > s['pts']):
                        badMatches.append(str(i) + ' ' + str(j))
        
        # find all pairs of good matches: pair of matches where the shape matrix edge/half-edge value corresponds with the 
        # lhs edge/half-edge value. must not contain matches that are in bad matches list
        goodMatches = []
        for indI in range(len(lhsMatrix) - 1):
            for indJ in range(indI + 1, len(lhsMatrix[indI])):
                lhsEdge1 = lhsMatrix[indI][indJ]
                lhsEdge2 = lhsMatrix[indJ][indI]
                for i in range(len(shapeMatrix)):
                    if str(indI) + ' ' + str(i) in badMatches:
                        continue
                    for j in range(len(shapeMatrix[i])):
                        if i == j:
                            continue
                        if str(indJ) + ' ' + str(j) in badMatches:
                            continue

                        matchEdge1 = shapeMatrix[i][j]
                        matchEdge2 = shapeMatrix[j][i]

                        if matchType == 'line':
                            if (lhsEdge1[0] == SKEW and matchEdge1[0] == SKEW):
                                pass
                            elif (((lhsEdge1[0] == PARALLEL or lhsEdge1[0] == SKEW) and lhsEdge1[0] != matchEdge1[0])
                            or (matchEdge1[0] == PARALLEL and (lhsEdge1[1] != OUTSIDE or lhsEdge2[1] != OUTSIDE))
                            or (lhsEdge1[0] > matchEdge1[0])
                            or lhsEdge1[1] > matchEdge1[1]
                            or lhsEdge2[1] > matchEdge2[1]):
                                continue
                        elif matchType == 'plane':
                            if ((lhsEdge1[0] == PARALLEL and lhsEdge1[0] != matchEdge1[0])
                            or (matchEdge1[0] == PARALLEL and (lhsEdge1[1] != OUTSIDE or lhsEdge2[1] != OUTSIDE))
                            or (lhsEdge1[0] > matchEdge1[0])
                            or lhsEdge1[1] > matchEdge1[1]
                            or lhsEdge2[1] > matchEdge2[1]):
                                continue

                        existing = False
                        for goodMatch in goodMatches:
                            if goodMatch[0][0] == indI and goodMatch[0][1] == i:
                                existing = True
                                goodMatch[1].append([indJ, j])
                        if not existing:
                            goodMatches.append([[indI, i],[[indJ, j]]])
        
        # go through the good matches list and compile a list of matches. Any two node matches in any of these matches must be good match.
        index = len(lhsMatrix) - 1
        roughMatches = []
        for goodMatch in reversed(goodMatches):
            if index != goodMatch[0][0]:
                index -= 1
                matchIndex = 0
                while matchIndex < len(roughMatches):
                    if len(roughMatches[matchIndex]) < len(lhsMatrix) - goodMatch[0][0] - 1:
                        roughMatches.pop(matchIndex)
                    else:
                        matchIndex += 1
            if goodMatch[0][0] == len(lhsMatrix) - 2:
                for m in goodMatch[1]:
                    roughMatches.append([goodMatch[0][1], m[1]])
            else:
                for matchInd in range(len(roughMatches)):
                    match = roughMatches[matchInd]
                    if len(match) > len(lhsMatrix) - goodMatch[0][0] - 1:
                        break
                    checkIndex = goodMatch[0][0] + 1
                    matchIndex = 0
                    for gMatch in goodMatch[1]:
                        if checkIndex == gMatch[0] and match[matchIndex] == gMatch[1]:
                            checkIndex += 1
                            matchIndex += 1
                        elif checkIndex < gMatch[0]:
                            break
                    if checkIndex == len(lhsMatrix):
                        newMatch = copy(match)
                        newMatch.insert(0, goodMatch[0][1])
                        roughMatches.append(newMatch)
        matchIndex = 0
        while matchIndex < len(roughMatches):
            if len(roughMatches[matchIndex]) < len(lhsMatrix):
                roughMatches.pop(matchIndex)
            else:
                matchIndex += 1

        # roughMatches = []
        # for goodMatch in goodMatches:
        #     if goodMatch[0][0] != 0:
        #         break
        #     matches = [[goodMatch[0][1]]]

        #     for missingIndex in range(1, len(nodeListObjs)):
        #         matchesCount = len(matches)
        #         for i in range(matchesCount):
        #             match = matches.pop(0)
        #             for validMatch in goodMatch[1]:
        #                 if validMatch[0] == missingIndex and validMatch[1] not in match:
        #                     newMatch = copy(match)
        #                     newMatch.append(validMatch[1])
        #                     matches.append(newMatch)
        #                 elif validMatch[0] > missingIndex:
        #                     break
        #     roughMatches += matches



        # for i in range(1, len(nodeListObjs) - 1):
        #     valitatingMatches = set()
        #     for goodMatch in goodMatches:
        #         if goodMatch[0][0] > i: break
        #         elif goodMatch[0][0] != i: continue
        #         matches = [[goodMatch[0][1]]]
        #         for missingIndex in range(i + 1, len(nodeListObjs)):
        #             matchesCount = len(matches)
        #             for j in range(matchesCount):
        #                 match = matches.pop(0)
        #                 for validMatch in goodMatch[1]:
        #                     if validMatch[0] == missingIndex and validMatch[1] not in match:
        #                         newMatch = copy(match)
        #                         newMatch.append(validMatch[1])
        #                         matches.append(newMatch)
        #                     elif validMatch[0] > missingIndex:
        #                         break
        #         for m in matches:
        #             valitatingMatches.add(str(m))
        #     matchI = 0
        #     while matchI < len(roughMatches):
        #         check = False
        #         if str(roughMatches[matchI][i:]) in valitatingMatches:
        #             check = True
        #             break
        #         # for valMatch in valitatingMatches:
        #         #     if roughMatches[matchI][i:] == valMatch:
        #         #         check = True
        #         #         break
        #         if not check:
        #             roughMatches.pop(matchI)
        #         else:
        #             matchI += 1
        return roughMatches
        
    # resolve end-points of special line-segment: line-segments that do not contain any intersection point on it
    def resolveSpecialLine(self, line, nodeInts, reverse, nodeI, matchNodeI):
        posMatch = []
        # if line.end is before the smallest intersection of the node
        if line.end < nodeInts[0][0]:
            checkEnd = False

            # check if there is any other line-segment in the node that contains the smallest intersection 
            # and the smallest intersection is not that line-segment's start point
            for lineCheck in nodeI.forms:
                if lineCheck.containsPoint(nodeInts[0][0]) and nodeInts[0][0] != lineCheck.start:
                    checkEnd = True
            
            # for each of the line-segment on the matching node
            for mLine in matchNodeI.forms:
                # remove lines that have different sort, have insufficient attribute, 
                # line starting point not smaller than smallest point match if not reverse or
                # line ending point not larger than smallest point match if reverse
                if line.sort.baseSort() != mLine.sort.baseSort() or (line.hasAttribute() 
                and not line.attribute.checkPartOf(mLine.attribute, flagDesc = False)):
                    continue
                if (not reverse) and mLine.start >= nodeInts[0][1]:
                    continue
                elif reverse and mLine.end <= nodeInts[0][1]:
                    continue

                # if there's no line that contain the smallest intersection, this line can match up to the smallest intersection match
                if not checkEnd:
                    if not reverse:
                        matchStart = mLine.start
                        matchEnd = nodeInts[0][1] if mLine.containsPoint(nodeInts[0][1]) else mLine.end
                    else:
                        matchStart = nodeInts[0][1] if mLine.containsPoint(nodeInts[0][1]) else mLine.start
                        matchEnd = mLine.end
                    newLine = mLine.__class__(mLine.sort, matchStart, matchEnd)
                # if there's another line that contain the smallest intersection, do not match if mLine contains the intersection
                else:
                    if checkEnd and mLine.containsPoint(nodeInts[0][1]):
                        continue
                    newLine = mLine
                posMatch.append(newLine)
        # if line.start is before the largest intersection of the node
        elif line.start > nodeInts[-1][0]:
            # check if there is any other line-segment in the node that contains the largest intersection 
            # and the largest intersection is not that line-segment's end point
            checkStart = False
            for lineCheck in nodeI.forms:
                if lineCheck.containsPoint(nodeInts[-1][0]) and nodeInts[-1][0] != lineCheck.end:
                    checkStart = True

            # for each of the line-segment on the matching node
            for mLine in matchNodeI.forms:
                # remove lines that have different sort, have insufficient attribute, 
                # line ending point not larger than largest point match if not reverse or
                # line starting point not smaller than largest point match if reverse
                if line.sort.baseSort() != mLine.sort.baseSort() or (line.hasAttribute() 
                and not line.attribute.checkPartOf(mLine.attribute, flagDesc = False)):
                    continue
                if (not reverse) and mLine.end <= nodeInts[-1][1]:
                    continue
                elif reverse and mLine.start >= nodeInts[-1][1]:
                    continue
                # if there's no line that contain the largest intersection, this line can match up to the largest intersection match
                if not checkStart:
                    if not reverse:
                        matchStart = nodeInts[-1][1] if mLine.containsPoint(nodeInts[-1][1]) else mLine.start
                        matchEnd = mLine.end
                    else:
                        matchStart = mLine.start
                        matchEnd = nodeInts[-1][1] if mLine.containsPoint(nodeInts[-1][1]) else mLine.end
                    newLine = mLine.__class__(mLine.sort, matchStart, matchEnd)
                # if there's another line that contain the largest intersection, do not match if mLine contains the intersection
                else:
                    if checkStart and mLine.containsPoint(nodeInts[-1][1]):
                        continue
                    newLine = mLine
                posMatch.append(newLine)
        # else -> line is between the smallest and biggest intersections
        else:
            index = 1
            while line.start > nodeInts[index][0]:
                index += 1
            # find the matches of the closest intersections that the line is between as lowerBound and upperBound
            if reverse:
                upperBound = nodeInts[0][1]
                for indexC in range(1,index):
                    if upperBound > nodeInts[indexC][1]:
                        upperBound = nodeInts[indexC][1]
                lowerBound = nodeInts[-1][1]
                for indexC in range(index,len(nodeInts)-1):
                    if lowerBound < nodeInts[indexC][1]:
                        lowerBound = nodeInts[indexC][1]
            else:
                lowerBound = nodeInts[0][1]
                for indexC in range(1,index):
                    if lowerBound < nodeInts[indexC][1]:
                        lowerBound = nodeInts[indexC][1]
                upperBound = nodeInts[-1][1]
                for indexC in range(index,len(nodeInts)-1):
                    if upperBound > nodeInts[indexC][1]:
                        upperBound = nodeInts[indexC][1]
            # check if there are any other lines that contains the two bounding intersections
            checkStart = False
            checkEnd = False
            for lineCheck in nodeI.forms:
                if lineCheck.containsPoint(nodeInts[index-1][0]) and nodeInts[index-1][0] != lineCheck.end:
                    checkStart = True
                if lineCheck.containsPoint(nodeInts[index][0]) and nodeInts[index][0] != lineCheck.start:
                    checkEnd = True

            # for each of the line-segment on the matching node
            for mLine in matchNodeI.forms:
                # remove lines that have different sort, have insufficient attribute, 
                # or doesn't have any part inside the bound
                if line.sort.baseSort() != mLine.sort.baseSort() or (line.hasAttribute() 
                and not line.attribute.checkPartOf(mLine.attribute, flagDesc = False)):
                    continue
                if mLine.end <= lowerBound or mLine.start >= upperBound:
                    continue
                
                # if mLine contains lowerBound
                if mLine.containsPoint(nodeInts[index-1][1]):
                    if checkStart:
                        continue
                    else:
                        matchStart = nodeInts[index-1][1]
                elif not reverse:
                    matchStart = mLine.start
                else:
                    matchStart = mLine.end

                if mLine.containsPoint(nodeInts[index][1]):
                    if checkEnd:
                        continue
                    else:
                        matchEnd = nodeInts[index][1]
                elif not reverse:
                    matchEnd = mLine.end
                else:
                    matchEnd = mLine.start
                newLine = mLine.__class__(mLine.sort, matchStart, matchEnd)
                posMatch.append(newLine)
        return posMatch
        
    # resolve end points of line-segments that contains at least one intersection in the node
    def resolveLine(self, line, reverse, nodeI, matchNodeI, transf):
        """
        resolve end points of line-segments that contains at least one intersection in the node
        """

        # check if the line-segment start and end points already have matches and add the unmatched one(s) to unmatchedPts
        checkStart = transf.contains(line.start)
        checkEnd = transf.contains(line.end)
        if checkStart and checkEnd:
            return True
        if checkStart:
            unmatchedPts = [line.end]
        elif checkEnd:
            unmatchedPts = [line.start]
        else:
            unmatchedPts = [line.start, line.end]
        
        
        hasMatch = False

        # get all existing node intersections with matches
        segmentInts = []
        nodeInts = []
        for nInt in nodeI.intersections:
            if nInt.endP:
                if line.containsPoint(nInt.value) and transf.contains(nInt.value):
                    pass
                else:
                    continue
            matchingInt = transf.transform(nInt.value)
            if not matchingInt:
                continue
            if line.start == nInt.value or line.end == nInt.value:
                alignment = BOUND
                segmentInts.append([nInt.value, matchingInt, alignment])
            elif line.containsPoint(nInt.value):
                alignment = INSIDE
                segmentInts.append([nInt.value, matchingInt, alignment])
            nodeInts.append([nInt.value, matchingInt])
        
        if (len(nodeInts) < 2):
            segmentInts = []
            nodeInts = []
            for nInt in nodeI.intersections:
                if nInt.endP:
                    if line.containsPoint(nInt.value) and transf.contains(nInt.value):
                        pass
                    else:
                        continue
                matchingInt = transf.allTransform(nInt.value)
                if not matchingInt:
                    continue
                if line.start == nInt.value or line.end == nInt.value:
                    alignment = BOUND
                    segmentInts.append([nInt.value, matchingInt, alignment])
                elif line.containsPoint(nInt.value):
                    alignment = INSIDE
                    segmentInts.append([nInt.value, matchingInt, alignment])
                nodeInts.append([nInt.value, matchingInt])

        # checking against all line-segments in the matching node
        for mLine in matchNodeI.forms:
            # if line's attribute is not a part of matching line's attribute --> continue
            if line.sort.baseSort() != mLine.sort.baseSort() or (line.hasAttribute() 
                and not line.attribute.checkPartOf(mLine.attribute, flagDesc = False)):
                continue
            
            # check if the matching line has the contains the matching intersections of intersections on line
            lineCheck = True
            for segInt in segmentInts:
                if (mLine.containsPoint(segInt[1]) 
                or (segInt[2] == BOUND and (mLine.start == segInt[1] or mLine.end == segInt[1]))):
                    continue
                lineCheck = False
                break
            if not lineCheck:
                continue
            hasMatch = True
            # for each unmatched points
            for pointIndex, point in enumerate(unmatchedPts):
                posMatch = None # possible match
                orderingPriority = False
                if point < segmentInts[0][0]:
                    if point > nodeInts[0][0]:
                        if reverse:
                            # point is smaller than first matched point on segment
                            # but bigger than first matched point on node
                            # and the node is matched in reverse
                            # --> matching point must be:
                            #       bigger than first matched point on matching segment
                            #       smaller or equals to first matched point on matching node
                            #       be on the matching segment
                            orderingPriority = '>'
                            for mInt in matchNodeI.intersections:
                                if (mInt.value > segmentInts[0][1]  
                                and mInt.value <= nodeInts[0][1]
                                and mLine.containsPoint(mInt.value)):
                                    posMatch = mInt.value
                        else:
                            # point is smaller than first matched point on segment
                            # but bigger than first matched point on node
                            # and the node is NOT matched in reverse
                            # --> matching point must be:
                            #       smaller than first matched point on matching segment
                            #       bigger or equals to first matched point on matching node
                            #       be on the matching segment
                            orderingPriority = '<'
                            for mInt in reversed(matchNodeI.intersections):
                                if (mInt.value < segmentInts[0][1]  
                                and mInt.value >= nodeInts[0][1]
                                and mLine.containsPoint(mInt.value)):
                                    posMatch = mInt.value
                    else:
                        if reverse:
                            # point is smaller than first matched point on segment,
                            # also smaller than first matched point on node 
                            # and the node is matched in reverse
                            # --> matching point must be:
                            #       bigger than first matched point on matching node
                            #       be on the matching segment
                            orderingPriority = '>'
                            for mInt in matchNodeI.intersections:
                                if (mInt.value > nodeInts[0][1]  
                                and mLine.containsPoint(mInt.value)):
                                    posMatch = mInt.value
                        else:
                            # point is smaller than first matched point on segment,
                            # also smaller than first matched point on node 
                            # and the node is NOT matched in reverse
                            # --> matching point must be:
                            #       smaller than first matched point on matching node
                            #       be on the matching segment
                            orderingPriority = '<'
                            for mInt in reversed(matchNodeI.intersections):
                                if (mInt.value < nodeInts[0][1]  
                                and mLine.containsPoint(mInt.value)):
                                    posMatch = mInt.value
                elif point > segmentInts[-1][0]:
                    if point < nodeInts[-1][0]:
                        if reverse:
                            # point is bigger than last matched point on segment,
                            # but smaller than last matched point on node 
                            # and the node is matched in reverse
                            # --> matching point must be:
                            #       smaller than last matched point on matching segment
                            #       bigger or equals to last matched point on matching node
                            #       be on the matching segment
                            orderingPriority = '<'
                            for mInt in reversed(matchNodeI.intersections):
                                if (mInt.value < segmentInts[-1][1]  
                                and mInt.value >= nodeInts[-1][1] ###
                                and mLine.containsPoint(mInt.value)):
                                    posMatch = mInt.value
                        else:
                            # point is bigger than last matched point on segment,
                            # but smaller than last matched point on node 
                            # and the node is NOT matched in reverse
                            # --> matching point must be:
                            #       bigger than last matched point on matching segment
                            #       smaller or equals to last matched point on matching node
                            #       be on the matching segment
                            orderingPriority = '>'
                            for mInt in matchNodeI.intersections:
                                if (mInt.value > segmentInts[-1][1]  
                                and mInt.value <= nodeInts[-1][1] ###
                                and mLine.containsPoint(mInt.value)):
                                    posMatch = mInt.value
                    else:
                        if reverse:
                            # point is bigger than last matched point on segment,
                            # also bigger than last matched point on node 
                            # and the node is matched in reverse
                            # --> matching point must be:
                            #       smaller than last matched point on matching node
                            #       be on the matching segment
                            orderingPriority = '<'
                            for mInt in reversed(matchNodeI.intersections):
                                if (mInt.value < nodeInts[-1][1]  
                                and mLine.containsPoint(mInt.value)):
                                    posMatch = mInt.value
                        else:
                            # point is bigger than last matched point on segment,
                            # also bigger than last matched point on node 
                            # and the node is NOT matched in reverse
                            # --> matching point must be:
                            #       bigger than last matched point on matching node
                            #       be on the matching segment
                            orderingPriority = '>'
                            for mInt in matchNodeI.intersections:
                                if (mInt.value > nodeInts[-1][1]  
                                and mLine.containsPoint(mInt.value)):
                                    posMatch = mInt.value
                else:
                    # point is in between the first and last matched points on segment
                    # narrow down the range in which the point can be 
                    index = 1
                    while point > segmentInts[index][0] and index < len(segmentInts):
                        index += 1
                    # the point must be now between segmentInts[index-1] and segmentInts[index]
                    # i.e. point is bigger than point on node at (index-1) 
                    #          and smaller than point on node at (index) 
                    if reverse:
                        # the node is matched in reverse
                        # --> matching point must be:
                        #       smaller than matched point on matching node at (index-1)
                        #       bigger than matched point on matching node at (index)
                        #       be on the matching segment
                        for mInt in reversed(matchNodeI.intersections):
                            if (mInt.value < segmentInts[index-1][1]  
                            and mInt.value > segmentInts[index][1]
                            and mLine.containsPoint(mInt.value)):
                                posMatch = mInt.value
                    else:
                        # the node is NOT matched in reverse
                        # --> matching point must be:
                        #       bigger than matched point on matching node at (index-1)
                        #       smaller than matched point on matching node at (index)
                        #       be on the matching segment
                        for mInt in matchNodeI.intersections:
                            if (mInt.value > segmentInts[index-1][1]  
                            and mInt.value < segmentInts[index][1]
                            and mLine.containsPoint(mInt.value)):
                                posMatch = mInt.value
                if posMatch is None:
                    continue
                # add the posible matches
                transf.add(point, posMatch, orderingPriority)
        for i in unmatchedPts:
            if not transf.contains(i):
                return False
        return hasMatch
        
    def resolveSingleIntersectionNodes(self, singleIntNodes, transf, noCheckPoint = False):
        """
        resolve nodes with only one single intersection
        """
        if len(singleIntNodes) == 0:
            return transf
        transf1 = copy(transf)
        for sn in singleIntNodes:
            # find the matches for reverse order
            transfIndex = 0
            while transfIndex < len(transf):
                tr = transf.pop(transfIndex)
                newTransf = self.resolveSingleIntersectionLine(sn[0],sn[1],tr,sn[2], True, noCheckPoint)
                for i in newTransf:
                    transf.insert(transfIndex, i)
                    transfIndex += 1

            # find the matches for normal order
            transfIndex1 = 0
            while transfIndex1 < len(transf1):
                tr = transf1.pop(transfIndex1)
                newTransf = self.resolveSingleIntersectionLine(sn[0],sn[1],tr,sn[2], False, noCheckPoint)
                for i in newTransf:
                    transf1.insert(transfIndex1, i)
                    transfIndex1 += 1
        transf += transf1
        return transf
    
    def resolveSingleIntersectionLine(self, nodeI, matchNodeI, tr, anchorIndex, reverse, noCheckPoint = False):
        """
        resolve nodes with only one single intersection
        """
        if len(nodeI.forms) > len(matchNodeI.forms) or len(nodeI.linePoints) > len(matchNodeI.linePoints):
            return []
        
        # check the orientation of the external intersections if this line has any as compared to the reverse parameter
        if len(nodeI.intersections) > 1:
            transformableIntersections = []
            for i in nodeI.intersections:
                match = tr.getExtraMatch(i.value)
                if match:
                    transformableIntersections.append([i.value, match])
            if len(transformableIntersections) > 1:
                intReverse = ((transformableIntersections[0][0] < transformableIntersections[1][0])
                            != (tr.apply(transformableIntersections[0][1]) < tr.apply(transformableIntersections[1][1])))
                for i in range(1, len(transformableIntersections) - 1):
                    intCurr = transformableIntersections[i]
                    intNext = transformableIntersections[i + 1]
                    if ((intCurr[0] < intNext[0]) != (intCurr[1] < intNext[1])) != intReverse:
                        intReverse = None
                        break
                if intReverse is not None and intReverse != reverse:
                    return []


        # the single intersection on the line
        anchorFrom = nodeI.intersections[anchorIndex].value
        anchorTo = tr.apply(nodeI.intersections[anchorIndex].value)
        # check if the anchor is on any of the line-segments
        anchorCheck = False
        for line in nodeI.forms:
            if line.containsPoint(anchorFrom):
                anchorCheck = True
                break


        lineMatches = []
        # checkPoint: one point that is not contained by this line (for clockwise orientation check)
        checkPoint = None
        for vect in tr.vects:
            if (not nodeI.nodeElement.containsPoint(vect[0])) and (not matchNodeI.nodeElement.containsPoint(vect[1])):
                checkPoint = vect
                break
        # if no point that is not on the node has matches 
        if noCheckPoint or checkPoint is None:
            # find all possible matches for each of the line-segment in the node.
            # only consider the lines relative position compared to the only intersection the line has.
            # the ordering of the line-segments are not considered yet
            for line in nodeI.forms:
                # line before anchor -> lineCheck = -1
                # line contains anchor -> lineCheck = 0
                # line after anchor -> lineCheck = 1
                if line.containsPoint(anchorFrom):
                    lineCheck = 0
                elif line.start > anchorFrom:
                    lineCheck = 1
                else:
                    lineCheck = -1

                posMatch = []
                # for each line-segments in matching node
                for mLine in matchNodeI.forms:
                    # ignore lines with different sort and attribute
                    if line.sort.baseSort() != mLine.sort.baseSort() or (line.hasAttribute() 
                    and not line.attribute.checkPartOf(mLine.attribute, flagDesc = False)):
                        continue

                    # find matching line-segment
                    if not reverse:
                        if lineCheck == 0 and mLine.containsPoint(anchorTo):
                            if line.start == anchorFrom:
                                if anchorTo == mLine.end: continue
                                newLine = mLine.__class__(mLine.sort, anchorTo, mLine.end)
                            elif line.end == anchorFrom:
                                if anchorTo == mLine.start: continue
                                newLine = mLine.__class__(mLine.sort, mLine.start, anchorTo)
                            else:
                                newLine = mLine
                        elif (not anchorCheck) and lineCheck == 1 and mLine.containsPoint(anchorTo):
                            if mLine.end == anchorTo: continue
                            newLine = mLine.__class__(mLine.sort, anchorTo, mLine.end)
                        elif (not anchorCheck) and lineCheck == -1 and mLine.containsPoint(anchorTo):
                            if mLine.start == anchorTo: continue
                            newLine = mLine.__class__(mLine.sort, mLine.start, anchorTo)
                        elif lineCheck == 1 and mLine.start > anchorTo:
                            newLine = mLine
                        elif lineCheck == -1 and mLine.end < anchorTo:
                            newLine = mLine
                        else:
                            continue
                    else:
                        if lineCheck == 0 and mLine.containsPoint(anchorTo):
                            if line.start == anchorFrom:
                                if anchorTo == mLine.start: continue
                                newLine = mLine.__class__(mLine.sort, mLine.start, anchorTo)
                            elif line.end == anchorFrom:
                                if anchorTo == mLine.end: continue
                                newLine = mLine.__class__(mLine.sort, anchorTo, mLine.end)
                            else:
                                newLine = mLine
                        elif (not anchorCheck) and lineCheck == 1 and mLine.containsPoint(anchorTo):
                            if mLine.start == anchorTo: continue
                            newLine = mLine.__class__(mLine.sort, mLine.start, anchorTo)
                        elif (not anchorCheck) and lineCheck == -1 and mLine.containsPoint(anchorTo):
                            if mLine.end == anchorTo: continue
                            newLine = mLine.__class__(mLine.sort, anchorTo, mLine.end)
                        elif lineCheck == 1 and mLine.end < anchorTo:
                            newLine = mLine
                        elif lineCheck == -1 and mLine.start > anchorTo:
                            newLine = mLine
                        else:
                            continue
                    posMatch.append(newLine)
                if len(posMatch) == 0:
                    return []
                # add the possible matches to lineMatches
                lineMatches.append((line, posMatch))
                
            
            indexing = [0] * len(lineMatches)
            transf = []
            transfIndex = 0

            # check all combinations of line matching to see whether they are valid: check the line-segments ordering
            while True:
                additionalTransf = tr.duplicate()
                addCheck = True
                for l in range(len(indexing)):
                    line = lineMatches[l][0]
                    posM = lineMatches[l][1][indexing[l]]
                    if not reverse:
                        additionalTransf.add(line.start, posM.start)
                        additionalTransf.add(line.end, posM.end)
                    else:
                        additionalTransf.add(line.start, posM.end)
                        additionalTransf.add(line.end, posM.start)
                    
                # check the ordering of the resolved points on the node for additionalTransf
                if not self.orderingCheck([additionalTransf], nodeI, True):
                    addCheck = False
                if addCheck:
                    transf.insert(transfIndex, additionalTransf)

                # advance the indexing
                transfIndex += 1
                indexing[0] += 1
                for i in range(len(indexing)-1):
                    if indexing[i] == len(lineMatches[i][1]):
                        indexing[i] = 0
                        indexing[i+1] += 1
                    else:
                        break
                if indexing[-1] == len(lineMatches[-1][1]):
                    break
            # resolve points that are on the node
            self.resolvePointsOnLine(nodeI, matchNodeI, reverse, transf)
            return transf
        
        # if there are points not on the node that already have matches
        else:
            # find each line-segment matches based on their clockwise/counterclockwise relation with checkPoint
            for line in nodeI.forms:
                posMatch = []

                # check if the line contains anchor
                containsAnchor = line.containsPoint(anchorFrom)

                # cPoint is the end point that is not anchor
                cPoint = line.end if line.start == anchorFrom else line.start

                # plane for lhs orientation check
                cPlane1 = plane3D(sortClass.getSort('__dummyPlane3D__'), [cPoint, anchorFrom, checkPoint[0]])

                # lhs orientation
                cwCheck = graphMap3D.cw(cPlane1, cPoint, anchorFrom, checkPoint[0])

                # for each of the matching lines
                for mLine in matchNodeI.forms:

                    # ignore lines of different sorts & attributes
                    if line.sort.baseSort() != mLine.sort.baseSort() or (line.hasAttribute() 
                    and not line.attribute.checkPartOf(mLine.attribute, flagDesc = False)):
                        continue

                    # nCPoint matching line's end point that is not anchor
                    nCPoint = mLine.end if mLine.start == anchorTo else mLine.start

                    # plane for shape orientation check
                    cPlane2 = plane3D(sortClass.getSort('__dummyPlane3D__'), [nCPoint, anchorTo, checkPoint[1]])
                    # if lhs line-segment contains anchor, the shape line-segment must contain anchor
                    # can match it with no further check
                    if containsAnchor:
                        if not mLine.containsPoint(anchorTo):
                            continue
                        # if anchorTo is either end-points
                        if anchorTo == mLine.end or anchorTo == mLine.start: 
                            if (graphMap3D.cw(cPlane2, nCPoint, anchorTo, checkPoint[1]) == cwCheck) != reverse:
                                newLine = mLine.__class__(mLine.sort, nCPoint, anchorTo)
                            else:
                                continue
                        else:
                            newLine = mLine.__class__(mLine.sort, mLine.start, mLine.end)

                        # newLine = mLine.__class__(mLine.sort, nCPoint, anchorTo)

                    # if lhs line-segment does not contain anchor but the shape line-segment does
                    elif mLine.containsPoint(anchorTo):
                        # check cw/ccw for correct orientation
                        if (graphMap3D.cw(cPlane2, nCPoint, anchorTo, checkPoint[1]) == cwCheck) != reverse:
                            newLine = mLine.__class__(mLine.sort, anchorTo, nCPoint)
                        elif anchorTo != mLine.end and anchorTo != mLine.start:
                            newLine = mLine.__class__(mLine.sort, anchorTo, mLine.end)
                        else:
                            continue
                    # if both do not contain anchor, check for orientation
                    elif (graphMap3D.cw(cPlane2, nCPoint, anchorTo, checkPoint[1]) == cwCheck) != reverse:
                        newLine = mLine
                    else:
                        continue
                    posMatch.append(newLine)
                if len(posMatch) == 0:
                    return []
                # add to lineMatches
                lineMatches.append((line, posMatch))
            
            indexing = [0] * len(lineMatches)
            transf = []
            transfIndex = 0
            # check all combinations of line matching to see whether they are valid: check the line-segments ordering
            while True:
                additionalTransf = tr.duplicate()
                addCheck = True
                for l in range(len(indexing)):
                    line = lineMatches[l][0]
                    posM = lineMatches[l][1][indexing[l]]
                    if anchorFrom == line.start:
                        if anchorTo != posM.start and anchorTo != posM.end:
                            if not reverse:
                                additionalTransf.add(line.end, posM.end)
                            else:
                                additionalTransf.add(line.end, posM.start)
                        elif anchorTo == posM.start:
                            additionalTransf.add(line.end, posM.end)
                        else:
                            additionalTransf.add(line.end, posM.start)
                    elif anchorFrom == line.end:
                        if anchorTo != posM.start and anchorTo != posM.end:
                            if not reverse:
                                additionalTransf.add(line.start, posM.start)
                            else:
                                additionalTransf.add(line.start, posM.end)
                        elif anchorTo == posM.start:
                            additionalTransf.add(line.start, posM.end)
                        else:
                            additionalTransf.add(line.start, posM.start)
                    elif line.start > anchorFrom:
                        if posM.start > anchorTo:
                            additionalTransf.add(line.start, posM.start)
                            additionalTransf.add(line.end, posM.end)
                        elif posM.end < anchorTo:
                            additionalTransf.add(line.start, posM.end)
                            additionalTransf.add(line.end, posM.start)
                        else:
                            raise NotImplementedError('Single Line Matching Special Case Not Implemented Yet')
                    elif line.end < anchorFrom:
                        if posM.end < anchorTo:
                            additionalTransf.add(line.start, posM.start)
                            additionalTransf.add(line.end, posM.end)
                        elif posM.start > anchorTo:
                            additionalTransf.add(line.start, posM.end)
                            additionalTransf.add(line.end, posM.start)
                    else:
                        if not reverse:
                            additionalTransf.add(line.start, posM.start)
                            additionalTransf.add(line.end, posM.end)
                        else:
                            additionalTransf.add(line.start, posM.end)
                            additionalTransf.add(line.end, posM.start)
                        
                if not self.orderingCheck([additionalTransf], nodeI, True):
                    addCheck = False
                
                if addCheck:
                    transf.insert(transfIndex, additionalTransf)
                transfIndex += 1
                indexing[0] += 1
                for i in range(len(indexing)-1):
                    if indexing[i] == len(lineMatches[i][1]):
                        indexing[i] = 0
                        indexing[i+1] += 1
                    else:
                        break
                if indexing[-1] == len(lineMatches[-1][1]):
                    break
            # resolve points that are on the node
            self.resolvePointsOnLine(nodeI, matchNodeI, reverse, transf)
            return transf

    def resolveNoIntersectionNodes(self, match, noIntNodes, transf):
        """
        resolve nodes with no intersection on it
        """
        if len(self.nodeList.lines) == 0:
            return transf


        transf1 = copy(transf)

        # # for each of the node
        # for noIntNode in noIntNodes:
        #     self.resolvePointsOnLine(noIntNode[0], noIntNode[1], True, transf)
        #     self.resolvePointsOnLine(noIntNode[0], noIntNode[1], False, transf1)

        orientChecks = []
        # for each of the node
        for noIntNode in noIntNodes:
            nodeI = noIntNode[0]
            matchNodeI = noIntNode[1]

            # reject matches where the matching node have fewer line-segments than the lhs node
            if len(nodeI.forms) > len(matchNodeI.forms) or len(nodeI.linePoints) > len(matchNodeI.linePoints):
                return False
            
            v0 = nodeI.nodeElement.root
            v1 = matchNodeI.nodeElement.root
            self.addOrientCheck(orientChecks, [v0,v1])

        # check the orientation
        if not self.orientationChecking(orientChecks):
            return False

        # resolve for each node
        for noIntNode in noIntNodes:
            nodeI = noIntNode[0]
            matchNodeI = noIntNode[1]
            
            # reverse
            transfIndex = 0
            while transfIndex < len(transf):
                tr = transf.pop(transfIndex)
                newTransf = self.resolveNoIntersectionLine(nodeI,matchNodeI,tr,True)
                for i in newTransf:
                    transf.insert(transfIndex, i)
                    transfIndex += 1
            self.resolvePointsOnLine(nodeI, matchNodeI, True, transf)
                    
            # normal orientation
            transfIndex1 = 0
            while transfIndex1 < len(transf1):
                tr = transf1.pop(transfIndex1)
                newTransf = self.resolveNoIntersectionLine(nodeI,matchNodeI,tr,False)
                for i in newTransf:
                    transf1.insert(transfIndex1, i)
                    transfIndex1 += 1
            self.resolvePointsOnLine(nodeI, matchNodeI, False, transf1)
        transf += transf1
        
        return transf

    def resolveNoIntersectionLine(self, nodeI, matchNodeI, tr, reverse):
        """
        resolve a node with no intersection on it
        """
        transf = []

        # check the orientation of the external intersections if this line has any as compared to the reverse parameter
        if len(nodeI.intersections) > 1:
            transformableIntersections = []
            for i in nodeI.intersections:
                match = tr.getExtraMatch(i.value)
                if match:
                    transformableIntersections.append([i.value, match])
            if len(transformableIntersections) > 1:
                intReverse = ((transformableIntersections[0][0] < transformableIntersections[1][0])
                            != (tr.apply(transformableIntersections[0][1]) < tr.apply(transformableIntersections[1][1])))
                for i in range(1, len(transformableIntersections) - 1):
                    intCurr = transformableIntersections[i]
                    intNext = transformableIntersections[i + 1]
                    if ((intCurr[0] < intNext[0]) != (intCurr[1] < intNext[1])) != intReverse:
                        intReverse = None
                        break
                if intReverse is not None and intReverse != reverse:
                    return []

        # all permutation of line-segment combination
        for permSet in itertools.permutations(list(range(len(matchNodeI.forms))), len(nodeI.forms)):
            check = True

            # reject matches with mismatching sorts and attributes
            for j in range(len(permSet)):
                # reject matches with different sort
                if nodeI.forms[j].sort != matchNodeI.forms[permSet[j]].sort:
                    check = False
                    break
                # check attributes
                if nodeI.forms[j].sort.type == 'sort-attribute' :
                    if (not reverse) and not nodeI.forms[j].attribute.checkPartOf(matchNodeI.forms[permSet[j]].attribute, flagDesc = False):
                        check = False
                        break
                    elif reverse and not nodeI.forms[j].attribute.checkPartOf(matchNodeI.forms[permSet[len(permSet) - j - 1]].attribute, flagDesc = False):
                        check = False
                        break
            if not check:
                continue
            
            newtr = tr.duplicate()

            # add the line according to the orientation
            if not reverse:
                for j in range(len(permSet)):
                    newtr.add(nodeI.forms[j].start, matchNodeI.forms[permSet[j]].start)
                    newtr.add(nodeI.forms[j].end, matchNodeI.forms[permSet[j]].end)
            else:
                for j in range(len(permSet)):
                    k = len(permSet) - j - 1
                    newtr.add(nodeI.forms[j].start, matchNodeI.forms[permSet[k]].end)
                    newtr.add(nodeI.forms[j].end, matchNodeI.forms[permSet[k]].start)
            if self.orderingCheck([newtr], nodeI, True):
                transf.append(newtr)
        return transf
        
    def resolveFreePoints(self, transf):
        """
        resolve points that are not contained by any node(line/plane)
        """

        # all lhs free points
        fPoints = []
        for p in self.nodeList.points:
            if p.oType == 'f':
                fPoints.append(p.value)
        if len(fPoints) == 0:
            return True

        # all shape free points
        mFPoints = []
        for p in self.matchingNodeList.points:
            if p.oType == 'f':
                mFPoints.append(p.value)
        if len(fPoints)>len(mFPoints):
            return False

        # match through permutation
        transfIndex = 0
        while transfIndex < len(transf):
            tr = transf.pop(transfIndex)
            
            for i in itertools.permutations(mFPoints, len(fPoints)):
                newTr = tr.duplicate()
                check = True
                for j in range(len(fPoints)):
                    if not newTr.add(fPoints[j], i[j]):
                        check = False
                if not check:
                    continue
                transf.insert(transfIndex, newTr)
                transfIndex += 1
        return True

    def resolvePointsOnPlane(self, nodeI, matchNodeI, transf):
        """
        resolve points that are contained by a plane but not contained by any line
        """

        # match through permutation
        transfIndex = 0
        while transfIndex < len(transf):
            tr = transf.pop(transfIndex)
            for subset in itertools.permutations([k for k in range(len(matchNodeI.points))], len(nodeI.points)):
                newTransf = tr.duplicate()
                check = True
                for i,j in enumerate(subset):
                    if not newTransf.add(nodeI.points[i].value, matchNodeI.points[j].value):
                        check = False
                        break
                if check:
                    transf.insert(transfIndex, newTransf)
                    transfIndex+=1        
        
    def resolvePointsOnLine(self, nodeI, matchNodeI, reverse, transf):
        """
        resolve points that are contained by a line-node
        """
        # for each point on the node
        for p in nodeI.linePoints:
            transfIndex = 0
            while transfIndex < len(transf):
                tr = transf.pop(transfIndex)

                # get all matched points
                # nodeInts = tr.getCheckPoints(nodeI.nodeElement)
                nodeInts = self.getCheckPoints(tr, nodeI, True)

                # find the range in which the point's match should be
                if len(nodeInts) == 0:
                    posRange = (None, None)
                elif not reverse:
                    if p.value < nodeInts[0][0]:
                        posRange = (None, nodeInts[0][1])
                    elif p.value > nodeInts[-1][0]:
                        posRange = (nodeInts[-1][1], None)
                    else:
                        rangeIndex = 1
                        while p.value > nodeInts[rangeIndex][0]:
                            rangeIndex += 1
                        posRange = (nodeInts[rangeIndex-1][1], nodeInts[rangeIndex][1])
                else:
                    if p.value < nodeInts[0][0]:
                        posRange = (nodeInts[0][1], None)
                    elif p.value > nodeInts[-1][0]:
                        posRange = (None, nodeInts[-1][1])
                    else:
                        rangeIndex = 1
                        while p.value > nodeInts[rangeIndex][0]:
                            rangeIndex += 1
                        posRange = (nodeInts[rangeIndex][1], nodeInts[rangeIndex-1][1])
                if (posRange[0] is not None and posRange[1] is not None 
                and posRange[0] > posRange[1]):
                    posRange = [posRange[1], posRange[0]]

                # match the points
                posMatch = []
                for mP in matchNodeI.linePoints:
                    if ((posRange[0] is not None and mP.value<posRange[0])
                    or (posRange[1] is not None and mP.value>posRange[1])):
                        continue
                    elif ((posRange[1] is None and mP.value==posRange[0])
                    or (posRange[0] is None and mP.value==posRange[1])):
                        if tr.containsMatch(mP.value):
                            continue
                    posMatch.append(mP.value)
                    
                # add the match
                for i in posMatch:
                    newTransf = tr.duplicate()
                    if not newTransf.add(p.value, i):
                        continue
                    transf.insert(transfIndex, newTransf)
                    transfIndex+=1
                    
    def resolvePlanes(self, transf, planeNode, matchingPlaneNode, match):
        """
        resolve plane-segments
        """
        # for each of the plane-segment in the plane node
        for plane in planeNode.forms:
            transfIndex = 0

            # go through each of the the transformation sets
            while transfIndex < len(transf):
                tr = transf.pop(transfIndex)
                
                pCheck = True
                # check each plane polygon for existing matches
                for poly in plane.boundaries[0].polygonal:
                    if not tr.contains(poly):
                        pCheck = False
                        break
                # if all plane polygons already have matches, add the transformation set and continue to the next one
                if pCheck:
                    transf.insert(transfIndex, tr)
                    transfIndex += 1      
                    continue

                # if not all plane polygons have matches
                index = 0
                while index < len(plane.boundaries[0].polygonal):
                    if tr.contains(plane.boundaries[0].polygonal[index]):
                        pCheck = True
                        break
                    index += 1
                if pCheck:
                    mPoly = tr.apply(plane.boundaries[0].polygonal[index])

                # for each of the shape plane-segments in the matching plane node, try to find a match
                for otherPlane in matchingPlaneNode.forms:
                    
                    # compare sort, attribute and number of polygons
                    if (otherPlane.sort.baseSort() != plane.sort.baseSort() or (plane.hasAttribute() 
                    and (not plane.attribute == otherPlane.attribute))
                    or len(plane.boundaries[0].polygonal) != len(otherPlane.boundaries[0].polygonal)):
                        continue
                    

                    # if no point on the plane is matched -> continue on to other planes
                    if not pCheck:
                        continue
                    posMatch = []
                    check = False
                    
                    # find the match of the lhs point at index on the rhs plane
                    mIndex = 0
                    while mIndex < len(otherPlane.boundaries[0].polygonal):
                        if otherPlane.boundaries[0].polygonal[mIndex] == mPoly:
                            check = True
                            break
                        mIndex += 1

                    # if the match of the lhs point at index is not on this plane segment, continue to other plane segment
                    if not check:
                        continue
                    
                    # check if the lhs plane segment is matched in same order with the rhs plane segment
                    i = index + 1
                    j = mIndex + 1
                    for k in range(1, len(plane.boundaries[0].polygonal)):
                        if i >= len(plane.boundaries[0].polygonal):
                            i -= len(plane.boundaries[0].polygonal)
                        if j >= len(plane.boundaries[0].polygonal):
                            j -= len(plane.boundaries[0].polygonal)
                            
                        if tr.contains(plane.boundaries[0].polygonal[i]):
                            newPoly = tr.apply(plane.boundaries[0].polygonal[i])
                            if otherPlane.boundaries[0].polygonal[j] != newPoly:
                                check = False
                                break
                        else:
                            posMatch.append([plane.boundaries[0].polygonal[i], otherPlane.boundaries[0].polygonal[j]])
                        i += 1
                        j += 1

                    # check if the lhs plane segment is matched in reverse order with the rhs plane segment
                    if not check:
                        check = True
                        i = index + 1
                        j = mIndex - 1
                        for k in range(1, len(plane.boundaries[0].polygonal)):
                            if i >= len(plane.boundaries[0].polygonal):
                                i -= len(plane.boundaries[0].polygonal)
                            if j < 0:
                                j += len(plane.boundaries[0].polygonal)
                            if tr.contains(plane.boundaries[0].polygonal[i]):
                                newPoly = tr.apply(plane.boundaries[0].polygonal[i])
                                if otherPlane.boundaries[0].polygonal[j] != newPoly:
                                    check = False
                                    break
                            else:
                                posMatch.append([plane.boundaries[0].polygonal[i], otherPlane.boundaries[0].polygonal[j]])
                            i += 1
                            j -= 1

                    # if it does not match in either same or reverse order, continue
                    if not check:
                        continue

                    # add match
                    newTr = tr.duplicate()
                    for pM in posMatch:
                        if not newTr.add(pM[0], pM[1]):
                            check = False
                            break
                    if not check:
                        continue
                    
                    
                    for i in range(len(self.nodeList.lines)):
                        nodeI = self.nodeList.lines[i]
                        # nodeInts = tr.getCheckPoints(nodeI.nodeElement)
                        nodeInts = self.getCheckPoints(tr, nodeI, False)
                        if len(nodeInts) < 2:
                            continue
                        
                        if nodeInts[0][1] > nodeInts[-1][1]:
                            reverse = True
                        else:
                            reverse = False
                            
                        polygons = []
                        for p in plane.boundaries[0].polygonal:
                            if nodeI.nodeElement.containsPoint(p):
                                polygons.append(p)
                        if len(polygons) == 0:
                            continue
                        
                        for i in polygons:
                            j = newTr.transform(i)
                            if j is False:
                                print('! ERROR: missing vector !',i)
                                newTr.printTransf()
                                raise ValueError('Missing vector in Transformation')
                            if i < nodeInts[0][0]:
                                if (reverse and j < nodeInts[0][1]) or (not reverse and j > nodeInts[0][1]):
                                    check = False
                                    break
                            elif i > nodeInts[-1][0]:
                                if (reverse and j > nodeInts[-1][1]) or (not reverse and j < nodeInts[-1][1]):
                                    check = False
                                    break
                            else:
                                index = 0
                                while index < len(nodeInts) and nodeInts[index][0] < i:
                                    index += 1
                                if not (nodeInts[index-1][1] <= j <= nodeInts[index][1]
                                or nodeInts[index][1] <= j <= nodeInts[index-1][1]):
                                    check = False
                                    break
                        if not check:
                            break
                    if not check:
                        continue
                    
                    transf.insert(transfIndex, newTr)
                    transfIndex += 1




                    # posMatch = []
                    # check = True
                    # for i in range(len(self.nodeList.lines)):
                    #     nodeI = self.nodeList.lines[i]
                    #     if not plane.containsLine(nodeI.nodeElement):
                    #         continue
    
                    #     polygons = []
                    #     for p in plane.boundaries[0].polygonal:
                    #         if nodeI.nodeElement.containsPoint(p):
                    #             polygons.append(p)
                                
                    #     if len(polygons) == 0:
                    #         continue
                    #     polygons = sorted(polygons)
                    #     checkProceed = False
                    #     for p in polygons:
                    #         if not tr.contains(p):
                    #             checkProceed = True
                    #     if not checkProceed:
                    #         continue
                    #     matchNodeI = self.matchingNodeList.lines[match[i]]
                        
                    #     nodeInts = tr.getCheckPoints(nodeI.nodeElement)
                    #     if nodeInts[0][1] > nodeInts[-1][1]:
                    #         reverse = True
                    #     else:
                    #         reverse = False
                            
                    #     mPolygons = []
                    #     for p in otherPlane.boundaries[0].polygonal:
                    #         if matchNodeI.nodeElement.containsPoint(p):
                    #             mPolygons.append(p)
                    #     if len(mPolygons) != len(polygons):
                    #         check = False
                    #         break
                    #     mPolygons = sorted(mPolygons)
                    #     if reverse:
                    #         mPolygons = list(reversed(mPolygons))
                    #     for pI in range(len(polygons)):
                    #         mP = tr.transform(polygons[pI])
                    #         if mP is not False:
                    #             if mPolygons[pI] == mP:
                    #                 continue

                    #             check = False
                    #             break
                    #         if polygons[pI] < nodeInts[0][0]:
                    #             if ((not reverse and mPolygons[pI] > nodeInts[0][1])
                    #             or (reverse and mPolygons[pI] < nodeInts[0][1])):
                    #                 check = False
                    #                 break
                    #         elif polygons[pI] > nodeInts[-1][0]:
                    #             if ((not reverse and mPolygons[pI] < nodeInts[-1][1])
                    #             or (reverse and mPolygons[pI] > nodeInts[-1][1])):
                    #                 check = False
                    #                 break
                    #         else:
                    #             intI = 0
                    #             while polygons[pI] > nodeInts[intI][0]:
                    #                 intI += 1
                    #             if not ((not reverse and nodeInts[intI-1][1] <= mPolygons[pI] <= nodeInts[intI][1])
                    #             or (reverse and nodeInts[intI][1] <= mPolygons[pI] <= nodeInts[intI-1][1])):
                    #                 check = False
                    #                 break
                    #         posMatch.append([polygons[pI], mPolygons[pI]])
                    # if not check:
                    #     continue
                    # newTr = tr.duplicate()
                    # for pM in posMatch:
                    #     newTr.add(pM[0], pM[1])
                    # transf.insert(transfIndex, newTr)
                    # transfIndex += 1      
                    
    def resolveCurves(self, transf, planeNode, matchingPlaneNode):
        """
        resolve curves (bezier curves, ellipses, circles only)
        """
        if len(planeNode.curves) == 0:
            return True
        if len(planeNode.curves) > len(matchingPlaneNode.curves):
            for i in range(len(transf)):
                transf.pop(0)
        if len(transf) == 0:
            return False
        tr = transf[0]
        circleList = []
        ellipseList = []
        bezierList = []
        insertSpotCircle = 0
        insertSpotEllipse = 0
        for i in planeNode.curves:
            containedVects = []
            if i.value.type == 'ind-bezierCurveP3D':
                bezierList.append(i)
                continue
            for j in tr.vects:
                if i.value.containsPoint(j[0]):
                    containedVects.append(j[0])
            if len(containedVects) >= 2:
                i.oType = 't'
                if i.value.type == 'ind-ellipseP3D':
                    ellipseList.insert(0,i)
                if i.value.type == 'ind-circleP3D' and not tr.contains(i.value.center):
                    circleList.insert(0,i)
            elif len(containedVects) == 1:
                i.oType = 'o'
                if i.value.type == 'ind-ellipseP3D':
                    ####
                    ellipseList.append(len(ellipseList)-insertSpotEllipse,i)
                if i.value.type == 'ind-circleP3D':
                    circleList.insert(len(circleList)-insertSpotCircle,i)
            else:
                if i.value.type == 'ind-ellipseP3D':
                    ellipseList.append(i)
                    insertSpotEllipse += 1
                if i.value.type == 'ind-circleP3D':
                    circleList.append(i)
                    insertSpotCircle += 1
            i.assocObj = containedVects
        otherCircleList = []
        otherEllipseList = []
        otherBezierList = []
        for i in matchingPlaneNode.curves:
            if i.value.type == 'ind-bezierCurveP3D':
                otherBezierList.append(i)
            elif i.value.type == 'ind-circleP3D':
                otherCircleList.append(i)
            elif i.value.type == 'ind-ellipseP3D':
                otherEllipseList.append(i)
                
                
        transfIndex = 0
        while transfIndex < len(transf):
            tr = transf.pop(transfIndex)
            for i in itertools.permutations(otherBezierList, len(bezierList)):
                newTr = tr.duplicate()
                check = True
                for j in range(len(bezierList)):
                    op1 = bezierList[j].misc[0]
                    op2 = bezierList[j].misc[1]
                    op3 = bezierList[j].misc[2]
                    mp1 = i[j].misc[0]
                    mp2 = i[j].misc[1]
                    mp3 = i[j].misc[2]
                    
                    if newTr.contains(op2):
                        if newTr.apply(op2) != mp2:
                            check = False
                            break
                        mp2 = None
                    if newTr.contains(op1):
                        p1 = newTr.apply(op1)
                        if p1 == mp1:
                            mp1 = None
                        elif p1 == mp3:
                            mp3 = mp1
                            mp1 = None
                        else:
                            check = False
                            break
                    if newTr.contains(op3):
                        p3 = newTr.apply(op3)
                        if p3 == mp3:
                            mp3 = None
                        elif mp1 is not None and p3 == mp1:
                            mp1 = mp3
                            mp3 = None
                        else:
                            check = False
                            break
                    if mp1 is not None:
                        newTr.add(op1, mp1)
                    if mp2 is not None:
                        newTr.add(op2, mp2)
                    if mp3 is not None:
                        newTr.add(op3, mp3)
                if not check:
                    continue
                transf.insert(transfIndex, newTr)
                transfIndex += 1
            
            
        transfIndex = 0
        while transfIndex < len(transf):
            tr = transf.pop(transfIndex)
            for i in itertools.permutations(otherCircleList, len(circleList)):
                newTr = tr.duplicate()
                check = True
                for j in range(len(circleList)):
                    nodeCent = circleList[j].misc[0]
                    nodeStart = circleList[j].misc[1]
                    nodeEnd = circleList[j].misc[2]
                    matchCent = i[j].misc[0]
                    matchStart = i[j].misc[1]
                    matchEnd = i[j].misc[2]
                    if circleList[j].oType == 't':
                        for v0 in circleList[j].assocObj:
                            v1 = tr.apply(v0)
                            if not i[j].value.containsPoint(v1):
                                check = False
                                break
                        if not check:
                            break
                        newTr.add(nodeCent, matchCent, override = True)
                    elif circleList[j].oType == 'o':
                        v1 = tr.apply(circleList[j].assocObj[0])
                        if (not i[j].value.containsPoint(v1) 
                        or not newTr.add(nodeCent, matchCent)):
                            check = False
                            break
                        if nodeStart != circleList[j].assocObj[0] and nodeEnd != circleList[j].assocObj[0]:
                            newTr.add(nodeStart, matchStart)
                            newTr.add(nodeEnd, matchEnd)
                        else:
                            nodeStart = nodeEnd if nodeStart == circleList[j].assocObj[0] else nodeStart
                            matchStart = matchEnd if matchStart == v1 else matchStart
                            newTr.add(nodeStart, matchStart)
                    else:
                        if (not newTr.add(nodeCent, matchCent)
                        or not newTr.add(nodeStart, matchStart) 
                        or not newTr.add(nodeEnd, matchEnd)):
                            check = False
                            break
                if not check:
                    continue
                transf.insert(transfIndex, newTr)
                transfIndex += 1

        transfIndex = 0
        while transfIndex < len(transf):
            tr = transf.pop(transfIndex)
            for i in itertools.permutations(otherEllipseList, len(ellipseList)):
                newTr = tr.duplicate()
                check = True
                for j in range(len(ellipseList)):
                    nodeF1 = ellipseList[j].misc[0]
                    nodeF2 = ellipseList[j].misc[1]
                    nodeStart = ellipseList[j].misc[2]
                    nodeEnd = ellipseList[j].misc[3]
                    matchF1 = i[j].misc[0]
                    matchF2 = i[j].misc[1]
                    matchStart = i[j].misc[2]
                    matchEnd = i[j].misc[3]
                    if ellipseList[j].oType == 't':
                        for v0 in ellipseList[j].assocObj:
                            v1 = tr.apply(v0)
                            if not i[j].value.containsPoint(v1):

                                check = False
                                break
                        if not check:
                            break
                        newTr.add(nodeF1, matchF1, override = True)
                        newTr.add(nodeF2, matchF2, override = True)
                    elif ellipseList[j].oType == 'o':
                        v1 = tr.apply(ellipseList[j].assocObj[0])
                        if (not i[j].value.containsPoint(v1) 
                        or not newTr.add(nodeF1, matchF1)
                        or not newTr.add(nodeF2, matchF2)):
                            check = False
                            break
                        if nodeStart != ellipseList[j].assocObj[0] and nodeEnd != ellipseList[j].assocObj[0]:
                            newTr.add(nodeStart, matchStart)
                            newTr.add(nodeEnd, matchEnd)
                        else:
                            nodeStart = nodeEnd if nodeStart == ellipseList[j].assocObj[0] else nodeStart
                            matchStart = matchEnd if matchStart == v1 else matchStart
                            newTr.add(nodeStart, matchStart)
                    else:
                        if (not newTr.add(nodeF1, matchF1)
                        or not newTr.add(nodeF2, matchF2) 
                        or not newTr.add(nodeStart, matchStart) 
                        or not newTr.add(nodeEnd, matchEnd)):
                            check = False
                            break
                if not check:
                    continue
                transf.insert(transfIndex, newTr)
                transfIndex += 1
        return True

    def resolveColinearEndPoints(self, transf):
        for colinearSet in self.colinearInts:
            pti, ptj, ptk, li, lj, lk = colinearSet
            for transfIndex in range(len(transf)):
                tr = transf.pop(0)
                trSet = transformSet()
                trSet.add(tr)
                trPti = tr.apply(pti.value)
                trPtj = tr.apply(ptj.value)
                trPtk = tr.apply(ptk.value)
                if trPti.colinear(trPtj, trPtk):
                    transf.append(tr)
                    continue
                if pti.endP and trPtj != trPtk:
                    self.adjustColinearEndPoints(tr, trSet, transf, trPtj, trPtk, pti.value, li)
                if ptj.endP and trPti != trPtk:
                    self.adjustColinearEndPoints(tr, trSet, transf, trPti, trPtk, ptj.value, lj)
                if ptk.endP and trPti != trPtj:
                    self.adjustColinearEndPoints(tr, trSet, transf, trPti, trPtj, ptk.value, lk)

    def adjustColinearEndPoints(self, tr, trSet, transf, anchor1, anchor2, movingPoint, movingNode):
        newLine = line3D(sortClass.getSort('__dummyLine3D__'), anchor1, anchor2)
        trLine = movingNode.nodeElement.edit(trSet)
        if newLine.intersect(trLine):
            newMovingPoint = newLine.intersection(trLine)
            check = True
            for lnSeg in movingNode.forms:
                if lnSeg.start == movingPoint or lnSeg.end == movingPoint:
                    trLnSeg = lnSeg.edit(trSet)
                    if not trLnSeg.containsPoint(newMovingPoint):
                        check = False
                        break
            if check:
                newTr = tr.duplicate()
                newTr.add(movingPoint, newMovingPoint, True)
                transf.append(newTr)

    def getCheckPoints(self, transf, lineNode, filtering = False):
        checkPoints = transf.getCheckPoints(lineNode.nodeElement)
        if filtering:
            i = 0
            while i < len(checkPoints):
                if len(checkPoints) <= 2:
                    break
                pt = checkPoints[i][0]
                isMin = True
                isMax = True
                for line in lineNode.forms:
                    if pt > line.start: isMin = False
                    if pt < line.end:   isMax = False
                    if not (isMin or isMax): break
                if not (isMin or isMax):
                    i += 1
                    continue
                for point in lineNode.containedPoints:
                    if pt > point.value: isMin = False
                    if pt < point.value: isMax = False
                    if not (isMin or isMax): break
                if isMin or isMax:
                    checkPoints.pop(i)
                    continue
                i += 1
        return checkPoints


    @staticmethod
    def getIntersection(line1, line2):
        """
        Get the intersection between two line-nodes
        """
        for i in line1.intersections:
            for j in line2.intersections:
                if i == j:
                # if i == j and (len(line1.forms) == 0 or len(line2.forms) == 0 or i.iType == 'i' or j.iType == 'i'):
                    return i
        return None
        
    def checkIntersectionPoints(self, transf):
        """
        Ensure all the existing points that lie on lines to have correct match 
        """
        ind = 0
        while ind < len(transf):
            tr = transf[ind]
            for p in self.nodeList.points:
                # ignore point objects that are not intersections
                if p.oType != 'i':
                    continue
                # transform the point objects
                transformedPoint = tr.allTransform(p.value)
                check = False

                # there must be a matching point with the same position as the transformed point
                for mP in self.matchingNodeList.points:
                    if mP.oType != 'i':
                        continue
                    if mP.value == transformedPoint:
                        check = True
                        break
                if not check:
                    transf.pop(ind)
                    ind -= 1
                    break
            ind += 1
        if len(transf) == 0:
            return False
        return True
        
    @staticmethod
    def normalizeLineEndPoints(nodeI, tr, reverse):
        for i in range(len(nodeI.forms)):
            li = nodeI.forms[i]
            trliEnd = tr.apply(li.end)
            for j in range(len(nodeI.forms)):
                if i == j:
                    continue
                lj = nodeI.forms[j]
                if li.attribute != lj.attribute or not li.end < lj.start:
                    continue
                trljStart = tr.apply(lj.start)
                if trliEnd > trljStart and not reverse:
                    nVect = trliEnd.add(trljStart).scale(c0p5)
                    tr.add(li.end, nVect, override = True)
                    tr.add(lj.start, nVect, override = True)
                if trliEnd < trljStart and reverse:
                    nVect = trliEnd.add(trljStart).scale(c0p5)
                    tr.add(li.end, nVect, override = True)
                    tr.add(lj.start, nVect, override = True)

    def orderingCheck(self, transf, nodeI, outsideCheck = False):
        """
        Check the ordering of the intersections on a specific line

        For normal case (line has 2 or more intersections on it):

        * intersections contained by any line-segment on the line: must retain their ordering among one another after transformation
        * intersections not contained by any line-segment: 2 cases:
            * lhs intersections that are between the contained intersections: no restriction
            * lhs intersections that are outside of contained intersections (smaller/bigger than the smallest/biggest contained intersections): their matches cannot be between the contained intersections (between the smallest/biggest contained intersections)
        
        For special cases (line has no intersection / line has only one intersection):

        * all intersections must retain their ordering
        """
        ind = 0
        containedInts = []
        externalInts = []
        # add intersections into containedInts and externalInts
        for ints in nodeI.intersections:
            if len(ints.assocObjs) < 2:
                continue
            if (not outsideCheck) and ints.iType == 'o':
                externalInts.append(ints.value)
            else:
                containedInts.append(ints.value)

        # if there are less than 2 contained intersections -> no ordering to be checked
        if len(containedInts) < 2:
            return True

        # sort the contained intersections
        containedInts = sorted(containedInts)

        # remove external intersections that are between the smallest and the biggest contained intersections
        exInd = 0
        while exInd < len(externalInts):
            if externalInts[exInd] > containedInts[0] and externalInts[exInd] < containedInts[-1]:
                externalInts.pop(exInd)
                continue
            exInd += 1
    
        # for each of the transformation set
        while ind < len(transf):
            tr = transf[ind]
            nodeInts = []
            # for each of the contained intersections, find the matching and add it to nodeInts
            for i in containedInts:
                if not tr.contains(i):
                    continue
                nodeInts.append([i,tr.apply(i)])

            # if there are fewer than 2 transformable intersections, there's no ordering to be checked 
            # -> continue to the next transformation set
            if len(nodeInts) < 2:
                ind += 1
                continue
            
            # check the first two contained intersections if the transformation reverse the intersection order or not
            reverse = False
            index = 1
            while index < len(nodeInts) and nodeInts[index][1] == nodeInts[index-1][1]:
                index += 1
            if index == len(nodeInts):
                ind += 1
                continue
            if nodeInts[index-1][1] < nodeInts[index][1]:
                reverse = False
            else:
                reverse = True
            
            # check the remaining intersection in pairs of 2 for the same ordering, if the ordering is violated,
            # remove the transformation
            check = True
            for index in range(1,len(nodeInts)):
                if ((not reverse and nodeInts[index-1][1] > nodeInts[index][1])
                or (reverse and nodeInts[index-1][1] < nodeInts[index][1])):
                    transf.pop(ind)
                    check = False
                    break
            if not check:
                continue
            
            # assign smallest, largest as smallest, largest contained intersections on the line
            smallest = nodeInts[0][1]
            largest = nodeInts[-1][1]
            if reverse:
                smallest, largest = largest, smallest

            # if any of the external intersection's match lies between the smallest and largest intersections 
            # -> discard the transformation
            for exInt in externalInts:
                exMatch = tr.getExtraMatch(exInt)
                if exMatch is None:
                    continue
                if exMatch > smallest and exMatch < largest:
                    transf.pop(ind)
                    check = False
                    break

            if check:
                ind += 1
        if len(transf) == 0 :
            return False
        return True
        
    def addOrientCheck(self, orientChecks, item):
        """
        add a point transformation for orientation checking
        """
        for i in range(len(orientChecks)):
            if orientChecks[i] == item:
                return
            elif orientChecks[i][0] > item[0]:
                orientChecks.insert(i, item)
                return
        orientChecks.append(item)
    

    def orientationChecking(self, orientChecks):
        """
        check the orientation of the points in orientChecks
        """
        reverse = None
        cwMatch = None
        if len(orientChecks) > 2:
            o = orientChecks[0]
            for index in range(2, len(orientChecks)):
                x = orientChecks[index-1]
                y = orientChecks[index]
                if o[0].colinear(x[0],y[0]):
                    if reverse is None:
                        a = o[0] < x[0]
                        b = o[1] < x[1]
                        reverse = (a!=b)
                    a = x[0] < y[0]
                    b = x[1] < y[1]
                    if reverse == (a==b):
                        return False
                    
                    # if (a and b) or not (a or b):
                    #     if reverse:
                    #         return False
                    # elif not reverse:
                    #     return False
                    
                else:
                    pl0 = plane3D(sortClass.getSort('__dummyPlane3D__'), [o[0],x[0],y[0]])
                    pl1 = plane3D(sortClass.getSort('__dummyPlane3D__'), [o[1],x[1],y[1]])
                    if cwMatch is None:
                        a = graphMap3D.cw(pl0,o[0],x[0],y[0])
                        b = graphMap3D.cw(pl1,o[1],x[1],y[1])
                        cwMatch = (a == b)
                    else:
                        a = graphMap3D.cw(pl0,o[0],x[0],y[0])
                        b = graphMap3D.cw(pl1,o[1],x[1],y[1])
                        if cwMatch != (a == b):
                            return False
        return True




    # ============================================================================================
    #                            PREDICATE SOLVING PROCESS
    # ============================================================================================

    def evaluatePredicate(self, transfSet, lhs, tlhs, form, param):   
        """
        resolve all predicates in the rule
        """
        if isinstance(param, dict):
            k = param
            param = paramSet()
            param.keys = k
        param.lhs = tlhs
        param.transform = transfSet
        graphTransf = self.getTransform(transfSet)
        self.postCheck = True


        for cond in self.predicates:
            # maxline predicate
            if cond[0] == 'maxline':
                for keyI in cond[1]:
                    ind = keyI.ind
                    transfInd = ind.transform(transfSet)
                    form.accept(self, direction = [cond[0], transfInd])
            # bound predicate
            elif cond[0] == 'bound':
                for keyI in cond[1]:
                    ind = keyI.ind.transform(transfSet)
                    transfStart = graphTransf.apply(keyI.ind.start)
                    transfEnd = graphTransf.apply(keyI.ind.end)
                    form.accept(self, direction = [cond[0], ind, keyI.properties, transfStart, transfEnd])
            # void predicate
            elif cond[0] == 'void':
                for bd in cond[1]:
                    # if there is a list of sort
                    if isinstance(bd[0], list):
                        sorts = bd[0]
                        bdr = bd[1:]

                        # create a plane boundary based on the vectors
                        bound = boundary()
                        if bdr[-1] == bdr[0]:
                            bdr.pop(-1)
                        for keyI in bdr:
                            newInd = graphTransf.transform(keyI)
                            if newInd is False:
                                graphTransf.printTransf()
                                raise ValueError('matching for vector '+str(keyI)+' is not found')
                            bound.add(newInd)
                        
                        # remove duplicate vertices, add plane border segment
                        i = 0
                        while i < len(bound.polygonal):
                            if bound.polygonal[i] == bound.polygonal[i-1]:
                                bound.polygonal.pop(i)
                            else:
                                i += 1
                        for i in range(len(bound.polygonal)):
                            bound.add(bSegment(bound.polygonal[i],bound.polygonal[i-1]))
                        
                        # create the plane-segment
                        planeSeg = planeSegment3D(lhs.sort, [bound], True)

                        # find all lines that are co-planar with the plane
                        lines = []
                        form.accept(self, direction = ['voidof', planeSeg, lines, sorts])
                        if not self.postCheck:
                            return False
                        
                        # check if any of the lines cut through the created plane-segment
                        check = planeSeg.classify_line(lines)
                        if not check.notInside():
                            self.postCheck = False
                    # if there's no sort list
                    else:
                        # create plane-segment boundary
                        bound = boundary()
                        if bd[-1] == bd[0]:
                            bd.pop(-1)
                        for keyI in bd:
                            newInd = graphTransf.transform(keyI)
                            if newInd is False:
                                graphTransf.printTransf()
                                raise ValueError('matching for vector '+str(keyI)+' is not found')
                            bound.add(newInd)
                        for i in range(len(bound.polygonal)):
                            bound.add(bSegment(bound.polygonal[i],bound.polygonal[i-1]))
                        # create plane-segment
                        planeSeg = planeSegment3D(lhs.sort, [bound], True)

                        # find all lines that are co-planar with the plane
                        lines = []
                        form.accept(self, direction = [cond[0], planeSeg, lines])
                        if not self.postCheck:
                            return False

                        # check if any of the lines cut through the created plane-segment
                        check = planeSeg.classify_line(lines)
                        if not check.notInside():
                            self.postCheck = False
            # inside predicate
            elif cond[0] == 'inside':
                for bd in cond[1]:
                    # if there is a list of sort
                    if isinstance(bd[0], list):
                        sorts = bd[0]
                        bdr = bd[1:]

                        # create a plane boundary based on the vectors
                        bound = boundary()
                        if bdr[-1] == bdr[0]:
                            bdr.pop(-1)
                        for keyI in bdr:
                            bound.add(keyI)
                        
                        # remove duplicate vertices, add plane border segment
                        i = 0
                        while i < len(bound.polygonal):
                            if bound.polygonal[i] == bound.polygonal[i-1]:
                                bound.polygonal.pop(i)
                            else:
                                i += 1
                        for i in range(len(bound.polygonal)):
                            bound.add(bSegment(bound.polygonal[i],bound.polygonal[i-1]))
                        
                        # create the plane-segment
                        planeSeg = planeSegment3D(lhs.sort, [bound], True)

                        # find all lines that are co-planar with the plane
                        lines = []
                        tlhs.accept(self, direction = ['insideof', planeSeg, lines, sorts])
                        if not self.postCheck:
                            return False
                        
                        # check if any of the lines cut through the created plane-segment
                        check = planeSeg.classify_line(lines)
                        if not check.isContained():
                            self.postCheck = False
                    # if there's no sort list
                    else:
                        # create plane-segment boundary
                        bound = boundary()
                        if bd[-1] == bd[0]:
                            bd.pop(-1)
                        for keyI in bd:
                            bound.add(keyI)
                        for i in range(len(bound.polygonal)):
                            bound.add(bSegment(bound.polygonal[i],bound.polygonal[i-1]))
                        # create plane-segment
                        planeSeg = planeSegment3D(lhs.sort, [bound], True)
                        # find all lines that are co-planar with the plane
                        lines = []
                        
                        tlhs.accept(self, direction = [cond[0], planeSeg, lines])
                        if not self.postCheck:
                            return False

                        # check if any of the lines cut through the created plane-segment
                        check = planeSeg.classify_line(lines)
                        if not check.isContained():
                            self.postCheck = False
            # outside predicate
            elif cond[0] == 'outside':
                for bd in cond[1]:
                    # if there is a list of sort
                    if isinstance(bd[0], list):
                        sorts = bd[0]
                        bdr = bd[1:]

                        # create a plane boundary based on the vectors
                        bound = boundary()
                        if bdr[-1] == bdr[0]:
                            bdr.pop(-1)
                        for keyI in bdr:
                            bound.add(keyI)

                        # remove duplicate vertices, add plane border segment
                        i = 0
                        while i < len(bound.polygonal):
                            if bound.polygonal[i] == bound.polygonal[i-1]:
                                bound.polygonal.pop(i)
                            else:
                                i += 1
                        for i in range(len(bound.polygonal)):
                            bound.add(bSegment(bound.polygonal[i],bound.polygonal[i-1]))
                        
                        # create the plane-segment
                        planeSeg = planeSegment3D(lhs.sort, [bound], True)

                        # find all lines that are co-planar with the plane
                        lines = []
                        tlhs.accept(self, direction = ['outsideof', planeSeg, lines, sorts])
                        if not self.postCheck:
                            return False
                        
                        # check if any of the lines cut through the created plane-segment
                        check = planeSeg.classify_line(lines)
                        if not check.notInside():
                            self.postCheck = False
                    # if there's no sort list
                    else:
                        # create plane-segment boundary
                        bound = boundary()
                        if bd[-1] == bd[0]:
                            bd.pop(-1)
                        for keyI in bd:
                            bound.add(keyI)
                        for i in range(len(bound.polygonal)):
                            bound.add(bSegment(bound.polygonal[i],bound.polygonal[i-1]))
                        # create plane-segment
                        planeSeg = planeSegment3D(lhs.sort, [bound], True)

                        # find all lines that are co-planar with the plane
                        lines = []
                        tlhs.accept(self, direction = [cond[0], planeSeg, lines])
                        if not self.postCheck:
                            return False
                        
                        # check if any of the lines cut through the created plane-segment
                        check = planeSeg.classify_line(lines)
                        if not check.notInside():
                            self.postCheck = False
            # embeds predicate
            elif cond[0] == 'embeds':
                for keyI in cond[1]:
                    bigObj = keyI[0].ind.transform(transfSet)
                    for i in range(1, len(keyI)):
                        smallObj = keyI[i].ind.transform(transfSet)
                        if 'ind-planeSegment' in bigObj.type:
                            if 'ind-lineSegment' in smallObj.type:
                                check = bigObj.classify_line([smallObj])
                                if check.notInside():
                                    self.postCheck = False
                            elif 'ind-point' in smallObj.type:
                                check = bigObj.containsPoint(smallObj.value)
                                if not check:
                                    self.postCheck = False
                        elif 'ind-lineSegment' in bigObj.type and 'ind-point' in smallObj.type:
                            if not bigObj.containsPoint(smallObj.value):
                                self.postCheck = False
                        else:
                            self.postCheck = False
            # shortes_line/longest_line predicate
            elif cond[0] == 'shortest_line' or cond[0] == 'longest_line':
                for keyI in cond[1]:
                    if isinstance(keyI, list):
                        sorts = keyI[0]
                        if isinstance(sorts, str):
                            sorts = [sorts]
                        for i in keyI[1:]:
                            nInd = i.ind.transform(transfSet)
                            tlhs.accept(self, direction = [cond[0], nInd.squareLength(), sorts])
                    else:
                        ind = keyI.ind.transform(transfSet)
                        tlhs.accept(self, direction = [cond[0], ind.squareLength(), [ind.sort.baseSort().name]])
            # nolabel predicate
            elif cond[0] == 'nolabel':
                for keyI in cond[1]:
                    ind = keyI.ind.transform(transfSet)
                    form.accept(self, direction = [cond[0], ind])
            # *description* predicate
            elif cond[0] == 'description':
                for keyI in cond[1]:
                    param.transform = transfSet
                    val = description(sortClass.getSort('__dummyDesc__'), keyI[0])
                    val = val.evaluateInd(param)
                    # '{}' -> equal to one of the values in the curly bracket
                    if keyI[1] == '{':
                        check = False
                        for comp in keyI[-1]:
                            
                            comparand = description(sortClass.getSort('__dummyDesc__'), comp)
                            comparand = comparand.evaluateInd(param)
                            if comparand == val:
                                check = True
                                break
                    # '[]' -> within the range of the square bracket
                    elif keyI[1] == '[':
                        if len(keyI[-1]) != 2:
                            raise ValueError('[ ] condition only accepts 2 arguments as min and max. e.g. [1,3] --> value between 1 and 3')
                        comp1 = description(sortClass.getSort('__dummyDesc__'), keyI[-1][0])
                        comp1 = comp1.evaluateInd(param)
                        comp2 = description(sortClass.getSort('__dummyDesc__'), keyI[-1][-1])
                        comp2 = comp2.evaluateInd(param)
                        if comp1 < val < comp2:
                            check = True
                        else:
                            check = False
                    # normal comparison
                    else:
                        comparand = description(sortClass.getSort('__dummyDesc__'), keyI[-1])
                        comparand = comparand.evaluateInd(param)
                        if keyI[1] == '=':
                            check = val == comparand
                        if keyI[1] == '<':
                            check = val < comparand
                        if keyI[1] == '>':
                            check = val > comparand
                        if keyI[1] == '<=':
                            check = val <= comparand
                        if keyI[1] == '>=':
                            check = val >= comparand
                        if keyI[1] == '<>':
                            check = not (val == comparand)
                    if not check:
                        self.postCheck = False
            if self.postCheck is False:
                return False
        return True

    # ============================================================================================
    #                               RHS/DIRECTIVES SOLVING PROCESS
    # ============================================================================================

    def solveDirectives(self, transfSet, lhs, tlhs, rhs, form, param):
        """
        solve the new lines/planes/points along with the directives in the rhs
        """
        allTransforms = [transfSet]
        if isinstance(param, dict):
            k = param
            param = paramSet()
            param.keys = k
        param.lhs = tlhs
        param.rhs = rhs
        # print('\n\n============================================\nResolving Order:')
        # for node in self.rhsSolvingOrder:
        #     print(node.nodeElement.printRhino())
        #     print(node.rhsInfo['case'])

        for node in self.rhsSolvingOrder:
            # print('\n____----____\n' + str(node.rhsInfo['case']))
            # print(node.nodeElement.printRhino())
            numTransfSet = len(allTransforms)
            if numTransfSet == 0:
                return allTransforms
            for i in range(numTransfSet):
                transf = allTransforms.pop(0)
                if node.nodeElement.type == 'ind-line3D':
                    newTransfs = self.solveRHSLine(node, transf, param)
                elif node.nodeElement.type == 'ind-plane3D':
                    newTransfs = self.solveRHSPlane(node, transf, param)
                elif node.nodeElement.type == 'ind-pointP3D':
                    newTransfs = self.solveRHSPoint(node, transf, param)
                else:
                    continue
                for new in newTransfs:
                    allTransforms.append(new)
        i = 0
        while i < len(allTransforms):
            tr = self.getTransform(allTransforms[i])
            # if there are one or more unresolved vectors --> check if we can resolve them by [0,1] point_on_line directive on an existing line
            if len(tr.unresolved) > 0:
                transfResolvable = True
                while len(tr.unresolved) > 0:
                    unresolvedVect = tr.unresolved[0]
                    vectResolvable = False
                    for lhsNode in self.nodeList.lines:
                        if not lhsNode.nodeElement.containsPoint(unresolvedVect):
                            continue
                        for lineSeg in lhsNode.forms:
                            if not lineSeg.containsPoint(unresolvedVect):
                                continue
                            if not (tr.contains(lineSeg.start) and tr.contains(lineSeg.end)):
                                continue
                            self.solvePointOnLine(tr, lineSeg, [coordinate(randint(0, 100000)/100000)], unresolvedVect)
                            vectResolvable = True
                            break
                        break
                    for rhsNode in self.nodeList.rhsLines:
                        if not rhsNode.nodeElement.containsPoint(unresolvedVect):
                            continue
                        for lineSeg in rhsNode.forms:
                            if not lineSeg.containsPoint(unresolvedVect):
                                continue
                            if not (tr.contains(lineSeg.start) and tr.contains(lineSeg.end)):
                                continue

                            self.solvePointOnLine(tr, lineSeg, [coordinate(randint(0, 100000)/100000)], unresolvedVect)
                            vectResolvable = True
                            break
                        break
                    if not vectResolvable:
                        transfResolvable = False
                        break
                if transfResolvable:
                    i += 1
                else:
                    print('\nOne or More Unresolved Vects Left.....')
                    for j in tr.vects:
                        print(j[0])
                    print(tr.unresolved)
                    allTransforms.pop(i)
            else: i += 1
        return allTransforms

    def solveRHSLine(self, node, transf, param):
        """
        solve a specific line node with a specific transformation set and update it
        """
        tr = self.getTransform(transf)
        param.transform = transf
        resultLines = []
        parallelProjection = None
        reverseCheck = None

        ##################### SOLVE LINE DIRECTION + ROOT #####################
        # TODO: TO BE TESTED
        # CASE_1PL_1PARALLEL_1INT
        # CASE_1PL_1PERP_1INT
        # CASE_1PL_1PERP_1PT (DONE)
        # CASE_1PARALLEL_1PT (DONE)
        # CASE_2PARALLEL (DONE)
        # CASE_ANGLE
        # CASE_POINT_LINE_DISTANCE (DONE)
        # CASE_LINE_LINE_DISTANCE (DONE)

        ###### NORMAL CASES ######
        ## solvable plane case ##
        if node.rhsInfo['case'] == graphNode3D.lineNode.line_casus.CASE_1PL_1PARALLEL_1INT:
            ### case: solvable plane, 1 parallel line, 1 intersecting line
            pl = node.rhsInfo['caseNodes'][0]
            ln = node.rhsInfo['caseNodes'][1]
            intLn = node.rhsInfo['caseNodes'][2]
            transformedPl = pl.nodeElement.edit(transf)
            transformedLn = ln.nodeElement.edit(transf)
            transformedIntLn = intLn.nodeElement.edit(transf)

            if transformedPl.parallelLine(transformedIntLn):
                return []


            # dir = parallelLineDir
            resultDir = transformedLn.direction

            # linePt = intersection of plane and intersecting line
            resultLinePoint = transformedPl.intersectionLine(transformedIntLn)

            resultLines.append(line3D(node.nodeElement.sort, resultLinePoint, resultLinePoint.add(resultDir)))

        elif node.rhsInfo['case'] == graphNode3D.lineNode.line_casus.CASE_1PL_1PERP_1INT:
            ### case: solvable plane, 1 perpendicular line, 1 intersecting line (co-planar perpendicular line)
            pl = node.rhsInfo['caseNodes'][0]
            ln = node.rhsInfo['caseNodes'][1]
            intLn = node.rhsInfo['caseNodes'][2]
            transformedPl = pl.nodeElement.edit(transf)
            transformedLn = ln.nodeElement.edit(transf)
            transformedIntLn = intLn.nodeElement.edit(transf)

            if transformedLn.direction.parallel(transformedPl.normal) or transformedPl.parallelLine(transformedIntLn):
                return []

            # dir = perpLineDir x planeNormal
            resultDir = transformedLn.direction.product(transformedPl.normal)

            # linePt = intersection of plane and intersecting line
            resultLinePoint = transformedPl.intersectionLine(transformedIntLn)

            resultLines.append(line3D(node.nodeElement.sort, resultLinePoint, resultLinePoint.add(resultDir)))

        elif node.rhsInfo['case'] == graphNode3D.lineNode.line_casus.CASE_1PL_1PERP_1PT:
            ### case: solvable plane, 1 perpendicular line, 1 contained point (co-planar perpendicular line)
            pl = node.rhsInfo['caseNodes'][0]
            ln = node.rhsInfo['caseNodes'][1]
            pt = node.rhsInfo['caseNodes'][2]
            transformedPl = pl.nodeElement.edit(transf)
            transformedLn = ln.nodeElement.edit(transf)

            if transformedLn.direction.parallel(transformedPl.normal):
                return []

            # dir = perpLineDir x planeNormal
            resultDir = transformedLn.direction.product(transformedPl.normal)

            # linePt = contained Pt
            resultLinePoint = tr.transform(pt)

            resultLines.append(line3D(node.nodeElement.sort, resultLinePoint, resultLinePoint.add(resultDir)))

        elif node.rhsInfo['case'] == graphNode3D.lineNode.line_casus.CASE_1PL_1PERP:
            ### case: solvable plane, 1 perpendicular line (not perpendicular to the plane, not co-planar)
            pl = node.rhsInfo['caseNodes'][0]
            ln = node.rhsInfo['caseNodes'][1]
            transformedPl = pl.nodeElement.edit(transf)
            transformedLn = ln.nodeElement.edit(transf)

            if transformedLn.direction.parallel(transformedPl.normal) or transformedPl.containsLine(transformedLn):
                return []

            # dir = perpLineDir x planeNormal
            resultDir = transformedLn.direction.product(transformedPl.normal)

            # linePt = intersection of plane and perpendicular line
            resultLinePoint = transformedPl.intersectionLine(transformedLn)

            resultLines.append(line3D(node.nodeElement.sort, resultLinePoint, resultLinePoint.add(resultDir)))

        ## no plane case ##
        elif node.rhsInfo['case'] == graphNode3D.lineNode.line_casus.CASE_1PARALLEL_1PT:
            ### case: solvable plane, 1 parallel line, 1 contained point
            ln = node.rhsInfo['caseNodes'][0]
            pt = node.rhsInfo['caseNodes'][1]
            transformedLn = ln.nodeElement.edit(transf)

            # dir = perpLineDir x planeNormal
            resultDir = transformedLn.direction
            resultLinePoint = tr.transform(pt)

            if not resultLinePoint:
                return []

            resultLines.append(line3D(node.nodeElement.sort, resultLinePoint, resultLinePoint.add(resultDir)))

        elif node.rhsInfo['case'] == graphNode3D.lineNode.line_casus.CASE_2PERP:
            ### case: 2 perpendicular lines not parallel to each other

            case0 = node.rhsInfo['caseNodes'][0]
            case1 = node.rhsInfo['caseNodes'][1]
            transformedCase0 = case0.nodeElement.edit(transf)
            transformedCase1 = case1.nodeElement.edit(transf)
            
            resultDir = transformedCase0.direction.product(transformedCase1.direction)
            resultLinePoint = None
            if case0.nodeElement.intersect(case1.nodeElement):
                if transformedCase0.intersect(transformedCase1):
                    resultLinePoint = transformedCase0.intersection(transformedCase1)
                else:
                    return []
            else:
                n = case0.nodeElement.direction.product(case1.nodeElement.direction)
                p1 = case0.nodeElement.root
                p2 = case1.nodeElement.root
                d1 = case0.nodeElement.direction
                d2 = case1.nodeElement.direction
                n1 = d1.product(n)
                resultLinePoint = p2.add(d2.scale((p1.subtract(p2).dotproduct(n1))/(d2.dotproduct(n1))))

            resultLines.append(line3D(node.nodeElement.sort, resultLinePoint, resultLinePoint.add(resultDir)))

        elif node.rhsInfo['case'] == graphNode3D.lineNode.line_casus.CASE_2PARALLEL:
            ### case: 2 parallel lines, direction from these two lines root to the new line must be perpendicular
            # MULTIPLE RESULTS POSSIBLE

            case0 = node.rhsInfo['caseNodes'][0].nodeElement
            case1 = node.rhsInfo['caseNodes'][1].nodeElement
            transformedCase0 = case0.edit(transf)
            transformedCase1 = case1.edit(transf)

            # find the plane that contains both of the parallel lines and its transformation
            pl = plane3D(sortClass.getSort('__dummyPlane3D__'), [case0.root, case0.root.add(case0.direction), case1.root])
            transformedPl = plane3D(sortClass.getSort('__dummyPlane3D__'), 
                            [transformedCase0.root,
                             transformedCase0.root.add(transformedCase0.direction),
                             transformedCase1.root])

            dist = transformedCase0.root.subtract(transformedCase1.root).length() / 2

            rootMidPoint = transformedCase0.root.add(transformedCase1.root).scale(c0p5)

            anchorVect = node.rhsInfo['caseNodes'][2]
            dirCheck = node.rhsInfo['caseNodes'][3]

            if anchorVect is not None:
                trAnchorVect = tr.transform(anchorVect)

                if trAnchorVect is False:
                    anchorCheck = self.replaceCaseAnchor(tr, node, 2, [pl], [transformedPl])
                    if anchorCheck is None:
                        anchorVect = None
                    else:
                        anchorVect, trAnchorVect = anchorCheck
                        dirCheck = node.rhsInfo['caseNodes'][3]

                if anchorVect is not None:
                    projTrAnchor0 = transformedPl.projection(trAnchorVect)
                    if projTrAnchor0 == trAnchorVect:
                        anchorCheck = self.replaceCaseAnchor(tr, node, 2, [pl], [transformedPl])
                        if anchorCheck is None:
                            anchorVect = None
                        else:
                            anchorVect, trAnchorVect = anchorCheck
                            dirCheck = node.rhsInfo['caseNodes'][3]
                            projTrAnchor0 = transformedPl.projection(trAnchorVect)

            if anchorVect is None:
                distDir = transformedPl.normal.scale(dist)

                # case 1: in accordance of the transformed plane normal direction
                nRoot = rootMidPoint.add(distDir)
                resultLines.append(line3D(node.nodeElement.sort, nRoot, nRoot.add(transformedCase0.direction)))

                # case 2: opposite of the transformed plane normal direction
                nRoot = rootMidPoint.subtract(distDir)
                resultLines.append(line3D(node.nodeElement.sort, nRoot, nRoot.add(transformedCase0.direction)))
            else:
                distDir = trAnchorVect.subtract(projTrAnchor0).unit().scale(dist)
                if not dirCheck:
                    distDir = distDir.scale(cn1)
                
                nRoot = rootMidPoint.add(distDir)
                resultLines.append(line3D(node.nodeElement.sort, nRoot, nRoot.add(transformedCase0.direction)))
            parallelProjection = node.rhsInfo['caseNodes'][0].forms + node.rhsInfo['caseNodes'][1].forms

        ###### DIRECTIVE CASES ######
        elif node.rhsInfo['case'] == graphNode3D.lineNode.line_casus.CASE_ANGLE:
            ### case: angle directive, solvable plane, 1 contained point
            # MULTIPLE RESULTS POSSIBLE
            dirStatement = node.rhsInfo['caseNodes'][0]
            bounds, check = self.evaluateDesc(dirStatement, param, special = True)
            if not check:
                return []
            new, base, angle = bounds # pylint: disable=unbalanced-tuple-unpacking

            while angle > pi:
                angle -= 2*pi

            new = new.ind
            base = base.ind
            transformedBase = base.edit(transf)

            pl = node.rhsInfo['caseNodes'][1]
            pt = node.rhsInfo['caseNodes'][2]

            transformedPl = pl.nodeElement.edit(transf)
            transformedPt = tr.transform(pt)


            # p0: projection of (0,0,0) onto cPlane
            p0 =  transformedPl.position(transformedPl.localize(vz))

            basePt1, basePt2, anchorVect, dirCheck = node.rhsInfo['caseNodes'][3:]
            if anchorVect is not None:
                trBasePt1 = tr.transform(basePt1)
                trBasePt2 = tr.transform(basePt2)
                trAnchor = tr.transform(anchorVect)
                if trBasePt1 is None or trBasePt2 is None or trAnchor is None:
                    anchorVect = None
                else:
                    # localized direction of transformedBase
                    # d1 = transformedPl.localize(trBasePt1.subtract(trBasePt2).unit())
                    d1 = transformedPl.localize(trBasePt2.subtract(trBasePt1).unit())

                    trAnchorVect = tr.transform(anchorVect)

                    if trAnchorVect is False:
                        return []

                    projTrAnchor = transformedPl.projection(trAnchorVect)
                    if projTrAnchor.colinear(trBasePt1, trBasePt2):
                        anchorVect = None

            # sin and cos of the angle
            cosAngle = cos(angle.value)
            sinAngle = sin(angle.value)


            if anchorVect is None:
                # localized direction of transformedBase
                d1 = transformedPl.localize(transformedBase.orientationDir())

                newDir = transformedPl.position(d1.__class__(cosAngle*d1.x-sinAngle*d1.y, sinAngle*d1.x+cosAngle*d1.y)).subtract(p0)
                resultLines.append(line3D(new.sort, transformedPt, transformedPt.add(newDir)))

                sinAngle = - sinAngle
                newDir = transformedPl.position(d1.__class__(cosAngle*d1.x-sinAngle*d1.y, sinAngle*d1.x+cosAngle*d1.y)).subtract(p0)
                resultLines.append(line3D(new.sort, transformedPt, transformedPt.add(newDir)))
            else:
                
                cwCheck = graphMap3D.cw(transformedPl, trBasePt1, trBasePt2, projTrAnchor)
                if cwCheck == dirCheck:
                    sinAngle = -sinAngle
                newDir = transformedPl.position(d1.__class__(cosAngle*d1.x-sinAngle*d1.y, sinAngle*d1.x+cosAngle*d1.y)).subtract(p0)
                resultLines.append(line3D(new.sort, transformedPt, transformedPt.add(newDir)))

        elif node.rhsInfo['case'] == graphNode3D.lineNode.line_casus.CASE_POINT_LINE_DISTANCE:
            ### case: point-line distance directive, solvable plane, 1 contained point
            # MULTIPLE RESULTS POSSIBLE
            dirStatement = node.rhsInfo['caseNodes'][0]
            bounds, check = self.evaluateDesc(dirStatement, param, special = True)
            if not check:
                return []
            new = bounds[0].ind
            base = bounds[1].ind
            dist = bounds[2]
            transformedBase = base.edit(transf).value

            pl = node.rhsInfo['caseNodes'][1]
            pt = node.rhsInfo['caseNodes'][2]

            transformedPl = pl.nodeElement.edit(transf)
            transformedPt = tr.transform(pt)
                
            if transformedPt is False:
                return []
            projTransformedPt = transformedPl.projection(transformedPt)
            if transformedPt != projTransformedPt:
                dist = dist * dist - projTransformedPt.subtract(transformedPt).squareLength()
                if dist < 0:
                    return []
                dist = dist.sqrt()
                transformedPt = projTransformedPt

            # p0: projection of (0,0,0) on cPlane
            p0 =  transformedPl.position(transformedPl.localize(vz))

            # localized direction from transformed base point to transformed cPoint
            d1 = transformedPl.localize(transformedBase.subtract(transformedPt))
            d1len = d1.length()

            #     (transformedPt)
            #           |\ <-- angle
            #           | \
            #           |  \
            #           |   \
            #           |    \
            #    result |     \
            #      line |      \ [d1len]
            #           |       \
            #           |        \
            #           |         \
            #           |          \
            #           |           \
            #           |_           \
            #           |_|_ _ _ _ _ _\ (transformedBase)
            #               [dist]

            # sin(angle) = dist/d1len
            sinAngle = dist / d1len

            # distance from projection of transformed base point onto the new line to transformed cPoint
            try:
                # cos(angle) = sqrt(d1len^2 + dist^2) / d1len ()
                cosAngle = (d1len*d1len - dist*dist).sqrt() / d1len
            except ValueError:
                return []
            

            anchorVect = node.rhsInfo['caseNodes'][3]
            dirCheck = node.rhsInfo['caseNodes'][4]

            if anchorVect is not None:
                trAnchorVect = tr.transform(anchorVect)

                if trAnchorVect is False:
                    return []

                projTrAnchor = transformedPl.projection(trAnchorVect)

                if projTrAnchor.colinear(transformedBase, transformedPt):
                    anchorVect = None

            if anchorVect is None:
                newDir = transformedPl.position(d1.__class__(cosAngle*d1.x-sinAngle*d1.y, sinAngle*d1.x+cosAngle*d1.y)).subtract(p0)
                resultLines.append(line3D(new.sort, transformedPt, transformedPt.add(newDir)))

                sinAngle = - sinAngle
                newDir = transformedPl.position(d1.__class__(cosAngle*d1.x-sinAngle*d1.y, sinAngle*d1.x+cosAngle*d1.y)).subtract(p0)
                resultLines.append(line3D(new.sort, transformedPt, transformedPt.add(newDir)))
            else:
                cwCheck = graphMap3D.cw(transformedPl, transformedBase, transformedPt, projTrAnchor)
                if cwCheck != dirCheck:
                    sinAngle = -sinAngle

                # direction of the transformed new line
                newDir = transformedPl.position(d1.__class__(cosAngle*d1.x-sinAngle*d1.y, sinAngle*d1.x+cosAngle*d1.y)).subtract(p0)

                resultLines.append(line3D(new.sort, transformedPt, transformedPt.add(newDir)))

        elif node.rhsInfo['case'] == graphNode3D.lineNode.line_casus.CASE_LINE_LINE_DISTANCE:
            ### case: line-line distance directive with no direction vector, solvable plane
            # MULTIPLE RESULTS POSSIBLE
            dirStatement = node.rhsInfo['caseNodes'][0]
            bounds, check = self.evaluateDesc(dirStatement, param, special = True)
            if not check:
                return []

            new = bounds[0].ind
            base = bounds[1].ind
            dist = bounds[2]
            transformedBase = base.edit(transf)

            cPlane = node.rhsInfo['caseNodes'][1].nodeElement
            trCPlane = cPlane.edit(transf)
            if trCPlane is None:
                trCPlane = cPlane

            if not cPlane.containsLine(base):
                base = base.__class__(base.sort, cPlane.projection(base.start), cPlane.projection(base.end))
                projStart = trCPlane.projection(transformedBase.start)
                dist = dist * dist - transformedBase.start.subtract(projStart).squareLength()
                if dist < c0:
                    return []
                dist = dist.sqrt()
                transformedBase = transformedBase.__class__(transformedBase.sort, projStart, trCPlane.projection(transformedBase.end))
            anchorVect = node.rhsInfo['caseNodes'][2]
            dirCheck = node.rhsInfo['caseNodes'][3]

            if anchorVect is not None:
                trAnchorVect = tr.transform(anchorVect)

                if trAnchorVect is False:
                    anchorCheck = self.replaceCaseAnchor(tr, node, 2, [cPlane, base], [trCPlane, transformedBase])
                    if anchorCheck is None:
                        anchorVect = None
                    else:
                        anchorVect, trAnchorVect = anchorCheck
                        dirCheck = node.rhsInfo['caseNodes'][3]
                if anchorVect is not None:
                    projTrAnchor0 = trCPlane.projection(trAnchorVect)
                    projTrAnchor1 = transformedBase.projection(trAnchorVect)

                    if projTrAnchor0 == projTrAnchor1:
                        anchorCheck = self.replaceCaseAnchor(tr, node, 2, [cPlane, base], [trCPlane, transformedBase])
                        if anchorCheck is None:
                            anchorVect = None
                        else:
                            anchorVect, trAnchorVect = anchorCheck
                            dirCheck = node.rhsInfo['caseNodes'][3]
                            projTrAnchor0 = trCPlane.projection(trAnchorVect)
                            projTrAnchor1 = transformedBase.projection(trAnchorVect)

            # if anchorVect and dirCheck ares None -> two result
            # * this would only happens if there is no existing vector in the transformation that has its projection on the plane
            # not on the base line --> the plane is not transformable since if it is, there's at least one other point that is 
            # on the plane but not on the base line
            if anchorVect is None:
                distDir = trCPlane.normal.product(transformedBase.unitDir).scale(dist)
                resultLines.append(line3D(node.nodeElement.sort, transformedBase.start.add(distDir), transformedBase.end.add(distDir)))
                resultLines.append(line3D(node.nodeElement.sort, transformedBase.start.subtract(distDir), transformedBase.end.subtract(distDir)))
            else:
                distDir = projTrAnchor0.subtract(projTrAnchor1).unit().scale(dist)
                if not dirCheck:
                    distDir = distDir.scale(cn1)
                resultLines.append(line3D(node.nodeElement.sort, transformedBase.start.add(distDir), transformedBase.end.add(distDir)))
            # special line end-points matching: if the new line's start/end points are projection of the base line's start/end points
            parallelProjection = [bounds[1].ind]

        elif node.rhsInfo['case'] == graphNode3D.lineNode.line_casus.CASE_LINE_LINE_DISTANCE_DIRECTION:
            ### case: line-line distance directive with direction vector
            dirStatement = node.rhsInfo['caseNodes'][0]
            bounds, check = self.evaluateDesc(dirStatement, param, special = True)
            if not check:
                return []

            new = bounds[0].ind
            base = bounds[1].ind
            dist = bounds[2]
            distDirection = bounds[3]

            transformedBase = base.edit(transf)

            if not isinstance(distDirection, vector3D):
                distDirection = vector3D(*vector3D.readVector(parseReader(distDirection))).unit()

            resultLines.append(line3D(node.nodeElement.sort, 
                                transformedBase.start.add(distDirection.scale(dist)),
                                transformedBase.end.add(distDirection.scale(dist))))

            # special line end-points matching: if the new line's start/end points are projection of the base line's start/end points
            parallelProjection = [base]

        elif node.rhsInfo['case'] == graphNode3D.lineNode.line_casus.CASE_DIRECTION:
            ### case: direction directive, 1 contained point
            dirStatement = node.rhsInfo['caseNodes'][0]
            bounds, check = self.evaluateDesc(dirStatement, param, special = True)

            if not check:
                return []

            if isinstance(bounds[1], vector3D):
                newDirection = bounds[1]
            else:
                newDirection = vector3D(*vector3D.readVector(parseReader(bounds[1]))).unit()

            containedPoint = node.rhsInfo['caseNodes'][1]
            trContainedPoint = tr.transform(containedPoint)
            if not trContainedPoint:
                return []
            reverseCheck = False
            resultLines.append(line3D(node.nodeElement.sort, trContainedPoint, trContainedPoint.add(newDirection)))

        ###### OTHER CASES ######s
        elif node.rhsInfo['case'] == graphNode3D.lineNode.line_casus.CASE_PROXY:
            ### case: line already defined through other line/plane cases
            try:
                resultLines.append(node.nodeElement.edit(transf))
            except IndexError:
                return []

        elif node.rhsInfo['case'] == graphNode3D.lineNode.line_casus.CASE_LHS:
            ### case: line in the lhs -> no need to solve any more line segment, only need to solve point_on_line
            try:
                resultLines.append(node.nodeElement.edit(transf))
                for directive in node.rhsInfo['directives']:
                    if directive[0] == 'point_on_line':
                        ind = directive[1]
                        bounds = directive[2]
                        if len(directive) == 4:
                            pointOnLineTarget = directive[3]
                        else:
                            pointOnLineTarget = None
                        ratio, evalCheck = self.evaluateDesc(bounds, param)
                        if not evalCheck:
                            return []
                        self.solvePointOnLine(tr, ind, ratio, pointOnLineTarget)
                return [transf]
            except IndexError:
                return []

        if resultLines is None:
            return []

        ################### SOLVE INTERSECTIONS + END-POINTS ###################
        resultTransformations = []
        originalTransf = transf
        for resultLine in resultLines:
            transf = originalTransf.duplicate()
            tr = self.getTransform(transf)
            # intersection with lhs nodes
            for lhsNode in self.nodeList.lines:
                if lhsNode.nodeElement.intersect(node.nodeElement):
                    transformedLHSNode = lhsNode.nodeElement.edit(transf)
                    if not transformedLHSNode.intersect(resultLine):
                        continue
                    intersection = lhsNode.nodeElement.intersection(node.nodeElement)
                    transformedInt = transformedLHSNode.intersection(resultLine)
                    tr.add(intersection, transformedInt, override = True)
            # intersection with solved rhs nodes
            for rhsNode in self.nodeList.rhsLines:
                try:
                    transformedRHSNode = rhsNode.nodeElement.edit(transf)
                except IndexError:
                    continue
                if rhsNode.nodeElement.intersect(node.nodeElement):
                    if not transformedRHSNode.intersect(resultLine):
                        continue
                    intersection = rhsNode.nodeElement.intersection(node.nodeElement)
                    transformedInt = transformedRHSNode.intersection(resultLine)
                    tr.add(intersection, transformedInt, override = True)

            # find parallel projection if indicated by the case
            if parallelProjection is not None:
                for baseLine in parallelProjection:
                    self.parallelProjection(tr, baseLine, resultLine, node)

            # checkPoints = tr.getCheckPoints(node.nodeElement)
            checkPoints = self.getCheckPoints(tr, node, False)


            for directive in node.rhsInfo['directives']:
                if directive[0] == 'point_on_line':
                    ind = directive[1]
                    lengthDir = None
                    for i in node.rhsInfo['directives']:
                        if i[0] == 'length' and i[1] == ind:
                            lengthDir = i
                    if not self.solveOneRHSLineSeg(tr, param, node, resultLine, ind,
                    checkPoints, False, lengthDirective = lengthDir):
                        return []
                    bounds = directive[2]
                    if len(directive) == 4:
                        pointOnLineTarget = directive[3]
                    else:
                        pointOnLineTarget = None

                    ratio, evalCheck = self.evaluateDesc(bounds, param)
                    if not evalCheck:
                        return []
                    self.solvePointOnLine(tr, ind, ratio, pointOnLineTarget)


            # check if the line is matched in reverse order or not
            if reverseCheck is not None:
                reverse = reverseCheck
            else:
                reverse = None
                # _ if there are 2 transformable points or more
                while len(checkPoints) > 1:
                    if checkPoints[0][1] == checkPoints[1][1]:
                        checkPoints.pop(1)
                        continue
                    d0 = checkPoints[1][0].subtract(checkPoints[0][0]) > vz
                    d1 = checkPoints[1][1].subtract(checkPoints[0][1]) > vz
                    reverse = d0 != d1
                    break

                # _ if there's no transformable point on the line (this case is probably not possible)
                if len(checkPoints) == 0:
                    # reverse = False
                    print('?????')
                    return []

                # _ if there's only one transformable point on the line
                elif len(checkPoints) == 1:
                    for vect in tr.vects:
                        # find a point projection on the line that is not
                        projvect0 = node.nodeElement.projection(vect[0])
                        if checkPoints[0][0] == projvect0:
                            continue
                        projvect1 = resultLine.projection(vect[1])
                        if checkPoints[0][1] == projvect1:
                            continue
                        d0 = projvect0.subtract(checkPoints[0][0]).unit() > vz
                        d1 = projvect1.subtract(checkPoints[0][1]).unit() > vz
                        # print(d0, d1)
                        # reverse = (d0.add(d1)).squareLength() == c0
                        reverse = d0 != d1
                        break

            # if reverse is None -> do both non-reverse and reverse transformation order
            # if reverse is False -> do only non-reverse
            # if reverse is True -> do only reverse
            if reverse is None or reverse == False:
                newTransf = transf.duplicate()
                newTr = self.getTransform(newTransf)
                param.transform = newTransf
                # solve the one with length directive first
                for directive in node.rhsInfo['directives']:
                    if directive[0] == 'length':
                        if not self.solveOneRHSLineSeg(newTr, param, node, resultLine, directive[1],
                        checkPoints, False, lengthDirective = directive):
                            return []
                for lineSeg in node.forms:
                    if not self.solveOneRHSLineSeg(newTr, param, node, resultLine, lineSeg, checkPoints, False):
                        return []
                resultTransformations.append(newTransf)

            if reverse is None or reverse == True:
                newTransf = transf.duplicate()
                newTr = self.getTransform(newTransf)
                param.transform = newTransf
                # solve the one with length directive first
                for directive in node.rhsInfo['directives']:
                    if directive[0] == 'length':
                        if not self.solveOneRHSLineSeg(newTr, param, node, resultLine, directive[1],
                        checkPoints, True, lengthDirective = directive):
                            return []
                for lineSeg in node.forms:
                    if not self.solveOneRHSLineSeg(newTr, param, node, resultLine, lineSeg, checkPoints, True):
                        return []
                resultTransformations.append(newTransf)
        # for transf in resultTransformations:
        #     tr = self.getTransform(transf)
        #     for directive in node.rhsInfo['directives']:
        #         if directive[0] == 'point_on_line':
        #             ind = directive[1]
        #             bounds = directive[2]
        #             ratio, evalCheck = self.evaluateDesc(bounds, param)
        #             if not evalCheck:
        #                 return []
        #             self.solvePointOnLine(tr, ind, ratio)
        return resultTransformations



    def solveRHSPlane(self, node, transf, param):
        """
        solve a specific plane node with a specific transformation set and update it
        """
        tr = self.getTransform(transf)
        param.transform = transf
        resultPlanes = []
        addDummy = True
        ##################### SOLVE LINE DIRECTION + ROOT #####################

        ###### OTHER PLANE CASES ######
        if node.rhsInfo['case'] == graphNode3D.planeNode.plane_casus.CASE_1PARALLEL_1PT:
            pl = node.rhsInfo['caseNodes'][0]
            pt = node.rhsInfo['caseNodes'][1]

            transformedPl = pl.nodeElement.edit(transf)
            transformedPt = tr.transform(pt)

            if transformedPt is False:
                return []

            testDir = vector3D(0,0,1)
            if testDir.parallel(transformedPl.normal):
                testDir = vector3D(1,0,0)

            u = testDir.product(transformedPl.normal)
            v = u.product(transformedPl.normal)

            pt0 = transformedPt.add(u)
            pt1 = transformedPt.add(v)

            resultPlanes.append(plane3D(sortClass.getSort('__dummyPlane3D__'), [transformedPt, pt0, pt1]))

        if node.rhsInfo['case'] == graphNode3D.planeNode.plane_casus.CASE_2PERP_1PT:
            case0 = node.rhsInfo['caseNodes'][0]
            case1 = node.rhsInfo['caseNodes'][1]
            pt = node.rhsInfo['caseNodes'][2]

            transformedCase0 = case0.nodeElement.edit(transf)
            transformedCase1 = case1.nodeElement.edit(transf)
            if transformedCase0.normal.parallel(transformedCase1.normal):
                return []
            transformedPt = tr.transform(pt)
            if transformedPt is False:
                return []
            
            pt0 = transformedPt.add(transformedCase0.normal)
            pt1 = transformedPt.add(transformedCase1.normal)

            resultPlanes.append(plane3D(sortClass.getSort('__dummyPlane3D__'), [transformedPt, pt0, pt1]))


        ###### LINE CASES ######
        if node.rhsInfo['case'] == graphNode3D.planeNode.plane_casus.CASE_1PERPLN_1PT:
            ln = node.rhsInfo['caseNodes'][0]
            pt = node.rhsInfo['caseNodes'][1]

            transformedLn = ln.nodeElement.edit(transf)
            transformedPt = tr.transform(pt)

            if transformedPt is False:
                return []

            testDir = vector3D(0,0,1)
            if testDir.parallel(transformedLn.direction):
                testDir = vector3D(1,0,0)
            u = testDir.product(transformedLn.direction)
            v = u.product(transformedLn.direction)

            pt0 = transformedPt.add(u)
            pt1 = transformedPt.add(v)

            resultPlanes.append(plane3D(sortClass.getSort('__dummyPlane3D__'), [transformedPt, pt0, pt1]))

        if node.rhsInfo['case'] == graphNode3D.planeNode.plane_casus.CASE_2PARALLELLN_1PT:
            case0 = node.rhsInfo['caseNodes'][0]
            case1 = node.rhsInfo['caseNodes'][1]
            pt = node.rhsInfo['caseNodes'][2]

            transformedCase0 = case0.nodeElement.edit(transf)
            transformedCase1 = case1.nodeElement.edit(transf)
            if transformedCase0.parallel(transformedCase1):
                return []
            transformedPt = tr.transform(pt)
            if transformedPt is False:
                return []
            
            pt0 = transformedPt.add(transformedCase0.direction)
            pt1 = transformedPt.add(transformedCase1.direction)

            resultPlanes.append(plane3D(sortClass.getSort('__dummyPlane3D__'), [transformedPt, pt0, pt1]))


        ###### DIRECTIVE CASES ######
        if node.rhsInfo['case'] == graphNode3D.planeNode.plane_casus.CASE_PLANE_PLANE_DISTANCE:
            dirStatement = node.rhsInfo['caseNodes'][0]
            bounds, check = self.evaluateDesc(dirStatement, param, special = True)
            if not check:
                return []

            new = bounds[0].ind
            base = bounds[1].ind
            dist = bounds[2]
            transformedBase = base.edit(transf)

            anchorVect = node.rhsInfo['caseNodes'][1]
            dirCheck = node.rhsInfo['caseNodes'][2]

            if anchorVect is not None:
                trAnchorVect = tr.transform(anchorVect)

                if trAnchorVect is False:
                    anchorCheck = self.replaceCaseAnchor(tr, node, 1, [base], [transformedBase])
                    if anchorCheck is None:
                        anchorVect = None
                    else:
                        anchorVect, trAnchorVect = anchorCheck
                        dirCheck = node.rhsInfo['caseNodes'][2]

                if anchorVect is not None:
                    projAnchor = transformedBase.projection(trAnchorVect)

                    if projAnchor == trAnchorVect:
                        anchorCheck = self.replaceCaseAnchor(tr, node, 1, [base], [transformedBase])
                        if anchorCheck is None:
                            anchorVect = None
                        else:
                            anchorVect, trAnchorVect = anchorCheck
                            dirCheck = node.rhsInfo['caseNodes'][2]
                            projAnchor = transformedBase.projection(trAnchorVect)

            if anchorVect is None:
                newRoot = transformedBase.pRoot.add(transformedBase.normal.scale(dist))
                resultPlanes.append(plane3D(sortClass.getSort('__dummyPlane3D__'), None, normal_n_origin = True,
                                    normal = transformedBase.normal, origin = newRoot))

                newRoot = transformedBase.pRoot.subtract(transformedBase.normal.scale(dist))
                resultPlanes.append(plane3D(sortClass.getSort('__dummyPlane3D__'), None, normal_n_origin = True,
                                    normal = transformedBase.normal, origin = newRoot))
            else:
                trAnchorVect = tr.transform(anchorVect)
                if trAnchorVect is False:
                    return []
                projAnchor = transformedBase.projection(trAnchorVect)
                distDir = trAnchorVect.subtract(projAnchor).unit().scale(dist)
                if not dirCheck:
                    distDir = distDir.scale(cn1)
                newRoot = transformedBase.pRoot.add(distDir)
                resultPlanes.append(plane3D(sortClass.getSort('__dummyPlane3D__'), None, normal_n_origin = True,
                                    normal = transformedBase.normal, origin = newRoot))


        ###### OTHER CASES ######
        if node.rhsInfo['case'] == graphNode3D.planeNode.plane_casus.CASE_DEFAULT:
            containedVecs = node.rhsInfo['caseNodes'][0]
            if len(containedVecs) == 1:
                newRoot = tr.transform(containedVecs[0])
                resultPlanes.append(plane3D(sortClass.getSort('__dummyPlane3D__'), None, normal_n_origin = True,
                                    normal = node.nodeElement.normal, origin = newRoot))
            elif len(containedVecs) == 2:
                pt1 = tr.transform(containedVecs[0])
                pt2 = tr.transform(containedVecs[1])
                ptDir = pt2.subtract(pt1)
                newNormal = node.nodeElement.normal
                if ptDir.dotproduct(newNormal) != 0:
                    newNormal = ptDir.product(vector3D(1,0,0))
                resultPlanes.append(plane3D(sortClass.getSort('__dummyPlane3D__'), None, normal_n_origin = True,
                                    normal = node.nodeElement.normal, origin = pt1))


        if node.rhsInfo['case'] == graphNode3D.planeNode.plane_casus.CASE_PROXY:
            addDummy = False
            resultPlanes.append(node.nodeElement.edit(transf))
        if node.rhsInfo['case'] == graphNode3D.planeNode.plane_casus.CASE_LHS:
            return [transf]

        if len(node.forms) == 0 and addDummy:
            u = node.nodeElement.u.scale(100).add(node.nodeElement.v)
            v = node.nodeElement.v.scale(100).add(node.nodeElement.u)

            pt0 = node.nodeElement.pRoot.add(u).add(v)

            check = 0
            while tr.contains(pt0, True):
                if check > 1:
                    u = u.add(node.nodeElement.u)
                    v = v.add(node.nodeElement.v)
                pt0 = pt0.add(u).add(v)
                check += 1
            pt1 = pt0.add(u)
            while tr.contains(pt1, True):
                pt1 = pt1.add(u)
            pt2 = pt0.add(v)
            while tr.contains(pt2, True):
                pt2 = pt2.add(v)
        resultTransformations = []
        oriTransf = transf
        for resultPlane in resultPlanes:
            transf = oriTransf.duplicate()
            tr = self.getTransform(transf)
            if len(node.forms) > 0:
                for planeSeg in node.forms:
                    for boundary in planeSeg.boundaries:
                        for p in boundary.polygonal:
                            check = False
                            if tr.contains(p):
                                continue
                            for vect in tr.vects:
                                projVect = planeSeg.projection(vect[0])
                                if p == projVect:
                                    tr.add(p, resultPlane.projection(vect[1]))
                                    check = True
                                    break
                            # if not check:
                            #     return []
            elif addDummy:
                nU = resultPlane.u.scale(100).add(resultPlane.v)
                nV = resultPlane.v.scale(100).add(resultPlane.u)
                nPt0 = resultPlane.pRoot.add(nU).add(nV)
                nPt1 = nPt0.add(nU)
                nPt2 = nPt0.add(nV)
                extraKey = str(node.nodeElement)
                tr.addExtra(pt0, nPt0, extraKey)
                tr.addExtra(pt1, nPt1, extraKey)
                tr.addExtra(pt2, nPt2, extraKey)
            resultTransformations.append(transf)
        return resultTransformations

    def solveRHSPoint(self, node, transf, param):
        """
        solve a specific point node with a specific transformation set and update it
        """
        tr = self.getTransform(transf)
        param.transform = transf
        if node.rhsInfo['case'] == graphNode3D.pointNode.point_casus.CASE_POINT_POINT_DISTANCE_DIRECTION:
            ### case: point-point distance directive with direction vector
            dirStatement = node.rhsInfo['caseNodes'][0]
            bounds, check = self.evaluateDesc(dirStatement, param, special = True)
            if not check:
                return []

            new = bounds[0].ind.value
            base = bounds[1].ind.value
            dist = bounds[2]
            distDirection = bounds[3]

            transformedBase = tr.transform(base)

            if not transformedBase:
                return []

            if not isinstance(distDirection, vector3D):
                distDirection = vector3D(*vector3D.readVector(parseReader(distDirection))).unit()

            tr.add(new, transformedBase.add(distDirection.scale(dist)))
            return [transf]
        elif node.rhsInfo['case'] == graphNode3D.pointNode.point_casus.CASE_POINT_POINT_DISTANCE:
            ### case: point-point distance directive with no direction vector
            dirStatement = node.rhsInfo['caseNodes'][0]
            bounds, check = self.evaluateDesc(dirStatement, param, special = True)
            if not check:
                return []

            new = bounds[0].ind.value
            base = bounds[1].ind.value
            dist = bounds[2]
            transformedBase = tr.transform(base)

            containingLine = node.rhsInfo['caseNodes'][1]
            trContainingLine = containingLine.edit(transf)

            anchorVect = node.rhsInfo['caseNodes'][2]
            dirCheck = node.rhsInfo['caseNodes'][3]

            if anchorVect is not None:
                trAnchorVect = tr.transform(anchorVect)

                perpPlane = plane3D(containingLine.sort, [], True, containingLine.unitDir, base)
                trPerpPlane = plane3D(trContainingLine.sort, [], True, trContainingLine.unitDir, transformedBase)
                if trAnchorVect is False or trPerpPlane.projection(trAnchorVect) == trAnchorVect:
                    anchorCheck = self.replaceCaseAnchor(tr, node, 2, [perpPlane], [trPerpPlane])
                    if anchorCheck is None:
                        anchorVect = None
                    else:
                        anchorVect, trAnchorVect = anchorCheck
                        dirCheck = node.rhsInfo['caseNodes'][2]
                if anchorVect is not None:
                    projTrAnchor = trPerpPlane.projection(trAnchorVect)

            projBase = trContainingLine.projection(transformedBase)
            distFromProjBase = (dist * dist - transformedBase.subtract(projBase).squareLength())
            if distFromProjBase < c0:
                return []
            distFromProjBase = distFromProjBase.sqrt()
            # if anchorVect and dirCheck ares None -> two result
            if anchorVect is None:
                distDir = trContainingLine.unitDir().scale(distFromProjBase)
                transf0 = transf.duplicate()
                tr0 = self.getTransform(transf0)
                tr0.add(new, projBase.add(distDir))
                transf1 = transf.duplicate()
                tr1 = self.getTransform(transf1)
                tr1.add(new, projBase.subtract(distDir))
                return [transf0, transf1]
            else:
                distDir = trAnchorVect.subtract(projTrAnchor).unit().scale(distFromProjBase)
                if not dirCheck:
                    distDir = distDir.scale(cn1)
                tr.add(new, projBase.add(distDir))
                return [transf]

        elif node.rhsInfo['case'] == graphNode3D.pointNode.point_casus.CASE_LINE_POINT_DISTANCE_INTERSECT:
            ### case: line-point distance directive
            dirStatement = node.rhsInfo['caseNodes'][0]
            bounds, check = self.evaluateDesc(dirStatement, param, special = True)
            if not check:
                return []

            new = bounds[0].ind.value
            base = bounds[1].ind
            dist = bounds[2]

            transformedBase = base.edit(transf)

            containingLine = node.rhsInfo['caseNodes'][1]
            trContainingLine = containingLine.edit(transf)

            if not trContainingLine.intersect(transformedBase):
                return []

            anchorVect = node.rhsInfo['caseNodes'][2]
            dirCheck = node.rhsInfo['caseNodes'][3]

            cPlane = plane3D(sortClass.getSort('__dummyPlane3D__'), [], True, containingLine.unitDir.product(base.unitDir), containingLine.root)
            trCPlane = plane3D(sortClass.getSort('__dummyPlane3D__'), [], True, trContainingLine.unitDir.product(transformedBase.unitDir), trContainingLine.root)
            if anchorVect is not None:
                trAnchorVect = tr.transform(anchorVect)
                if trAnchorVect is False:
                    anchorCheck = self.replaceCaseAnchor(tr, node, 2, [cPlane, base], [trCPlane, transformedBase])
                    if anchorCheck is None:
                        anchorVect = None
                    else:
                        anchorVect, trAnchorVect = anchorCheck
                        dirCheck = node.rhsInfo['caseNodes'][2]
                if anchorVect is not None:
                    projTrAnchor0 = trCPlane.projection(trAnchorVect)
                    projTrAnchor1 = transformedBase.projection(trAnchorVect)
                    if projTrAnchor0 == projTrAnchor1 or trAnchorVect == projTrAnchor1 or projTrAnchor0 == trAnchorVect:
                        anchorCheck = self.replaceCaseAnchor(tr, node, 2, [cPlane, base], [trCPlane, transformedBase])
                        if anchorCheck is None:
                            anchorVect = None
                        else:
                            anchorVect, trAnchorVect = anchorCheck
                            dirCheck = node.rhsInfo['caseNodes'][2]
                            projTrAnchor0 = trCPlane.projection(trAnchorVect)
                            projTrAnchor1 = transformedBase.projection(trAnchorVect)

            # if anchorVect and dirCheck ares None -> two result
            # However, this should never happen since the 2 end points of containingLine can always be the anchor
            if anchorVect is None:
                moveDir = transformedBase.unitDir.product(trCPlane.normal).scale(dist)
                movedtrBase0 = line3D(base.sort, transformedBase.root.add(moveDir), transformedBase.root.add(transformedBase.direction).add(moveDir))
                transf0 = transf.duplicate()
                tr0 = self.getTransform(transf0)
                tr0.add(new, movedtrBase0.intersection(trContainingLine))
                movedtrBase1 = line3D(base.sort, transformedBase.root.subtract(moveDir), transformedBase.root.add(transformedBase.direction).subtract(moveDir))
                transf1 = transf.duplicate()
                tr1 = self.getTransform(transf1)
                tr0.add(new, movedtrBase1.intersection(trContainingLine))
                return [transf0, transf1]
            # if there is anchorVect -> only one result
            else:
                moveDir = projTrAnchor0.subtract(projTrAnchor1).unit().scale(dist)
                movedtrBase = line3D(base.sort, transformedBase.root.add(moveDir), transformedBase.root.add(transformedBase.direction).add(moveDir))
                if not dirCheck:
                    distDir = distDir.scale(cn1)
                tr.add(new, movedtrBase.intersection(trContainingLine))
                return [transf]

        elif node.rhsInfo['case'] == graphNode3D.pointNode.point_casus.CASE_LINE_POINT_DISTANCE_SKEW:
            ### case: line-point distance directive
            dirStatement = node.rhsInfo['caseNodes'][0]
            bounds, check = self.evaluateDesc(dirStatement, param, special = True)
            if not check:
                return []

            new = bounds[0].ind.value
            base = bounds[1].ind
            dist = bounds[2]

            transformedBase = base.edit(transf)

            containingLine = node.rhsInfo['caseNodes'][1]
            trContainingLine = containingLine.edit(transf)

            if trContainingLine.parallel(transformedBase):
                return []

            anchorVect = node.rhsInfo['caseNodes'][2]
            dirCheck = node.rhsInfo['caseNodes'][3]

            n = trContainingLine.direction.product(transformedBase.direction)
            p1 = transformedBase.root
            p2 = trContainingLine.root
            d1 = transformedBase.direction
            d2 = trContainingLine.direction
            n1 = d1.product(n)
            n2 = d2.product(n)
            trBasePt = p1.add(d1.scale((p2.subtract(p1).dotproduct(n2))/(d1.dotproduct(n2))))
            trContainingLinePt = p2.add(d2.scale((p1.subtract(p2).dotproduct(n1))/(d2.dotproduct(n1))))

            movedtrBaseX = line3D(base.sort, trContainingLinePt, trContainingLinePt.add(transformedBase.direction))

            cPlane = plane3D(sortClass.getSort('__dummyPlane3D__'), [], True, containingLine.unitDir.product(base.unitDir), containingLine.root)
            trCPlane = plane3D(sortClass.getSort('__dummyPlane3D__'), [], True, n, trContainingLine.root)

            if anchorVect is not None:
                trAnchorVect = tr.transform(anchorVect)
                if trAnchorVect is False:
                    anchorCheck = self.replaceCaseAnchor(tr, node, 2, [cPlane, base], [trCPlane, movedtrBaseX])
                    if anchorCheck is None:
                        anchorVect = None
                    else:
                        anchorVect, trAnchorVect = anchorCheck
                        dirCheck = node.rhsInfo['caseNodes'][2]
                if anchorVect is not None:
                    projTrAnchor0 = trCPlane.projection(trAnchorVect)
                    projTrAnchor1 = movedtrBaseX.projection(trAnchorVect)
                    if projTrAnchor0 == projTrAnchor1 or trAnchorVect == projTrAnchor1 or projTrAnchor0 == trAnchorVect:
                        anchorCheck = self.replaceCaseAnchor(tr, node, 2, [cPlane, base], [trCPlane, movedtrBaseX])
                        if anchorCheck is None:
                            anchorVect = None
                        else:
                            anchorVect, trAnchorVect = anchorCheck
                            dirCheck = node.rhsInfo['caseNodes'][2]
                            projTrAnchor0 = trCPlane.projection(trAnchorVect)
                            projTrAnchor1 = movedtrBaseX.projection(trAnchorVect)

            projectionDist = c0
            if trBasePt != trContainingLinePt:
                projectionDist = trContainingLinePt.subtract(trBasePt).squareLength()
            moveDist = (dist * dist - projectionDist).sqrt()
            # if anchorVect and dirCheck ares None -> two result
            if anchorVect is None:
                moveDir = transformedBase.unitDir.product(trCPlane.normal).scale(moveDist)
                movedtrBase0 = line3D(base.sort, transformedBase.root.add(moveDir), transformedBase.root.add(transformedBase.direction).add(moveDir))
                transf0 = transf.duplicate()
                tr0 = self.getTransform(transf0)
                tr0.add(new, movedtrBase0.intersection(trContainingLine))
                movedtrBase1 = line3D(base.sort, transformedBase.root.subtract(moveDir), transformedBase.root.add(transformedBase.direction).subtract(moveDir))
                transf1 = transf.duplicate()
                tr1 = self.getTransform(transf1)
                tr0.add(new, movedtrBase1.intersection(trContainingLine))
                return [transf0, transf1]
            else:
                moveDir = projTrAnchor0.subtract(projTrAnchor1).unit().scale(moveDist)
                movedtrBase = line3D(base.sort, trContainingLinePt.add(moveDir), trContainingLinePt.add(transformedBase.direction).add(moveDir))
                if not dirCheck:
                    distDir = distDir.scale(cn1)
                tr.add(new, movedtrBase.intersection(trContainingLine))
                return [transf]




    def solveOneRHSLineSeg(self, tr, param, node, resultLine, lineSeg, checkPoints, reverse, lengthDirective = None):
        startSolved = tr.contains(lineSeg.start)
        endSolved   = tr.contains(lineSeg.end)
        # print('~~~___',lineSeg.printRhino(), startSolved, endSolved)
        # if both start and end points are already in the transformation set -> return
        if startSolved and endSolved:
            # if there's a length directive, check the matched line if it has the correct length 
            if lengthDirective is not None:

                # get the length
                dirStatement = lengthDirective[-1]
                bounds, check = self.evaluateDesc(dirStatement, param)
                if not check:
                    return False
                length = bounds[-1]

                # if the length matches -> return True
                matchStart = tr.transform(lineSeg.start)
                matchEnd = tr.transform(lineSeg.end)
                oriSqLength = matchEnd.subtract(matchStart).squareLength()
                if oriSqLength == length * length:
                    return True

                # if the length does not match -> readjust the point that is added last (or return False)
                revMod = cn1 if reverse else c1
                for i in tr.vects:
                    if i[0] == lineSeg.start:
                        # adjust end point
                        readjusted = lineSeg.end
                        newMatch = matchStart.add(resultLine.unitDir.scale(revMod).scale(length))
                        break
                    elif i[0] == lineSeg.end:
                        readjusted = lineSeg.start
                        newMatch = matchEnd.subtract(resultLine.unitDir.scale(revMod).scale(length))
                        break
                tr.add(readjusted, newMatch, override = True)
            return True

        # find the line length
        # line case 1: solve with length from length directive
        length = None
        if lengthDirective is not None:
            dirStatement = lengthDirective[-1]
            bounds, check = self.evaluateDesc(dirStatement, param)
            if not check:
                return False
            length = bounds[-1]
        if length is None and len(node.rhsInfo['parallel']) > 0:
            # line case 2: a parallel line that has both start and end points being the projections 
            #              of the new line start and end points
            for parallelNode in node.rhsInfo['parallel']:
                # if line is not transformable, continue
                try:
                    parallelNode.nodeElement.edit(tr)
                except IndexError:
                    continue
                # find a parallel line that has equivalent start and end points (same length)
                for parallelLine in parallelNode.forms:
                    if (lineSeg.projection(parallelLine.start) == lineSeg.start
                    and lineSeg.projection(parallelLine.end  ) == lineSeg.end):
                        nStart = tr.transform(parallelLine.start)
                        nEnd = tr.transform(parallelLine.end)
                        if not nStart or not nEnd:
                            continue
                        tr.add(lineSeg.start, resultLine.projection(nStart))
                        tr.add(lineSeg.end, resultLine.projection(nEnd))
                        return True

            # line case 3: the start and end points are projections of some other parallel lines start/end points
            for parallelNode in node.rhsInfo['parallel']:
                # if line is not transformable, continue
                try:
                    parallelNode.nodeElement.edit(tr)
                except IndexError:
                    continue
                # find a line end-point that its projection is the start/end point of the new line
                for parallelLine in parallelNode.forms:

                    projS = lineSeg.projection(parallelLine.start)
                    projE = lineSeg.projection(parallelLine.end)
                    nStart = tr.transform(parallelLine.start)
                    nEnd = tr.transform(parallelLine.end)

                    if projS == lineSeg.start and nStart != False:
                        tr.add(lineSeg.start, resultLine.projection(nStart))
                    elif projS == lineSeg.end and nStart != False:
                        tr.add(lineSeg.end, resultLine.projection(nStart))

                    if projE == lineSeg.start and nEnd != False:
                        tr.add(lineSeg.start, resultLine.projection(nEnd))
                    elif projE == lineSeg.end and nEnd != False:
                        tr.add(lineSeg.end, resultLine.projection(nEnd))
            if tr.contains(lineSeg.start) and tr.contains(lineSeg.end):
                return True
        # line case 4: a perpendicular line with a known end point intersecting at the unmatched point
        if length is None and len(node.rhsInfo['perpendicular']) > 0 and (startSolved or endSolved):
            tobeSolved = lineSeg.end if startSolved else lineSeg.start
            for pNode in node.rhsInfo['perpendicular']:
                if not pNode.nodeElement.containsPoint(tobeSolved):
                    continue
                # if it's a intersection with another line and that line has a legit case but is not yet matched
                # -> wait until that line is matched
                if (pNode.rhsInfo and pNode.rhsInfo['case'] != graphNode3D.lineNode.line_casus.CASE_PROXY
                and pNode.rhsInfo['case'] != graphNode3D.lineNode.line_casus.CASE_LHS 
                and node not in pNode.rhsInfo['caseNodes']):
                    return True

                for l in pNode.forms:
                    if tr.contains(l.start):
                        tr.add(tobeSolved, resultLine.projection(tr.transform(l.start)))
                        return True
                    if tr.contains(l.end):
                        tr.add(tobeSolved, resultLine.projection(tr.transform(l.end)))
                        return True
        # case 5: if start or end is between two checkpoints
        # if end is in between two checkpoints
        if length is None and not endSolved and len(checkPoints) > 1:
            for i in range(len(checkPoints) - 1):
                if checkPoints[i][0] < lineSeg.end and checkPoints[i + 1][0] > lineSeg.end:
                    checkPointDist0 = lineSeg.end.subtract(checkPoints[i][0]).length()
                    checkPointDist1 = lineSeg.end.subtract(checkPoints[i+1][0]).length()
                    ratio = checkPointDist0 / (checkPointDist0 + checkPointDist1)
                    # if checkPointDist0 == checkPointDist1:
                    #     ratio = c0p5
                    # else:
                    #     ratio = coordinate(randint(0, 100000)/100000)
                    tr.add(lineSeg.end, checkPoints[i][1].add(checkPoints[i+1][1].subtract(checkPoints[i][1]).scale(ratio)))
                    endSolved = True

        if length is None and not startSolved and len(checkPoints) > 1:
            for i in range(len(checkPoints) - 1):
                if checkPoints[i][0] < lineSeg.start and checkPoints[i + 1][0] > lineSeg.start:
                    checkPointDist0 = lineSeg.start.subtract(checkPoints[i][0]).length()
                    checkPointDist1 = lineSeg.start.subtract(checkPoints[i + 1][0]).length()
                    ratio = checkPointDist0 / (checkPointDist0 + checkPointDist1)
                    # if checkPointDist0 == checkPointDist1:
                    #     ratio = c0p5
                    # else:
                    #     ratio = coordinate(randint(0, 100000)/100000)
                    tr.add(lineSeg.start, checkPoints[i][1].add(checkPoints[i+1][1].subtract(checkPoints[i][1]).scale(ratio)))
                    startSolved = True
        if startSolved and endSolved:
            return True

        # line case 6: the start or end points are projections of some random points
        for vec in tr.vects:
            vecProj = lineSeg.projection(vec[0])
            if not startSolved and vecProj == lineSeg.start:
                tr.add(lineSeg.start, resultLine.projection(vec[1]))
                startSolved = True
            if not endSolved and vecProj == lineSeg.end:
                tr.add(lineSeg.end, resultLine.projection(vec[1]))
                endSolved = True
            if startSolved and endSolved:
                return True

        # leftover case: solve with length, same length as its pretransformed segment
        if length is None:
            length = lineSeg.length()

        # revMod = -1 if reverse, revMod = 1 if not reverse
        revMod = cn1 if reverse else c1
        # if only start has match -> find end
        if startSolved:
            matchStart = tr.transform(lineSeg.start)
            matchEnd = matchStart.add(resultLine.unitDir.scale(revMod).scale(length))
            tr.add(lineSeg.end, matchEnd, override = True)
        # if only end has match -> find start
        elif endSolved:
            matchEnd = tr.transform(lineSeg.end)
            matchStart = matchEnd.subtract(resultLine.unitDir.scale(revMod).scale(length))
            tr.add(lineSeg.start, matchStart, override = True)
        # if neither start or end has match
        else:
            containedCheckPoints = []
            minPoint = [None]
            maxPoint = [None]
            for i in checkPoints:
                if lineSeg.start < i[0] < lineSeg.end:
                    containedCheckPoints.append(i)
                elif i[0] < lineSeg.start:
                    if minPoint[0] is None or minPoint[0] < i[0]:
                        minPoint = i
                elif i[0] > lineSeg.end:
                    if maxPoint[0] is None or maxPoint[0] > i[0]:
                        maxPoint = i
            if reverse:
                length = c0 - length

            # if exactly 1 contained point -> that point is the midpoint of the new line 
            if len(containedCheckPoints) == 1:
                midPoint = containedCheckPoints[0][1]
            # if more than 1 contained points -> mid point of the first and last contained points
            elif len(containedCheckPoints) > 1:
                midPoint = containedCheckPoints[0][1].add(containedCheckPoints[-1][1]).scale(c0p5)
            # if there are both minPoint and maxPoint -> midpoint of minpoint and maxpoint
            elif minPoint[0] is not None and maxPoint[0] is not None:
                midPoint = minPoint[1].add(maxPoint[1]).scale(c0p5)
            elif len(checkPoints) > 1:
                midPoint = checkPoints[0][1].add(checkPoints[-1][1]).scale(c0p5)
            elif len(checkPoints) == 1:
                midPoint = checkPoints[0][1]
            else:
                return False
            tr.add(lineSeg.start, midPoint.subtract(resultLine.unitDir.scale(length/2)), override = True)
            tr.add(lineSeg.end, midPoint.add(resultLine.unitDir.scale(length/2)), override = True)
        return True

    def parallelProjection(self, tr, base, resultLine, lineNode):
        projS = lineNode.nodeElement.projection(base.start)
        projE = lineNode.nodeElement.projection(base.end)
        for ln in lineNode.forms:
            if projS == ln.start:
                tr.add(ln.start, resultLine.projection(tr.apply(base.start)))
            elif projS == ln.end:
                tr.add(ln.end, resultLine.projection(tr.apply(base.start)))
            if projE == ln.start:
                tr.add(ln.start, resultLine.projection(tr.apply(base.end)))
            elif projE == ln.end:
                tr.add(ln.end, resultLine.projection(tr.apply(base.end)))

    def replaceCaseAnchor(self, tr, node, anchorIndex, baseSet, trBaseSet):
        for vec in tr.vects:
            if len(baseSet) == 1:
                proj0 = baseSet[0].projection(vec[0])
                if proj0 == vec[0]:
                    continue
                trproj0 = trBaseSet[0].projection(vec[1])
                if trproj0 == vec[1]:
                    continue

                orientationAnchor = vec[0].subtract(proj0).unit()
                if 'line' in node.nodeElement.type:
                    root0 = node.nodeElement.root
                elif 'point' in node.nodeElement.type:
                    root0 = node.nodeElement.value
                else:
                    root0 = node.nodeElement.pRoot
                projRoot0 = baseSet[0].projection(root0)
                orientationNew = root0.subtract(projRoot0).unit()
                dirCheck = orientationAnchor == orientationNew

                node.rhsInfo['caseNodes'][anchorIndex] = vec[0]
                node.rhsInfo['caseNodes'][anchorIndex + 1] = dirCheck
                return vec
            elif len(baseSet) == 2:
                proj0 = baseSet[0].projection(vec[0])
                proj1 = baseSet[1].projection(vec[0])
                if proj0 == proj1:
                    continue
                trproj0 = trBaseSet[0].projection(vec[1])
                trproj1 = trBaseSet[1].projection(vec[1])
                if trproj0 == trproj1:
                    continue

                orientationAnchor = proj0.subtract(proj1).unit()

                if 'line' in node.nodeElement.type:
                    root0 = node.nodeElement.root
                elif 'point' in node.nodeElement.type:
                    root0 = node.nodeElement.value
                root1 = baseSet[1].root
                orientationNew = root0.subtract(root1).unit()
                dirCheck = orientationAnchor == orientationNew

                node.rhsInfo['caseNodes'][anchorIndex] = vec[0]
                node.rhsInfo['caseNodes'][anchorIndex + 1] = dirCheck
                return vec
            elif len(baseSet) == 3:
                pass
        return None

    # def replacePointCaseAnchor(self, tr, node, anchorIndex, baseSet, trBaseSet):
    #     for vec in tr.vects:
    #         if vec[0] == node.nodeElement.value:
    #             continue
    #         if len(baseSet) == 1:
    #             base = baseSet[0]
    #             trBase = trBaseSet[0]
    #             proj0 = base.projection(vec[0])
    #             if proj0 == vec[0]:
    #                 continue
    #             trproj0 = trBase.projection(vec[1])
    #             if trproj0 == vec[1]:
    #                 continue
    #             orientationAnchor = vec[0].subtract(proj0).unit()
    #             root0 = node.nodeElement.value
    #             projRoot0 = base.projection(root0)
    #             orientationNew = root0.subtract(projRoot0).unit()
    #             dirCheck = orientationAnchor == orientationNew

    #             node.rhsInfo['caseNodes'][anchorIndex] = vec[0]
    #             node.rhsInfo['caseNodes'][anchorIndex + 1] = dirCheck
    #         elif len(baseSet) == 2:
    #             proj0 = baseSet[0].projection(vec[0])
    #             proj1 = baseSet[1].projection(vec[0])
    #             if proj0 == proj1:
    #                 continue
    #             trproj0 = trBaseSet[0].projection(vec[1])
    #             trproj1 = trBaseSet[1].projection(vec[1])
    #             if trproj0 == trproj1:
    #                 continue

    #             orientationAnchor = proj0.subtract(proj1).unit()

    #             root0 = node.nodeElement.root
    #             root1 = baseSet[1].root
    #             orientationNew = root0.subtract(root1).unit()
    #             dirCheck = orientationAnchor == orientationNew

    #             node.rhsInfo['caseNodes'][anchorIndex] = vec[0]
    #             node.rhsInfo['caseNodes'][anchorIndex + 1] = dirCheck

    #         return vec
    #     return None


    def getTransform(self, tfs):
        """
        get the graphTransform from the transformation set
        """
        string = self.id.split("=|=")
        if len(string) > 1:
            tfs.current = string[0]
        return tfs.retrieveTransform(graphTransform3D)

    def visitEnter(self, obj, direction = 0):
        """
        visitEnter for each type of predicate/point_on_line
        """
        if direction[0] == 'maxline':
            if (obj.__class__ == direction[1].__class__ 
            and obj.sort == direction[1].sort 
            and (not obj.equals(direction[1]))
            and (obj.compareValue(direction[1]) == 0)
            and ((obj.containsPoint(direction[1].start) and obj.start < direction[1].start)
            or (obj.containsPoint(direction[1].end) and obj.end > direction[1].end))):
                self.postCheck = False
                return False
        elif direction[0] == 'bound':
            if (obj.__class__ == direction[1].__class__ 
            and obj.sort == direction[1].sort 
            and obj.contains(direction[1])):
                if 0 in direction[2] and obj.start != direction[3] and obj.end != direction[3]:
                    self.postCheck = False
                    return False
                if 1 in direction[2] and obj.start != direction[4] and obj.end != direction[4]:
                    self.postCheck = False
                    return False
                    
        elif direction[0] == 'void':
            if 'ind-planeSegment' in obj.type :
                if ('3D' in obj.type) == ('2D' in direction[1].type):
                    raise ValueError('There exists a mix of 2D and 3D objects in the shape')
                if not plane3D.compareValue(obj,direction[1]) == 0:
                    return True
                check = obj.classify(direction[1])
                if check.disjoint() or check.touches():
                    return True
                self.postCheck = False
                return False
            if 'ind-lineSegment' in obj.type:
                if not plane3D.containsLine(direction[1], obj):
                    return True
                direction[2].append(bSegment(obj.start, obj.end))
            if 'ind-point' in obj.type:
                if not plane3D.containsPoint(direction[1], obj.value):
                    return True
                check = False
                for line in direction[1].boundaries[0].canonical:
                    if line.containsPoint(obj.value):
                        check = True
                if check:
                    return True
                if direction[1].boundaries[0].containsPointInside(obj.value):
                    self.postCheck = False
                    return False
        elif direction[0] == 'voidof':
            if 'ind-planeSegment' in obj.type :
                if not (obj.sort.name in direction[-1] or obj.sort.baseSort().name in direction[-1]):
                    return True
                if ('3D' in obj.type) == ('2D' in direction[1].type):
                    raise ValueError('There exists a mix of 2D and 3D objects in the shape')
                if not plane3D.compareValue(obj,direction[1]) == 0:
                    return True
                check = obj.classify(direction[1])
                if check.disjoint() or check.touches():
                    return True
                self.postCheck = False
                return False
            if 'ind-lineSegment' in obj.type:
                if not (obj.sort.name in direction[-1] or obj.sort.baseSort().name in direction[-1]):
                    return True
                if not plane3D.containsLine(direction[1], obj):
                    return True
                direction[2].append(bSegment(obj.start, obj.end))
            if 'ind-point' in obj.type:
                if not (obj.sort.name in direction[-1] or obj.sort.baseSort().name in direction[-1]):
                    return True
                if not plane3D.containsPoint(direction[1], obj.value):
                    return True
                check = False
                for line in direction[1].boundaries[0].canonical:
                    if line.containsPoint(obj.value):
                        check = True
                if check:
                    return True
                if direction[1].boundaries[0].containsPointInside(obj.value):
                    self.postCheck = False
                    return False
        elif direction[0] == 'inside':
            if 'ind-planeSegment' in obj.type :
                if ('3D' in obj.type) == ('2D' in direction[1].type):
                    raise ValueError('There exists a mix of 2D and 3D objects in the shape')
                if not plane3D.compareValue(obj,direction[1]) == 0:
                    self.postCheck = False
                    return False
                check = obj.classify(direction[1])
                if check.isContained():
                    return True
                self.postCheck = False
                return False
            if 'ind-lineSegment' in obj.type:
                if not plane3D.containsLine(direction[1], obj):
                    self.postCheck = False
                    return False
                direction[2].append(bSegment(obj.start, obj.end))
            if 'ind-point' in obj.type:
                if not plane3D.containsPoint(direction[1], obj.value):
                    self.postCheck = False
                    return False
                check = False
                for line in direction[1].boundaries[0].canonical:
                    if line.containsPoint(obj.value):
                        check = True
                if check:
                    return True
                if direction[1].boundaries[0].containsPointInside(obj.value):
                    return True
                self.postCheck = False
                return False
        elif direction[0] == 'insideof':
            if 'ind-planeSegment' in obj.type :
                if not (obj.sort.name in direction[-1] or obj.sort.baseSort().name in direction[-1]):
                    return True
                if ('3D' in obj.type) == ('2D' in direction[1].type):
                    raise ValueError('There exists a mix of 2D and 3D objects in the shape')
                if not plane3D.compareValue(obj,direction[1]) == 0:
                    self.postCheck = False
                    return False
                check = obj.classify(direction[1])
                if check.isContained():
                    return True
                self.postCheck = False
                return False
            if 'ind-lineSegment' in obj.type:
                if not (obj.sort.name in direction[-1] or obj.sort.baseSort().name in direction[-1]):
                    return True
                if not plane3D.containsLine(direction[1], obj):
                    self.postCheck = False
                    return False
                direction[2].append(bSegment(obj.start, obj.end))
            if 'ind-point' in obj.type:
                if not (obj.sort.name in direction[-1] or obj.sort.baseSort().name in direction[-1]):
                    return True
                if not plane3D.containsPoint(direction[1], obj.value):
                    self.postCheck = False
                    return False
                check = False
                for line in direction[1].boundaries[0].canonical:
                    if line.containsPoint(obj.value):
                        check = True
                if check:
                    return True
                if direction[1].boundaries[0].containsPointInside(obj.value):
                    return True
                self.postCheck = False
                return False
        elif direction[0] == 'outside':
            if 'ind-planeSegment' in obj.type :
                if ('3D' in obj.type) == ('2D' in direction[1].type):
                    raise ValueError('There exists a mix of 2D and 3D objects in the shape')
                if not plane3D.compareValue(obj,direction[1]) == 0:
                    return True
                check = obj.classify(direction[1])
                if check.disjoint() or check.touches():
                    return True
                self.postCheck = False
                return False
            if 'ind-lineSegment' in obj.type:
                if not plane3D.containsLine(direction[1], obj):
                    return True
                direction[2].append(bSegment(obj.start, obj.end))
            if 'ind-point' in obj.type:
                if not plane3D.containsPoint(direction[1], obj.value):
                    return True
                check = False
                for line in direction[1].boundaries[0].canonical:
                    if line.containsPoint(obj.value):
                        check = True
                if check:
                    return True
                if direction[1].boundaries[0].containsPointInside(obj.value):
                    self.postCheck = False
                    return False
        elif direction[0] == 'outsideof':
            if 'ind-planeSegment' in obj.type :
                if not (obj.sort.name in direction[-1] or obj.sort.baseSort().name in direction[-1]):
                    return True
                if ('3D' in obj.type) == ('2D' in direction[1].type):
                    raise ValueError('There exists a mix of 2D and 3D objects in the shape')
                if not plane3D.compareValue(obj,direction[1]) == 0:
                    return True
                check = obj.classify(direction[1])
                if check.disjoint() or check.touches():
                    return True
                self.postCheck = False
                return False
            if 'ind-lineSegment' in obj.type:
                if not (obj.sort.name in direction[-1] or obj.sort.baseSort().name in direction[-1]):
                    return True
                if not plane3D.containsLine(direction[1], obj):
                    return True
                direction[2].append(bSegment(obj.start, obj.end))
            if 'ind-point' in obj.type:
                if not (obj.sort.name in direction[-1] or obj.sort.baseSort().name in direction[-1]):
                    return True
                if not plane3D.containsPoint(direction[1], obj.value):
                    return True
                check = False
                for line in direction[1].boundaries[0].canonical:
                    if line.containsPoint(obj.value):
                        check = True
                if check:
                    return True
                if direction[1].boundaries[0].containsPointInside(obj.value):
                    self.postCheck = False
                    return False
        elif direction[0] == 'shortest_line':
            if len(direction) == 3:
                if 'ind-lineSegment' not in obj.type:
                    return True
                if (obj.sort.name in direction[-1] or
                obj.sort.baseSort().name in direction[-1]) and obj.squareLength() < direction[1]:
                    self.postCheck = False
                    return False
            else:
                if 'ind-lineSegment' in obj.type and obj.squareLength() < direction[1]:
                    self.postCheck = False
                    return False
        elif direction[0] == 'longest_line':
            if len(direction) == 3:
                if 'ind-lineSegment' not in obj.type:
                    return True
                if (obj.sort.name in direction[-1] or
                obj.sort.baseSort().name in direction[-1]) and obj.squareLength() > direction[1]:
                    self.postCheck = False
                    return False
            else:
                if 'ind-lineSegment' in obj.type and obj.squareLength() > direction[1]:
                    self.postCheck = False
                    return False
        elif direction[0] == 'nolabel':
            if 'ind-lineSegment' in obj.type and obj.type == direction[1].type and obj.contains(direction[1]):
                if obj.hasAttribute():
                    obj.attribute.accept(self, direction = ['nolabel*'])
            if 'ind-planeSegment' in obj.type and obj.type == direction[1].type and obj.contains(direction[1]):
                if obj.hasAttribute():
                    obj.attribute.accept(self, direction = ['nolabel*'])
            if 'ind-point' in obj.type and obj.type == direction[1].type and obj == direction[1]:
                if obj.hasAttribute():
                    obj.attribute.accept(self, direction = ['nolabel*'])
        elif direction[0] == 'nolabel*':
            if obj.type == 'ind-label' or obj.type == 'ind-description':
                self.postCheck = False
                return False
        elif direction[0] == 'containsCheck':
            if 'ind-lineSegment' in obj.type:
                if len(direction) > 3 and (direction[3] is not None) and (obj == direction[3]):
                    return False
                elif obj.start == direction[1] or obj.end == direction[1]:
                    direction[2][0] = True
            if 'ind-point' in obj.type:
                if obj.value == direction[1]:
                    direction[2][0] = True
        return True
        
    def visitLeave(self, obj, direction = 0):
        return
        
        
    @staticmethod
    def cw(pl,A0,B0,C0):
        """
        Check if ABC is a counter-clockwise angle on plane pl
        """
        A = pl.localize(A0)
        B = pl.localize(B0)
        C = pl.localize(C0)
        return (C.y-A.y) * (B.x-A.x) > (B.y-A.y) * (C.x-A.x)
        
    @staticmethod
    def evaluateDesc(boundList, param, special = False):
        """
        Evaluate all the descriptions in a predicate/directive 
        """
        newList = []
        check = True
        for i in boundList:
            if isinstance(i, description):
                try:
                    val = i.evaluateInd(param).getValue()[1:-1]
                except Exception:
                    check = False
                    break
                if special:
                    try:
                        newList.append(coordinate(val))
                    except:
                        newList.append(val)
                else:
                    newList.append(coordinate(val))
            elif isinstance(i,int) or isinstance(i,float):
                newList.append(coordinate(i))
            else:
                newList.append(i)
        return newList, check

    @staticmethod
    def checkNewDir(oldDir, newDir, oldAnchorDir, newAnchorDir):
        oldDir = oldDir.unit()
        newDir = newDir.unit()

        oldAnchorDir = oldAnchorDir.unit()
        newAnchorDir = newAnchorDir.unit()

        cDirCheck = oldDir.add(oldAnchorDir).squareLength()
        nDirCheck = newDir.add(newAnchorDir).squareLength()
        
        if cDirCheck != nDirCheck:
            return newDir.scale(-1)
        return newDir

    
    @staticmethod
    def evaluateBounds(bounds, tfSet):
        if isinstance(bounds[-1], list):
            raise NotImplementedError('evaluate bounds - obsolete syntax')
        return bounds[-1]
        
    @staticmethod
    def solvePointOnLine(graphTransf, obj, bounds, target = None):
        """
        solving point_on_line directive
        """
        i = 0
        if target is not None:
            startP = obj.oStart
            endP = obj.oEnd
            if obj.containsPoint(target) and (not startP == target) and (not endP == target):
                if target.subtract(endP).length() < target.subtract(startP).length():
                    startP, endP = endP, startP
                nStart = graphTransf.apply(startP)
                nEnd = graphTransf.apply(endP)
                p = nStart.add(nEnd.subtract(nStart).scale(bounds[-1]))
                graphTransf.add(target, p)
            return
        while i < len(graphTransf.unresolved):
            u = graphTransf.unresolved[i]
            if obj.containsPoint(u) and (not obj.start == u) and (not obj.end == u):
                startP = obj.oStart
                endP = obj.oEnd
                if u.subtract(endP).length() < u.subtract(startP).length():
                    startP, endP = endP, startP
                # if adaptive:
                #     for j in graphTransf.vects:
                #         if obj.containsPoint(j[0]):
                #             if j[0] < u:
                #                 if j[0] > startP:
                #                     startP = j[0]
                #             else:
                #                 if j[0] < endP:
                #                     endP = j[0]
                graphTransf.unresolved.pop(i)
                nStart = graphTransf.apply(startP)
                nEnd = graphTransf.apply(endP)
                p = nStart.add(nEnd.subtract(nStart).scale(bounds[-1]))
                # if singleRatio:
                #     p = nStart.add(nEnd.subtract(nStart).scale(bounds[-1]))
                # else:
                #     rand = randint(0, 100000)
                #     ratio = (bounds[-1]-bounds[-2])*rand/100000 + bounds[-2]
                #     p = nStart.add(nEnd.subtract(nStart).scale(ratio))
                graphTransf.add(u, p)
            else:
                i += 1
